Package: resemble
Type: Package
Title: Similarity Retrieval and Local Learning for Spectral
        Chemometrics
Version: 3.0.0
Date: 2026-04-15
Authors@R: c(
  person(given = "Leonardo",
         family = "Ramirez-Lopez",
         role = c("aut", "cre"),
         email = "ramirez.lopez.leo@gmail.com",
         comment = c(ORCID = "0000-0002-5369-5120")),
  person(given = "Antoine",
         family = "Stevens",
         role = c("aut", "ctb"),
         comment = c(ORCID = "0000-0002-1588-7519")),
  person(given = "Claudio",
         family = "Orellano",
         role = c("ctb"))
  )
BugReports: https://github.com/l-ramirez-lopez/resemble/issues
Description: 
    Functions for dissimilarity analysis and machine learning in complex
    spectral data sets, including memory-based learning (MBL), optimal subset
    search and selection, and retrieval-based modelling with model libraries.
    Supports local learning, optimisation of spectral libraries, and
    ensemble prediction from precomputed models.
    Most of these functions are based on the methods presented in 
    Ramirez-Lopez et al. (2013) <doi:10.1016/j.geoderma.2012.12.014>.
License: MIT + file LICENSE
URL: https://l-ramirez-lopez.github.io/resemble/,
        https://github.com/l-ramirez-lopez/resemble
Depends: R (>= 4.2.0)
Imports: foreach, iterators, RhpcBLASctl, Rcpp (>= 1.0.3), mathjaxr (>=
        1.0), lifecycle (>= 0.2.0)
Suggests: prospectr, parallel, doParallel, testthat, quarto, knitr
LinkingTo: Rcpp, RcppArmadillo
RdMacros: mathjaxr
VignetteBuilder: quarto
NeedsCompilation: yes
Repository: CRAN
RoxygenNote: 7.3.3
Encoding: UTF-8
Config/VersionName: vortex
Packaged: 2026-04-20 18:50:53 UTC; leo
Author: Leonardo Ramirez-Lopez [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5369-5120>),
  Antoine Stevens [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-1588-7519>),
  Claudio Orellano [ctb]
Maintainer: Leonardo Ramirez-Lopez <ramirez.lopez.leo@gmail.com>
Date/Publication: 2026-04-20 20:40:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-26 01:44:36 UTC; unix
Archs: resemble.so.dSYM

Package: resemble
Type: Package
Title: Similarity Retrieval and Local Learning for Spectral
        Chemometrics
Version: 3.0.1
Date: 2026-09-01
Authors@R: c(
  person(given = "Leonardo",
         family = "Ramirez-Lopez",
         role = c("aut", "cre"),
         email = "ramirez.lopez.leo@gmail.com",
         comment = c(ORCID = "0000-0002-5369-5120")),
  person(given = "Antoine",
         family = "Stevens",
         role = c("aut", "ctb"),
         comment = c(ORCID = "0000-0002-1588-7519")),
  person(given = "Claudio",
         family = "Orellano",
         role = c("ctb"))
  )
BugReports: https://github.com/l-ramirez-lopez/resemble/issues
Description: 
    Functions for dissimilarity analysis and machine learning in complex
    spectral data sets, including memory-based learning (MBL), optimal subset
    search and selection, and retrieval-based modelling with model libraries.
    Supports local learning, optimisation of spectral libraries, and
    ensemble prediction from precomputed models. Most of these functions are
    based on the methods presented in Ramirez-Lopez et al. (2013)
    <doi:10.1016/j.geoderma.2012.12.014>, Ramirez-Lopez et al. (2026a)
    <doi:10.1016/j.aca.2026.345682>, and Ramirez-Lopez et al. (2026b)
    <doi:10.1016/j.aca.2026.345651>.
License: MIT + file LICENSE
URL: https://l-ramirez-lopez.github.io/resemble/,
        https://github.com/l-ramirez-lopez/resemble
Depends: R (>= 4.2.0)
Imports: foreach, iterators, RhpcBLASctl, Rcpp (>= 1.0.3), mathjaxr (>=
        1.0), lifecycle (>= 0.2.0)
Suggests: prospectr, parallel, doParallel, testthat, quarto, knitr
LinkingTo: Rcpp, RcppArmadillo
RdMacros: mathjaxr
VignetteBuilder: quarto
NeedsCompilation: yes
Repository: CRAN
Encoding: UTF-8
Config/VersionName: tarragona
Config/roxygen2/version: 8.0.0
Packaged: 2026-08-30 20:59:21 UTC; leo
Author: Leonardo Ramirez-Lopez [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-5369-5120>),
  Antoine Stevens [aut, ctb] (ORCID:
    <https://orcid.org/0000-0002-1588-7519>),
  Claudio Orellano [ctb]
Maintainer: Leonardo Ramirez-Lopez <ramirez.lopez.leo@gmail.com>
Date/Publication: 2026-08-31 16:30:02 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-08-31 19:02:38 UTC; unix
Archs: resemble.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(gesearch,default)
S3method(gesearch,formula)
S3method(plot,gesearch)
S3method(plot,liblex)
S3method(plot,mbl)
S3method(plot,ortho_projection)
S3method(predict,gesearch)
S3method(predict,liblex)
S3method(predict,ortho_projection)
S3method(predict,plsr)
S3method(predict,resemble_model)
S3method(print,diss_correlation)
S3method(print,diss_method)
S3method(print,diss_pca)
S3method(print,diss_pls)
S3method(print,fit_gpr)
S3method(print,fit_method)
S3method(print,fit_pls)
S3method(print,fit_wapls)
S3method(print,gesearch)
S3method(print,liblex)
S3method(print,mbl)
S3method(print,model_control)
S3method(print,ncomp_selection)
S3method(print,neighbors)
S3method(print,neighbors_diss)
S3method(print,neighbors_k)
S3method(print,ortho_projection)
S3method(print,resemble_dissimilarity)
S3method(print,resemble_model)
export(diss_correlation)
export(diss_cosine)
export(diss_euclidean)
export(diss_evaluate)
export(diss_mahalanobis)
export(diss_pca)
export(diss_pls)
export(dissimilarity)
export(fit_gpr)
export(fit_pls)
export(fit_wapls)
export(gesearch)
export(gesearch_control)
export(get_predictions)
export(liblex)
export(liblex_control)
export(mbl)
export(mbl_control)
export(model)
export(model_control)
export(ncomp_by_cumvar)
export(ncomp_by_opc)
export(ncomp_by_var)
export(ncomp_fixed)
export(neighbors_diss)
export(neighbors_k)
export(ortho_projection)
export(search_neighbors)
export(sid)
export(sim_eval)
import(Rcpp)
import(foreach)
import(grDevices)
import(graphics)
import(iterators)
import(lifecycle)
import(mathjaxr)
importFrom(RhpcBLASctl,blas_get_num_procs)
importFrom(RhpcBLASctl,blas_set_num_threads)
importFrom(foreach,"%dopar%")
importFrom(foreach,foreach)
importFrom(graphics,barplot)
importFrom(graphics,grid)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(graphics,segments)
importFrom(lifecycle,deprecate_soft)
importFrom(mathjaxr,preview_rd)
importFrom(stats,.MFclass)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,bw.nrd0)
importFrom(stats,complete.cases)
importFrom(stats,cor)
importFrom(stats,cov)
importFrom(stats,delete.response)
importFrom(stats,density)
importFrom(stats,diffinv)
importFrom(stats,median)
importFrom(stats,model.extract)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,na.fail)
importFrom(stats,na.omit)
importFrom(stats,na.pass)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,reshape)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,terms)
importFrom(stats,update)
importFrom(utils,flush.console)
importFrom(utils,setTxtProgressBar)
importFrom(utils,tail)
importFrom(utils,txtProgressBar)
useDynLib(resemble, .registration=TRUE)

# `resemble` Similarity Retrieval and Local Learning for Spectral Chemometrics

<!-- badges: start -->

![R-CMD-check](https://github.com/l-ramirez-lopez/resemble/actions/workflows/R-CMD-check.yaml/badge.svg)
[![codecov](https://codecov.io/gh/l-ramirez-lopez/resemble/badge.svg)](https://app.codecov.io/gh/l-ramirez-lopez/resemble)
[![CRAN
status](https://www.r-pkg.org/badges/version/resemble?v=2.png)](https://CRAN.R-project.org/package=resemble)
[![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/resemble?v=2.png)](https://CRAN.R-project.org/package=resemble)

<img align="right" src="./man/figures/logo.png" width="15%">

<!-- badges: end -->

*Last update: 2026-04-20*

Version: 3.0.0 – vortex

<em>
<p align="right">

Think Globally, Fit Locally (Saul and Roweis, 2003)
</p>

</em>

## About

The `resemble` package provides computationally efficient methods for
dissimilarity analysis and predictive modelling with complex spectral
data. Its core functionality includes memory-based learning (MBL),
evolutionary subset search and selection, and retrieval-based modelling
using pre-computed model libraries. The package is designed to support
local modelling, spectral library optimization, and model-based
prediction in large and heterogeneous spectral data sets.

## Documentation

The package includes comprehensive vignettes covering all major
functionality:

1.  **Essential concepts and setup**: Introduction, data preparation,
    and notation
2.  **Dimensionality reduction**: PCA and PLS projections with
    `ortho_projection()`
3.  **Estimating dissimilarity between spectra**: Dissimilarity methods
    and evaluation
4.  **Nearest neighbor search**: Finding similar spectra with
    `search_neighbors()`
5.  **Simple global models**: Global calibration with `model()`
6.  **Classical memory-based learning**: Per-query local modelling with
    `mbl()`
7.  **Evolutionary subset search**: Domain-adaptive calibration with
    `gesearch()`
8.  **Building a library of models**: Pre-computed experts with
    `liblex()`

## What’s new in version 3.0

Version 3.0 is a major release with a redesigned API, new modelling
functions, and improved computational efficiency.

**New modelling functions:**

- `liblex()`: Builds a library of reusable localized models (experts)
  that can be stored and reused for prediction without refitting. Based
  on Ramirez-Lopez et al. (2026b).

- `gesearch()`: Evolutionary algorithm for selecting optimal subsets
  from large spectral libraries to build context-specific calibrations.
  Based on Ramirez-Lopez et al. (2026a).

- `model()`: Fits global PLS or GPR calibration models with
  cross-validation.

**Redesigned dissimilarity interface:**

The dissimilarity system now uses constructor functions:

- `diss_pca()`, `diss_pls()`: Mahalanobis distance in projection space
- `diss_correlation()`: Correlation-based dissimilarity (including
  moving window)
- `diss_euclidean()`, `diss_mahalanobis()`, `diss_cosine()`: Distance
  metrics

Component selection via `ncomp_by_var()`, `ncomp_by_cumvar()`,
`ncomp_by_opc()`, or `ncomp_fixed()`.

**Redesigned neighbor and fitting interfaces:**

- `neighbors_k()`, `neighbors_diss()`: Neighbor selection constructors
- `fit_pls()`, `fit_wapls()`, `fit_gpr()`: Local fitting constructors
  (replace `local_fit_*()` functions)

**Breaking changes in `mbl()`:**

- `k`, `k_diss`, `k_range` replaced by `neighbors` argument
- `method` renamed to `fit_method`
- `center` and `scale` removed; now controlled within constructors

See `NEWS.md` for full details on deprecated and removed functions.

## Core functionality

**Dimensionality reduction:**

- `ortho_projection()`: PCA or PLS projection with multiple algorithms
  (SVD, NIPALS, SIMPLS)

**Computing dissimilarity matrices:**

- `dissimilarity()`: Main interface for dissimilarity computation
- `diss_pca()`, `diss_pls()`, `diss_correlation()`, `diss_euclidean()`,
  `diss_mahalanobis()`, `diss_cosine()`: Method constructors
- `diss_evaluate()`: Evaluate dissimilarity matrices using side
  information

**Neighbor search:**

- `search_neighbors()`: Efficient k-nearest neighbor retrieval

**Modelling spectral data:**

- `model()`: Global PLS or GPR calibration
- `mbl()`: Memory-based learning for per-query local modelling
- `gesearch()`: Evolutionary subset selection for domain-adaptive
  calibration
- `liblex()`: Pre-computed library of local experts for fast prediction

## Installation

Install from CRAN:

``` r
install.packages("resemble")
```

Or install the development version from GitHub:

``` r
# install.packages("devtools")
devtools::install_github("l-ramirez-lopez/resemble")
```

The package requires a C++ compiler. On Windows, install
[Rtools](https://cran.r-project.org/bin/windows/Rtools/). On macOS, you
may need to install `gfortran` and `clang` from [CRAN
tools](https://cran.r-project.org/bin/macosx/tools/).

## Example: Memory-based learning with `mbl()`

``` r
library(resemble)
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
 detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
 m = 1, p = 1, w = 7
)

# Split into training and test sets
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]
test_x <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]
test_y <- NIRsoil$CEC[NIRsoil$train == 0 & !is.na(NIRsoil$CEC)]

# Memory-based learning with Gaussian process regression
sbl <- mbl(
 Xr = train_x,
 Yr = train_y,
 Xu = test_x,
 neighbors = neighbors_k(seq(50, 130, by = 20)),
 diss_method = diss_pca(ncomp = ncomp_by_opc(40)),
 fit_method = fit_gpr(),
 control = mbl_control(validation_type = "NNv")
)
sbl
plot(sbl)
get_predictions(sbl)
```

<p align="center">

<img src="./man/figures/mbl.png" width="80%">
</p>

## Example: Pre-computed model library with `liblex()`

`liblex()` builds a library of local experts that can be reused for
prediction without refitting:

``` r
# Build model library
model_lib <- liblex(
 Xr = train_x,
 Yr = train_y,
 neighbors = neighbors_k(c(40, 60, 80)),
 diss_method = diss_correlation(ws = 27, scale = TRUE),
 fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "mpls"),
 control = liblex_control(tune = TRUE)
)

# Predict new observations
predictions <- predict(model_lib, test_x)
```

## Example: Evolutionary subset selection with `gesearch()`

`gesearch()` selects optimal subsets from large spectral libraries:

``` r
# Search for optimal calibration subset
gs <- gesearch(
 Xr = train_x, 
 Yr = train_y,
 Xu = test_x,
 k = 50, 
 b = 100, 
 retain = 0.97,
 target_size = 200,
 fit_method = fit_pls(ncomp = 15, method = "mpls"),
 optimization = c("reconstruction", "similarity"),
 seed = 42
)

# Predict using selected subset
preds <- predict(gs, test_x)
plot(gs)
```

## Memory-based learning overview

Memory-based learning (MBL, a.k.a. instance-based learning or local
modelling) is a non-linear lazy learning approach. For each prediction,
the algorithm:

1.  Finds the k-nearest neighbors in the reference set
2.  Fits a local model using those neighbors
3.  Predicts the response for the target observation

The `mbl()` function offers three regression methods for local models:

- `fit_gpr()`: Gaussian process regression with linear kernel
- `fit_pls()`: Partial least squares
- `fit_wapls()`: Weighted average PLS (Shenk et al., 1997)

<p align="center">

<img src="./vignettes/MBL.gif" width="50%">
</p>

## Citing the package

``` r
citation(package = "resemble")
```

## News: Memory-based learning and `resemble`

- **2026.05:** [van Leeuwen et al.,
  2026](https://doi.org/10.1016/j.geoderma.2026.117804)  
  used `resemble` for principal component Mahalanobis nearest-neighbour
  search to extract spectrally similar samples from the KSSL library for
  MIR model calibration in Dutch soils.

- **2026.04:** [Irving et al.,
  2026](https://doi.org/10.1016/j.seh.2026.100205)  
  used `resemble` in modelling workflows for infrared spectroscopy
  prediction of soil microbial properties across Australian soils.

- **2026.03:** [Shrestha et al.,
  2026](https://doi.org/10.1016/j.geodrs.2026.e01063)  
  used `resemble` in a hybrid localisation workflow to predict
  farm-scale soil cadmium from a regional spectral library; LOCAL models
  with MIR data performed best.

- **2025.10:** [Summerauer et al.,
  2025](https://doi.org/10.5194/egusphere-2025-4625) used `resemble` for
  MBL modelling of soil properties from infrared spectra across tropical
  hillslopes in Eastern Africa.

- **2025.05:** [Sun and Shi,
  2025](https://doi.org/10.1016/j.geoderma.2025.117298) combined
  spectral and geographical similarity for SOC prediction; local PLSR
  outperformed global models.

- **2025.03:** [Breure et al.,
  2025](https://doi.org/10.1038/s41467-025-57355-y)  
  used `resemble` for local VNIR modelling of soil carbon fractions (POC
  and MAOC) across European agricultural soils, published in *Nature
  Communications*.

- **2025.03:** [Purushothaman et al.,
  2025](https://doi.org/10.1109/TGRS.2025.3569059) applied MBL to
  AVIRIS-NG hyperspectral data for soil property prediction in India.

- **2025.01:** [Dai et al.,
  2025](https://doi.org/10.1016/j.still.2024.106297) used MBL for POC
  and MAOC prediction from VNIR in Guangdong.

- **2024.12:** [Asrat et al.,
  2024](https://doi.org/10.1016/j.geoderma.2024.117116) MBL for local
  calibration sample selection in the Moroccan Soil Spectral Library.

- **2024.09:** [Barbetti et al.,
  2024](https://doi.org/10.1109/TAFE.2024.3449215) MBL to detect SOC
  changes in long-term experiments using vis–NIR.

- **2023.11:** [Wang et al.,
  2023](https://doi.org/10.1016/j.geoderma.2023.116752) N-MBL (MBL + RF
  within local fitting) for regional vis–NIR models.

- **2022:** [Sanderman et al., 2022](https://doi.org/10.1002/saj2.20513)
  evaluated transferability of large MIR spectral databases across
  instruments.

- **2022.01:** [Ng et al.,
  2022](https://doi.org/10.1016/j.geoderma.2021.115501) showed that MBL
  yields better local SOC predictions than spiking approaches.

- **2021.10:** [Ramirez-Lopez et al.,
  2021](https://soil.copernicus.org/articles/7/693/2021/) MBL to predict
  soil properties in Africa.

- **2020.08:** Charlotte Rivard’s MIR MBL tutorial:
  <https://whrc.github.io/Soil-Predictions-MIR/>

- **2020.01:** [Sanderman et al.,
  2020](https://doi.org/10.1002/saj2.20009) MIR spectroscopy for
  prediction of soil health indicators; MBL and Cubist excelled.

- **2019.03:** [Ramirez-Lopez et al.,
  2019](https://doi.org/10.1111/ejss.12752) MBL in digital soil mapping
  at farm scale.

- **2019.03:** [Jaconi et al.,
  2019](https://doi.org/10.1016/j.geoderma.2018.11.042) MBL for
  national-scale NIR texture predictions in Germany.

- **2018.01:** [Dotto et al.,
  2018](https://doi.org/10.1016/j.geoderma.2017.11.028) MBL for SOC
  prediction in Brazil.

- **2016.04:** [Viscarra Rossel et al.,
  2016](https://doi.org/10.3390/rs8040341) memory-based learning for
  soil property prediction.

- **2014.03:** First CRAN release of `resemble`.

## Related packages

- [`prospectr`](https://github.com/l-ramirez-lopez/prospectr): Signal
  processing and chemometrics for spectroscopy

## Contributing

Contributions are welcome! Please read our Contributing Guidelines
(available in the GitHub repo) before submitting pull requests.

This project follows a Code of Conduct available in the GitHub repo.

## Bug reports

Report issues at
[GitHub](https://github.com/l-ramirez-lopez/resemble/issues) or contact
the maintainer (<ramirez.lopez.leo@gmail.com>).

## References

Lobsey, C. R., Viscarra Rossel, R. A., Roudier, P., & Hedley, C. B.
2017. rs-local data-mines information from spectral libraries to improve
local calibrations. European Journal of Soil Science, 68(6), 840-852.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte,
J.A.M., Scholten, T. 2013. The spectrum-based learner: A new local
approach for modeling soil vis-NIR spectra of complex data sets.
Geoderma 195-196, 268-279.

Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
libraries are too complex to search: Evolutionary subset selection for
domain-adaptive calibration. Analytica Chimica Acta, under review.

Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C., Perez-Fernandez,
E., Plans, M., Breure, T., Behrens, T., Viscarra Rossel, R., & Peng, Y.
(2026b). Rethinking local spectral modelling: From per-query refitting
to model libraries. Analytica Chimica Acta, under review.

Saul, L. K., & Roweis, S. T. 2003. Think globally, fit locally:
unsupervised learning of low dimensional manifolds. Journal of machine
learning research, 4(Jun), 119-155.

Shenk, J., Westerhaus, M., and Berzaghi, P. 1997. Investigation of a
LOCAL calibration procedure for near infrared instruments. Journal of
Near Infrared Spectroscopy, 5, 223-232.

# `resemble` Similarity Retrieval and Local Learning for Spectral Chemometrics

<!-- badges: start -->

![R-CMD-check](https://github.com/l-ramirez-lopez/resemble/actions/workflows/R-CMD-check.yaml/badge.svg)
[![codecov](https://codecov.io/gh/l-ramirez-lopez/resemble/badge.svg)](https://app.codecov.io/gh/l-ramirez-lopez/resemble)
[![CRAN
status](https://www.r-pkg.org/badges/version/resemble?v=2.png)](https://CRAN.R-project.org/package=resemble)
[![Downloads](https://cranlogs.r-pkg.org/badges/grand-total/resemble?v=2.png)](https://CRAN.R-project.org/package=resemble)

\<img6.5src=“./man/figures/logo.png” width=“15%”\>

<!-- badges: end -->

*Last update: 2026-08-30*

Version: 3.0.1 – tarragona

<em>
<p align="right">

Think Globally, Fit Locally (Saul and Roweis, 2003)
</p>

</em>

## About

The `resemble` package provides computationally efficient methods for
dissimilarity analysis and predictive modelling with complex spectral
data. Its core functionality includes memory-based learning (MBL),
evolutionary subset search and selection, and retrieval-based modelling
using pre-computed model libraries. The package is designed to support
local modelling, spectral library optimization, and model-based
prediction in large and heterogeneous spectral data sets.

## Documentation

The package includes comprehensive vignettes covering all major
functionality:

1.  **Essential concepts and setup**: Introduction, data preparation,
    and notation
2.  **Dimensionality reduction**: PCA and PLS projections with
    `ortho_projection()`
3.  **Estimating dissimilarity between spectra**: Dissimilarity methods
    and evaluation
4.  **Nearest neighbor search**: Finding similar spectra with
    `search_neighbors()`
5.  **Simple global models**: Global calibration with `model()`
6.  **Classical memory-based learning**: Per-query local modelling with
    `mbl()`
7.  **Evolutionary subset search**: Domain-adaptive calibration with
    `gesearch()`
8.  **Building a library of models**: Pre-computed experts with
    `liblex()`

## What’s new in version 3.0

Version 3.0 is a major release with a redesigned API, new modelling
functions, and improved computational efficiency.

**New modelling functions:**

- `liblex()`: Builds a library of reusable localized models (experts)
  that can be stored and reused for prediction without refitting. Based
  on Ramirez-Lopez et al. (2026b).

- `gesearch()`: Evolutionary algorithm for selecting optimal subsets
  from large spectral libraries to build context-specific calibrations.
  Based on Ramirez-Lopez et al. (2026a).

- `model()`: Fits global PLS or GPR calibration models with
  cross-validation.

**Redesigned dissimilarity interface:**

The dissimilarity system now uses constructor functions:

- `diss_pca()`, `diss_pls()`: Mahalanobis distance in projection space
- `diss_correlation()`: Correlation-based dissimilarity (including
  moving window)
- `diss_euclidean()`, `diss_mahalanobis()`, `diss_cosine()`: Distance
  metrics

Component selection via `ncomp_by_var()`, `ncomp_by_cumvar()`,
`ncomp_by_opc()`, or `ncomp_fixed()`.

**Redesigned neighbor and fitting interfaces:**

- `neighbors_k()`, `neighbors_diss()`: Neighbor selection constructors
- `fit_pls()`, `fit_wapls()`, `fit_gpr()`: Local fitting constructors
  (replace `local_fit_*()` functions)

**Breaking changes in `mbl()`:**

- `k`, `k_diss`, `k_range` replaced by `neighbors` argument
- `method` renamed to `fit_method`
- `center` and `scale` removed; now controlled within constructors

See `NEWS.md` for full details on deprecated and removed functions.

## Core functionality

**Dimensionality reduction:**

- `ortho_projection()`: PCA or PLS projection with multiple algorithms
  (SVD, NIPALS, SIMPLS)

**Computing dissimilarity matrices:**

- `dissimilarity()`: Main interface for dissimilarity computation
- `diss_pca()`, `diss_pls()`, `diss_correlation()`, `diss_euclidean()`,
  `diss_mahalanobis()`, `diss_cosine()`: Method constructors
- `diss_evaluate()`: Evaluate dissimilarity matrices using side
  information

**Neighbor search:**

- `search_neighbors()`: Efficient k-nearest neighbor retrieval

**Modelling spectral data:**

- `model()`: Global PLS or GPR calibration
- `mbl()`: Memory-based learning for per-query local modelling
- `gesearch()`: Evolutionary subset selection for domain-adaptive
  calibration
- `liblex()`: Pre-computed library of local experts for fast prediction

## Installation

Install from CRAN:

``` r
install.packages("resemble")
```

Or install the development version from GitHub:

``` r
# install.packages("devtools")
devtools::install_github("l-ramirez-lopez/resemble")
```

The package requires a C++ compiler. On Windows, install
[Rtools](https://cran.r-project.org/bin/windows/Rtools/). On macOS, you
may need to install `gfortran` and `clang` from [CRAN
tools](https://cran.r-project.org/bin/macosx/tools/).

## Example: Memory-based learning with `mbl()`

``` r
library(resemble)
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Split into training and test sets
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]
test_y  <- NIRsoil$CEC[NIRsoil$train == 0 & !is.na(NIRsoil$CEC)]

# Memory-based learning using the LOCAL lagorithm of Shenk et al. (1997)
local_shenk <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  neighbors = neighbors_k(seq(50, 130, by = 20)),
  diss_method = diss_correlation(center = FALSE),
  fit_method = fit_wapls(3, 15),
  gh = TRUE
)
local_shenk
plot(local_shenk)
get_predictions(local_shenk)
```

<p align="center">

<img src="./man/figures/mbl_example.png" width="80%">
</p>

## Example: Pre-computed model library with `liblex()`

`liblex()` builds a library of local experts that can be reused for
prediction without refitting:

``` r
# Build model library
model_lib <- liblex(
 Xr = train_x,
 Yr = train_y,
 neighbors = neighbors_k(c(40, 60, 80)),
 diss_method = diss_correlation(ws = 27, scale = TRUE),
 fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "mpls"),
 control = liblex_control(tune = TRUE)
)

# Predict new observations
predictions <- predict(model_lib, test_x)
```

## Example: Evolutionary subset selection with `gesearch()`

`gesearch()` selects optimal subsets from large spectral libraries:

``` r
# Search for optimal calibration subset
gs <- gesearch(
 Xr = train_x, 
 Yr = train_y,
 Xu = test_x,
 k = 50, 
 b = 100, 
 retain = 0.97,
 target_size = 200,
 fit_method = fit_pls(ncomp = 15, method = "mpls"),
 optimization = c("reconstruction", "similarity"),
 seed = 42
)

# Predict using selected subset
preds <- predict(gs, test_x)
plot(gs)
```

## Memory-based learning overview

Memory-based learning (MBL, a.k.a. instance-based learning or local
modelling) is a non-linear lazy learning approach. For each prediction,
the algorithm:

1.  Finds the k-nearest neighbors in the reference set
2.  Fits a local model using those neighbors
3.  Predicts the response for the target observation

The `mbl()` function offers three regression methods for local models:

- `fit_gpr()`: Gaussian process regression with linear kernel
- `fit_pls()`: Partial least squares
- `fit_wapls()`: Weighted average PLS (Shenk et al., 1997)

<p align="center">

<img src="./vignettes/MBL.gif" width="50%">
</p>

## Citing the package

``` r
citation(package = "resemble")
```

## News: Memory-based learning and `resemble`

- **2026.04:** [Greenberg et al.,
  2026](https://doi.org/10.1016/j.geoderma.2026.117829)  
  used `resemble` for MBL-PLSR modelling of portable X-ray fluorescence
  (pXRF) spectra, showing improved soil property predictions relative to
  global PLSR across in-situ and ex-situ measurement settings.

- **2026.05:** [Ramirez-Lopez et al.,
  2026](https://doi.org/10.1016/j.aca.2026.345682) introduced `liblex`,
  a retrieval-gated modelling framework that converts spectral libraries
  into libraries of precomputed local experts for scalable memory-based
  learning. For this, `resemble` was used.

- **2026.05:** [Ramirez-Lopez et al.,
  2026](https://doi.org/10.1016/j.aca.2026.345651) introduced
  `gesearch`, an evolutionary subset-selection method for identifying
  target-domain-relevant samples from complex spectral libraries. For
  this, `resemble` was used.

- **2026.05:** [van Leeuwen et al.,
  2026](https://doi.org/10.1016/j.geoderma.2026.117804)  
  used `resemble` for principal component Mahalanobis nearest-neighbour
  search to extract spectrally similar samples from the KSSL library for
  MIR model calibration in Dutch soils.

- **2026.04:** [Irving et al.,
  2026](https://doi.org/10.1016/j.seh.2026.100205)  
  used `resemble` in modelling workflows for infrared spectroscopy
  prediction of soil microbial properties across Australian soils.

- **2026.03:** [Shrestha et al.,
  2026](https://doi.org/10.1016/j.geodrs.2026.e01063)  
  used `resemble` in a hybrid localisation workflow to predict
  farm-scale soil cadmium from a regional spectral library; LOCAL models
  with MIR data performed best.

- **2025.10:** [Summerauer et al.,
  2025](https://doi.org/10.5194/egusphere-2025-4625) used `resemble` for
  MBL modelling of soil properties from infrared spectra across tropical
  hillslopes in Eastern Africa.

- **2025.05:** [Sun and Shi,
  2025](https://doi.org/10.1016/j.geoderma.2025.117298) combined
  spectral and geographical similarity for SOC prediction; local PLSR
  outperformed global models.

- **2025.03:** [Breure et al.,
  2025](https://doi.org/10.1038/s41467-025-57355-y)  
  used `resemble` for local VNIR modelling of soil carbon fractions (POC
  and MAOC) across European agricultural soils, published in *Nature
  Communications*.

- **2025.03:** [Purushothaman et al.,
  2025](https://doi.org/10.1109/TGRS.2025.3569059) applied MBL to
  AVIRIS-NG hyperspectral data for soil property prediction in India.

- **2025.01:** [Dai et al.,
  2025](https://doi.org/10.1016/j.still.2024.106297) used MBL for POC
  and MAOC prediction from VNIR in Guangdong.

- **2024.12:** [Asrat et al.,
  2024](https://doi.org/10.1016/j.geoderma.2024.117116) MBL for local
  calibration sample selection in the Moroccan Soil Spectral Library.

- **2024.09:** [Barbetti et al.,
  2024](https://doi.org/10.1109/TAFE.2024.3449215) MBL to detect SOC
  changes in long-term experiments using vis–NIR.

- **2023.11:** [Wang et al.,
  2023](https://doi.org/10.1016/j.geoderma.2023.116752) N-MBL (MBL + RF
  within local fitting) for regional vis–NIR models.

- **2022:** [Sanderman et al., 2022](https://doi.org/10.1002/saj2.20513)
  evaluated transferability of large MIR spectral databases across
  instruments.

- **2022.01:** [Ng et al.,
  2022](https://doi.org/10.1016/j.geoderma.2021.115501) showed that MBL
  yields better local SOC predictions than spiking approaches.

- **2021.10:** [Ramirez-Lopez et al.,
  2021](https://soil.copernicus.org/articles/7/693/2021/) MBL to predict
  soil properties in Africa.

- **2020.08:** Charlotte Rivard’s MIR MBL tutorial:
  <https://whrc.github.io/Soil-Predictions-MIR/>

- **2020.01:** [Sanderman et al.,
  2020](https://doi.org/10.1002/saj2.20009) MIR spectroscopy for
  prediction of soil health indicators; MBL and Cubist excelled.

- **2019.03:** [Ramirez-Lopez et al.,
  2019](https://doi.org/10.1111/ejss.12752) MBL in digital soil mapping
  at farm scale.

- **2019.03:** [Jaconi et al.,
  2019](https://doi.org/10.1016/j.geoderma.2018.11.042) MBL for
  national-scale NIR texture predictions in Germany.

- **2018.01:** [Dotto et al.,
  2018](https://doi.org/10.1016/j.geoderma.2017.11.028) MBL for SOC
  prediction in Brazil.

- **2016.04:** [Viscarra Rossel et al.,
  2016](https://doi.org/10.3390/rs8040341) memory-based learning for
  soil property prediction.

- **2014.03:** First CRAN release of `resemble`.

## Related packages

- [`prospectr`](https://github.com/l-ramirez-lopez/prospectr): Signal
  processing and chemometrics for spectroscopy

## Contributing

Contributions are welcome! Please read our Contributing Guidelines
(available in the GitHub repo) before submitting pull requests.

This project follows a Code of Conduct available in the GitHub repo.

## Bug reports

Report issues at
[GitHub](https://github.com/l-ramirez-lopez/resemble/issues) or contact
the maintainer (<ramirez.lopez.leo@gmail.com>).

## References

Lobsey, C. R., Viscarra Rossel, R. A., Roudier, P., & Hedley, C. B.
2017. rs-local data-mines information from spectral libraries to improve
local calibrations. European Journal of Soil Science, 68(6), 840-852.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte,
J.A.M., Scholten, T. 2013. The spectrum-based learner: A new local
approach for modeling soil vis-NIR spectra of complex data sets.
Geoderma 195-196, 268-279.

Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
libraries are too complex to search: Evolutionary subset selection for
domain-adaptive calibration. Analytica Chimica Acta, under review.

Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C., Perez-Fernandez,
E., Plans, M., Breure, T., Behrens, T., Viscarra Rossel, R., & Peng, Y.
(2026b). Rethinking local spectral modelling: From per-query refitting
to model libraries. Analytica Chimica Acta, under review.

Saul, L. K., & Roweis, S. T. 2003. Think globally, fit locally:
unsupervised learning of low dimensional manifolds. Journal of machine
learning research, 4(Jun), 119-155.

Shenk, J., Westerhaus, M., and Berzaghi, P. 1997. Investigation of a
LOCAL calibration procedure for near infrared instruments. Journal of
Near Infrared Spectroscopy, 5, 223-232.

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| eval: true
# Retain components that individually explain at least 1% of variance
proj_var <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))
proj_var


## -----------------------------------------------------------------------------
#| eval: true
# Retain enough components to explain at least 99% of variance
proj_cumvar <- ortho_projection(train_x, ncomp = ncomp_by_cumvar(0.99))
proj_cumvar


## -----------------------------------------------------------------------------
#| eval: true
# Optimal component selection using side information
# using default max_ncomp which is 40
proj_opc <- ortho_projection(train_x, Yr = train_y, ncomp = ncomp_by_opc())
proj_opc


## -----------------------------------------------------------------------------
# this needs to be passed to the ncomp argument in ortho_projection()
ncomp_by_opc(max_ncomp = 15)


## -----------------------------------------------------------------------------
#| eval: false
# # Using ncomp_fixed()
# proj_fixed <- ortho_projection(train_x, ncomp = ncomp_fixed(10))
# 
# # Equivalent: passing an integer directly
# proj_fixed <- ortho_projection(train_x, ncomp = 10)


## -----------------------------------------------------------------------------
#| results: hide
# PCA with default component selection (ncomp_by_var(0.01))
pca_train <- ortho_projection(train_x, method = "pca")
pca_train


## -----------------------------------------------------------------------------
#| label: fig-pca-variance
#| fig-cap: "Individual contribution to the explained variance for each component (left) and cumulative variance explained by the principal components (right)."
#| fig-width: 7
#| fig-height: 3
plot(pca_train)


## -----------------------------------------------------------------------------
#| results: hide
# PCA using NIPALS
pca_nipals_train <- ortho_projection(train_x, method = "pca_nipals")


## -----------------------------------------------------------------------------
#| results: hide
# PLS using SIMPLS with 10 components
pls_train <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = 10
)
pls_train


## -----------------------------------------------------------------------------
#| results: hide
# PCA with optimal component selection
pca_opc <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40)
)
pca_opc


## -----------------------------------------------------------------------------
#| label: fig-pca-opc
#| fig-cap: "RMSD between observations and their nearest neighbors as a function of the number of principal components."
#| fig-width: 5
#| fig-height: 4
plot(pca_opc, col = "#F59E0B")


## -----------------------------------------------------------------------------
#| results: hide
# Fit PLS projection model
pls_model <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

# Project test data
pls_test_scores <- predict(pls_model, newdata = test_x)


## -----------------------------------------------------------------------------
#| label: fig-pc-proj
#| fig-width: 4.5
#| fig-height: 5
#| fig-cap: "Projection of the training observations in the first two PLS components (blue) and projection of the test observations onto the same latent space (red)."
plot(
  pls_model$scores[, 1:2], 
  col = rgb(0.231, 0.51, 0.965, 0.3), 
  pch = 16, cex = 1.5,
  xlab = "PLS1", ylab = "PLS2"
)
grid(lty = 1)
points(
  pls_test_scores[, 1:2], 
  col = rgb(0.961, 0.620, 0.043, 0.7), 
  pch = 16, cex = 1.5
)
legend(
  "topright",
  legend = c("Training projection", "Predicted projection"),
  col = c(
    rgb(0.231, 0.51, 0.965, 0.3),
    rgb(0.961, 0.620, 0.043, 0.7)
  ),
  pch = 16,
  pt.cex = 1.5,
  box.col = "black"
)


## -----------------------------------------------------------------------------
#| results: hide
# Project training and test data simultaneously
pca_both <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)


## -----------------------------------------------------------------------------
#| results: hide
#| eval: false
# train_y2 <- NIRsoil$Nt[NIRsoil$train == 1]
# train_y3 <- NIRsoil$CEC[NIRsoil$train == 1]
# 
# # PLS with multivariate side information
# pls_multi <- ortho_projection(
#   train_x,
#   Xu = test_x,
#   Yr = cbind(train_y, train_y2, train_y3),
#   method = "pca",
#   ncomp = ncomp_by_opc(40),
#   scale = TRUE
# )
# 
# plot(pls_multi)

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)


## ----neighbors-k-example, eval = FALSE----------------------------------------
# neighbors_k(k = 50)
# neighbors_k(k = c(40, 60, 80, 100))


## ----neighbors-diss-example, eval = FALSE-------------------------------------
# neighbors_diss(threshold = 0.3)
# neighbors_diss(threshold = c(0.1, 0.2, 0.3), k_min = 10, k_max = 150)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
set.seed(8011)
rnd_idc <- sample(nrow(test_x), 3)

k_fixed <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2, return_projection = TRUE),
  neighbors = neighbors_k(30)
)


## -----------------------------------------------------------------------------
k_diss <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2),
  neighbors = neighbors_diss(threshold = 0.25)
)


## -----------------------------------------------------------------------------
test_scores_indices <- grep("^Xu_", rownames(k_fixed$projection$scores))


## -----------------------------------------------------------------------------
#| label: fig-neighbors
#| fig-cap: "Nearest neighbors identified using the same dissimilarity metric but retained using two different selection methods."
#| fig-width: 8.5
#| fig-height: 4.5

old_par <- par(no.readonly = TRUE)
on.exit(par(old_par), add = TRUE)

par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Fixed k-nearest neighbors"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_fixed$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Neighbors selected by a \nthreshold dissimilarity"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_diss$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)


## ----knn-pca, eval = FALSE----------------------------------------------------
# # matrix of neighbors
# k_fixed$neighbors
# 
# # matrix of neighbor distances (dissimilarity scores)
# k_fixed$neighbors_diss
# 
# # the index (in the training set) of the first two closest neighbors found in
# # training for the first observation in testing:
# k_fixed$neighbors[1:2, 1, drop = FALSE]
# 
# # the distances of the two closest neighbors found in
# # training for the first observation in testing:
# k_fixed$neighbors_diss[1:2, 1, drop = FALSE]
# 
# # the indices in training that fall in any of the
# # neighborhoods of testing
# k_fixed$unique_neighbors


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-other-methods
# using PC dissimilarity with optimal selection of components
knn_opc <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(),
    scale = TRUE,
    return_projection = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using PLS dissimilarity with optimal selection of components
knn_pls <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pls(
    ncomp = ncomp_by_opc(),
    scale = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using correlation dissimilarity
knn_cor <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(),
  neighbors = neighbors_k(50)
)

# using moving window correlation dissimilarity
knn_mw <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(ws = 51),
  neighbors = neighbors_k(50)
)


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-threshold-example
#| results: hide
# a dissimilarity threshold
d_th <- 1

# the minimum number of observations required in each neighborhood
k_min <- 20

# the maximum number of observations allowed in each neighborhood
k_max <- 300

dnn_pca <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(scale = TRUE),
  neighbors = neighbors_diss(threshold = d_th, k_min = k_min, k_max = k_max)
)

# matrix of neighbors. The minimum number of indices is 20 (given by k_min)
# and the maximum number of indices is 300 (given by k_max).
# NAs indicate "not a neighbor"
dnn_pca$neighbors

# this reports how many neighbors were found for each observation in 
# testing using the input distance threshold (column n_k) and how 
# many were finally selected (column final_n_k)
dnn_pca$k_diss_info

# matrix of neighbor distances
dnn_pca$neighbors_diss

# the indices in training that fall in any of the 
# neighborhoods of testing
dnn_pca$unique_neighbors


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: fig-knn-hist
#| fig-cap: "Histogram of the final neighborhood sizes after applying the dissimilarity threshold and the minimum and maximum neighborhood size constraints."
#| fig-width: 5
#| fig-height: 4
hist(
  dnn_pca$k_diss_info$final_n_k,
  breaks = 20,
  xlab = "Final neighborhood size",
  main = "",
  col = "dodgerblue"
)


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-spike-example
# the indices of the observations that we want to "invite" to every neighborhood
forced_guests <- c(1, 5, 8, 9)

# using PC dissimilarity with optimal selection of components
knn_spiked <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(20)
  ),
  Yr = train_y,
  neighbors = neighbors_k(50),
  spike = forced_guests
)

# check the first 8 neighbors found in training for the 
# first 2 observations in testing
knn_spiked$neighbors[1:8, 1:2]

## -----------------------------------------------------------------------------
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
#| label: fit-methods
# PLS with 15 components
fit_pls(ncomp = 15)

# SIMPLS with scaling
fit_pls(ncomp = 15, method = "simpls", scale = TRUE)

# mPLS with scaling
fit_pls(ncomp = 15, method = "mpls")

# GPR with default noise variance
fit_gpr()


## -----------------------------------------------------------------------------
#| label: data-prep
#| message: false
library(prospectr)

wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split data
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

# Remove missing values
ok_train <- !is.na(train_y)
ok_test  <- !is.na(test_y)
train_x <- train_x[ok_train, ]
train_y <- train_y[ok_train]
test_x  <- test_x[ok_test, ]
test_y  <- test_y[ok_test]


## -----------------------------------------------------------------------------
#| label: fit-pls

set.seed(1124) # guarantee same CV splits for all methods
pls_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_pls(ncomp = 12, method = "mpls", scale = TRUE),
 control = model_control(validation_type = "lgo", number = 10)
)

pls_mod


## -----------------------------------------------------------------------------
#| label: fit-gpr
set.seed(1124) # guarantee same CV splits for all methods
gpr_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_gpr(noise_variance = 0.5),
 control = model_control(validation_type = "lgo", number = 10)
)

gpr_mod


## -----------------------------------------------------------------------------
#| label: predict
# Predict using optimal number of components (from CV)
pls_pred <- predict(pls_mod, newdata = test_x)

# Predict with GPR
gpr_pred <- predict(gpr_mod, newdata = test_x)

# Compare predictions
data.frame(
  observed = test_y,
  pls = pls_pred[, which(pls_mod$cv_results$optimal)],
  gpr = as.vector(gpr_pred)
) |> head(10)


## -----------------------------------------------------------------------------
#| label: validation
# Function to compute stats
eval_pred <- function(obs, pred) {
  data.frame(
    rmse = sqrt(mean((obs - pred)^2)),
    r2 = cor(obs, pred)^2
  )
}


## -----------------------------------------------------------------------------
## PLS evaluation
eval_pred(test_y, pls_pred[, which(pls_mod$cv_results$optimal)])


## -----------------------------------------------------------------------------
## GPR evaluation
eval_pred(test_y, gpr_pred)

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| eval: true
citation(package = "resemble")


## -----------------------------------------------------------------------------
#| label: libraries
#| message: false
library(resemble)
library(prospectr)


## -----------------------------------------------------------------------------
#| message: false
#| results: hide
data(NIRsoil)
dim(NIRsoil)
str(NIRsoil)


## -----------------------------------------------------------------------------
#| label: NIRsoil
#| message: false
# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)


## -----------------------------------------------------------------------------
#| label: fig-plotspectra
#| fig-cap: "Raw spectral absorbance data (top) and first derivative of the absorbance spectra (bottom)."
#| fig-align: center
#| fig-width: 7
#| fig-height: 7
#| echo: false
# Hex sticker palette#
bg_dark <- NA #"#0F172A"      # background
blue <- "#3B82F6"         # border / spiral high
amber <- "#F59E0B"        # spiral low
amber_light <- "#FBBF24"  # text accent
slate <- "#64748B"        # muted elements

old_par <- par("mfrow", "mar", "bg")
par(mfrow = c(2, 1), mar = c(4, 4, 1, 4), bg = bg_dark)

new_wavs <- as.matrix(as.numeric(colnames(NIRsoil$spc_p)))

text_col <- "black"#"white"

# Plot 1: Raw spectra
plot(range(wavs), range(NIRsoil$spc), col = NA,
     xlab = "",
     ylab = "Absorbance",
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = wavs, y = t(NIRsoil$spc), 
         lty = 1, col = paste0(blue, "33"))

# Plot 2: First derivative
plot(range(new_wavs), range(NIRsoil$spc_p), col = NA,
     xlab = "Wavelengths, nm",
     ylab = expression(d(detrended~A)/d*lambda),
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = new_wavs, y = t(NIRsoil$spc_p), 
         lty = 1, col = paste0(amber, "33"))

par(old_par)


## -----------------------------------------------------------------------------
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
diss_correlation()


## -----------------------------------------------------------------------------
diss_correlation(center = TRUE, scale = TRUE)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: pca-diss
#| eval: TRUE

# Default: variance-based component selection (ncomp_by_var(0.01))
d_pca <- dissimilarity(train_x, diss_method = diss_pca())

# With OPC-based component selection (requires Yr)
d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30))
)

# Between training and test sets
d_pca_tr_ts <- dissimilarity(
 train_x,
 Xu = test_x,
 Yr = train_y,
 diss_method = diss_pca(
   ncomp = ncomp_by_opc(30),
   return_projection = TRUE
 )
)

d_pca_tr_ts


## -----------------------------------------------------------------------------
# Example of the first 5 rows and columns of the dissimilarity 
# matrix between training and test sets
# Rows: training observations; 
# Columns: test observations
d_pca_tr_ts$dissimilarity[1:5, 1:5]


## -----------------------------------------------------------------------------
d_pca_tr_ts$projection


## ----pls-diss, eval = FALSE---------------------------------------------------
# # Default: OPC-based component selection
# d_pls <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   Yr = train_y,
#   diss_method = diss_pls()
# )
# 
# # Fixed number of components
# d_pls_fixed <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   Yr = train_y,
#   diss_method = diss_pls(ncomp = 15)
# )


## ----cor-diss, eval = FALSE---------------------------------------------------
# # Standard correlation dissimilarity
# d_cor <- dissimilarity(train_x, diss_method = diss_correlation())
# d_cor


## -----------------------------------------------------------------------------
correlation_baser <- 0.5 * (1 - cor(t(train_x)))
corrrelation_resemble <- dissimilarity(
  train_x, diss_method = diss_correlation(center = FALSE)
)
# the maximum discreepancy between the two implementations should 
# be very close to zero
max(abs(corrrelation_resemble$dissimilarity - correlation_baser))


## -----------------------------------------------------------------------------
#| eval: false
# # Compare computational speed: resemble vs base R correlation dissimilarity
# n_iter <- 50
# 
# # Transpose once before timing (fair comparison)
# train_x_t <- t(train_x)
# 
# # resemble implementation
# time_resemble <- system.time({
#   for (i in seq_len(n_iter)) {
#     cor_resemble <- dissimilarity(train_x, diss_method = diss_correlation(center = FALSE))
#   }
# })
# 
# # base R implementation
# time_base <- system.time({
#   for (i in seq_len(n_iter)) {
#     cor_base <- 0.5 * (1 - cor(train_x_t))
#   }
# })
# 
# # Results
# data.frame(
#   method = c("resemble", "base R"),
#   elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
#   per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
# )


## ----cor-diss-mw, eval = FALSE------------------------------------------------
# # Moving window correlation (window size must be odd)
# d_cor_mw <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   diss_method = diss_correlation(ws = 51)
# )
# d_cor_mw


## ----omp-threads-use, eval = FALSE--------------------------------------------
# # Use 4 threads
# Sys.setenv(OMP_NUM_THREADS = 4)
# 
# # Or use all available cores
# Sys.setenv(OMP_NUM_THREADS = parallel::detectCores())


## ----omp-threads-restrict, eval = FALSE---------------------------------------
# # Limit OpenMP threads
# Sys.setenv(OMP_NUM_THREADS = 1)
# 
# # If using OpenBLAS
# Sys.setenv(OPENBLAS_NUM_THREADS = 1)
# 
# # If using MKL
# Sys.setenv(MKL_NUM_THREADS = 1)


## ----blas-threads, eval = FALSE-----------------------------------------------
# RhpcBLASctl::blas_set_num_threads(1)
# RhpcBLASctl::omp_set_num_threads(1)


## ----euclid-diss, eval = FALSE------------------------------------------------
# d_euclid <- dissimilarity(train_x, Xu = test_x, diss_method = diss_euclidean())
# d_euclid


## -----------------------------------------------------------------------------
ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
ed_baser <- as.matrix(dist(train_x, method = "euclidean"))

# scaling the base R Euclidean distance by the number of variables:
ed_baser_scaled <- sqrt((ed_baser^2) / ncol(train_x))

# differences between the two implementations
max(abs(ed_baser_scaled - ed_resemble$dissimilarity))


## -----------------------------------------------------------------------------
#| label: euclid-speed
#| eval: false
# # Compare computational speed: resemble vs base R
# n_iter <- 50
# 
# # resemble implementation
# time_resemble <- system.time({
#   for (i in seq_len(n_iter)) {
#     ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
#   }
# })
# 
# # base R implementation
# time_base <- system.time({
#   for (i in seq_len(n_iter)) {
#     ed_base <- as.matrix(dist(train_x, method = "euclidean"))
#   }
# })
# 
# # Results
# data.frame(
#   method = c("resemble", "base R"),
#   elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
#   per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
# )


## ----cosine-diss, eval = FALSE------------------------------------------------
# d_cosine <- dissimilarity(train_x, Xu = test_x, diss_method = diss_cosine())
# d_cosine


## -----------------------------------------------------------------------------
# Compute dissimilarity matrices using different methods
d_pca_var <- dissimilarity(train_x, diss_method = diss_pca(scale = TRUE))

d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30), scale = TRUE)
)

d_pls_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE)
)

d_cor <- dissimilarity(
  train_x,
  diss_method = diss_correlation(ws = 51, scale = TRUE)
)

d_euclid <- dissimilarity(
  train_x,
  diss_method = diss_euclidean(scale = TRUE)
)

d_cosine <- dissimilarity(
  train_x,
  diss_method = diss_cosine(scale = TRUE)
)


## -----------------------------------------------------------------------------
# Evaluate each method using Ciso as side information
side_info <- as.matrix(train_y)

eval_results <- list(
  "PCA (var)" = diss_evaluate(
    d_pca_var$dissimilarity, side_info
  ),
  "PCA (opc)" = diss_evaluate(
    d_pca_opc$dissimilarity, side_info
  ),
  "PLS (opc)" = diss_evaluate(
    d_pls_opc$dissimilarity, side_info
  ),
  correlation = diss_evaluate(
    d_cor$dissimilarity, side_info
  ),
  Euclidean = diss_evaluate(
    d_euclid$dissimilarity, side_info
  ),
  cosine = diss_evaluate(
    d_cosine$dissimilarity, side_info
  )
)

# Extract RMSD and correlation for each method
comparison <- do.call(rbind, lapply(names(eval_results), function(nm) {
  data.frame(
    method = nm,
    rmsd   = eval_results[[nm]]$eval[, "rmsd"],
    r      = eval_results[[nm]]$eval[, "r"]
  )
}))
comparison


## -----------------------------------------------------------------------------
#| label: diss-eval-plot
#| fig-cap: "Comparison of dissimilarity methods based on first nearest-neighbor evaluation using `Ciso` as side information. In each panel, the x-axis shows the `Ciso` value of the nearest neighbor identified from the spectral dissimilarity matrix, and the y-axis shows the observed `Ciso` value of the corresponding sample. The dashed line represents perfect agreement (y = x)."
#| fig-align: "center"
#| fig-width: 8
#| fig-height: 6
#| echo: true
#| fig-show: hold
blue  <- "#3B82F6"
amber <- "#F59E0B"
slate_grid <- "#33415540"

old_par <- par(mfrow = c(2, 3), mar = c(4, 4, 3, 1))

for (nm in names(eval_results)) {
  xy <- eval_results[[nm]]$first_nn
  plot(
    xy[, 2], xy[, 1],
    xlab = "Ciso (1-NN)",
    ylab = "Ciso",
    col = adjustcolor(blue, alpha.f = 0.5),
    pch = 16,
    main = nm
  )
  grid(lty = 1, col = slate_grid)
  abline(0, 1, col = amber, lwd = 2)
}

par(old_par)

## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)
library(prospectr)
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| label: data-prep
#| message: false
library(resemble)
library(prospectr)

data(NIRsoil)

# Wavelengths
wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split into training and test sets
# Note: missing values in the response are allowed in liblex
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

cat("Training set:", nrow(train_x), "observations\n")
cat("Test set:", nrow(test_x), "observations\n")


## -----------------------------------------------------------------------------
#| label: neighbors-example1
# Fixed neighborhood sizes to evaluate
neighbors_k(k = seq(40, 120, by = 20))


## -----------------------------------------------------------------------------
#| label: neighbors-example2
# Threshold-based selection with bounds
neighbors_diss(
  threshold = seq(0.05, 0.5, length.out = 10), 
  k_min = 40, 
  k_max = 150
)


## -----------------------------------------------------------------------------
#| label: diss-example
# Moving-window correlation dissimilarity (recommended for spectra)
diss_correlation(ws = 37, scale = TRUE)

# PCA-based Mahalanobis distance with optimized components
diss_pca()


## -----------------------------------------------------------------------------
#| label: fit-method-example
# waPLS with modified PLS algorithm and scaling
fit_wapls(
  min_ncomp = 3, 
  max_ncomp = 15, 
  scale = TRUE, 
  method = "mpls"
)


## -----------------------------------------------------------------------------
#| label: control-example
#| eval: false
# # Build library with hyperparameter tuning
# liblex_control(mode = "build", tune = TRUE)


## -----------------------------------------------------------------------------
#| eval: false
# ciso_lib_k <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 80, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_k


## -----------------------------------------------------------------------------
#| label: liblex-fixed-k
#| cache: true
#| echo: false
#| message: false
#| warning: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_k


## -----------------------------------------------------------------------------
#| label: fig-liblex
#| fig-cap: "Top: Best model obtained for each neighborhood size (based on the RMSE). Bottom: Centroids of the neighborhoods, usled later in prediction to selet the models to be used."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 6.5
plot(ciso_lib_k)


## -----------------------------------------------------------------------------
#| label: liblex-threshold
#| cache: true
#| echo: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_thr


## -----------------------------------------------------------------------------
#| cache: true
#| eval: false
# ciso_lib_thr <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_diss(
#     threshold = seq(0.05, 0.5, length.out = 5),
#     k_min = 40,
#     k_max = 150
#   ),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_thr


## -----------------------------------------------------------------------------
#| label: fig-centroids
#| fig-cap: "Neighborhood centroids (blue) compared to test spectra (red)"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center
wavs_pr <- as.numeric(colnames(ciso_lib_k$scaling$local_x_center))

matplot(
  wavs_pr,
  t(test_x),
  col = rgb(1, 0, 0, 0.3),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance"
)

matlines(
  wavs_pr,
  t(ciso_lib_k$scaling$local_x_center),
  col = rgb(0, 0, 1, 0.3),
  lty = 1
)

grid(lty = 1)

legend(
  "topright",
  legend = c("Samples to predict", "Neighborhood centroids"),
  col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
  lty = 1,
  lwd = 2,
  bty = "n"
)


## -----------------------------------------------------------------------------
#| label: fig-regression-coefficients
#| fig-cap: "Regression coefficients (top) and intercept distribution (bottom) of the local experts"
#| fig-width: 8
#| fig-height: 7
#| fig-align: center

par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Regression coefficients across wavelengths
models <- ciso_lib_k$coefficients

matplot(
  x = wavs_pr,
  y = t(models$B),
  type = "l",
  lty = 1,
  col = rgb(0.23, 0.51, 0.96, 0.15),
  xlab = "Wavelength (nm)",
  ylab = "Regression coefficient",
  main = paste0("Regression coefficients (", nrow(models$B), " experts)")
)

# Add mean coefficient profile
lines(wavs_pr, colMeans(models$B), col = "red", lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Individual experts", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96, 0.6), "red"),
  lty = 1,
  lwd = c(1, 2),
  bty = "n"
)

# Distribution of intercepts
plot(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96),
  lwd = 2,
  xlab = "Intercept value",
  ylab = "Density",
  main = "Distribution of intercepts"
)
polygon(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96, 0.2),
  border = NA
)
abline(v = mean(models$B0, na.rm = TRUE), col = "red", lty = 2, lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Density", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96), "red"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)

par(mfrow = c(1, 1))


## -----------------------------------------------------------------------------
#| label: kmeans-anchors
# Select 350 representative anchors using k-means sampling
# on the first 20 principal components
set.seed(1124)
kms <- naes(
  train_x, 
  k = 350, 
  pc = 20, 
  iter.max = 100,
  .center = TRUE, 
  .scale = TRUE
)

anchor_km <- kms$model

cat("Selected", length(anchor_km), "anchors via k-means\n")


## -----------------------------------------------------------------------------
#| label: liblex-anchored
#| cache: true
#| echo: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_anchored


## -----------------------------------------------------------------------------
#| cache: true
#| eval: false
# ciso_lib_anchored <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 80, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   anchor_indices = anchor_km,
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_anchored


## -----------------------------------------------------------------------------
# Number of experts in the anchored library
n_experts <- nrow(ciso_lib_anchored$coefficients$B)
cat("Number of experts in the anchored library:", n_experts, "\n")


## -----------------------------------------------------------------------------
#| label: fig-centroids-anchored
#| fig-cap: "Neighborhood centroids from anchored library compared to test spectra"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center

wavs_pr <- as.numeric(colnames(ciso_lib_anchored$scaling$local_x_center))
n_experts <- nrow(ciso_lib_anchored$scaling$local_x_center)
n_test <- nrow(test_x)

# Plot test spectra first (background)
matplot(
  wavs_pr,
  t(test_x),
  col = rgb(0.8, 0.2, 0.2, 0.2),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance",
  main = paste0("Coverage: ", n_experts, " experts vs ", n_test, " test samples")
)

# Overlay centroids
matlines(
  wavs_pr,
  t(ciso_lib_anchored$scaling$local_x_center),
  col = rgb(0.23, 0.51, 0.96, 0.3),
  lty = 1
)
grid(lty = 1)

legend(
  "topright",
  legend = c(
    paste0("Test samples (n = ", n_test, ")"),
    paste0("Centroids (n = ", n_experts, ")")
  ),
  col = c(
    rgb(0.8, 0.2, 0.2, 0.5),
    rgb(0.23, 0.51, 0.96, 0.5)
  ),
  lty = 1,
  lwd = c(1, 1, 2, 2),
  bty = "n"
)


## -----------------------------------------------------------------------------
#| label: predict-basic
ciso_pred <- predict(ciso_lib_k, newdata = test_x, verbose = FALSE)

# Prediction output structure
names(ciso_pred)

# Main predictions with uncertainty
head(ciso_pred$predictions)


## -----------------------------------------------------------------------------
#| label: weighting-options
#| eval: false
# # Gaussian kernel (default)
# predict(ciso_lib_k, newdata = test_x, weighting = "gaussian")
# 
# # Tricubic kernel (similar to LOCAL algorithm)
# predict(ciso_lib_k, newdata = test_x, weighting = "tricube")
# 
# # Equal weights
# predict(ciso_lib_k, newdata = test_x, weighting = "none")


## -----------------------------------------------------------------------------
#| label: enforce-indices
#| eval: false
# # Always include specific experts in predictions
# predict(ciso_lib_k, newdata = test_x, enforce_indices = c(1, 5, 10))


## -----------------------------------------------------------------------------
#| label: evaluation
pred_values <- ciso_pred$predictions$pred
pred_sd <- ciso_pred$predictions$pred_sd

# Performance metrics (excluding missing values)
complete <- !is.na(test_y)
r2 <- cor(pred_values[complete], test_y[complete])^2
rmse <- sqrt(mean((pred_values[complete] - test_y[complete])^2))

cat("R²:  ", round(r2, 3), "\n")
cat("RMSE:", round(rmse, 3), "\n")


## -----------------------------------------------------------------------------
#| label: uncertainty-filter
# Filter predictions by uncertainty threshold
unc_threshold <- quantile(pred_sd[complete], 0.75)
reliable <- complete & (pred_sd < unc_threshold)

r2_filtered <- cor(pred_values[reliable], test_y[reliable])^2
rmse_filtered <- sqrt(mean((pred_values[reliable] - test_y[reliable])^2))

cat("After filtering high-uncertainty predictions:\n")
cat("  Retained:", sum(reliable), "/", sum(complete), "observations\n")
cat("  R²:      ", round(r2_filtered, 3), "\n")
cat("  RMSE:    ", round(rmse_filtered, 3), "\n")


## -----------------------------------------------------------------------------
#| label: fig-validation
#| fig-cap: "Predicted versus observed values"
#| fig-width: 6
#| fig-height: 5
#| fig-align: center
lims <- range(pred_values[complete], test_y[complete], na.rm = TRUE)

plot(
  pred_values[complete],
  test_y[complete],
  pch = 16,
  col = rgb(0, 0, 0, 0.5),
  xlab = "Predicted",
  ylab = "Observed",
  xlim = lims,
  ylim = lims
)

abline(0, 1, col = "red", lwd = 2)
grid(lty = 1)


## -----------------------------------------------------------------------------
#| label: parallel-example
#| eval: false
# library(doParallel)
# 
# # Register parallel backend
# n_cores <- min(parallel::detectCores() - 1, 4)
# cl <- makeCluster(n_cores)
# registerDoParallel(cl)
# 
# # Build library with parallel processing
# ciso_lib_parallel <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 120, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(
#     tune = TRUE,
#     allow_parallel = TRUE,
#     chunk_size = 10
#   )
# )
# 
# # Clean up
# stopCluster(cl)
# registerDoSEQ()

## -----------------------------------------------------------------------------
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
#| label: fit-methods
#| eval: true

# Creates an object with instructions to build PLS models
my_pls <- fit_pls(ncomp = 15)
my_pls

# Creates an object with instructions to build WAPLS models
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 20)
my_wapls

# Creates an object with instructions to build GPR models
my_gpr <- fit_gpr()
my_gpr


## -----------------------------------------------------------------------------
#| label: validation-control
#| eval: true
# Create an object with instructions to conduct both validation types
# "NNv" and "local_cv"
two_val_control <- mbl_control(
  validation_type = c("NNv", "local_cv"),
  number = 10,
  p = 0.75
)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: mbl-local
#| eval: true
#| results: hide
# Define the neighborhood sizes to test
my_ks <- seq(80, 160, by = 40)

# Define how to use the dissimilarity information (ignore it)
ignore_diss <- "none"

# Define the regression method to be used at each neighborhood
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 25, scale = TRUE)

# For the moment use only "NNv" validation (faster)
nnv_val_control <- mbl_control()

# Predict Total Carbon
# (remove missing values)
local_ciso <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_k(my_ks),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  gh = TRUE,
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: fig-localresultsciso
#| fig-cap: "MBL results for Total Carbon predictions using the LOCAL algorithm. NNv: nearest-neighbor cross-validation."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4
plot(local_ciso, main = "")
local_ciso


## -----------------------------------------------------------------------------
#| label: bestk
#| eval: true
#| echo: false
#| results: hide
bestk <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bestk <- local_ciso$validation_results$nearest_neighbor_validation$k[bestk]


## -----------------------------------------------------------------------------
#| label: get-predictions
#| eval: true
#| results: hide
#| fig-show: hide
bki <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bk <- local_ciso$validation_results$nearest_neighbor_validation$k[bki]

# All the prediction results are stored in:
local_ciso$results

# The get_predictions function makes easier to retrieve the
# predictions from the previous object
ciso_hat <- as.matrix(get_predictions(local_ciso))[, bki]


## -----------------------------------------------------------------------------
#| label: fig-plot-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation."
#| eval: true
#| fig-width: 5
#| fig-height: 5
# Plot predicted vs reference
rng <- range(ciso_hat, test_y, na.rm = TRUE)
plot(ciso_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a fixed k", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid(lty = 1)
abline(0, 1, col = "red")


## -----------------------------------------------------------------------------
#| eval: true
# Prediction RMSE:
sqrt(mean((ciso_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_hat, test_y, use = "complete.obs")^2


## -----------------------------------------------------------------------------
#| label: mbl-diss-threshold
#| eval: true
#| results: hide
#| fig-show: hide
# Create a vector of dissimilarity thresholds to evaluate
# since the correlation dissimilarity will be used
# these thresholds need to be > 0 and <= 1
dths <- seq(0.025, 0.15, by = 0.025)

# Indicate the minimum and maximum sizes allowed for the neighborhood
k_min <- 30
k_max <- 150

local_ciso_diss <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dths, k_min = k_min, k_max = k_max),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: plot-diss-results
#| eval: false
# plot(local_ciso_diss)


## -----------------------------------------------------------------------------
#| label: bestd
#| eval: true
#| echo: false
#| results: hide
bestd <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bestd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bestd]


## -----------------------------------------------------------------------------
#| label: show-diss-results
#| eval: true
local_ciso_diss


## -----------------------------------------------------------------------------
#| label: diss-predictions
#| eval: true
#| results: hide
#| fig-show: hide
# Best distance threshold
bdi <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bdi]

# Predictions for the best distance
ciso_diss_hat <- as.matrix(get_predictions(local_ciso_diss))[, bdi]


## -----------------------------------------------------------------------------
#| label: fig-diss-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation and distance thresholds."
#| eval: true
# Plot predicted vs reference
plot(ciso_diss_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a distance threshold \nfor neighbor retrieval", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid()
abline(0, 1, col = "red")

## -----------------------------------------------------------------------------
# RMSE
sqrt(mean((ciso_diss_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_diss_hat, test_y, use = "complete.obs")^2


## -----------------------------------------------------------------------------
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_cec <- NIRsoil$CEC[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_cec  <- NIRsoil$CEC[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: cec-examples
#| results: hide
#| eval: true
#| fig-show: hide
# Define the WAPLS fitting method
my_wapls <- fit_wapls(min_ncomp = 2, max_ncomp = 25, scale = FALSE)

# mbl_cor: LOCAL algorithm with correlation dissimilarity
dth_cor <- seq(0.01, 0.3, by = 0.08)
mbl_cor <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_cor, k_min = 80, k_max = 150),
  diss_method = diss_correlation(),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pc: PCA dissimilarity with dissimilarity matrix as predictors
dth_pc <- seq(0.05, 1, by = 0.4)
mbl_pc <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pls: PLS dissimilarity without dissimilarity predictors
mbl_pls <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_gpr: Gaussian process regression with PCA dissimilarity as predictors
mbl_gpr <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = fit_gpr(),
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: collect-predictions
#| eval: true
#| fig-show: hide
# Get the indices of the best results according to
# nearest neighbor validation statistics
c_val_name <- "validation_results"
c_nn_val_name <- "nearest_neighbor_validation"

bi_cor <- which.min(mbl_cor[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pc  <- which.min(mbl_pc[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pls <- which.min(mbl_pls[[c_val_name]][[c_nn_val_name]]$rmse)
bi_gpr <- which.min(mbl_gpr[[c_val_name]][[c_nn_val_name]]$rmse)

preds <- cbind(
  get_predictions(mbl_cor)[, bi_cor],
  get_predictions(mbl_pc)[, bi_pc],
  get_predictions(mbl_pls)[, bi_pls],
  get_predictions(mbl_gpr)[, bi_gpr]
)

colnames(preds) <- c("mbl_cor", "mbl_pc", "mbl_pls", "mbl_gpr")
preds <- as.matrix(preds)

# R2s
cor(test_cec, preds, use = "complete.obs")^2

# RMSEs
colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5


## -----------------------------------------------------------------------------
#| label: cec-plots-code
#| eval: false
#| fig-show: hide
# old_par <- par("mfrow", "mar")
# par(mfrow = c(2, 2))
# 
# plot(test_cec, preds[, "mbl_cor"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_cor (LOCAL)")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_pc"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_pc")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_pls"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_pls")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_gpr"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_gpr")
# abline(0, 1, col = "red")
# 
# par(old_par)


## -----------------------------------------------------------------------------
#| label: fig-mblcomparisons
#| fig-cap: "CEC prediction results for the different MBL configurations tested"
#| fig-align: center
#| fig-width: 8
#| fig-height: 8
#| echo: false
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2), pch = 16, mar = c(4, 4, 4, 4))

my_cols <- c(
  "mbl_cor" = "#D42B08CC",
  "mbl_pc"  = "#750E3380",
  "mbl_pls" = "#EFBF4780",
  "mbl_gpr" = "#5177A180"
)

# R2s
r2s <- drop(round(cor(test_cec, preds, use = "complete.obs")^2, 2))

# RMSEs
rmses <- round(colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5, 2)

plot_titles <- c(
  "mbl_cor" = "mbl_cor (LOCAL)",
  "mbl_pc"  = "mbl_pc",
  "mbl_pls" = "mbl_pls",
  "mbl_gpr" = "mbl_gpr"
)

p <- sapply(
  colnames(preds),
  FUN = function(label, y, yhats, cols, rsq, rmse, titles) {
    plot(
      x = y,
      y = yhats[, label],
      xlim = range(y, na.rm = TRUE),
      ylim = range(y, na.rm = TRUE),
      xlab = "CEC, meq/100g",
      ylab = "Predicted CEC, meq/100g",
      col = cols[label]
    )
    title(titles[label])
    title(paste0("\n\n\n RMSE: ", rmse[label], "; R²: ", rsq[label]), cex.main = 1)
    grid(col = "#80808080", lty = 1)
    abline(0, 1, col = "#FF1A0080")
  },
  y = test_cec,
  yhats = preds,
  cols = my_cols,
  rsq = r2s,
  rmse = rmses,
  titles = plot_titles
)

par(old_par)


## -----------------------------------------------------------------------------
#| label: yu-argument
#| eval: true
# Use Yu argument to validate the predictions
pc_pred_cec_yu <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_k(80),
  diss_method = diss_pca(scale = TRUE),
  diss_usage = "none",
  verbose = FALSE,
  control = mbl_control()
)

pc_pred_cec_yu


## -----------------------------------------------------------------------------
#| label: parallel-example
#| eval: false
# # Running the mbl function using multiple cores
# 
# # Execute with two cores, if available, ...
# n_cores <- 2
# 
# # ... if not then go with 1 core
# if (parallel::detectCores() < 2) {
#   n_cores <- 1
# }
# 
# # Set the number of cores
# library(doParallel)
# clust <- makeCluster(n_cores)
# registerDoParallel(clust)
# 
# # Alternatively:
# # library(doSNOW)
# # clust <- makeCluster(n_cores, type = "SOCK")
# # registerDoSNOW(clust)
# # getDoParWorkers()
# 
# pc_pred_cec <- mbl(
#   Xr = train_x[!is.na(train_cec), ],
#   Yr = train_cec[!is.na(train_cec)],
#   Xu = test_x,
#   neighbors = neighbors_k(seq(40, 100, by = 10)),
#   diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
#   diss_usage = "none",
#   control = mbl_control()
# )
# 
# # Go back to sequential processing
# registerDoSEQ()
# try(stopCluster(clust))
# 
# pc_pred_cec

library(ggplot2)
library(hexSticker)
library(sysfonts)
library(showtext)
library(ggtext)

font_add_google("Sora", "sora")
showtext_auto()
showtext_opts(dpi = 300)


n_points <- 200
r_min <- 1.5
r_max <- 5.5
n_turns <- 12  

radii <- seq(r_min, r_max, length.out = n_points)

# Uniform angular spacing (constant angle increment)
angles <- seq(0, n_turns * 2 * pi, length.out = n_points)

df <- data.frame(
  x = radii * cos(angles),
  y = radii * sin(angles),
  radius = radii,
  size = rep(1 * c(0.5, 1, 2, 1.5, 2.5, 1, 3, 2), length.out = n_points)
)
df$col_val <- df$radius

p <- ggplot(df, aes(x = x, y = y, size = size, colour = col_val)) +
  geom_point(alpha = 0.65, stroke = 0) +
  coord_fixed() +
  scale_colour_gradient(low = "#F59E0B", high = "#60A5FA")+
  scale_size_identity() +
  theme_void() +
  theme(
    panel.background = element_rect(fill = "transparent", colour = NA),
    plot.background = element_rect(fill = "transparent", colour = NA)
  ) +
  guides(size = "none", colour = "none")

p

# --- Sticker ---
library(ggtext)

s <- sticker(
  p,
  package = "",
  s_x = 1,
  s_y = 1,
  s_width = 1.8,
  s_height = 1.8,
  h_fill = "#020617",
  h_color = "#3B82F6",
  h_size = 1.4,
  filename = "resemble_hex.png",
  dpi = 300, 
  url = "https://github.com/l-ramirez-lopez/resemble", 
  u_color = "white", 
  u_size = 1.1
)

s <- s +
  geom_richtext(
    aes(x = 1, y = 1),
    label = "<span style='color:#F8FAFC;'>rese</span><span style='color:#F59E0B;'>mbl</span><span style='color:#F8FAFC;'>e</span>",
    family = "outfit",
    fontface = "bold",
    size = 9,
    fill = NA,
    label.color = NA
  )

ggsave("man/figures/logo.png", s, width = 43.94, height = 50.8, units = "mm", dpi = 300, bg = "transparent")

## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| message: false
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Missing values in the response are allowed
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

md <- dissimilarity(
  test_x,
  apply(test_x, 2, median),
  diss_method = diss_correlation(ws = 101, center = T, scale = T)
)

threshold <- 0.4


## -----------------------------------------------------------------------------
#| label: fig-diss-threshold
#| fig-cap: "Left: Dissimilarity of the test spectra to the median test spectrum. Samples above the threshold (dashed line and red) were excluded from the target set as a simple attempt to reduce spectral heterogeneity and potential non-linearities in the spectra-response relationship. Right: Spectra of the samples above (red, to be excluded) and below (blue, to be kept) the dissimilarity threshold."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 2, 1))

plot(
  seq_along(md$dissimilarity),
  md$dissimilarity, 
  ylim = c(0, 1), 
  col = "steelblue",
  pch = 16,
  cex = 1.5,
  xlab = "Sample index", 
  ylab = "Dissimilarity to median spectrum"
)
points(
  which(md$dissimilarity >= threshold),
  md$dissimilarity[md$dissimilarity >= threshold],
  col = "tomato",
  pch = 16,
  cex = 1.5
)
abline(h = threshold, col = "tomato", lty = 2, lwd = 2)
grid(lty = 1)

matplot(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity >= threshold, ]),
  col = "tomato",
  ylab = "Preprocessed spectra",
  xlab = "Wavelengths, nm",
  lty = 1,
  type = "l"
)
grid(lty = 1)
matlines(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity < threshold, ]),
  col = "steelblue",
  lty = 1,
  type = "l"
)

par(op)


## -----------------------------------------------------------------------------
keep <- md$dissimilarity < threshold

cat("Number of samples retained:", sum(keep))

test_x_test <- test_x[keep, ]
test_y_test <- test_y[keep]

# sample a very small subset to guide the search, 
# and use the rest for testing the final model
set.seed(1124)
ref_ind <- sample(sum(keep), 8)
test_x_ref <- test_x_test[ref_ind, ]
test_y_ref <- test_y_test[ref_ind]

test_x_test <- test_x_test[-ref_ind, ]
test_y_test <- test_y_test[-ref_ind]


## -----------------------------------------------------------------------------
set.seed(1124)

pls_model <- model(
  Xr = rbind(
    train_x[!is.na(train_y), ],
    test_x_ref[!is.na(test_y_ref), ]
  ), 
  Yr = c(
    train_y[!is.na(train_y)],
    test_y_ref[!is.na(test_y_ref)]
  ),
  fit_method = fit_pls(ncomp = 15, method = "simpls", scale = TRUE),
  control = model_control(validation_type = "lgo", number = 10)
)

best_ncomp <- which(pls_model$cv_results$optimal)

pred <- predict(pls_model, test_x_test, ncomp = best_ncomp)


## -----------------------------------------------------------------------------
reg_metrics <- function(obs, pred, na.rm = TRUE) {
  if (na.rm) {
    ok <- complete.cases(obs, pred)
    obs <- obs[ok]
    pred <- pred[ok]
  }
  c(RMSE = sqrt(mean((obs - pred)^2)), R2 = cor(obs, pred)^2)
}


## -----------------------------------------------------------------------------
reg_metrics(test_y_test, pred)


## -----------------------------------------------------------------------------
my_control <- gesearch_control(
  retain_by = "probability",
  tune = FALSE,
  number = 10L,
  p = 0.75
)


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false 
gs <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = test_y_ref, # the available target domain response (accepts NAs)
  k = 50, b = 100, retain = 0.95,
  target_size = 120,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "response"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = test_y_ref, # the available target domain response (accepts NAs)
#   k = 50, b = 100, retain = 0.95,
#   target_size = 120,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "response"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
#| eval: false
# plot(gs)


## -----------------------------------------------------------------------------
#| label: fig-evolution-gs
#| echo: false 
#| fig-cap: "Evolution of the three weakness metrics (response, reconstruction and dissimilarity) used in the search."
#| fig-align: "center"
#| fig-width: 8.5
#| fig-height: 5.5
plot(gs, main = NA)


## -----------------------------------------------------------------------------
# the internal leave-group out CV results for the data found
gs$validation_results[[1]]$results$train


## -----------------------------------------------------------------------------
# the CV on the target samples that drove the search (note that this is not a proper test set, as these samples were used to guide the search, but it can provide some insight into the performance of the final model on the target domain)
gs$validation_results[[1]]$results$test


## -----------------------------------------------------------------------------
best_ncomp <- which.min(gs$validation_results[[1]]$results$train$rmse_cv)


## -----------------------------------------------------------------------------
pred_gs <- predict(gs, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs[, best_ncomp])


## -----------------------------------------------------------------------------
#| label: fig-pred-comparison
#| fig-cap: "Comparison of observed versus predicted total carbon values. Left: model fitted using all available training samples. Right: model fitted using the subset of samples selected by gesearch()."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 3, 1))

rng <- range(pred, pred_gs, test_y_test, na.rm = TRUE)

plot(
  pred, test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with all training \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

plot(
  pred_gs[, best_ncomp], test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with gesearch-selected \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

par(op)


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false 
gs_nr <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs_nr <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = NULL, # NO available target domain response (accepts NAs)
#   k = 50, b = 80, retain = 0.98,
#   target_size = 160,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
best_ncomp_nr <- which.min(gs_nr$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr
pred_gs_nr <- predict(gs_nr, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr[, best_ncomp_nr])


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false 
gs_nr2 <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  Yu_lims = c(0, 5), # the response range in the target domain
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "range"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs_nr2 <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = NULL, # NO available target domain response (accepts NAs)
#   Yu_lims = c(0, 5), # the response range in the target domain
#   k = 50, b = 80, retain = 0.98,
#   target_size = 160,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "range"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
best_ncomp_nr2 <- which.min(gs_nr2$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr2
pred_gs_nr2 <- predict(gs_nr2, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr2[, best_ncomp_nr2])


## -----------------------------------------------------------------------------
#| label: parallel-gesearch
#| eval: false
# # Running gesearch() using multiple cores
# 
# # Execute with two cores, if available, ...
# n_cores <- 2
# 
# # ... if not then go with 1 core
# if (parallel::detectCores() < 2) {
#   n_cores <- 1
# }
# 
# # Set the number of cores
# library(doParallel)
# clust <- makeCluster(n_cores)
# registerDoParallel(clust)
# 
# # Alternatively:
# # library(doSNOW)
# # clust <- makeCluster(n_cores, type = "SOCK")
# # registerDoSNOW(clust)
# # getDoParWorkers()
# 
# my_control <- gesearch_control(
#   retain_by = "probability",
#   tune = FALSE,
#   number = 10L,
#   p = 0.75
# )
# 
# gs_p <- gesearch(
#   Xr = train_x[!is.na(train_y), ],
#   Yr = train_y[!is.na(train_y)],
#   Xu = test_x_ref,
#   Yu = test_y_ref,
#   k = 50,
#   b = 100,
#   retain = 0.98,
#   target_size = 80,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "response"),
#   control = my_control,
#   seed = 1124,
#   pchunks = 2L
# )
# 
# # Go back to sequential processing
# registerDoSEQ()
# try(stopCluster(clust))
# 
# gs_p

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| eval: true
# Retain components that individually explain at least 1% of variance
proj_var <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))
proj_var


## -----------------------------------------------------------------------------
#| eval: true
# Retain enough components to explain at least 99% of variance
proj_cumvar <- ortho_projection(train_x, ncomp = ncomp_by_cumvar(0.99))
proj_cumvar


## -----------------------------------------------------------------------------
#| eval: true
# Optimal component selection using side information
# using default max_ncomp which is 40
proj_opc <- ortho_projection(train_x, Yr = train_y, ncomp = ncomp_by_opc())
proj_opc


## -----------------------------------------------------------------------------
# this needs to be passed to the ncomp argument in ortho_projection()
ncomp_by_opc(max_ncomp = 15)


## -----------------------------------------------------------------------------
#| eval: false
# # Using ncomp_fixed()
# proj_fixed <- ortho_projection(train_x, ncomp = ncomp_fixed(10))
# 
# # Equivalent: passing an integer directly
# proj_fixed <- ortho_projection(train_x, ncomp = 10)


## -----------------------------------------------------------------------------
#| results: hide
# PCA with default component selection (ncomp_by_var(0.01))
pca_train <- ortho_projection(train_x, method = "pca")
pca_train


## -----------------------------------------------------------------------------
#| label: fig-pca-variance
#| fig-cap: "Individual contribution to the explained variance for each component (left) and cumulative variance explained by the principal components (right)."
#| fig-width: 7
#| fig-height: 3
plot(pca_train)


## -----------------------------------------------------------------------------
#| results: hide
# PCA using NIPALS
pca_nipals_train <- ortho_projection(train_x, method = "pca_nipals")


## -----------------------------------------------------------------------------
#| results: hide
# PLS using SIMPLS with 10 components
pls_train <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = 10
)
pls_train


## -----------------------------------------------------------------------------
#| results: hide
# PCA with optimal component selection
pca_opc <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40)
)
pca_opc


## -----------------------------------------------------------------------------
#| label: fig-pca-opc
#| fig-cap: "RMSD between observations and their nearest neighbors as a function of the number of principal components."
#| fig-width: 5
#| fig-height: 4
plot(pca_opc, col = "#F59E0B")


## -----------------------------------------------------------------------------
#| results: hide
# Fit PLS projection model
pls_model <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

# Project test data
pls_test_scores <- predict(pls_model, newdata = test_x)


## -----------------------------------------------------------------------------
#| label: fig-pc-proj
#| fig-width: 4.5
#| fig-height: 5
#| fig-cap: "Projection of the training observations in the first two PLS components (blue) and projection of the test observations onto the same latent space (red)."
plot(
  pls_model$scores[, 1:2], 
  col = rgb(0.231, 0.51, 0.965, 0.3), 
  pch = 16, cex = 1.5,
  xlab = "PLS1", ylab = "PLS2"
)
grid(lty = 1)
points(
  pls_test_scores[, 1:2], 
  col = rgb(0.961, 0.620, 0.043, 0.7), 
  pch = 16, cex = 1.5
)
legend(
  "topright",
  legend = c("Training projection", "Predicted projection"),
  col = c(
    rgb(0.231, 0.51, 0.965, 0.3),
    rgb(0.961, 0.620, 0.043, 0.7)
  ),
  pch = 16,
  pt.cex = 1.5,
  box.col = "black"
)


## -----------------------------------------------------------------------------
#| results: hide
# Project training and test data simultaneously
pca_both <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)


## -----------------------------------------------------------------------------
#| results: hide
#| eval: false
# train_y2 <- NIRsoil$Nt[NIRsoil$train == 1]
# train_y3 <- NIRsoil$CEC[NIRsoil$train == 1]
# 
# # PLS with multivariate side information
# pls_multi <- ortho_projection(
#   train_x,
#   Xu = test_x,
#   Yr = cbind(train_y, train_y2, train_y3),
#   method = "pca",
#   ncomp = ncomp_by_opc(40),
#   scale = TRUE
# )
# 
# plot(pls_multi)

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
diss_correlation()


## -----------------------------------------------------------------------------
diss_correlation(center = TRUE, scale = TRUE)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: pca-diss
#| eval: TRUE

# Default: variance-based component selection (ncomp_by_var(0.01))
d_pca <- dissimilarity(train_x, diss_method = diss_pca())

# With OPC-based component selection (requires Yr)
d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30))
)

# Between training and test sets
d_pca_tr_ts <- dissimilarity(
 train_x,
 Xu = test_x,
 Yr = train_y,
 diss_method = diss_pca(
   ncomp = ncomp_by_opc(30),
   return_projection = TRUE
 )
)

d_pca_tr_ts


## -----------------------------------------------------------------------------
# Example of the first 5 rows and columns of the dissimilarity 
# matrix between training and test sets
# Rows: training observations; 
# Columns: test observations
d_pca_tr_ts$dissimilarity[1:5, 1:5]


## -----------------------------------------------------------------------------
d_pca_tr_ts$projection


## ----pls-diss, eval = FALSE---------------------------------------------------
# # Default: OPC-based component selection
# d_pls <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   Yr = train_y,
#   diss_method = diss_pls()
# )
# 
# # Fixed number of components
# d_pls_fixed <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   Yr = train_y,
#   diss_method = diss_pls(ncomp = 15)
# )


## ----cor-diss, eval = FALSE---------------------------------------------------
# # Standard correlation dissimilarity
# d_cor <- dissimilarity(train_x, diss_method = diss_correlation())
# d_cor


## -----------------------------------------------------------------------------
correlation_baser <- 0.5 * (1 - cor(t(train_x)))
corrrelation_resemble <- dissimilarity(
  train_x, diss_method = diss_correlation(center = FALSE)
)
# the maximum discreepancy between the two implementations should 
# be very close to zero
max(abs(corrrelation_resemble$dissimilarity - correlation_baser))


## -----------------------------------------------------------------------------
#| eval: false
# # Compare computational speed: resemble vs base R correlation dissimilarity
# n_iter <- 50
# 
# # Transpose once before timing (fair comparison)
# train_x_t <- t(train_x)
# 
# # resemble implementation
# time_resemble <- system.time({
#   for (i in seq_len(n_iter)) {
#     cor_resemble <- dissimilarity(train_x, diss_method = diss_correlation(center = FALSE))
#   }
# })
# 
# # base R implementation
# time_base <- system.time({
#   for (i in seq_len(n_iter)) {
#     cor_base <- 0.5 * (1 - cor(train_x_t))
#   }
# })
# 
# # Results
# data.frame(
#   method = c("resemble", "base R"),
#   elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
#   per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
# )


## ----cor-diss-mw, eval = FALSE------------------------------------------------
# # Moving window correlation (window size must be odd)
# d_cor_mw <- dissimilarity(
#   train_x,
#   Xu = test_x,
#   diss_method = diss_correlation(ws = 51)
# )
# d_cor_mw


## ----omp-threads-use, eval = FALSE--------------------------------------------
# # Use 4 threads
# Sys.setenv(OMP_NUM_THREADS = 4)
# 
# # Or use all available cores
# Sys.setenv(OMP_NUM_THREADS = parallel::detectCores())


## ----omp-threads-restrict, eval = FALSE---------------------------------------
# # Limit OpenMP threads
# Sys.setenv(OMP_NUM_THREADS = 1)
# 
# # If using OpenBLAS
# Sys.setenv(OPENBLAS_NUM_THREADS = 1)
# 
# # If using MKL
# Sys.setenv(MKL_NUM_THREADS = 1)


## ----blas-threads, eval = FALSE-----------------------------------------------
# RhpcBLASctl::blas_set_num_threads(1)
# RhpcBLASctl::omp_set_num_threads(1)


## ----euclid-diss, eval = FALSE------------------------------------------------
# d_euclid <- dissimilarity(train_x, Xu = test_x, diss_method = diss_euclidean())
# d_euclid


## -----------------------------------------------------------------------------
ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
ed_baser <- as.matrix(dist(train_x, method = "euclidean"))

# scaling the base R Euclidean distance by the number of variables:
ed_baser_scaled <- sqrt((ed_baser^2) / ncol(train_x))

# differences between the two implementations
max(abs(ed_baser_scaled - ed_resemble$dissimilarity))


## -----------------------------------------------------------------------------
#| label: euclid-speed
#| eval: false
# # Compare computational speed: resemble vs base R
# n_iter <- 50
# 
# # resemble implementation
# time_resemble <- system.time({
#   for (i in seq_len(n_iter)) {
#     ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
#   }
# })
# 
# # base R implementation
# time_base <- system.time({
#   for (i in seq_len(n_iter)) {
#     ed_base <- as.matrix(dist(train_x, method = "euclidean"))
#   }
# })
# 
# # Results
# data.frame(
#   method = c("resemble", "base R"),
#   elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
#   per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
# )


## ----cosine-diss, eval = FALSE------------------------------------------------
# d_cosine <- dissimilarity(train_x, Xu = test_x, diss_method = diss_cosine())
# d_cosine


## -----------------------------------------------------------------------------
# Compute dissimilarity matrices using different methods
d_pca_var <- dissimilarity(train_x, diss_method = diss_pca(scale = TRUE))

d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30), scale = TRUE)
)

d_pls_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE)
)

d_cor <- dissimilarity(
  train_x,
  diss_method = diss_correlation(ws = 51, scale = TRUE)
)

d_euclid <- dissimilarity(
  train_x,
  diss_method = diss_euclidean(scale = TRUE)
)

d_cosine <- dissimilarity(
  train_x,
  diss_method = diss_cosine(scale = TRUE)
)


## -----------------------------------------------------------------------------
# Evaluate each method using Ciso as side information
side_info <- as.matrix(train_y)

eval_results <- list(
  "PCA (var)" = diss_evaluate(
    d_pca_var$dissimilarity, side_info
  ),
  "PCA (opc)" = diss_evaluate(
    d_pca_opc$dissimilarity, side_info
  ),
  "PLS (opc)" = diss_evaluate(
    d_pls_opc$dissimilarity, side_info
  ),
  correlation = diss_evaluate(
    d_cor$dissimilarity, side_info
  ),
  Euclidean = diss_evaluate(
    d_euclid$dissimilarity, side_info
  ),
  cosine = diss_evaluate(
    d_cosine$dissimilarity, side_info
  )
)

# Extract RMSD and correlation for each method
comparison <- do.call(rbind, lapply(names(eval_results), function(nm) {
  data.frame(
    method = nm,
    rmsd   = eval_results[[nm]]$eval[, "rmsd"],
    r      = eval_results[[nm]]$eval[, "r"]
  )
}))
comparison


## -----------------------------------------------------------------------------
#| label: diss-eval-plot
#| fig-cap: "Comparison of dissimilarity methods based on first nearest-neighbor evaluation using `Ciso` as side information. In each panel, the x-axis shows the `Ciso` value of the nearest neighbor identified from the spectral dissimilarity matrix, and the y-axis shows the observed `Ciso` value of the corresponding sample. The dashed line represents perfect agreement (y = x)."
#| fig-align: "center"
#| fig-width: 8
#| fig-height: 6
#| echo: true
#| fig-show: hold
blue  <- "#3B82F6"
amber <- "#F59E0B"
slate_grid <- "#33415540"

old_par <- par(mfrow = c(2, 3), mar = c(4, 4, 3, 1))

for (nm in names(eval_results)) {
  xy <- eval_results[[nm]]$first_nn
  plot(
    xy[, 2], xy[, 1],
    xlab = "Ciso (1-NN)",
    ylab = "Ciso",
    col = adjustcolor(blue, alpha.f = 0.5),
    pch = 16,
    main = nm
  )
  grid(lty = 1, col = slate_grid)
  abline(0, 1, col = amber, lwd = 2)
}

par(old_par)

## -----------------------------------------------------------------------------
#| message: false
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
library(resemble)


## -----------------------------------------------------------------------------
#| message: false
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Missing values in the response are allowed
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

md <- dissimilarity(
  test_x,
  apply(test_x, 2, median),
  diss_method = diss_correlation(ws = 101, center = T, scale = T)
)

threshold <- 0.4


## -----------------------------------------------------------------------------
#| label: fig-diss-threshold
#| fig-cap: "Left: Dissimilarity of the test spectra to the median test spectrum. Samples above the threshold (dashed line and red) were excluded from the target set as a simple attempt to reduce spectral heterogeneity and potential non-linearities in the spectra-response relationship. Right: Spectra of the samples above (red, to be excluded) and below (blue, to be kept) the dissimilarity threshold."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 2, 1))

plot(
  seq_along(md$dissimilarity),
  md$dissimilarity, 
  ylim = c(0, 1), 
  col = "steelblue",
  pch = 16,
  cex = 1.5,
  xlab = "Sample index", 
  ylab = "Dissimilarity to median spectrum"
)
points(
  which(md$dissimilarity >= threshold),
  md$dissimilarity[md$dissimilarity >= threshold],
  col = "tomato",
  pch = 16,
  cex = 1.5
)
abline(h = threshold, col = "tomato", lty = 2, lwd = 2)
grid(lty = 1)

matplot(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity >= threshold, ]),
  col = "tomato",
  ylab = "Preprocessed spectra",
  xlab = "Wavelengths, nm",
  lty = 1,
  type = "l"
)
grid(lty = 1)
matlines(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity < threshold, ]),
  col = "steelblue",
  lty = 1,
  type = "l"
)

par(op)


## -----------------------------------------------------------------------------
keep <- md$dissimilarity < threshold

cat("Number of samples retained:", sum(keep))

test_x_test <- test_x[keep, ]
test_y_test <- test_y[keep]

# sample a very small subset to guide the search, 
# and use the rest for testing the final model
set.seed(1124)
ref_ind <- sample(sum(keep), 8)
test_x_ref <- test_x_test[ref_ind, ]
test_y_ref <- test_y_test[ref_ind]

test_x_test <- test_x_test[-ref_ind, ]
test_y_test <- test_y_test[-ref_ind]


## -----------------------------------------------------------------------------
set.seed(1124)

pls_model <- model(
  Xr = rbind(
    train_x[!is.na(train_y), ],
    test_x_ref[!is.na(test_y_ref), ]
  ), 
  Yr = c(
    train_y[!is.na(train_y)],
    test_y_ref[!is.na(test_y_ref)]
  ),
  fit_method = fit_pls(ncomp = 15, method = "simpls", scale = TRUE),
  control = model_control(validation_type = "lgo", number = 10)
)

best_ncomp <- which(pls_model$cv_results$optimal)

pred <- predict(pls_model, test_x_test, ncomp = best_ncomp)


## -----------------------------------------------------------------------------
reg_metrics <- function(obs, pred, na.rm = TRUE) {
  if (na.rm) {
    ok <- complete.cases(obs, pred)
    obs <- obs[ok]
    pred <- pred[ok]
  }
  c(RMSE = sqrt(mean((obs - pred)^2)), R2 = cor(obs, pred)^2)
}


## -----------------------------------------------------------------------------
reg_metrics(test_y_test, pred)


## -----------------------------------------------------------------------------
my_control <- gesearch_control(
  retain_by = "probability",
  tune = FALSE,
  number = 10L,
  p = 0.75
)


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false
gs <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = test_y_ref, # the available target domain response (accepts NAs)
  k = 50, b = 20, retain = 0.95,
  target_size = 120,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "response"),
  control = my_control,
  seed = 1124,
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = test_y_ref, # the available target domain response (accepts NAs)
#   k = 50, b = 100, retain = 0.95,
#   target_size = 120,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "response"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
#| eval: false
# plot(gs)


## -----------------------------------------------------------------------------
#| label: fig-evolution-gs
#| echo: false 
#| fig-cap: "Evolution of the three weakness metrics (response, reconstruction and dissimilarity) used in the search."
#| fig-align: "center"
#| fig-width: 8.5
#| fig-height: 5.5
plot(gs, main = NA)


## -----------------------------------------------------------------------------
# the internal leave-group out CV results for the data found
gs$validation_results[[1]]$results$train


## -----------------------------------------------------------------------------
# the CV on the target samples that drove the search (note that this is not a proper test set, as these samples were used to guide the search, but it can provide some insight into the performance of the final model on the target domain)
gs$validation_results[[1]]$results$test


## -----------------------------------------------------------------------------
best_ncomp <- which.min(gs$validation_results[[1]]$results$train$rmse_cv)


## -----------------------------------------------------------------------------
pred_gs <- predict(gs, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs[, best_ncomp])


## -----------------------------------------------------------------------------
#| label: fig-pred-comparison
#| fig-cap: "Comparison of observed versus predicted total carbon values. Left: model fitted using all available training samples. Right: model fitted using the subset of samples selected by gesearch()."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 3, 1))

rng <- range(pred, pred_gs, test_y_test, na.rm = TRUE)

plot(
  pred, test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with all training \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

plot(
  pred_gs[, best_ncomp], test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with gesearch-selected \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

par(op)


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false
gs_nr <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  k = 50, b = 15, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity"),
  control = my_control,
  seed = 1124,
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs_nr <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = NULL, # NO available target domain response (accepts NAs)
#   k = 50, b = 80, retain = 0.98,
#   target_size = 160,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
best_ncomp_nr <- which.min(gs_nr$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr
pred_gs_nr <- predict(gs_nr, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr[, best_ncomp_nr])


## -----------------------------------------------------------------------------
#| eval: true
#| echo: false
gs_nr2 <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  Yu_lims = c(0, 5), # the response range in the target domain
  k = 50, b = 15, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "range"),
  control = my_control,
  seed = 1124,
  verbose = FALSE
)


## -----------------------------------------------------------------------------
#| eval: false
# gs_nr2 <- gesearch(
#   Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
#   Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
#   Xu = test_x_ref, # the available target domain spectra
#   Yu = NULL, # NO available target domain response (accepts NAs)
#   Yu_lims = c(0, 5), # the response range in the target domain
#   k = 50, b = 80, retain = 0.98,
#   target_size = 160,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "range"),
#   control = my_control,
#   seed = 1124
# )


## -----------------------------------------------------------------------------
best_ncomp_nr2 <- which.min(gs_nr2$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr2
pred_gs_nr2 <- predict(gs_nr2, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr2[, best_ncomp_nr2])


## -----------------------------------------------------------------------------
#| label: parallel-gesearch
#| eval: false
# # Running gesearch() using multiple cores
# 
# # Execute with two cores, if available, ...
# n_cores <- 2
# 
# # ... if not then go with 1 core
# if (parallel::detectCores() < 2) {
#   n_cores <- 1
# }
# 
# # Set the number of cores
# library(doParallel)
# clust <- makeCluster(n_cores)
# registerDoParallel(clust)
# 
# # Alternatively:
# # library(doSNOW)
# # clust <- makeCluster(n_cores, type = "SOCK")
# # registerDoSNOW(clust)
# # getDoParWorkers()
# 
# my_control <- gesearch_control(
#   retain_by = "probability",
#   tune = FALSE,
#   number = 10L,
#   p = 0.75
# )
# 
# gs_p <- gesearch(
#   Xr = train_x[!is.na(train_y), ],
#   Yr = train_y[!is.na(train_y)],
#   Xu = test_x_ref,
#   Yu = test_y_ref,
#   k = 50,
#   b = 100,
#   retain = 0.98,
#   target_size = 80,
#   fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
#   optimization = c("reconstruction", "similarity", "response"),
#   control = my_control,
#   seed = 1124,
#   pchunks = 2L
# )
# 
# # Go back to sequential processing
# registerDoSEQ()
# try(stopCluster(clust))
# 
# gs_p

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## ----neighbors-k-example, eval = FALSE----------------------------------------
# neighbors_k(k = 50)
# neighbors_k(k = c(40, 60, 80, 100))


## ----neighbors-diss-example, eval = FALSE-------------------------------------
# neighbors_diss(threshold = 0.3)
# neighbors_diss(threshold = c(0.1, 0.2, 0.3), k_min = 10, k_max = 150)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
set.seed(8011)
rnd_idc <- sample(nrow(test_x), 3)

k_fixed <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2, return_projection = TRUE),
  neighbors = neighbors_k(30)
)


## -----------------------------------------------------------------------------
k_diss <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2),
  neighbors = neighbors_diss(threshold = 0.25)
)


## -----------------------------------------------------------------------------
test_scores_indices <- grep("^Xu_", rownames(k_fixed$projection$scores))


## -----------------------------------------------------------------------------
#| label: fig-neighbors
#| fig-cap: "Nearest neighbors identified using the same dissimilarity metric but retained using two different selection methods."
#| fig-width: 8.5
#| fig-height: 4.5

old_par <- par(no.readonly = TRUE)
on.exit(par(old_par), add = TRUE)

par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Fixed k-nearest neighbors"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_fixed$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Neighbors selected by a \nthreshold dissimilarity"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_diss$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)


## ----knn-pca, eval = FALSE----------------------------------------------------
# # matrix of neighbors
# k_fixed$neighbors
# 
# # matrix of neighbor distances (dissimilarity scores)
# k_fixed$neighbors_diss
# 
# # the index (in the training set) of the first two closest neighbors found in
# # training for the first observation in testing:
# k_fixed$neighbors[1:2, 1, drop = FALSE]
# 
# # the distances of the two closest neighbors found in
# # training for the first observation in testing:
# k_fixed$neighbors_diss[1:2, 1, drop = FALSE]
# 
# # the indices in training that fall in any of the
# # neighborhoods of testing
# k_fixed$unique_neighbors


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-other-methods
# using PC dissimilarity with optimal selection of components
knn_opc <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(),
    scale = TRUE,
    return_projection = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using PLS dissimilarity with optimal selection of components
knn_pls <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pls(
    ncomp = ncomp_by_opc(),
    scale = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using correlation dissimilarity
knn_cor <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(),
  neighbors = neighbors_k(50)
)

# using moving window correlation dissimilarity
knn_mw <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(ws = 51),
  neighbors = neighbors_k(50)
)


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-threshold-example
#| results: hide
# a dissimilarity threshold
d_th <- 1

# the minimum number of observations required in each neighborhood
k_min <- 20

# the maximum number of observations allowed in each neighborhood
k_max <- 300

dnn_pca <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(scale = TRUE),
  neighbors = neighbors_diss(threshold = d_th, k_min = k_min, k_max = k_max)
)

# matrix of neighbors. The minimum number of indices is 20 (given by k_min)
# and the maximum number of indices is 300 (given by k_max).
# NAs indicate "not a neighbor"
dnn_pca$neighbors

# this reports how many neighbors were found for each observation in 
# testing using the input distance threshold (column n_k) and how 
# many were finally selected (column final_n_k)
dnn_pca$k_diss_info

# matrix of neighbor distances
dnn_pca$neighbors_diss

# the indices in training that fall in any of the 
# neighborhoods of testing
dnn_pca$unique_neighbors


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: fig-knn-hist
#| fig-cap: "Histogram of the final neighborhood sizes after applying the dissimilarity threshold and the minimum and maximum neighborhood size constraints."
#| fig-width: 5
#| fig-height: 4
hist(
  dnn_pca$k_diss_info$final_n_k,
  breaks = 20,
  xlab = "Final neighborhood size",
  main = "",
  col = "dodgerblue"
)


## -----------------------------------------------------------------------------
#| eval: TRUE
#| label: knn-spike-example
# the indices of the observations that we want to "invite" to every neighborhood
forced_guests <- c(1, 5, 8, 9)

# using PC dissimilarity with optimal selection of components
knn_spiked <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(20)
  ),
  Yr = train_y,
  neighbors = neighbors_k(50),
  spike = forced_guests
)

# check the first 8 neighbors found in training for the 
# first 2 observations in testing
knn_spiked$neighbors[1:8, 1:2]

## -----------------------------------------------------------------------------
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
#| label: fit-methods
# PLS with 15 components
fit_pls(ncomp = 15)

# SIMPLS with scaling
fit_pls(ncomp = 15, method = "simpls", scale = TRUE)

# mPLS with scaling
fit_pls(ncomp = 15, method = "mpls")

# GPR with default noise variance
fit_gpr()


## -----------------------------------------------------------------------------
#| label: data-prep
#| message: false
library(prospectr)

wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split data
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

# Remove missing values
ok_train <- !is.na(train_y)
ok_test  <- !is.na(test_y)
train_x <- train_x[ok_train, ]
train_y <- train_y[ok_train]
test_x  <- test_x[ok_test, ]
test_y  <- test_y[ok_test]


## -----------------------------------------------------------------------------
#| label: fit-pls

set.seed(1124) # guarantee same CV splits for all methods
pls_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_pls(ncomp = 12, method = "mpls", scale = TRUE),
 control = model_control(validation_type = "lgo", number = 10)
)

pls_mod


## -----------------------------------------------------------------------------
#| label: fit-gpr
set.seed(1124) # guarantee same CV splits for all methods
gpr_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_gpr(noise_variance = 0.5),
 control = model_control(validation_type = "lgo", number = 10)
)

gpr_mod


## -----------------------------------------------------------------------------
#| label: predict
# Predict using optimal number of components (from CV)
pls_pred <- predict(pls_mod, newdata = test_x)

# Predict with GPR
gpr_pred <- predict(gpr_mod, newdata = test_x)

# Compare predictions
data.frame(
  observed = test_y,
  pls = pls_pred[, which(pls_mod$cv_results$optimal)],
  gpr = as.vector(gpr_pred)
) |> head(10)


## -----------------------------------------------------------------------------
#| label: validation
# Function to compute stats
eval_pred <- function(obs, pred) {
  data.frame(
    rmse = sqrt(mean((obs - pred)^2)),
    r2 = cor(obs, pred)^2
  )
}


## -----------------------------------------------------------------------------
## PLS evaluation
eval_pred(test_y, pls_pred[, which(pls_mod$cv_results$optimal)])


## -----------------------------------------------------------------------------
## GPR evaluation
eval_pred(test_y, gpr_pred)

## -----------------------------------------------------------------------------
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)


## -----------------------------------------------------------------------------
#| message: false
#| echo: false
library(resemble)


## -----------------------------------------------------------------------------
#| label: fit-methods
#| eval: true

# Creates an object with instructions to build PLS models
my_pls <- fit_pls(ncomp = 15)
my_pls

# Creates an object with instructions to build WAPLS models
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 20)
my_wapls

# Creates an object with instructions to build GPR models
my_gpr <- fit_gpr()
my_gpr


## -----------------------------------------------------------------------------
#| label: validation-control
#| eval: true
# Create an object with instructions to conduct both validation types
# "NNv" and "local_cv"
two_val_control <- mbl_control(
  validation_type = c("NNv", "local_cv"),
  number = 10,
  p = 0.75
)


## -----------------------------------------------------------------------------
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: mbl-local
#| eval: true
#| results: hide
# Define the neighborhood sizes to test
my_ks <- seq(80, 160, by = 40)

# Define how to use the dissimilarity information (ignore it)
ignore_diss <- "none"

# Define the regression method to be used at each neighborhood
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 25, scale = TRUE)

# For the moment use only "NNv" validation (faster)
nnv_val_control <- mbl_control()

# Predict Total Carbon
# (remove missing values)
local_ciso <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_k(my_ks),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  gh = TRUE,
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: fig-localresultsciso
#| fig-cap: "MBL results for Total Carbon predictions using the LOCAL algorithm. NNv: nearest-neighbor cross-validation."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4
plot(local_ciso, main = "")
local_ciso


## -----------------------------------------------------------------------------
#| label: bestk
#| eval: true
#| echo: false
#| results: hide
bestk <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bestk <- local_ciso$validation_results$nearest_neighbor_validation$k[bestk]


## -----------------------------------------------------------------------------
#| label: get-predictions
#| eval: true
#| results: hide
#| fig-show: hide
bki <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bk <- local_ciso$validation_results$nearest_neighbor_validation$k[bki]

# All the prediction results are stored in:
local_ciso$results

# The get_predictions function makes easier to retrieve the
# predictions from the previous object
ciso_hat <- as.matrix(get_predictions(local_ciso))[, bki]


## -----------------------------------------------------------------------------
#| label: fig-plot-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation."
#| eval: true
#| fig-width: 5
#| fig-height: 5
# Plot predicted vs reference
rng <- range(ciso_hat, test_y, na.rm = TRUE)
plot(ciso_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a fixed k", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid(lty = 1)
abline(0, 1, col = "red")


## -----------------------------------------------------------------------------
#| eval: true
# Prediction RMSE:
sqrt(mean((ciso_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_hat, test_y, use = "complete.obs")^2


## -----------------------------------------------------------------------------
#| label: mbl-diss-threshold
#| eval: true
#| results: hide
#| fig-show: hide
# Create a vector of dissimilarity thresholds to evaluate
# since the correlation dissimilarity will be used
# these thresholds need to be > 0 and <= 1
dths <- seq(0.025, 0.15, by = 0.025)

# Indicate the minimum and maximum sizes allowed for the neighborhood
k_min <- 30
k_max <- 150

local_ciso_diss <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dths, k_min = k_min, k_max = k_max),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: plot-diss-results
#| eval: false
# plot(local_ciso_diss)


## -----------------------------------------------------------------------------
#| label: bestd
#| eval: true
#| echo: false
#| results: hide
bestd <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bestd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bestd]


## -----------------------------------------------------------------------------
#| label: show-diss-results
#| eval: true
local_ciso_diss


## -----------------------------------------------------------------------------
#| label: diss-predictions
#| eval: true
#| results: hide
#| fig-show: hide
# Best distance threshold
bdi <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bdi]

# Predictions for the best distance
ciso_diss_hat <- as.matrix(get_predictions(local_ciso_diss))[, bdi]


## -----------------------------------------------------------------------------
#| label: fig-diss-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation and distance thresholds."
#| eval: true
# Plot predicted vs reference
plot(ciso_diss_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a distance threshold \nfor neighbor retrieval", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid()
abline(0, 1, col = "red")

## -----------------------------------------------------------------------------
# RMSE
sqrt(mean((ciso_diss_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_diss_hat, test_y, use = "complete.obs")^2


## -----------------------------------------------------------------------------
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_cec <- NIRsoil$CEC[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_cec  <- NIRsoil$CEC[NIRsoil$train == 0]


## -----------------------------------------------------------------------------
#| label: cec-examples
#| results: hide
#| eval: true
#| fig-show: hide
# Define the WAPLS fitting method
my_wapls <- fit_wapls(min_ncomp = 2, max_ncomp = 25, scale = FALSE)

# mbl_cor: LOCAL algorithm with correlation dissimilarity
dth_cor <- seq(0.01, 0.3, by = 0.08)
mbl_cor <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_cor, k_min = 80, k_max = 150),
  diss_method = diss_correlation(),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pc: PCA dissimilarity with dissimilarity matrix as predictors
dth_pc <- seq(0.05, 1, by = 0.4)
mbl_pc <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pls: PLS dissimilarity without dissimilarity predictors
mbl_pls <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_gpr: Gaussian process regression with PCA dissimilarity as predictors
mbl_gpr <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = fit_gpr(),
  control = nnv_val_control
)


## -----------------------------------------------------------------------------
#| label: collect-predictions
#| eval: true
#| fig-show: hide
# Get the indices of the best results according to
# nearest neighbor validation statistics
c_val_name <- "validation_results"
c_nn_val_name <- "nearest_neighbor_validation"

bi_cor <- which.min(mbl_cor[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pc  <- which.min(mbl_pc[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pls <- which.min(mbl_pls[[c_val_name]][[c_nn_val_name]]$rmse)
bi_gpr <- which.min(mbl_gpr[[c_val_name]][[c_nn_val_name]]$rmse)

preds <- cbind(
  get_predictions(mbl_cor)[, bi_cor],
  get_predictions(mbl_pc)[, bi_pc],
  get_predictions(mbl_pls)[, bi_pls],
  get_predictions(mbl_gpr)[, bi_gpr]
)

colnames(preds) <- c("mbl_cor", "mbl_pc", "mbl_pls", "mbl_gpr")
preds <- as.matrix(preds)

# R2s
cor(test_cec, preds, use = "complete.obs")^2

# RMSEs
colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5


## -----------------------------------------------------------------------------
#| label: cec-plots-code
#| eval: false
#| fig-show: hide
# old_par <- par("mfrow", "mar")
# par(mfrow = c(2, 2))
# 
# plot(test_cec, preds[, "mbl_cor"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_cor (LOCAL)")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_pc"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_pc")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_pls"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_pls")
# abline(0, 1, col = "red")
# 
# plot(test_cec, preds[, "mbl_gpr"],
#      xlab = "CEC, meq/100g",
#      ylab = "Predicted CEC, meq/100g",
#      main = "mbl_gpr")
# abline(0, 1, col = "red")
# 
# par(old_par)


## -----------------------------------------------------------------------------
#| label: fig-mblcomparisons
#| fig-cap: "CEC prediction results for the different MBL configurations tested"
#| fig-align: center
#| fig-width: 8
#| fig-height: 8
#| echo: false
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2), pch = 16, mar = c(4, 4, 4, 4))

my_cols <- c(
  "mbl_cor" = "#D42B08CC",
  "mbl_pc"  = "#750E3380",
  "mbl_pls" = "#EFBF4780",
  "mbl_gpr" = "#5177A180"
)

# R2s
r2s <- drop(round(cor(test_cec, preds, use = "complete.obs")^2, 2))

# RMSEs
rmses <- round(colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5, 2)

plot_titles <- c(
  "mbl_cor" = "mbl_cor (LOCAL)",
  "mbl_pc"  = "mbl_pc",
  "mbl_pls" = "mbl_pls",
  "mbl_gpr" = "mbl_gpr"
)

p <- sapply(
  colnames(preds),
  FUN = function(label, y, yhats, cols, rsq, rmse, titles) {
    plot(
      x = y,
      y = yhats[, label],
      xlim = range(y, na.rm = TRUE),
      ylim = range(y, na.rm = TRUE),
      xlab = "CEC, meq/100g",
      ylab = "Predicted CEC, meq/100g",
      col = cols[label]
    )
    title(titles[label])
    title(paste0("\n\n\n RMSE: ", rmse[label], "; R²: ", rsq[label]), cex.main = 1)
    grid(col = "#80808080", lty = 1)
    abline(0, 1, col = "#FF1A0080")
  },
  y = test_cec,
  yhats = preds,
  cols = my_cols,
  rsq = r2s,
  rmse = rmses,
  titles = plot_titles
)

par(old_par)


## -----------------------------------------------------------------------------
#| label: yu-argument
#| eval: true
# Use Yu argument to validate the predictions
pc_pred_cec_yu <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_k(80),
  diss_method = diss_pca(scale = TRUE),
  diss_usage = "none",
  verbose = FALSE,
  control = mbl_control()
)

pc_pred_cec_yu


## -----------------------------------------------------------------------------
#| label: parallel-example
#| eval: false
# # Running the mbl function using multiple cores
# 
# # Execute with two cores, if available, ...
# n_cores <- 2
# 
# # ... if not then go with 1 core
# if (parallel::detectCores() < 2) {
#   n_cores <- 1
# }
# 
# # Set the number of cores
# library(doParallel)
# clust <- makeCluster(n_cores)
# registerDoParallel(clust)
# 
# # Alternatively:
# # library(doSNOW)
# # clust <- makeCluster(n_cores, type = "SOCK")
# # registerDoSNOW(clust)
# # getDoParWorkers()
# 
# pc_pred_cec <- mbl(
#   Xr = train_x[!is.na(train_cec), ],
#   Yr = train_cec[!is.na(train_cec)],
#   Xu = test_x,
#   neighbors = neighbors_k(seq(40, 100, by = 10)),
#   diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
#   diss_usage = "none",
#   control = mbl_control()
# )
# 
# # Go back to sequential processing
# registerDoSEQ()
# try(stopCluster(clust))
# 
# pc_pred_cec

## -----------------------------------------------------------------------------
#| message: false
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
library(resemble)
library(prospectr)


## -----------------------------------------------------------------------------
#| label: data-prep
#| message: false
library(resemble)
library(prospectr)

data(NIRsoil)

# Wavelengths
wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split into training and test sets
# Note: missing values in the response are allowed in liblex
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

cat("Training set:", nrow(train_x), "observations\n")
cat("Test set:", nrow(test_x), "observations\n")


## -----------------------------------------------------------------------------
#| label: neighbors-example1
# Fixed neighborhood sizes to evaluate
neighbors_k(k = seq(40, 120, by = 20))


## -----------------------------------------------------------------------------
#| label: neighbors-example2
# Threshold-based selection with bounds
neighbors_diss(
  threshold = seq(0.05, 0.5, length.out = 10), 
  k_min = 40, 
  k_max = 150
)


## -----------------------------------------------------------------------------
#| label: diss-example
# Moving-window correlation dissimilarity (recommended for spectra)
diss_correlation(ws = 37, scale = TRUE)

# PCA-based Mahalanobis distance with optimized components
diss_pca()


## -----------------------------------------------------------------------------
#| label: fit-method-example
# waPLS with modified PLS algorithm and scaling
fit_wapls(
  min_ncomp = 3, 
  max_ncomp = 15, 
  scale = TRUE, 
  method = "mpls"
)


## -----------------------------------------------------------------------------
#| label: control-example
#| eval: false
# # Build library with hyperparameter tuning
# liblex_control(mode = "build", tune = TRUE)


## -----------------------------------------------------------------------------
#| eval: false
# ciso_lib_k <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 80, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_k


## -----------------------------------------------------------------------------
#| label: liblex-fixed-k
#| echo: false
#| message: false
#| warning: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_k


## -----------------------------------------------------------------------------
#| label: fig-liblex
#| fig-cap: "Top: Best model obtained for each neighborhood size (based on the RMSE). Bottom: Centroids of the neighborhoods, usled later in prediction to selet the models to be used."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 6.5
plot(ciso_lib_k)


## -----------------------------------------------------------------------------
#| label: liblex-threshold
#| echo: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_thr


## -----------------------------------------------------------------------------
#| eval: false
# ciso_lib_thr <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_diss(
#     threshold = seq(0.05, 0.5, length.out = 5),
#     k_min = 40,
#     k_max = 150
#   ),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_thr


## -----------------------------------------------------------------------------
#| label: fig-centroids
#| fig-cap: "Neighborhood centroids (blue) compared to test spectra (red)"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center
wavs_pr <- as.numeric(colnames(ciso_lib_k$scaling$local_x_center))

matplot(
  wavs_pr,
  t(test_x),
  col = rgb(1, 0, 0, 0.3),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance"
)

matlines(
  wavs_pr,
  t(ciso_lib_k$scaling$local_x_center),
  col = rgb(0, 0, 1, 0.3),
  lty = 1
)

grid(lty = 1)

legend(
  "topright",
  legend = c("Samples to predict", "Neighborhood centroids"),
  col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
  lty = 1,
  lwd = 2,
  bty = "n"
)


## -----------------------------------------------------------------------------
#| label: fig-regression-coefficients
#| fig-cap: "Regression coefficients (top) and intercept distribution (bottom) of the local experts"
#| fig-width: 8
#| fig-height: 7
#| fig-align: center

par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Regression coefficients across wavelengths
models <- ciso_lib_k$coefficients

matplot(
  x = wavs_pr,
  y = t(models$B),
  type = "l",
  lty = 1,
  col = rgb(0.23, 0.51, 0.96, 0.15),
  xlab = "Wavelength (nm)",
  ylab = "Regression coefficient",
  main = paste0("Regression coefficients (", nrow(models$B), " experts)")
)

# Add mean coefficient profile
lines(wavs_pr, colMeans(models$B), col = "red", lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Individual experts", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96, 0.6), "red"),
  lty = 1,
  lwd = c(1, 2),
  bty = "n"
)

# Distribution of intercepts
plot(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96),
  lwd = 2,
  xlab = "Intercept value",
  ylab = "Density",
  main = "Distribution of intercepts"
)
polygon(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96, 0.2),
  border = NA
)
abline(v = mean(models$B0, na.rm = TRUE), col = "red", lty = 2, lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Density", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96), "red"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)

par(mfrow = c(1, 1))


## -----------------------------------------------------------------------------
#| label: kmeans-anchors
# Select 350 representative anchors using k-means sampling
# on the first 20 principal components
set.seed(1124)
kms <- naes(
  train_x, 
  k = 350, 
  pc = 20, 
  iter.max = 100,
  .center = TRUE, 
  .scale = TRUE
)

anchor_km <- kms$model

cat("Selected", length(anchor_km), "anchors via k-means\n")


## -----------------------------------------------------------------------------
#| label: liblex-anchored
#| echo: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_anchored


## -----------------------------------------------------------------------------
#| eval: false
# ciso_lib_anchored <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 80, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   anchor_indices = anchor_km,
#   control = liblex_control(tune = TRUE)
# )
# 
# ciso_lib_anchored


## -----------------------------------------------------------------------------
# Number of experts in the anchored library
n_experts <- nrow(ciso_lib_anchored$coefficients$B)
cat("Number of experts in the anchored library:", n_experts, "\n")


## -----------------------------------------------------------------------------
#| label: fig-centroids-anchored
#| fig-cap: "Neighborhood centroids from anchored library compared to test spectra"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center

wavs_pr <- as.numeric(colnames(ciso_lib_anchored$scaling$local_x_center))
n_experts <- nrow(ciso_lib_anchored$scaling$local_x_center)
n_test <- nrow(test_x)

# Plot test spectra first (background)
matplot(
  wavs_pr,
  t(test_x),
  col = rgb(0.8, 0.2, 0.2, 0.2),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance",
  main = paste0("Coverage: ", n_experts, " experts vs ", n_test, " test samples")
)

# Overlay centroids
matlines(
  wavs_pr,
  t(ciso_lib_anchored$scaling$local_x_center),
  col = rgb(0.23, 0.51, 0.96, 0.3),
  lty = 1
)
grid(lty = 1)

legend(
  "topright",
  legend = c(
    paste0("Test samples (n = ", n_test, ")"),
    paste0("Centroids (n = ", n_experts, ")")
  ),
  col = c(
    rgb(0.8, 0.2, 0.2, 0.5),
    rgb(0.23, 0.51, 0.96, 0.5)
  ),
  lty = 1,
  lwd = c(1, 1, 2, 2),
  bty = "n"
)


## -----------------------------------------------------------------------------
#| label: predict-basic
ciso_pred <- predict(ciso_lib_k, newdata = test_x, verbose = FALSE)

# Prediction output structure
names(ciso_pred)

# Main predictions with uncertainty
head(ciso_pred$predictions)


## -----------------------------------------------------------------------------
#| label: weighting-options
#| eval: false
# # Gaussian kernel (default)
# predict(ciso_lib_k, newdata = test_x, weighting = "gaussian")
# 
# # Tricubic kernel (similar to LOCAL algorithm)
# predict(ciso_lib_k, newdata = test_x, weighting = "tricube")
# 
# # Equal weights
# predict(ciso_lib_k, newdata = test_x, weighting = "none")


## -----------------------------------------------------------------------------
#| label: enforce-indices
#| eval: false
# # Always include specific experts in predictions
# predict(ciso_lib_k, newdata = test_x, enforce_indices = c(1, 5, 10))


## -----------------------------------------------------------------------------
#| label: evaluation
pred_values <- ciso_pred$predictions$pred
pred_sd <- ciso_pred$predictions$pred_sd

# Performance metrics (excluding missing values)
complete <- !is.na(test_y)
r2 <- cor(pred_values[complete], test_y[complete])^2
rmse <- sqrt(mean((pred_values[complete] - test_y[complete])^2))

cat("R²:  ", round(r2, 3), "\n")
cat("RMSE:", round(rmse, 3), "\n")


## -----------------------------------------------------------------------------
#| label: uncertainty-filter
# Filter predictions by uncertainty threshold
unc_threshold <- quantile(pred_sd[complete], 0.75)
reliable <- complete & (pred_sd < unc_threshold)

r2_filtered <- cor(pred_values[reliable], test_y[reliable])^2
rmse_filtered <- sqrt(mean((pred_values[reliable] - test_y[reliable])^2))

cat("After filtering high-uncertainty predictions:\n")
cat("  Retained:", sum(reliable), "/", sum(complete), "observations\n")
cat("  R²:      ", round(r2_filtered, 3), "\n")
cat("  RMSE:    ", round(rmse_filtered, 3), "\n")


## -----------------------------------------------------------------------------
#| label: fig-validation
#| fig-cap: "Predicted versus observed values"
#| fig-width: 6
#| fig-height: 5
#| fig-align: center
lims <- range(pred_values[complete], test_y[complete], na.rm = TRUE)

plot(
  pred_values[complete],
  test_y[complete],
  pch = 16,
  col = rgb(0, 0, 0, 0.5),
  xlab = "Predicted",
  ylab = "Observed",
  xlim = lims,
  ylim = lims
)

abline(0, 1, col = "red", lwd = 2)
grid(lty = 1)


## -----------------------------------------------------------------------------
#| label: parallel-example
#| eval: false
# library(doParallel)
# 
# # Register parallel backend
# n_cores <- min(parallel::detectCores() - 1, 4)
# cl <- makeCluster(n_cores)
# registerDoParallel(cl)
# 
# # Build library with parallel processing
# ciso_lib_parallel <- liblex(
#   Xr = train_x,
#   Yr = train_y,
#   neighbors = neighbors_k(seq(40, 120, by = 20)),
#   diss_method = diss_correlation(ws = 37, scale = TRUE),
#   fit_method = fit_wapls(
#     min_ncomp = 3,
#     max_ncomp = 15,
#     scale = TRUE,
#     method = "mpls"
#   ),
#   control = liblex_control(
#     tune = TRUE,
#     allow_parallel = TRUE,
#     chunk_size = 10
#   )
# )
# 
# # Clean up
# stopCluster(cl)
# registerDoSEQ()

## -----------------------------------------------------------------------------
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)


## -----------------------------------------------------------------------------
#| eval: true
citation(package = "resemble")


## -----------------------------------------------------------------------------
#| label: libraries
#| message: false
library(resemble)
library(prospectr)


## -----------------------------------------------------------------------------
#| message: false
#| results: hide
data(NIRsoil)
dim(NIRsoil)
str(NIRsoil)


## -----------------------------------------------------------------------------
#| label: NIRsoil
#| message: false
# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)


## -----------------------------------------------------------------------------
#| label: fig-plotspectra
#| fig-cap: "Raw spectral absorbance data (top) and first derivative of the absorbance spectra (bottom)."
#| fig-align: center
#| fig-width: 7
#| fig-height: 7
#| echo: false
# Hex sticker palette#
bg_dark <- NA #"#0F172A"      # background
blue <- "#3B82F6"         # border / spiral high
amber <- "#F59E0B"        # spiral low
amber_light <- "#FBBF24"  # text accent
slate <- "#64748B"        # muted elements

old_par <- par("mfrow", "mar", "bg")
par(mfrow = c(2, 1), mar = c(4, 4, 1, 4), bg = bg_dark)

new_wavs <- as.matrix(as.numeric(colnames(NIRsoil$spc_p)))

text_col <- "black"#"white"

# Plot 1: Raw spectra
plot(range(wavs), range(NIRsoil$spc), col = NA,
     xlab = "",
     ylab = "Absorbance",
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = wavs, y = t(NIRsoil$spc), 
         lty = 1, col = paste0(blue, "33"))

# Plot 2: First derivative
plot(range(new_wavs), range(NIRsoil$spc_p), col = NA,
     xlab = "Wavelengths, nm",
     ylab = expression(d(detrended~A)/d*lambda),
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = new_wavs, y = t(NIRsoil$spc_p), 
         lty = 1, col = paste0(amber, "33"))

par(old_par)


## -----------------------------------------------------------------------------
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

#' @title Control parameters for gesearch
#'
#' @description
#' Creates a control object specifying algorithm parameters for
#' \code{\link{gesearch}}.
#'
#' @param retain_by A character string specifying how training observations are
#'   selected at each iteration:
#'   \describe{
#'     \item{\code{"probability"}}{(default) Retains observations with errors
#'       below a percentile estimated from a given probability. More robust to
#'       outliers.}
#'     \item{\code{"proportion"}}{Retains a fixed proportion of observations
#'       with lowest errors.}
#'   }
#' @param percentile_type An integer between 1 and 9 specifying the quantile
#'   algorithm when \code{retain_by = "probability"}. Passed to the \code{type}
#'   argument of \code{\link[stats]{quantile}}. Default is 7. Ignored when
#'   \code{retain_by = "proportion"}.
#' @param tune A logical indicating whether to tune regression parameters
#'   (e.g., number of PLS components) via cross-validation at each iteration.
#'   Increases computation time substantially. Default is \code{FALSE}.
#' @param number An integer specifying the number of groups for leave-group-out
#'   cross-validation. Default is 10. This is used for validating the final 
#'   models built the samples found. When \code{tune = TRUE}, this value is also 
#'   used in the internal CV for tuning regression parameters at each iteration. 
#' @param p A numeric value in (0, 1) specifying the proportion of observations
#'   per group in leave-group-out cross-validation. Default is 0.75. When 
#'   \code{tune = TRUE}, this value is also used in the internal CV for tuning 
#'   regression parameters at each iteration. 
#' @param stagnation_limit An integer specifying the maximum number of
#'   consecutive iterations with no change in gene pool size before early
#'   termination. Prevents infinite loops when target size cannot be reached.
#'   Default is 5.
#' @param allow_parallel A logical indicating whether to enable parallel
#'   processing for internal resampling and calibration. The parallel backend
#'   must be registered by the user. Default is \code{TRUE}.
#' @param blas_threads An integer specifying the number of BLAS threads to use
#'   during computation. Default is 1, which avoids multi-threaded OpenBLAS
#'   overhead on Linux. Requires \pkg{RhpcBLASctl}. See Details.
#' @details
#' ## Retention strategies
#'
#' When \code{retain_by = "probability"} (default), observations with errors
#' below a percentile threshold are retained. The percentile is computed using
#' \code{\link[stats]{quantile}} with \code{probs} set to the \code{retain}
#' value from \code{\link{gesearch}}. This approach is more robust when outlier
#' observations have extreme error values.
#'
#' When \code{retain_by = "proportion"}, a fixed fraction of observations
#' (specified by the \code{retain} argument in \code{\link{gesearch}}) with the
#' lowest associated errors are kept at each iteration.
#'
#' ## Cross-validation for tuning
#'
#' When \code{tune = TRUE}, leave-group-out cross-validation is used to select
#' optimal regression parameters at each iteration. The \code{number} argument
#' controls how many CV groups are formed, and \code{p} controls the proportion
#' of observations in each group.
#' 
#' ## BLAS threading
#'
#' On Linux systems with multi-threaded OpenBLAS, the default thread count
#' can cause significant overhead for algorithms that perform many small
#' matrix operations (like the iterative PLS fits in \code{gesearch}).
#' Setting \code{blas_threads = 1} (the default) eliminates this overhead.
#'
#' This setting requires the \pkg{RhpcBLASctl} package. If not installed,
#' the parameter is ignored and a message is displayed. The original thread
#' count is restored when \code{\link{gesearch}} completes.
#' 
#' @return A list of class \code{"gesearch_control"} containing the specified
#'   parameters.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @references
#' Lobsey, C.R., Viscarra Rossel, R.A., Roudier, P., Hedley, C.B. 2017.
#' rs-local data-mines information from spectral libraries to improve local
#' calibrations. European Journal of Soil Science 68:840-852.
#'
#' @seealso \code{\link{gesearch}}, \code{\link[stats]{quantile}}
#'
#' @examples
#' # Default parameters (probability-based retention)
#' gesearch_control()
#'
#' # Proportion-based retention
#' gesearch_control(retain_by = "proportion")
#'
#' # Enable parameter tuning with custom CV settings
#' gesearch_control(tune = TRUE, number = 5, p = 0.8)
#'
#' @export

gesearch_control <- function(
    # Retention strategy
  retain_by = c("probability", "proportion"),
  percentile_type = 7L,
  # Cross-validation tuning
  tune = FALSE,
  number = 10L,
  p = 0.75,
  # Algorithm behavior
  stagnation_limit = 5L,
  allow_parallel = TRUE, 
  blas_threads = 1L
) {
  
  retain_by <- match.arg(retain_by)
  
  # Validate percentile_type
  if (retain_by == "probability") {
    if (!is.numeric(percentile_type) || length(percentile_type) != 1L ||
        !percentile_type %in% 1L:9L) {
      stop("'percentile_type' must be an integer between 1 and 9", call. = FALSE)
    }
    percentile_type <- as.integer(percentile_type)
  } else {
    percentile_type <- NULL
  }
  
  # Validate tune
  if (!is.logical(tune) || length(tune) != 1L || is.na(tune)) {
    stop("'tune' must be TRUE or FALSE", call. = FALSE)
  }
  
  # Validate number
  if (!is.numeric(number) || length(number) != 1L || number < 1L) {
    stop("'number' must be a positive integer", call. = FALSE)
  }
  number <- as.integer(number)
  
  # Validate p
  if (!is.numeric(p) || length(p) != 1L || p <= 0 || p >= 1) {
    stop("'p' must be a numeric value in (0, 1)", call. = FALSE)
  }
  
  # Validate stagnation_limit
  if (!is.numeric(stagnation_limit) || length(stagnation_limit) != 1L ||
      stagnation_limit < 1L || stagnation_limit != as.integer(stagnation_limit)) {
    stop("'stagnation_limit' must be a positive integer", call. = FALSE)
  }
  stagnation_limit <- as.integer(stagnation_limit)
  
  # Validate allow_parallel
  if (!is.logical(allow_parallel) || length(allow_parallel) != 1L ||
      is.na(allow_parallel)) {
    stop("'allow_parallel' must be TRUE or FALSE", call. = FALSE)
  }
  
  structure(
    list(
      retain_by = retain_by,
      percentile_type = percentile_type,
      tune = tune,
      number = number,
      p = p,
      stagnation_limit = stagnation_limit,
      allow_parallel = allow_parallel, 
      blas_threads = blas_threads
    ),
    class = "gesearch_control"
  )
}

#' @title Get neighbor information
#' @description Gathers neighborhood information for Xu observations in Xr.
#' Required during local regressions.
#' @details
#' For local PCA and PLS distances, local dissimilarity matrices are not
#' computed here (cheaper to compute during local regressions). Instead,
#' global distances are output for later local dissimilarity computation.
#' @keywords internal
#' @noRd
get_neighbor_info <- function(
    Xr, Xu, 
    diss_method,
    Yr = NULL,
    neighbors,
    spike = NULL,
    return_dissimilarity = FALSE,
    diss_usage = "none"
) {
  
  n_xr <- nrow(Xr)
  
  # Determine if using local dissimilarities
  is_local <- inherits(diss_method, "diss_local_pca") || 
    inherits(diss_method, "diss_local_pls")
  
  if (is_local) {
    # For local methods, use pre_k for initial neighbor search
    neighbors <- neighbors_k(diss_method$pre_k)
    return_diss <- FALSE
  } else {
    return_diss <- return_dissimilarity
  }
  
  # Search neighbors
  info_neighbors <- search_neighbors(
    Xr = Xr, 
    Xu = Xu, 
    diss_method = diss_method,
    Yr = Yr,
    neighbors = neighbors,
    spike = spike,
    return_dissimilarity = return_diss
  )
  
  # Compute Xr-Xr dissimilarity if needed for predictors
  if (diss_usage == "predictors" && !is_local) {
    is_ortho <- inherits(diss_method, "diss_pca") || 
      inherits(diss_method, "diss_pls")
    if (is_ortho) {
      # Convert scores to Mahalanobis space, compute Euclidean on Xr portion
      scores_mahal <- euclid_to_mahal(info_neighbors$projection$scores)
      info_neighbors$diss_xr_xr <- f_diss(
        scores_mahal[1:n_xr, , drop = FALSE],
        diss_method = "euclid",
        center = FALSE, 
        scale = FALSE
      )
    } else {
      # Non-orthogonal methods: compute directly on scaled data
      center <- diss_method$center %||% TRUE
      scale <- diss_method$scale %||% FALSE
      
      Xr_scaled <- scale(
        rbind(Xr, Xu),
        center = center,
        scale = scale
      )[1:n_xr, , drop = FALSE]
      
      info_neighbors$diss_xr_xr <- dissimilarity(
        Xr = Xr_scaled,
        diss_method = diss_method
      )$dissimilarity
    }
  }
  
  info_neighbors
}


#' @title Cross validation for PLS regression
#' @description for internal use only!
#' @keywords internal
#' @noRd
pls_cv <- function(
    x, y, ncomp,
    method = c("pls", "wapls"),
    center = TRUE, scale,
    min_component = 1,
    new_x = matrix(0, 1, 1),
    weights = NULL,
    p = 0.75,
    number = 10,
    group = NULL,
    retrieve = TRUE,
    tune = TRUE,
    max_iter = 1, tol = 1e-6,
    seed = NULL,
    algorithm = c("pls", "mpls", "simpls")
) {
  min_allowed <- (floor(min(ncol(x), nrow(x)) * p)) - 1
  
  if (min_allowed < ncomp) {
    ncomp <- min_allowed
  }
  
  modified <- ifelse(algorithm == "mpls", TRUE, FALSE)

  cv_samples <- sample_stratified(
    y = y,
    p = p,
    number = number,
    group = group,
    replacement = FALSE,
    seed = seed
  )
  
  if (method == "wapls" & retrieve & tune) {
    pls_c <- c(min_component, ncomp)
    seq_pls <- min(pls_c):max(pls_c)
    search_grid <- expand.grid(
      min_component = seq_pls,
      max_component = seq_pls,
      KEEP.OUT.ATTRS = FALSE
    )
    search_grid <- as.matrix(search_grid[search_grid$min_component < search_grid$max_component, ])
    rownames(search_grid) <- 1:nrow(search_grid)
    fit_method <- "completewapls1"
  } else {
    search_grid <- matrix(0, 0, 0)
    fit_method <- method
  }

  cv_results <- opls_cv_cpp(
    X = x,
    Y = y,
    scale = scale,
    method = fit_method,
    mindices = cv_samples$hold_in,
    pindices = cv_samples$hold_out,
    min_component = min_component,
    ncomp = ncomp,
    new_x = new_x,
    maxiter = max_iter,
    tol = tol,
    wapls_grid = search_grid,
    algorithm = algorithm
  )
  
  val <- NULL
  val$resamples <- cv_samples$hold_out
  
  if (method == "pls") {
    val$cv_results <- data.frame(
      ncomp = 1:ncomp,
      rmse_cv = rowMeans(cv_results$rmse_seg),
      st_rmse_cv = rowMeans(cv_results$st_rmse_seg),
      rmse_sd_cv = get_col_sds(t(cv_results$rmse_seg)),
      r2_cv = rowMeans(cv_results$rsq_seg)
    )
    best_pls_c <- val$cv_results$ncomp[which(val$cv_results$rmse_cv == min(val$cv_results$rmse_cv))]
    val$best_pls_c <- best_pls_c
    
    if (retrieve) {
      if (!is.null(weights)) {
        x <- sweep(x, 1, weights, "*")
        y <- y * weights
      }
      if (tune) {
        val$models <- opls_get_basics(
          X = x,
          Y = y,
          scale = scale,
          ncomp = best_pls_c,
          maxiter = max_iter,
          tol = tol,
          algorithm = algorithm
        )
      } else {
        val$models <- opls_get_basics(
          X = x,
          Y = y,
          scale = scale,
          ncomp = ncomp,
          maxiter = max_iter,
          tol = tol,
          algorithm = algorithm
        )
      }
    }
  }
  
  if (method == "wapls" & retrieve & !tune) {
    val$compweights <- cv_results$compweights
    val$cv_results <- data.frame(
      min_component = min_component,
      max_component = ncomp,
      rmse_cv = mean(cv_results$rmse_seg),
      st_rmse_cv = mean(cv_results$st_rmse_seg),
      rmse_sd_cv = sd(cv_results$rmse_seg),
      r2_cv = mean(cv_results$rsq_seg)
    )
    
    
    if (!is.null(weights)) {
      x <- sweep(x, 1, weights, "*") ### MODIFIED
      y <- y * weights
    }
    val$models <- opls_get_basics(
      X = x,
      Y = y,
      scale = scale,
      ncomp = ncomp,
      maxiter = max_iter,
      tol = tol,
      algorithm = algorithm
    )
  }
  
  if (method == "wapls" & retrieve & tune) {
    val$cv_results <- data.frame(search_grid,
                                 rmse_cv = rowMeans(cv_results$rmse_seg),
                                 st_rmse_cv = rowMeans(cv_results$st_rmse_seg),
                                 rmse_sd_cv = get_col_sds(t(cv_results$rmse_seg)),
                                 r2_cv = rowMeans(cv_results$rsq_seg)
    )
    opmls <- which.min(val$cv_results$rmse_cv)
    val$best_pls_c$min_component <- val$cv_results$min_component[opmls]
    val$best_pls_c$max_component <- val$cv_results$max_component[opmls]
    val$compweights <- cv_results$compweights
    
    
    if (!is.null(weights)) {
      x <- sweep(x, 1, weights, "*")
      y <- y * weights
    }
    
    tmp_compweights <- val$compweights[val$best_pls_c$min_component:val$best_pls_c$max_component]
    val$compweights[] <- 0
    tmp_compweights <- tmp_compweights / sum(tmp_compweights)
    val$compweights[val$best_pls_c$min_component:val$best_pls_c$max_component] <- tmp_compweights
    
    val$models <- opls_get_basics(
      X = x,
      Y = y,
      scale = scale,
      ncomp = ncomp,
      maxiter = max_iter,
      tol = tol,
      algorithm = algorithm
    )
  }
  val
}


#' @title Internal function for computing the weights of the PLS components
#' necessary for weighted average PLS
#' @description internal
#' @keywords internal
#' @param pls_model either an object returned by the \code{pls_cv} function or an
#' object as returned by the \code{opls_get_basics} function which contains a pls model.
#' @param original_x the original spectral matrix which was used for calibrating the
#' pls model.
#' @param type type of weight to be computed. The only available option (for
#' the moment) is \code{"w1"}. See details on the \code{mbl} function where it
#' is explained how \code{"w1"} is computed within the \code{"wapls"}
#' regression.
#' @param new_x a vector of a new spectral observation. When "w1" is selected, new_x
#' must be specified.
#' @param pls_c a vector of length 2 which contains both the minimum and maximum
#' number of PLS components for which the weights must be computed.
#' @return \code{get_wapls_weights} returns a vector of weights for each PLS
#' component specified
#' @noRd
#' @author Leonardo Ramirez-Lopez and Antoine Stevens
get_wapls_weights <- function(pls_model, original_x, type = "w1", new_x = NULL, pls_c) {
  if (!is.null(pls_model$models)) {
    rmse_cv <- pls_model$cv_results$rmse_cv
    pls_model <- pls_model$models
    is_plsCv <- TRUE
  } else {
    is_plsCv <- FALSE
  }
  if (max(pls_c) > max(pls_model$ncomp)) {
    stop("the maximum number of PLS components specified in pls_c exceeds the number of PLS components contained in pls_model!")
  }
  if (type == "w1" & is.null(new_x)) {
    stop("new_x is missng! When type = 'w2', a vector of a new spectral sample must be specified")
  }
  if (type == "w1" & (length(new_x) != ncol(original_x))) {
    stop("the length of the vector new_x does not match with the number of variables of original_x")
  }
  if (length(pls_c) != 2) {
    stop("'pls_c' must be a vector of length 2 which specifies both the minimum and the maximum number of PLS components")
  }
  if (pls_c[[1]] > pls_c[[2]]) {
    stop("for the vector of 'pls_c' the first value must be the minimum number of PLS components")
  }
  
  min_component <- pls_c[[1]]
  max_component <- pls_c[[2]]
  
  do_scaling <- ifelse(nrow(pls_model$transf$Xscale) == 1, TRUE, FALSE)
  do_centering <- ifelse(nrow(pls_model$transf$Xcenter) == 1, TRUE, FALSE)
  
  if (do_scaling) {
    scaling_vec <- pls_model$transf$Xscale[1, ]
  } else {
    scaling_vec <- matrix(NA, 0, 0)
  }
  
  if (do_centering) {
    centering_vec <- pls_model$transf$Xcenter[1, ]
  } else {
    scaling_vec <- matrix(NA, 0, 0)
  }
  
  
  whgt <- get_local_pls_weights(
    projection_mat = pls_model$projection_mat,
    xloadings = pls_model$X_loadings,
    coefficients = pls_model$coefficients,
    new_x = new_x,
    min_component = min_component,
    max_component = max_component,
    scale = do_scaling,
    Xcenter = centering_vec,
    Xscale = scaling_vec
  )[1, ]
  
  return(whgt)
}


#' @title An iterator for local prediction data in mbl
#' @description internal function. It collects only the data necessary to
#' execute a local prediction for the mbl function based on a list of neighbors.
#' Not valid for local dissmilitary (e.g. for ortho_diss(...., .local = TRUE))
#' @param Xr the Xr matrix in mbl.
#' @param Xu the Xu matrix in mbl. Default \code{NULL}. If not provided, the 
#' function will iterate for each \code{\{Yr, Xr\}} to get the respective neighbors.
#' @param Yr the Yr matrix in mbl.
#' @param Yu the Yu matrix in mbl. Default \code{NULL}. 
#' @param diss_usage a character string indicating if the dissimilarity data
#' will be used as predictors ("predictors") or not ("none").
#' @param neighbor_indices a matrix with the indices of neighbors of every Xu
#' found in Xr.
#' @param neighbor_diss a matrix with the dissimilarity socres for the neighbors
#' of every Xu found in Xr. This matrix is organized in the same way as
#' \code{neighbor_indices}.
#' @param diss_xr_xr a dissimilarity matrix between sampes in Xr.
#' @param group a factor representing the group labels of Xr.
#' @return an object of \code{class} iterator giving the following list:
#' \itemize{
#' \item{ith_xr: the Xr data of the neighbors for the ith observation (if
#' \code{diss_usage = "predictors"}, this data is combined with the local
#' dissmilarity scores of the neighbors of Xu (or Xr if Xu was not provided))}
#' \item{ith_yr: the Yr data of the neighbors for the ith observation}
#' \item{ith_xu: the ith Xu observation (or Xr if Xu was not provided). 
#' If \code{diss_usage = "predictors"}, this data is combined with the local 
#' dissmilarity scores to its Xr neighbors.}
#' \item{ith_yu: the ith Yu observation (or Yr observation if Xu was not provided).}
#' \item{ith_neigh_diss: the dissimilarity scores of the neighbors for the ith
#' observation.}
#' \item{ith_group: the group labels for ith_xr.}
#' \item{n_k: the number of neighbors.}
#' }
#' @details isubset will look at the order of knn in each col of D and
#' re-organize the rows of x accordingly
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
ith_mbl_neighbor <- function(
    Xr, Xu = NULL, Yr, Yu = NULL,
    diss_usage = "none",
    neighbor_indices,
    neighbor_diss = NULL,
    diss_xr_xr = NULL,
    group = NULL
) {
  k_neighbors <- colSums(!is.na(neighbor_indices))
  iter_k_neighbors <- iter(k_neighbors, by = "cell", recycle=T)
  iter_neighbors <- iter(neighbor_indices, by = "col", recycle=T)
  
  if (is.null(Xu)) {
    iter_xu <- iter(Xr, by = "row" , recycle=T)
    iter_yu <- iter(Yr, by = "cell", recycle=T)
    iterate_yu <- TRUE
  } else {
    iter_xu <- iter(Xu, by = "row", recycle=T)
    iter_yu <- iter(Yu, by = "cell", recycle=T)
    iterate_yu <- !is.null(Yu)
  }
  neighbor_diss <- t(neighbor_diss)
  iter_xr_xu_diss <- iter(neighbor_diss, by = "row", recycle=T)
  group
  nextEl <- function() {
    ith_neighborhood <- nextElem(iter_neighbors)
    ith_ks <- 1:nextElem(iter_k_neighbors)
    ith_neighborhood <- ith_neighborhood[ith_ks]
    
    ith_xr_neighbors <- Xr[ith_neighborhood, , drop = FALSE]
    
    ith_yr_neighbors <- Yr[ith_neighborhood, , drop = FALSE]
    ith_xu <- nextElem(iter_xu)
    ith_neigh_diss <- nextElem(iter_xr_xu_diss)[, ith_ks, drop = FALSE]
    
    if (iterate_yu) {
      ith_yu <- nextElem(iter_yu)
    } else {
      ith_yu <- NULL
    }
    
    if (diss_usage == "predictors") {
      ith_local_xr_xr_diss <- diss_xr_xr[ith_neighborhood, ith_neighborhood]
      colnames(ith_local_xr_xr_diss) <- paste0("k_", ith_ks)
      ith_xr_neighbors <- cbind(ith_local_xr_xr_diss, ith_xr_neighbors)
      ith_xu <- cbind(ith_neigh_diss, ith_xu)
    }
    
    if (!is.null(group)) {
      ith_group <- factor(group[ith_neighborhood])
    } else {
      ith_group <- NULL
    }
    
    it_neigh <- list(
      ith_xr = ith_xr_neighbors,
      ith_yr = ith_yr_neighbors,
      ith_xu = ith_xu,
      ith_yu = ith_yu,
      ith_neig_indices = ith_neighborhood,
      ith_neigh_diss = ith_neigh_diss,
      local_index_nearest = which.min(ith_neigh_diss)[1],
      ith_group = ith_group,
      n_k = max(ith_ks)
    )
    it_neigh
  }
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}


# #' @title Get local neighbors based on orthogonal dissimilarity
# #' @description Internal function for obtaining local neighbors based on
# #' dissimilarity matrices from orthogonal projections. These neighbors are
# #' obtained from an orthogonal projection on a set of precomputed neighbors.
# #' Used internally by \code{mbl()}.
# #' @param ith_xr The set of neighbors of a Xu observation found in Xr.
# #' @param ith_xu The Xu observation.
# #' @param ith_yr The response values of the neighbors.
# #' @param ith_yu The response value of the Xu observation.
# #' @param diss_usage Character indicating if dissimilarity data will be used
# #'   as predictors ("predictors") or not ("none").
# #' @param ith_neig_indices Vector of original indices of the Xr neighbors.
# #' @param neighbors A neighbor selection object from \code{neighbors_k()} or
# #'   \code{neighbors_diss()}.
# #' @param spike Vector of observation indices forced to be retained as neighbors.
# #' @param diss_method A local dissimilarity method object (\code{diss_local_pca()}
# #'   or \code{diss_local_pls()}).
# #' @param ith_group Vector containing group labels for \code{ith_xr}.
# #' @param mbl_is_parallel Logical indicating if \code{mbl()} is running in
# #'   parallel. If \code{TRUE}, nested parallelism is prevented by forcing
# #'   \code{allow_parallel = FALSE} in the dissimilarity computation.
# #' @return A list containing:
# #' \itemize{
# #'   \item{ith_xr: Xr data of neighbors (combined with local dissimilarity
# #'     scores if \code{diss_usage = "predictors"})}
# #'   \item{ith_yr: Yr data of neighbors}
# #'   \item{ith_xu: Xu observation (combined with local dissimilarity scores
# #'     if \code{diss_usage = "predictors"})}
# #'   \item{ith_yu: Yu observation}
# #'   \item{ith_neig_indices: Original indices of selected neighbors}
# #'   \item{ith_neigh_diss: Dissimilarity scores of neighbors}
# #'   \item{local_index_nearest: Index of nearest neighbor}
# #'   \item{ith_group: Group labels for ith_xr}
# #'   \item{n_k: Number of neighbors}
# #'   \item{ith_ncomp: Number of components used}
# #' }
# #' @author
# #' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
# #' @keywords internal
# #' @noRd
# get_ith_local_neighbors <- function(
#     ith_xr, ith_xu, ith_yr, ith_yu = NULL,
#     diss_usage = "none",
#     ith_neig_indices,
#     neighbors,
#     spike = NULL, 
#     diss_method,
#     ith_group = NULL,
#     mbl_is_parallel = FALSE
# ) {
#   is_predictors <- diss_usage == "predictors"
#   
#   # Prevent nested parallelism if mbl is running in parallel
#   if (mbl_is_parallel) {
#     diss_method$allow_parallel <- FALSE
#   }
#   
#   # Handle spike indices (only positive values are forced in)
#   if (!is.null(spike)) {
#     n_spike_hold <- sum(spike > 0)
#     if (n_spike_hold > 0) {
#       reorg_spike <- seq_len(n_spike_hold)
#     } else {
#       reorg_spike <- NULL
#     }
#   } else {
#     reorg_spike <- NULL
#   }
#   
#   # Convert neighbors to internal format
#   if (inherits(neighbors, "neighbors_k")) {
#     k <- max(neighbors$k)
#     k_diss <- NULL
#     k_range <- NULL
#   } else {
#     k <- NULL
#     k_diss <- max(neighbors$threshold)
#     k_range <- c(neighbors$k_min, neighbors$k_max)
#   }
#   
#   # Compute local dissimilarity
#   local_diss <- ortho_diss(
#     Xr = ith_xr, 
#     Xu = ith_xu,
#     Yr = ith_yr,
#     diss_method = diss_method,
#     compute_all = is_predictors,
#     return_projection = FALSE
#   )
#   
#   if (is_predictors) {
#     indx_xu <- ncol(local_diss$dissimilarity)
#     
#     loc_neigh <- diss_to_neighbors(
#       local_diss$dissimilarity[-indx_xu, indx_xu, drop = FALSE],
#       k = k, 
#       k_diss = k_diss, 
#       k_range = k_range,
#       spike = reorg_spike,
#       return_dissimilarity = FALSE
#     )
#     
#     neigh_indcs <- loc_neigh$neighbors
#     ith_local_xr_xr_diss <- local_diss$dissimilarity[neigh_indcs, neigh_indcs]
#     colnames(ith_local_xr_xr_diss) <- paste0("k_", seq_len(ncol(ith_local_xr_xr_diss)))
#     
#     ith_xr <- cbind(ith_local_xr_xr_diss, ith_xr[neigh_indcs, ])
#     ith_yr <- ith_yr[neigh_indcs, , drop = FALSE]
#     ith_xu <- cbind(t(loc_neigh$neighbors_diss), ith_xu)
#   } else {
#     loc_neigh <- diss_to_neighbors(
#       local_diss$dissimilarity,
#       k = k, 
#       k_diss = k_diss, 
#       k_range = k_range,
#       spike = reorg_spike,
#       return_dissimilarity = FALSE
#     )
#     
#     ith_xr <- ith_xr[loc_neigh$neighbors, ]
#     ith_yr <- ith_yr[loc_neigh$neighbors, , drop = FALSE]
#   }
#   
#   loc_neighbors <- ith_neig_indices[loc_neigh$neighbors]
#   ith_neigh_diss <- loc_neigh$neighbors_diss
#   
#   if (!is.null(ith_group)) {
#     ith_group <- ith_group[loc_neigh$neighbors]
#   }
#   
#   list(
#     ith_xr = ith_xr,
#     ith_yr = ith_yr,
#     ith_xu = ith_xu,
#     ith_yu = ith_yu,
#     ith_neig_indices = loc_neighbors,
#     ith_neigh_diss = ith_neigh_diss,
#     local_index_nearest = which.min(ith_neigh_diss)[1],
#     ith_group = ith_group,
#     n_k = length(loc_neighbors),
#     ith_ncomp = local_diss$ncomp
#   )
# }


#' @title Standard deviation of columns
#' @description For internal use only!
#' @keywords internal
#' @noRd
get_col_sds <- function(x) {
  return(as.vector(get_column_sds(x)))
}


#' @title Local multivariate regression
#' @description internal
#' @keywords internal
#' @noRd
fit_and_predict <- function(
    x, y, pred_method, scale = FALSE, weights = NULL,
    newdata, pls_c = NULL, CV = FALSE,
    tune = FALSE, number = 10, p = 0.75,
    group = NULL, noise_variance = 0.001,
    range_prediction_limits = TRUE,
    pls_max_iter = 1, pls_tol = 1e-6, algorithm, seed = NULL
) {
  if (is.null(weights)) {
    weights <- 1
  }
  
  if (any(get_col_sds(x) == 0)) {
    warning("Variables with zero variance. Data will not be scaled")
    scale <- FALSE
  }
  
  results <- cv_val <- pred <- NULL
  if (pred_method == "gpr") {
    if (CV) {
      cv_val <- gaussian_pr_cv(
        x = x,
        y = y,
        scale = scale,
        weights = weights,
        p = p,
        number = number,
        group = group,
        retrieve = "final_model",
        seed = seed
      )
      fit <- cv_val$model
    } else {
      x <- sweep(x, 1, weights, "*")
      y <- y * weights
      fit <- gaussian_process(
        X = x,
        Y = y,
        noisev = noise_variance,
        scale = scale
      )
    }
    
    pred <- predict_gaussian_process(
      Xz = fit$Xz,
      alpha = fit$alpha,
      newdata = newdata,
      scale = fit$is_scaled,
      Xcenter = fit$Xcenter,
      Xscale = fit$Xscale,
      Ycenter = fit$Ycenter,
      Yscale = fit$Yscale
    )[[1]]
  }

  if (pred_method == "pls") {
    if (CV) {
      cv_val <- pls_cv(
        x = x,
        y = y,
        ncomp = pls_c,
        method = "pls",
        center = TRUE,
        scale = scale,
        weights = weights,
        p = p,
        number = number,
        group = group,
        retrieve = TRUE,
        tune = tune,
        max_iter = pls_max_iter,
        tol = pls_tol,
        algorithm = algorithm,
        seed = seed
      )
      fit <- cv_val$models
      ncomp <- cv_val$best_pls_c
    } else {
      x <- sweep(x, 1, weights, "*")
      y <- y * weights
      fit <- opls_get_basics(
        X = x,
        Y = y,
        scale = scale,
        ncomp = pls_c,
        maxiter = pls_max_iter,
        tol = pls_tol,
        algorithm = algorithm
      )
      ncomp <- pls_c
    }
    pred <- predict_opls(
      bo = fit$bo,
      b = fit$coefficients,
      ncomp = ncomp,
      newdata = newdata,
      scale = scale,
      Xscale = fit$transf$Xscale
    )[, ncomp]
  }
  
  if (pred_method == "wapls") {
    if (CV) {
      cv_val <- pls_cv(
        x = x, y = y,
        ncomp = pls_c[[2]],
        method = "wapls",
        center = TRUE,
        scale = scale,
        min_component = pls_c[[1]],
        new_x = newdata,
        weights = weights,
        p = p,
        number = number,
        group = group,
        retrieve = TRUE,
        tune = tune,
        max_iter = pls_max_iter,
        tol = pls_tol,
        algorithm = algorithm,
        seed = seed
      )
      
      fit <- cv_val$models
      if (!tune) {
        cv_val$cv_results <- data.frame(
          min_component = pls_c[[1]],
          max_component = pls_c[[2]],
          rmse_cv = cv_val$cv_results$rmse_cv,
          st_rmse_cv = cv_val$cv_results$st_rmse_cv,
          rmse_sd_cv = cv_val$cv_results$rmse_sd_cv,
          r2_cv = cv_val$cv_results$r2_cv
        )
      }
      
      
      ## here all the weights are output from 1 to plsMax
      w <- cv_val$compweights
    } else {
      x <- sweep(x, 1, weights, "*")
      y <- y * weights
      fit <- opls_get_basics(
        X = x,
        Y = y,
        scale = scale,
        ncomp = pls_c[[2]],
        maxiter = pls_max_iter,
        tol = pls_tol,
        algorithm = algorithm
      )
      
      # compute weights for PLS components selected (from plsMin to plsMax)
      w <- rep(0, pls_c[[2]])
      w[pls_c[[1]]:pls_c[[2]]] <- get_wapls_weights(
        pls_model = fit,
        original_x = x,
        type = "w1",
        new_x = newdata,
        pls_c = pls_c
      )
    }
    
    # compute the weighted average of the multiple PLS predictions
    new_x_prediction <- predict_opls(
      bo = fit$bo,
      b = fit$coefficients,
      ncomp = pls_c[[2]],
      newdata = newdata,
      scale = scale,
      Xscale = fit$transf$Xscale
    )
    pred <- sum(new_x_prediction * w) # weighted preds
  }
  
  if (range_prediction_limits) {
    if (pred > max(y)) {
      pred <- max(y)
    }
    if (pred < min(y)) {
      pred <- min(y)
    }
  }
  
  results$validation <- cv_val
  results$prediction <- pred
  
  results
}

#' @title Cross validation for Gaussian process regression
#' @description internal
#' @keywords internal
#' @noRd
gaussian_pr_cv <- function(
    x,
    y,
    scale,
    weights = NULL,
    p = 0.75,
    number = 10,
    group = NULL,
    noise_variance = 0.001,
    retrieve = c("final_model", "none"),
    seed = NULL
) {
  
  ## Create the resampling groups
  cv_samples <- sample_stratified(
    y = y,
    p = p,
    number = number,
    group = group,
    replacement = FALSE,
    seed = seed
  )
  
  cv_results <- gaussian_process_cv(
    X = x,
    Y = y,
    mindices = cv_samples$hold_in,
    pindices = cv_samples$hold_out,
    noisev = noise_variance,
    scale = scale
  )
  
  val <- NULL
  val$resamples <- cv_samples$hold_out
  val$cv_results <- data.frame(
    rmse_cv = mean(cv_results$rmse_seg),
    st_rmse_cv = mean(cv_results$st_rmse_seg),
    rmse_sd_cv = sd(cv_results$rmse_seg),
    r2_cv = mean(cv_results$rsq_seg)
  )
  
  if (retrieve == "final_model") {
    if (is.null(weights)) {
      x <- sweep(x, 1, weights, "*")
      y <- y * weights
    }
    val$model <- gaussian_process(
      X = x,
      Y = y,
      noisev = noise_variance,
      scale = scale
    )
  }
  val
}


# =============================================================================
# GH distance computation
# =============================================================================

.compute_gh <- function(proj, n_xr) {
  scores <- proj$scores
  
  # Compute centroid from Xr scores only
  scores_center <- colMeans(scores[seq_len(n_xr), , drop = FALSE])
  
  # Mahalanobis distance from center
  gh_diss <- f_diss(
    Xr = scores,
    Xu = matrix(scores_center, nrow = 1),
    diss_method = "mahalanobis",
    center = FALSE,
    scale = FALSE
  )
  gh_vec <- as.vector(gh_diss)
  
  n_total <- nrow(scores)
  
  result <- list(gh_Xr = gh_vec[seq_len(n_xr)])
  
  if (n_total > n_xr) {
    result$gh_Xu <- gh_vec[(n_xr + 1L):n_total]
  }
  
  result
}

#' @title Cross-validated PLS fitting for gesearch
#' 
#' @description
#' Internal function that fits a PLS model with cross-validation for use in
#' `gesearch()`. Handles both `fit_pls()` and `fit_wapls()` methods.
#'
#' @param X Numeric matrix of predictor variables (observations in rows).
#' @param Y Numeric matrix or vector of response values.
#' @param indices Optional integer vector specifying row indices to use for
#'   training. If `NULL`, all rows are used.
#' @param fit_method A `fit_method` object created by [fit_pls()] or
#'   [fit_wapls()] specifying the PLS fitting parameters.
#' @param number Integer. Number of cross-validation iterations.
#' @param group Optional factor assigning group labels to observations for
#'   leave-group-out cross-validation.
#' @param retrieve Logical. If `TRUE`, return resampling indices.
#' @param tune Logical. If `TRUE`, tune the number of PLS components via
#'   cross-validation.
#' @param seed Optional integer for reproducible resampling.
#' @param p Numeric. Proportion of observations to retain in each
#'   cross-validation fold.
#'
#' @return An object of class `"plsr"` containing:
#' \describe{
#'   \item{models}{Fitted PLS model from `opls_get_basics()`.}
#'   \item{cv_results}{Cross-validation results.}
#'   \item{train_resamples}{Resampling indices (if `retrieve = TRUE`).}
#'   \item{fit_method}{The `fit_method` object used.}
#'   \item{Xscale}{Scaling parameters for predictors.}
#'   \item{call}{The matched call.}
#' }
#'
#' @seealso [gesearch()], [fit_pls()], [fit_wapls()]
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
get_plsr <- function(
    X, Y,
    indices = NULL,
    fit_method,
    number = NULL,
    group = NULL,
    retrieve = FALSE,
    tune = TRUE,
    seed = NULL,
    p = NULL
) {
  
  if (!inherits(fit_method, "fit_method")) {
    stop("'fit_method' must be a fit_*() object", call. = FALSE)
  }
  
  # Ensure matrices
  X <- as.matrix(X)
  Y <- as.matrix(Y)
  
  # Subset rows for training
  if (is.null(indices)) {
    idx <- seq_len(nrow(X))
  } else {
    idx <- indices
  }
  
  # Align optional grouping to the selected rows
  grp <- if (!is.null(group)) group[idx] else NULL
  
  # Determine ncomp range based on fit_method type
  if (inherits(fit_method, "fit_wapls")) {
    min_ncomp <- fit_method$min_ncomp
    max_ncomp <- fit_method$max_ncomp
  } else if (inherits(fit_method, "fit_pls")) {
    min_ncomp <- 1L
    max_ncomp <- fit_method$ncomp
  } else {
    stop("'fit_method' must be fit_pls() or fit_wapls()", call. = FALSE)
  }
  
  # ---- Cross-validated PLS -------------------------------------------------
  pls_args <- list(
    x = X[idx, , drop = FALSE],
    y = Y[idx, , drop = FALSE],
    ncomp = max_ncomp,
    method = fit_method$fit_method,
    center = fit_method$center,
    scale = fit_method$scale,
    min_component = min_ncomp,
    number = number,
    group = grp,
    retrieve = retrieve,
    tune = tune,
    max_iter = fit_method$max_iter,
    tol = fit_method$tol,
    seed = seed,
    algorithm = fit_method$method
  )
  
  if (!is.null(p)) pls_args$p <- p
  
  plsval <- do.call(pls_cv, pls_args)
  
  # ---- Fit model for prediction --------------------------------------------
  plsval$models <- opls_get_basics(
    X = X[idx, , drop = FALSE],
    Y = Y[idx, , drop = FALSE],
    ncomp = max_ncomp,
    scale = fit_method$scale,
    maxiter = fit_method$max_iter,
    tol = fit_method$tol, 
    algorithm = fit_method$method
  )
  
  # ---- Map resample indices back to original row indices -------------------
  if (!is.null(indices) && !is.null(plsval$resamples)) {
    rnms <- rownames(plsval$resamples)
    plsval$resamples <- apply(plsval$resamples, 2L, function(r) indices[r])
    rownames(plsval$resamples) <- rnms
  }
 
  # ---- Assemble output -----------------------------------------------------
  obj <- list(
    models = plsval$models,
    cv_results = plsval$cv_results,
    train_resamples = plsval$resamples,
    fit_method = fit_method,
    Xscale = plsval$models$transf$Xscale,
    call = match.call()
  )
  class(obj) <- "plsr"
  obj
}



#' @title Predict from a plsr model (internal)
#' 
#' @description
#' Internal S3 method. Generates predictions for new data using a `"plsr"`
#' object (coefficients and intercepts produced by `opls_get_basics()`).
#'
#' @param object A `"plsr"` object returned by `get_plsr()`.
#' @param newdata Numeric matrix or data frame of predictors.
#' @param ncomp Integer specifying the number of components to use. Defaults
#'   to the maximum number available in the model.
#' @param ... Unused. Reserved for future extensions.
#'
#' @return A numeric matrix of predictions with columns named
#'   `"pls1"`, `"pls2"`, etc.
#'
#' @details
#' Delegates to `predict_opls()`. Row names of `newdata` are preserved in
#' the output if present.
#'
#' @seealso `get_plsr()`, `fit_pls()`, `fit_wapls()`
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
predict.plsr <- function(object, newdata, ncomp = NULL, ...) {
  if (!inherits(object, "plsr")) {
    stop("'object' must be of class 'plsr'", call. = FALSE)
  }
  
  Xnew <- as.matrix(newdata)
  
  # Determine ncomp from fit_method if not provided
  if (is.null(ncomp)) {
    if (inherits(object$fit_method, "fit_wapls")) {
      ncomp <- object$fit_method$max_ncomp
    } else if (inherits(object$fit_method, "fit_pls")) {
      ncomp <- object$fit_method$ncomp
    } else {
      ncomp <- ncol(object$models$coefficients)
    }
  }
  
  yhat <- predict_opls(
    bo = object$models$bo,
    b = object$models$coefficients,
    newdata = Xnew,
    ncomp = ncomp,
    scale = object$fit_method$scale,
    Xscale = object$Xscale
  )
  
  colnames(yhat) <- paste0("ncomp", seq_len(ncol(yhat)))
  rownames(yhat) <- if (!is.null(rownames(newdata))) {
    rownames(newdata)
  } else {
    seq_len(nrow(yhat))
  }
  
  yhat
}

#' @title Validation metrics for PLS predictions (internal)
#' @description Internal helper to compute per-component metrics
#'   (RMSE, R², mean error) comparing predicted columns vs a single
#'   observed response.
#'
#' @param predicted Matrix/data frame of predictions
#'   (\code{n} rows × \code{k} components).
#' @param observed Numeric vector or single-column matrix with observed values.
#'
#' @return A data frame with columns:
#'   \itemize{
#'   \item \code{npls}: component index (1..\eqn{k})
#'   \item \code{r2}: squared correlation between predictions and observed
#'   \item \code{rmse}: root mean squared error
#'   \item \code{me}: mean error (signed bias)
#'   }
#'
#' @details Only complete cases in \code{observed} are used. The residuals are
#'   computed as \eqn{\hat{y}_{i,k} - y_i}.
#'
#' @seealso \code{\link{predict.plsr}}
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
get_validation_metrics <- function(predicted, observed) {
  # predicted  : matrix/data.frame of predictedictions (rows = samples, cols = comps)
  # observed : vector or single-column matrix with ground observed
  P <- as.matrix(predicted)
  Y <- as.matrix(observed)
  
  if (ncol(Y) != 1L)
    stop("`observed` must be a vector or a single-column matrix.")
  
  ok <- complete.cases(Y)
  if (!any(ok))
    return(data.frame(ncomp = integer(), rmse = numeric(), r2 = numeric()))
  
  P_ok <- P[ok, , drop = FALSE]
  y_ok <- Y[ok, 1, drop = TRUE]
  
  resid <- sweep(P_ok, 1, y_ok, FUN = "-")
  me    <- colMeans(resid)
  rmse  <- sqrt(colMeans(resid^2))
  r2   <- as.vector(cor(P_ok, y_ok)^2)
  
  data.frame(
    ncomp = seq_len(ncol(P)),
    r2   = r2,
    rmse = rmse,
    me   = me,
    row.names = NULL
  )
}

#' @title Tidy and label PLS model components (internal)
#' 
#' @description
#' Internal helper that renames and transposes selected fields of a PLS
#' object to a consistent shape, assigning row and column names based on
#' predictor names and the number of components.
#'
#' @param pls_obj A list-like PLS object containing elements such as
#'   `coefficients`, `projection_mat`, `vip`, `selectivity_ratio`,
#'   `X_loadings`, `Y_loadings`, `weights`, `scores`, and `ncomp`.
#' @param x_names Character vector of predictor variable names.
#'
#' @return The modified `pls_obj` with consistent names and matrix orientations.
#'
#' @details
#' Matrices `coefficients`, `projection_mat`, `vip`, and `selectivity_ratio`
#' are transposed to match downstream conventions (rows = components,
#' cols = variables). Component indices are assigned as `1:ncomp`; scores
#' are labelled with sample and component indices.
#'
#' @seealso `get_plsr()`
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
assign_pls_names <- function(pls_obj, x_names) {
  # Transpose to convention: rows = components, cols = variables
  pls_obj$coefficients <- t(pls_obj$coefficients)
  pls_obj$projection_mat <- t(pls_obj$projection_mat)
  pls_obj$vip <- t(pls_obj$vip)
  pls_obj$selectivity_ratio <- t(pls_obj$selectivity_ratio)
  
  # Assign column names from predictor names
  colnames(pls_obj$coefficients) <-
    colnames(pls_obj$X_loadings) <-
    colnames(pls_obj$projection_mat) <-
    colnames(pls_obj$vip) <-
    colnames(pls_obj$selectivity_ratio) <-
    colnames(pls_obj$weights) <- x_names
  
  # Assign row names (component indices)
  comp_names <- seq_len(pls_obj$ncomp)
  
  rownames(pls_obj$coefficients) <-
    colnames(pls_obj$bo) <-
    rownames(pls_obj$X_loadings) <-
    rownames(pls_obj$Y_loadings) <-
    rownames(pls_obj$projection_mat) <-
    rownames(pls_obj$vip) <-
    rownames(pls_obj$selectivity_ratio) <-
    rownames(pls_obj$weights) <- comp_names
  
  rownames(pls_obj$bo) <- "Intercepts"
  colnames(pls_obj$Y_loadings) <- "Loadings"
  
  # Scores: rows = samples, cols = components
  colnames(pls_obj$scores) <- comp_names
  rownames(pls_obj$scores) <- seq_len(nrow(pls_obj$scores))
  
  pls_obj
}

#' @title Local fitting method constructors
#' @name fit_methods
#' @aliases fit_pls
#' @aliases fit_wapls
#' @aliases fit_gpr
#' @description
#' 
#' \loadmathjax
#' 
#' These functions create configuration objects that specify how local 
#' regression models are fitted within the \code{\link{mbl}} function.
#' @usage
#' fit_pls(ncomp, method = c("pls", "mpls", "simpls"), 
#'         scale = FALSE, max_iter = 100L, tol = 1e-6)
#'
#' fit_wapls(min_ncomp, max_ncomp, method = c("mpls", "pls", "simpls"),
#'           scale = FALSE, max_iter = 100L, tol = 1e-6)
#'
#' fit_gpr(noise_variance = 0.001, center = TRUE, scale = TRUE)
#'
#' @param ncomp an integer indicating the number of PLS components to use
#' in local regressions when \code{fit_pls} is used.
#' @param min_ncomp an integer indicating the minimum number of PLS components
#' to use in local regressions when \code{fit_wapls} is used. See details.
#' @param max_ncomp an integer indicating the maximum number of PLS components
#' to use in local regressions when \code{fit_wapls} is used. See details.
#' @param method a character string indicating the PLS algorithm to use. 
#' Options are:
#' \itemize{
#'   \item{\code{'pls'}: standard PLS using covariance between X and Y for 
#'     weight computation (NIPALS algorithm).}
#'   \item{\code{'mpls'}: modified PLS using correlation between X and Y for 
#'     weight computation (NIPALS algorithm). See Shenk and Westerhaus (1991).}
#'   \item{\code{'simpls'}: SIMPLS algorithm (de Jong, 1993). Computationally 
#'     faster as it avoids iterative X deflation. Parameters \code{max_iter} 
#'     and \code{tol} are ignored when this method is used.}
#' }
#' Default is \code{'pls'} for \code{fit_pls} and \code{'mpls'} for 
#' \code{fit_wapls}.
#' @param scale logical indicating whether predictors must be scaled. 
#' Default is \code{FALSE} for PLS methods and \code{TRUE} for GPR.
#' @param max_iter an integer indicating the maximum number of iterations 
#' for convergence in the NIPALS algorithm. Only used when 
#' \code{method = 'pls'} or \code{method = 'mpls'}. Default is 100.
#' @param tol a numeric value indicating the convergence tolerance for 
#' calculating scores in the NIPALS algorithm. Only used when 
#' \code{method = 'pls'} or \code{method = 'mpls'}. Default is 1e-6.
#' @param noise_variance a numeric value indicating the variance of the noise
#' for Gaussian process local regressions (\code{fit_gpr}). Default is 0.001.
#' @param center logical indicating whether predictors should be centered 
#' before fitting. Only used for \code{fit_gpr}. Default is \code{TRUE}.
#'
#' @details
#' These functions create configuration objects that are passed to 
#' \code{\link{mbl}} to specify how local regression models are fitted.
#'
#' There are three fitting methods available:
#' 
#' \subsection{Partial least squares (\code{fit_pls})}{
#' Uses orthogonal scores partial least squares regression. Three algorithm 
#' variants are available:
#' \itemize{
#'   \item{\strong{Standard PLS} (\code{method = 'pls'}): Uses the NIPALS 
#'     algorithm with covariance-based weights.}
#'   \item{\strong{Modified PLS} (\code{method = 'mpls'}): Uses the NIPALS 
#'     algorithm with correlation-based weights. Proposed by Shenk and 
#'     Westerhaus (1991), this approach gives equal influence to all 
#'     predictors regardless of their variance scale.}
#'   \item{\strong{SIMPLS} (\code{method = 'simpls'}): Uses the SIMPLS 
#'     algorithm (de Jong, 1993), which deflates the cross-product matrix 
#'     rather than X itself. This is computationally faster, especially for 
#'     wide matrices, and produces identical predictions to standard PLS.}
#' }
#' The only parameter to optimise is the number of PLS components 
#' (\code{ncomp}).
#' }
#' 
#' \subsection{Weighted average PLS (\code{fit_wapls})}{
#' This method was developed by Shenk et al. (1997) and is used as the 
#' regression method in the LOCAL algorithm. It fits multiple PLS models 
#' using different numbers of components (from \code{min_ncomp} to 
#' \code{max_ncomp}). The final prediction is a weighted average of 
#' predictions from all models, where the weight for component \mjeqn{j}{j} 
#' is:
#'
#' \mjdeqn{w_{j} = \frac{1}{s_{1:j} \times g_{j}}}{w_j = 1/(s_{1:j} * g_j)}
#'
#' where \mjeqn{s_{1:j}}{s_{1:j}} is the root mean square of the spectral 
#' reconstruction error of the target observation(s) when \mjeqn{j}{j} PLS 
#' components are used, and \mjeqn{g_{j}}{g_j} is the root mean square of 
#' the squared regression coefficients for the \mjeqn{j}{j}th component.
#'
#' The same algorithm variants (\code{'pls'}, \code{'mpls'}, \code{'simpls'}) 
#' are available. The default is \code{'mpls'} following the original LOCAL 
#' implementation.
#' }
#' 
#' \subsection{Gaussian process regression (\code{fit_gpr})}{
#' Gaussian process regression is a non-parametric Bayesian method 
#' characterised by a mean and covariance function. This implementation uses 
#' a dot product covariance.
#'
#' The prediction vector \mjeqn{A}{A} is computed from training data 
#' (\mjeqn{X}{X}, \mjeqn{Y}{Y}) as:
#'
#' \mjdeqn{A = (X X^{T} + \sigma^2 I)^{-1} Y}{A = (X X^T + sigma^2 I)^{-1} Y}
#'
#' where \mjeqn{\sigma^2}{sigma^2} is the noise variance and \mjeqn{I}{I} is 
#' the identity matrix. Prediction for a new observation 
#' \mjeqn{x_{u}}{x_u} is:
#'
#' \mjdeqn{\hat{y}_{u} = x_{u} X^{T} A}{hat y_u = x_u X^T A}
#'
#' The only parameter is the noise variance (\code{noise_variance}).
#' }
#'
#' @return An object of class \code{c("fit_<method>", "fit_method")} 
#' containing the specified parameters. This object is passed to 
#' \code{\link{mbl}} to configure local model fitting.
#' 
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#' 
#' @references
#' de Jong, S. (1993). SIMPLS: An alternative approach to partial least 
#' squares regression. Chemometrics and Intelligent Laboratory Systems, 
#' 18(3), 251-263.
#' 
#' Rasmussen, C.E., Williams, C.K. (2006). Gaussian Processes for Machine 
#' Learning. MIT Press.
#' 
#' Shenk, J.S., & Westerhaus, M.O. (1991). Populations structuring of near 
#' infrared spectra and modified partial least squares regression. Crop 
#' Science, 31(6), 1548-1555.
#'
#' Shenk, J., Westerhaus, M., & Berzaghi, P. (1997). Investigation of a LOCAL
#' calibration procedure for near infrared instruments. Journal of Near 
#' Infrared Spectroscopy, 5, 223-232.
#' 
#' Westerhaus, M. (2014). Eastern Analytical Symposium Award for outstanding 
#' achievements in near infrared spectroscopy: my contributions to near 
#' infrared spectroscopy. NIR news, 25(8), 16-20.
#' 
#' @seealso \code{\link{mbl}}
#' 
#' @examples
#' # PLS with 10 components using standard algorithm
#' fit_pls(ncomp = 10)
#' 
#' # PLS with modified algorithm (correlation-based weights)
#' fit_pls(ncomp = 10, method = "mpls")
#' 
#' # PLS with SIMPLS (faster, no iteration)
#' fit_pls(ncomp = 10, method = "simpls")
#' 
#' # Weighted average PLS (LOCAL-style)
#' fit_wapls(min_ncomp = 3, max_ncomp = 12)
#' 
#' # Weighted average PLS with SIMPLS
#' fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "simpls")
#' 
#' # Gaussian process regression
#' fit_gpr()
#' fit_gpr(noise_variance = 0.01)
#' 
NULL


# =============================================================================
# Fitting method constructors
# =============================================================================
# Public:
#   fit_pls()
#   fit_wapls()
#   fit_gpr()
#
# Internal:
#   .new_fit_method()    - shared constructor
#   print.fit_method     - shared print (fallback)
#   print.fit_pls
#   print.fit_wapls
#   print.fit_gpr


# -----------------------------------------------------------------------------
# Internal constructor
# -----------------------------------------------------------------------------
.new_fit_method <- function(fit_method, ...) {
  
  structure(
    list(fit_method = fit_method, ...),
    class = c(paste0("fit_", fit_method), "fit_method")
  )
}


# -----------------------------------------------------------------------------
# fit_pls
# -----------------------------------------------------------------------------
#' @export
fit_pls <- function(
    ncomp,
    method = c("pls", "mpls", "simpls"),
    scale = FALSE,
    max_iter = 100L,
    tol = 1e-6
) {
  if (missing(ncomp)) {
    stop("'ncomp' is required.", call. = FALSE)
  }
  
  if (!is.numeric(ncomp) || length(ncomp) != 1L || ncomp < 1L) {
    stop("'ncomp' must be a positive integer.", call. = FALSE)
  }
  
  if (!is.logical(scale) || length(scale) != 1L) {
    stop("'scale' must be TRUE or FALSE.", call. = FALSE)
  }
  
  method <- match.arg(method)
  
  .new_fit_method(
    "pls",
    ncomp    = as.integer(ncomp),
    method   = method,
    scale    = scale,
    max_iter = as.integer(max_iter),
    tol      = tol
  )
}

#' @noRd
#' @export
print.fit_pls <- function(x, ...) {
  cat(
    "Fitting method: pls\n",
    "  ncomp    :", x$ncomp, "\n",
    "  method   :", x$method, "\n",
    "  scale    :", x$scale, "\n",
    "  max_iter :", x$max_iter, "\n",
    "  tol      :", x$tol, "\n"
  )
  invisible(x)
}


# -----------------------------------------------------------------------------
# fit_wapls
# -----------------------------------------------------------------------------
#' @export
fit_wapls <- function(
    min_ncomp,
    max_ncomp,
    method = c("mpls", "pls", "simpls"),
    scale = FALSE,
    max_iter = 100L,
    tol = 1e-6
) {
  if (missing(min_ncomp) || missing(max_ncomp)) {
    stop("Both 'min_ncomp' and 'max_ncomp' are required.", call. = FALSE)
  }
  
  if (!is.numeric(min_ncomp) || length(min_ncomp) != 1L || min_ncomp < 1L) {
    stop("'min_ncomp' must be a positive integer.", call. = FALSE)
  }
  
  if (!is.numeric(max_ncomp) || length(max_ncomp) != 1L || max_ncomp < 1L) {
    stop("'max_ncomp' must be a positive integer.", call. = FALSE)
  }
  
  if (min_ncomp >= max_ncomp) {
    stop("'min_ncomp' must be less than 'max_ncomp'.", call. = FALSE)
  }
  
  if (!is.logical(scale) || length(scale) != 1L) {
    stop("'scale' must be TRUE or FALSE.", call. = FALSE)
  }
  
  method <- match.arg(method)
  
  .new_fit_method(
    "wapls",
    min_ncomp = as.integer(min_ncomp),
    max_ncomp = as.integer(max_ncomp),
    method    = method,
    scale     = scale,
    max_iter  = as.integer(max_iter),
    tol       = tol
  )
}

#' @noRd
#' @export
print.fit_wapls <- function(x, ...) {
  cat(
    "Fitting method: wapls\n",
    "  min_ncomp :", x$min_ncomp, "\n",
    "  max_ncomp :", x$max_ncomp, "\n",
    "  method    :", x$method, "\n",
    "  scale     :", x$scale, "\n",
    "  max_iter  :", x$max_iter, "\n",
    "  tol       :", x$tol, "\n"
  )
  invisible(x)
}



# -----------------------------------------------------------------------------
# fit_gpr
# -----------------------------------------------------------------------------

#' @export
fit_gpr <- function(
    noise_variance = 0.001,
    center = TRUE,
    scale = TRUE
) {
  if (!is.numeric(noise_variance) || length(noise_variance) != 1L) {
    stop("'noise_variance' must be a single numeric value.", call. = FALSE)
  }
  
  if (noise_variance <= 0) {
    stop("'noise_variance' must be positive.", call. = FALSE)
  }
  
  if (!is.logical(center) || length(center) != 1L) {
    stop("'center' must be TRUE or FALSE.", call. = FALSE)
  }
  
  if (!is.logical(scale) || length(scale) != 1L) {
    stop("'scale' must be TRUE or FALSE.", call. = FALSE)
  }
  
  .new_fit_method(
    "gpr",
    noise_variance = noise_variance,
    center         = center,
    scale          = scale
  )
}

#' @noRd
#' @export
print.fit_gpr <- function(x, ...) {
  cat(
    "Fitting method: gpr\n",
    "  noise_variance :", x$noise_variance, "\n",
    "  center         :", x$center, "\n",
    "  scale          :", x$scale, "\n"
  )
  invisible(x)
}


# -----------------------------------------------------------------------------
# Fallback print for fit_method
# -----------------------------------------------------------------------------
#' @noRd
#' @export
print.fit_method <- function(x, ...) {
  cat("Fitting method:", x$fit_method, "\n")
  for (nm in setdiff(names(x), "fit_method")) {
    cat(" ", nm, ":", x[[nm]], "\n")
  }
  invisible(x)
}

context("test-liblex")

library(foreach)
library(RhpcBLASctl)
registerDoSEQ()

# =============================================================================
# Setup helper
# =============================================================================

.setup_liblex_data <- function() {
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  # Preprocess spectra
  sg_det <- prospectr::savitzkyGolay(
    prospectr::detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1
  test_idx <- NIRsoil$train == 0
  
  list(
    train_x = NIRsoil$spc_pr[train_idx, ],
    train_y = NIRsoil$Ciso[train_idx],
    test_x = NIRsoil$spc_pr[test_idx, ],
    test_y = NIRsoil$Ciso[test_idx]
  )
}

local_liblex_setup <- function(env = parent.frame()) {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  library(prospectr)
  data("NIRsoil", package = "prospectr")
  
  NIRsoil$spc_pr <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  
  env$train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
  env$train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
  env$test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
  env$test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
}

check_liblex_predictions <- function(model, test_x, test_y, r2_min = 0.82, rmse_max = 0.6) {
  y_hat <- predict(model, test_x, verbose = FALSE)
  r2 <- cor(y_hat$predictions$pred, test_y, use = "complete.obs")^2
  rmse <- sqrt(mean((y_hat$predictions$pred - test_y)^2, na.rm = TRUE))
  
  expect_true(r2 > r2_min, info = paste("R2 too low:", round(r2, 3)))
  expect_true(rmse < rmse_max, info = paste("RMSE too high:", round(rmse, 3)))
}


# =============================================================================
# Input validation tests - Xr
# =============================================================================

test_that("liblex requires Xr to have column names", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  train_x <- d$train_x
  colnames(train_x) <- NULL
  
  expect_error(
    liblex(
      Xr = train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30)
    ),
    "Xr.*must have column names"
  )
})


test_that("liblex requires Xr to be numeric", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  train_x <- d$train_x
  train_x[1, 1] <- "not_numeric"
  
  expect_error(
    liblex(
      Xr = train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30)
    ),
    "Xr.*must be numeric"
  )
})


# =============================================================================
# Input validation tests - Yr
# =============================================================================

test_that("liblex requires Yr to be numeric", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = as.character(d$train_y),
      neighbors = neighbors_k(30)
    ),
    "Yr.*must be.*numeric"
  )
})


test_that("liblex requires Yr length to match Xr rows", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y[1:10],
      neighbors = neighbors_k(30)
    ),
    "'Yr' must have the same number of observations as 'Xr'"
  )
})


test_that("liblex rejects multi-column Yr", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = cbind(d$train_y, d$train_y),
      neighbors = neighbors_k(30)
    ),
    "'Yr' must be a numeric vector or single-column matrix"
  )
})


# =============================================================================
# Input validation tests - neighbors
# =============================================================================

test_that("liblex requires neighbors argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y
    ),
    "neighbors.*required"
  )
})

test_that("liblex with neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  result <- liblex(
    Xr = d$train_x,
    Yr = d$train_y,
    neighbors = neighbors_diss(
      threshold = seq(0.05, 0.3, length.out = 3),
      k_min = 20,
      k_max = 40
    ),
    diss_method = diss_correlation(ws = 27, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 8, method = "mpls"),
    control = liblex_control(tune = TRUE),
    verbose = FALSE
  )
  
  expect_s3_class(result, "liblex")
  expect_true("diss_threshold" %in% names(result$results))
})


test_that("liblex with neighbors_diss predictions work", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  lib <- liblex(
    Xr = d$train_x,
    Yr = d$train_y,
    neighbors = neighbors_diss(
      threshold = c(0.1, 0.2, 0.3),
      k_min = 20,
      k_max = 40
    ),
    diss_method = diss_correlation(ws = 27, scale = TRUE),
    fit_method = fit_pls(ncomp = 6),
    control = liblex_control(tune = FALSE),
    verbose = FALSE
  )
  
  preds <- predict(lib, d$test_x, verbose = FALSE)
  
  expect_true(is.data.frame(preds$predictions))
  expect_equal(nrow(preds$predictions), nrow(d$test_x))
})

test_that("liblex requires neighbors_k object", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = 30
    ),
    "neighbors.*must be.*neighbors_k"
  )
})


test_that("liblex requires neighbors values >= 4", {
  
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(3)
    ),
    "All values in 'k' must be at least 4"
  )
})


# =============================================================================
# Input validation tests - diss_method
# =============================================================================

test_that("liblex validates diss_method type", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      diss_method = "euclidean"
    ),
    "diss_method.*must be.*diss_\\*"
  )
})


test_that("liblex validates precomputed diss_method matrix is numeric", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  n <- nrow(d$train_x)
  diss_mat <- matrix("a", n, n)
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      diss_method = diss_mat
    ),
    "diss_method.*matrix must be numeric"
  )
})


test_that("liblex validates precomputed diss_method matrix is square", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  n <- nrow(d$train_x)
  diss_mat <- matrix(0, n, n - 5)
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      diss_method = diss_mat
    ),
    "diss_method.*matrix must be square"
  )
})


test_that("liblex validates precomputed diss_method matrix dimensions", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  n <- nrow(d$train_x)
  diss_mat <- matrix(0, n - 10, n - 10)
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      diss_method = diss_mat
    ),
    "diss_method.*dimensions must match"
  )
})


# =============================================================================
# Input validation tests - fit_method
# =============================================================================

test_that("liblex validates fit_method type", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      fit_method = "wapls"
    ),
    "fit_method.*must be.*fit_\\*"
  )
})


# =============================================================================
# Input validation tests - anchor_indices
# =============================================================================

test_that("liblex validates anchor_indices are numeric without NA", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      anchor_indices = c(1, 2, NA)
    ),
    "anchor_indices.*numeric vector without NA"
  )
})


test_that("liblex validates anchor_indices are within bounds", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      anchor_indices = c(1, 2, nrow(d$train_x) + 10)
    ),
    "anchor_indices.*must be between 1 and nrow"
  )
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      anchor_indices = c(0, 1, 2)
    ),
    "anchor_indices.*must be between 1 and nrow"
  )
})


test_that("liblex validates anchor_indices does not exceed 90% of data", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  n <- nrow(d$train_x)
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      anchor_indices = seq_len(n)  # 100% of data
    ),
    "anchor_indices.*exceeds 90%"
  )
})


# =============================================================================
# Input validation tests - gh
# =============================================================================

test_that("liblex validates gh parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      gh = "yes"
    ),
    "gh.*must be TRUE or FALSE"
  )
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      gh = NA
    ),
    "gh.*must be TRUE or FALSE"
  )
})


# =============================================================================
# Input validation tests - control
# =============================================================================

test_that("liblex validates control parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      control = list(mode = "build")
    ),
    "control.*must be created by liblex_control"
  )
})


# =============================================================================
# Input validation tests - verbose
# =============================================================================

test_that("liblex validates verbose parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      verbose = "yes"
    ),
    "'verbose' must be a single TRUE or FALSE value"
  )
})


# =============================================================================
# Input validation tests - group
# =============================================================================

test_that("liblex validates group length", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      group = factor(rep(1:5, 10)),
      verbose = FALSE
    ),
    "group.*must have length equal to nrow"
  )
})


# =============================================================================
# Basic functionality tests (skip on CRAN)
# =============================================================================

test_that("liblex runs with default parameters", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  # Use subset for speed
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
  expect_true("dissimilarity" %in% names(model))
  expect_true("fit_method" %in% names(model))
  expect_true("coefficients" %in% names(model))
  expect_true("optimal_params" %in% names(model))
})


test_that("liblex runs with diss_pca", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_pca(ncomp = ncomp_by_var(0.99)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex runs with diss_pls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_pls(ncomp = 10),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex runs with diss_correlation", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_correlation(ws = 27, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex runs with diss_euclidean", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_euclidean(center = TRUE, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex runs with fit_pls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_pls(ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex runs with anchor_indices", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  anchor_idx <- sample(seq_along(idx), 30)
  
  expect_message(
    model <- liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(c(20, 30)),
      anchor_indices = anchor_idx,
      fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
      verbose = FALSE
    ),
    NA
  )
  
  expect_s3_class(model, "liblex")
  expect_equal(model$anchor_indices, anchor_idx)
})


test_that("liblex runs with gh = FALSE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    gh = FALSE,
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
  expect_null(model$gh)
})


test_that("liblex runs with gh = TRUE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    gh = TRUE,
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
  expect_true("gh" %in% names(model))
  expect_true("gh_Xr" %in% names(model$gh))
  expect_true("projection" %in% names(model$gh))
})


test_that("liblex allows missing values in Yr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  # Introduce some NAs
  train_y <- d$train_y[idx]
  train_y[sample(length(train_y), 5)] <- NA
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = train_y,
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


test_that("liblex with simpls method works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10, method = "simpls"),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})


# =============================================================================
# Output structure tests
# =============================================================================

test_that("liblex output contains expected elements", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expected_elements <- c(
    "dissimilarity", "fit_method", "gh", "results", "best",
    "optimal_params", "residuals", "coefficients", "vips",
    "selectivity_ratios", "scaling", "neighborhood_stats", "anchor_indices"
  )
  
  for (elem in expected_elements) {
    expect_true(elem %in% names(model), info = paste("Missing element:", elem))
  }
  
  # Check coefficients structure
  expect_true("B0" %in% names(model$coefficients))
  expect_true("B" %in% names(model$coefficients))
  
  # Check optimal_params structure
  expect_true("k" %in% names(model$optimal_params))
})


# =============================================================================
# predict.liblex input validation tests
# =============================================================================


test_that("predict.liblex validates probs parameter", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], probs = c(-0.1, 0.5)),
    "probs.*values in \\[0, 1\\]"
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], probs = c(0.5, 1.5)),
    "probs.*values in \\[0, 1\\]"
  )
})


test_that("predict.liblex validates range_prediction_limits", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], range_prediction_limits = "yes"),
    "range_prediction_limits.*must be TRUE or FALSE"
  )
})


test_that("predict.liblex validates residual_cutoff", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], residual_cutoff = -1),
    "residual_cutoff.*positive"
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], residual_cutoff = 0),
    "residual_cutoff.*positive"
  )
})


test_that("predict.liblex validates enforce_indices", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], enforce_indices = c(0, 1)),
    "enforce_indices.*positive integers"
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], enforce_indices = c(1, 1000)),
    "enforce_indices.*exceeding number of models"
  )
})


test_that("predict.liblex validates verbose", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], verbose = "yes"),
    "verbose.*must be TRUE or FALSE"
  )
})


test_that("predict.liblex validates newdata has required variables", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  # Remove some columns from test data
  test_subset <- d$test_x[1:5, 1:10]
  
  expect_error(
    predict(model, newdata = test_subset, verbose = FALSE),
    "Missing predictor variables"
  )
})


test_that("predict.liblex validates adaptive_bandwidth", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], adaptive_bandwidth = "yes"),
    "adaptive_bandwidth.*must be TRUE or FALSE"
  )
})


test_that("predict.liblex validates reliability_weighting", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], reliability_weighting = "yes"),
    "reliability_weighting.*must be TRUE or FALSE"
  )
})


# =============================================================================
# predict.liblex functionality tests
# =============================================================================

test_that("predict.liblex works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  preds <- predict(model, newdata = d$test_x[1:10, ], verbose = FALSE)
  
  expect_type(preds, "list")
  expect_true("predictions" %in% names(preds))
  expect_true("neighbors" %in% names(preds))
  expect_true("expert_predictions" %in% names(preds))
  
  # Check predictions structure
  expect_true("pred" %in% names(preds$predictions))
  expect_true("pred_sd" %in% names(preds$predictions))
  expect_equal(nrow(preds$predictions), 10)
})


test_that("predict.liblex works with different weighting functions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  weighting_methods <- c(
    "gaussian", "tricube", "triweight", "triangular",
    "quartic", "parabolic", "cauchy", "none"
  )
  
  for (w in weighting_methods) {
    preds <- predict(
      model, 
      newdata = d$test_x[1:5, ], 
      weighting = w,
      verbose = FALSE
    )
    expect_equal(nrow(preds$predictions), 5, info = paste("Weighting:", w))
  }
})


test_that("predict.liblex includes GH distance when available", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    gh = TRUE,
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  preds <- predict(model, newdata = d$test_x[1:10, ], verbose = FALSE)
  
  expect_true("gh" %in% names(preds$predictions))
  expect_equal(length(preds$predictions$gh), 10)
})


test_that("predict.liblex with range_prediction_limits clips predictions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 10),
    verbose = FALSE
  )
  
  preds <- predict(
    model, 
    newdata = d$test_x[1:10, ], 
    range_prediction_limits = TRUE,
    verbose = FALSE
  )
  
  # Check that below_min and above_max flags exist
  expect_true("below_min" %in% names(preds$predictions))
  expect_true("above_max" %in% names(preds$predictions))
  expect_true("min_yr" %in% names(preds$predictions))
  expect_true("max_yr" %in% names(preds$predictions))
})



































test_that("liblex with mpls, diss_scale=TRUE, fit_scale=FALSE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = FALSE, method = "mpls"),
    verbose = FALSE,
    control = liblex_control(tune = TRUE)
  )
  
  check_liblex_predictions(model, test_x, test_y)
})

test_that("liblex with simpls, diss_scale=TRUE, fit_scale=FALSE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = FALSE, method = "simpls"),
    verbose = FALSE,
    control = liblex_control(tune = FALSE)
  )
  
  check_liblex_predictions(model, test_x, test_y)
})

test_that("liblex with pls, diss_scale=TRUE, fit_scale=FALSE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = FALSE, method = "pls"),
    verbose = FALSE,
    control = liblex_control(tune = TRUE)
  )
  
  check_liblex_predictions(model, test_x, test_y, r2_min = 0.82, rmse_max = 0.65)
})

test_that("liblex with pls, diss_scale=TRUE, fit_scale=TRUE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = TRUE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = TRUE, method = "pls"),
    verbose = FALSE,
    control = liblex_control(tune = TRUE)
  )

  check_liblex_predictions(model, test_x, test_y, r2_min = 0.82, rmse_max = 0.75)
})

test_that("liblex with pls, diss_scale=FALSE, fit_scale=FALSE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = FALSE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = FALSE, method = "pls"),
    verbose = FALSE,
    control = liblex_control(tune = TRUE)
  )
  check_liblex_predictions(model, test_x, test_y, r2_min = 0.69, rmse_max = 1.3)
})

test_that("liblex with pls, diss_scale=FALSE, fit_scale=TRUE", {
  skip_on_cran()
  local_liblex_setup()
  
  model <- liblex(
    Xr = train_x, Yr = train_y,
    neighbors = neighbors_k(c(30, 40)),
    diss_method = diss_correlation(ws = 31, scale = FALSE),
    fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, scale = TRUE, method = "pls"),
    verbose = FALSE,
    control = liblex_control(tune = TRUE)
  )
  check_liblex_predictions(model, test_x, test_y)
})


# =============================================================================
# Additional validation tests for liblex
# =============================================================================

# -----------------------------------------------------------------------------
# anchor_indices validation
# -----------------------------------------------------------------------------

test_that("liblex rejects duplicate anchor_indices", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  
  expect_error(
    liblex(
      Xr = d$train_x,
      Yr = d$train_y,
      neighbors = neighbors_k(30),
      anchor_indices = c(1, 2, 3, 3, 5),
      verbose = FALSE
    ),
    "anchor_indices.*contains duplicate"
  )
})

test_that("liblex warns when max k exceeds anchor count", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  # Use small anchor set with large k
  expect_warning(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(c(20, 25)),
      anchor_indices = 1:15,
      fit_method = fit_wapls(3, 10),
      verbose = FALSE
    ),
    "exceeds number of anchors"
  )
})

# -----------------------------------------------------------------------------
# chunk_size validation
# -----------------------------------------------------------------------------

test_that("liblex errors when chunk_size exceeds data size", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(30),
      fit_method = fit_wapls(3, 10),
      control = liblex_control(chunk_size = 1000),
      verbose = FALSE
    ),
    "chunk_size.*cannot exceed"
  )
})

# -----------------------------------------------------------------------------
# verbose validation (additional coverage)
# -----------------------------------------------------------------------------

test_that("liblex rejects NA verbose", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(30),
      verbose = NA
    ),
    "verbose.*must be "
  )
})

test_that("liblex rejects vector verbose", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(30),
      verbose = c(TRUE, FALSE)
    ),
    "'verbose' must be a single TRUE or FALSE value"
  )
})

# -----------------------------------------------------------------------------
# Precomputed dissimilarity matrix validation
# -----------------------------------------------------------------------------

test_that("liblex validates precomputed matrix row count", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  
  # Wrong number of rows
  diss_mat <- matrix(0, n - 5, n)
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(30),
      diss_method = diss_mat,
      verbose = FALSE
    ),
    "'diss_method' matrix must be square"
  )
})

test_that("liblex validates precomputed matrix diagonal is zero", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  
  # Non-zero diagonal
  diss_mat <- matrix(runif(n * n), n, n)
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(30),
      diss_method = diss_mat,
      verbose = FALSE
    ),
    "diagonal.*zeros"
  )
})

test_that("liblex validates precomputed matrix with anchor_indices ncol", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  anchor_idx <- 1:20
  
  # Wrong number of columns for anchor case
  diss_mat <- matrix(0, n, n)  # Should be n x length(anchor_idx)
  diag(diss_mat) <- 0
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(15),
      anchor_indices = anchor_idx,
      diss_method = diss_mat,
      verbose = FALSE
    ),
    "ncol equal to length\\(anchor_indices\\)"
  )
})

test_that("liblex validates precomputed matrix anchor diagonal is zero", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  anchor_idx <- 1:20
  
  # Correct dimensions but non-zero anchor diagonal
  diss_mat <- matrix(runif(n * length(anchor_idx)), n, length(anchor_idx))

  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = d$train_y[idx],
      neighbors = neighbors_k(15),
      anchor_indices = anchor_idx,
      diss_method = diss_mat,
      verbose = FALSE
    ),
    "diss_method\\[anchor_indices, \\].*zeros on the diagonal"
  )
})

# -----------------------------------------------------------------------------
# Yr validation for anchors
# -----------------------------------------------------------------------------

test_that("liblex requires at least 3 non-missing Yr for anchors", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  
  # Set almost all Yr to NA
  train_y <- rep(NA_real_, length(idx))
  train_y[1:2] <- d$train_y[idx][1:2]  # Only 2 non-NA
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = train_y,
      neighbors = neighbors_k(30),
      verbose = FALSE
    ),
    "Each 'side_info' variable must have at least 4 non-missing values"
  )
})

# -----------------------------------------------------------------------------
# max_k exceeds valid Yr count
# -----------------------------------------------------------------------------

test_that("liblex errors when max_k exceeds non-missing Yr count", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  
  # Set most Yr to NA so valid count < k
  train_y <- d$train_y[idx]
  train_y[6:50] <- NA
  
  expect_error(
    liblex(
      Xr = d$train_x[idx, ],
      Yr = train_y,
      neighbors = neighbors_k(10),
      verbose = FALSE
    ),
    "no complete element pairs"
  )
})

# -----------------------------------------------------------------------------
# validate mode output structure
# -----------------------------------------------------------------------------

test_that("liblex validate mode returns correct structure", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  result <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(3, 10),
    control = liblex_control(mode = "validate"),
    verbose = FALSE
  )
  
  expect_s3_class(result, "liblex")
  expect_null(result$coefficients)
  expect_true("results" %in% names(result))
  expect_true("best" %in% names(result))
  expect_true("residuals" %in% names(result))
  expect_true("neighborhood_stats" %in% names(result))
})

test_that("liblex validate mode with rownames preserves them", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  train_x <- d$train_x[idx, ]
  rownames(train_x) <- paste0("obs_", seq_len(nrow(train_x)))
  
  result <- liblex(
    Xr = train_x,
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(3, 10),
    control = liblex_control(mode = "validate"),
    verbose = FALSE
  )
  
  # Check that neighborhood_stats has proper row names
  expect_true(all(grepl("^obs_", rownames(result$neighborhood_stats[[1]]))))
})

# =============================================================================
# predict.liblex additional validation tests
# =============================================================================

test_that("predict.liblex errors when precomputed diss requires diss_method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  
  # Build with precomputed matrix
  diss_mat <- as.matrix(dist(d$train_x[idx, 1:10]))
  diss_mat <- diss_mat / max(diss_mat)
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(20),
    diss_method = diss_mat,
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], verbose = FALSE),
    "precomputed dissimilarity.*diss_method.*required"
  )
})

test_that("predict.liblex warns when overriding stored diss_method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_correlation(ws = 27),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_warning(
    predict(
      model, 
      newdata = d$test_x[1:5, ], 
      diss_method = diss_correlation(ws = 31),  # Different
      verbose = FALSE
    ),
    "Overriding stored"
  )
})

test_that("predict.liblex validates diss_method is diss_* object", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:50]
  n <- length(idx)
  
  # Build with precomputed matrix
  diss_mat <- as.matrix(dist(d$train_x[idx, 1:10]))
  diss_mat <- diss_mat / max(diss_mat)
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(20),
    diss_method = diss_mat,
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_error(
    predict(model, newdata = d$test_x[1:5, ], diss_method = "euclidean", verbose = FALSE),
    "diss_method.*must be a diss_\\*"
  )
})

# -----------------------------------------------------------------------------
# verbose output in predict.liblex
# -----------------------------------------------------------------------------

test_that("predict.liblex verbose output for correlation", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_correlation(ws = 27),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_output(
    predict(model, newdata = d$test_x[1:5, ], verbose = TRUE),
    "correlation dissimilarity.*window size"
  )
})

test_that("predict.liblex verbose output for correlation full window", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_correlation(),  # No ws
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_output(
    predict(model, newdata = d$test_x[1:5, ], verbose = TRUE),
    "full window"
  )
})

test_that("predict.liblex verbose output for other diss methods", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    diss_method = diss_pca(ncomp = 10),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_output(
    predict(model, newdata = d$test_x[1:5, ], verbose = TRUE),
    "pca dissimilarity"
  )
})


# -----------------------------------------------------------------------------
# residual_cutoff functionality
# -----------------------------------------------------------------------------

test_that("predict.liblex with residual_cutoff penalizes high-residual models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  # With residual cutoff
  preds_cutoff <- predict(
    model, 
    newdata = d$test_x[1:5, ], 
    residual_cutoff = 0.1,
    verbose = FALSE
  )
  
  # Without residual cutoff
  preds_no_cutoff <- predict(
    model, 
    newdata = d$test_x[1:5, ], 
    verbose = FALSE
  )
  
  # Predictions may differ due to penalization
  expect_equal(nrow(preds_cutoff$predictions), 5)
  expect_equal(nrow(preds_no_cutoff$predictions), 5)
})

# -----------------------------------------------------------------------------
# enforce_indices functionality
# -----------------------------------------------------------------------------

test_that("predict.liblex with enforce_indices prepends models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  enforce_idx <- c(1, 5, 10)
  
  preds <- predict(
    model, 
    newdata = d$test_x[1:5, ],
    enforce_indices = enforce_idx,
    verbose = FALSE
  )
  
  expect_equal(nrow(preds$predictions), 5)
  
  # Check that enforced indices are in neighbors
  idc <- lapply(
    preds$neighbors$indices, 
    FUN = function(x, indices) all(x[1:length(indices)] == indices), 
    indices = enforce_idx
  )
  idc <- do.call(c, idc)
  
  expect_true(all(idc))
})

# =============================================================================
# plot.liblex tests
# =============================================================================

test_that("plot.liblex works with default parameters", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30, 40)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_silent(plot(model))
})

test_that("plot.liblex works with what = 'rmse'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30, 40)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_silent(plot(model, what = "rmse"))
})

test_that("plot.liblex works with what = 'r2'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30, 40)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_silent(plot(model, what = "r2"))
})

test_that("plot.liblex works with what = 'residuals'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 30, 40)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_silent(plot(model, what = "residuals"))
})

test_that("plot.liblex works with neighbors_diss", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_diss(
      threshold = c(0.1, 0.2, 0.3),
      k_min = 20,
      k_max = 40
    ),
    diss_method = diss_correlation(ws = 27),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_silent(plot(model))
  expect_silent(plot(model, what = "rmse"))
})


test_that("plot.liblex validates what parameter", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20)),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_null(plot(model))

  model$scaling$local_x_center <- model$results <- NULL  # Remove results to trigger error
  expect_error(
    plot(model),
    "Nothing to plot"
  )
})

# =============================================================================
# anchor_indices with non-projection dissimilarity methods
# =============================================================================

test_that("liblex with anchor_indices and diss_euclidean prescales correctly", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  anchor_idx <- sample(seq_along(idx), 30)
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 25)),
    anchor_indices = anchor_idx,
    diss_method = diss_euclidean(center = TRUE, scale = TRUE),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
  expect_equal(model$anchor_indices, anchor_idx)
})

test_that("liblex with anchor_indices and diss_cosine prescales correctly", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_liblex_data()
  idx <- which(!is.na(d$train_y))[1:80]
  anchor_idx <- sample(seq_along(idx), 30)
  
  model <- liblex(
    Xr = d$train_x[idx, ],
    Yr = d$train_y[idx],
    neighbors = neighbors_k(c(20, 25)),
    anchor_indices = anchor_idx,
    diss_method = diss_cosine(center = TRUE, scale = FALSE),
    fit_method = fit_wapls(3, 10),
    verbose = FALSE
  )
  
  expect_s3_class(model, "liblex")
})

context("test-gesearch")
library(foreach)
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# =============================================================================
# Setup helper
# =============================================================================

.setup_gesearch_data <- function() {
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  # Preprocess spectra
  sg_det <- prospectr::savitzkyGolay(
    prospectr::detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  list(
    train_x = NIRsoil$spc_pr[train_idx, ],
    train_y = NIRsoil$Ciso[train_idx],
    test_x = NIRsoil$spc_pr[test_idx, ],
    test_y = NIRsoil$Ciso[test_idx]
  )
}


# =============================================================================
# gesearch input validation tests
# =============================================================================

test_that("gesearch validates input dimensions", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  # nrow(Xr) must equal length(Yr)
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y[1:10],
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "nrow.*Xr.*must equal"
  )
  
  # ncol(Xr) must equal ncol(Xu)
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x[, 1:10],
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "ncol.*Xr.*must equal.*ncol.*Xu"
  )
})


test_that("gesearch validates k parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = nrow(d$train_x) + 10,
      b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "k.*must be less than"
  )
})


test_that("gesearch validates target_size parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  # target_size must be >= k
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 50, b = 10, target_size = 30,
      fit_method = fit_pls(ncomp = 5)
    ),
    "target_size.*must be.*>= k"
  )
  
  # target_size cannot exceed nrow(Xr)
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = nrow(d$train_x) + 100,
      fit_method = fit_pls(ncomp = 5)
    ),
    "target_size.*cannot be greater"
  )
})


test_that("gesearch validates retain parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, retain = 0, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "retain.*must be in"
  )
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, retain = 1.5, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "retain.*must be in"
  )
})


test_that("gesearch validates optimization parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "invalid_option"
    ),
    "Invalid.*optimization"
  )
})


test_that("gesearch requires Yu for response optimization", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      Yu = NULL,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "response"
    ),
    "Yu.*required.*response"
  )
})


test_that("gesearch validates fit_method", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = "pls"
    ),
    "fit_method.*must be"
  )
})


test_that("gesearch rejects missing values in Xu", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  test_x_na <- d$test_x
  test_x_na[1, 1] <- NA
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = test_x_na,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "Xu.*missing values"
  )
})


test_that("gesearch rejects infinite values", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  test_x_inf <- d$test_x
  test_x_inf[1, 1] <- Inf
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = test_x_inf,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "Infinite values"
  )
})


test_that("gesearch validates Xu has at least 3 rows", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x[1:2, ],
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "Xu must have at least 3 rows"
  )
})


test_that("gesearch validates verbose parameter", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = "yes"
    ),
    "verbose.*must be logical"
  )
})


# =============================================================================
# Basic functionality tests (skip on CRAN)
# =============================================================================

test_that("gesearch runs with reconstruction optimization", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
  expect_true("x_local" %in% names(gs))
  expect_true("y_local" %in% names(gs))
  expect_true("indices" %in% names(gs))
  expect_true("final_models" %in% names(gs))
  expect_true("validation_results" %in% names(gs))
  expect_gte(length(gs$indices), 50)
})


test_that("gesearch runs with different PLS methods", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  # mpls
  gs_mpls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "mpls"),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  expect_s3_class(gs_mpls, "gesearch")
  
  # simpls
  gs_simpls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "simpls"),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  expect_s3_class(gs_simpls, "gesearch")
})


test_that("gesearch runs with response optimization", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "response",
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
  expect_gte(length(gs$indices), 50)
})


test_that("gesearch runs with similarity optimization", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "similarity",
    verbose = FALSE,
    seed = 42
  )
  expect_s3_class(gs, "gesearch")
})


test_that("gesearch runs with combined optimization criteria", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = c("reconstruction", "similarity"),
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
})


test_that("gesearch with scale = TRUE works and predict works too", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, scale = TRUE),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
})


# =============================================================================
# predict.gesearch tests
# =============================================================================



test_that("predict.gesearch validations", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_error(
    predict(gs, newdata = d$test_x, type = "invalid"),
    "type.*must be"
  )
  
  preds <- predict(gs, newdata = d$test_x)
  
  expect_type(preds, "list")
  expect_equal(nrow(preds[[1]]), nrow(d$test_x))
  expect_equal(ncol(preds[[1]]), 10)  # ncomp columns
  expect_error(predict(gs), "newdata.*required")
  
  
})


test_that("predict.gesearch with what = 'all_generations' works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  preds <- predict(gs, newdata = d$test_x, what = "all_generations")
  
  expect_type(preds, "list")
  expect_true(length(preds) == gs$complete_iter)
  expect_true(all(grepl("^generation_", names(preds))))
})


test_that("predict.gesearch warns without intermediate_models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    intermediate_models = FALSE,
    verbose = FALSE,
    seed = 42
  )
  
  expect_warning(
    predict(gs, newdata = d$test_x, what = "all_generations"),
    "intermediate models"
  )
})


# =============================================================================
# plot.gesearch tests
# =============================================================================

test_that("plot.gesearch works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_silent(plot(gs, which = "weakness"))
  expect_silent(plot(gs, which = "removed"))
})


# =============================================================================
# Reproducibility tests
# =============================================================================

test_that("gesearch is reproducible with seed", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs1 <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  gs2 <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_equal(gs1$indices, gs2$indices)
  expect_equal(gs1$complete_iter, gs2$complete_iter)
})


# =============================================================================
# Output structure tests
# =============================================================================

test_that("gesearch output contains expected elements", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expected_elements <- c(
    "x_local", "y_local", "indices", "complete_iter",
    "iter_weakness", "samples", "n_removed", "control",
    "fit_method", "validation_results", "final_models", "seed"
  )
  
  for (elem in expected_elements) {
    expect_true(elem %in% names(gs), info = paste("Missing element:", elem))
  }
  
  # Check dimensions
  expect_equal(nrow(gs$x_local), length(gs$indices))
  expect_equal(nrow(gs$y_local), length(gs$indices))
  
  # Check validation results structure
  expect_true("results" %in% names(gs$validation_results[[1]]))
  expect_true("train" %in% names(gs$validation_results[[1]]$results))
  expect_true("test" %in% names(gs$validation_results[[1]]$results))
})


test_that("gesearch n_removed is a data.frame with correct structure", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs$n_removed, "data.frame")
  expect_true(all(c("iteration", "removed", "cumulative") %in% names(gs$n_removed)))
  expect_equal(nrow(gs$n_removed), gs$complete_iter)
  
  # Cumulative should be non-decreasing
  expect_true(all(diff(gs$n_removed$cumulative) >= 0))
})


test_that("gesearch with intermediate_models stores generations", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  expect_true("intermediate_models" %in% names(gs))
  expect_type(gs$intermediate_models, "list")
  expect_gt(length(gs$intermediate_models), 0)
  
  # Check structure of intermediate models
  for (gen in gs$intermediate_models) {
    expect_true("subset_size" %in% names(gen))
    expect_true("pls_models" %in% names(gen))
    expect_true("validation" %in% names(gen))
  }
})















test_that("gesearch produces similar results with simpls and pls for optimization = 'response'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs_simpls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "simpls"),
    optimization = "response",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  gs_pls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "pls"),
    optimization = "response",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  expect_true(identical(gs_simpls$indices, gs_pls$indices))
})



test_that("gesearch produces different results with pls and mpls for optimization = 'response'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs_pls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "pls"),
    optimization = "response",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  gs_mpls <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    Yu = d$test_y,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10, method = "mpls"),
    optimization = "response",
    intermediate_models = TRUE,
    verbose = FALSE,
    seed = 42
  )
  
  expect_false(identical(gs_mpls$indices, gs_pls$indices))
})















































































































# =============================================================================
# gesearch_control validation tests
# =============================================================================

test_that("gesearch validates control class", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      control = list(tune = FALSE)
    ),
    "'control' must be created by gesearch_control\\(\\)"
  )
})

# =============================================================================
# Multi-column Yr validation tests
# =============================================================================

test_that("gesearch validates multi-column Yr has column names", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  # Multi-column Yr without column names
  Yr_multi <- cbind(d$train_y, d$train_y * 2)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Yr must have column names when it has multiple columns"
  )
})

test_that("gesearch validates multi-column Yr row count", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  # Multi-column Yr with wrong row count
  Yr_multi <- cbind(y1 = d$train_y[1:10], y2 = d$train_y[1:10])
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "nrow\\(Xr\\) must equal nrow\\(Yr\\)"
  )
})

# =============================================================================
# Multi-column Yu validation tests
# =============================================================================

test_that("gesearch validates Yu length matches Xu rows", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      Yu = d$test_y[1:5],
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "nrow\\(Xu\\) must equal length\\(Yu\\)"
  )
})

test_that("gesearch rejects Yu_lims with multi-column Yu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  Yu_multi <- cbind(y1 = d$test_y, y2 = d$test_y * 2)
  Yr_multi <- cbind(y1 = d$train_y, y2 = d$train_y * 2)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      Yu = Yu_multi,
      Yu_lims = c(0, 10),
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'Yu_lims' only supported for single-column Yu"
  )
})

test_that("gesearch validates multi-column Yu ncol matches Yr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  Yr_multi <- cbind(y1 = d$train_y, y2 = d$train_y * 2)
  Yu_wrong <- cbind(y1 = d$test_y, y2 = d$test_y * 2, y3 = d$test_y * 3)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      Yu = Yu_wrong,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "ncol\\(Yr\\) must equal ncol\\(Yu\\)"
  )
})

test_that("gesearch validates multi-column Yu nrow matches Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  Yr_multi <- cbind(y1 = d$train_y, y2 = d$train_y * 2)
  Yu_wrong <- cbind(y1 = d$test_y[1:5], y2 = d$test_y[1:5] * 2)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      Yu = Yu_wrong,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "nrow\\(Xu\\) must equal nrow\\(Yu\\)"
  )
})

test_that("gesearch validates multi-column Yu has column names", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  Yr_multi <- cbind(y1 = d$train_y, y2 = d$train_y * 2)
  Yu_no_names <- cbind(d$test_y, d$test_y * 2)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      Yu = Yu_no_names,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Yu must have column names when it has multiple columns"
  )
})

test_that("gesearch validates Yu and Yr column names match", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  Yr_multi <- cbind(y1 = d$train_y, y2 = d$train_y * 2)
  Yu_wrong_names <- cbind(a = d$test_y, b = d$test_y * 2)
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = Yr_multi,
      Xu = d$test_x,
      Yu = Yu_wrong_names,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Column names of Yr and Yu must match"
  )
})

# =============================================================================
# fit_method validation tests
# =============================================================================

test_that("gesearch rejects non-fit_pls fit_method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_gpr(),
      verbose = FALSE
    ),
    "Only fit_pls\\(\\) is supported"
  )
})

test_that("gesearch warns when tune is used with reconstruction", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_warning(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 20, retain = 0.90,
      target_size = 350,
      fit_method = fit_pls(ncomp = 10),
      optimization = "reconstruction",
      control = gesearch_control(tune = TRUE),
      verbose = FALSE,
      seed = 42
    ),
    "PLS components are not tuned when optimization includes 'reconstruction'"
  )
})

# =============================================================================
# Local CV validation tests
# =============================================================================

test_that("gesearch errors when local CV leaves fewer than 3 observations", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      Yu = d$test_y,
      k = 10,  # small k
      b = 10,
      target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "response",
      control = gesearch_control(tune = TRUE, p = 0.95),  # leaves ~5% for hold-out
      verbose = FALSE
    ),
    "Local cross-validation requires at least 3 observations"
  )
})

test_that("gesearch errors when ncomp exceeds available samples", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 10,  # small k
      b = 10,
      target_size = 350,
      fit_method = fit_pls(ncomp = 15),  # more components than k
      control = gesearch_control(tune = FALSE),
      verbose = FALSE
    ),
    "More PLS components than available observations"
  )
})

# =============================================================================
# Range optimization validation tests
# =============================================================================

test_that("gesearch requires Yu_lims for range optimization", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "range",
      verbose = FALSE
    ),
    "'Yu_lims' is required for 'range' optimization"
  )
})

test_that("gesearch validates Yu_lims length", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "range",
      Yu_lims = c(0, 5, 10),
      verbose = FALSE
    ),
    "'Yu_lims' must be a numeric vector of length 2"
  )
})

test_that("gesearch with range optimization works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = c("reconstruction", "range"),
    Yu_lims = range(d$train_y),
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
})

# =============================================================================
# gesearch.formula tests
# =============================================================================

test_that("gesearch.formula validates formula class", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  expect_error(
    gesearch(
      formula = "y ~ spc",  # character, not formula
      train = train_df,
      test = test_df,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5)
    ),
    "'formula' argument must be a formula object"
  )
})

test_that("gesearch.formula requires fit_method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  expect_error(
    gesearch(
      formula = y ~ spc,
      train = train_df,
      test = test_df,
      k = 30, b = 10, target_size = 350
    ),
    "'fit_method' is missing"
  )
})

test_that("gesearch.formula warns when response missing in test", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(spc = I(d$test_x))  # No y column
  
  expect_warning(
    gs <- resemble::gesearch(
      formula = y ~ spc,
      train = train_df,
      test = test_df,
      k = 30, b = 20, retain = 0.90,
      target_size = 350,
      fit_method = fit_pls(ncomp = 10),
      optimization = "reconstruction",
      verbose = FALSE,
      seed = 42
    ),
    "y not found in test; assigned NA."
  )
  
  expect_s3_class(gs, "gesearch")
})

test_that("gesearch.formula errors when response optimization without response in test", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(spc = I(d$test_x))  # No y column
  
  expect_error(
    gesearch(
      formula = y ~ spc,
      train = train_df,
      test = test_df,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      optimization = "response"
    ),
    "'optimization = \"response\"' requires response values in test"
  )
})

test_that("gesearch.formula works with data.frame input", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  gs <- gesearch(
    formula = y ~ spc,
    train = train_df,
    test = test_df,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_s3_class(gs, "gesearch")
  expect_s3_class(gs, "gesearch.formula")
  expect_true("formula" %in% names(gs))
  expect_true("dataclasses" %in% names(gs))
})

# =============================================================================
# predict.gesearch formula model tests
# =============================================================================

test_that("predict.gesearch validates newdata for formula models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  gs <- gesearch(
    formula = y ~ spc,
    train = train_df,
    test = test_df,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  # Wrong type for newdata
  expect_error(
    predict(gs, newdata = as.list(d$test_x)),
    "'newdata' must be a data.frame or matrix"
  )
})

test_that("predict.gesearch validates missing variables for formula models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  gs <- gesearch(
    formula = y ~ spc,
    train = train_df,
    test = test_df,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  # Missing predictor variables
  wrong_df <- data.frame(wrong_name = I(d$test_x))
  
  expect_error(
    predict(gs, newdata = wrong_df),
    "Missing predictor variables"
  )
})

test_that("predict.gesearch works with formula model and data.frame", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  gs <- gesearch(
    formula = y ~ spc,
    train = train_df,
    test = test_df,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  preds <- predict(gs, newdata = test_df)
  
  expect_type(preds, "list")
  expect_equal(nrow(preds[[1]]), nrow(d$test_x))
})

test_that("predict.gesearch works with formula model and matrix", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_df <- data.frame(y = d$train_y, spc = I(d$train_x))
  test_df <- data.frame(y = d$test_y, spc = I(d$test_x))
  
  gs <- gesearch(
    formula = y ~ spc,
    train = train_df,
    test = test_df,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  # Predict with matrix
  preds <- predict(gs, newdata = d$test_x)
  
  expect_type(preds, "list")
  expect_equal(nrow(preds[[1]]), nrow(d$test_x))
})

test_that("predict.gesearch rejects non-matrix newdata for non-formula models", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  gs <- gesearch(
    Xr = d$train_x,
    Yr = d$train_y,
    Xu = d$test_x,
    k = 30, b = 20, retain = 0.90,
    target_size = 350,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    verbose = FALSE,
    seed = 42
  )
  
  expect_error(
    predict(gs, newdata = as.data.frame(d$test_x)),
    "'newdata' must be a matrix"
  )
})

# =============================================================================
# Verbose output tests
# =============================================================================

test_that("gesearch verbose output works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  expect_output(
    gs <- gesearch(
      Xr = d$train_x,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 20, retain = 0.90,
      target_size = 350,
      fit_method = fit_pls(ncomp = 10),
      optimization = "reconstruction",
      verbose = TRUE,
      seed = 42
    ),
    "Generation"
  )
})

# =============================================================================
# Missing Xr/Yr validation tests
# =============================================================================

test_that("gesearch rejects missing values in Xr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_x_na <- d$train_x
  train_x_na[1, 1] <- NA
  
  expect_error(
    gesearch(
      Xr = train_x_na,
      Yr = d$train_y,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'Xr' contains missing values"
  )
})

test_that("gesearch rejects missing values in Yr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_gesearch_data()
  
  train_y_na <- d$train_y
  train_y_na[1] <- NA
  
  expect_error(
    gesearch(
      Xr = d$train_x,
      Yr = train_y_na,
      Xu = d$test_x,
      k = 30, b = 10, target_size = 350,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'Yr' contains missing values"
  )
})

context("fit-and-k-constructors")

# =============================================================================
# neighbors_* constructor tests
# =============================================================================

test_that("neighbors_k constructor works", {
  ng <- neighbors_k(50)
  expect_s3_class(ng, c("neighbors_k", "neighbors"))
  expect_equal(ng$k, 50L)
  expect_equal(ng$method, "k")
  
  ng2 <- neighbors_k(c(100, 50, 80))
  expect_equal(ng2$k, c(50L, 80L, 100L))  # sorted
})


test_that("neighbors_k validates inputs", {
  expect_error(neighbors_k(), "required")
  expect_error(neighbors_k("a"), "numeric")
  expect_error(neighbors_k(3), "at least 4")
  expect_error(neighbors_k(c(10, 2)), "at least 4")
  expect_error(neighbors_k(10.5), "integer values")
})


test_that("neighbors_diss constructor works", {
  ng <- neighbors_diss(0.3)
  expect_s3_class(ng, c("neighbors_diss", "neighbors"))
  expect_equal(ng$threshold, 0.3)
  expect_equal(ng$k_min, 4L)
  expect_equal(ng$k_max, Inf)
  expect_equal(ng$method, "diss")
  
  ng2 <- neighbors_diss(c(0.3, 0.1, 0.2), k_min = 10, k_max = 150)
  expect_equal(ng2$threshold, c(0.1, 0.2, 0.3))  # sorted
  expect_equal(ng2$k_min, 10L)
  expect_equal(ng2$k_max, 150L)
})


test_that("neighbors_diss validates inputs", {
  expect_error(neighbors_diss(), "required")
  expect_error(neighbors_diss("a"), "numeric")
  expect_error(neighbors_diss(-0.1), "positive")
  expect_error(neighbors_diss(0.1, k_min = 2), "at least 4")
  expect_error(neighbors_diss(0.1, k_min = 10, k_max = 5), "greater than")
})


# =============================================================================
# fit_* constructor tests
# =============================================================================

test_that("fit_pls constructor works", {
  m <- fit_pls(ncomp = 10)
  expect_s3_class(m, c("fit_pls", "fit_method"))
  expect_equal(m$ncomp, 10L)
  expect_equal(m$method, "pls")
  expect_false(m$scale)
  expect_equal(m$max_iter, 100L)
  expect_equal(m$tol, 1e-6)
  
  m2 <- fit_pls(ncomp = 5, method = "mpls", scale = TRUE)
  expect_equal(m2$method, "mpls")
  expect_true(m2$scale)
  
  m3 <- fit_pls(ncomp = 5, method = "simpls")
  expect_equal(m3$method, "simpls")
})


test_that("fit_pls validates inputs", {
  expect_error(fit_pls(), "required")
  expect_error(fit_pls(0), "positive integer")
  expect_error(fit_pls("a"), "positive integer")
  expect_error(fit_pls(10, scale = "yes"), "TRUE or FALSE")
})


test_that("fit_wapls constructor works", {
  m <- fit_wapls(min_ncomp = 3, max_ncomp = 15)
  expect_s3_class(m, c("fit_wapls", "fit_method"))
  expect_equal(m$min_ncomp, 3L)
  expect_equal(m$max_ncomp, 15L)
  expect_equal(m$method, "mpls")  # default for wapls
  expect_false(m$scale)
  
  m2 <- fit_wapls(min_ncomp = 5, max_ncomp = 20, method = "simpls", scale = TRUE)
  expect_equal(m2$method, "simpls")
  expect_true(m2$scale)
})


test_that("fit_wapls validates inputs", {
  expect_error(fit_wapls(), "required")
  expect_error(fit_wapls(min_ncomp = 5), "required")
  expect_error(fit_wapls(min_ncomp = 0, max_ncomp = 10), "positive integer")
  expect_error(fit_wapls(min_ncomp = 10, max_ncomp = 5), "less than")
  expect_error(fit_wapls(min_ncomp = 5, max_ncomp = 5), "less than")
})


test_that("fit_gpr constructor works", {
  m <- fit_gpr()
  expect_s3_class(m, c("fit_gpr", "fit_method"))
  expect_equal(m$noise_variance, 0.001)
  expect_true(m$center)
  expect_true(m$scale)
  
  m2 <- fit_gpr(noise_variance = 0.01, center = FALSE, scale = FALSE)
  expect_equal(m2$noise_variance, 0.01)
  expect_false(m2$center)
  expect_false(m2$scale)
})


test_that("fit_gpr validates inputs", {
  expect_error(fit_gpr(noise_variance = 0), "positive")
  expect_error(fit_gpr(noise_variance = -1), "positive")
  expect_error(fit_gpr(noise_variance = "a"), "single numeric")
  expect_error(fit_gpr(center = "yes"), "TRUE or FALSE")
  expect_error(fit_gpr(scale = "yes"), "TRUE or FALSE")
})

context("test-fit-constructors")

# =============================================================================
# fit_pls constructor tests
# =============================================================================

test_that("fit_pls constructor works", {
  f1 <- fit_pls(ncomp = 10)
  expect_s3_class(f1, "fit_pls")
  expect_s3_class(f1, "fit_method")
  expect_equal(f1$ncomp, 10L)
  expect_equal(f1$method, "pls")
  expect_false(f1$scale)
  expect_equal(f1$max_iter, 100L)
  expect_equal(f1$tol, 1e-6)
  
  f2 <- fit_pls(ncomp = 15, method = "mpls", scale = TRUE)
  expect_equal(f2$ncomp, 15L)
  expect_equal(f2$method, "mpls")
  expect_true(f2$scale)
  
  f3 <- fit_pls(ncomp = 10, method = "simpls")
  expect_equal(f3$method, "simpls")
})


test_that("fit_pls requires ncomp", {
  expect_error(fit_pls(), "ncomp.*required")
})


test_that("fit_pls validates ncomp", {
  expect_error(fit_pls(ncomp = 0), "positive integer")
  expect_error(fit_pls(ncomp = -5), "positive integer")
  expect_error(fit_pls(ncomp = "ten"), "positive integer")
  expect_error(fit_pls(ncomp = c(5, 10)), "positive integer")
})


test_that("fit_pls validates scale", {
  expect_error(fit_pls(ncomp = 10, scale = "yes"), "TRUE or FALSE")
  expect_error(fit_pls(ncomp = 10, scale = c(TRUE, FALSE)), "TRUE or FALSE")
})


test_that("fit_pls validates method", {
  expect_error(fit_pls(ncomp = 10, method = "invalid"))
})


test_that("print.fit_pls works", {
  f <- fit_pls(ncomp = 10, method = "mpls")
  expect_output(print(f), "Fitting method: pls")
  expect_output(print(f), "ncomp.*: 10")
  expect_output(print(f), "method.*: mpls")
})


# =============================================================================
# fit_wapls constructor tests
# =============================================================================

test_that("fit_wapls constructor works", {
  f1 <- fit_wapls(min_ncomp = 3, max_ncomp = 15)
  expect_s3_class(f1, "fit_wapls")
  expect_s3_class(f1, "fit_method")
  expect_equal(f1$min_ncomp, 3L)
  expect_equal(f1$max_ncomp, 15L)
  expect_equal(f1$method, "mpls")  # default for wapls
  expect_false(f1$scale)
  
  f2 <- fit_wapls(min_ncomp = 5, max_ncomp = 20, method = "simpls", scale = TRUE)
  expect_equal(f2$method, "simpls")
  expect_true(f2$scale)
})


test_that("fit_wapls requires both min_ncomp and max_ncomp", {
  expect_error(fit_wapls(min_ncomp = 3), "required")
  expect_error(fit_wapls(max_ncomp = 15), "required")
})


test_that("fit_wapls validates min_ncomp < max_ncomp", {
  expect_error(fit_wapls(min_ncomp = 15, max_ncomp = 10), "less than")
  expect_error(fit_wapls(min_ncomp = 10, max_ncomp = 10), "less than")
})


test_that("fit_wapls validates ncomp values", {
  expect_error(fit_wapls(min_ncomp = 0, max_ncomp = 10), "positive integer")
  expect_error(fit_wapls(min_ncomp = 3, max_ncomp = -5), "positive integer")
})


test_that("print.fit_wapls works", {
  f <- fit_wapls(min_ncomp = 3, max_ncomp = 15)
  expect_output(print(f), "Fitting method: wapls")
  expect_output(print(f), "min_ncomp.*: 3")
  expect_output(print(f), "max_ncomp.*: 15")
})


# =============================================================================
# fit_gpr constructor tests
# =============================================================================

test_that("fit_gpr constructor works", {
  f1 <- fit_gpr()
  expect_s3_class(f1, "fit_gpr")
  expect_s3_class(f1, "fit_method")
  expect_equal(f1$noise_variance, 0.001)
  expect_true(f1$center)
  expect_true(f1$scale)
  
  f2 <- fit_gpr(noise_variance = 0.01, center = FALSE, scale = FALSE)
  expect_equal(f2$noise_variance, 0.01)
  expect_false(f2$center)
  expect_false(f2$scale)
})


test_that("fit_gpr validates noise_variance", {
  expect_error(fit_gpr(noise_variance = 0), "positive")
  expect_error(fit_gpr(noise_variance = -0.01), "positive")
  expect_error(fit_gpr(noise_variance = "small"), "numeric")
  expect_error(fit_gpr(noise_variance = c(0.001, 0.01)), "single numeric")
})


test_that("fit_gpr validates center and scale", {
  expect_error(fit_gpr(center = "yes"), "TRUE or FALSE")
  expect_error(fit_gpr(scale = "yes"), "TRUE or FALSE")
})


test_that("print.fit_gpr works", {
  f <- fit_gpr(noise_variance = 0.01)
  expect_output(print(f), "Fitting method: gpr")
  expect_output(print(f), "noise_variance.*: 0.01")
})

# test-model.R
# Unit tests for model(), model_control(), and predict.resemble_model()

context("test-model")

# =============================================================================
# Setup helper
# =============================================================================

.setup_model_data <- function(n_xr = 80, n_xu = 20, preprocess = TRUE) {
  data("NIRsoil", package = "prospectr")
  
  if (preprocess) {
    NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  }
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ][seq_len(n_xu), ]
  Xr <- Xr[!is.na(Yr), ][seq_len(n_xr), ]
  Yu <- Yu[!is.na(Yu)][seq_len(n_xu)]
  Yr <- Yr[!is.na(Yr)][seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}


# =============================================================================
# model_control tests
# =============================================================================

test_that("model_control constructor works", {
  ctrl <- model_control()
  
  expect_type(ctrl, "list")
  expect_s3_class(ctrl, "model_control")
  expect_equal(ctrl$validation_type, "lgo")
  expect_equal(ctrl$number, 10L)
  expect_equal(ctrl$p, 0.75)
})


test_that("model_control accepts valid arguments", {
  ctrl_none <- model_control(validation_type = "none")
  expect_equal(ctrl_none$validation_type, "none")
  
  ctrl_custom <- model_control(validation_type = "lgo", number = 5, p = 0.8)
  expect_equal(ctrl_custom$number, 5L)
  expect_equal(ctrl_custom$p, 0.8)
})


test_that("model_control validates inputs", {
  expect_error(model_control(validation_type = "invalid"), "arg")
  expect_error(model_control(number = 0), "positive integer")
  expect_error(model_control(number = -5), "positive integer")
  expect_error(model_control(number = "ten"), "positive integer")
  expect_error(model_control(p = 0), "between 0 and 1")
  expect_error(model_control(p = 1), "between 0 and 1")
  expect_error(model_control(p = 1.5), "between 0 and 1")
  expect_error(model_control(p = -0.5), "between 0 and 1")
})


test_that("print.model_control works", {
  ctrl <- model_control()
  expect_output(print(ctrl), "validation_type")
  expect_output(print(ctrl), "number")
  expect_output(print(ctrl), "p")
  
  ctrl_none <- model_control(validation_type = "none")
  output <- capture.output(print(ctrl_none))
  expect_true(any(grepl("none", output)))
})


# =============================================================================
# model() input validation tests
# =============================================================================

test_that("model validates Xr and Yr dimensions", {
  skip_on_cran()
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(5), nrow = 5)
  
  expect_error(
    model(Xr, Yr, fit_method = fit_pls(ncomp = 3), verbose = FALSE),
    "same number of rows"
  )
})


test_that("model rejects Yr with multiple columns", {
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(20), nrow = 10, ncol = 2)
  
  expect_error(
    model(Xr, Yr, fit_method = fit_pls(ncomp = 3), verbose = FALSE),
    "exactly one column"
  )
})


test_that("model rejects NA values", {
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(10), nrow = 10)
  
  Xr_na <- Xr
  Xr_na[1, 1] <- NA
  expect_error(
    model(Xr_na, Yr, fit_method = fit_pls(ncomp = 3), verbose = FALSE),
    "missing values"
  )
  
  Yr_na <- Yr
  Yr_na[1] <- NA
  expect_error(
    model(Xr, Yr_na, fit_method = fit_pls(ncomp = 3), verbose = FALSE),
    "missing values"
  )
})


test_that("model rejects invalid fit_method", {
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(10), nrow = 10)
  
  expect_error(
    model(Xr, Yr, fit_method = list(method = "pls"), verbose = FALSE),
    "fit_method.*fit_\\*\\(\\)"
  )
})


test_that("model rejects fit_wapls", {
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(10), nrow = 10)
  
  expect_error(
    model(Xr, Yr, fit_method = fit_wapls(3, 10), verbose = FALSE),
    "fit_wapls.*not supported"
  )
})


test_that("model rejects invalid control", {
  Xr <- matrix(rnorm(100), nrow = 10)
  Yr <- matrix(rnorm(10), nrow = 10)
  
  expect_error(
    model(Xr, Yr, fit_method = fit_pls(ncomp = 3), 
          control = list(validation_type = "lgo"), verbose = FALSE),
    "model_control"
  )
})


# =============================================================================
# Basic model functionality tests - PLS
# =============================================================================

test_that("model with fit_pls without CV works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, scale = FALSE),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
  expect_equal(mod$n_obs, nrow(d$Xr))
  expect_equal(mod$n_vars, ncol(d$Xr))
  expect_null(mod$cv_results)
  expect_true(!is.null(mod$model))
})


test_that("model with fit_pls with CV works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  ctrl <- model_control(validation_type = "lgo", number = 5, p = 0.75)
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 8, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
  expect_true(!is.null(mod$cv_results))
  expect_equal(nrow(mod$cv_results), 8)
  expect_true("rmse" %in% names(mod$cv_results))
  expect_true("r2" %in% names(mod$cv_results))
  expect_true("optimal" %in% names(mod$cv_results))
  expect_equal(sum(mod$cv_results$optimal), 1)
  expect_equal(which(mod$cv_results$optimal), which.min(mod$cv_results$rmse))
})


test_that("model with fit_pls (mpls method) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, method = "mpls"),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
})


test_that("model with fit_pls (simpls method) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, method = "simpls"),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
})


# =============================================================================
# Basic model functionality tests - GPR
# =============================================================================

test_that("model with fit_gpr without CV works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data(n_xr = 60)
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_gpr(noise_variance = 0.01, scale = TRUE),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
  expect_equal(mod$n_obs, nrow(d$Xr))
  expect_null(mod$cv_results)
  expect_true(!is.null(mod$model))
})


test_that("model with fit_gpr with CV works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data(n_xr = 60)
  
  ctrl <- model_control(validation_type = "lgo", number = 5, p = 0.7)
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_gpr(noise_variance = 0.001, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(mod, "resemble_model")
  expect_true(!is.null(mod$cv_results))
  
  # GPR CV results: single row (no components)
  expect_equal(nrow(mod$cv_results), 1)
  expect_true("rmse" %in% names(mod$cv_results))
  expect_true("r2" %in% names(mod$cv_results))
  expect_true(mod$cv_results$rmse > 0)
})


# =============================================================================
# predict.resemble_model tests
# =============================================================================

test_that("predict works for PLS model", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  
  preds <- predict(mod, d$Xu)
  
  expect_true(is.matrix(preds))
  expect_equal(nrow(preds), nrow(d$Xu))
  expect_equal(ncol(preds), 6)
  expect_equal(colnames(preds), paste0("ncomp", 1:6))
})


test_that("predict respects ncomp argument for PLS", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 10),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  # Request specific number of components
  preds_5 <- predict(mod, d$Xu, ncomp = 5)
  expect_equal(ncol(preds_5), 1)
  
  # Vector of components
  preds_vec <- predict(mod, d$Xu, ncomp = c(2, 5, 8))
  expect_equal(ncol(preds_vec), 3)
  expect_equal(colnames(preds_vec), c("ncomp2", "ncomp5", "ncomp8"))
})


test_that("predict rejects ncomp exceeding model", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 5),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_error(predict(mod, d$Xu, ncomp = 10), "exceeds")
})


test_that("predict works for GPR model", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data(n_xr = 60)
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_gpr(noise_variance = 0.01, scale = TRUE),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  preds <- predict(mod, d$Xu)
  
  expect_true(is.matrix(preds))
  expect_equal(nrow(preds), nrow(d$Xu))
  expect_equal(ncol(preds), 1)
  expect_equal(colnames(preds), "predicted")
})


test_that("predict rejects wrong number of columns", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 5),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  Xu_wrong <- matrix(rnorm(100), nrow = 10, ncol = 10)
  expect_error(predict(mod, Xu_wrong), "columns")
})


test_that("predict preserves rownames from newdata", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data(n_xu = 5)
  rownames(d$Xu) <- paste0("sample_", 1:5)
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 5),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  preds <- predict(mod, d$Xu)
  expect_equal(rownames(preds), paste0("sample_", 1:5))
})


# =============================================================================
# Expected results tests (skipped on CRAN)
# =============================================================================
## Sanity checks ensuring results stay within plausible bounds rather 
## than testing for exact values. This catches regressions where something 
## breaks catastrophically.

test_that("model delivers expected results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  ctrl <- model_control(validation_type = "lgo", number = 10, p = 0.75)
  
  # PLS
  pls_mod <- model(
    Xr = Xr,
    Yr = Yr,
    fit_method = fit_pls(ncomp = 15, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # GPR
  gpr_mod <- model(
    Xr = Xr,
    Yr = Yr,
    fit_method = fit_gpr(noise_variance = 0.001, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # Check PLS CV RMSE bounds
  expect_true(all(pls_mod$cv_results$rmse < 6))
  expect_true(min(pls_mod$cv_results$rmse) < 4.5)
  
  # Check PLS CV R2 bounds
  expect_true(max(pls_mod$cv_results$r2) > 0.5)
  
  # Check GPR CV RMSE bounds
  expect_true(gpr_mod$cv_results$rmse < 5)
  expect_true(gpr_mod$cv_results$rmse > 1)
  
  # Check GPR CV R2 bounds
  expect_true(gpr_mod$cv_results$r2 > 0.4)
  
  # Predictions should be in reasonable range
  pls_preds <- predict(pls_mod, Xu)
  gpr_preds <- predict(gpr_mod, Xu)
  
  # Check prediction correlation with Yu
  pls_cor <- cor(pls_preds[, which.min(pls_mod$cv_results$rmse)], Yu)
  gpr_cor <- cor(gpr_preds[, 1], Yu)
  
  
  expect_true(pls_cor > 0.6)
  expect_true(gpr_cor > 0.5)
})


test_that("model with different PLS methods delivers consistent results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  ctrl <- model_control(validation_type = "none")
  
  # NIPALS (default)
  mod_nipals <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, method = "pls"),
    control = ctrl,
    verbose = FALSE
  )
  
  # SIMPLS
  mod_simpls <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, method = "simpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  # MPLS
  mod_mpls <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 6, method = "mpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  # Predictions should be similar (not identical due to different algorithms)
  preds_nipals <- predict(mod_nipals, d$Xu)
  preds_simpls <- predict(mod_simpls, d$Xu)
  preds_mpls <- predict(mod_mpls, d$Xu)
  
  # Correlation between methods should be very high
  expect_true(cor(preds_nipals[, 6], preds_simpls[, 6]) > 0.99)
  expect_true(cor(preds_nipals[, 6], preds_mpls[, 6]) > 0.99)
})


test_that("model predictions are numerically consistent across calls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 5),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  preds1 <- predict(mod, d$Xu)
  preds2 <- predict(mod, d$Xu)
  
  expect_equal(preds1, preds2)
})


test_that("model stores fit_method and control in output", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  fm <- fit_pls(ncomp = 6, scale = TRUE, method = "simpls")
  ctrl <- model_control(validation_type = "none")
  
  mod <- model(d$Xr, d$Yr, fit_method = fm, control = ctrl, verbose = FALSE)
  
  expect_equal(mod$fit_method$ncomp, 6)
  expect_equal(mod$fit_method$scale, TRUE)
  expect_equal(mod$fit_method$method, "simpls")
  expect_equal(mod$control$validation_type, "none")
})


test_that("model stores call attribute", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_model_data()
  
  mod <- model(
    Xr = d$Xr,
    Yr = d$Yr,
    fit_method = fit_pls(ncomp = 5),
    control = model_control(validation_type = "none"),
    verbose = FALSE
  )
  
  expect_true(!is.null(attr(mod, "call")))
  expect_true(inherits(attr(mod, "call"), "call"))
})

#' @title Control parameters for global model fitting
#' @name model_control
#' @description
#' Specifies cross-validation settings for the \code{\link{model}} function.
#'
#' @usage 
#' 
#' model_control(validation_type = c("lgo", "none"), number = 10L, p = 0.75)
#' 
#' @param validation_type a character string specifying the validation method:
#'   \itemize{
#'     \item{\code{"lgo"}: Leave-group-out cross-validation. At each iteration,
#'       a proportion \code{p} of observations is retained for training and the
#'       remainder is used for validation. This is repeated \code{number} times.}
#'     \item{\code{"none"}: No cross-validation is performed.}
#'   }
#' @param number an integer indicating the number of cross-validation 
#'   iterations. Only used when \code{validation_type = "lgo"}. Default is 10.
#' @param p a numeric value between 0 and 1 indicating the proportion of 
#'   observations to retain for training at each cross-validation iteration. 
#'   Only used when \code{validation_type = "lgo"}. Default is 0.75.
#' 
#' @return A list of class \code{"model_control"} containing the specified 
#'   parameters.
#'
#' @seealso \code{\link{model}}
#'
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @examples
#' # Default settings (leave-group-out CV with 10 iterations)
#' model_control()
#' 
#' # No cross-validation
#' model_control(validation_type = "none")
#' 
#' # Custom CV settings
#' model_control(validation_type = "lgo", number = 20, p = 0.80)
#'
#' @export
model_control <- function(
    validation_type = c("lgo", "none"),
    number = 10L,
    p = 0.75
) {
  validation_type <- match.arg(validation_type)
  
  
  if (!is.numeric(number) || length(number) != 1L || number < 1L) {
    stop("'number' must be a positive integer.", call. = FALSE)
  }
  
  if (!is.numeric(p) || length(p) != 1L || p <= 0 || p >= 1) {
    stop("'p' must be a numeric value between 0 and 1 (exclusive).", 
         call. = FALSE)
  }
  
  structure(
    list(
      validation_type = validation_type,
      number          = as.integer(number),
      p               = p
    ),
    class = "model_control"
  )
}


#' @noRd
#' @export
print.model_control <- function(x, ...) {
  cat("Model control parameters:\n")
  cat("  validation_type :", x$validation_type, "\n")
  if (x$validation_type != "none") {
    cat("  number          :", x$number, "\n")
    cat("  p               :", x$p, "\n")
  }
  invisible(x)
}

#' @title Print method for an object of class \code{ortho_projection}
#' @description Prints the contents of an object of class \code{ortho_projection}
#' @aliases print.ortho_projection
#' @usage \method{print}{ortho_projection}(x, ...)
#' @param x an object of class \code{ortho_projection} (as returned by the
#' \code{ortho_projection} function).
#' @param ... arguments to be passed to methods (not yet functional).
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @export

print.ortho_projection <- function(x, ...) {
  cat("\n", "Method: ", x$method)
  cat("\n", "Number of components retained: ", x$ncomp, "\n")
  cat(" Number of observations and number of original variables: ", c(nrow(x$scores), ncol(x$X_loadings)), "\n")
  cat(paste0(" Original variance in Xr: ", round(x$variance$original_x_var, 3), "\n"))

  if (any(grepl("Xu", rownames(x$scores)))) {
    mss <- c("\n", "Explained variance in X {Xr; Xu}: \n")
  } else {
    mss <- c("\n", "Explained variance in Xr: \n")
  }

  cat("\n", "Explained variances, ratio of explained variance, cumulative explained variance:", "\n")
  cat(mss)
  print(x$variance$x_var, digits = 3)

  if (x$method %in% c("pls", "mpls")) {
    cat("\n", "Explained variance in Yr: \n")
    print(x$variance$y_var, digits = 3)
  }
  invisible(x)
}

## tools:
## right triangular prism (errror, plsmin, plsmax, k)
##  weighted sd enemble
## vips analysis
## bo analysis
## animation of the vip (reverse from global to local)
## SD vips vs R2 at different neighborhood sizes 

## TODO:
# - add an option to indicate wheter selectivityR and vip must be calculated during building... 
#   if TRUE, opls2 is used (slow) if FALSE opls3 is used (fast)
# 
## downscaling the library to rule-based models


# rgrid <- matrix(NA, max(pls.c), max(pls.c))
# 
# for(pp in 1:105){
#   rgrid[predperformance$minpls[pp],
#         predperformance$maxpls[pp]] <- predperformance$rmse[pp]
# }
# 
# 



# Xr = Xr, 
# Yr = Yr, 
# minF = minF, 
# maxF = maxF, 
# emgrid = emgrid,
# scale = scale, 
# maxiter = maxiter, 
# tol = tol, 
# regression = regression, 
# pc_selection = pc_selection,
# k_kidxmat = kidxmat[ik,],
# k_kidxgrop = kidxgrop[ik,])
# 
# 


#' @title Iterator for Grouped Subsets by Index
#' @description
#' Internal helper that returns an iterator object over subsets of `x` and `y`,
#' grouped by precomputed index sets (e.g., k-nearest neighbors), with optional
#' dissimilarity matrix augmentation.
#'
#' @param x A matrix or data frame of predictors (observations in rows).
#' @param y A vector or matrix of responses, same number of rows as `x`.
#' @param kindx A list or matrix of indices (e.g., kNN indices) used to form 
#' subsets.
#' @param kgroup A vector of group indices into `kindx` to select for each 
#' iteration.
#' @param D (Optional) A square dissimilarity matrix corresponding to `x`, used 
#' to augment the feature set if provided.
#'
#' @return
#' An object of class `isubset`, `abstractiter`, `iter` that produces, on each
#' call to `nextElem()`:
#' \itemize{
#'   \item \code{x}: Subset of predictors (optionally prepended with local 
#'   dissimilarities).
#'   \item \code{y}: Corresponding subset of responses.
#'   \item \code{xval}: The test observation (optionally prepended with 
#'   dissimilarity vector).
#' }
#'
#' @details
#' On each iteration:
#' \itemize{
#'   \item A group of indices is selected using \code{kgroup} from \code{kindx}.
#'   \item Corresponding rows are extracted from \code{x} and \code{y}.
#'   \item If \code{D} is provided, the subset of the dissimilarity matrix for
#'   the current group is used as augmented features.
#' }
#' @author Leonardo Ramirez-Lopez
#' @noRd
#' @keywords internal
ith_subsets_by_group <- function(
    x, 
    y, 
    kindx,
    kgroup,
    D = NULL
){
  
  it_kindx <- iter(kindx, by = "column")
  it_kgroup <- iter(kgroup, by = "column")
  it_xval <- iter(x, by = "row")
  it_D <- iter(D, by = "row")
  
  nextEl <- function() {
    knns <- nextElem(it_kindx)[nextElem(it_kgroup)]
    if(is.null(D)){
      list(x = x[knns, , drop = FALSE], 
           y = y[knns, , drop = FALSE], 
           xval = nextElem(it_xval)) 
    } else {
      idsm <- D[knns, knns]
      list(x = cbind(idsm, x[knns, , drop = FALSE]), 
           y = y[knns, , drop = FALSE], 
           xval = t(c(nextElem(it_D)[knns],
                      nextElem(it_xval))))
    }
  }
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}


#' @title Grouped KNN Subset Iterator with Chunk Size
#'
#' @description
#' Creates an iterator over grouped subsets of rows from `x` and `y` based on
#' nearest neighbor indices (`kindx`) and group selectors (`kgroup`). The
#' iteration is performed in consecutive column batches of size up to
#' `chunk_size` from the input matrices, and optionally incorporates a
#' dissimilarity matrix `D`.
#'
#' @param x A numeric matrix (n × p) representing the full reference dataset.
#' @param y A numeric matrix or vector (n × 1) of associated response values.
#' @param kindx An integer matrix (k × m) where each column contains indices of
#'   the `k` nearest neighbors for the corresponding observation.
#' @param kgroup A logical or integer matrix (k × m) indicating a group
#'   (subset) of rows from `kindx` to be selected for each observation.
#' @param neighborhood_sizes An integer vector of length `m` indicating the 
#' number of neighbors to consider for each observation. This is constant if the 
#' method of retaining by a fix number of neighbors is employed. If the method 
#' to retain neighbors is based on a distance threshold, this vector will have 
#' varying values.   
#' @param D Optional numeric matrix (n × n) representing a dissimilarity or
#'   distance matrix. If provided, its values for selected neighbors will be
#'   embedded into the `xval` vector.
#' @param chunk_size Integer. The maximum number of columns from `kindx` and
#'   `kgroup` to include in each batch. Defaults to 1. The last batch may be
#'   smaller if the total number of columns is not a multiple of `chunk_size`.
#'
#' @return An iterator object that returns, at each call of `nextElem()`, a list
#'   of lists. Each inner list contains:
#'   - `x`: the reference subset (`k × p` or `k × (p + k)` if `D` used),
#'   - `y`: the corresponding responses (`k × 1`),
#'   - `xval`: the target sample (`1 × p`) or extended with dissimilarities.
#'
#' @details
#' This iterator is useful when local models need to be built for each
#' observation using a selected group of its neighbors (`kgroup`) from `kindx`,
#' and optionally enhanced with pairwise distances via `D`.
#'
#' When `D` is provided, the function will include the distances between the
#' target observation and its selected neighbors in the `xval` element of each
#' result.
#'
#' @seealso [iterators::iter()], [iterators::nextElem()]
#' @noRd
#' @keywords internal
ith_subsets_by_group_list <- function(
    x, y, kindx, kgroup, neighborhood_sizes, D = NULL, chunk_size = 1
) {
  stopifnot(nrow(x) == nrow(y), ncol(kindx) == ncol(kgroup))
  if (!is.null(D)) stopifnot(nrow(D) == nrow(x), ncol(D) == nrow(x))
  
  m <- ncol(kindx)
  chunk_size <- as.integer(chunk_size); stopifnot(chunk_size > 0L)
  
  col_indices <- split(seq_len(m), ceiling(seq_len(m) / chunk_size))
  it_index <- iter(col_indices)
  
  nextEl <- function() {
    idx <- nextElem(it_index)
    
    kindx_sub  <- kindx[,  idx, drop = FALSE]
    kgroup_sub <- kgroup[, idx, drop = FALSE]
    kneighborhoods_sub <- neighborhood_sizes[idx]
    
    # Anchor sample per column = first row of kindx
    anchor_idx <- as.integer(kindx_sub[1L, ])
    xval_sub   <- x[anchor_idx, , drop = FALSE]
    
    chunks <- vector("list", length(idx))
    for (i in seq_along(idx)) {
      ith_k <- seq_len(kneighborhoods_sub[i])
      sel <- as.logical(kgroup_sub[, i])
      sel[-ith_k] <- FALSE # keep the non-neighbors out
      sel[is.na(sel)] <- FALSE
      sel[1L] <- FALSE  # never include anchor among neighbours
      
      knns <- as.integer(kindx_sub[sel, i])
      
      if (is.null(D)) {
        chunks[[i]] <- list(
          x = x[knns, , drop = FALSE],
          y = y[knns, , drop = FALSE],
          xval = xval_sub[i, , drop = FALSE]
        )
      } else {
        idsm  <- D[knns, knns, drop = FALSE]
        xdata <- cbind(idsm, x[knns, , drop = FALSE])
        chunks[[i]] <- list(
          x = xdata,
          y = y[knns, , drop = FALSE],
          xval = t(c(D[anchor_idx[i], knns], xval_sub[i, , drop = FALSE]))
        )
      }
    }
    chunks
  }
  
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubsetgroup", "abstractiter", "iter")
  obj
}


# ith_subsets_by_group_list <- function(
    #     x, y, kindx, kgroup, D = NULL, chunk_size = 1
# ) {
#   stopifnot(nrow(x) == nrow(y), ncol(kindx) == ncol(kgroup))
#   if (!is.null(D)) stopifnot(nrow(D) == nrow(x), ncol(D) == nrow(x))
#   
#   m <- ncol(kindx)
#   chunk_size <- as.integer(chunk_size); stopifnot(chunk_size > 0L)
#   
#   col_indices <- split(seq_len(m), ceiling(seq_len(m) / chunk_size))
#   it_index <- iter(col_indices)
#   
#   nextEl <- function() {
#     idx <- nextElem(it_index)
#     
#     kindx_sub  <- kindx[,  idx, drop = FALSE]
#     kgroup_sub <- kgroup[, idx, drop = FALSE]
#     
#     # Anchor sample per column = first row of kindx
#     anchor_idx <- as.integer(kindx_sub[1L, ])
#     xval_sub   <- x[anchor_idx, , drop = FALSE]
#     
#     chunks <- vector("list", length(idx))
#     for (i in seq_along(idx)) {
#       sel <- as.logical(kgroup_sub[, i]); sel[is.na(sel)] <- FALSE
#       sel[1L] <- FALSE  # never include anchor among neighbours
#       
#       knns <- as.integer(kindx_sub[sel, i])
#       
#       if (is.null(D)) {
#         chunks[[i]] <- list(
#           x = x[knns, , drop = FALSE],
#           y = y[knns, , drop = FALSE],
#           xval = xval_sub[i, , drop = FALSE]
#         )
#       } else {
#         idsm  <- D[knns, knns, drop = FALSE]
#         xdata <- cbind(idsm, x[knns, , drop = FALSE])
#         chunks[[i]] <- list(
#           x = xdata,
#           y = y[knns, , drop = FALSE],
#           xval = t(c(D[anchor_idx[i], knns], xval_sub[i, , drop = FALSE]))
#         )
#       }
#     }
#     chunks
#   }
#   
#   obj <- list(nextElem = nextEl)
#   class(obj) <- c("isubsetgroup", "abstractiter", "iter")
#   obj
# }




#' @title Iterator for Subsets by Index
#' @description
#' Internal helper to iterate over subsets of `x` and `y` based on `kindx`
#' indices. Optionally augments `x` with a dissimilarity matrix `D` and
#' filters out rows where `y` is NA.
#'
#' @param x A matrix or data frame of predictors (rows = observations).
#' @param y A vector or matrix of responses (same rows as `x`).
#' @param kindx A list or matrix of row indices (e.g., nearest neighbors).
#' @param D Optional square dissimilarity matrix matching rows of `x`.
#'
#' @return
#' An iterator object of class `isubset`, `abstractiter`, `iter`. Each
#' call to `nextElem()` produces:
#' \itemize{
#'   \item \code{x}: A subset of `x`, optionally prepended with dissimilarities.
#'   \item \code{y}: Corresponding subset of `y`, with NA rows removed.
#' }
#'
#' @details
#' For each iteration, indices from `kindx` are used to extract rows
#' from `x` and `y`. If `D` is given, `x` is augmented with local
#' dissimilarities from `D[knns, knns]`. Observations where `y` is NA
#' are excluded from both `x` and `y` in the returned subset.
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
ith_subsets <- function(
    x, 
    y, 
    kindx,
    D = NULL
){
  it_kindx <- iter(kindx, by = "column")
  it_xval <- iter(x, by = "row")
  it_D <- iter(D, by = "row")
  
  nextEl <- function() {
    knns <- nextElem(it_kindx)
    ixr <- x[knns, ]
    iyr <- y[knns, , drop = FALSE]
    if (!is.null(D)) {
      ixr <- cbind(D[knns, knns], ixr)
    }
    list(
      x = ixr[!is.na(iyr), ], 
      y = iyr[!is.na(iyr), , drop = FALSE]
    )
  }
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}



#' @title Chunked Iterator for Subsetting by Nearest Neighbors
#' @description
#' Creates an iterator that returns batched subsets of observations for
#' local modeling. Each batch corresponds to a set of consecutive indices
#' (rows) of size up to `chunk_size`. For each observation in the batch,
#' its corresponding nearest neighbors are used to extract the `x`, `y`,
#' and optional `D` values.
#'
#' @param x A numeric matrix of predictors (n × p), where `n` is the number of
#'   observations and `p` the number of features.
#' @param y A numeric vector or matrix of responses of length `n` or
#'   dimension `n × 1`.
#' @param kindx An integer matrix of size `k × n`, where each column contains
#'   the indices of the `k` nearest neighbors for the corresponding observation.
#' @param neighborhood_sizes An integer vector of length `m` indicating the 
#' number of neighbors to consider for each observation. This is constant if the 
#' method of retaining by a fix number of neighbors is employed. If the method 
#' to retain neighbors is based on a distance threshold, this vector will have 
#' varying values.   
#' @param D Optional numeric matrix of pairwise dissimilarities (n × n).
#'   If provided, for each subset the dissimilarities among the selected
#'   neighbors are prepended as additional columns to the predictors.
#' @param chunk_size Integer. Maximum number of observations per batch.
#'   The `n` observations are split into consecutive batches of size
#'   `chunk_size`, except for the final batch which may be smaller.
#'
#' @return An iterator object compatible with `nextElem()` from the
#'   **iterators** package. Each call to `nextElem()` returns a list of
#'   sublists (one per observation in the batch), where each sublist contains:
#'   - `x`: predictor matrix for that local subset (with optional dissimilarity)
#'   - `y`: response vector/matrix for that local subset
#'
#' @examples
#' \donttest{
#' library(iterators)
#' it <- ith_subsets_list(Xr, Yr, kindx, D, chunk_size = 100)
#' 
#' while (TRUE) {
#'   chunk <- tryCatch(nextElem(it), error = function(e) {
#'     if (inherits(e, "StopIteration")) return(NULL)
#'     stop(e)
#'   })
#'   if (is.null(chunk)) break
#'   print(length(chunk))  # Number of subsets in the batch
#' }
#' }
#' @noRd
#' @keywords internal
ith_subsets_list <- function(x, y, kindx, neighborhood_sizes, D = NULL, chunk_size = 1) {
  stopifnot(nrow(x) == nrow(y))
  if (!is.null(D)) stopifnot(nrow(D) == nrow(x), ncol(D) == nrow(x))
  
  # We iterate over columns of kindx (one local model per column)
  m <- ncol(kindx)
  stopifnot(is.numeric(chunk_size), length(chunk_size) == 1L)
  chunk_size <- as.integer(chunk_size)
  stopifnot(chunk_size > 0L)
  
  # Build consecutive column-index chunks of size `chunk_size`
  col_chunks <- split(seq_len(m), ceiling(seq_len(m) / chunk_size))
  it_chunks <- iter(col_chunks)
  
  nextEl <- function() {
    idx <- nextElem(it_chunks)  # indices of columns in `kindx`
    lapply(idx, function(i) {
      knns <- as.integer(kindx[seq_len(neighborhood_sizes[i]), i])
      ixr  <- x[knns, , drop = FALSE]
      iyr  <- y[knns, , drop = FALSE]
      
      if (!is.null(D)) {
        ixr <- cbind(D[knns, knns, drop = FALSE], ixr)
      }
      
      keep <- !is.na(iyr)
      list(
        x = ixr[keep, , drop = FALSE],
        y = iyr[keep, , drop = FALSE]
      )
    })
  }
  
  obj <- list(nextElem = nextEl)
  class(obj) <- c("chunked_isubset", "abstractiter", "iter")
  obj
}



#' @title Iterator for Prediction Subsets (PLS Local Models)
#' @description
#' Internal helper to iterate over prediction inputs for local PLS models.
#'
#' @param plslib Matrix of PLS regression vectors (one per reference spectrum).
#' @param Xu Matrix of new observations to predict (rows = samples).
#' @param xunn List of integer vectors containing nearest neighbor indices for 
#'   each row of Xu.
#' @param xscale Matrix of scaling vectors for each reference model.
#' @param dxrxu Optional dissimilarity matrix (ref vs. new samples).
#'
#' @return
#' An iterator of class `isubset`, `abstractiter`, `iter`. Each call
#' to `nextElem()` produces:
#' \itemize{
#'   \item \code{iplslib}: Local PLS regression vectors for neighbors.
#'   \item \code{ixscale}: Scaling vectors for those models.
#'   \item \code{ixunn}: Indices of nearest neighbors for prediction.
#'   \item \code{ixu}: New observation from `Xu` to predict.
#'   \item \code{idxrxu}: (Optional) dissimilarity vector to neighbors.
#' }
#'
#' @details
#' For each new sample in `Xu`, retrieves its nearest neighbors, along
#' with corresponding PLS vectors, scaling vectors, and optional
#' dissimilarity values. Intended for local PLS prediction routines.
#' @author Leonardo Ramirez-Lopez
#' @noRd
#' @keywords internal
ith_pred_subsets <- function(
    plslib,
    Xu,
    xunn,
    xscale,
    dxrxu = NULL
) {
  n <- length(xunn)
  i <- 0L
  
  nextEl <- function() {
    i <<- i + 1L
    if (i > n) stop("StopIteration", call. = FALSE)
    
    ixunn <- xunn[[i]]
    ixu <- Xu[i, , drop = TRUE]
    
    if (!is.null(dxrxu)) {
      idxrxu <- dxrxu[ixunn, i]
    } else {
      idxrxu <- NULL
    }
    
    list(
      iplslib = plslib[ixunn, , drop = FALSE],
      ixscale = xscale[ixunn, , drop = FALSE],
      ixunn = ixunn,
      ixu = ixu,
      idxrxu = idxrxu
    )
  }
  
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}

# knns <- kindx[,..ii..]
# iyr <- y[knns, ,drop = FALSE]
# ixr <- cbind(D[knns, knns], x[knns, ])
# ixr <- ixr[!is.na(iyr), ]
# iyr <- iyr[!is.na(iyr), , drop = FALSE]
# 




# #' @title Quantile Stats of Neighbor Responses
# #' @description
# #' Computes quantiles of the response values among the filtered
# #' nearest neighbors for a specific sample.
# #'
# #' @param ..i.. Index of the sample being analyzed.
# #' @param kidxmat Matrix of nearest neighbor indices (rows = neighbors).
# #' @param kidxgrop Matrix mask to filter out same-group neighbors (TRUE = keep).
# #' @param Yr Numeric response vector or 1-col matrix.
# #' @param k Vector of neighbor counts for each sample.
# #'
# #' @return
# #' A matrix of quantiles (0%, 5%, 25%, 50%, 75%, 95%, 100%) of the
# #' response values among the valid nearest neighbors.
# #'
# #' @details
# #' For the i-th sample, this function selects the top-k neighbors for
# #' each sample, filters using the group mask, and computes quantile
# #' statistics from their corresponding response values in `Yr`.
# #' @author Leonardo Ramirez-Lopez
# #' @keywords internal
# i_nn_stats <- function(..i.., kidxmat, kidxgrop, Yr, k) {
#   ik <- k[..i..]
#   
#   jstats <- sapply(
#     1:ncol(kidxmat),
#     FUN = function(..j.., kidxmat, kidxgrop, Yr, ik){
#       inn <- kidxmat[1:ik,..j..]
#       inn <- inn[kidxgrop[1:ik,..j..]]
#       quantile(Yr[inn,], c(0, 0.05, 0.25, 0.5, 0.75, 0.95, 1.00))
#       #sd(Yr[inn,])
#     }, 
#     kidxgrop = kidxgrop,
#     kidxmat = kidxmat,
#     Yr = Yr,
#     ik = ik
#   )
#   t(jstats)
# }

#' @title Internal: Fit weighted local PLS models for all reference samples
#'
#' @description
#' Computes predictions from weighted local PLS models for each observation in
#' the reference set \code{Xr}, based on its nearest neighbors and optionally
#' using a dissimilarity matrix as extra predictive information.
#'
#' @param ..k.. Index (integer) indicating which element of \code{k} to use
#'   for the number of neighbors.
#' @param Xr A numeric matrix of predictor variables for the reference data
#'   (rows = observations, columns = variables).
#' @param Yr A numeric vector or single-column matrix of response values
#'   corresponding to \code{Xr}.
#' @param k A matrix where each row corresponds to a given neighbor size which 
#' can be constant if the method of retaining neighbors is based on a fix number 
#' of neighbors, or varying if the method of retaining neighbors is based on a 
#' disimilarity threshold.
#' @param ncomp_min Minimum number of PLS components to use.
#' @param ncomp_max Maximum number of PLS components to use.
#' @param emgrid A binary matrix indicating which component combinations to
#'   evaluate. Each row represents a combination; columns correspond to
#'   components from \code{ncomp_min} to \code{ncomp_max}.
#' @param scale Logical; if \code{TRUE}, scale predictors before regression.
#' @param max_iter Maximum number of iterations for the PLS algorithm.
#' @param tol Convergence tolerance for the PLS algorithm.
#' @param algorithm Character string specifying the PLS fitting algorithm to 
#' use. Options are \code{'mpls'} (default), \code{'pls'} or \code{'nipals'}.
#' @param kidxmat Integer matrix of neighbor indices (rows = neighbors,
#'   columns = observations).
#' @param kidxgrop Logical matrix indicating which neighbors should be
#'   excluded (e.g., same group as target for leave-group-out validation).
#' @param dissimilarity_mat Optional numeric matrix of dissimilarities between
#'   reference observations. Used to augment predictors when dissimilarity
#'   is used as additional predictive information. Default is \code{NULL}.
#' @param pb A progress bar object (e.g., from \code{txtProgressBar()}) or
#'   \code{NULL} to suppress progress display.
#' @param chunk_size Integer; number of observations to process per parallel
#'   chunk. Default is \code{1L}.
#' @param ... Further arguments (currently unused).
#'
#' @return A numeric matrix of predicted values. Rows correspond to component
#'   combinations in \code{emgrid}; columns correspond to observations in
#'   \code{Xr}.
#'
#' @details
#' For each reference observation in \code{Xr}, the function:
#' \enumerate{
#'   \item Builds a local training subset using \code{ith_subsets_by_group_list()}
#'   \item Fits a local weighted PLS model via \code{ith_local_fit()}
#'   \item Produces predictions for each component combination in \code{emgrid}
#' }
#'
#' Parallel computation is handled via \code{foreach} and \code{\%dopar\%}.
#' Register a parallel backend (e.g., via \pkg{doParallel}) before calling
#' \code{liblex()} to enable parallelization.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @keywords internal
#' @noRd
.get_all_fits <- function(
    ..k..,
    Xr,
    Yr,
    k,
    ncomp_min,
    ncomp_max,
    emgrid,
    scale,
    max_iter,
    tol,
    algorithm,
    kidxmat,
    kidxgrop,
    dissimilarity_mat = NULL,
    pb = NULL,
    chunk_size = 1L,
    ...
) {
  
  neighborhood_sizes <- k[..k.., ]
  ik <- seq_len(max(neighborhood_sizes))
  
  if (!is.null(pb)) {
    setTxtProgressBar(pb, ..k..)
  }
  
  n_xr <- nrow(Xr)
  ith_preds_template <- matrix(
    NA_real_,
    nrow = nrow(emgrid),
    ncol = chunk_size
  )
  
  innpreds <- foreach(
    i = seq_len(n_xr),
    ksubsets = ith_subsets_by_group_list(
      x = Xr,
      y = Yr,
      kindx = kidxmat[ik, , drop = FALSE],
      kgroup = kidxgrop[ik, , drop = FALSE],
      neighborhood_sizes = neighborhood_sizes, 
      D = dissimilarity_mat,
      chunk_size = chunk_size
    ),
    .export = c(
      "ith_pred_subsets",
      "ith_subsets",
      "ith_subsets_by_group",
      ".get_all_fits",
      "ith_local_fit",
      "opls"
    )
  ) %dopar% {
    ith_preds <- ith_preds_template
    for (j in seq_along(ksubsets)) {
      ith_preds[, j] <- ith_local_fit(
        X = ksubsets[[j]]$x,
        Y = ksubsets[[j]]$y,
        xval = ksubsets[[j]]$xval,
        emgrid = emgrid,
        ncomp_max = ncomp_max,
        ncomp_min = ncomp_min,
        scale = scale,
        max_iter = max_iter,
        tol = tol, 
        algorithm = algorithm
      )
    }
    if (j < ncol(ith_preds)) {
      ith_preds <- ith_preds[, seq_len(j), drop = FALSE]
    }
    ith_preds
  }
  
  innpreds <- do.call("cbind", innpreds)
  innpreds
}

# #' @title Internal: Fit a local weighted PLS model and predict for a query point
# #'
# #' @description
# #' Fits a local Partial Least Squares (PLS) model using a neighborhood subset
# #' and computes a weighted prediction for a target sample. The weighting is
# #' done over multiple components using a provided evaluation grid.
# #'
# #' @param ilocalsubset A list with elements:
# #' \itemize{
# #'   \item{\code{x}}: matrix of predictors from the local neighborhood.
# #'   \item{\code{y}}: vector or 1-col matrix of corresponding responses.
# #'   \item{\code{xval}}: query sample (usually one row) to predict.
# #' }
# #' @param min_component Minimum number of PLS components to use in prediction.
# #' @param max_component Maximum number of PLS components to fit.
# #' @param emgrid A numeric matrix used to weight component-wise predictions.
# #' @param scale Logical; whether to scale predictors before PLS fitting.
# #' @param maxiter Maximum number of iterations for the PLS algorithm.
# #' @param tol Convergence tolerance for the PLS algorithm.
# #' @param pc_selection A list defining the principal component selection
# #'        strategy. (Included for consistency, not used directly here.)
# #' @param ... Additional arguments (currently unused).
# #'
# #' @return
# #' A numeric matrix of weighted predictions (rows = grid rows, 1 column).
# #'
# #' @details
# #' The function performs the following steps:
# #' \enumerate{
# #'   \item Fits a PLS model on `ilocalsubset$x` and `ilocalsubset$y` with
# #'         up to `max_component` components.
# #'   \item Computes component weights for `ilocalsubset$xval` using
# #'         \code{get_local_pls_weights()}.
# #'   \item Predicts component-wise responses using \code{predict_opls()}.
# #'   \item Applies a weighted average of predictions using `emgrid` and the
# #'         component weights.
# #' }
# #' The output is a matrix of weighted predictions using the weighted model grid.
# #'
# #' @author Leonardo Ramirez-Lopez
# #'
# #' @keywords internal
# ith_local_fit <- function(
    #     ilocalsubset,
#     min_component,
#     max_component,
#     emgrid,
#     scale, 
#     maxiter, 
#     tol, 
#     pc_selection,
#     ...
# ){ 
#   
#   ipreds <- run_ipls_prediction(
#     X = ilocalsubset$x,
#     Y = ilocalsubset$y,
#     xval = ilocalsubset$xval,
#     emgrid = emgrid,
#     max_component = max_component,
#     min_component = min_component,
#     scale = scale,
#     maxiter = maxiter,
#     tol = tol
#   )
#   ipreds
# }


# #' @title Internal: Compute weighted final PLS model outputs
# #' 
# #' @description
# #' Fits a PLS model on a local neighborhood and computes a weighted combination
# #' of regression coefficients, intercept, VIP scores, selectivity ratios, and
# #' predictor scales based on component-specific weights.
# #' 
# #' @param ilocalsubset A list with local subset data:
# #' \itemize{
# #'   \item \code{x}: matrix of predictors from the local neighborhood.
# #'   \item \code{y}: response vector or 1-col matrix for local samples.
# #' }
# #' @param min_component Integer; minimum number of components used for output.
# #' @param max_component Integer; maximum number of components to fit in PLS.
# #' @param scale Logical; whether to scale predictors before PLS fitting.
# #' @param maxiter Integer; maximum number of iterations for the PLS algorithm.
# #' @param tol Numeric; convergence tolerance for the PLS algorithm.
# #' @param ... Additional arguments (currently unused).
# #' 
# #' @return
# #' A numeric vector containing the following, in order:
# #' \itemize{
# #'   \item Intercept term (weighted average of component intercepts)
# #'   \item Weighted regression coefficients
# #'   \item Weighted VIP scores (normalized by column SDs)
# #'   \item Weighted selectivity ratios (normalized by column SDs)
# #'   \item Predictor scaling values (from training transformation)
# #' }
# #' 
# #' @details
# #' This function fits a PLS model on the local neighborhood using
# #' \code{opls_get_all()}, computes local weights with
# #' \code{get_local_pls_weights()}, and combines all component-wise metrics
# #' into a final weighted output.
# #' 
# #' Component indices are selected using the range
# #' \code{min_component:max_component}. Weighting is applied to each component
# #' across coefficients, VIPs, and selectivity ratios. All component-wise scores
# #' are normalized before aggregation. The result can be used as a compressed
# #' representation of the final prediction model.
# #' 
# #' Note: the prediction target is assumed to be \code{x[1, ]} of the subset.
# #' 
# #' @author Leonardo Ramirez-Lopez
# #' 
# #' @keywords internal
# final_fits <- function(
    #     ilocalsubset,
#     min_component, 
#     max_component, 
#     scale, 
#     maxiter, 
#     tol, 
#     ...
# ){
#   
#   result <- final_fits_cpp(
#     X = ilocalsubset$x,
#     Y = ilocalsubset$y,
#     new_x = ilocalsubset$x[1, , drop = FALSE],
#     min_component,
#     max_component,
#     scale = scale,
#     maxiter = maxiter,
#     tol = tol
#   )
#   unlist(result, recursive = FALSE, use.names = FALSE)
# }


# #' @title Internal: Predict using local PLS coefficients and optional
# #' dissimilarities
# #'
# #' @description
# #' Computes a prediction for a new observation using a local PLS model
# #' represented by coefficients (`plslib`). The prediction is based on
# #' inverse-scaled feature values. If a dissimilarity vector is provided, it is
# #' appended to the input features before inverse scaling.
# #'
# #' @param plslib A matrix of PLS model coefficients. First column is the
# #'        intercept; remaining columns are coefficients for scaled features.
# #' @param xscale A matrix of scaling values for each feature (same dimensions
# #'        as \code{plslib[,-1]}).
# #' @param Xu A numeric vector (or single-row matrix) representing the query
# #'        sample to be predicted.
# #' @param xunn Currently unused; placeholder for nearest neighbor indices.
# #' @param dxrxu Optional numeric vector of dissimilarities between `Xu` and
# #'        reference samples. If provided, used as additional predictive
# #'        features.
# #' @param ... Additional arguments (currently ignored).
# #'
# #' @return A numeric scalar containing the predicted response for `Xu`.
# #'
# #' @details
# #' The function constructs a synthetic predictor vector by dividing
# #' `xscale` by `Xu` (or by the concatenation of `dxrxu` and `Xu`), then
# #' inverting the result. This reflects a distance- or similarity-based
# #' transformation. The resulting vector is then multiplied element-wise with
# #' the PLS coefficients, and the sum is added to the intercept to compute
# #' the final prediction.
# #'
# #' @author Leonardo Ramirez-Lopez
# #'
# #' @keywords internal
# ith_pred <- function(plslib, xscale, Xu, xunn, dxrxu = NULL, ...){
#   if (is.null(dxrxu)) {
#     ixus <- sweep(xscale, MARGIN = 2, STATS =  Xu, FUN = "/")
#     ixus <- 1 / ixus
#     ipred <- plslib[, 1] + rowSums(plslib[, -1] * ixus)
#   } else {
#     dxu <- c(dxrxu, Xu)
#     ixus <- sweep(xscale, MARGIN = 2, STATS =  dxu, FUN = "/")
#     ixus <- 1 / ixus
#     ipred <- plslib[, 1] + rowSums(plslib[, -1] * ixus)
#   }
#   ipred
# }





#' Compute weighted quantiles
#'
#' Computes weighted quantiles from a numeric vector and associated weights.
#' Observations are sorted, weights are normalized to sum to 1, and
#' cumulative weights are used to determine quantile positions via linear
#' interpolation.
#' @usage
#' weighted_quantiles(x, weights,
#'     probs = c(0.05, 0.25, 0.5, 0.75, 0.95))
#' @param x A numeric vector of observations.
#' @param weights A numeric vector of non-negative weights, the same length as `x`.
#'   Weights are internally normalized to sum to 1.
#' @param probs A numeric vector of probabilities in \eqn{[0, 1]} specifying
#'   the quantile levels to compute for the response values within each
#'   neighbourhood (e.g., \code{c(0.25, 0.5, 0.75)} for quartiles).
#' @param exclude_last Logical; if `TRUE`, the last non-NA observation is excluded 
#' from each observation to avoid edge effects.
#'
#' @return A numeric vector of weighted quantiles corresponding to `probs`.
#'
#' @details
#' The function uses a weighted version of the empirical cumulative
#' distribution function (CDF). It sorts `x` and applies normalized weights
#' to compute cumulative sums, which are then used to interpolate the
#' quantiles at specified `probs`. Ties in cumulative weights are averaged.
#'
#' This function is designed for internal use and is not exported.
#'
#' @examples
#' x <- c(1, 2, 3, 4, 5)
#' w <- c(1, 1, 2, 2, 4)
#' weighted_quantiles(x, w, probs = c(0.25, 0.5, 0.75))
#' @noRd
#' @keywords internal
weighted_quantiles <- function(
    x,
    weights,
    probs = c(0.05, 0.25, 0.5, 0.75, 0.95), 
    exclude_last = TRUE
) {
  stopifnot(is.matrix(x), is.matrix(weights))
  stopifnot(all(dim(x) == dim(weights)))
  stopifnot(all(probs >= 0 & probs <= 1))

  n_rows <- nrow(x)
  n_probs <- length(probs)
  result <- matrix(NA_real_, nrow = n_rows, ncol = n_probs)
  colnames(result) <- paste0("q", probs * 100)
  
  # Collapse duplicate weights by averaging associated x values
  collapse_duplicates <- function(w, x) {
    uw <- unique(w)
    agg_x <- vapply(uw, function(wi) {
      mean(x[w == wi])
    }, numeric(1))
    list(w = uw, x = agg_x)
  }
  
  for (i in seq_len(n_rows)) {
    xi <- x[i, ]
    wi <- weights[i, ]
    
    # Remove NA values
    valid <- !(is.na(xi) | is.na(wi))
    if (exclude_last) {
      valid[sum(valid)] <- FALSE
    }
    xi <- xi[valid]
    wi <- wi[valid]
    
    if (length(xi) == 0 || sum(wi) == 0) next
    
    # Order by xi
    ord <- order(xi)
    xi <- xi[ord]
    wi <- wi[ord]
    
    # Normalize weights
    w_cum <- cumsum(wi) / sum(wi)
    
    # Collapse duplicated cumulative weights
    if (any(duplicated(w_cum))) {
      collapsed <- collapse_duplicates(w_cum, xi)
      w_cum <- collapsed$w
      xi <- collapsed$x
    }
    
    # Compute quantiles via interpolation
    qvals <- numeric(n_probs)
    for (j in seq_along(probs)) {
      p <- probs[j]
      if (p <= min(w_cum)) {
        qvals[j] <- xi[1]
      } else if (p >= max(w_cum)) {
        qvals[j] <- xi[length(xi)]
      } else {
        qvals[j] <- approx(w_cum, xi, xout = p)$y
      }
    }
    
    result[i, ] <- qvals
  }
  
  return(result)
}

#' Cap ncomp bounds against data dimensions
#' @param ncomp_obj A ncomp_selection object
#' @param n_obs Number of observations
#' @param n_vars Number of variables
#' @param verbose Logical; print messages when capping
#' @return Validated ncomp_obj (possibly with capped max_ncomp)
#' @keywords internal
#' @noRd
#' @keywords internal
#' @noRd
cap_ncomp_to_data <- function(ncomp_obj, n_obs, n_vars, verbose = TRUE) {
  max_dim <- min(n_obs, n_vars)
  
  if (inherits(ncomp_obj, "ncomp_fixed")) {
    if (ncomp_obj$ncomp > max_dim) {
      stop("ncomp_fixed(", ncomp_obj$ncomp, ") exceeds maximum (",
           max_dim, ")", call. = FALSE)
    }
  } else if (inherits(ncomp_obj, "ncomp_selection")) {
    if (ncomp_obj$max_ncomp > max_dim) {
      if (verbose) {
        message("max_ncomp capped from ", ncomp_obj$max_ncomp, " to ", max_dim)
      }
      ncomp_obj$max_ncomp <- max_dim
    }
  }
  
  ncomp_obj
}

# =============================================================================
# Correlation dissimilarity — aligned with shared constructor pattern
# =============================================================================
# Public:   diss_correlation()   — method constructor
# Internal: .cor_diss_compute()  — raw computation, not exported


#' @title Correlation dissimilarity method constructor
#'
#' @description
#' Creates a configuration object that fully specifies a correlation (or moving
#' correlation) dissimilarity method. Pass the result to \code{dissimilarity()}
#' to compute the dissimilarity matrix.
#'
#' @param ws Either \code{NULL} (default) or an odd integer greater than 2.
#'   When \code{NULL}, standard Pearson correlation dissimilarity is used.
#'   When an odd integer is provided, a moving (rolling) correlation
#'   dissimilarity is computed using a window of that size.
#' @param center Logical. Should the data be mean-centered before computing
#'   dissimilarities? Centering is applied jointly to \code{Xr} and \code{Xu}
#'   (if provided) based on their combined column means. Default is \code{TRUE}.
#' @param scale Logical. Should the data be scaled (divided by column standard
#'   deviations) before computing dissimilarities? Scaling is applied jointly
#'   to \code{Xr} and \code{Xu} (if provided). Default is \code{FALSE}.
#'
#' @return An object of class \code{c("diss_correlation", "diss_method")} — a
#'   list holding the validated method parameters. Intended to be passed to
#'   \code{dissimilarity()}, not used directly.
#'
#' @details
#' The correlation dissimilarity between two observations \eqn{x_i} and
#' \eqn{x_j} is:
#'
#' \deqn{d(x_i, x_j) = \frac{1}{2}(1 - \rho(x_i, x_j))}
#'
#' where \eqn{\rho} is the Pearson correlation coefficient. This is used when
#' \code{ws = NULL}.
#'
#' When \code{ws} is specified, the moving correlation dissimilarity is:
#'
#' \deqn{d(x_i, x_j; ws) = \frac{1}{2\,ws} \sum_{k=1}^{p - ws}
#'   \bigl(1 - \rho(x_{i,(k:k+ws)},\, x_{j,(k:k+ws)})\bigr)}
#'
#' where \eqn{ws} is the window size and \eqn{p} is the number of variables.
#'
#' @section Parallel execution:
#' The underlying C++ implementation uses OpenMP for parallel computation.
#' Thread count is controlled by the \code{OMP_NUM_THREADS} environment
#' variable. To limit threads (e.g., when calling from within a parallel
#' backend):
#' \preformatted{
#' Sys.setenv(OMP_NUM_THREADS = 1)
#' }
#'
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @seealso \code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
#'   \code{\link{diss_mahalanobis}}, \code{\link{diss_cosine}}
#'
#' @examples
#' # Standard correlation dissimilarity
#' m <- diss_correlation()
#'
#' # Moving correlation with window size 41
#' m <- diss_correlation(ws = 41)
#'
#' # Without centering
#' m <- diss_correlation(center = FALSE)
#'
#' @export
diss_correlation <- function(
    ws     = NULL,
    center = TRUE,
    scale  = FALSE
) {
  
  # --- validate ws -----------------------------------------------------------
  if (!is.null(ws)) {
    if (!is.numeric(ws) || length(ws) != 1L) {
      stop("'ws' must be a single integer value or NULL.")
    }
    ws <- as.integer(ws)
    if (ws < 3L) {
      stop("'ws' must be an odd integer greater than 2.")
    }
    if (ws %% 2L == 0L) {
      stop("'ws' must be an odd integer (e.g. 3, 5, 11, 41, ...).")
    }
  }
  
  # --- delegate center/scale validation and object construction --------------
  .new_diss_method(
    "correlation",
    center = center,
    scale  = scale,
    extra  = list(ws = ws)
  )
}


# --- print method ------------------------------------------------------------
#' @export
print.diss_correlation <- function(x, ...) {
  ws_label <- if (is.null(x$ws)) "full (no moving window)" else x$ws
  cat(
    "Dissimilarity method: correlation\n",
    "  window size (ws) :", ws_label, "\n",
    "  center           :", x$center, "\n",
    "  scale            :", x$scale, "\n"
  )
  invisible(x)
}


# =============================================================================
# Internal compute function — NOT exported
# =============================================================================
#
# Called by dissimilarity() after data-level validation has been done there.
# Assumes:
#   - Xr and Xu are numeric matrices with no NAs
#   - ncol(Xu) == ncol(Xr) if Xu is provided
#   - method$ws < ncol(Xr) if ws is not NULL (checked in dissimilarity())
#
.cor_diss_compute <- function(Xr, Xu = NULL, method) {
  
  stopifnot(inherits(method, "diss_correlation"))
  
  ws     <- method$ws
  center <- method$center
  scale  <- method$scale
  
  # --- preprocessing ---------------------------------------------------------
  if (center || scale) {
    X <- rbind(Xr, Xu)
    
    if (center) {
      X <- sweep(X, MARGIN = 2, STATS = colMeans(X), FUN = "-")
    }
    if (scale) {
      col_sds <- get_col_sds(X)
      if (any(col_sds == 0)) {
        stop("Cannot scale: one or more columns have zero standard deviation.")
      }
      X <- sweep(X, MARGIN = 2, STATS = col_sds, FUN = "/")
    }
    
    if (!is.null(Xu)) {
      n_xu <- nrow(Xu)
      Xu   <- X[(nrow(X) - n_xu + 1L):nrow(X), , drop = FALSE]
      Xr   <- X[1L:(nrow(X) - n_xu), , drop = FALSE]
    } else {
      Xr <- X
    }
    rm(X)
  }
  
  # --- zero-sd guard (row-level for correlation) -----------------------------
  xr_row_sds <- get_col_sds(t(Xr))
  if (any(xr_row_sds == 0)) {
    stop(
      sprintf(
        "Correlation cannot be computed: Xr contains %d observation(s) with zero variance.",
        sum(xr_row_sds == 0)
      )
    )
  }
  
  if (!is.null(Xu)) {
    xu_row_sds <- get_col_sds(t(Xu))
    if (any(xu_row_sds == 0)) {
      stop(
        sprintf(
          "Correlation cannot be computed: Xu contains %d observation(s) with zero variance.",
          sum(xu_row_sds == 0)
        )
      )
    }
  }
  
  # --- resolve full-correlation window size ----------------------------------
  if (is.null(ws)) {
    ws <- ncol(Xr)
  }
  
  # --- dispatch to C++ backend -----------------------------------------------
  if (!is.null(Xu)) {
    result <- moving_cor_diss_xy(
      Xu, Xr, ws,
      compute_block_rows(dim(Xu)),
      compute_block_rows(dim(Xr))
    )
    rownames(result) <- paste("Xr", seq_len(nrow(Xr)), sep = "_")
    colnames(result) <- paste("Xu", seq_len(nrow(Xu)), sep = "_")
  } else {
    result <- moving_cor_diss_self_f64(Xr, ws, compute_block_rows(dim(Xr)))
    rownames(result) <- colnames(result) <- paste("Xr", seq_len(nrow(Xr)), sep = "_")
  }
  
  result
}


#' Heuristic for cache-aware tile height (block_rows)
#'
#' @description
#' Choose a tile height for tiled rolling-window correlation/distance kernels,
#' balancing per-tile cache footprint against matrix width and available
#' threads. Cross-platform (macOS, Linux, Windows).
#'
#' @details
#' Let \eqn{b} be the tile height and \eqn{T} the number of columns.
#' The working set per tile (in elements) is approximated as:
#'
#' \deqn{M(b) \approx 3 b^2 + 2 b T} if \code{include_squares = TRUE}
#' (materialised \eqn{X_i^2}, \eqn{X_j^2}), and
#' \deqn{M(b) \approx 3 b^2} otherwise.
#'
#' A per-thread budget \eqn{K} (elements) is obtained from \code{target_mb},
#' divided by the effective OpenMP thread count. The quadratic bound yields
#' \deqn{b_T = \frac{-T + \sqrt{T^2 + 3K}}{3}}
#' (or \eqn{b_T = \sqrt{K/3}} when \code{include_squares = FALSE}).
#'
#' The final choice is
#' \code{block_rows = round_to_64( min( max(64, m / min_tiles), b_T ) )},
#' clamped to \code{[64, min(max_block_cap, m)]}.
#'
#' If OpenMP is unavailable (\code{capabilities("openmp") == FALSE}),
#' the budget assumes a single thread.
#'
#' @param dimensions Integer vector of length two \code{c(m, T)}:
#'   number of rows and columns of \eqn{X}.
#' @param include_squares Logical. If \code{TRUE}, assumes the kernel
#'   materialises \code{Xi_sq} and \code{Xj_sq}; if \code{FALSE}, uses the
#'   lighter memory model.
#' @param target_mb Numeric. Target MiB of cache to devote per thread to one
#'   tile. Default is 8.
#' @param min_tiles Integer. Aim for at least this many tiles along the
#'   row dimension (default \code{12L}).
#' @param max_block_cap Integer. Hard upper bound for \code{block_rows}
#'   (default \code{1024L}).
#'
#' @return Integer scalar, the recommended \code{block_rows} (multiple of 64).
#'
#' @section Thread detection:
#' Thread count is determined by:
#' \enumerate{
#'   \item \code{OMP_NUM_THREADS} environment variable (if set)
#'   \item \code{parallel::detectCores(logical = FALSE)} (physical cores)
#'   \item Falls back to 1 if detection fails or OpenMP is unavailable
#' }
#'
#' @keywords internal
#' @noRd
compute_block_rows <- function(
    dimensions,
    include_squares = TRUE,
    target_mb       = 8,
    min_tiles       = 12L,
    max_block_cap   = 1024L
) {
  mm <- as.integer(dimensions[1])
  tt <- as.integer(dimensions[2])
  
  max_block <- min(max_block_cap, mm)
  if (max_block < 64L) return(max_block)
  
  # --- detect effective thread count -----------------------------------------
  threads <- suppressWarnings(as.integer(Sys.getenv("OMP_NUM_THREADS", NA)))
  if (is.na(threads) || threads < 1L) {
    threads <- parallel::detectCores(logical = FALSE)
    if (is.na(threads) || threads < 1L) threads <- 1L
  }
  if (!isTRUE(capabilities("openmp"))) threads <- 1L
  
  # --- per-thread element budget (double -> 8 bytes) -------------------------
  elt_size <- 8
  per_thread_mb <- max(8, as.numeric(target_mb) / threads)
  K <- (per_thread_mb * 1024^2) / elt_size
  
  # --- working-set model bound -----------------------------------------------
  b_T <- if (isTRUE(include_squares)) {
    (-tt + sqrt(tt * tt + 3 * K)) / 3
  } else {
    sqrt(K / 3)
  }
  
  # --- ensure enough tiles along rows ----------------------------------------
  b_m   <- mm / as.numeric(min_tiles)
  b_raw <- min(max(64, b_m), b_T)
  
  block <- round_to_64(b_raw)
  block <- max(64L, min(as.integer(block), max_block))
  block
}


#' Round to nearest multiple of 64
#' @param x Numeric value to round.
#' @param mult Integer multiple (default 64).
#' @return Integer rounded to nearest multiple.
#' @keywords internal
#' @noRd
round_to_64 <- function(x, mult = 64L) {
  as.integer(mult * round(x / mult))
}

#' @title Evolutionary sample search for context-specific calibrations
#' @name gesearch
#' @aliases gesearch gesearch.default gesearch.formula
#' @aliases predict.gesearch plot.gesearch
#'
#' @description
#' Implements an evolutionary search algorithm that selects a subset from large
#' reference datasets (e.g., spectral libraries) to build context-specific
#' calibrations. The algorithm iteratively removes weak or non-informative
#' samples based on prediction error, spectral reconstruction error, or
#' dissimilarity criteria. This implementation is based on the methods proposed 
#' in Ramirez-Lopez et al. (2026a).
#' @usage
#'
#' \method{gesearch}{default}(Xr, Yr, Xu, Yu = NULL, Yu_lims = NULL,
#'          k, b, retain = 0.95, target_size = k,
#'          fit_method = fit_pls(ncomp = 10),
#'          optimization = "reconstruction",
#'          group = NULL, control = gesearch_control(),
#'          intermediate_models = FALSE,
#'          verbose = TRUE, seed = NULL, pchunks = 1L, ...)
#'
#' \method{gesearch}{formula}(formula, train, test, k, b, target_size, fit_method,
#'          ..., na_action = na.pass)
#'
#' \method{predict}{gesearch}(object, newdata, type = "response",
#'          what = c("final", "all_generations"), ...)
#'
#' \method{plot}{gesearch}(x, which = c("weakness", "removed"), ...)
#'
#' @param Xr A numeric matrix of predictor variables for the reference data
#'   (observations in rows, variables in columns).
#' @param Yr A numeric vector or single-column matrix of response values
#'   corresponding to \code{Xr}. Only one response variable is supported.
#' @param Xu A numeric matrix of predictor variables for target observations
#'   (same structure as \code{Xr}).
#' @param Yu An optional numeric vector or single-column matrix of response 
#'   values for \code{Xu}. Required when \code{optimization} includes 
#'   \code{"response"}. Default is \code{NULL}.
#' @param Yu_lims A numeric vector of length 2 specifying expected response
#'   limits for the target population. Used with \code{optimization = "range"}.
#' @param k An integer specifying the number of samples in each resampling
#'   subset (gene size).
#' @param b An integer specifying the target average number of times each
#'   training sample is evaluated per iteration. Higher values (e.g., >40)
#'   produce more stable results but increase computation time.
#' @param retain A numeric value in (0, 1] specifying the proportion of samples
#'   retained per iteration. Default is 0.95. Values >0.9 are recommended for
#'   stability. See \code{\link{gesearch_control}} for retention strategy.
#' @param target_size An integer specifying the target number of selected
#'   samples (gene pool size). Must be >= \code{k}. Default is \code{k}.
#' @param fit_method A fit method object created with \code{\link{fit_pls}}.
#'   Specifies the regression model and scaling used during the search.
#'   Currently only \code{fit_pls()} is supported.
#' @param optimization A character vector specifying optimization criteria:
#'   \itemize{
#'     \item \code{"reconstruction"}: (default) Retains samples based on
#'       spectral reconstruction error of \code{Xu} in PLS space.
#'     \item \code{"response"}: Retains samples based on RMSE of predicting
#'       \code{Yu}. Requires \code{Yu}.
#'     \item \code{"similarity"}: Retains samples based on Mahalanobis distance
#'       between \code{Xu} and training samples in PLS score space.
#'     \item \code{"range"}: Removes samples producing predictions outside
#'       \code{Yu_lims}.
#'   }
#'   Multiple criteria can be combined, e.g.,
#'   \code{c("reconstruction", "similarity")}.
#' @param group An optional factor assigning group labels to training
#'   observations. Used for leave-group-out cross-validation to avoid
#'   pseudo-replication.
#' @param control A list created with \code{\link{gesearch_control}} containing
#'   additional algorithm parameters.
#' @param intermediate_models A logical indicating whether to store models for
#'   each intermediate generation. Default is \code{FALSE}.
#' @param verbose A logical indicating whether to print progress information.
#'   Default is \code{TRUE}.
#' @param seed An integer for random number generation to ensure
#'   reproducibility. Default is \code{NULL}.
#' @param pchunks An integer specifying the chunk size used for memory-efficient
#' parallel processing. Larger values divide the workload into smaller pieces,
#' which can help reduce memory pressure. Default is 1L.
#' @param formula A \code{\link[stats]{formula}} defining the model.
#' @param train A data.frame containing training data with model variables.
#' @param test A data.frame containing test data with model variables.
#' @param na_action A function for handling missing values in training data.
#'   Default is \code{\link[stats]{na.pass}}.
#' @param object A fitted \code{gesearch} object (for \code{predict}).
#' @param newdata A matrix or data.frame of new observations. For formula-fitted
#'   models, a data.frame containing all predictor variables is accepted. For
#'   non-formula models, a matrix is required.
#' @param type A character string specifying the prediction type. Currently only
#'   \code{"response"} is supported.
#' @param what A character string specifying which models to use for prediction:
#'   \code{"final"} (default) for predictions from final models only, or
#'   \code{"all_generations"} for predictions from all intermediate generations
#'   plus the final models.
#' @param x A \code{gesearch} object (for \code{plot}).
#' @param which Character string specifying what to plot: 
#'   \code{"weakness"} (maximum weakness scores per generation) or 
#'   \code{"removed"} (cumulative samples removed).
#' @param ... Additional arguments passed to methods.
#'
#' @details
#' The \code{gesearch} algorithm requires a large reference dataset (\code{Xr})
#' where the sample search is conducted, target observations (\code{Xu}), and
#' three tuning parameters: \code{k}, \code{b}, and \code{retain}.
#'
#' The target observations (\code{Xu}) should represent the population of
#' interest. These may be selected via algorithms like Kennard-Stone when
#' response values are unavailable.
#'
#' The algorithm iteratively removes weak samples from \code{Xr} based on:
#' \itemize{
#'   \item Increased RMSE when predicting \code{Yu}
#'   \item Increased PLS reconstruction error on \code{Xu}
#'   \item Increased dissimilarity to \code{Xu} in PLS space
#' }
#'
#' A resampling scheme identifies samples that consistently appear in
#' high-error subsets. These are labeled weak and removed. The process
#' continues until approximately \code{target_size} samples remain.
#'
#' The `gesearch()` function also returns a final model fitted on the selected
#' samples, which can be used for prediction. This model is internally validated
#' by cross-validation using only the selected samples from the training/reference
#' set. If `Yu` is available, a model fitted only on the selected reference samples
#' is first used to predict the target samples. The final model is then refitted
#' using both the selected reference samples and the target samples used to guide
#' the search, provided that response values are available for those target samples.
#'
#' ## Parameter guidance
#' \itemize{
#'   \item \code{k}: Number of samples per resampling subset. See Lobsey et
#'     al. (2017) for guidance.
#'   \item \code{b}: Resampling intensity. Higher values increase stability
#'     but computational cost.
#'   \item \code{retain}: Proportion retained per iteration. Values >0.9
#'     recommended.
#' }
#' 
#' ## Prediction
#' The \code{predict} method generates predictions from a fitted 
#' \code{gesearch} object. If the model was fitted with a formula, 
#' \code{newdata} is validated and transformed to the appropriate model matrix.
#' 
#' When \code{what = "all_generations"}, the return value is a named list with
#' one element per generation, where each element contains a prediction 
#' matrix. This option requires \code{intermediate_models = TRUE} during 
#' fitting.
#'
#' @return
#' \strong{For \code{gesearch}:} A list of class \code{"gesearch"} containing:
#' \itemize{
#'   \item \code{x_local}: Matrix of predictors for selected samples.
#'   \item \code{y_local}: Vector of responses for selected samples.
#'   \item \code{indices}: Indices of selected samples from original training set.
#'   \item \code{complete_iter}: Number of completed iterations.
#'   \item \code{iter_weakness}: List with iteration-level weakness statistics.
#'   \item \code{samples}: List of sample indices retained at each iteration.
#'   \item \code{n_removed}: data.frame of samples removed per iteration.
#'   \item \code{control}: Copy of control parameters.
#'   \item \code{fit_method}: Fit constructor from \code{fit_method}.
#'   \item \code{validation_results}: Cross-validation in the training only set 
#'   validation on the test set using models built only with the samples found.
#'   \item \code{final_models}: Final PLS model containing coefficients, loadings, 
#'     scores, VIP, and selectivity ratios.
#'   \item \code{intermediate_models}: List of models per generation (if
#'     \code{intermediate_models = TRUE}).
#'   \item \code{seed}: RNG seed used.
#' }
#' 
#' \strong{For \code{predict.gesearch}:}
#' \itemize{
#'   \item If \code{what = "final"}: a prediction matrix with 
#'     \code{nrow(newdata)} rows and one column per PLS component.
#'   \item If \code{what = "all_generations"}: a named list of generations, 
#'     where each generation contains a prediction matrix as above.
#' }
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez},
#' Claudio Orellano,
#' \href{https://orcid.org/0000-0001-5416-4520}{Craig Lobsey},
#' \href{https://orcid.org/0000-0003-1540-4748}{Raphael Viscarra Rossel}
#'
#' @references
#' Lobsey, C.R., Viscarra Rossel, R.A., Roudier, P., Hedley, C.B. 2017.
#' rs-local data-mines information from spectral libraries to improve local
#' calibrations. European Journal of Soil Science 68:840-852.
#'
#' Kennard, R.W., Stone, L.A. 1969. Computer aided design of experiments.
#' Technometrics 11:137-148.
#'
#' Rajalahti, T., Arneberg, R., Berven, F.S., Myhr, K.M., Ulvik, R.J.,
#' Kvalheim, O.M. 2009. Biomarker discovery in mass spectral profiles by means
#' of selectivity ratio plot. Chemometrics and Intelligent Laboratory Systems
#' 95:35-48.
#' 
#' Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
#' Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
#' Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
#' libraries are too complex to search: Evolutionary subset selection for
#' domain-adaptive calibration. \emph{Analytica Chimica Acta}, under review.
#'
#' @seealso
#' \code{\link{fit_pls}}, \code{\link{gesearch_control}}, \code{\link{mbl}}
#'
#' @examples
#' \dontrun{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess
#' sg_det <- savitzkyGolay(
#'   detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
#'   m = 1, p = 1, w = 7
#' )
#' NIRsoil$spc_pr <- sg_det
#'
#' # Split data
#' train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$Ciso), ]
#' train_y <- NIRsoil$Ciso[NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)]
#' test_x <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$Ciso), ]
#' test_y <- NIRsoil$Ciso[NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)]
#'
#' # Basic search with reconstruction and similarity optimizations
#' gs <- gesearch(
#'   Xr = train_x, Yr = train_y,
#'   Xu = test_x, Yu = test_y,
#'   k = 50, b = 100, retain = 0.97,
#'   target_size = 200,
#'   fit_method = fit_pls(ncomp = 15, method = "mpls"),
#'   optimization = c("reconstruction", "similarity"),
#'   control = gesearch_control(retain_by = "probability"),
#'   seed = 42
#' )
#'
#' # Predict
#' preds <- predict(gs, test_x)
#'
#' # Plot progress
#' plot(gs)
#' plot(gs, which = "removed")
#'
#' # With reconstruction and response optimization (requires Yu)
#' gs_response <- gesearch(
#'   Xr = train_x, Yr = train_y,
#'   Xu = test_x, Yu = test_y,
#'   k = 50, b = 100, retain = 0.97,
#'   target_size = 200,
#'   fit_method = fit_pls(ncomp = 15),
#'   optimization = c("reconstruction", "response"),
#'   seed = 42
#' )
#'
#' # Parallel processing
#' library(doParallel)
#' n_cores <- min(2, parallel::detectCores() - 1)
#' cl <- makeCluster(n_cores)
#' registerDoParallel(cl)
#'
#' gs_parallel <- gesearch(
#'   Xr = train_x, Yr = train_y,
#'   Xu = test_x,
#'   k = 50, b = 100, retain = 0.97,
#'   target_size = 200,
#'   fit_method = fit_pls(ncomp = 15),
#'   pchunks = 3,
#'   seed = 42
#' )
#'
#' stopCluster(cl)
#' registerDoSEQ()
#' }
#'
#' @importFrom stats as.formula update na.pass model.frame model.matrix
#' @importFrom stats model.extract terms .MFclass delete.response
#' @export gesearch
#' 
NULL
######################################################################
# resemble
# Copyright (C) 2026 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################

#' @aliases gesearch
"gesearch" <-
  function(...) {
    UseMethod("gesearch")
  }


#' @aliases gesearch
#' @export
#'
## add an argument to initualize with a given amount of observations, e.g. if 
## the library is 30.000 you could initialize with 15.000 observations and to 
## compensate you can increase the number of resampling iterations
## perhaps a number of initial resampling iterations can also work say we 
## start with 10.000 for the first loop and then it goes to what b specifies
gesearch.default <- function(
    Xr,
    Yr,
    Xu,
    Yu = NULL,
    Yu_lims = NULL,
    k,
    b,
    retain = 0.95,
    target_size = k,
    fit_method = fit_pls(ncomp = 10),
    optimization = "reconstruction",
    group = NULL,
    control = gesearch_control(),
    intermediate_models = FALSE,
    verbose = TRUE,
    seed = NULL,
    pchunks = 1L,
    ...
) {
  crossover <- TRUE

  dots <- list(...)
  if ("formula" %in% names(dots)) {
    if (!inherits(dots$formula, "formula")) {
      stop("The 'formula' argument must be a formula object.", call. = FALSE)
    }
  }
  
  
  # --- Input validation ---
  if (!inherits(Xr, c("matrix", "data.frame", "formula"))) {
    stop(
      "You are attempting to pass a ", class(Xr)[1L], " as 'Xr' or 'formula'.\n",
      "Expected a numeric matrix for gesearch.default() or a formula for gesearch.formula().",
      call. = FALSE
    )
  }
  
  Yr <- as.matrix(Yr)

  # --- control validation ---
  if (!inherits(control, "gesearch_control")) {
    stop("'control' must be created by gesearch_control()", call. = FALSE)
  }
  
  if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
    old_blas_threads <- blas_get_num_procs()
    if (old_blas_threads != control$blas_threads) {
      blas_set_num_threads(control$blas_threads)
      on.exit(blas_set_num_threads(old_blas_threads), add = TRUE)
    }
  } else if (Sys.info()["sysname"] == "Linux" && control$blas_threads == 1L) {
    message("Tip: Install 'RhpcBLASctl' for optimal performance on Linux:\n",
            "  install.packages('RhpcBLASctl')")
  }
  
  if (ncol(Yr) == 1L) {
    if (nrow(Xr) != length(Yr)) {
      stop("nrow(Xr) must equal length(Yr)", call. = FALSE)
    }
  } else {
    if (is.null(colnames(Yr))) {
      stop("Yr must have column names when it has multiple columns", call. = FALSE)
    }
    if (nrow(Xr) != nrow(Yr)) {
      stop("nrow(Xr) must equal nrow(Yr)", call. = FALSE)
    }
  }
  
  if (target_size > nrow(Xr)) {
    stop("'target_size' cannot be greater than nrow(Xr)", call. = FALSE)
  }
  
  if (!is.logical(verbose)) {
    stop("'verbose' must be logical", call. = FALSE)
  }
  
  if (nrow(Xu) < 3L) {
    stop("Xu must have at least 3 rows", call. = FALSE)
  }
  
  if (!is.null(Yu)) {
    Yu <- as.matrix(Yu)
    if (ncol(Yu) == 1L) {
      if (nrow(Xu) != length(Yu)) {
        stop("nrow(Xu) must equal length(Yu)", call. = FALSE)
      }
    } else {
      if (!is.null(Yu_lims)) {
        stop("'Yu_lims' only supported for single-column Yu", call. = FALSE)
      }
      if (ncol(Yr) != ncol(Yu)) {
        stop("ncol(Yr) must equal ncol(Yu)", call. = FALSE)
      }
      if (nrow(Xu) != nrow(Yu)) {
        stop("nrow(Xu) must equal nrow(Yu)", call. = FALSE)
      }
      if (is.null(colnames(Yu))) {
        stop("Yu must have column names when it has multiple columns", call. = FALSE)
      }
      if (!all(colnames(Yr) == colnames(Yu))) {
        stop("Column names of Yr and Yu must match", call. = FALSE)
      }
    }
  }
  
  # Yu required with non-missing values for response optimization
  if ("response" %in% optimization) {
    if (is.null(Yu) || !any(!is.na(Yu))) {
      stop("'Yu' with non-missing values required when optimization includes 'response'", 
           call. = FALSE)
    }
  }
  
  # No missing values allowed in Xr, Yr, or Xu
  if (anyNA(Xr)) {
    stop("'Xr' contains missing values.", call. = FALSE)
  }
  if (anyNA(Yr)) {
    stop("'Yr' contains missing values.", call. = FALSE)
  }
  if (anyNA(Xu)) {
    stop("'Xu' contains missing values.", call. = FALSE)
  }
  
  # No infinite values in any input
  if (any(is.infinite(Xr)) || any(is.infinite(Yr)) || 
      any(is.infinite(Xu)) || (!is.null(Yu) && any(is.infinite(Yu)))) {
    stop("Infinite values detected in input data", call. = FALSE)
  }
  
  # Xr and Xu must have same number of variables
  if (ncol(Xr) != ncol(Xu)) {
    stop("ncol(Xr) must equal ncol(Xu)", call. = FALSE)
  }
  
  # Validate k (gene size)
  k <- as.integer(round(k))
  if (k >= nrow(Xr)) {
    stop("'k' must be less than nrow(Xr)", call. = FALSE)
  }
  
  # Validate target_size
  target_size <- as.integer(round(target_size))
  if (target_size < k) {
    stop("'target_size' must be >= k", call. = FALSE)
  }
  
  # Validate retain proportion
  if (retain <= 0 || retain > 1) {
    stop("'retain' must be in (0, 1]", call. = FALSE)
  }
  
  # Validate optimization options
  # Note: "fresponse" is undocumented (internal/experimental)
  allowed <- c("response", "reconstruction", "similarity", "range", "fresponse")
  bad <- setdiff(optimization, allowed)
  if (length(bad) > 0L) {
    stop(
      sprintf(
        "Invalid 'optimization' option(s): %s. Allowed: %s",
        paste(shQuote(bad), collapse = ", "),
        paste(shQuote(allowed), collapse = ", ")
      ),
      call. = FALSE
    )
  }
  
  # Validate fit_method
  # TODO: Currently only fit_pls() is supported; fit_wapls(), fit_gpr() may be added later
  if (!inherits(fit_method, "fit_method")) {
    stop("'fit_method' must be a fit method object (e.g., fit_pls())", call. = FALSE)
  }
  
  if (!inherits(fit_method, "fit_pls")) {
    stop("Only fit_pls() is supported for the moment", call. = FALSE)
  }
  
  # Tuning not applied during reconstruction optimization
  if ("reconstruction" %in% optimization && control$tune) {
    warning(
      "PLS components are not tuned when optimization includes 'reconstruction'; ",
      "using fixed ncomp from 'fit_method'. The 'tune' option was ignored.",
      call. = FALSE
    )
  }
  
  
  if (control$tune) {
    min_samples <- floor(min(k, dim(Xr)) * control$p) - 1
    min_cv_samples <- floor(min(k, dim(Xr)) * (1 - control$p))
    if (min_cv_samples < 3) {
      stop(paste0(
        "Local cross-validation requires at least 3 observations in ",
        "the hold-out set, the current cross-validation parameters ",
        "leave less than 3 observations for one or more resampling ",
        "iterations."
      ))
    }
  } else {
    min_samples <- floor(min(k, dim(Xr))) - 1
  }
  
  # Validate ncomp vs available samples
  # TODO: Add fit_wapls support when implemented
  if (inherits(fit_method, "fit_pls")) {
    max_ncomp <- fit_method$ncomp
  } else if (inherits(fit_method, "fit_wapls")) {
    max_ncomp <- fit_method$max_ncomp
  }
  
  if (min_samples < max_ncomp) {
    stop(
      "More PLS components than available observations in some resampling iterations. ",
      "If tuning is enabled, some observations are held out for validation.",
      call. = FALSE
    )
  }
  
  call_f <- match.call()
  
  Xr <- as.matrix(Xr)
  Xu <- as.matrix(Xu)
  
  # Proportion to remove each iteration
  r <- 1 - round(retain, .Machine$sizeof.longdouble)
  
  # Index into the spectral library (SL); we operate on indices, not copies
  sl_idx <- seq_len(nrow(Xr))
  
  # Initialize K as full SL
  k_idx <- sl_idx
  
  # Iteration counter
  outer_idx <- 0L
  
  # Sample ranking vectors (by RMSE)
  U <- V <- rep(0, length(sl_idx))
  
  # Iteration tracking
  s_to_drop <- NULL
  cuts <- NULL
  max_weakness_score <- NULL
  pp <- r
  
  # # Progress indicators
  # cat_progress <- cat_iter(c("\\", "|", "/", "-"))
  # cat_progress2 <- cat_iter(c("\\", "|", "/", "-"))
  
  # Define once outside the loop
  if (verbose) {
    # spinner <- c("⠋", "⠙", "⠹", "⠸", "⠼", "⠴", "⠦", "⠧", "⠇", "⠏")
    spinner <- c(">  ", ">> ", ">>>", "   ")
    # spinner <- c("|", "/", "-", "\\")
    # spinner <- c(".", "o", "O", "o")
    # spinner <- c("[=   ]", "[ =  ]", "[  = ]", "[   =]", "[  = ]", "[ =  ]")
    # spinner <- c(".  ", ".. ", "...", "   ")
    # spinner <- c(">  ", ">> ", ">>>", "   ")
    # spinner <- c("[-]", "[\\]", "[|]", "[/]")
    
    n_spin <- length(spinner)
    prev_msg_len <- 0L
  }
  
  
  pool_sizes_log <- integer(0L)
  complete_iterations <- 0L
  selected <- list()
  # Step 2-6: Main evolutionary loop (see Step 7 for termination)
  while (length(k_idx) > (target_size * (1 + pp)) & sum(!is.na(sl_idx)) > target_size) {
    # Step 1: Update K to contain only non-dropped samples
    k_idx <- sl_idx[!is.na(sl_idx)]
    selected <- c(selected, list(k_idx))
    
    outer_idx <- outer_idx + 1L
    
    # Number of samples to remove this iteration
    cull_quantity <- round(length(k_idx) * round(r, .Machine$sizeof.longdouble))
    
    # Step 2-4: Resampling iterations (B times)
    B <- round(length(k_idx) * b / k)
    
    # Progress output
    # if (verbose) {
    #   pool_label <- if (outer_idx == 1L) "Initial set of" else "Active"
    #   
    #   msg <- paste0(
    #     "Generation ", outer_idx, ":\t",
    #     "\033[34m", cat_progress(), " ", pool_label, " genes (samples): ", length(k_idx), "\t\033[39m",
    #     "\033[34m", cat_progress2(), " Individuals to evaluate: ", B, "\033[39m"
    #   )
    #   
    #   # Pad to fixed width and return cursor to start
    #   cat(sprintf("\r%-120s", msg))
    #   flush.console()
    # }
    if (verbose) {
      spin <- spinner[(outer_idx %% n_spin) + 1L]
      pool_label <- if (outer_idx == 1L) "Initial" else "Active"
      
      # Build message without ANSI codes for length calculation
      msg_plain <- sprintf(
        "Generation %d: %s genes (samples): %d %s | Individuals: %d %s",
        outer_idx, pool_label, length(k_idx), spin, B, spin
      )
      
      # Build message with ANSI codes for display
      msg <- sprintf(
        "\rGeneration %d: \033[34m%s genes (samples): %d %s\033[39m | \033[34mIndividuals: %d %s\033[39m",
        outer_idx, pool_label, length(k_idx), spin, B, spin
      )
      
      # Pad with spaces to overwrite previous longer message
      padding <- max(0L, prev_msg_len - nchar(msg_plain))
      cat(msg, strrep(" ", padding), sep = "")
      flush.console()
      
      prev_msg_len <- nchar(msg_plain)
    }
    
    # Step 2: Sample training subsets of size k without replacement
    if (!is.null(seed)) {
      set.seed(seed + B)
    }
    
    if (outer_idx == 1L) {
      ismpl <- replicate(
        n = B,
        expr = sample(x = k_idx, size = k, replace = FALSE)
      )
    } else {
      # Replace genes dropped in last iteration with NA
      ismpl[ismpl %in% which(new_nas)] <- NA
      
      if (crossover) {
        ismpl <- replicate(
          n = B,
          expr = combine_individuals(ismpl, k, k_idx)
        )
      } else {
        ismpl <- replicate(
          n = B,
          expr = sample(x = k_idx, size = k, replace = FALSE)
        )
      }
    }
    
    # Iterator for memory-efficient parallel chunking:
    # Takes only needed slices from the matrix rather than loading
    # the full matrix into memory for each parallel worker
    if (control$allow_parallel && getDoParRegistered()) {
      leftc <- B %% pchunks
      if (leftc > 0L) {
        req_iter <- floor(B / pchunks) + leftc
      } else {
        req_iter <- B / pchunks
      }
    } else {
      req_iter <- B
      pchunks <- 1L
    }
    
    itersubs <- ithrssubsets(
      x = Xr,
      y = Yr,
      group = group,
      indx = ismpl,
      chunksize = pchunks
    )
    
    # Step 3: Calibrate PLS model using selected samples
    # Step 4: Validate on target samples
    # Step 5: Repeat steps 2-4 B times
    
    # Validate Yu_lims for range optimization
    if ("range" %in% optimization) {
      if (is.null(Yu_lims)) {
        stop("'Yu_lims' is required for 'range' optimization", call. = FALSE)
      }
      if (length(Yu_lims) != 2L) {
        stop("'Yu_lims' must be a numeric vector of length 2", call. = FALSE)
      }
    }
    
    # Set response limits for range filtering
    if (!is.null(Yu_lims) && "range" %in% optimization) {
      user_min_y <- min(Yu_lims)
      user_max_y <- max(Yu_lims)
    } else {
      user_min_y <- NULL
      user_max_y <- NULL
      optimization <- optimization[optimization != "range"]
    }
    
    results_df <- biter(
      itersubs = itersubs,
      Xu = Xu,
      Yu = Yu,
      y_lim_left = user_min_y,
      y_lim_right = user_max_y,
      iter_sequence = seq_len(req_iter),
      n = nrow(ismpl),
      optimization = optimization,
      ncomp = fit_method$ncomp,
      tune = control$tune,
      p = control$p,
      number = control$number,
      scale = fit_method$scale,
      max_iter = 1L,
      tol = 1e-6,
      seed = seed,
      algorithm = fit_method$method,
      allow_parallel = control$allow_parallel,
      pchunks = pchunks
    )
    
    # Step 6: Rank samples by weakness and identify candidates for removal
    drop_results <- apply(
      results_df$weakness_scores_subset,
      MARGIN = 2L,
      FUN = drop_indices,
      uu = U,
      vv = V,
      r = r,
      obs_indices = sl_idx,
      resampling_indices = results_df$sampleidx,
      retain_by = control$retain_by,
      cull_quantity = cull_quantity,
      percentile_type = control$percentile_type,
      max_ncomp = fit_method$ncomp
    )
    
    # Check for interruptions from drop_indices
    was_interrupted <- vapply(drop_results, function(x) x$interrupted, logical(1L))
    
    if (any(was_interrupted)) {
      # Clean up analysis names for display
      analysis_name <- gsub("rmse_|score_dis|residual_| deviation_from", "", names(drop_results))
      analysis_name <- gsub("_", " ", analysis_name)
      
      # Report which analyses were interrupted
      for (i in which(was_interrupted)) {
        message(analysis_name[i], ": ", drop_results[[i]]$message)
      }
      break
    }
    
    pool_sizes_log <- c(pool_sizes_log, length(k_idx))
    
    # Accumulate cutoff and weakness scores across iterations
    cuts <- rbind(cuts, vapply(drop_results, function(x) x$cutoff, numeric(1L)))
    max_weakness_score <- rbind(
      max_weakness_score, 
      vapply(drop_results, function(x) x$max_weakness_score, numeric(1L))
    )
    
    # Mark dropped samples as NA
    new_nas <- !complete.cases(
      vapply(drop_results, function(x) x$obs_indices, numeric(length(sl_idx)))
    )
    sl_idx[new_nas] <- NA
    
    complete_iterations <- complete_iterations + 1L
    s_to_drop <- c(s_to_drop, sum(is.na(sl_idx)))
    
    # Check for stagnation (gene pool size unchanged)
    if (length(pool_sizes_log) >= control$stagnation_limit) {
      recent_sizes <- tail(pool_sizes_log, control$stagnation_limit)
      if (all(recent_sizes == recent_sizes[1L])) {
        cat(
          "\n\033[31mEarly termination: Active genes remained at ",
          recent_sizes[1L], " for ", control$stagnation_limit,
          " consecutive iterations. Stopping.\033[39m\n",
          sep = ""
        )
        break
      }
    }
  }
  # Finalize k_idx after last iteration
  k_idx <- sl_idx[!is.na(sl_idx)]
  selected <- c(selected, list(k_idx))
  
  # Ensure final selection meets target_size
  
  msizes <- vapply(selected, length, integer(1L))
  k_idx <- selected[[max(which(msizes >= target_size))]]
  selected <- selected[msizes >= target_size]
  
  
  models_per_generation <- vector("list", length(selected) - 1)
  if (intermediate_models) {
    if (verbose) {
      cat("\nFitting intermediate models...")
    }
    for (i in 1:(length(selected) - 1)) {
      ith_selected <- selected[[i]]
      
      # Fit models for this generation
      pls_models <- list()
      validation_results <- list()
      for (j in seq_len(ncol(Yr))) {
        # Fit and validate PLS model for response j
        pls_val <- get_plsr(
          X = Xr,
          Y = Yr[, j, drop = FALSE],
          indices = ith_selected,
          fit_method = fit_method,
          number = control$number,
          group = group[ith_selected],
          retrieve = FALSE,
          tune = TRUE,
          seed = seed,
          p = control$p
        )
        validation <- list(
          val_info = list(train_resamples = pls_val$train_resamples),
          results = list(train = pls_val$cv_results)
        )
        
        # Add test set predictions if Yu available
        if (!is.null(Yu)) {
          yuhat <- predict(pls_val, newdata = Xu, ncomp = fit_method$ncomp)
          
          validation$val_info$test_predictions <- yuhat
          validation$results$test <- get_validation_metrics(yuhat, Yu)
          
          tmpx <- rbind(Xr[ith_selected, , drop = FALSE], Xu)
          tmpy <- c(Yr[ith_selected, j], Yu[, j])
          complete_idx <- complete.cases(tmpy)
          tmpx <- tmpx[complete_idx, , drop = FALSE]
          tmpy <- tmpy[complete_idx]
        } else {
          tmpx <- Xr[ith_selected, , drop = FALSE]
          tmpy <- Yr[ith_selected, j]
        }
        
        # Fit final PLS model for this generation
        intermediate_pls <- assign_pls_names(opls_get_all(
          X = tmpx,
          Y = as.matrix(tmpy),
          ncomp = fit_method$ncomp,
          scale = fit_method$scale,
          maxiter = 1L,
          tol = 1e-6,
          algorithm = if (fit_method$fit_method == "mpls") "mpls" else "pls"
        ), x_names = colnames(Xr))
        
        pls_models[[j]] <- intermediate_pls
        validation_results[[j]] <- validation
      }
      
      # Name models by response variable if available
      if (!is.null(colnames(Yu))) {
        names(pls_models) <- names(validation_results) <- colnames(Yu)
      }
      models_per_generation[[i]] <- list(
        subset_size = length(ith_selected),
        pls_models = pls_models,
        validation = validation_results
      )
    }
  }

  # Progress message for final model fitting
  if (verbose) {
    cat("\nFitting final model on", length(k_idx), "selected genes...\n")
    if (!is.null(Yu) && sum(complete.cases(Yu)) > 0L) {
      cat("and", sum(complete.cases(Yu)), "target genes with available response values...\n")
    }
  }
  
  # Fit final models on selected samples
  pls_models <- list()
  validation_results <- list()
  
  for (j in seq_len(ncol(Yr))) {
    # Fit and validate PLS model for response j
    pls_val <- get_plsr(
      X = Xr,
      Y = Yr[, j, drop = FALSE],
      indices = k_idx,
      fit_method = fit_method,
      number = control$number,
      group = group[k_idx],
      retrieve = FALSE,
      tune = TRUE,
      seed = seed,
      p = control$p
    )
    
    validation <- list(
      val_info = list(train_resamples = pls_val$train_resamples),
      results = list(train = pls_val$cv_results)
    )
    
    # Add test set predictions if Yu available
    if (!is.null(Yu)) {
      yuhat <- predict(pls_val, newdata = Xu, ncomp = fit_method$ncomp)
      
      validation$val_info$test_predictions <- yuhat
      validation$results$test <- get_validation_metrics(yuhat, Yu)
      
      tmpx <- rbind(Xr[k_idx, , drop = FALSE], Xu)
      tmpy <- c(Yr[k_idx, j], Yu[, j])
      complete_idx <- complete.cases(tmpy)
      tmpx <- tmpx[complete_idx, , drop = FALSE]
      tmpy <- tmpy[complete_idx]
    } else {
      tmpx <- Xr[k_idx, , drop = FALSE]
      tmpy <- Yr[k_idx, j]
    }

    # Fit final PLS model
    finalpls <- assign_pls_names(opls_get_all(
      X = tmpx,
      Y = as.matrix(tmpy),
      ncomp = fit_method$ncomp,
      scale = fit_method$scale,
      maxiter = 1L,
      tol = 1e-6,
      algorithm = if (fit_method$fit_method == "mpls") "mpls" else "pls"
    ), x_names = colnames(Xr))
    
    pls_models[[j]] <- finalpls
    validation_results[[j]] <- validation
  }
  
  
  
  
  
  
  # Name models by response variable if available
  if (!is.null(colnames(Yu))) {
    names(pls_models) <- colnames(Yu)
    names(validation_results) <- colnames(Yu)
  }
  
  # Compile iteration weakness statistics
  iter_weakness <- list(
    maximum = data.frame(
      iteration = seq_len(nrow(max_weakness_score)),
      max_weakness_score
    ),
    cutoff = data.frame(
      iteration = seq_len(nrow(max_weakness_score)),
      cuts
    )
  )
  
  names(selected) <- paste0("iteration_", seq_along(selected))
  
  # Assemble results
  resultsList <- list(
    x_local = Xr[k_idx, , drop = FALSE],
    y_local = Yr[k_idx, , drop = FALSE],
    indices = k_idx,
    complete_iter = complete_iterations,
    iter_weakness = iter_weakness,
    samples = selected,
    n_removed = data.frame(
      iteration = seq_along(s_to_drop),
      removed = c(s_to_drop[1L], diff(s_to_drop)),
      cumulative = s_to_drop
    ),
    control = control,
    fit_method = fit_method,
    validation_results = validation_results,
    final_models = pls_models,
    intermediate_models = models_per_generation,
    seed = seed
  )
  
  if (!intermediate_models) {
    resultsList <- resultsList[!names(resultsList) %in% "intermediate_models"]
  }
  
  attr(resultsList, "call") <- call_f
  class(resultsList) <- c("gesearch", "list")
  
  resultsList
}


#' @aliases gesearch
#' @importFrom stats quantile complete.cases diffinv na.pass model.extract 
#' model.frame model.matrix
#' @export
gesearch.formula <- function(
    formula,
    train,
    test,
    k,
    b,
    target_size,
    fit_method,
    ...,
    na_action = na.pass
) {
  if (!inherits(formula, "formula")) {
    stop("'formula' must be a formula object", call. = FALSE)
  }
  
  call_f <- match.call()
  call_env <- parent.frame() 
  
  if (missing(fit_method)) {
    stop("'fit_method' is missing", call. = FALSE)
  }
  
  mf <- match.call(expand.dots = FALSE)
  
  if (!"na_action" %in% names(mf)) {
    mf[["na_action"]] <- formals(gesearch.formula)[["na_action"]]
  }
  
  # Get the model frame
  mr <- match(x = c("formula", "train", "na_action"), table = names(mf))
  mu <- match(x = c("formula", "test"), table = names(mf))
  
  mfr <- mf[c(1L, mr)]
  mfu <- mf[c(1L, mu)]
  
  names(mfr)[names(mfr) == "na_action"] <- "na.action"
  names(mfr)[names(mfr) == "train"] <- "data"
  names(mfu)[names(mfu) == "test"] <- "data"
  
  yname <- all.vars(formula, functions = FALSE, max.names = 1L)
  
  mfr[[1L]] <- mfu[[1L]] <- as.name("model.frame")
  
  input_list <- list(...)
  
  # Handle missing response in test data
  if (!yname %in% colnames(eval(mfu$data, envir = call_env))) {
    if ("optimization" %in% names(input_list)) {
      if (input_list$optimization == "response") {
        stop("'optimization = \"response\"' requires response values in test", 
             call. = FALSE)
      }
    }
    
    test <- local({
      tmp <- make.unique(c(names(test), ".tmp_y"))[length(names(test)) + 1L]
      test[[tmp]] <- NA
      names(test)[names(test) == tmp] <- yname
      test
    })
    warning(yname, " not found in test; assigned NA.", call. = FALSE)
  }
  
  mfu <- model.frame(mfu, data = test, na.action = NULL)
  mfr <- eval(mfr, call_env)
  
  trms <- attr(mfr, "terms")
  formulaclasses <- list(attr(trms, "dataClasses"))
  
  attr(trms, "intercept") <- 0L
  xr <- model.matrix(trms, model.frame(mfr, drop.unused.levels = TRUE))
  yr <- model.extract(mfr, "response")
  
  xu <- model.matrix(trms, mfu)
  yu <- model.extract(mfu, "response")
  
  # Validate response for response optimization
  if ("optimization" %in% names(input_list)) {
    if ("response" %in% input_list$optimization) {
      if (sum(is.na(yu)) == length(yu)) {
        stop("'optimization = \"response\"' requires response values in test", 
             call. = FALSE)
      }
    }
  }
  
  rsl <- gesearch(
    Xr = xr,
    Yr = yr,
    Xu = xu,
    Yu = yu,
    k = k,
    b = b,
    target_size = target_size,
    fit_method = fit_method,
    ...
  )
  rsl$formula <- formula
  rsl$dataclasses <- formulaclasses
  
  attr(rsl, "call") <- call_f
  class(rsl) <- c("gesearch", "gesearch.formula", "list")
  
  rsl
}

## ## THIS IS TO ALLOW A LIST OF FORMULAS TO BE USED FOR MULTI-RESPONSE MODELING
## @aliases gesearch
## @importFrom stats quantile complete.cases diffinv na.pass model.extract
## model.frame model.matrix
## @export
# gesearch.list <- function(
#     formula,
#     train,
#     test,
#     k,
#     b,
#     target_size,
#     fit_method,
#     ...,
#     na_action = na.pass
# ) {
#   # Validate formula list
#   is_formula <- vapply(formula, inherits, logical(1L), what = "formula")
#   if (any(!is_formula)) {
#     stop("All elements in 'formula' must be formula objects", call. = FALSE)
#   }
#   
#   right_side <- vapply(formula, function(x) labels(terms(x)), character(1L))
#   left_side <- vapply(formula, function(x) all.vars(update(x, . ~ 1)), character(1L))
#   
#   mresponse <- paste(c("matrix_response", left_side), collapse = "_")
#   mfml <- as.formula(paste(mresponse, "~", unique(right_side)))
#   
#   if (length(unique(right_side)) > 1L) {
#     stop("Right-hand side variables must be identical across all formulas", 
#          call. = FALSE)
#   }
#   
#   call_f <- match.call()
#   
#   
#   definition <- sys.function(sys.parent())
#   mf <- match.call(expand.dots = FALSE)
#   formals <- formals(definition)
#   
#   mf[["formula"]] <- mfml
#   
#   if (!"na_action" %in% names(mf)) {
#     mf[["na_action"]] <- formals[["na_action"]]
#     match.call(definition, mf, TRUE)
#   }
#   
#   # Get the model frame
#   mr <- match(x = c("formula", "train", "na_action"), table = names(mf))
#   mu <- match(x = c("formula", "test"), table = names(mf))
#   
#   mfr <- mf[c(1L, mr)]
#   mfu <- mf[c(1L, mu)]
#   
#   names(mfr)[names(mfr) == "na_action"] <- "na.action"
#   names(mfr)[names(mfr) == "train"] <- "data"
#   names(mfu)[names(mfu) == "test"] <- "data"
#   
#   yname <- left_side
#   
#   mfr[[1L]] <- mfu[[1L]] <- as.name("model.frame")
#   
#   input_list <- list(...)
#   
#   # Handle missing response in test data
#   if (!any(yname %in% colnames(eval(mfu$data)))) {
#     if ("optimization" %in% names(input_list)) {
#       if (input_list$optimization == "response") {
#         stop("'optimization = \"response\"' requires response values in test", 
#              call. = FALSE)
#       }
#     }
#     
#     missing_vars <- setdiff(yname, names(test))
#     if (length(missing_vars) > 0L) {
#       test[missing_vars] <- NA
#       warning(
#         paste(missing_vars, collapse = ", "), " not found in test; assigned NA.",
#         call. = FALSE
#       )
#     }
#   }
#   
#   train[[mresponse]] <- as.matrix(train[, left_side])
#   test[[mresponse]] <- as.matrix(test[, left_side])
#   
#   mfu <- model.frame(mfu, data = test, na.action = NULL)
#   mfr <- model.frame(mfr, data = train, na.action = NULL)
#   
#   trms <- attr(mfr, "terms")
#   
#   # Build formula classes
#   formulaclasses <- vector("list", length(yname))
#   wformulaclasses <- attr(trms, "dataClasses")
#   wformulaclasses[[1L]] <- "nmatrix.1"
#   for (i in seq_along(yname)) {
#     formulaclasses[[i]] <- wformulaclasses
#     names(formulaclasses[[i]])[1L] <- yname[i]
#   }
#   
#   attr(trms, "intercept") <- 0L
#   xr <- model.matrix(trms, model.frame(mfr, drop.unused.levels = TRUE))
#   yr <- model.extract(mfr, "response")
#   
#   xu <- model.matrix(trms, mfu)
#   yu <- model.extract(mfu, "response")
#   
#   if (!inherits(yu, "matrix")) {
#     yu <- as.matrix(yu)
#   }
#   
#   # Validate response for response optimization
#   for (i in seq_len(ncol(yu))) {
#     if ("optimization" %in% names(input_list) && 
#         sum(is.na(yu[, i])) == length(yu[, i])) {
#       stop(
#         "'optimization = \"response\"' requires response values in test. ",
#         "Check: ", colnames(yu)[i],
#         call. = FALSE
#       )
#     }
#   }
#   
#   rsl <- gesearch(
#     Xr = xr,
#     Yr = yr,
#     Xu = xu,
#     Yu = yu,
#     k = k,
#     b = b,
#     target_size = target_size,
#     fit_method = fit_method,
#     ...
#   )
#   rsl$formula <- formula
#   rsl$dataclasses <- formulaclasses
#   
#   attr(rsl, "call") <- call_f
#   class(rsl) <- c("gesearch", "gesearch.list", "list")
#   
#   rsl
# }


#' @aliases gesearch
#' @importFrom stats .MFclass terms delete.response model.frame model.matrix
#' @export
predict.gesearch <- function(
    object, 
    newdata, 
    type = "response", 
    what = c("final", "all_generations"), 
    ...
) {
  what <- match.arg(what)
  # Validate type
  allowed_types <- c("response")
  if (!is.character(type) || length(type) != 1L || is.na(type) || 
      !(type %in% allowed_types)) {
    stop(
      "'type' must be one of: ", 
      paste(shQuote(allowed_types), collapse = ", "), 
      call. = FALSE
    )
  }
  
  if (missing(newdata)) {
    stop("'newdata' is required", call. = FALSE)
  }
  
  # Handle formula-fitted models
  if (!is.null(object$formula)) {
    dcls <- object$dataclasses[[1L]][-1L]
    
    if (!inherits(newdata, "matrix") && !inherits(newdata, "data.frame")) {
      stop("'newdata' must be a data.frame or matrix for formula-fitted models", 
           call. = FALSE)
    }
    
    if (inherits(newdata, "data.frame")) {
      missing_vars <- setdiff(names(dcls), names(newdata))
      if (length(missing_vars) > 0L) {
        stop("Missing predictor variables: ", paste(missing_vars, collapse = ", "), 
             call. = FALSE)
      }
    }
    
    # Handle non-numeric predictors
    if (any(dcls != "numeric")) {
      if (inherits(newdata, "matrix") && length(dcls) == 1L) {
        if (.MFclass(newdata) == dcls) {
          pnames <- gsub(
            names(dcls), "",
            colnames(object$final_models[[1L]]$coefficients)
          )
          if (all(pnames %in% colnames(newdata))) {
            newdata_temp <- newdata
            newdata <- data.frame(rep(NA, nrow(newdata)))
            colnames(newdata) <- names(object$dataclasses[[1L]])[1L]
            newdata[[names(dcls)]] <- newdata_temp[, pnames]
          } else {
            stop("Missing predictor variables", call. = FALSE)
          }
        }
      }
    }
    
    oterms <- terms(object$formula)
    oterms <- delete.response(oterms)
    attr(oterms, "intercept") <- 0L
    mf <- model.frame(oterms, newdata)
    newdata <- model.matrix(oterms, model.frame(mf, drop.unused.levels = TRUE))
  }
  
  if (!inherits(newdata, "matrix")) {
    stop("'newdata' must be a matrix", call. = FALSE)
  }
  
  # Generate predictions from final models
  if (type == "response") {
    preds <- vector("list", length(object$final_models))
    for (i in seq_along(object$final_models)) {
      preds[[i]] <- predict_opls(
        bo = object$final_models[[i]]$bo,
        b = t(object$final_models[[i]]$coefficients),
        ncomp = object$final_models[[i]]$ncomp,
        newdata = newdata,
        scale = object$fit_method$scale,
        Xscale = object$final_models[[i]]$transf$Xscale
      )
      rownames(preds[[i]]) <- rownames(newdata)
      colnames(preds[[i]]) <- paste0("ncomp_", seq_len(ncol(preds[[i]])))
    }
    names(preds) <- names(object$final_models)
  }
  
  # Include intermediate generation predictions if requested
  if (type == "response" && what == "all_generations") {
    if (is.null(object$intermediate_models)) {
      warning(
        "Model object does not contain intermediate models; ",
        "predictions generated only from final model.",
        call. = FALSE
      )
    } else {
      n_intermediate <- length(object$intermediate_models)
      intermediate_preds <- vector("list", n_intermediate + 1L)
      
      for (j in seq_len(n_intermediate)) {
        n_models <- length(object$intermediate_models[[j]]$pls_models)
        intermediate_preds[[j]] <- vector("list", n_models)
        
        for (i in seq_len(n_models)) {
          intermediate_preds[[j]][[i]] <- predict_opls(
            bo = object$intermediate_models[[j]]$pls_models[[i]]$bo,
            b = t(object$intermediate_models[[j]]$pls_models[[i]]$coefficients),
            ncomp = object$intermediate_models[[j]]$pls_models[[i]]$ncomp,
            newdata = newdata,
            scale = object$fit_method$scale,
            Xscale = object$intermediate_models[[j]]$pls_models[[i]]$transf$Xscale
          )
          rownames(intermediate_preds[[j]][[i]]) <- rownames(newdata)
          colnames(intermediate_preds[[j]][[i]]) <- paste0(
            "ncomp_", seq_len(ncol(intermediate_preds[[j]][[i]]))
          )
        }
        names(intermediate_preds[[j]]) <- names(
          object$intermediate_models[[j]]$pls_models
        )
      }
      
      # Add final model predictions as last generation
      intermediate_preds[[n_intermediate + 1L]] <- preds
      names(intermediate_preds) <- paste0("generation_", seq_along(intermediate_preds))
      preds <- intermediate_preds
    }
  }
  
  preds
}


#' @aliases gesearch
#' @export
plot.gesearch <- function(x, which = c("weakness", "removed"), ...) {
  which <- match.arg(which)
  
  # Save and restore graphical parameters on exit
  opar <- par(no.readonly = TRUE)
  on.exit(par(opar), add = TRUE)
  
  if (which == "weakness") {
    nbox <- ceiling(sqrt(ncol(x$iter_weakness$maximum) - 1L))
    par(mfrow = c(nbox, nbox), mar = c(4, 5.5, 2.5, 2))
  }
  
  # Process plot arguments
  plot_dots <- list(...)
  
  if (!"col" %in% names(plot_dots)) {
    plot_dots$col <- "dodgerblue"
  }
  
  if (!"type" %in% names(plot_dots)) {
    plot_dots$type <- "b"
  }
  
  # Remove axis labels from dots (set explicitly below)
  plot_dots$xlab <- NULL
  plot_dots$ylab <- NULL
  
  # Extract or set main title
  if ("main" %in% names(plot_dots)) {
    main <- plot_dots$main
    plot_dots$main <- NULL
  } else {
    main <- "gesearch results"
  }
  
  grid_col <- rgb(0.3, 0.3, 0.3, 0.3)
  
  # Plot weakness scores
  if (which == "weakness") {
    pnames <- colnames(x$iter_weakness$maximum)[-1L]
    for (i in seq_along(pnames)) {
      ith_ylab <- gsub("_", " ", pnames[i])
      ith_ylab <- gsub("rmse ", "", ith_ylab)
      ith_ylab <- paste0("Max. ", ith_ylab)
      do.call(
        plot,
        c(
          list(
            x = x$iter_weakness$maximum[[1L]],
            y = x$iter_weakness$maximum[[1L + i]],
            xlab = "Generation",
            ylab = ith_ylab
          ),
          plot_dots
        )
      )
      grid(col = grid_col, lty = 1L)
    }
  }
  
  # Plot cumulative removals
  if (which == "removed") {
    do.call(
      plot,
      c(
        list(
          x = x$n_removed$iteration,
          y = x$n_removed$cumulative,
          xlab = "Generation",
          ylab = "Samples removed (cumulative)"
        ),
        plot_dots
      )
    )
    grid(col = grid_col, lty = 1L)
  }
  
  mtext(main, outer = TRUE, cex = 2, line = -2)
  invisible(NULL)
}

#' @title From dissimilarity matrix to neighbors
#' @param diss_matrix a matrix representing the dissimilarities between
#' observations in a matrix \code{Xu} and observations in another matrix
#' \code{Xr}. \code{Xr} in rows \code{Xu} in columns.
#' @param k an integer value indicating the k-nearest neighbors of each
#' observation in \code{Xu} that must be selected from \code{Xr}.
#' @param k_diss an integer value indicating a dissimilarity treshold.
#' For each observation in \code{Xu}, its nearest neighbors in \code{Xr}
#' are selected as those for which their dissimilarity to \code{Xu} is below
#' this \code{k_diss} threshold. This treshold depends on the corresponding
#' dissimilarity metric specified in \code{diss_method}. Either \code{k} or
#' \code{k_diss} must be specified.
#' @param k_range an integer vector of length 2 which specifies the minimum
#' (first value) and the maximum (second value) number of neighbors to be
#' retained when the \code{k_diss} is given.
#' @param spike a vector of integers indicating what observations in \code{Xr}
#' (and \code{Yr}) must be 'forced' to always be part of all the neighborhoods.
#' @param return_dissimilarity logical indicating if the input dissimilarity
#' must be mirroed in the output.
#' @param skip_first a logical indicating whether to skip the first neighbor or
#' not. Default is \code{FALSE}. This is used when the search is being conducted
#' in symmetric matrix of distances (i.e. to avoid that the nearest neighbor of
#' each observation is itself).
#' @param keep_self an logical indicating that the indices of the observations 
#' themselves (i.e. the diagonal of the dissimilarity matrix) must be considered 
#' when the spike argument has negative values to exclude samples from the 
#' neighborhood. 
#' @description internal
#' @noRd
#' @keywords internal
diss_to_neighbors <- function(diss_matrix,
                              k = NULL, k_diss = NULL, k_range = NULL,
                              spike = NULL,
                              return_dissimilarity = FALSE,
                              skip_first = FALSE, 
                              keep_self = FALSE) {
  if (!is.null(spike)) {
    
    spike_hold <- spike[spike > 0]
    spike_rm <- -spike[spike < 0]
    
    f_order_neigh <- function(x, s_hold, s_remove, skip_first = FALSE, keep_self = FALSE) {
      x <- order(x)
      if (skip_first) {
        x <- x[-1]
      }
      x <- c(s_hold, x[!x %in% s_hold])
      if (keep_self && x[1] %in% s_remove) {
        keep_self_index <- x[1]
        x <- x[!x %in% s_remove]
        x <- c(keep_self_index, x)
      } else {
        x <- x[!x %in% s_remove]
      }
      x
    }
    
    # f_order_neigh <- function(x, s_hold, s_remove, skip_first = FALSE, keep_self = FALSE) {
    #   x <- order(x)
    #   if (skip_first) {
    #     x <- x[-1]
    #   }
    #   x <- c(s_hold, x[!x %in% s_hold])
    #   if (keep_self) {
    #     keep_self_index <- x[1]
    #   } else {
    #     keep_self_index <- NULL
    #   }
    #   x <- x[!x %in% s_remove]
    #   x <- c(keep_self_index, x)
    #   x
    # }
    
    neighbor_indcs <- apply(
      diss_matrix,
      MARGIN = 2,
      FUN = f_order_neigh,
      s_hold = spike_hold,
      s_remove = spike_rm,
      skip_first = skip_first, 
      keep_self = keep_self
    )
    
    if (is.list(neighbor_indcs)) {
      tlength <- sapply(neighbor_indcs, FUN = length)
      ni <- matrix(NA, max(tlength), ncol(diss_matrix))
      for (i in 1:ncol(ni)) {
        ni[1:tlength[i], i] <- neighbor_indcs[[i]]
      }
      neighbor_indcs <- ni
      rm(ni)
    }
    
    stats <- seq(1,
                 nrow(diss_matrix) * ncol(diss_matrix),
                 by = nrow(diss_matrix)
    ) - 1
    neighbors_diss <- sweep(neighbor_indcs,
                            MARGIN = 2,
                            STATS = stats,
                            FUN = "+"
    )
    neighbors_diss[] <- diss_matrix[as.vector(neighbors_diss)]
  } else {
    neighbor_indcs <- apply(diss_matrix, MARGIN = 2, FUN = order)
    neighbors_diss <- apply(diss_matrix, MARGIN = 2, FUN = sort)
    if (skip_first) {
      neighbor_indcs <- neighbor_indcs[-1, ]
      neighbors_diss <- neighbors_diss[-1, ]
    }
  }
  
  
  if (!is.null(k)) {
    neighbor_indcs <- neighbor_indcs[1:k, , drop = FALSE]
    neighbors_diss <- neighbors_diss[1:k, , drop = FALSE]
  } else {
    k_min <- min(k_range)
    k_max <- max(k_range)
    neighbors <- neighbors_diss <= k_diss
    if (!is.null(spike)) {
      n_spike_hold <- sum(spike > 0)
      if (n_spike_hold > 0) {
        neighbors[1:n_spike_hold, ] <- TRUE
      }
    }
    n_neighbors <- original_n_neighbors <- colSums(neighbors, na.rm = TRUE)
    n_neighbors[original_n_neighbors < k_min] <- k_min
    n_neighbors[original_n_neighbors > k_max] <- k_max
    neighbor_indcs <- neighbor_indcs[1:max(n_neighbors), , drop = FALSE]
    neighbors_diss <- neighbors_diss[1:max(n_neighbors), , drop = FALSE]
    
    
    neigh_list <- lapply(n_neighbors, FUN = function(x) 1:x)
    sts <- (max(n_neighbors) * (1:length(n_neighbors))) - max(n_neighbors)
    neighbor_vec_indcs <- sort(unlist(Map("+", neigh_list, sts)))
    neighbor_indcs[-neighbor_vec_indcs] <- NA
    neighbors_diss[-neighbor_vec_indcs] <- NA
  }
  rownames(neighbor_indcs) <- paste0("k_", 1:nrow(neighbor_indcs))
  rownames(neighbors_diss) <- rownames(neighbor_indcs)
  
  # colnames(neighbor_indcs) <- paste0("Xu_", 1:ncol(neighbor_indcs))
  colnames(neighbors_diss) <- colnames(neighbor_indcs)
  
  
  results <- list(
    neighbors_diss = neighbors_diss,
    neighbors = neighbor_indcs,
    unique_neighbors = sort(unique(as.vector(neighbor_indcs)))
  )
  
  if (!is.null(k_diss)) {
    results$k_diss_info <- data.frame(
      Xu_index = 1:ncol(diss_matrix),
      n_k = original_n_neighbors,
      final_n_k = n_neighbors
    )
  }
  
  if (return_dissimilarity) {
    results$dissimilarity <- diss_matrix
  }
  results
}

#' @title A function for computing the spectral information divergence between
#' spectra (sid)
#' @description
#' \loadmathjax
#' \ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#experimental'><img src='figures/lifecycle-experimental.svg' alt='Experimental lifecycle'></a>}}{\strong{Experimental}}
#'
#' This function computes the spectral information divergence/dissimilarity between
#' spectra based on the kullback-leibler divergence algorithm (see details).
#' @usage
#' sid(Xr, Xu = NULL,
#'     mode = "density",
#'     center = FALSE, scale = FALSE,
#'     kernel = "gaussian",
#'     n = if(mode == "density") round(0.5 * ncol(Xr)),
#'     bw = "nrd0",
#'     reg = 1e-04,
#'     ...)
#' @param Xr a matrix containing the spectral (reference) data.
#' @param Xu an optional matrix containing the spectral data of a second set of
#' observations.
#' @param mode the method to be used for computing the spectral information
#' divergence. Options are \code{"density"} (default) for computing the divergence
#' values on the density distributions of the spectral observations, and
#' \code{"feature"} for computing the divergence vales on the spectral variables.
#' See details.
#' @param center a logical indicating if the computations must be carried out on
#' the centred \code{X} and \code{Xu} (if specified) matrices. If
#' \code{mode = "feature"} centring is not carried out since this option does
#' not accept negative values which are generated after centring the matrices.
#' Default is FALSE. See details.
#' @param scale a logical indicating if the computations must be carried out on
#' the variance scaled \code{X} and \code{Xu} (if specified) matrices. Default
#' is TRUE.
#' @param kernel if \code{mode = "density"} a character string indicating the
#' smoothing kernel to be used. It must be one of \code{"gaussian"} (default),
#' \code{"rectangular"}, \code{"triangular"}, \code{"epanechnikov"},
#' \code{"biweight"}, \code{"cosine"} or \code{"optcosine"}. See the
#' \code{\link[stats]{density}} function of the \code{stats} package.
#' @param n if \code{mode = "density"} a numerical value indicating the number
#' of equally spaced points at which the density is to be estimated. See the
#' \code{\link[stats]{density}} function of the \code{stats} package for further
#' details. Default is \code{round(0.5 * ncol(X))}.
#' @param bw if \code{mode = "density"} a numerical value indicating the
#' smoothing kernel bandwidth to be used. Optionally the character string
#' \code{"nrd0"} can be used, it computes the bandwidth using the \code{bw.nrd0}
#' function of the \code{stats} package (see \code{bw.nrd0}). See the
#' \code{\link[stats]{density}} and the \code{bw.nrd0} functions for more
#' details. By default \code{"nrd0"} is used, in this case the bandwidth is
#' computed as \code{bw.nrd0(as.vector(X))}, if \code{Xu} is specified the
#' bandwidth is computed as \code{bw.nrd0(as.vector(rbind(X, Xu)))}.
#' @param reg a numerical value larger than 0 which indicates a regularization
#' parameter. Values (probabilities) below this threshold are replaced by this
#' value for numerical stability. Default is 1e-4.
#' @param ... additional arguments to be passed to the
#' \code{\link[stats]{density}} function of the base package.
#' @details
#' This function computes the spectral information divergence (distance)
#' between spectra.
#' When \code{mode = "density"}, the function first computes the probability
#' distribution of each spectrum which result in a matrix of density
#' distribution estimates. The density distributions of all the observations in
#' the datasets are compared based on the kullback-leibler divergence algorithm.
#' When \code{mode = "feature"}, the kullback-leibler divergence between all
#' the observations is computed directly on the spectral variables.
#' The spectral information divergence (SID) algorithm (Chang, 2000) uses the
#' Kullback-Leibler divergence (\mjeqn{KL}{KL}) or relative entropy
#' (Kullback and Leibler, 1951) to account for the vis-NIR information provided
#' by each spectrum. The SID between two spectra (\mjeqn{x_{i}}{x_i} and
#' \mjeqn{x_{j}}{x_j}) is computed as follows:
#'
#' \mjdeqn{sid(x_{i},x_{j}) = KL(x_{i} \left |\right | x_{j}) + KL(x_{j} \left |\right | x_{i})}{sid(x_i,x_j) = KL(x_i  |\ | x_j) + KL(x_j  |\ | x_i)}
#'
#' \mjdeqn{sid(x_{i},x_{j}) = \sum_{l=1}^{k} p_l \ log(\frac{p_l}{q_l}) + \sum_{l=1}^{k} q_l \ log(\frac{q_l}{p_l})}{sid(x_i,x_j) = \sum_{l=1}^{k} p_l \ log(p_l/q_l) + \sum_{l=1}^{k} q_l \ log(q_l/p_l)}
#'
#' where \mjeqn{k}{k} represents the number of variables or spectral features,
#' \mjeqn{p}{p} and \mjeqn{q}{q} are the probability vectors of \mjeqn{x_{i}}{x_i} and
#' \mjeqn{x_{i}}{x_j} respectively which are calculated as:
#'
#' \mjdeqn{p = \frac{x_i}{\sum_{l=1}^{k} x_{i,l}}}{p = x_i/{\sum_{l=1}^{k} x_{i,l}}}
#'
#' \mjdeqn{q = \frac{x_j}{\sum_{l=1}^{k} x_{j,l}}}{q = x_j1/\sum_{l=1}^{k} x_{j,l}}
#'
#' From the above equations it can be seen that the original SID algorithm
#' assumes that all the components in the data matrices are nonnegative.
#' Therefore centering cannot be applied when \code{mode = "feature"}. If a
#' data matrix with negative values is provided and \code{mode = "feature"},
#' the \code{sid} function automatically scales the matrix as follows:
#'
#' \mjdeqn{X_s = \frac{X-min(X)}{max(X)-min(X)}}{X_s = {X-min(X)}/{max(X)-min(X)}}
#'
#' or
#'
#' \mjdeqn{X_{s} = \frac{X-min(X, Xu)}{max(X, Xu)-min(X, Xu)}}{X_{s} = {X-min(X, Xu)}/{max(X, Xu)-min(X, Xu)}}
#'
#' \mjdeqn{Xu_{s} = \frac{Xu-min(X, Xu)}{max(X, Xu)-min(X, Xu)}}{Xu_{s} = {Xu-min(X, Xu)}/{max(X, Xu)-min(X, Xu)}}
#'
#' if \code{Xu} is specified. The 0 values are replaced by a regularization
#' parameter (\code{reg} argument) for numerical stability.
#' The default of the \code{sid} function is to compute the SID based on the
#' density distributions of the spectra (\code{mode = "density"}). For each
#' spectrum in \code{X} the density distribution is computed using the
#' \code{\link[stats]{density}} function of the \code{stats} package.
#' The 0 values of the estimated density distributions of the spectra are
#' replaced by a regularization parameter (\code{"reg"} argument) for numerical
#' stability. Finally the divergence between the computed spectral histogramas
#' is computed using the SID algorithm. Note that if \code{mode = "density"},
#' the \code{sid} function will accept negative values and matrix centering
#' will be possible.
#' @return a \code{list} with the following components:
#' \itemize{
#'  \item{\code{sid}: if only \code{"X"} is specified (i.e. \code{Xu = NULL}),
#'  a square symmetric matrix of SID distances between all the components in
#'  \code{"X"}. If both \code{"X"} and \code{"Xu"} are specified, a matrix
#'  of SID distances between the components in \code{"X"} and the components
#'  in \code{"Xu"}) where the rows represent the objects in \code{"X"} and the
#'  columns represent the objects in \code{"Xu"}}
#'  \item{\code{Xr}: the (centered and/or scaled if specified) spectral
#'  \code{X} matrix}
#'  \item{\code{Xu}: the (centered and/or scaled if specified) spectral
#'  \code{Xu} matrix}
#'  \item{\code{densityDisXr}: if \code{mode = "density"}, the computed
#'  density distributions of \code{Xr}}
#'  \item{\code{densityDisXu}: if \code{mode = "density"}, the computed
#'  density distributions of \code{Xu}}
#'  }
#' @references Chang, C.I. 2000. An information theoretic-based approach to
#' spectral variability, similarity and discriminability for hyperspectral
#' image analysis. IEEE Transactions on Information Theory 46, 1927-1932.
#' @seealso \code{\link[stats]{density}}
#' @author Leonardo Ramirez-Lopez
#' @importFrom stats bw.nrd0 density
#' @examples
#' \donttest{
#' library(prospectr)
#'
#' data(NIRsoil)
#'
#' Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
#' Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
#' Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
#' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#'
#' Xu <- Xu[!is.na(Yu), ]
#' Xr <- Xr[!is.na(Yr), ]
#'
#' # Example 1
#' # Compute the SID distance between all the observations in Xr
#' xr_sid <- sid(Xr)
#' xr_sid
#'
#' # Example 2
#' # Compute the SID distance between the observations in Xr and the observations
#' # in Xu
#' xr_xu_sid <- sid(Xr, Xu)
#' xr_xu_sid
#' }
#' @export

#######################################################################
# resemble
# Copyright (C) 2014 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#######################################################################

## History:
## 09.03.2014 Leo     Doc examples  were formated with a max. line width
## 23.05.2020 Leo     - Argument scaled renamed to scale
##                    - default for scale has changed from TRUE to FALSE
##                    - the dimnames of the resulting matrix are now
##                      Xr_1... Xr_n (previusly Xr.1... Xr.n)


sid <- function(Xr, Xu = NULL,
                mode = "density",
                center = FALSE, scale = FALSE,
                kernel = "gaussian",
                n = if (mode == "density") round(0.5 * ncol(Xr)),
                bw = "nrd0",
                reg = 1e-4, ...) {

  # Sanity checks
  if (!is.logical(center)) {
    stop("'center' argument must be logical")
  }

  if (!is.logical(scale)) {
    stop("'scale' argument must be logical")
  }

  if (!is.numeric(reg)) {
    stop("'reg' argument must be numeric")
  }

  if (length(reg) > 1) {
    stop("'reg' must be a single numeric value")
  }

  n <- n
  X <- Xr
  rm(Xr)
  if (is.null(Xu)) {
    if (mode == "density") {
      match.arg(kernel, c("gaussian", "rectangular", "triangular", "biweight", "epanechnikov", "cosine", "optcosine"))

      if (center) {
        X <- sweep(x = X, MARGIN = 2, FUN = "-", STATS = colMeans(X))
      }
      if (scale) {
        X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = apply(X, 2, sd))
      }
      Xo <- X

      if (bw == "nrd0") {
        bw <- bw.nrd0(as.vector(X))
      }

      rng <- c(0.65 * min(X), 1.35 * max(X))
      # rng <- c(0.80 * min(X), 1.2 * max(X))
      str <- seq(rng[1], rng[2], diff(rng) / (n - 1))

      densM <- matrix(0, nrow(X), length(str))
      colnames(densM) <- str
      for (i in 1:nrow(X))
      {
        densM[i, ] <- density(X[i, ], from = rng[1], to = rng[2], n = n, kernel = kernel, bw = bw, ...)$y
      }
      X <- densM

      reg <- 10^-4
      w <- X < reg
      X[w] <- reg
      rm(densM)

      ps <- sweep(x = (X), MARGIN = 1, FUN = "/", STATS = rowSums((X)))
      qs <- sweep(x = (X), MARGIN = 1, FUN = "/", STATS = rowSums((X)))


      abKld <- matrix(NA, nrow(X), nrow(X))
      for (i in 1:nrow(X))
      {
        dvsn <- sweep(x = ps, MARGIN = 2, FUN = "/", STATS = qs[i, ])
        abKld[i, ] <- rowSums(ps * (log(dvsn)))
      }
      klKl <- (abKld + t(abKld)) / 2
      colnames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")
      rownames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")
      rownames(X) <- rownames(klKl)
      return(list(sid = klKl, Xr = Xo, densityDisXr = X, xdval = str))
    }

    if (mode == "feature") {
      if (center) {
        message("centering has been disregarded since when mode = 'feature', the function does not accept negative values (which are generated by centering)")
      }
      if (sum(abs(X)) != (sum(X))) {
        X <- (X - min(X)) / diff(range(X))
        message(paste("the 'X' matrix was rescaled from 0 to 1 values, since it contained negative values.", if (scale) {
          "After this the variables were scaled to unit variance."
        }))
      }
      if (scale) {
        X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = apply(X, 2, sd))
      }
      Xo <- X

      reg <- 10^-4
      w <- X < reg
      X[w] <- reg

      ps <- sweep(x = (X), MARGIN = 1, FUN = "/", STATS = rowSums((X)))
      qs <- sweep(x = (X), MARGIN = 1, FUN = "/", STATS = rowSums((X)))


      abKld <- matrix(NA, nrow(X), nrow(X))
      for (i in 1:nrow(X))
      {
        dvsn <- sweep(x = ps, MARGIN = 2, FUN = "/", STATS = qs[i, ])
        abKld[i, ] <- rowSums(ps * (log(dvsn)))
      }
      klKl <- (abKld + t(abKld)) / 2
      colnames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")
      rownames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")
      return(list(sid = klKl, Xr = X))
    }
  }


  if (!is.null(Xu)) {
    if (ncol(X) != ncol(Xu)) {
      stop("The number of variables in X must be equal to the number of variables in Xu")
    }

    if (mode == "density") {
      match.arg(kernel, c("gaussian", "rectangular", "triangular", "biweight", "epanechnikov", "cosine", "optcosine"))

      X <- rbind(X, Xu)
      if (center) {
        X <- sweep(x = X, MARGIN = 2, FUN = "-", STATS = colMeans(X))
      }
      if (scale) {
        X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = apply(X, 2, sd))
      }
      Xo <- X


      if (bw == "nrd0") {
        bw <- bw.nrd0(as.vector(X))
      }

      rng <- c(0.65 * min(X), 1.35 * max(X))
      # rng <- c(0.80 * min(X), 1.2 * max(X))
      str <- seq(rng[1], rng[2], diff(rng) / (n - 1))

      densM <- matrix(0, nrow(X), length(str))
      colnames(densM) <- str
      for (i in 1:nrow(X))
      {
        densM[i, ] <- density(X[i, ], from = rng[1], to = rng[2], n = n, kernel = kernel, bw = bw, ...)$y
      }

      reg <- 10^-4
      w <- densM < reg
      densM[w] <- reg

      Xu <- densM[(nrow(X) - nrow(Xu) + 1):nrow(densM), , drop = FALSE]
      X <- densM[1:(nrow(densM) - nrow(Xu)), , drop = FALSE]
      rm(densM)

      ps <- sweep(x = X, MARGIN = 1, FUN = "/", STATS = rowSums(X))
      qs <- sweep(x = Xu, MARGIN = 1, FUN = "/", STATS = rowSums(Xu))

      klKl <- matrix(NA, nrow(X), nrow(Xu))
      for (i in 1:nrow(Xu))
      {
        dvsnA <- sweep(x = ps, MARGIN = 2, FUN = "/", STATS = qs[i, ])
        dvsnB <- sweep(x = ps^-1, MARGIN = 2, FUN = "/", STATS = (qs[i, ]^-1))
        klKl[, i] <- (rowSums(ps * log(dvsnA)) + colSums(qs[i, ] * t(log(dvsnB)))) / 2
      }
      colnames(klKl) <- paste("Xu", 1:nrow(Xu), sep = "_")
      rownames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")
      rownames(X) <- rownames(klKl)
      rownames(Xu) <- colnames(klKl)

      return(list(
        sid = klKl, Xr = Xo[1:nrow(X), , drop = FALSE], Xu = Xo[(nrow(Xo) - nrow(Xu) + 1):nrow(Xo), , drop = FALSE],
        densityDisXr = X, densityDisXu = Xu,
        xdval = str
      ))
    }

    if (mode == "feature") {
      X <- rbind(X, Xu)
      if (center) {
        warning("centering was not applied since when mode = 'feature', the function does not accept negative values (which are generated by centering)")
      }
      if (sum(abs(X)) != (sum(X))) {
        X <- (X - min(X)) / diff(range(X))
        message(paste("since either or both 'X' and 'Xu' contained negative values, these matrices were rescaled  to a 0 to 1 range (see details on the 'sid' function).", if (scale) {
          "After this the variables were scaled to unit variance."
        }, "Finally the matrix was split back into 'X' and 'Xu'."))
      }
      if (scale) {
        X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = apply(X, 2, sd))
      }

      Xu <- X[(nrow(X) - nrow(Xu) + 1):nrow(X), ]
      X <- X[1:(nrow(X) - nrow(Xu)), ]

      ps <- sweep(x = X, MARGIN = 1, FUN = "/", STATS = rowSums(X))
      qs <- sweep(x = Xu, MARGIN = 1, FUN = "/", STATS = rowSums(Xu))

      klKl <- matrix(NA, nrow(X), nrow(Xu))
      for (i in 1:nrow(Xu))
      {
        dvsnA <- sweep(x = ps, MARGIN = 2, FUN = "/", STATS = qs[i, ])
        dvsnB <- sweep(x = ps^-1, MARGIN = 2, FUN = "/", STATS = (qs[i, ]^-1))
        klKl[, i] <- (rowSums(ps * log(dvsnA)) + colSums(qs[i, ] * t(log(dvsnB)))) / 2
      }

      colnames(klKl) <- paste("Xu", 1:nrow(Xu), sep = "_")
      rownames(klKl) <- paste("Xr", 1:nrow(X), sep = "_")

      return(list(sid = klKl, Xr = X, Xu = Xu))
    }
  }
}

#' @title A function to create calibration and validation sample sets for
#' leave-group-out cross-validation
#' @description for internal use only! This is stratified sampling based on the
#' values of a continuous response variable (y). If group is provided, the
#' sampling is done based on the groups and the average of y per group. This
#' function is used to create calibration and validation groups for
#' leave-group-out cross-validations (or
#' leave-group-of-groups-out cross-validation if group argument is provided).
#' @param y a matrix of one column with the response variable.
#' @param p the percentage of samples (or groups if group argument is used) to
#' retain in the validation_indices set
#' @param number the number of sample groups to be crated
#' @param group the labels for each sample in \code{y} indicating the group each
#' observation belongs to.
#' @param replacement A logical indicating sample replacements for the
#' calibration set are required.
#' @param seed an integer for random number generator (default \code{NULL}).
#' @return a list with two matrices (\code{hold_in} and
#' \code{hold_out}) giving the indices of the observations in each
#' column. The number of columns represents the number of sampling repetitions.
#' @keywords internal
#' @noRd

sample_stratified <- function(
    y, p, number, group = NULL, replacement = FALSE, seed = NULL
) {
  
  ## If the percentage of samples to build the hold_in subset is below 50% of
  ## the total number of samples, the selection is based on the number of samples
  ## to retain.
  ## On the other hand, if the percentage of samples to build the hold_in subset
  ## is equal or above 50% of the total number of samples, the selection is based
  ## on the number of samples to exclude.

  # Account for machine precision for calculating the number of samples to take
  # later. The reason is that the floor function can return wrong results, e.g.
  # floor((1 - 0.9) * 50) returns (incorrectly) 4. The following way to round
  # the number of samples to take should solve this issue
  machine_precision <- ifelse(
    is.null(.Machine$sizeof.longdouble) || .Machine$sizeof.longdouble == 0,
    8,
    .Machine$sizeof.longdouble - 2
  )

  if (p < 0.5) {
    p_to_sample <- round(p, machine_precision)
    do_sampling_for <- "calibration"
  } else {
    p_to_sample <- round(1 - p, machine_precision)
    do_sampling_for <- "validation"
  }

  # Stratified sampling if no group is given
  if (is.null(group)) {
    # Number of samples that must be sampled
    nv <- floor(p_to_sample * nrow(y))
    nv <- ifelse(nv < 1, 1, nv)

    # Calculate the number of samples in the validation and calibration sets,
    # respectively.
    n_val <- ifelse(p >= 0.5, nv, nrow(y) - nv)
    n_cal <- ifelse(replacement, nrow(y), nrow(y) - n_val)

    # Compute the strata, including how many samples must be taken in each of them.
    strata_category <- optim_sample_strata(
      y = y,
      n = nv
    )

    calibration_indices <- matrix(NA, n_cal, number)
    validation_indices <- matrix(NA, n_val, number)

    colnames(calibration_indices) <- colnames(validation_indices) <- paste0(
      "Resample_",
      seq(1:number)
    )
    rownames(calibration_indices) <- paste0("index_", seq(1:n_cal))
    rownames(validation_indices) <- paste0("index_", seq(1:nrow(validation_indices)))

    # The seed is set both here and if there is a group given. Alternatively,
    # one could set the seed before the if(is.null(group)) statement. However,
    # the way it is done now should ensure that future changes to the code do not
    # change the results of the sampling with the same seeds.
    if (!is.null(seed)) {
      set.seed(seed)
    }

    # Obtain the samples drawn from the calculated strata for a total number of
    # 'number' times.
    for (jj in 1:number) {
      strata_samples <- get_samples_from_strata(
        y = y,
        original_order = strata_category$sample_strata$original_order,
        strata = strata_category$sample_strata$strata,
        samples_per_strata = strata_category$samples_to_get,
        replacement = replacement,
        sampling_for = do_sampling_for
      )
      calibration_indices[, jj] <- strata_samples$calibration
      validation_indices[, jj] <- strata_samples$validation
    }
  } else {
    # Stratified sampling if a group is given.
    # The idea is to compute the means of each group and proceed to do the
    # stratified sampling based on the vector of means instead. In particular,
    # this ensures that all members of a groups will always be in the same
    # validation or calibration set.
    y_groups <- data.frame(
      y = y,
      group = factor(group),
      original_order = 1:length(y)
    )
    gr_levels <- levels(y_groups$group)
    n_levels <- nlevels(y_groups$group)
    # Compute means per group and make the stratified sampling based on that
    # vector of means
    aggregated_y <- tapply(
      X = y_groups$y,
      INDEX = y_groups$group,
      FUN = mean
    )

    # Number of groups that must be sampled
    nv <- floor(p_to_sample * n_levels)
    nv <- ifelse(nv < 1, 1, nv)

    # Calculate the number of samples in the validation and calibration sets,
    # respectively.
    n_val <- ifelse(p >= 0.5, nv, nrow(y) - nv)
    n_cal <- ifelse(replacement, nrow(y), nrow(y) - n_val)

    # Compute the strata, including how many groups must be taken in each of them.
    strata_category <- optim_sample_strata(
      y = aggregated_y,
      n = nv
    )

    calibration_groups_indices <- validation_groups_indices <- NULL
    # The seed is set both here and if there is no group given. Alternatively,
    # one could set the seed before the if(is.null(group)) statement. However,
    # the way it is done now should ensure that future changes to the code do not
    # change the results of the sampling with the same seeds.
    if (!is.null(seed)) {
      set.seed(seed)
    }

    # Obtain the groups drawn from the calculated strata for a total number of
    # 'number' times.
    for (jj in 1:number) {
      strata_samples <- get_samples_from_strata(
        y = aggregated_y,
        original_order = strata_category$sample_strata$original_order,
        strata = strata_category$sample_strata$strata,
        samples_per_strata = strata_category$samples_to_get,
        sampling_for = do_sampling_for,
        replacement = replacement
      )

      # Since we sampled based on the group levels, we must convert these levels
      # back to the sample indices.
      if (replacement) {
        sel_sample_indices <- lapply(1:length(strata_samples$calibration),
          FUN = function(ith, full_groups, cal_groups) {
            ## this equivalent to extract from
            ## y_groups$original_order
            which(full_groups %in% cal_groups[ith])
          },
          full_groups = y_groups$group,
          cal_groups = gr_levels[strata_samples$calibration]
        )
        sel_sample_indices <- do.call("c", sel_sample_indices)
      } else {
        sel_sample_indices <- y_groups$original_order[y_groups$group %in% gr_levels[strata_samples$calibration]]
      }

      calibration_groups_indices[[jj]] <- sel_sample_indices
      validation_groups_indices[[jj]] <- y_groups$original_order[y_groups$group %in% gr_levels[strata_samples$validation]]
    }
    # Put everything together into the calibration indices.
    lengths_list_calibration_indices <- lengths(calibration_groups_indices)
    lengths_list_validation_indices <- lengths(validation_groups_indices)
    calibration_indices <- matrix(NA, max(lengths_list_calibration_indices), number)
    validation_indices <- matrix(NA, max(lengths_list_validation_indices), number)
    colnames(calibration_indices) <- colnames(validation_indices) <- paste0(
      "Resample_",
      seq(1:number)
    )
    rownames(calibration_indices) <- paste0("index_", 1:nrow(calibration_indices))
    rownames(validation_indices) <- paste0("index_", 1:nrow(validation_indices))
    if (!is.null(seed)) {
      set.seed(seed + 1)
    }
    # Groups can be of different sizes, so the
    for (jj in 1:number) {
      ## randomly select the replacements for the missing indexes
      ## replacements are necessary!
      n_cal_missing <- max(lengths_list_calibration_indices) - lengths_list_calibration_indices[jj]
      jj_add_list_calibration_indices <- sample(calibration_groups_indices[[jj]],
        size = n_cal_missing,
        replace = TRUE
      )

      n_val_missing <- max(lengths_list_validation_indices) - lengths_list_validation_indices[jj]
      jj_add_list_validation_indices <- sample(validation_groups_indices[[jj]],
        size = n_val_missing,
        replace = TRUE
      )

      calibration_indices[1:lengths_list_calibration_indices[jj], jj] <- calibration_groups_indices[[jj]]
      validation_indices[1:lengths_list_validation_indices[jj], jj] <- validation_groups_indices[[jj]]
      calibration_indices[-(1:lengths_list_calibration_indices[jj]), jj] <- jj_add_list_calibration_indices
      validation_indices[-(1:lengths_list_validation_indices[jj]), jj] <- jj_add_list_validation_indices
    }
  }

  list(
    hold_in = calibration_indices,
    hold_out = validation_indices
  )
}

#' @title A function to assign values to sample distribution strata
#' @description for internal use only! This function takes a continuous variable,
#' creates n strata based on its distribution and assigns the corresponding starta
#' to every value.
#' @param y a matrix of one column with the response variable.
#' @param n the number of strata.
#' @return a data table with the input \code{y} and the corresponding strata to
#' every value.
#' @noRd
#' @keywords internal
get_sample_strata <- function(y, n = NULL, probs = NULL) {
  if (!is.null(n) & !is.null(probs)) {
    stop(
      "both n and probs have been passed to the function, only one of them can be accepted"
    )
  }

  if (!is.null(n)) {
    probs <- seq(0, 1, length = (n + 1))
  }

  y_strata <- unique(
    quantile(
      y,
      probs = probs,
      names = FALSE
    )
  )

  strata_labels <- 1:max(length(y_strata) - 1, 1)
  y_cuts <- cut(
    y,
    breaks = y_strata,
    labels = strata_labels,
    include.lowest = TRUE,
    right = FALSE # Use left closed intervals for compatibility with Julia code
  )

  strata_category <- data.frame(
    original_order = 1:length(y),
    strata = y_cuts
  )
  strata_category
}

#' @title A function to construct an optimal strata for the samples, based on
#' the distribution of the given y.
#' @description for internal use only! This function computes the optimal strata
#' from the distribution of the given y
#' @param y a matrix of one column with the response variable.
#' @param n number of samples that must be sampled.
#' @return a list with two \code{data.frame} objects: \code{sample_strata} contains
#' the optimal strata, whereas \code{samples_to_get} contains information on how
#' many samples per stratum are supposed to be drawn.
#' @keywords internal
#' @noRd
optim_sample_strata <- function(y, n) {
  sample_strata <- get_sample_strata(y, n)
  n_strata <- length(unique(sample_strata$strata))
  iter <- 1
  table_strata <- table(sample_strata$strata)

  # If the number of strata is lower than number of samples, we must correct the
  # strata such that we have the correct number of samples in each stratum.
  #
  # The idea of the following loop is to find the optimal number of strata to
  # properly collect random samples from each of them, as there might be some
  # strata that do not contain any samples. Therefore, we have to widen the
  # strata. As a consequence, the number of strata will be reduced.
  if (n_strata < n || min(table_strata, na.rm = TRUE) < 3) {
    keep_going <- TRUE
    new_n <- n
    new_min_samples_per_strata <- 2 # There must be at least 2 samples per strata
    while (keep_going) {
      # n_vec <- 1:new_n

      # We have to increase the minimum number of samples per strata as we decrease
      # the number of strata. We increase by two; one for the drawing the sample
      # set, one for drawing the replacement set.
      new_min_samples_per_strata <- new_min_samples_per_strata + 2

      # s_missing <- n_vec[-as.numeric(names(table_strata))]
      # v_missing <- rep(0, length(s_missing))
      # names(v_missing) <- s_missing
      # table_strata <- c(table_strata, v_missing)

      # Assign a new number of strata
      new_n <- ceiling(n / (iter + 1))

      # Get the new sample strata for the reduced number of strata
      sample_strata <- get_sample_strata(y, new_n)
      table_strata <- table(sample_strata$strata)

      # Check that each strata contains at least the previously set minimum
      # number of samples. Must be strictly less; otherwise eliminate randomness
      # from the sampling.
      condition_1 <- new_min_samples_per_strata < min(table_strata)
      # Check that all the strata contain samples
      condition_2 <- length(table_strata) == new_n
      # Check that we can further reduce the number of strata without issue
      condition_3 <- new_min_samples_per_strata >= n

      if ((condition_1 & condition_2) | condition_3) {
        break
      }
      iter <- iter + 1
    }

    # Describes how many samples to get for each strata. We start by assigning
    # the same number of samples in each stratum. For some strata, we then reduce
    # this number, as there will (in most cases) more total samples in all strata
    # than the total number of samples available.
    samples_to_get <- data.frame(
      strata = levels(sample_strata$strata),
      samples_to_get = rep(
        new_min_samples_per_strata / 2,
        nlevels(sample_strata$strata)
      )
    )
    # Compute how many samples are currently assigned in samples_to_get.
    total_samples <- sum(samples_to_get$samples_to_get)
    # Derive the difference between the total samples in all strata and the number
    # of available samples. In case this number is larger then zero, we must remove
    # some samples to get from the strata.
    to_remove <- total_samples - n
    if (to_remove > 0) {
      # Remove one sample to be drawn from strata with least samples in them
      highest_freqs <- names(sort(table_strata, decreasing = FALSE)[1:to_remove])
      samples_to_get$samples_to_get[samples_to_get$strata %in% highest_freqs] <-
        samples_to_get$samples_to_get[samples_to_get$strata %in% highest_freqs] - 1
    }
  } else {
    # In case the strata already satisfies the above requirements of having exactly
    # n strata and at least 3 samples in each stratum, we do not have to correct
    # the strata further and can get exactly 1 sample in each stratum.
    samples_to_get <- data.frame(
      strata = levels(sample_strata$strata),
      samples_to_get = 1
    )
  }
  # Return the strata and how many samples per stratum to get.
  list(
    sample_strata = sample_strata,
    samples_to_get = samples_to_get
  )
}

#' @title A function for stratified calibration/validation sampling
#' @description for internal use only! This function selects samples
#' based on provided strata.
#' @param original_order a matrix of one column with the response variable.
#' @param strata the number of strata.
#' @param sampling_for sampling to select the calibration samples ("calibration")
#' or sampling to select the validation samples ("validation").
#' @param replacement logical indicating if sampling with replacement must be
#' done.
#' @return a list with the indices of the calibration and validation samples.
#' @keywords internal
#' @noRd
get_samples_from_strata <- function(
    y,
    original_order,
    strata,
    samples_per_strata,
    sampling_for = c("calibration", "validation"),
    replacement = FALSE
) {
  # For validation, we double the samples to get per strata in case of replacement.
  # The reason is that we will sample two distinct sets of samples; one denotes
  # the validation indices, while the other describes the replacement indices
  # that will replace the drawn samples. This avoids drawing with replacement
  # for the validation set, which would result in over-optimistic estimates
  # of the RMSE.
  val_replacement <- replacement & sampling_for == "validation"
  if (val_replacement) {
    samples_per_strata$samples_to_get <- 2 * samples_per_strata$samples_to_get
  }

  # This functions draws samples from a given single stratum. We do not draw with
  # replacement here, as the replacements are done manually later.
  # For selecting the replacement samples in cases where a strata has only one
  # sample, the replacement sample is randomly selected from the data
  get_random_sample <- function(x, ns) {
    if (length(x) == 1) {
      # this is required to keep the name of the strata, otherwise it fails
      x <- c(x, x)
    }
    sample(x, size = ns)
  }
  max_samples <- max(samples_per_strata$samples_to_get)
  vec_samples <- rep(NA, max_samples)

  # Here, we draw samples from each strata. In case of sampling for validation
  # with replacement, we draw twice as many samples compared to the other methods.
  # These drawn samples will be split into two sets, one denoting the validation
  # indices, whereas the second one describes the replacement indices that replace
  # the drawn samples.
  strata_samples <- lapply(levels(strata),
    FUN = function(strata,
                   original_order,
                   samples_per_strata,
                   vec_samples,
                   replacement,
                   ii) {
      ith_n <- samples_per_strata$samples_to_get[samples_per_strata$strata == ii]
      ith_set <- original_order[which(strata == ii)]
      ith_sel <- get_random_sample(ith_set, ith_n)
      if (replacement) {
        # Split the drawn samples into the two above described sets. The sets are
        # both contained in vec_samples, but these vectors will be split in half
        # later. Note that we must take special care with the indexing here, to
        # ensure that the samples are put into the correct set.
        ln <- (length(ith_sel) / 2)
        vec_samples[1:ln] <- ith_sel[1:ln]
        vec_samples[(length(vec_samples) - ln + 1):length(vec_samples)] <- ith_sel[-(1:ln)]
      } else {
        vec_samples[1:length(ith_sel)] <- ith_sel
      }
      vec_samples
    },
    strata = strata,
    original_order = original_order,
    samples_per_strata = samples_per_strata,
    vec_samples = vec_samples,
    replacement = val_replacement
  )

  # Put all drawn samples into one single matrix
  strata_samples <- do.call("rbind", strata_samples)
  if (val_replacement) {
    # The samples in case of validation sampling with replacement are separated
    # here into validation and replacement sets. The rows of the matrix
    # samples are the above sampled indices. Hence, the columns correspond to
    # above constructed separation, where the first half are describing the
    # validations set and the rest describe the replacement set.
    col_s <- 1:(ncol(strata_samples) / 2)
    col_replacement <- -col_s
    strata_samples <- cbind(
      sort(strata_samples[, col_s]),
      sort(strata_samples[, col_replacement])
    )
  } else {
    # In any other case, we can simply put them into a matrix
    strata_samples <- as.matrix(sort(strata_samples))
  }


  if (sampling_for == "validation") {
    if (replacement) {
      # Ensure that the validation and replacement sets are distinct.
      unique_sample_strata <- which(strata_samples[, 1] == strata_samples[, 2])
      if (length(unique_sample_strata > 0)) {
        solve_replacement <- sample(original_order[-strata_samples[, 2]], length(unique_sample_strata))
        strata_samples[unique_sample_strata, 2] <- solve_replacement
      }
      replacement_indices <- strata_samples[, 2]
    } else {
      replacement_indices <- NULL
    }
    # These are the calibration and validation sets. The replacement indices
    # are added later to the keep parameter in case of validation sampling
    # with replacement.
    keep <- original_order[!original_order %in% strata_samples[, 1]]
    exclude <- as.vector(sort(strata_samples[, 1]))
  }


  if (sampling_for == "calibration") {
    # For calibration, we can already use the constructed calibration and validation
    # sets. In case of sampling with replacements, the replacement indices are
    # added afterwards
    keep <- strata_samples[, 1]
    exclude <- original_order[!original_order %in% keep]
    if (replacement) {
      # replacement_indices <- sample(keep, length(original_order) - length(keep), replace = TRUE)
      # We draw the replacement indices for calibration from a quartile strata,
      # based on the calibration set.
      replacement_indices <- get_quartile_samples(y[keep], n = length(original_order) - length(keep))
      replacement_indices <- keep[replacement_indices]
    } else {
      replacement_indices <- NULL
    }
  }
  # original
  # if (sampling_for == "calibration") {
  #   keep <- strata_samples[, 1]
  #   exclude <- original_order[!original_order %in% keep]
  #   if (replacement) {
  #     (length(original_order) +length(keep)) / (length(original_order) - length(keep))
  #     ip <-
  #      replacement_indices <-  sample_stratified(y = y[keep, ], p = ip, number = 1, seed = seed)
  #     replacement_indices <- sample(keep, length(original_order) - length(keep), replace = TRUE)
  #   } else {
  #     replacement_indices <- NULL
  #   }
  # }

  # Add replacement indices to the calibration set.
  keep <- sort(as.vector(c(keep, replacement_indices)))

  strata_samples <- list(
    calibration = keep,
    validation = exclude
  )
  strata_samples
}




get_quartile_samples <- function(y, n) {
  quartile_probs <- seq(0, 1, 0.25)
  if (length(y) > 1) {
    strata <- get_sample_strata(y, probs = quartile_probs)
  } else {
    return(NULL)
  }
  nn <- floor(n / length(quartile_probs))

  if (nn < 1) {
    nn <- 1
  }

  ## samples per stratum
  ssampling <- data.frame(
    stratum = sort(as.numeric(unique(strata$strata))),
    n = nn
  )

  # I think this is impossible to reach.
  # if (sum(ssampling$n) > n) {
  #   repl <- length(ssampling$stratum) < sum(ssampling$n) - n
  #   removes <- sample(ssampling$stratum,
  #                     sum(ssampling$n) - n,
  #                     replace = repl)
  #   removes <- table(removes)
  #   for (i in as.numeric(names(adds))) {
  #     if (!is.na(adds[i])) {
  #       i_add <- adds[i]
  #     } else {
  #       i_add <- 0
  #     }
  #     ssampling$n[i] <- ssampling$n[ssampling$stratum == i] - i_add
  #   }
  # }

  if (sum(ssampling$n) < n) {
    repl <- length(ssampling$stratum) < n - sum(ssampling$n)
    adds <- sample(ssampling$stratum, n - sum(ssampling$n), replace = repl)
    if (length(adds) > 1) {
      adds <- table(adds)
      for (i in as.numeric(names(adds))) {
        if (!is.na(adds[i])) {
          i_add <- adds[i]
        } else {
          i_add <- 0
        }
        ssampling$n[i] <- ssampling$n[ssampling$stratum == i] + i_add
      }
    } else {
      ssampling$n[ssampling$stratum == adds] <- ssampling$n[ssampling$stratum == adds] + 1
    }
  }

  strat_samples <- lapply(
    ssampling$stratum,
    FUN = function(stratum_data, stratum_n, i) {
      smp <- stratum_data$original_order[stratum_data$strata == i]
      i_n <- stratum_n$n[stratum_n$stratum == i]
      sample(smp,
        i_n,
        replace = ifelse(i_n > length(smp), TRUE, FALSE)
      )
    },
    stratum_data = strata,
    stratum_n = ssampling
  )

  strat_samples <- sort(unlist(strat_samples))

  if (length(strat_samples) > n) {
    overs <- length(strat_samples) - n
    strat_samples <- strat_samples[-sample(length(strat_samples), overs)]
  }

  if (length(strat_samples) < n) {
    unders <- n - length(strat_samples)
    strat_samples <- c(strat_samples, sort(sample(strat_samples, unders, replace = TRUE)))
  }
  strat_samples
}

#' @title Get the package version info
#' @description returns package info.
#' @param pkg the package name i.e "resemble"
#' @keywords internal
pkg_info <- function(pkg = "resemble") {
  fld <- c("Version", "Config/VersionName", "URL")
  pinfo <- read.dcf(system.file("DESCRIPTION", package = pkg), fields = fld)
  pinfo
}

#' @title evaluation of multiple distances obtained with multiple PCs
#' @description internal
#' @keywords internal
#' @noRd
eval_multi_pc_diss <- function(scores,
                               side_info,
                               from = 1,
                               to = ncol(scores),
                               steps = 1,
                               method = c("pc", "pls"),
                               check_dims = TRUE) {
  if (ncol(side_info) == 1) {
    extract_sim_results <- function(x) {
      ## this takes either rmsd or kappa as only one will be in the matrix
      measure_names <- colnames(x$eval)[colnames(x$eval) %in% c("rmsd", "kappa")]
      return(list(
        result = x$eval[, measure_names],
        measure_names = measure_names
      ))
    }
    n_cols_results <- 1
  } else {
    extract_sim_results <- function(x) {
      measure_names <- c(
        paste0("rmsd_", names(x$eval[, "rmsd"])),
        "mean_standardized_rmsd_Yr"
      )
      return(list(
        result = c(
          x$eval[, "rmsd"],
          x$global_eval[, "mean_standardized_rmsd"]
        ),
        measure_names = measure_names
      ))
    }
    n_cols_results <- ncol(side_info) + 1
  }

  eval_pcs <- seq(from = 1, to = to, by = steps)
  results <- matrix(NA, length(eval_pcs), n_cols_results)

  std_scores <- sweep(scores,
    MARGIN = 2,
    STATS = get_column_sds(scores),
    FUN = "/"
  )

  if (check_dims) {
    if (nrow(side_info) != nrow(scores)) {
      stop(paste0(
        "Number of observations in 'scores' do not match with the ",
        "number of observations in 'side_info'"
      ))
    }
  } else {
    std_scores <- std_scores[1:nrow(side_info), , drop = FALSE]
  }

  d <- 0
  for (i in eval_pcs) {
    d <- d + fast_diss_vector(
      std_scores[, i, drop = TRUE]
    )
    tmp <- diss_evaluate(
      diss = d,
      side_info = side_info
    )
    ith_result <- extract_sim_results(tmp)
    results[i, 1:n_cols_results] <- unlist(ith_result$result)
  }

  colnames(results) <- ith_result$measure_names
  eval_pcs <- matrix(eval_pcs, dimnames = list(eval_pcs, method))
  results <- cbind(pc = eval_pcs, results)

  if (ncol(side_info) == 1) {
    colnames(results) <- gsub("rmsd", "rmsd_Yr", colnames(results))
  }

  if ("kappa" %in% ith_result$measure_names) {
    best_pc <- eval_pcs[which.max(results[, "kappa"])]
  }
  if ("mean_standardized_rmsd_Yr" %in% ith_result$measure_names) {
    best_pc <- eval_pcs[which.min(results[, "mean_standardized_rmsd_Yr"])]
  }
  if ("rmsd" %in% ith_result$measure_names) {
    best_pc <- eval_pcs[which.min(results[, "rmsd_Yr"])]
  }

  return(list(results = results, best_pc = best_pc))
}

#' @title Memory-based learning (mbl)
#' @name mbl
#' @aliases mbl plot.mbl get_predictions
#' @description
#' \loadmathjax
#' Memory-based learning (a.k.a. instance-based learning or local regression)
#' is a non-linear lazy learning approach for predicting a response variable
#' from predictor variables. For each observation in a prediction set, a local
#' regression is fitted using a subset of similar observations (nearest
#' neighbors) from a reference set. This function does not produce a global
#' model.
#'
#' @usage
#' mbl(Xr, Yr, Xu, Yu = NULL,
#'     neighbors,
#'     diss_method = diss_pca(ncomp = ncomp_by_opc()),
#'     diss_usage = c("none", "predictors", "weights"),
#'     fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
#'     spike = NULL, group = NULL,
#'     gh = FALSE,
#'     control = mbl_control(),
#'     verbose = TRUE, seed = NULL, ...)
#'
#' @usage \method{plot}{mbl}(x, what = c("validation", "gh"), metric = "rmse", ncomp = c(1, 2), ...)
#'
#' @usage get_predictions(x)
#' 
#' @param Xr A matrix of predictor variables for the reference data
#'   (observations in rows, variables in columns). Column names are required.
#' @param Yr A numeric vector or single-column matrix of response values
#'   corresponding to \code{Xr}. NA values are not permitted.
#' @param Xu A matrix of predictor variables for the data to be predicted
#'   (observations in rows, variables in columns). Must have the same column
#'   names as \code{Xr}.
#' @param Yu An optional numeric vector or single-column matrix of response
#'   values corresponding to \code{Xu}. Used for computing prediction
#'   statistics. Default is \code{NULL}.
#' @param neighbors A neighbor selection object specifying how to select
#'   neighbors. Use \code{\link{neighbors_k}()} for fixed k-nearest neighbors
#'   or \code{\link{neighbors_diss}()} for dissimilarity threshold-based
#'   selection.
#' @param diss_method A dissimilarity method object or a precomputed
#'   dissimilarity matrix. Available constructors:
#'   \itemize{
#'     \item \code{\link{diss_pca}()}: Mahalanobis distance in PCA score space. 
#'     This is the default where the number of components is optimized using 
#'     side information (see \code{ncomp_by_opc}()).
#'     \item \code{\link{diss_pls}()}: Mahalanobis distance in PLS score space
#'     \item \code{\link{diss_euclidean}()}: Euclidean distance
#'     \item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
#'     \item \code{\link{diss_cosine}()}: Cosine dissimilarity
#'     \item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
#'   }
#'   A precomputed matrix can also be passed. When \code{diss_usage =
#'   "predictors"}, it must be square with dimensions
#'   \code{(nrow(Xr) + nrow(Xu))} and zeros on the diagonal. Otherwise, it must
#'   have \code{nrow(Xr)} rows and \code{nrow(Xu)} columns.
#' @param diss_usage How dissimilarity information is used in local models:
#'   \itemize{
#'     \item \code{"none"} (default): dissimilarities used only for neighbor
#'       selection
#'     \item \code{"predictors"}: local dissimilarity matrix columns added as
#'       predictors
#'     \item \code{"weights"}: neighbors weighted by dissimilarity using a
#'       tricubic function
#'   }
#' @param fit_method A local fitting method object. Available constructors:
#'   \itemize{
#'     \item \code{\link{fit_pls}()}: Partial least squares regression
#'     \item \code{\link{fit_wapls}()}: Weighted average PLS (default)
#'     \item \code{\link{fit_gpr}()}: Gaussian process regression
#'   }
#' @param spike An integer vector indicating indices of observations in
#'   \code{Xr} to force into (positive values) or exclude from (negative values)
#'   all neighborhoods. Default is \code{NULL}. Spiking does not change
#'   neighborhood size; forced observations displace the most distant neighbors.
#' @param gh Logical indicating whether to compute global Mahalanobis (GH)
#'   distances. Default is \code{FALSE}. GH distances measure how far each
#'   observation lies from the center of the reference set in PLS score space.
#'   The computation uses a fixed methodology: PLS projection with the number
#'   of components selected via \code{\link{ncomp_by_opc}()} (capped at 40).
#'   This is independent of the \code{diss_method} argument.
#' @param group An optional factor assigning group labels to \code{Xr}
#'   observations (e.g., measurement batches). Used to avoid pseudo-replication
#'   in cross-validation: when one observation is held out, all observations
#'   from its group are also removed.
#' @param control A list from \code{\link{mbl_control}()} specifying validation
#'   type, tuning options, and other settings.
#' @param verbose Logical indicating whether to display a progress bar.
#'   Default is \code{TRUE}. Not shown during parallel execution.
#' @param seed An integer for random number generation, enabling reproducible
#'   cross-validation results. Default is \code{NULL}.
#'   
#' @param x An object of class \code{mbl} (as returned by \code{\link{mbl}}).
#' @param what Character vector specifying what to plot. Options are
#'   \code{"validation"} (validation statistics) and/or \code{"gh"} (PLS scores
#'   used for GH distance computation). Default is both.
#' @param metric Character string specifying which validation statistic to plot.
#'   Options are \code{"rmse"}, \code{"st_rmse"}, or \code{"r2"}. Only used when
#'   \code{"validation"} is in \code{what}.
#' @param ncomp Integer vector of length 1 or 2 specifying which PLS components
#'   to plot. Default is \code{c(1, 2)}. Only used when \code{"gh"} is in
#'   \code{what}.
#' @param ... Additional arguments passed to \code{\link[graphics]{plot}}. 
#' Currently unused for `mbl()`.
#' 
#' @param ... Additional arguments (currently unused).
#' 
#' @details
#' ## Spiking
#' The \code{spike} argument forces specific reference observations into or out
#' of neighborhoods. Positive indices are always included; negative indices are
#' always excluded. When observations are forced in, the most distant neighbors
#' are displaced to maintain neighborhood size. See Guerrero et al. (2010).
#'
#' ## Dissimilarity usage
#' When \code{diss_usage = "predictors"}, the local dissimilarity matrix columns
#' are appended as additional predictor variables, which can improve predictions
#' (Ramirez-Lopez et al., 2013a).
#'
#' When \code{diss_usage = "weights"}, neighbors are weighted using a tricubic
#' function (Cleveland and Devlin, 1988; Naes et al., 1990):
#'
#' \mjdeqn{W_{j} = (1 - v^{3})^{3}}{W_j = (1 - v^3)^3}
#'
#' where \mjeqn{v = d(xr_i, xu_j) / \max(d)}{v = d(xr_i, xu_j) / max(d)}.
#'
#' ## GH distance
#' The global Mahalanobis distance (GH) measures how far each observation lies
#' from the center of the reference set. It is always computed using a PLS
#' projection with the number of components optimized via
#' \code{\link{ncomp_by_opc}()} (maximum 40 components or \code{nrow(Xr)},
#' whichever is smaller). This methodology is fixed and independent of the
#' \code{diss_method} specified for neighbor selection.
#'
#' GH distances are useful for identifying extrapolation: observations with
#' high GH values lie far from the calibration space and may yield unreliable
#' predictions.
#'
#' ## Grouping
#' The \code{group} argument enables leave-group-out cross-validation. When
#' \code{validation_type = "local_cv"} in \code{\link{mbl_control}()}, the
#' \code{p} parameter refers to the proportion of groups (not observations)
#' retained per iteration.
#' 
#' ## Deprecated arguments
#' The following arguments from previous versions of \code{resemble} are no
#' longer supported and will throw an error if used: \code{k}, \code{k_diss},
#' \code{k_range}, \code{method}, \code{pc_selection}, \code{center},
#' \code{scale}, and \code{documentation}. See the current argument list for
#' their replacements.
#'
#' @return 
#' ## mbl  
#' For \code{mbl()}, a list of class \code{mbl} containing:
#' \itemize{
#'   \item \code{control}: control parameters from \code{control}
#'   \item \code{fit_method}: fit constructor from \code{fit_method} 
#'   \item \code{Xu_neighbors}: list with neighbor indices and dissimilarities
#'   \item \code{dissimilarities}: dissimilarity method and matrix (if
#'     \code{return_dissimilarity = TRUE} in \code{control})
#'   \item \code{n_predictions}: number of predictions made
#'   \item \code{gh}: GH distances for \code{Xr} and \code{Xu} (if
#'     \code{gh = TRUE})
#'   \item \code{validation_results}: validation statistics by method
#'   \item \code{results}: list of data.frame objects with predictions, one per
#'     neighborhood size
#'   \item \code{seed}: the seed value used
#' }
#'
#' Each results table contains:
#' \itemize{
#'   \item \code{o_index}: observation index
#'   \item \code{k}: number of neighbors used
#'   \item \code{k_diss}, \code{k_original}: (\code{neighbors_diss} only)
#'     threshold and original count
#'   \item \code{ncomp}: (\code{fit_pls} only) number of PLS components
#'   \item \code{min_ncomp}, \code{max_ncomp}: (\code{fit_wapls} only)
#'     component range
#'   \item \code{yu_obs}, \code{pred}: observed and predicted values
#'   \item \code{yr_min_obs}, \code{yr_max_obs}: response range in neighborhood
#'   \item \code{index_nearest_in_Xr}, \code{index_farthest_in_Xr}: neighbor
#'     indices
#'   \item \code{y_nearest}, \code{y_farthest}: neighbor response values
#'   \item \code{diss_nearest}, \code{diss_farthest}: neighbor dissimilarities
#'   \item \code{y_nearest_pred}: (NNv validation) leave-one-out prediction
#'   \item \code{loc_rmse_cv}, \code{loc_st_rmse_cv}: (local_cv validation) CV
#'     statistics
#'   \item \code{loc_ncomp}: (local dissimilarity only) components used locally
#' }
#' 
#' ## Get predictions
#'
#' The \code{get_predictions()} function extracts predicted values from an
#' object of class \code{mbl}. It returns a \code{data.frame} containing the
#' predictions.
#' 
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} and
#' Antoine Stevens
#'
#' @references
#' Cleveland, W. S., and Devlin, S. J. 1988. Locally weighted regression: an
#' approach to regression analysis by local fitting. Journal of the American
#' Statistical Association 83:596-610.
#'
#' Guerrero, C., Zornoza, R., Gomez, I., Mataix-Beneyto, J. 2010. Spiking of
#' NIR regional models using observations from target sites: Effect of model
#' size on prediction accuracy. Geoderma 158:66-77.
#'
#' Naes, T., Isaksson, T., Kowalski, B. 1990. Locally weighted regression and
#' scatter correction for near-infrared reflectance data. Analytical Chemistry
#' 62:664-673.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199:43-53.
#'
#' Rasmussen, C.E., Williams, C.K. 2006. Gaussian Processes for Machine
#' Learning. MIT Press.
#'
#' Shenk, J., Westerhaus, M., Berzaghi, P. 1997. Investigation of a LOCAL
#' calibration procedure for near infrared instruments. Journal of Near
#' Infrared Spectroscopy 5:223-232.
#'
#' @seealso
#' \code{\link{mbl_control}}, \code{\link{neighbors_k}},
#' \code{\link{neighbors_diss}}, \code{\link{diss_pca}}, \code{\link{diss_pls}},
#' \code{\link{fit_pls}}, \code{\link{fit_wapls}}, \code{\link{fit_gpr}},
#' \code{\link{search_neighbors}}
#'
#' @examples
#' \dontrun{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess: detrend + first derivative with Savitzky-Golay
#' sg_det <- savitzkyGolay(
#'   detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
#'   m = 1, p = 1, w = 7
#' )
#' NIRsoil$spc_pr <- sg_det
#'
#' # Split data
#' test_x <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]
#' test_y <- NIRsoil$CEC[NIRsoil$train == 0 & !is.na(NIRsoil$CEC)]
#' train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
#' train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]
#'
#' # Example 1: Spectrum-based learner (Ramirez-Lopez et al., 2013)
#' ctrl <- mbl_control(validation_type = "NNv")
#'
#' sbl <- mbl(
#'   Xr = train_x,
#'   Yr = train_y,
#'   Xu = test_x,
#'   neighbors = neighbors_k(seq(40, 140, by = 20)),
#'   diss_method = diss_pca(ncomp = ncomp_by_opc(40)),
#'   fit_method = fit_gpr(),
#'   control = ctrl
#' )
#' sbl
#' plot(sbl)
#' get_predictions(sbl)
#'
#' # Example 2: With known Yu
#' sbl_2 <- mbl(
#'   Xr = train_x,
#'   Yr = train_y,
#'   Xu = test_x,
#'   Yu = test_y,
#'   neighbors = neighbors_k(seq(40, 140, by = 20)),
#'   fit_method = fit_gpr(),
#'   control = ctrl
#' )
#' plot(sbl_2)
#'
#' # Example 3: LOCAL algorithm (Shenk et al., 1997)
#' local_algo <- mbl(
#'   Xr = train_x,
#'   Yr = train_y,
#'   Xu = test_x,
#'   Yu = test_y,
#'   neighbors = neighbors_k(seq(40, 140, by = 20)),
#'   diss_method = diss_correlation(),
#'   diss_usage = "none",
#'   fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
#'   control = ctrl
#' )
#' plot(local_algo)
#'
#' # Example 4: Using dissimilarity as predictors
#' local_algo_2 <- mbl(
#'   Xr = train_x,
#'   Yr = train_y,
#'   Xu = test_x,
#'   Yu = test_y,
#'   neighbors = neighbors_k(seq(40, 140, by = 20)),
#'   diss_method = diss_pca(ncomp = ncomp_by_opc(40)),
#'   diss_usage = "predictors",
#'   fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
#'   control = ctrl
#' )
#' plot(local_algo_2)
#'
#' # Example 5: Parallel execution
#' library(doParallel)
#' n_cores <- min(2, parallel::detectCores() - 1)
#' clust <- makeCluster(n_cores)
#' registerDoParallel(clust)
#'
#' local_algo_par <- mbl(
#'   Xr = train_x,
#'   Yr = train_y,
#'   Xu = test_x,
#'   Yu = test_y,
#'   neighbors = neighbors_k(seq(40, 140, by = 20)),
#'   diss_method = diss_correlation(),
#'   fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
#'   control = ctrl
#' )
#'
#' registerDoSEQ()
#' try(stopCluster(clust))
#' }
#'
#' @export

######################################################################
# resemble
# Copyright (C) 2026 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################

mbl <- function(
  Xr, Yr, Xu, Yu = NULL,
  neighbors,
  diss_method = diss_pca(ncomp = ncomp_by_opc()),
  diss_usage = c("none", "predictors", "weights"),
  fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
  spike = NULL,
  group = NULL,
  gh = FALSE,
  control = mbl_control(),
  verbose = TRUE,
  seed = NULL,
  ...
) {
  f_call <- match.call()
  dots <- list(...)
  # ---------------------------------------------------------------------------
  # Block removed arguments
  # ---------------------------------------------------------------------------
  
  if ("k" %in% names(dots) || "k_diss" %in% names(dots) || "k_range" %in% names(dots)) {
    stop(
      "Arguments 'k', 'k_diss', 'k_range' have been removed.\n",
      "Use neighbors_k() or neighbors_diss() instead.\n",
      "Example: mbl(..., neighbors = neighbors_k(c(40, 80, 120)))",
      call. = FALSE
    )
  }
  
  if ("method" %in% names(dots)) {
    stop(
      "Argument 'method' has been renamed to 'fit_method'.\n",
      "Use fit_pls(), fit_wapls(), or fit_gpr() constructors.\n",
      "Example: mbl(..., fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15))",
      call. = FALSE
    )
  }
  
  if ("pc_selection" %in% names(dots))  {
    stop(
      "Argument 'pc_selection' has been removed.\n",
      "Component selection is now specified in diss_*() constructors.\n",
      "Example: mbl(..., diss_method = diss_pca(ncomp = ncomp_by_opc(40)))",
      call. = FALSE
    )
  }
  
  if ("center" %in% names(dots) || "scale" %in% names(dots)) {
    stop(
      "Arguments 'center' and 'scale' have been removed.\n",
      "These are now set in diss_*() and fit_*() constructors.\n",
      "Example: diss_pca(center = TRUE, scale = FALSE)",
      call. = FALSE
    )
  }
  
  if ("documentation" %in% names(dots)) {
    stop(
      "Argument 'documentation' has been removed.",
      call. = FALSE
    )
  }
  
  
  # ---------------------------------------------------------------------------
  # Validate constructor arguments
  # ---------------------------------------------------------------------------
  if (missing(neighbors)) {
    stop("'neighbors' is required. Use neighbors_k() or neighbors_diss().", 
         call. = FALSE)
  }
  
  if (!inherits(neighbors, "neighbors")) {
    stop(
      "'neighbors' must be created by neighbors_k() or neighbors_diss().",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Validate inputs
  # ---------------------------------------------------------------------------
  # --- control validation ---
  if (!inherits(control, "mbl_control")) {
    stop("'control' must be created by mbl_control()", call. = FALSE)
  }
  
  # ---------------------------------------------------------------------------
  # Set BLAS threads to reduce overhead (restored on exit)
  # ---------------------------------------------------------------------------
  if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
    old_blas_threads <- blas_get_num_procs()
    if (old_blas_threads != control$blas_threads) {
      blas_set_num_threads(control$blas_threads)
      on.exit(blas_set_num_threads(old_blas_threads), add = TRUE)
    }
  } else if (Sys.info()["sysname"] == "Linux" && control$blas_threads == 1L) {
    message(
      "Tip: Install 'RhpcBLASctl' for optimal performance on Linux:\n",
      "  install.packages('RhpcBLASctl')"
    )
  }
  
  
  if (!is.logical(verbose) || length(verbose) != 1L) {
    stop("'verbose' must be TRUE or FALSE.", call. = FALSE)
  }
  
  if (!is.matrix(Xr)) {
    Xr <- as.matrix(Xr)
  }
  
  if (!is.matrix(Xu)) {
    Xu <- as.matrix(Xu)
  }
  
  if (!is.null(Yu)) {
    if (!is.matrix(Yu)) {
      Yu <- as.matrix(Yu)
    }
  }
  
  if (!is.matrix(Yr)) {
    Yr <- as.matrix(Yr)
  }
  
  
  
  n_xr <- nrow(Xr)
  n_xu <- nrow(Xu)
  n_total <- n_xr + n_xu
  
  if (ncol(Xr) != ncol(Xu)) {
    stop("ncol(Xr) must equal ncol(Xu).", call. = FALSE)
  }
  
  if (ncol(Xr) < 4L) {
    stop("Xr must have at least 4 columns.", call. = FALSE)
  }
  
  if (NROW(Yr) != nrow(Xr)) {
    stop("length(Yr) must equal nrow(Xr).", call. = FALSE)
  }
  
  if (any(is.na(Yr))) {
    stop("NA values in Yr are not supported.", call. = FALSE)
  }
  
  if (!is.null(Yu) && NROW(Yu) != nrow(Xu)) {
    stop("length(Yu) must equal nrow(Xu).", call. = FALSE)
  }
  
  
  # ---------------------------------------------------------------------------
  # Validate diss_method
  # ---------------------------------------------------------------------------
  if (is.matrix(diss_method)) {
    # User passed a precomputed dissimilarity matrix
    diss_matrix <- diss_method
    diss_method <- NULL
    # center <- FALSE
    # scale <- FALSE
  } else if (!inherits(diss_method, "diss_method")) {
    stop(
      "'diss_method' must be a diss_*() constructor or a numeric matrix.\n",
      "Available constructors: diss_pca(), diss_pls(), diss_correlation(), ",
      "diss_euclidean(), diss_cosine(), diss_mahalanobis()",
      call. = FALSE
    )
  } else {
    # ---------------------------------------------------------------------------
    # Validate ncomp against data dimensions
    # ---------------------------------------------------------------------------
    if (!is.null(diss_method$ncomp)) {
      max_ncomp <- .get_max_ncomp(diss_method$ncomp)
      max_allowed <- min(n_total, ncol(Xr))
      
      if (max_ncomp > max_allowed) {
        warning(
          "Requested ncomp (", max_ncomp, ") exceeds max allowed by data dimensions (",
          max_allowed, ").\nIt will be capped to ", max_allowed, ".",
          call. = FALSE
        )
      }
    }
    diss_matrix <- NULL
    # center <- if (!is.null(diss_method$center)) diss_method$center else TRUE
    # scale <- if (!is.null(diss_method$scale)) diss_method$scale else FALSE
  }
  
  if (!inherits(fit_method, "fit_method")) {
    stop(
      "'fit_method' must be created by a fit_*() constructor.\n",
      "Available: fit_pls(), fit_wapls(), fit_gpr()",
      call. = FALSE
    )
  }
  
  diss_usage <- match.arg(diss_usage)
  
  # ---------------------------------------------------------------------------
  # Convert neighbors to internal format
  # ---------------------------------------------------------------------------
  if (inherits(neighbors, "neighbors_k")) {
    k <- neighbors$k
    k_diss <- NULL
    k_range <- NULL
    k_max <- max(k)
    k_diss_max <- NULL
    
    # Validate k against data
    if (k_max > n_xr) {
      stop(
        "k (", k_max, ") cannot exceed nrow(Xr) (", n_xr, ").",
        call. = FALSE
      )
    }
  } else {
    # neighbors_diss
    k <- NULL
    k_diss <- neighbors$threshold
    k_range <- c(neighbors$k_min, neighbors$k_max)
    k_max <- NULL
    k_diss_max <- max(k_diss)
    
    if (is.infinite(k_range[2])) {
      if (verbose) {
        message(
          "setting 'k_max' (", k_range[2], ") to nrow(Xr) (", n_xr, ").",
          call. = FALSE
        )
      }
      k_range[2] <- n_xr
    }
    
    # Validate k_range against data
    if (k_range[2] > n_xr) {
      stop(
        "k_max (", k_range[2], ") cannot exceed nrow(Xr) (", n_xr, ").",
        call. = FALSE
      )
    }
  }
  
  # ---------------------------------------------------------------------------
  # Extract from fit_method constructor
  # ---------------------------------------------------------------------------
  is_fit_pls <- inherits(fit_method, "fit_pls")
  is_fit_wapls <- inherits(fit_method, "fit_wapls")
  is_fit_gpr <- inherits(fit_method, "fit_gpr")
  
  # For fit_and_predict()
  if (is_fit_pls) {
    pred_method <- "pls"
    pls_c <- fit_method$ncomp
  } else if (is_fit_wapls) {
    pred_method <- "wapls"
    pls_c <- c(fit_method$min_ncomp, fit_method$max_ncomp)
  } else {
    pred_method <- "gpr"
    pls_c <- NULL
  }
  
  noise_variance <- fit_method$noise_variance
  fit_scale <- fit_method$scale
  
  # ---------------------------------------------------------------------------
  # Parallel setup
  # ---------------------------------------------------------------------------
  "%mydo%" <- get("%do%")
  if (control$allow_parallel && getDoParRegistered()) {
    "%mydo%" <- get("%dopar%")
  }
  
  
  # ---------------------------------------------------------------------------
  # Validate and set names
  # ---------------------------------------------------------------------------
  if (is.null(colnames(Xr))) {
    stop("Xr must have column names.", call. = FALSE)
  }
  
  if (is.null(colnames(Xu))) {
    stop("Xu must have column names.", call. = FALSE)
  }
  
  if (!identical(colnames(Xr), colnames(Xu))) {
    stop("Column names of Xr and Xu must be identical.", call. = FALSE)
  }
  
  rownames(Xr) <- seq_len(nrow(Xr))
  rownames(Xu) <- seq_len(nrow(Xu))
  
  # ---------------------------------------------------------------------------
  # Validation type
  # ---------------------------------------------------------------------------
  validation_type <- control$validation_type
  is_local_cv <- "local_cv" %in% validation_type
  is_nnv <- "NNv" %in% validation_type
  
  if (is_local_cv && is_nnv) {
    validation_type <- "both"
  }
  
  # ---------------------------------------------------------------------------
  # Validate group
  # ---------------------------------------------------------------------------
  if (!is.null(group)) {
    if (length(group) != nrow(Xr)) {
      stop("length(group) must equal nrow(Xr).", call. = FALSE)
    }
    group <- as.factor(group)
  }
  
  # ---------------------------------------------------------------------------
  # Setup for Xu
  # ---------------------------------------------------------------------------
  if (is_nnv && n_xu < 3L) {
    stop(
      "Nearest neighbor validation requires at least 3 observations in Xu.",
      call. = FALSE
    )
  }
  
  if (!is.null(Yu)) {
    Yu <- as.matrix(Yu)
    if (nrow(Yu) != n_xu) {
      stop("nrow(Yu) must equal nrow(Xu).", call. = FALSE)
    }
  }
  
  pre_nms_ng <- "Xu_"
  ln <- n_xu
  first_nn <- 1L
  y_output_name <- "yu_obs"
  y_hat_output_name <- "pred"
  val_summary_name <- "Yu_prediction_statistics"
  
  # ---------------------------------------------------------------------------
  # Validate fit_method ncomp against data dimensions
  # ---------------------------------------------------------------------------
  if (is_fit_pls) {
    if (fit_method$ncomp > min(n_xr, ncol(Xr))) {
      warning(
        "Requested ncomp (", fit_method$ncomp, ") in fit_pls() exceeds max allowed (",
        min(n_xr, ncol(Xr)), ").\nIt will be capped internally.",
        call. = FALSE
      )
    }
  } else if (is_fit_wapls) {
    max_allowed <- min(n_xr, ncol(Xr))
    if (fit_method$max_ncomp > max_allowed) {
      warning(
        "Requested max_ncomp (", fit_method$max_ncomp, ") in fit_wapls() exceeds max allowed (",
        max_allowed, ").\nIt will be capped internally.",
        call. = FALSE
      )
    }
  }
  
  has_projection <- FALSE
  # Copy diss_method to avoid modifying user's input
  diss_method_internal <- diss_method
  # ---------------------------------------------------------------------------
  # Compute neighborhoods
  # ---------------------------------------------------------------------------
  if (is.null(diss_matrix)) {
    # Force return_projection = TRUE when needed for diss_usage = "predictors"
    if (diss_usage == "predictors") {
      if (inherits(diss_method_internal, "diss_pca") || 
          inherits(diss_method_internal, "diss_pls")) {
        if (!isTRUE(diss_method_internal$return_projection)) {
          diss_method_internal$return_projection <- TRUE
        }
      }
    }
    # Using diss_method constructor
    neighborhoods <- get_neighbor_info(
      Xr = Xr, 
      Xu = Xu,
      diss_method = diss_method_internal, 
      Yr = Yr,
      neighbors = neighbors,
      spike = spike,
      return_dissimilarity = control$return_dissimilarity,
      diss_usage = diss_usage
    )
    
    diss_xr_xu <- neighborhoods$dissimilarity
    
    if (!is.null(neighborhoods$projection)) {
      diss_projection <- neighborhoods$projection
      has_projection <- TRUE
    } else {
      has_projection <- FALSE
    }
    
  } else {
    # Using precomputed diss_matrix
    has_projection <- FALSE
    diss_xr_xr <- NULL
    dim_diss <- dim(diss_matrix)
    
    if (diss_usage == "predictors") {
      if (dim_diss[1] != n_total || dim_diss[2] != n_total || 
          any(diag(diss_matrix) != 0)) {
        stop(
          "When diss_usage = 'predictors', the dissimilarity matrix must be ",
          "square (", n_total, " x ", n_total, ") with zeros on the diagonal.",
          call. = FALSE
        )
      }
  
      diss_xr_xr <- diss_matrix[1:n_xr, 1:n_xr]
      diss_xr_xu <- diss_matrix[1:n_xr, (n_xr + 1):n_total]
      
    } else {
      # diss_usage %in% c("weights", "none")
      if (dim_diss[1] != n_xr || dim_diss[2] != n_xu) {
        stop(
          "When diss_usage = 'weights' or 'none', the dissimilarity matrix ",
          "must have nrow(Xr) (", n_xr, ") rows and nrow(Xu) (", n_xu, ") columns.",
          call. = FALSE
        )
      }
      diss_xr_xu <- diss_matrix
    }
    
    neighborhoods <- diss_to_neighbors(
      diss_xr_xu,
      k = k_max, 
      k_diss = k_diss_max,
      k_range = k_range,
      spike = spike,
      return_dissimilarity = control$return_dissimilarity,
      skip_first = FALSE
    )
    
    neighborhoods$diss_xr_xr <- diss_xr_xr
  }
  
  # ---------------------------------------------------------------------------
  # Compute GH distances (independent of diss_method)
  # ---------------------------------------------------------------------------
  if (gh) {
    # FIXME: this is not ideal, as it requires an additional projection step.
    # in addition the optimisation of components is hard coded
    gh_projection <- ortho_projection(
      Xr = Xr, 
      Xu = Xu,
      Yr = Yr,
      ncomp = ncomp_by_opc(min(n_xr, 40L)),
      method = "pls",
      scale = fit_scale
    )
    
    gh_center <- colMeans(gh_projection$scores[1:n_xr, , drop = FALSE])
    
    neighborhoods$gh <- list(
      gh_Xr = as.vector(f_diss(
        gh_projection$scores[1:n_xr, , drop = FALSE],
        Xu = t(gh_center),
        diss_method = "mahalanobis",
        center = FALSE, 
        scale = FALSE
      )),
      gh_Xu = as.vector(f_diss(
        gh_projection$scores[(n_xr + 1):n_total, , drop = FALSE],
        Xu = t(gh_center),
        diss_method = "mahalanobis",
        center = FALSE, 
        scale = FALSE
      )),
      projection = gh_projection
    )
  } else {
    neighborhoods$gh <- NULL
  }
  
  # ---------------------------------------------------------------------------
  # Determine smallest neighborhood size
  # ---------------------------------------------------------------------------
  if (inherits(neighbors, "neighbors_k")) {
    min_k <- min(neighbors$k)
    smallest_neighborhood <- neighborhoods$neighbors[1:min_k, , drop = FALSE]
    smallest_n_neighbors <- colSums(!is.na(smallest_neighborhood))
  } else {
    # neighbors_diss
    min_threshold <- min(neighbors$threshold)
    min_diss <- neighborhoods$neighbors_diss <= min_threshold
    
    if (!is.null(spike)) {
      spike_in <- sum(spike > 0)
      if (spike_in > 0) {
        min_diss[1:spike_in, ] <- TRUE
      }
    }
    
    smallest_neighborhood <- neighborhoods$neighbors
    smallest_neighborhood[!min_diss] <- NA
    smallest_n_neighbors <- colSums(!is.na(smallest_neighborhood))
    smallest_n_neighbors[smallest_n_neighbors < neighbors$k_min] <- neighbors$k_min
    smallest_n_neighbors[smallest_n_neighbors > neighbors$k_max] <- neighbors$k_max
  }
  
  if (is_local_cv) {
    min_n_samples <- floor(min(smallest_n_neighbors) * control$p) - 1L
    
    if (inherits(neighbors, "neighbors_k")) {
      min_neighborhood <- min(neighbors$k)
    } else {
      min_neighborhood <- neighbors$k_min
    }
    min_cv_samples <- floor(min_neighborhood * (1 - control$p))
    
    if (min_cv_samples < 3L) {
      stop(
        "Local cross-validation requires at least 3 observations in ",
        "the hold-out set. The current cross-validation parameters ",
        "leave less than 3 observations in some neighborhoods.",
        call. = FALSE
      )
    }
  } else {
    min_n_samples <- smallest_n_neighbors - 1L
  }
  
  # ---------------------------------------------------------------------------
  # Validate PLS components vs neighborhood size
  # ---------------------------------------------------------------------------
  if (is_fit_pls || is_fit_wapls) {
    if (is_fit_pls) {
      max_pls <- fit_method$ncomp
    } else {
      max_pls <- fit_method$max_ncomp
    }
    
    if (any(min_n_samples < max_pls)) {
      stop(
        "More PLS components than observations in some neighborhoods.\n",
        "If 'local_cv' is being used, consider that some observations ",
        "in the neighborhoods are held out for local validation.",
        call. = FALSE
      )
    }
  }
  
  # ---------------------------------------------------------------------------
  # Prepare iteration neighborhoods
  # ---------------------------------------------------------------------------
  is_local_diss <- !is.null(diss_method) && 
    (inherits(diss_method, "diss_local_pca") || 
       inherits(diss_method, "diss_local_pls"))
  
  if (is_local_diss) {
    iter_neighborhoods <- ith_mbl_neighbor(
      Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu,
      diss_usage = "none",
      neighbor_indices = neighborhoods$neighbors,
      neighbor_diss = neighborhoods$neighbors_diss,
      group = group
    )
  } else {
    iter_neighborhoods <- ith_mbl_neighbor(
      Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu,
      diss_usage = diss_usage,
      neighbor_indices = neighborhoods$neighbors,
      neighbor_diss = neighborhoods$neighbors_diss,
      diss_xr_xr = neighborhoods$diss_xr_xr,
      group = group
    )
  }
  
  r_fields <- c(
    "o_index", "k_diss", "k_original", "k", "ncomp", "min_ncomp", "max_ncomp",
    y_output_name, y_hat_output_name, "yr_min_obs", "yr_max_obs",
    "index_nearest_in_Xr", "index_farthest_in_Xr",
    "y_nearest", "y_nearest_pred",
    "y_farthest", "diss_nearest", "diss_farthest",
    "loc_rmse_cv", "loc_st_rmse_cv", "loc_ncomp", "rep"
  )
  
  n_ith_result <- ifelse(
    inherits(neighbors, "neighbors_k"), 
    length(neighbors$k), 
    length(neighbors$threshold)
  )
  
  template_pred_results <- as.data.frame(
    matrix(
      NA, n_ith_result, length(r_fields),
      dimnames = list(NULL, r_fields)
    )
  )
  
  template_pred_results$rep[1] <- 0
  
  if (inherits(neighbors, "neighbors_k")) {
    template_pred_results$k <- neighbors$k
  } else {
    template_pred_results$k_diss <- neighbors$threshold
  }
  
  pg_bar_width <- 10
  n_characters <- nchar(n_xu)
  n_iter <- n_xu
  to_erase <- pg_bar_width + (2 * n_characters) + 8
  to_erase <- paste(rep(" ", to_erase), collapse = "")
  
  if (verbose) {
    cat("\033[32m\033[3mPredicting...\n\033[23m\033[39m")
  }
  
  pred_obs <- foreach(
    i = seq_len(n_iter),
    ith_observation = iter_neighborhoods,
    .inorder = FALSE,
    .export = c(
      "dissimilarity", "fit_and_predict", "pls_cv",
      "get_col_sds", "get_wapls_weights",
      "sample_stratified", "gaussian_pr_cv"
    ),
    .noexport = c("Xr", "Xu")
  ) %mydo% {
  
    ith_pred_results <- template_pred_results
    additional_results <- NULL
    ith_pred_results$o_index[] <- i
    
    # if (is_local_diss) {
    #   ith_observation <- get_ith_local_neighbors(
    #     ith_xr = ith_observation$ith_xr,
    #     ith_xu = ith_observation$ith_xu,
    #     ith_yr = ith_observation$ith_yr,
    #     ith_yu = ith_observation$ith_yu,
    #     diss_usage = diss_usage,
    #     ith_neig_indices = ith_observation$ith_neig_indices,
    #     neighbors = neighbors,
    #     spike = spike,
    #     diss_method = diss_method,
    #     ith_group = ith_observation$ith_group,
    #     mbl_is_parallel = control$allow_parallel
    #   )
    #   ith_pred_results$loc_ncomp[] <- ith_observation$ith_ncomp
    #   additional_results$ith_neig_indices <- ith_observation$ith_neig_indices
    #   additional_results$ith_neigh_diss <- ith_observation$ith_neigh_diss
    # }
    
    if (verbose) {
      cat(paste0("\033[34m\033[3m", i, "/", n_iter, "\033[23m\033[39m"))
      pb <- txtProgressBar(width = pg_bar_width, char = "\033[34m_\033[39m")
    }
    
    if (inherits(neighbors, "neighbors_diss")) {
      ith_diss <- ith_observation$ith_neigh_diss
      if (!is.null(spike)) {
        n_spike_hold <- sum(spike > 0)
        if (n_spike_hold > 0) {
          ith_diss[seq_len(n_spike_hold)] <- 0
        }
      }
      k_diss_vec <- neighbors$threshold
      k_min <- neighbors$k_min
      k_max <- neighbors$k_max
      ith_pred_results$k_original <- sapply(k_diss_vec, FUN = function(x, d) sum(d < x), d = ith_diss)
      ith_pred_results$k <- ith_pred_results$k_original
      ith_pred_results$k[ith_pred_results$k_original < k_min] <- k_min
      ith_pred_results$k[ith_pred_results$k_original > k_max] <- k_max
    } else {
      ith_pred_results$k <- neighbors$k
    }
    
    
    for (kk in seq_len(nrow(ith_pred_results))) {
      if (verbose) {
        setTxtProgressBar(pb, kk / nrow(ith_pred_results))
      }
      
      # If the sample has not been predicted before,
      # then create a model and predict it (useful only when k_diss is used)
      current_k <- ith_pred_results$k[kk]
      
      # Skip refitting if neighborhood size unchanged from previous iteration
      # (optimization for neighbors_diss where different thresholds may prorduce 
      # same k)
      if (current_k != ifelse(kk == 1, 0, ith_pred_results$k[kk - 1])) {
        
        # Subset predictors: when diss_usage == "predictors", include local 
        # dissimilarity columns alongside original spectral variables
        if (diss_usage == "predictors") {
          keep_cols <- c(
            1:current_k,
            (1 + ith_observation$n_k):ncol(ith_observation$ith_xr)
          )
          i_k_xr <- ith_observation$ith_xr[1:current_k, keep_cols]
          i_k_xu <- ith_observation$ith_xu[, keep_cols, drop = FALSE]
        } else {
          i_k_xr <- ith_observation$ith_xr[1:current_k, ]
          i_k_xu <- ith_observation$ith_xu
        }
        
        # for extracting some basic stats
        i_k_yr <- ith_observation$ith_yr[first_nn:current_k, , drop = FALSE]
        
        # i_k_yu <- ith_observation$ith_yu
        kth_diss <- ith_observation$ith_neigh_diss[first_nn:current_k]
        i_idx <- ith_observation$ith_neig_indices[first_nn:current_k]
        
        
        ith_pred_results$rep[kk] <- 0
        ith_yr_range <- range(i_k_yr)
        ith_pred_results$yr_min_obs[kk] <- ith_yr_range[1]
        ith_pred_results$yr_max_obs[kk] <- ith_yr_range[2]
        ith_pred_results$diss_farthest[kk] <- max(kth_diss)
        ith_pred_results$diss_nearest[kk] <- min(kth_diss)
        ith_pred_results$y_farthest[kk] <- i_k_yr[which.max(kth_diss)]
        ith_pred_results$y_nearest[kk] <- i_k_yr[which.min(kth_diss)]
        ith_pred_results$index_nearest_in_Xr[kk] <- i_idx[which.min(kth_diss)]
        ith_pred_results$index_farthest_in_Xr[kk] <- i_idx[which.max(kth_diss)]
        
        # Full neighbor set for model fitting (including spiked observations).
        # Stats above excluded spike via first_nn offset. This set is then
        # subsetted at each k iteration.
        i_k_yr <- ith_observation$ith_yr[1:current_k, , drop = FALSE]
        
        
        if (!is.null(group)) {
          i_k_group <- factor(ith_observation$ith_group[1:current_k])
        } else {
          i_k_group <- NULL
        }
        
        if (diss_usage == "weights") {
          # Weights are defined according to a tricubic function
          # as in Cleveland and Devlin (1988) and Naes and Isaksson (1990).
          std_kth_diss <- kth_diss / max(kth_diss)
          kth_weights <- (1 - (std_kth_diss^3))^3
          kth_weights[which(kth_weights == 0)] <- 1e-04
        } else {
          kth_weights <- rep(1, current_k)
        }

        # Local model fit and prediction
        i_k_pred <- fit_and_predict(
          x = i_k_xr,
          y = i_k_yr,
          pred_method = pred_method,
          scale = fit_scale,
          pls_c = pls_c,
          weights = kth_weights,
          newdata = i_k_xu,
          CV = is_local_cv,
          tune = control$tune_locally,
          group = i_k_group,
          p = control$p,
          number = control$number,
          noise_variance = noise_variance,
          range_prediction_limits = control$range_prediction_limits,
          pls_max_iter = fit_method$max_iter,
          pls_tol = fit_method$tol,
          seed = seed,
          algorithm = fit_method$method
        )
        
        ith_pred_results[[y_hat_output_name]][kk] <- i_k_pred$prediction
        
        # Track selected PLS components for NNv validation
        selected_pls <- NULL
        
        if (is_local_cv) {
          # Select best model: tuned or fixed components
          if (control$tune_locally) {
            best_row <- which.min(i_k_pred$validation$cv_results$rmse_cv)
          } else {
            best_row <- ifelse(is_fit_pls, fit_method$ncomp, 1)
          }
          # Store optimized component counts
          if (is_fit_pls) {
            ith_pred_results$ncomp[kk] <- i_k_pred$validation$cv_results$ncomp[best_row]
            selected_pls <- ith_pred_results$ncomp[kk]
          }
          if (is_fit_wapls) {
            ith_pred_results$min_ncomp[kk] <- i_k_pred$validation$cv_results$min_component[best_row]
            ith_pred_results$max_ncomp[kk] <- i_k_pred$validation$cv_results$max_component[best_row]
            selected_pls <- i_k_pred$validation$cv_results[best_row, 1:2]
          }
          
          # Store local CV statistics
          ith_pred_results$loc_rmse_cv[kk] <- i_k_pred$validation$cv_results$rmse_cv[best_row]
          ith_pred_results$loc_st_rmse_cv[kk] <- i_k_pred$validation$cv_results$st_rmse_cv[best_row]
        } else {
          # No local CV: use fixed component counts from fit_method
          if (is_fit_pls) {
            ith_pred_results$ncomp[kk] <- fit_method$ncomp
            selected_pls <- ith_pred_results$ncomp[kk]
          }
          if (is_fit_wapls) {
            ith_pred_results$min_ncomp[kk] <- fit_method$min_ncomp
            ith_pred_results$max_ncomp[kk] <- fit_method$max_ncomp
            selected_pls <- c(fit_method$min_ncomp, fit_method$max_ncomp)
          }
        }
        
        
        if (is_nnv) {
          # Leave-nearest-neighbor-out validation: exclude nearest neighbor (or its group)
          # and predict its value using the remaining neighbors
          if (!is.null(group)) {
            out_group <- which(i_k_group == i_k_group[[ith_observation$local_index_nearest]])
          } else {
            out_group <- ith_observation$local_index_nearest
          }
          
          nearest_pred <- fit_and_predict(
            x = i_k_xr[-out_group, ],
            y = i_k_yr[-out_group, , drop = FALSE],
            pred_method = pred_method,
            scale = fit_scale,
            pls_c = selected_pls,
            noise_variance = noise_variance,
            newdata = i_k_xr[ith_observation$local_index_nearest, , drop = FALSE],
            CV = FALSE,
            tune = FALSE,
            range_prediction_limits = control$range_prediction_limits,
            pls_max_iter = fit_method$max_iter,
            pls_tol = fit_method$tol,
            seed = seed,
            algorithm = fit_method$method
          )$prediction
          
          ith_pred_results$y_nearest_pred[kk] <- nearest_pred / kth_weights[1]
        }
      } else {
        # Neighborhood size unchanged: copy previous results
        ith_k_diss <- ith_pred_results$k_diss[kk]
        ith_pred_results[kk, ] <- ith_pred_results[kk - 1, ]
        ith_pred_results$rep[kk] <- 1
        ith_pred_results$k_diss[kk] <- ith_k_diss
      }
    }
    
    if (verbose) {
      if (kk == nrow(ith_pred_results) & i != n_iter) {
        cat("\r", to_erase, "\r")
      }
      
      if (i == n_iter) {
        cat("\n")
      }
      # do not use close() (it prints a new line)
      ## close(pb)
    }
    list(
      results = ith_pred_results,
      additional_results = additional_results
    )
  }
  
  # Reorder results by observation index (foreach may return out of order)
  iteration_order <- sapply(
    pred_obs,
    FUN = function(x) x$results$o_index[1]
  )
  pred_obs <- pred_obs[order(iteration_order, decreasing = FALSE)]
  
  # Combine all iteration results into single table
  results_table <- do.call(
    "rbind", lapply(pred_obs, FUN = function(x) x$results)
  )
  
  # No k adjustment needed when Xu is provided (always the case now)
  fix_k <- 0
  
  if (is_local_diss) {
    # Reconstruct local dissimilarity matrix from iteration results
    diss_xr_xu <- do.call(
      "cbind",
      lapply(iteration_order,
             FUN = function(ii, x, m) {
               idc <- x[[ii]]$additional_results$ith_neig_indices
               d <- x[[ii]]$additional_results$ith_neigh_diss
               m[idc] <- d
               m
             },
             x = pred_obs,
             m = matrix(NA, nrow(Xr), 1)
      )
    )
    class(diss_xr_xu) <- c("local_ortho_diss", "matrix")
    dimnames(diss_xr_xu) <- list(
      paste0("Xr_", seq_len(nrow(diss_xr_xu))),
      paste0(pre_nms_ng, seq_len(ncol(diss_xr_xu)))
    )
    
    # Reconstruct neighbor indices from iteration results
    neighborhoods$neighbors <- do.call(
      "cbind", lapply(iteration_order,
                      FUN = function(ii, x, m) {
                        idc <- x[[ii]]$additional_results$ith_neig_indices
                        m[seq_len(length(idc))] <- idc
                        m
                      },
                      x = pred_obs,
                      m = matrix(NA, max(results_table$k), 1)
      )
    )
  }
  
  
  out <- c(
    if (is.null(Yu)) "yu_obs",
    if (all(is.na(results_table$k_original))) "k_original",
    if (!is_nnv) "y_nearest_pred",
    if (!is_fit_wapls) c("min_ncomp", "max_ncomp"),
    if (!is_fit_pls) "ncomp",
    if (!is_local_cv) c("loc_rmse_cv", "loc_st_rmse_cv"),
    if (!is_local_diss) "loc_ncomp",
    "rep"
  )
  
  results_table <- results_table[, !(names(results_table) %in% out), drop = FALSE]
  
  
  if (inherits(neighbors, "neighbors_diss")) {
    # Split results by dissimilarity threshold
    k_diss_vec <- neighbors$threshold
    results_table <- lapply(
      k_diss_vec,
      FUN = function(x, tbl, sel) tbl[tbl[[sel]] == x, ],
      tbl = results_table,
      sel = "k_diss"
    )
    names(results_table) <- paste0("k_diss_", k_diss_vec)
    
    # Compute percentage of observations bounded by k_range
    k_min <- neighbors$k_min
    k_max <- neighbors$k_max
    p_bounded <- sapply(
      results_table,
      FUN = function(x, k_min, k_max) {
        sum(x$k_original <= k_min | x$k_original >= k_max)
      },
      k_min = k_min,
      k_max = k_max
    )
    col_ks <- data.frame(
      k_diss = k_diss_vec,
      p_bounded = paste0(round(100 * p_bounded / n_xu, 3), "%")
    )
  } else {
    # Split results by k
    k_vec <- neighbors$k - fix_k
    results_table <- lapply(
      k_vec,
      FUN = function(x, tbl, sel) tbl[tbl[[sel]] == x, ],
      tbl = results_table,
      sel = "k"
    )
    names(results_table) <- paste0("k_", k_vec)
    col_ks <- data.frame(k = k_vec)
  }
  
  
  if (is_nnv) {
    # Compute nearest neighbor validation statistics
    nn_stats <- function(x) {
      nn_rmse <- (mean((x$y_nearest - x$y_nearest_pred)^2))^0.5
      nn_st_rmse <- nn_rmse / diff(range(x$y_nearest))
      nn_rsq <- (cor(x$y_nearest, x$y_nearest_pred))^2
      c(nn_rmse = nn_rmse, nn_st_rmse = nn_st_rmse, nn_rsq = nn_rsq)
    }
    loc_nn_res <- do.call("rbind", lapply(results_table, FUN = nn_stats))
    loc_nn_res <- cbind(
      col_ks,
      rmse = loc_nn_res[, "nn_rmse"],
      st_rmse = loc_nn_res[, "nn_st_rmse"],
      r2 = loc_nn_res[, "nn_rsq"]
    )
  } else {
    loc_nn_res <- NULL
  }
  
  if (is_local_cv) {
    # Compute mean local cross-validation statistics
    mean_loc_res <- function(x) {
      mean_loc_rmse <- mean(x$loc_rmse_cv)
      mean_loc_st_rmse <- mean(x$loc_st_rmse_cv)
      c(loc_rmse = mean_loc_rmse, loc_st_rmse = mean_loc_st_rmse)
    }
    loc_res <- do.call("rbind", lapply(results_table, mean_loc_res))
    loc_res <- cbind(
      col_ks,
      rmse = loc_res[, "loc_rmse"],
      st_rmse = loc_res[, "loc_st_rmse"]
    )
  } else {
    loc_res <- NULL
  }
  
  if (!is.null(Yu)) {
    # Add observed values to results and compute prediction statistics
    for (i in seq_along(results_table)) {
      results_table[[i]][[y_output_name]] <- Yu
    }
    yu_stats <- function(x, y_hat, y) {
      y_rmse <- mean((x[[y_hat]] - x[[y]])^2, na.rm = TRUE)^0.5
      y_st_rmse <- y_rmse / diff(range(x[[y_hat]]), na.rm = TRUE)
      y_rsq <- cor(x[[y_hat]], x[[y]], use = "complete.obs")^2
      c(rmse = y_rmse, st_rmse = y_st_rmse, rsq = y_rsq)
    }
    pred_res <- do.call(
      "rbind",
      lapply(results_table, yu_stats, y_hat = y_hat_output_name, y = y_output_name)
    )
    pred_res <- cbind(
      col_ks,
      rmse = pred_res[, "rmse"],
      st_rmse = pred_res[, "st_rmse"],
      r2 = pred_res[, "rsq"]
    )
  } else {
    pred_res <- NULL
  }
  
  diss_method_name <- if (is_local_diss) {
    paste0(class(diss_method)[1], " (locally computed)")
  } else {
    class(diss_method)[1]
  }
  
  if (control$return_dissimilarity) {
    diss_list <- list(
      diss_method = diss_method,  
      diss_usage = diss_usage,  
      diss_xr_xu = diss_xr_xu
    )
    if (has_projection) {
      diss_list$projection <- diss_projection
    }
  } else {
    diss_method_return <- NULL
    if (inherits(diss_method, "diss_method")) {
      diss_method_return <- diss_method
    } 
    diss_list <- list(
      diss_method = diss_method_return,  
      diss_usage = diss_usage
    )
  }
  
  
  colnames(neighborhoods$neighbors) <- paste0(pre_nms_ng, seq_len(ln))
  rownames(neighborhoods$neighbors) <- paste0("k_", seq_len(nrow(neighborhoods$neighbors)))
  
  val_list <- structure(
    list(
      loc_res,
      loc_nn_res,
      pred_res
    ),
    names = c(
      "local_cross_validation",
      "nearest_neighbor_validation",
      val_summary_name
    )
  )
  
  results_list <- list(
    control = control,
    diss_usage = diss_usage,
    fit_method = fit_method,
    spike = ifelse(is.null(spike), FALSE, TRUE), 
    dissimilarities = diss_list,
    Xu_neighbors = list(
      neighbors = neighborhoods$neighbors,
      neighbors_diss = neighborhoods$neighbors_diss
    ),
    n_predictions = n_xu,
    gh = neighborhoods$gh,
    validation_results = val_list,
    results = results_table,
    seed = seed
  )
  
  attr(results_list, "call") <- f_call
  class(results_list) <- c("mbl", "list")
  
  results_list
}

#' @rdname mbl
#' @export
plot.mbl <- function(x,
                     what = c("validation", "gh"),
                     metric = "rmse",
                     ncomp = c(1, 2),
                     ...) {
  # Validate arguments
  what <- match.arg(what, c("validation", "gh"), several.ok = TRUE)
  metric <- match.arg(metric, c("rmse", "st_rmse", "r2"))
  
  if (!inherits(x, "mbl")) {
    stop("'x' must be an object of class 'mbl'.", call. = FALSE)
  }
  
  # Save and restore graphics parameters
  
  opar <- par("mfrow", "mar")
  on.exit(par(opar))
  
  # Set up multi-panel if needed
  if (length(what) == 2 && !is.null(x$gh)) {
    par(mfrow = c(1, 2))
  }
  
  
  # Process plot arguments
  plot_args <- .process_plot_args(list(...))
  
  # Plot validation results
  if ("validation" %in% what) {
    .plot_validation(x, metric, plot_args)
  }
  
  # Plot GH scores
  if ("gh" %in% what) {
    if (is.null(x$gh)) {
      message("GH distance not available in this object.")
    } else {
      .plot_gh_scores(x, ncomp, plot_args)
    }
  }
  
  mtext(plot_args$main, outer = TRUE, cex = 2, line = -2)
  invisible(x)
}


# =============================================================================
# Internal plotting helpers
# =============================================================================

#' Process plot arguments with defaults
#' @keywords internal
#' @noRd
.process_plot_args <- function(dots) {
  args <- dots
  
  # Extract or set main title
  if ("main" %in% names(args)) {
    main <- args$main
    args$main <- NULL
  } else {
    main <- "Memory-based learning results"
  }
  
  # Set defaults
  if (!"col.axis" %in% names(args)) {
    args$col.axis <- grey(0.3)
  }
  if (!"pch" %in% names(args)) {
    args$pch <- 16
  }
  
  # Remove arguments that are set internally
  internal_args <- c("col", "xlab", "ylab", "type", "ylim", "xlim")
  args <- args[!names(args) %in% internal_args]
  
  args$main <- main
  args
}


#' Plot validation results
#' @keywords internal
#' @noRd
.plot_validation <- function(object, metric, plot_args) {
  # Collect validation results
  val_data <- list()
  colors <- character()
  
  if (!is.null(object$validation_results$nearest_neighbor_validation)) {
    val_data$NNv <- cbind(
      object$validation_results$nearest_neighbor_validation,
      val = "NNv"
    )
    colors <- c(colors, "dodgerblue")
  }
  
  if (!is.null(object$validation_results$local_cross_validation)) {
    val_data$local_cv <- cbind(
      object$validation_results$local_cross_validation,
      r2 = NA,
      val = "local_cv"
    )
    colors <- c(colors, "green4")
  }
  
  if (!is.null(object$validation_results$Yu_prediction_statistics)) {
    val_data$Yu <- cbind(
      object$validation_results$Yu_prediction_statistics,
      val = "Yu prediction"
    )
    colors <- c(colors, "red")
  }
  
  if (!is.null(object$validation_results$Yr_fitted_statistics)) {
    val_data$Yr <- cbind(
      object$validation_results$Yr_fitted_statistics,
      val = "Yr fitted"
    )
    colors <- c(colors, "orange")
  }
  
  if (length(val_data) == 0) {
    par(mfrow = c(1, 1))
    message("No validation results to plot.")
    return(invisible(NULL))
  }
  
  tpl <- do.call(rbind, val_data)
  col_names <- colnames(tpl)
  
  # Determine ID variable
  id_var <- if ("k" %in% col_names) "k" else "k_diss"
  
  # Select columns for metric
  keep_cols <- c(id_var, metric, "val")
  keep_cols <- keep_cols[keep_cols %in% col_names]
  tpl_subset <- data.frame(tpl)[, keep_cols, drop = FALSE]
  
  # Reshape to wide format
  to_plot <- reshape(
    tpl_subset,
    timevar = "val",
    idvar = id_var,
    direction = "wide"
  )
  colnames(to_plot) <- gsub("[.]", " ", colnames(to_plot))
  
  # Remove r2 for local_cv (not available)
  if (metric == "r2") {
    drop_col <- grepl("local_cv", colnames(to_plot))
    to_plot <- to_plot[, !drop_col, drop = FALSE]
    colors <- colors[colors != "green4"]
  }
  
  # Plot
  y_range <- range(to_plot[, -1], na.rm = TRUE)
  y_range[2] <- y_range[2] * 1.1
  
  do.call("matplot", c(
    list(
      x = to_plot[, 1],
      y = to_plot[, -1, drop = FALSE],
      type = "b",
      xlab = id_var,
      ylab = metric,
      ylim = y_range,
      col = colors
    ),
    plot_args[!names(plot_args) %in% "main"]
  ))
  
  grid(col = rgb(0.3, 0.3, 0.3, 0.1), lty = 1, lwd = 1)
  mtext("Validation results", col = grey(0.3))
  
  legend(
    "topright",
    legend = colnames(to_plot)[-1],
    col = colors,
    pch = plot_args$pch,
    box.lty = 0,
    bg = NA
  )
  
  invisible(NULL)
}


#' Plot GH distance scores
#' @keywords internal
#' @noRd
.plot_gh_scores <- function(object, ncomp, plot_args) {
  xr_scores <- object$gh$projection$scores
  n_components <- object$gh$projection$ncomp
  
  # Transform to Mahalanobis space if multicomponent
  if (n_components > 1) {
    xr_scores <- euclid_to_mahal(xr_scores)
    xr_scores <- sweep(xr_scores, 2, colMeans(xr_scores), "-")
  }
  
  # Split Xr and Xu scores
  xu_idx <- grep("Xu_", rownames(xr_scores))
  xr_idx <- grep("Xr_", rownames(xr_scores))
  
  xu_scores <- xr_scores[xu_idx, , drop = FALSE]
  xr_scores <- xr_scores[xr_idx, , drop = FALSE]
  
  # Colors
  xr_col <- rgb(0, 0, 0.4, 0.5)
  xu_col <- rgb(1, 0, 0, 0.5)
  
  if (n_components == 1) {
    .plot_gh_1d(xr_scores, xu_scores, xr_col, xu_col, plot_args)
  } else {
    .plot_gh_2d(xr_scores, xu_scores, ncomp, xr_col, xu_col, plot_args)
  }
  
  invisible(NULL)
}


#' Plot 1D GH scores
#' @keywords internal
#' @noRd
.plot_gh_1d <- function(xr_scores, xu_scores, xr_col, xu_col, plot_args) {
  scores_combined <- c(xr_scores[, 1], xu_scores[, 1])
  set_labels <- c(rep("Xr", nrow(xr_scores)), rep("Xu", nrow(xu_scores)))
  
  df <- data.frame(
    index = seq_along(scores_combined),
    score = scores_combined,
    set = set_labels
  )
  df <- df[order(df$score), ]
  df$index <- seq_len(nrow(df))
  
  rng <- range(scores_combined)
  rng <- 1.2 * c(-max(abs(rng)), max(abs(rng)))
  
  do.call("plot", c(
    list(
      x = df$index[df$set == "Xr"],
      y = df$score[df$set == "Xr"],
      ylim = rng,
      col = xr_col,
      xlab = "Ordered PLS values",
      ylab = "PLS 1"
    ),
    plot_args[!names(plot_args) %in% "main"]
  ))
  
  points(
    df$index[df$set == "Xu"],
    df$score[df$set == "Xu"],
    col = xu_col,
    pch = plot_args$pch
  )
  
  mtext("Partial least squares scores", col = grey(0.3))
  grid(col = rgb(0.3, 0.3, 0.3, 0.1), lty = 1, lwd = 1)
  legend("topleft", legend = c("Xr", "Xu"), col = c(xr_col, xu_col),
         pch = plot_args$pch, cex = 0.8, box.lty = 0, bg = NA)
}


#' Plot 2D GH scores
#' @keywords internal
#' @noRd
.plot_gh_2d <- function(xr_scores, xu_scores, ncomp, xr_col, xu_col, plot_args) {
  rng <- 1.2 * range(xr_scores[, ncomp], xu_scores[, ncomp])
  
  xlab <- paste0("PLS ", ncomp[1], " (standardized)")
  ylab <- paste0("PLS ", ncomp[2], " (standardized)")
  
  do.call("plot", c(
    list(
      x = xr_scores[, ncomp[1]],
      y = xr_scores[, ncomp[2]],
      xlab = xlab,
      ylab = ylab,
      xlim = rng,
      ylim = rng,
      col = xr_col
    ),
    plot_args[!names(plot_args) %in% "main"]
  ))
  
  points(xu_scores[, ncomp, drop = FALSE], col = xu_col, pch = plot_args$pch)
  
  mtext("Partial least squares (Mahalanobis space)", col = grey(0.3))
  grid(col = rgb(0.3, 0.3, 0.3, 0.1), lty = 1, lwd = 1)
  
  legend("topright", legend = c("Xr", "Xu"), col = c(xr_col, xu_col),
         pch = plot_args$pch, cex = 0.8, box.lty = 0, bg = NA)
  
  # Draw reference circles
  max_score <- ceiling(max(abs(xr_scores[, ncomp])))
  for (r in seq_len(max_score)) {
    theta <- seq(0, 2 * pi, length.out = 101)
    lines(r * cos(theta), r * sin(theta),
          col = rgb(0.3, 0.3, 0.3, 0.3), lty = 1, lwd = 0.5)
  }
}



#' @aliases mbl
#' @export
get_predictions <- function(x) {
  if (is.na(match("mbl", class(x)))) {
    stop("the object is not of class 'mbl'")
  }
  
  ext_pred <- function(x, ...) {
    prediction <- x$pred
    return(prediction)
  }
  predictions <- data.frame(sapply(x$results, ext_pred))
  return(predictions)
}

context("test-mbl")
# tests/testthat/test-mbl.R
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# =============================================================================
# Setup helper
# =============================================================================

.setup_mbl_data <- function(n_xr = 40, n_xu = 20, preprocess = FALSE) {
  data("NIRsoil", package = "prospectr")
  
  if (preprocess) {
    NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  }
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ][seq_len(n_xu), ]
  Xr <- Xr[!is.na(Yr), ][seq_len(n_xr), ]
  Yu <- Yu[!is.na(Yu)][seq_len(n_xu)]
  Yr <- Yr[!is.na(Yr)][seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}



# =============================================================================
# mbl_control tests
# =============================================================================

test_that("mbl_control constructor works", {
  ctrl <- mbl_control()
  
  expect_type(ctrl, "list")
  expect_false(ctrl$return_dissimilarity)
  expect_equal(ctrl$validation_type, "NNv")
  expect_true(ctrl$tune_locally)
  expect_equal(ctrl$number, 10L)
  expect_equal(ctrl$p, 0.75)
  expect_true(ctrl$range_prediction_limits)
  expect_true(ctrl$allow_parallel)
  expect_equal(ctrl$blas_threads, 1L)
})


test_that("mbl_control validates inputs", {
  expect_error(mbl_control(return_dissimilarity = "yes"), "TRUE or FALSE")
  expect_error(
    mbl_control(validation_type = "invalid"),
    "NNv.*local_cv.*none"
  )
  expect_error(mbl_control(tune_locally = "yes"), "TRUE or FALSE")
  expect_error(mbl_control(validation_type = "local_cv", number = 0), "positive integer")
  expect_error(mbl_control(validation_type = "local_cv", p = 1.5), "between 0 and 1")
  expect_error(mbl_control(validation_type = "local_cv", p = 0), "between 0 and 1")
  expect_error(mbl_control(allow_parallel = "yes"), "TRUE or FALSE")
  expect_error(mbl_control(blas_threads = 0), "positive integer")
})


test_that("mbl_control rejects 'none' combined with other types", {
  expect_error(
    mbl_control(validation_type = c("NNv", "none")),
    "cannot combine 'none'"
  )
})


# =============================================================================
# Basic mbl functionality tests
# =============================================================================

test_that("mbl with fit_gpr and neighbors_k works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.5,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_gpr(),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
  expect_true("validation_results" %in% names(result))
  expect_true("results" %in% names(result))
  expect_true("Xu_neighbors" %in% names(result))
})


test_that("mbl with fit_pls and neighbors_k works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.5,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
  expect_true("validation_results" %in% names(result))
})


test_that("mbl with fit_pls (mpls method) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_pls(ncomp = 5, method = "mpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_pls (simpls method) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_pls(ncomp = 5, method = "simpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_wapls and neighbors_k works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.5,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_wapls (mpls method) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 5, method = "mpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_wapls (simpls method) works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 5, method = "simpls"),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


# =============================================================================
# neighbors_diss tests
# =============================================================================

test_that("mbl with fit_gpr and neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.5,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 15, k_max = 30),
    fit_method = fit_gpr(),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_pls and neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 15, k_max = 30),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with fit_wapls and neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 15, k_max = 30),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


# =============================================================================
# Group argument tests
# =============================================================================

test_that("mbl with group argument works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.5,
    allow_parallel = FALSE
  )
  
  group <- rep(c(1, 2), length.out = nrow(d$Xr))
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    group = group,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


# =============================================================================
# Dissimilarity method tests
# =============================================================================

test_that("mbl with diss_pca works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_pca(ncomp = 10),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with diss_pls works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_pls(ncomp = 10),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with diss_correlation works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_correlation(ws = 11),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with diss_euclidean works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_euclidean(center = TRUE, scale = TRUE),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with diss_cosine works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_cosine(),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


# =============================================================================
# External dissimilarity matrix tests
# =============================================================================

test_that("mbl with external dissimilarity matrix works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  tol <- 1e-8
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  # Compute external dissimilarity matrix
  combined <- rbind(d$Xr, d$Xu)
  ext_diss <- dissimilarity(
    Xr = combined,
    diss_method = diss_correlation(center = FALSE, scale = FALSE)
  )$dissimilarity
  diag(ext_diss) <- 0
  
  # mbl with external matrix
  result_ext <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = ext_diss,
    diss_usage = "predictors",
    fit_method = fit_gpr(),
    control = ctrl,
    verbose = FALSE
  )
  
  # mbl with internal computation (same method)
  result_int <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_correlation(center = FALSE, scale = FALSE),
    diss_usage = "predictors",
    fit_method = fit_gpr(),
    control = ctrl,
    verbose = FALSE
  )
  
  # Results should be identical
  nnv_ext <- result_ext$validation_results$nearest_neighbor_validation
  nnv_int <- result_int$validation_results$nearest_neighbor_validation
  
  expect_lt(sum(abs(nnv_ext - nnv_int)), tol)
})


# =============================================================================
# diss_usage tests
# =============================================================================

test_that("mbl with diss_usage = 'predictors' works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_pca(ncomp = 10),
    diss_usage = "predictors",
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


test_that("mbl with diss_usage = 'weights' works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    diss_method = diss_pca(ncomp = 10),
    diss_usage = "weights",
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
})


# =============================================================================
# spike argument tests
# =============================================================================

test_that("mbl with spike argument works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_gpr(),
    spike = 1:5,
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
  expect_true(result$spike)
})


# =============================================================================
# gh argument tests
# =============================================================================

test_that("mbl with gh = TRUE works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(c(25, 35)),
    fit_method = fit_pls(ncomp = 5),
    gh = TRUE,
    control = ctrl,
    verbose = FALSE
  )
  
  expect_s3_class(result, "mbl")
  expect_true(!is.null(result$gh))
})


# =============================================================================
# Expected results tests (skipped on CRAN)
# =============================================================================
## it's a sanity check ensuring results stay within plausible bounds rather 
## than testing for exact values. This catches regressions where something 
## breaks catastrophically 
test_that("mbl delivers expected results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.8,
    tune_locally = TRUE,
    allow_parallel = FALSE
  )
  
  k_test <- c(40, 150)
  tseed <- 141020
  
  # GPR
  set.seed(tseed)
  gpr <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_gpr(noise_variance = 0.0001, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # PLS
  set.seed(tseed)
  pls <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_pls(ncomp = 10, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # WAPLS
  set.seed(tseed)
  wapls <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15, scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # MPLS
  set.seed(tseed)
  mpls <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_pls(ncomp = 10, method = "mpls", scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # WAMPLS
  set.seed(tseed)
  wampls <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "mpls", scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  
  wampls_simpls <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "simpls", scale = TRUE),
    control = ctrl,
    verbose = FALSE
  )
  
  # Check local CV RMSE bounds
  expect_true(all(
    gpr$validation_results$local_cross_validation$rmse < 2.5 &
      gpr$validation_results$local_cross_validation$rmse > 1.5
  ))
  
  expect_true(all(
    pls$validation_results$local_cross_validation$rmse < 2 &
      pls$validation_results$local_cross_validation$rmse > 1.4
  ))
  
  expect_true(all(
    wapls$validation_results$local_cross_validation$rmse < 1.8 &
      wapls$validation_results$local_cross_validation$rmse > 1.5
  ))
  
  expect_true(all(
    mpls$validation_results$local_cross_validation$rmse < 2 &
      mpls$validation_results$local_cross_validation$rmse > 1.5
  ))
  
  expect_true(all(
    wampls$validation_results$local_cross_validation$rmse < 1.9 &
      wampls$validation_results$local_cross_validation$rmse > 1.5
  ))
  # WAMPLS SIMPLS bounds (should be similar to WAPLS/WAMPLS)
  expect_true(all(
    wampls_simpls$validation_results$local_cross_validation$rmse < 1.9 &
      wampls_simpls$validation_results$local_cross_validation$rmse > 1.5
  ))
  
  
  # Check NNv R2 bounds
  expect_true(all(gpr$validation_results$nearest_neighbor_validation$r2 > 0.50))
  expect_true(all(pls$validation_results$nearest_neighbor_validation$r2 > 0.74))
  expect_true(all(wapls$validation_results$nearest_neighbor_validation$r2 > 0.80))
  
  # Check Yu prediction R2 bounds
  expect_true(all(gpr$validation_results$Yu_prediction_statistics$r2 > 0.67))
  expect_true(all(pls$validation_results$Yu_prediction_statistics$r2 > 0.60))
  expect_true(all(wapls$validation_results$Yu_prediction_statistics$r2 > 0.69))
  expect_true(all(wampls_simpls$validation_results$nearest_neighbor_validation$r2 > 0.78))
  expect_true(all(wampls_simpls$validation_results$Yu_prediction_statistics$r2 > 0.65))
})


test_that("mbl with neighbors_diss delivers expected results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.8,
    tune_locally = TRUE,
    allow_parallel = FALSE
  )
  
  tseed <- 141020
  
  # GPR with neighbors_diss
  set.seed(tseed)
  gpr_diss <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 20, k_max = 100),
    fit_method = fit_gpr(),
    control = ctrl,
    verbose = FALSE
  )
  
  # PLS with neighbors_diss
  set.seed(tseed)
  pls_diss <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 20, k_max = 100),
    fit_method = fit_pls(ncomp = 10),
    control = ctrl,
    verbose = FALSE
  )
  
  # WAPLS with neighbors_diss
  set.seed(tseed)
  wapls_diss <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_diss(threshold = 0.1, k_min = 20, k_max = 100),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
    control = ctrl,
    verbose = FALSE
  )
  
  # Check bounds
  expect_true(all(
    gpr_diss$validation_results$local_cross_validation$rmse < 2.8 &
      gpr_diss$validation_results$local_cross_validation$rmse > 2.5
  ))
  
  expect_true(all(
    pls_diss$validation_results$local_cross_validation$rmse < 1.8 &
      pls_diss$validation_results$local_cross_validation$rmse > 1.4
  ))
  
  expect_true(all(
    wapls_diss$validation_results$local_cross_validation$rmse < 2 &
      wapls_diss$validation_results$local_cross_validation$rmse > 1.4
  ))
  
  expect_true(all(gpr_diss$validation_results$nearest_neighbor_validation$r2 > 0.76))
  expect_true(all(pls_diss$validation_results$nearest_neighbor_validation$r2 > 0.81))
  expect_true(all(wapls_diss$validation_results$nearest_neighbor_validation$r2 > 0.81))
  
  expect_true(all(gpr_diss$validation_results$Yu_prediction_statistics$r2 > 0.60))
  expect_true(all(pls_diss$validation_results$Yu_prediction_statistics$r2 > 0.60))
  expect_true(all(wapls_diss$validation_results$Yu_prediction_statistics$r2 > 0.65))
})


test_that("mbl with group argument delivers expected results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  NIRsoil$spc <- prospectr::savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4, 
    p = 0.8,
    tune_locally = TRUE,
    allow_parallel = FALSE
  )
  
  k_test <- c(40, 150)
  tseed <- 141020
  
  xgroup <- rep(seq_len(floor(nrow(Xr) / 2)), each = 2)
  
  set.seed(tseed)
  pls_group <- mbl(
    Xr = Xr, Yr = Yr, Xu = Xu, Yu = Yu,
    neighbors = neighbors_k(k_test),
    fit_method = fit_pls(ncomp = 10),
    control = ctrl, 
    group = xgroup,
    verbose = FALSE
  )
  
  expect_true(all(
    pls_group$validation_results$local_cross_validation$rmse < 2 &
      pls_group$validation_results$local_cross_validation$rmse > 1.4
  ))
  
  expect_true(all(pls_group$validation_results$nearest_neighbor_validation$r2 > 0.7))
})


.with_test_device <- function(code) {
  tf <- tempfile(fileext = ".pdf")
  grDevices::pdf(tf)
  on.exit({
    grDevices::dev.off()
    unlink(tf)
  }, add = TRUE)
  force(code)
}

.build_mbl_for_plot <- function(gh = FALSE) {
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    allow_parallel = FALSE
  )
  
  mbl(
    Xr = d$Xr,
    Yr = d$Yr,
    Xu = d$Xu,
    Yu = d$Yu,
    neighbors = neighbors_k(25),
    fit_method = fit_pls(ncomp = 5),
    gh = gh,
    control = ctrl,
    verbose = FALSE,
    seed = 42
  )
}

test_that("plot.mbl rejects non-mbl objects", {
  expect_error(
    plot(structure(list(), class = "not_mbl")),
    "'x' is a list."
  )
})

test_that("plot.mbl validation plot works and restores par", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  obj <- .build_mbl_for_plot(gh = FALSE)
  
  old_par <- par(c("mfrow", "mar"))
  
  .with_test_device({
    expect_invisible(
      plot(obj, what = "validation", metric = "rmse", main = "MBL validation")
    )
  })
  
  expect_equal(par("mfrow"), old_par$mfrow)
  expect_equal(par("mar"), old_par$mar)
})

test_that("plot.mbl GH plot works when gh is available", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  obj <- .build_mbl_for_plot(gh = TRUE)
  
  .with_test_device({
    expect_invisible(
      plot(obj, what = "gh", ncomp = c(1, 2), main = "MBL GH")
    )
  })
})

test_that("plot.mbl combined validation and GH plots work", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  obj <- .build_mbl_for_plot(gh = TRUE)
  
  old_par <- par(c("mfrow", "mar"))
  
  .with_test_device({
    expect_invisible(
      plot(
        obj,
        what = c("validation", "gh"),
        metric = "rmse",
        ncomp = c(1, 2),
        main = "MBL combined"
      )
    )
  })
  
  expect_equal(par("mfrow"), old_par$mfrow)
  expect_equal(par("mar"), old_par$mar)
})

test_that("plot.mbl reports missing GH distances", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  obj <- .build_mbl_for_plot(gh = FALSE)
  
  .with_test_device({
    expect_message(
      plot(obj),
      "GH distance not available in this object."
    )
  })
})

test_that(".plot_gh_1d works on simple scores", {
  xr_scores <- matrix(c(-2, -1, 0, 1), ncol = 1)
  xu_scores <- matrix(c(-1.5, 0.5), ncol = 1)
  
  plot_args <- list(
    pch = 16,
    col.axis = grey(0.3),
    main = "1D GH"
  )
  
  .with_test_device({
    expect_true(
      is.list(
        invisible(
          resemble:::.plot_gh_1d(
            xr_scores = xr_scores,
            xu_scores = xu_scores,
            xr_col = rgb(0, 0, 0.4, 0.5),
            xu_col = rgb(1, 0, 0, 0.5),
            plot_args = plot_args
          )
        )
      )
    )
  })
})

test_that(".plot_gh_2d works on simple scores", {
  xr_scores <- cbind(c(-2, -1, 1, 2), c(-1, 1, -1, 1))
  xu_scores <- cbind(c(-0.5, 0.5), c(0.25, -0.25))
  
  plot_args <- list(
    pch = 16,
    col.axis = grey(0.3),
    main = "2D GH"
  )
  
  .with_test_device({
    expect_null(
      invisible(
        resemble:::.plot_gh_2d(
          xr_scores = xr_scores,
          xu_scores = xu_scores,
          ncomp = c(1, 2),
          xr_col = rgb(0, 0, 0.4, 0.5),
          xu_col = rgb(1, 0, 0, 0.5),
          plot_args = plot_args
        )
      )
    )
  })
})


# =============================================================================
# Deprecated argument tests
# =============================================================================

test_that("mbl rejects deprecated k, k_diss, k_range arguments", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      k = 30,
      verbose = FALSE
    ),
    "Arguments 'k', 'k_diss', 'k_range' have been removed"
  )
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      k_diss = 0.5,
      verbose = FALSE
    ),
    "Arguments 'k', 'k_diss', 'k_range' have been removed"
  )
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      k_range = c(10, 50),
      verbose = FALSE
    ),
    "Arguments 'k', 'k_diss', 'k_range' have been removed"
  )
})

test_that("mbl rejects deprecated method argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      method = "pls",
      verbose = FALSE
    ),
    "Argument 'method' has been renamed to 'fit_method'"
  )
})

test_that("mbl rejects deprecated pc_selection argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      pc_selection = list(method = "opc"),
      verbose = FALSE
    ),
    "Argument 'pc_selection' has been removed"
  )
})

test_that("mbl rejects deprecated center and scale arguments", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      center = TRUE,
      verbose = FALSE
    ),
    "Arguments 'center' and 'scale' have been removed"
  )
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      scale = FALSE,
      verbose = FALSE
    ),
    "Arguments 'center' and 'scale' have been removed"
  )
})

test_that("mbl rejects deprecated documentation argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      documentation = "test",
      verbose = FALSE
    ),
    "Argument 'documentation' has been removed"
  )
})

# =============================================================================
# Input validation tests
# =============================================================================

test_that("mbl requires neighbors argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'neighbors' is required"
  )
})

test_that("mbl validates neighbors class", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = c(20, 30),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'neighbors' must be created by neighbors_k\\(\\) or neighbors_diss\\(\\)"
  )
})

test_that("mbl validates control class", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      control = list(validation_type = "NNv"),
      verbose = FALSE
    ),
    "'control' must be created by mbl_control\\(\\)"
  )
})

test_that("mbl validates verbose argument", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = "yes"
    ),
    "'verbose' must be TRUE or FALSE"
  )
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = c(TRUE, FALSE)
    ),
    "'verbose' must be TRUE or FALSE"
  )
})

test_that("mbl validates ncol(Xr) == ncol(Xu)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu[, 1:10],
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "ncol\\(Xr\\) must equal ncol\\(Xu\\)"
  )
})

test_that("mbl validates Xr has at least 4 columns", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  Xr_small <- d$Xr[, 1:3]
  Xu_small <- d$Xu[, 1:3]
  colnames(Xr_small) <- colnames(Xu_small) <- paste0("V", 1:3)
  
  expect_error(
    mbl(
      Xr = Xr_small, Yr = d$Yr, Xu = Xu_small,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 2),
      verbose = FALSE
    ),
    "Xr must have at least 4 columns"
  )
})

test_that("mbl validates Yr length matches Xr rows", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr[1:10], Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "length\\(Yr\\) must equal nrow\\(Xr\\)"
  )
})

test_that("mbl rejects NA values in Yr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  Yr_na <- d$Yr
  Yr_na[1] <- NA
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = Yr_na, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "NA values in Yr are not supported"
  )
})

test_that("mbl validates Yu length matches Xu rows", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu[1:5],
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "length\\(Yu\\) must equal nrow\\(Xu\\)"
  )
})

test_that("mbl validates diss_method type", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      diss_method = "pca",
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "'diss_method' must be a diss_\\*\\(\\) constructor or a numeric matrix"
  )
})

test_that("mbl validates fit_method type", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = "pls",
      verbose = FALSE
    ),
    "'fit_method' must be created by a fit_\\*\\(\\) constructor"
  )
})

test_that("mbl validates k does not exceed nrow(Xr)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(100),  # exceeds nrow(Xr) = 40
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "k.*cannot exceed nrow\\(Xr\\)"
  )
})

test_that("mbl validates k_max in neighbors_diss does not exceed nrow(Xr)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_diss(threshold = 0.5, k_min = 10, k_max = 100),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "k_max.*cannot exceed nrow\\(Xr\\)"
  )
})

test_that("mbl messages when k_max is infinite in neighbors_diss", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_message(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_diss(threshold = 0.5, k_min = 10, k_max = Inf),
      fit_method = fit_pls(ncomp = 5),
      control = mbl_control(validation_type = "NNv", allow_parallel = FALSE),
      verbose = FALSE
    ),
    "setting 'k_max'"
  )
})

test_that("mbl warns when diss_method ncomp exceeds data dimensions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_warning(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      diss_method = diss_pca(ncomp = 1000),  # exceeds dimensions
      fit_method = fit_pls(ncomp = 5),
      control = mbl_control(validation_type = "NNv", allow_parallel = FALSE),
      verbose = FALSE
    ),
    "Requested ncomp.*exceeds max allowed"
  )
})

test_that("mbl stops when fit_pls ncomp exceeds data dimensions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 26),  # exceeds dimensions
      control = mbl_control(validation_type = "NNv"),
      verbose = FALSE
    ),
    "More PLS components than observations"
  )
})

test_that("mbl stops when fit_wapls max_ncomp exceeds data dimensions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 26),  # exceeds dimensions
      control = mbl_control(validation_type = "NNv", allow_parallel = FALSE),
      verbose = FALSE
    ),
    "More PLS components than observations"
  )
})

# =============================================================================
# Precomputed dissimilarity matrix validation
# =============================================================================

test_that("mbl validates precomputed matrix for diss_usage = 'predictors'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  n_xr <- nrow(d$Xr)
  n_xu <- nrow(d$Xu)
  
  # Wrong dimensions
  bad_diss <- matrix(0, n_xr, n_xu)
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      diss_method = bad_diss,
      diss_usage = "predictors",
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "When diss_usage = 'predictors'.*square"
  )
})

test_that("mbl validates precomputed matrix for diss_usage = 'weights'", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  n_xr <- nrow(d$Xr)
  n_xu <- nrow(d$Xu)
  n_total <- n_xr + n_xu
  
  # Wrong dimensions (square instead of n_xr x n_xu)
  bad_diss <- matrix(0, n_total, n_total)
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      diss_method = bad_diss,
      diss_usage = "weights",
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "When diss_usage = 'weights' or 'none'.*must have nrow\\(Xr\\).*rows"
  )
})

# =============================================================================
# Local CV validation tests
# =============================================================================

test_that("mbl errors when local CV leaves fewer than 3 observations", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  # Very small k with high p leaves too few for hold-out
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(5),  # very small
      fit_method = fit_pls(ncomp = 2),
      control = mbl_control(
        validation_type = "local_cv",
        p = 0.95,  # leaves only ~5% for hold-out
        allow_parallel = FALSE
      ),
      verbose = FALSE
    ),
    "Local cross-validation requires at least 3 observations"
  )
})

test_that("mbl errors when PLS components exceed neighborhood size", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(8),  # small neighborhood
      fit_method = fit_pls(ncomp = 10),  # more components than neighbors
      control = mbl_control(
        validation_type = "local_cv",
        p = 0.5,
        allow_parallel = FALSE
      ),
      verbose = FALSE
    ),
    "More PLS components than observations"
  )
})

# =============================================================================
# return_dissimilarity tests
# =============================================================================

test_that("mbl with return_dissimilarity = TRUE returns dissimilarity", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    return_dissimilarity = TRUE,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
    neighbors = neighbors_k(25),
    diss_method = diss_pca(ncomp = 10),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_true("diss_xr_xu" %in% names(result$dissimilarities))
  expect_true(is.matrix(result$dissimilarities$diss_xr_xu$dissimilarity))
})

test_that("mbl with return_dissimilarity and projection returns projection", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "NNv",
    return_dissimilarity = TRUE,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
    neighbors = neighbors_k(25),
    diss_method = diss_pca(ncomp = 10, return_projection = TRUE),
    diss_usage = "predictors",
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  expect_true("projection" %in% names(result$dissimilarities))
})

# =============================================================================
# get_predictions tests
# =============================================================================

test_that("get_predictions works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
    neighbors = neighbors_k(c(25, 30)),
    fit_method = fit_pls(ncomp = 5),
    control = mbl_control(validation_type = "NNv", allow_parallel = FALSE),
    verbose = FALSE
  )
  
  preds <- get_predictions(result)
  
  expect_true(is.data.frame(preds))
  expect_equal(nrow(preds), nrow(d$Xu))
  expect_equal(ncol(preds), 2)  # Two k values
})

test_that("get_predictions rejects non-mbl objects", {
  expect_error(
    get_predictions(list(a = 1)),
    "the object is not of class 'mbl'"
  )
})

# =============================================================================
# plot.mbl local_cv validation tests
# =============================================================================

test_that("plot.mbl with local_cv validation works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = "local_cv",
    number = 4,
    p = 0.7,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(25),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  .with_test_device({
    expect_invisible(
      plot(result, what = "validation", metric = "rmse")
    )
  })
})

test_that("plot.mbl with both validation types works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  ctrl <- mbl_control(
    validation_type = c("NNv", "local_cv"),
    number = 4,
    p = 0.7,
    allow_parallel = FALSE
  )
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr, Xu = d$Xu, Yu = d$Yu,
    neighbors = neighbors_k(25),
    fit_method = fit_pls(ncomp = 5),
    control = ctrl,
    verbose = FALSE
  )
  
  .with_test_device({
    expect_invisible(
      plot(result, what = "validation", metric = "st_rmse")
    )
  })
})

# =============================================================================
# Column names validation tests
# =============================================================================

test_that("mbl requires Xr to have column names", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  Xr_no_names <- d$Xr
  colnames(Xr_no_names) <- NULL
  
  expect_error(
    mbl(
      Xr = Xr_no_names, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Xr must have column names"
  )
})

test_that("mbl requires Xu to have column names", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  Xu_no_names <- d$Xu
  colnames(Xu_no_names) <- NULL
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = Xu_no_names,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Xu must have column names"
  )
})

test_that("mbl requires identical column names in Xr and Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  Xu_diff_names <- d$Xu
  colnames(Xu_diff_names) <- paste0("different_", seq_len(ncol(Xu_diff_names)))
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = Xu_diff_names,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      verbose = FALSE
    ),
    "Column names of Xr and Xu must be identical"
  )
})

# =============================================================================
# group validation tests
# =============================================================================

test_that("mbl validates group length", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu,
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      group = rep(1, 10),  # wrong length
      verbose = FALSE
    ),
    "length\\(group\\) must equal nrow\\(Xr\\)"
  )
})

# =============================================================================
# NNv minimum observations test
# =============================================================================

test_that("mbl with NNv requires at least 3 observations in Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_mbl_data()
  
  expect_error(
    mbl(
      Xr = d$Xr, Yr = d$Yr, Xu = d$Xu[1:2, ],
      neighbors = neighbors_k(25),
      fit_method = fit_pls(ncomp = 5),
      control = mbl_control(validation_type = "NNv", allow_parallel = FALSE),
      verbose = FALSE
    ),
    "Nearest neighbor validation requires at least 3 observations"
  )
})

#' @title iterates over elements to print
#' @description internal. used for printing
#' @keywords internal
#' @noRd
cat_iter <- function(x) {
  iter_x <- iter(x, by = "cell", recycle = TRUE)
  
  next_e <- function() {
    nextElem(iter_x)
  }
  next_e
  class(next_e) <- c("isubset", "abstractiter", "iter")
  next_e
}

#' @title fits pls models at each \emph{gesearch} iteration
#' @description internal
#' @keywords internal
#' @noRd
biter <- function(
    itersubs,
    Xu,
    Yu,
    y_lim_left,
    y_lim_right,
    iter_sequence,
    n,
    optimization,
    ncomp,
    tune,
    p,
    number,
    scale,
    max_iter,
    tol,
    seed = NULL,
    algorithm, 
    allow_parallel, 
    pchunks = 1
) {
  
  "%mydo%" <- get("%do%")
  if (allow_parallel & getDoParRegistered()) {
    "%mydo%" <- get("%dopar%")
  }
  
  if (is.null(Yu)) {
    n_yu <- 1   
  } else {
    n_yu <- ncol(Yu) 
  }
  
  soptions <- c(
    "rmse_response", 
    "rmse_reconstruction", 
    "score_dissimilarity",
    "deviation_from_boundaries",
    "residual_variance"
  )
  noptions <- length(soptions)
  
  results_temp <- matrix(NA, pchunks, n + (n_yu * noptions))
  
  if (!is.null(colnames(Yu))) {
    nms <- rep(colnames(Yu), each = noptions)
  } else {
    nms <- rep(1:n_yu, each = noptions)
  }
  
  nms <- paste0(
    soptions, 
    "_", 
    nms)
  
  colnames(results_temp) <- c(nms, paste0("sample_", 1:n))
  #   sample = its$idx
  its <- NULL
  rdf <- foreach(
    idx = iter_sequence,
    its = itersubs, 
    .inorder = FALSE,
    .noexport = c("Xr")
  ) %mydo% {
    
    ith_results <- results_temp[1:ncol(its$idx), , drop = FALSE]
    for (h in 1:ncol(its$idx)) {
      hsub_subset <- its$unique_idx %in% its$idx[, h]
      
      # #
      # # Step 2 - sample a training data set of size k from K without replacement
      # #
      # selected.idx <- sample(k_idx,
      #                        size = k,
      #                        replace = FALSE)
      #
      # # retrieve the spectra and dependent variable from the SL for the training
      # set selection
      # X <- Xr[selected.idx,]
      # Y <- Yr[selected.idx]
      
      # dataset <- cbind(X, Y)
      #
      # Step 3 calibrate a PLS model using the selected SL samples
      #
      for (i in  1:ncol(its$y)) {
        if (tune) {
          # perform a leave-group-out cross.validation of the selected SL (k)
          # observations to determine a suitable number of pls factors
          plsv <- pls_cv(
            x = its$x[hsub_subset, , drop = FALSE],
            y = its$y[hsub_subset, i, drop = FALSE],
            ncomp = ncomp,
            method = "pls",
            center = TRUE,
            scale = scale,
            min_component = 1,
            p = p,
            number = number,
            group = its$group,
            retrieve = FALSE,
            tune = TRUE,
            max_iter = max_iter,
            tol = tol, 
            seed = seed, 
            algorithm = algorithm
          )
          
          selected_pls_c <- which.min(plsv$cv_results$rmse_cv)[1]
        } else {
          selected_pls_c <- ncomp
        }
        gs_pls <- opls_gesearch(
          Xr = its$x[hsub_subset, , drop = FALSE],
          Yr = its$y[hsub_subset, i, drop = FALSE],     
          Xu = Xu, 
          ncomp = selected_pls_c,
          scale = scale,     
          response = "response" %in% optimization | "range" %in% optimization,
          reconstruction = "reconstruction" %in% optimization,
          similarity = "similarity" %in% optimization,
          fresponse = "fresponse" %in% optimization,
          algorithm =  algorithm
        )
        
        deviation_from_boundaries <- residual_variance <- rmse_reconstruction <- rmse_response <- score_dissimilarity <- NULL
        if ("similarity" %in% optimization) {
          score_dissimilarity <- gs_pls$score_dissimilarity[[1]]
        } else {
          score_dissimilarity <- NA
        }
        
        if ("response" %in% optimization) {
          # It is possible we get NA predictions, catch this case and ignore this
          # observation test
          y_pred <- gs_pls$pred_response[, 1]
          if (any(is.na(y_pred))) {
            
          } else {
            rmse_response <- sqrt(mean((y_pred - Yu[, i])^2, na.rm = TRUE))
          }
        } else {
          rmse_response <- NA
        }
        
        if ("reconstruction" %in% optimization) {
          rmse_reconstruction <- gs_pls$rmse_reconstruction[[1]]
        } else {
          rmse_reconstruction <- NA
        }
        
        if ("range" %in% optimization) {
          out_g_left <- gs_pls$pred_response[which(gs_pls$pred_response < y_lim_left), ]
          out_g_right <- gs_pls$pred_response[which(gs_pls$pred_response > y_lim_right), ]
          ws_left <- sum(y_lim_left - out_g_left) / length(gs_pls$pred_response)
          ws_right <- sum(out_g_right - y_lim_right) / length(gs_pls$pred_response)
          deviation_from_boundaries <- ws_out <- ws_left + ws_right
          
        } else {
          deviation_from_boundaries <- NA
        }
        
        if ("fresponse" %in% optimization) {
          residual_variance <- gs_pls$residual_variance[[1]]
        } else {
          residual_variance <- NA
        }
        
        
        # combine the validation results and the observations that were selected for this
        # model
        ith_results[h, (i * noptions) - 4] <- rmse_response
        ith_results[h, (i * noptions) - 3] <- rmse_reconstruction
        ith_results[h, (i * noptions) - 2] <- score_dissimilarity
        ith_results[h, (i * noptions) - 1] <- deviation_from_boundaries
        ith_results[h, (i * noptions)] <- residual_variance
      }
      ith_results[h, -c(1:(n_yu * noptions))] <- its$unique_idx[hsub_subset]
    }
    ith_results
  }
  
  # compile a data frame of all results of the sampling iterations
  rdf <- do.call("rbind", rdf)
  
  obj_cols <- NULL
  if ("response" %in% optimization) {
    obj_cols <- grep("rmse_response", colnames(rdf))
  } 
  
  if ("reconstruction" %in% optimization) {
    obj_cols <- sort(c(obj_cols, grep("rmse_reconstruction", colnames(rdf))))
  }
  
  if ("similarity" %in% optimization) {
    obj_cols <- sort(c(obj_cols, grep("score_dissimilarity", colnames(rdf))))
  }
  
  if ("range" %in% optimization) {
    obj_cols <- sort(c(obj_cols, grep("deviation_from_boundaries", colnames(rdf))))
  }
  
  
  if ("fresponse" %in% optimization) {
    obj_cols <- sort(c(obj_cols, grep("residual_variance", colnames(rdf))))
  }
  
  weakness_scores_subset <- rdf[, obj_cols, drop = FALSE]
  rdf <- rdf[, -grep("^rmse_|^score_|^deviation_|^residual_", colnames(rdf)), drop = FALSE]
  return(list(
    weakness_scores_subset = weakness_scores_subset,
    sampleidx = rdf
  ))
}

#' @title iterator that subsets a matrix based on input indices for pls modeling
#' at each \emph{gesearch} iteration
#' @description internal
#' @param x an input matrix of predictors
#' @param y a response variable
#' @param group a variable giving the groups/calsses to which the observations
#' belong to (used for avoiding pseudo-replication during validation)
#' @param indx the indices to retrieve
#' @param chunksize the number of elements of \code{indx} to return with each 
#' iteration.
#' @return an object of \code{class} iterator giving a \code{list} with (a) a
#' subset of the input matrix with rows re-arranged according to `indx`
#' @details this is designed to be used within (parallel) loops to avoid sending
#' entire matrices to each core. It just sends the subset of that is required.
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
ithrssubsets <- function(
    x,
    y,
    group = NULL,
    indx, 
    chunksize = 1L
) {
  it_indx <- iter(indx, by = "column", chunksize = chunksize)
  
  if (is.null(group)) {
    this_next_element <- function() {
      ss <- nextElem(it_indx)
      unique_ss <- unique(as.vector(ss))
      list(
        x = x[unique_ss, , drop = FALSE],
        y = y[unique_ss, , drop = FALSE],
        group = NULL,
        idx = ss, 
        unique_idx = unique_ss
      )
    }
  } else {
    nextEl <- function() {
      ss <- nextElem(it_indx)
      unique_ss <- unique(as.vector(ss))
      list(
        x = x[unique_ss, , drop = FALSE],
        y = y[unique_ss, , drop = FALSE],
        group = factor(group[unique_ss]),
        idx = ss, 
        unique_idx = unique_ss
      )
    }
  }
  obj <- list(nextElem = this_next_element)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}

#' @title drop the samples that consistently appear in high error models
#' @description internal
#' @keywords internal
#' @noRd
drop_indices <- function(
    uu, 
    vv,
    r,
    obs_indices,
    resampling_indices, 
    weakness_subset,
    retain_by, 
    cull_quantity, 
    percentile_type, 
    max_ncomp
) {
  
  # vv <- as.vector(table(resampling_indices))
  # uu <- vv <- rep(0, n)
  
  for (i in 1:nrow(resampling_indices)) {
    uu[resampling_indices[i, ]] <- uu[resampling_indices[i, ]] + weakness_subset[i]
    vv[resampling_indices[i, ]] <- vv[resampling_indices[i, ]] + 1
  }
  
  vv[vv == 0] <- NA
  # normalise U using V
  uu <- uu / vv
  
  # uu <- sapply(obs_indices, 
  #              FUN = function(weakness, indices, ii) {
  #                mean(weakness[arrayInd(which(indices == ii), .dim = nrow(indices))])
  #              },
  #              weakness = weakness_subset, 
  #              indices = resampling_indices)
  # uu[is.nan(uu)] <- NA
  
  u_df <- data.frame(
    idx = obs_indices,
    weakness_score = uu
  )
  
  # now order / rank by weakness
  u_ordered <- u_df[order(u_df$weakness_score, decreasing = TRUE), ]
  u_ordered <- u_ordered[!is.na(u_ordered$weakness_score), ]
  
  uws <- length(unique(u_ordered$weakness_score))
  if (uws == 1) {
    results <- list(interrupted = FALSE,
                    s_to_drop = 0,
                    cutoff = 0,
                    max_weakness_score = 0,
                    obs_indices = obs_indices)
    return(results)
  } else {
    if (retain_by == "proportion") {
      # select the poorest performing SL samples, the amount is determined by
      # cull_quantity calculated earlier
      if (uws < cull_quantity) {
        cull_quantity <- uws
      }
      worst_reference_set <- u_ordered[1:cull_quantity, ]
      ok_reference_set <- u_ordered[-c(1:cull_quantity), ]
      cutoff <- min(worst_reference_set$weakness_score)
      pp <- 1 - nrow(ok_reference_set) / nrow(u_ordered)
    } else {
      ## Instead removing a given number of samples
      ## remove samples that are above a given cutoff prob
      cutoff <- quantile(uu, 1 - r, na.rm = TRUE, type = percentile_type)
      worst_reference_set <- u_ordered[u_ordered$weakness_score >= cutoff, ]
      ok_reference_set <- u_ordered[u_ordered$weakness_score < cutoff, ]
      if (nrow(ok_reference_set) < max_ncomp) {
        mss <- (paste0(
          "Iteration interrupted as the number of selected observations (",
          nrow(ok_reference_set),
          ") is lower than the number of pls factors (", max_ncomp, ")"
        ))
        return(list(interrupted = TRUE,
                    message = mss))
      }
      pp <- 1 - nrow(ok_reference_set) / nrow(u_ordered)
    }
    obs_indices[worst_reference_set$idx] <- NA
    results <- list(interrupted = FALSE,
                    s_to_drop = nrow(worst_reference_set),
                    cutoff = cutoff,
                    max_weakness_score = max(ok_reference_set$weakness_score),
                    obs_indices = obs_indices)
    return(results)
  }
}

#' @title combine two randomly selected individuals
#' @description internal
#' @keywords internal
#' @noRd
combine_individuals <- function(x, k, k_idx) {
  # Select two individuals out of x (i.e. columns), randomly (without replacements)
  sel_individuals <- sample(1:ncol(x), 2, replace = FALSE)
  # Combine the two samples and remove duplicated entries and NAs
  combined_individuals <- unique(na.omit(c(x[, sel_individuals])))
  # If the combined individual has more than k samples, we draw k samples from 
  # the combined gene pool
  # Otherwise, do mutation: randomly selected genes from the global gene pool 
  # are used to fill up the number of genes to k
  if (length(combined_individuals) < k) {
    mutation_gene_pool <- setdiff(k_idx, combined_individuals)
    c(
      combined_individuals, 
      sample(
        mutation_gene_pool, 
        k - length(combined_individuals), 
        replace = FALSE
      )
    )
  } else {
    sample(combined_individuals, k, replace = FALSE)
  }
}

#' @title iterator for nearest neighbor subsets
#' @param x a reference matrix
#' @param xu a second matrix
#' @param y a matrix of side information
#' @param kindx a matrix of nearest neighbor indices
#' @param na_rm logical indicating whether NAs must be removed.
#' @description internal
#' @keywords internal
#' @noRd
ith_subsets_ortho_diss <- function(x,
                                   xu = NULL,
                                   y,
                                   kindx,
                                   na_rm = FALSE) {
  it_kindx <- iter(kindx, by = "column")
  it_xu <- iter(xu, by = "row")

  nextEl <- function() {
    knns <- nextElem(it_kindx)
    if (na_rm) {
      knns <- knns[!is.na(knns)]
    }
    lsubst <- list(x = x[knns, ])
    if (!is.null(y)) {
      lsubst$y <- y[knns, , drop = FALSE]
    }
    if (!is.null(xu)) {
      lsubst$xu <- nextElem(it_xu)
    }
    return(lsubst)
  }
  obj <- list(nextElem = nextEl)
  class(obj) <- c("isubset", "abstractiter", "iter")
  obj
}

# #' @title local ortho dissimilarity matrices initialized by a global
# #' dissimilarity matrix
# #' @param k_index_matrix a matrix of nearest neighnbor indices
# #' @param Xr argument passed to ortho_projection
# #' @param Yr argument passed to ortho_projection
# #' @param Xu argument passed to ortho_projection
# #' @param diss_method argument passed to ortho_projection
# #' @param pc_selection argument passed to ortho_projection
# #' @param center argument passed to ortho_projection
# #' @param scale argument passed to ortho_projection
# #' @description internal
# #' @keywords internal
# #' @noRd
# local_ortho_diss <- function(k_index_matrix, Xr, Yr, Xu,
#                              diss_method, pc_selection, center, scale,
#                              allow_parallel, ...) {
#   "%mydo%" <- get("%do%")
#   if (allow_parallel & getDoParRegistered()) {
#     "%mydo%" <- get("%dopar%")
#   }
# 
#   if (!is.null(Xu)) {
#     n_xu <- nrow(Xu)
#   } else {
#     n_xu <- 0
#   }
# 
#   n_xr <- nrow(Xr)
# 
#   if (diss_method %in% c("pca", "pca.nipals")) {
#     use_distance_method <- "euclid"
#   } else {
#     use_distance_method <- "mahalanobis"
#   }
# 
#   if (n_xr + n_xu == ncol(k_index_matrix)) {
#     ## squared matrix for all possible pair-wise dissimilarities
#     distnc <- matrix(NA, ncol(k_index_matrix), ncol(k_index_matrix))
#     diag(distnc) <- 0
#     ith_subsets <- ith_subsets_ortho_diss(
#       x = Xr, y = Yr,
#       kindx = k_index_matrix
#     )
#     n_rows_d <- ncol(k_index_matrix)
#     # neighbors (excluding target sample )
#     n_neighbors <- nrow(k_index_matrix) - 1
#     has_target_indices <- TRUE
#   } else {
#     ## matrix for dissimilarities between Xr and Xu
#     distnc <- matrix(NA, n_xr, ncol(k_index_matrix))
#     ith_subsets <- ith_subsets_ortho_diss(
#       x = Xr, xu = Xu, y = Yr,
#       kindx = k_index_matrix,
#       na_rm = TRUE
#     )
#     n_rows_d <- n_xr
#     n_neighbors <- nrow(k_index_matrix)
#     has_target_indices <- FALSE
#   }
# 
# 
#   dist_vec <- rep(NA, n_neighbors + 2)
# 
#   ith_local_subset <- NULL
#   ## for each computations
#   i <- NULL
#   local_d <- foreach(
#     i = 1:ncol(k_index_matrix),
#     ith_local_subset = ith_subsets,
#     .combine = cbind,
#     .inorder = FALSE,
#     .export = c(
#       "ortho_projection",
#       "pc_projection",
#       "pls_projection",
#       "diss_evaluate",
#       "dist_vec"
#     ),
#     .packages = c("resemble")
#   ) %mydo% { # Parallel
#     ith_projection <- ortho_projection(
#       Xr = ith_local_subset$x,
#       Yr = ith_local_subset$y,
#       Xu = ith_local_subset$xu,
#       method = diss_method,
#       pc_selection = pc_selection,
#       center = center,
#       scale = scale,
#       ...
#     )
# 
#     if (diss_method %in% c("pca", "pca.nipals")) {
#       scores <- sweep(ith_projection$scores,
#         MARGIN = 2,
#         STATS = ith_projection$scores_sd,
#         FUN = "/"
#       )
#     } else {
#       scores <- ith_projection$scores
#     }
# 
#     ith_dist <- dist_vec
#     ith_dist[3:length(ith_dist)] <- f_diss(
#       Xr = scores[1, , drop = FALSE],
#       Xu = scores[-1, , drop = FALSE],
#       diss_method = use_distance_method,
#       center = FALSE,
#       scale = FALSE
#     )
#     ith_dist[1] <- i
#     ith_dist[2] <- ith_projection$ncomp
#     ith_dist
#   }
#   ## end of foreach computations
# 
# 
#   local_d <- local_d[-1, order(local_d[1, ])]
#   local_n_components <- local_d[1, ]
#   local_d <- local_d[-1, ]
#   add_row_indices <- ((0:(ncol(k_index_matrix) - 1)) * n_rows_d)
#   # necessary to remove the xi-to-xi from the indices of nearest neighnbors
#   if (has_target_indices) {
#     k_index_matrix <- k_index_matrix[-1, ]
#   }
#   vec_positions <- sweep(k_index_matrix,
#     MARGIN = 2,
#     STATS = add_row_indices, FUN = "+"
#   )
#   distnc[vec_positions] <- as.vector(local_d)
#   return(list(
#     dissimilarity_m = distnc,
#     local_ncomp = data.frame(local_n_components)
#   ))
# }



#' @title format internal messages
#' @param xr_xu_names the names of Xr and Xu
#' @description internal
#' @keywords internal
#' @noRd
format_xr_xu_indices <- function(xr_xu_names) {
  xu_insufficient <- gsub("Xu_", "", xr_xu_names[grep("^Xu_", xr_xu_names)])
  xr_insufficient <- gsub("Xr_", "", xr_xu_names[grep("^Xr_", xr_xu_names)])


  xu_mss <- paste0(c(
    ifelse(length(xu_insufficient) > 0,
      "\nXu: ", ""
    ),
    paste(xu_insufficient, collapse = ", ")
  ),
  collapse = " "
  )
  xr_mss <- paste0(c(
    ifelse(length(xr_insufficient) > 0,
      "\nXr: ", ""
    ),
    paste(xr_insufficient, collapse = ", ")
  ),
  collapse = " "
  )

  return(list(
    xr_mss = xr_mss,
    xu_mss = xu_mss
  ))
}

#' @title Search neighbors in a reference set
#'
#' @description
#' \loadmathjax
#' Searches for the nearest neighbors of observations in a reference set or
#' between two sets of observations.
#' 
#' @usage
#' search_neighbors(Xr, Xu = NULL,
#'                  diss_method = diss_pca(), Yr = NULL,
#'                  neighbors, spike = NULL,
#'                  return_dissimilarity = FALSE, 
#'                  k, k_diss, k_range, pc_selection, 
#'                  center, scale, documentation, ...
#'                  )
#'     
#' @param Xr A numeric matrix of reference observations (rows) and variables
#'   (columns) where the neighbor search is conducted.
#' @param Xu Optional matrix of observations for which neighbors are to be
#'   searched in \code{Xr}.
#' @param diss_method A dissimilarity method object created by one of:
#'   \itemize{
#'     \item \code{\link{diss_pca}()}: Mahalanobis distance in PCA space
#'     \item \code{\link{diss_pls}()}: Mahalanobis distance in PLS space
#'     \item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
#'     \item \code{\link{diss_euclidean}()}: Euclidean distance
#'     \item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
#'     \item \code{\link{diss_cosine}()}: Cosine dissimilarity
#'   }
#'   Default is \code{diss_pca()}.
#' @param Yr Optional response matrix. Required for PLS methods and when using
#'   \code{ncomp_by_opc()}.
#' @param neighbors A neighbor selection object created by:
#'   \itemize{
#'     \item \code{\link{neighbors_k}()}: Select k nearest neighbors
#'     \item \code{\link{neighbors_diss}()}: Select neighbors by dissimilarity threshold
#'   }
#' @param spike Optional integer vector indicating observations in \code{Xr}
#'   to force into (positive indices) or exclude from (negative indices)
#'   neighborhoods.
#' @param return_dissimilarity Logical indicating whether to return the
#'   dissimilarity matrix. Default is \code{FALSE}.
#'   
#' @param k Deprecated.
#' @param k_diss Deprecated.
#' @param k_range Deprecated.
#' @param pc_selection Deprecated.
#' @param center Deprecated.
#' @param scale Deprecated.
#' @param documentation Deprecated.
#' @param ... Additional arguments (currently unused).
#'
#' @details
#' This function is useful for reducing large reference sets by identifying
#' only relevant neighbors before running \code{\link{mbl}}.
#'
#' If \code{Xu} is not provided, the function searches for neighbors within
#' \code{Xr} itself (excluding self-matches). If \code{Xu} is provided,
#' neighbors of each observation in \code{Xu} are searched in \code{Xr}.
#'
#' The \code{spike} argument allows forcing specific observations into or out
#' of all neighborhoods. Positive indices are always included; negative indices
#' are always excluded.
#'
#' @return A list containing:
#' \describe{
#'   \item{neighbors}{Matrix of \code{Xr} indices for each query observation's
#'     neighbors, sorted by dissimilarity (columns = query observations).}
#'   \item{neighbors_diss}{Matrix of dissimilarity scores corresponding to
#'     \code{neighbors}.}
#'   \item{unique_neighbors}{Vector of unique \code{Xr} indices that appear
#'     in any neighborhood.}
#'   \item{k_diss_info}{If \code{neighbors_diss()} was used, a \code{data.frame} 
#'     with columns for observation index, number of neighbors found, and final
#'     number after applying bounds.}
#'   \item{dissimilarity}{If \code{return_dissimilarity = TRUE}, the full
#'     dissimilarity object.}
#'   \item{projection}{If the dissimilarity method includes
#'     \code{return_projection = TRUE}, the projection object.}
#'   \item{gh}{If the dissimilarity method includes \code{gh = TRUE}, the
#'     GH distances.}
#' }
#'
#' @seealso \code{\link{dissimilarity}}, \code{\link{mbl}}, 
#'   \code{\link{neighbors_k}}, \code{\link{neighbors_diss}}
#'
#' @references
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
#' 268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199, 43-53.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
#' Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
#' Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
#' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#'
#' Xu <- Xu[!is.na(Yu), ]
#' Yu <- Yu[!is.na(Yu)]
#' Xr <- Xr[!is.na(Yr), ]
#' Yr <- Yr[!is.na(Yr)]
#'
#' # Correlation-based neighbor search with k neighbors
#' ex1 <- search_neighbors(
#'   Xr = Xr, Xu = Xu,
#'   diss_method = diss_correlation(),
#'   neighbors = neighbors_k(40)
#' )
#'
#' # PCA-based with OPC selection
#' ex2 <- search_neighbors(
#'   Xr = Xr, Xu = Xu,
#'   diss_method = diss_pca(
#'     ncomp = ncomp_by_opc(40),
#'     scale = TRUE,
#'     return_projection = TRUE
#'   ),
#'   Yr = Yr,
#'   neighbors = neighbors_k(50)
#' )
#'
#' # Observations not in any neighborhood
#' setdiff(seq_len(nrow(Xr)), ex2$unique_neighbors)
#'
#' # Dissimilarity threshold-based selection
#' ex3 <- search_neighbors(
#'   Xr = Xr, Xu = Xu,
#'   diss_method = diss_pls(
#'     ncomp = ncomp_by_opc(40),
#'     scale = TRUE
#'   ),
#'   Yr = Yr,
#'   neighbors = neighbors_diss(threshold = 0.5, k_min = 10, k_max = 100)
#' )
#' }
#'
#' @export
search_neighbors <- function(
    Xr,
    Xu = NULL,
    diss_method = diss_pca(),
    Yr = NULL,
    neighbors,
    spike = NULL,
    return_dissimilarity = FALSE,
    # --- Deprecated arguments ---
    k, k_diss, k_range, pc_selection, center, scale, documentation, ...
) {
  
  
  # ---------------------------------------------------------------------------
  # Block removed arguments
  # ---------------------------------------------------------------------------
  if (!missing(k) || !missing(k_diss) || !missing(k_range)) {
    stop(
      "Arguments 'k', 'k_diss', 'k_range' have been removed.\n",
      "Use neighbors_k() or neighbors_diss() instead.\n",
      "Example: search_neighbors(..., neighbors = neighbors_k(40))",
      call. = FALSE
    )
  }
  
  if (!missing(pc_selection)) {
    stop(
      "Argument 'pc_selection' has been removed.\n",
      "Component selection is now specified in diss_*() constructors.\n",
      "Example: diss_pca(ncomp = ncomp_by_opc(40))",
      call. = FALSE
    )
  }
  
  if (!missing(center) || !missing(scale)) {
    stop(
      "Arguments 'center' and 'scale' have been removed.\n",
      "These are now set in diss_*() constructors.\n",
      "Example: diss_pca(center = TRUE, scale = FALSE)",
      call. = FALSE
    )
  }
  
  if (!missing(documentation)) {
    stop(
      "Argument 'documentation' has been removed.",
      call. = FALSE
    )
  }
  
  if (length(list(...)) > 0) {
    stop(
      "Unknown arguments passed via '...'.\n",
      "Arguments like 'ws', 'return_projection', 'gh' are now set in ",
      "diss_*() constructors.",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Handle legacy character-based diss_method
  # ---------------------------------------------------------------------------
  if (is.character(diss_method)) {
    stop(
      "Character-based 'diss_method' is no longer supported.\n\n",
      "Use method constructors instead:\n\n",
      "  Old API                          -> New API\n",
      "  -------                             -------\n",
      "  diss_method = \"pca\"               -> diss_pca()\n",
      "  diss_method = \"pca.nipals\"        -> diss_pca(method = \"pca_nipals\")\n",
      "  diss_method = \"pls\"               -> diss_pls()\n",
      "  diss_method = \"mpls\"              -> diss_pls(method = \"mpls\")\n",
      "  diss_method = \"euclid\"            -> diss_euclidean()\n",
      "  diss_method = \"cosine\"            -> diss_cosine()\n",
      "  diss_method = \"cor\"               -> diss_correlation()\n",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Validate diss_method
  # ---------------------------------------------------------------------------
  if (!inherits(diss_method, "diss_method")) {
    stop(
      "'diss_method' must be a dissimilarity method object.\n",
      "Use one of: diss_pca(), diss_pls(), diss_euclidean(), ",
      "diss_mahalanobis(), diss_cosine(), diss_correlation().",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Validate neighbors
  # ---------------------------------------------------------------------------
  if (missing(neighbors)) {
    stop(
      "'neighbors' is required. Use neighbors_k() or neighbors_diss().",
      call. = FALSE
    )
  }
  
  if (!inherits(neighbors, "neighbors")) {
    stop(
      "'neighbors' must be created by neighbors_k() or neighbors_diss().",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Convert neighbors to internal format
  # ---------------------------------------------------------------------------
  n_xr <- nrow(Xr)
  
  if (inherits(neighbors, "neighbors_k")) {
    k <- max(neighbors$k)
    k_diss <- NULL
    k_range <- NULL
    
    if (k > n_xr) {
      stop("'k' (", k, ") cannot exceed nrow(Xr) (", n_xr, ").", call. = FALSE)
    }
  } else {
    # neighbors_diss
    k <- NULL
    k_diss <- max(neighbors$threshold)
    k_range <- c(neighbors$k_min, neighbors$k_max)

    if (is.infinite(k_range[2])) {
      
      if (is.null(Xu)) {
        max_neighbors  <- n_xr
        mss <- "nrow(Xr) - 1 (excluding self-matches)"
      } else {
        max_neighbors <- n_xr - 1
        mss <- "nrow(Xr)"
      }
      message(
        "setting 'k_max' (", k_range[2], ") to ", mss,  " (", n_xr, ")."
      )
      k_range[2] <- n_xr
    }
        
    if (k_range[2] > n_xr) {
      stop(
        "'k_max' (", k_range[2], ") cannot exceed nrow(Xr) (", n_xr, ").",
        call. = FALSE
      )
    }
  }
  
  # ---------------------------------------------------------------------------
  # Validate spike
  # ---------------------------------------------------------------------------
  if (!is.null(spike)) {
    if (!is.vector(spike) || !all(spike == as.integer(spike))) {
      stop("'spike' must be an integer vector.", call. = FALSE)
    }
    if (length(spike) >= n_xr) {
      stop("'spike' length must be less than nrow(Xr).", call. = FALSE)
    }
    if (max(abs(spike)) > n_xr) {
      stop("'spike' contains out-of-bounds indices.", call. = FALSE)
    }
    
    spike_hold <- spike[spike > 0L]
    spike_rm <- abs(spike[spike < 0L])
    
    if (length(intersect(spike_hold, spike_rm)) > 0L) {
      stop("'spike' contains contradictory indices.", call. = FALSE)
    }
    
    min_k <- if (!is.null(k)) min(neighbors$k) else neighbors$k_min
    if (min_k <= length(spike_hold)) {
      stop(
        "Minimum k must exceed length of positive spike indices.",
        call. = FALSE
      )
    }
    
    spike <- sort(unique(as.integer(spike)))
  }
  
  # ---------------------------------------------------------------------------
  # Compute dissimilarity
  # ---------------------------------------------------------------------------
  dsm <- dissimilarity(
    Xr = Xr,
    Xu = Xu,
    diss_method = diss_method,
    Yr = Yr
  )
  
  # ---------------------------------------------------------------------------
  # Find neighbors
  # ---------------------------------------------------------------------------
  skip_first <- is.null(Xu)
  
  results <- diss_to_neighbors(
    dsm$dissimilarity,
    k = k,
    k_diss = k_diss,
    k_range = k_range,
    spike = spike,
    return_dissimilarity = return_dissimilarity,
    skip_first = skip_first
  )
  
  # Column names
  mprefix <- if (skip_first) "Xr" else "Xu"
  colnames(results$neighbors) <- paste0(mprefix, seq_len(ncol(results$neighbors)))
  
  # ---------------------------------------------------------------------------
  # Add optional outputs from dissimilarity
  # ---------------------------------------------------------------------------
  if (return_dissimilarity) {
    results$dissimilarity <- dsm
  }
  
  if (!is.null(dsm$projection)) {
    results$projection <- dsm$projection
  }
  
  if (!is.null(dsm$gh)) {
    results$gh <- dsm$gh
  }
  
  results
}

#' @title Euclidean, Mahalanobis and cosine dissimilarity measurements
#' @description
#' \loadmathjax
#' This function is used to compute the dissimilarity between observations
#' based on Euclidean or Mahalanobis distance measures or on cosine
#' dissimilarity measures (a.k.a spectral angle mapper).
#' @usage
#' f_diss(Xr, Xu = NULL, diss_method = "euclid",
#'        center = TRUE, scale = FALSE)
#' @param Xr a matrix containing the (reference) data.
#' @param Xu an optional matrix containing data of a second set of observations
#' (samples).
#' @param diss_method the method for computing the dissimilarity between
#' observations.
#' Options are \code{"euclid"} (Euclidean distance), \code{"mahalanobis"}
#' (Mahalanobis distance) and \code{"cosine"} (cosine distance, a.k.a spectral
#' angle mapper). See details.
#' @param center a logical indicating if the spectral data \code{Xr} (and
#' \code{Xu} if specified) must be centered. If \code{Xu} is provided, the data
#' is scaled on the basis of \mjeqn{Xr \cup Xu}{Xr U Xu}.
#' @param scale a logical indicating if \code{Xr} (and \code{Xu} if specified)
#' must be scaled. If \code{Xu} is provided the data is scaled on the basis
#' of \mjeqn{Xr \cup Xu}{Xr U Xu}.
#' @details
#' The results obtained for Euclidean dissimilarity are equivalent to those
#' returned by the [stats::dist()] function, but are scaled
#' differently. However, \code{f_diss} is considerably faster (which can be
#' advantageous when computing dissimilarities for very large matrices). The
#' final scaling of the dissimilarity scores in \code{f_diss} where
#' the number of variables is used to scale the squared dissimilarity scores. See
#' the examples section for a comparison between [stats::dist()] and
#' \code{f_diss}.
#'
#' In the case of both the Euclidean and Mahalanobis distances, the scaled
#' dissimilarity matrix \mjeqn{D}{D} between between observations in a given
#' matrix \mjeqn{X}{X} is computed as follows:
#'
#' \mjdeqn{d(x_i, x_j)^{2} = \sum (x_i - x_j)M^{-1}(x_i - x_j)^{\mathrm{T}}}{d(x_i, x_j)^{2} = \sum (x_i - x_j)M^{-1}(x_i - x_j)^T}
#' \mjdeqn{d_{scaled}(x_i, x_j) = \sqrt{\frac{1}{p}d(x_i, x_j)^{2}}}{d_scaled (x_i, x_j) = sqrt(1/p d(x_i, x_j)^2)}
#'
#' where \mjeqn{p}{p} is the number of variables in \mjeqn{X}{X}, \mjeqn{M}{M} is the identity
#' matrix in the case of the Euclidean distance and the variance-covariance
#' matrix of \mjeqn{X}{X} in the case of the Mahalanobis distance. The Mahalanobis
#' distance can also be viewed as the Euclidean distance after applying a
#' linear transformation of the original variables. Such a linear transformation
#' is done by using a factorization of the inverse covariance matrix as
#' \mjeqn{M^{-1} = W^{T}W}{M^-1 = W^TW}, where \mjeqn{M}{M} is merely the square root of
#' \mjeqn{M^{-1}}{M^{-1}} which can be found by using a singular value decomposition.
#'
#' Note that when attempting to compute the Mahalanobis distance on a dataset
#' with highly correlated variables (i.e. spectral variables) the
#' variance-covariance matrix may result in a singular matrix which cannot be
#' inverted and therefore the distance cannot be computed.
#' This is also the case when the number of observations in the dataset is
#' smaller than the number of variables.
#'
#' For the computation of the Mahalanobis distance, the mentioned method is
#' used.
#'
#' The cosine dissimilarity \mjeqn{c}{c} between two observations
#' \mjeqn{x_i}{x_i} and \mjeqn{x_j}{x_j} is computed as follows:
#'
#' \mjdeqn{c(x_i, x_j) = cos^{-1}{\frac{\sum_{k=1}^{p}x_{i,k} x_{j,k}}{\sqrt{\sum_{k=1}^{p} x_{i,k}^{2}} \sqrt{\sum_{k=1}^{p} x_{j,k}^{2}}}}}{c(x_i, x_j) = cos^{-1} ((sum_(k=1)^p x_(i,k) x_(j,k))/(sum_(k=1)^p x_(i,k) sum_(k=1)^p x_(j,k)))}
#'
#' where \mjeqn{p}{p} is the number of variables of the observations.
#' The function does not accept input data containing missing values.
#' NOTE: The computed distances are divided by the number of variables/columns
#' in \code{Xr}.
#'
#' @return
#' a matrix of the computed dissimilarities.
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} and Antoine Stevens
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
#' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#'
#' # Euclidean distances between all the observations in Xr
#'
#' ed <- f_diss(Xr = Xr, diss_method = "euclid")
#'
#' # Equivalence with the dist() function of R base
#' ed_dist <- (as.matrix(dist(Xr))^2 / ncol(Xr))^0.5
#' round(ed_dist - ed, 5)
#'
#' # Comparing the computational time
#' iter <- 20
#' tm <- proc.time()
#' for (i in 1:iter) {
#'   f_diss(Xr)
#' }
#' f_diss_time <- proc.time() - tm
#'
#' tm_2 <- proc.time()
#' for (i in 1:iter) {
#'   dist(Xr)
#' }
#' dist_time <- proc.time() - tm_2
#'
#' f_diss_time
#' dist_time
#'
#' # Euclidean distances between observations in Xr and observations in Xu
#' ed_xr_xu <- f_diss(Xr, Xu)
#'
#' # Mahalanobis distance computed on the first 20 spectral variables
#' md_xr_xu <- f_diss(Xr[, 1:20], Xu[, 1:20], "mahalanobis")
#'
#' # Cosine dissimilarity matrix
#' cdiss_xr_xu <- f_diss(Xr, Xu, "cosine")
#' }
#' @noRd
#' @keywords internal

######################################################################
# resemble
# Copyright (C) 2014 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################

## History:
## 09.03.2014 Leo     The line rslt[is.na(rslt)] <- 0 was added in order
##                    to deal with NaNs produced by the C++ code
## 18.11.2015 Leo     Bug fixed. Code crashed When Xr was of one row
## 18.07.2019 Leo     A bug in the scaling of the euclidean distances in f_diss was detected and fixed. The distance ratios
##                    (between observations) were correctly calculated, but the final scaling of the results was not properly
##                    done. The distance between Xi and Xj were scaled by taking the squared root of the mean of the squared
##                    differences and dividing it by the number of variables i.e. sqrt(mean((Xi-Xj)^2))/ncol(Xi), however the
##                    correct calculation is done by taking the mean of the squared differences, dividing it by the number of
##                    variables and then compute the squared root i.e. sqrt(mean((Xi-Xj)^2)/ncol(Xi)).
##                    This bug had no effect on the computations of the nearest neighbors.
## 21.04.2020 Leo     styler applied and Argument scaled renamed to scale
##                    the dimnames of the resulting matrix are now Xr_1... Xr_n (previusly Xr.1... Xr.n)
## 03.07.2020 Leo     FIXME: with cosine  diss between the same observation in some
##                    cases NaN are returned and also values around 1e-08

f_diss <- function(Xr, Xu = NULL, diss_method = "euclid",
                   center = TRUE, scale = FALSE) {
  if (!is.null(Xu)) {
    if (ncol(Xu) != ncol(Xr)) {
      stop("The number of columns (variables) in Xr must be equal to \n the number of columns (variables) in Xu")
    }
    if (sum(is.na(Xu)) > 0) {
      stop("Input data contains missing values")
    }
  }
  if (sum(is.na(Xr)) > 0) {
    stop("Matrices with missing values are not accepted")
  }

  n_method <- diss_method
  if (!n_method %in% c("euclid", "mahalanobis", "cosine")) {
    stop("'diss_method' must be one of: 'euclid', 'mahalanobis' or'cosine'")

    if (length(n_method) > 1) {
    } else {
      n_method <- diss_method[[1]]
      message(paste("More than one diss_method was specified, only", n_method, "was used."))
    }
  }

  if (!is.logical(center)) {
    stop("'center' must be logical")
  }

  if (!is.logical(scale)) {
    stop("'scale' must be logical")
  }

  if (center | scale | n_method %in% c("mahalanobis", "euclid")) {
    X <- rbind(Xr, Xu)

    if (center) {
      X <- sweep(x = X, MARGIN = 2, FUN = "-", STATS = colMeans(X))
    }

    if (scale) {
      X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = get_col_sds(X))
    }

    if (n_method == "mahalanobis") {
      if (nrow(X) < ncol(X)) {
        stop("For computing the Mahalanobis distance, the total number of observations (rows) \n must be larger than the number of variables (columns).")
      }
      X <- try(euclid_to_mahal(X, sm_method = "svd"), TRUE)
      if (!is.matrix(X)) {
        stop("The covariance matrix (for the computation of the Mahalanobis distance) is exactly singular. \n Try another method.")
      }
      n_method <- "euclid"
    }

    if (!is.null(Xu)) {
      Xu <- X[(nrow(X) - nrow(Xu) + 1):nrow(X), , drop = FALSE]
      Xr <- X[1:(nrow(X) - nrow(Xu)), , drop = FALSE]
    } else {
      Xr <- X
    }
    rm(X)
  }

  if (!is.null(Xu)) {
    ## FIXME check numerical precision in Rcpp
    ## in some cases it returns 0s as -1e-14
    ## perhaps due to reuse memory?
    rslt <- abs(fast_diss(Xu, Xr, n_method))
    if (n_method == "euclid") {
      rslt <- sqrt(rslt / ncol(Xr))
    }
    colnames(rslt) <- paste("Xu", 1:nrow(Xu), sep = "_")
    rownames(rslt) <- paste("Xr", 1:nrow(Xr), sep = "_")
  } else {
    ## FIXME check numerical precision in Rcpp
    ## in some cases it returns 0s as -1e-14
    ## perhaps due to reuse memory?
    rslt <- abs(fast_diss(Xr, Xr, n_method))
    if (n_method == "euclid") {
      rslt <- sqrt(rslt / ncol(Xr))
    }
    rownames(rslt) <- paste("Xr", 1:nrow(Xr), sep = "_")
    colnames(rslt) <- rownames(rslt)
  }
  if (diss_method == "cosine") {
    rslt[is.nan(rslt)] <- 0
  }

  rslt
}


#' @title A function for transforming a matrix from its Euclidean space to
#' its Mahalanobis space
#' @description For internal use only
#' @keywords internal
#' @importFrom stats cov
#' @noRd
euclid_to_mahal <- function(X, sm_method = c("svd", "eigen")) {
  nms <- dimnames(X)
  if (ncol(X) > nrow(X)) {
    stop("In order to project the matrix to a Mahalanobis space, the number of observations of the input matrix must larger than its number of variables")
  }

  if (length(sm_method) > 1) {
    sm_method <- sm_method[1]
  }
  if (!(sm_method %in% c("svd", "eigen"))) {
    stop("sm_method must be one of 'svd', 'eigen'")
  }

  X <- as.matrix(X)
  vcv <- cov(X)
  sq_vcv <- sqrt_sm(vcv, method = sm_method)
  sq_S <- solve(sq_vcv)
  ms_x <- X %*% sq_S
  dimnames(ms_x) <- nms

  ms_x
}


#' @title Square root of (square) symmetric matrices
#' @description For internal use only
#' @keywords internal
#' @noRd
sqrt_sm <- function(X, method = c("svd", "eigen")) {
  if (!isSymmetric(X)) {
    stop("X must be a square symmetric matrix")
  }
  if (length(method) > 1) {
    method <- method[1]
  }
  if (!(method %in% c("svd", "eigen"))) {
    stop("method must be one of 'svd', 'eigen'")
  }

  if (method == "svd") {
    ## REPLACE BY  arma::svd(U, S, V, X, "dc")
    out <- svd(X)
    D <- diag(out$d)
    U <- out$v
    return(U %*% (D^0.5) %*% t(U))
  }

  if (method == "eigen") {
    out <- eigen(X)
    D <- diag(out$values)
    U <- out$vectors
    return(U %*% (D^0.5) %*% t(U))
  }
}

#' @title Component selection methods
#'
#' @description
#' Constructor functions for specifying how to select the number of components
#' in projection-based dissimilarity methods ([diss_pca()], [diss_pls()]).
#'
#' @param min_var Numeric in (0, 1]. Minimum variance a single component must
#'   explain to be retained.
#' @param min_cumvar Numeric in (0, 1]. Minimum cumulative variance that the
#'   retained components must explain.
#' @param max_ncomp Positive integer. Maximum number of components to
#'   compute or evaluate.
#' @param ncomp Positive integer. Exact number of components to use.
#'
#' @details
#' Four selection methods are available:
#'
#' \describe{
#'   \item{`ncomp_by_var()`}{Retains components that individually explain at
#'     least `min_var` proportion of variance.}
#'   \item{`ncomp_by_cumvar()`}{Retains the minimum number of components whose
#'     combined explained variance reaches `min_cumvar`.
#'   }
#'   \item{`ncomp_by_opc()`}{Optimized principal component selection based on
#'     side information (Ramirez-Lopez et al., 2013). The optimal number of
#'     components minimizes the RMSD between each observation's response and
#'     its nearest neighbor's response in the projected space. Requires `Yr`.}
#'   \item{`ncomp_fixed()`}{Uses exactly `ncomp` components with no automatic
#'     selection. Equivalent to passing an integer directly.}
#' }
#'
#' At runtime, `max_ncomp` is capped at `min(max_ncomp, nrow(X), ncol(X))`.
#'
#' @return
#' An object of class `"ncomp_selection"` with a subclass indicating the
#' method:
#' \itemize{
#'   \item `ncomp_by_var`: class `c("ncomp_by_var", "ncomp_selection")`
#'   \item `ncomp_by_cumvar`: class `c("ncomp_by_cumvar", "ncomp_selection")`
#'   \item `ncomp_by_opc`: class `c("ncomp_by_opc", "ncomp_selection")`
#'   \item `ncomp_fixed`: class `c("ncomp_fixed", "ncomp_selection")`
#' }
#'
#' @references
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
#' 268-279.
#'
#' @seealso [diss_pca()], [diss_pls()], [dissimilarity()]
#'
#' @examples
#' # Retain components explaining >= 1% variance each
#' ncomp_by_var(0.01)
#'
#' # Retain enough components for 99% cumulative variance
#' ncomp_by_cumvar(0.99)
#'
#' # Optimize using side information (requires Yr)
#' ncomp_by_opc(max_ncomp = 40)
#'
#' # Fix at exactly 10 components
#' ncomp_fixed(10)
#'
#' # Usage in dissimilarity constructors
#' diss_pca(ncomp = ncomp_by_var(0.01))
#' diss_pca(ncomp = ncomp_by_opc())
#' diss_pca(ncomp = 10)  
#'
#' @name ncomp_selection
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
NULL

#' @rdname ncomp_selection
#' @export
ncomp_by_var <- function(min_var = 0.01, max_ncomp = 40L) {
  
  if (!is.numeric(min_var) || length(min_var) != 1L ||
      min_var <= 0 || min_var > 1) {
    stop("'min_var' must be a single number in (0, 1].", call. = FALSE)
  }
  .validate_max_ncomp(max_ncomp)
  
  structure(
    list(
      min_var = min_var,
      max_ncomp = as.integer(max_ncomp)
    ),
    class = c("ncomp_by_var", "ncomp_selection")
  )
}

#' @rdname ncomp_selection
#' @export
ncomp_by_cumvar <- function(min_cumvar = 0.99, max_ncomp = 40L) {
  if (!is.numeric(min_cumvar) || length(min_cumvar) != 1L ||
      min_cumvar <= 0 || min_cumvar > 1) {
    stop("'min_cumvar' must be a single number in (0, 1].", call. = FALSE)
  }
  .validate_max_ncomp(max_ncomp)
  
  structure(
    list(
      min_cumvar = min_cumvar,
      max_ncomp = as.integer(max_ncomp)
    ),
    class = c("ncomp_by_cumvar", "ncomp_selection")
  )
}

#' @rdname ncomp_selection
#' @export
ncomp_by_opc <- function(max_ncomp = 40L) {
  .validate_max_ncomp(max_ncomp)
  
  structure(
    list(max_ncomp = as.integer(max_ncomp)),
    class = c("ncomp_by_opc", "ncomp_selection")
  )
}

#' @rdname ncomp_selection
#' @export
ncomp_fixed <- function(ncomp) {
  if (missing(ncomp)) {
    stop("'ncomp' is required.", call. = FALSE)
  }
  if (!is.numeric(ncomp) || length(ncomp) != 1L || ncomp < 1L) {
    stop("'ncomp' must be a single positive integer.", call. = FALSE)
  }
  
  structure(
    list(ncomp = as.integer(ncomp)),
    class = c("ncomp_fixed", "ncomp_selection")
  )
}

# --- internal validator ------------------------------------------------------
.validate_max_ncomp <- function(max_ncomp) {
  if (!is.numeric(max_ncomp) || length(max_ncomp) != 1L || max_ncomp < 1L) {
    stop("'max_ncomp' must be a single positive integer.", call. = FALSE)
  }
  invisible(NULL)
}

# --- print method ------------------------------------------------------------
#' @export
print.ncomp_selection <- function(x, ...) {
  label <- switch(
    class(x)[[1L]],
    ncomp_by_var = sprintf(
      "by per-component variance >= %.2f (max: %d)",
      x$min_var, x$max_ncomp
    ),
    ncomp_by_cumvar = sprintf(
      "by cumulative variance >= %.2f (max: %d)",
      x$min_cumvar, x$max_ncomp
    ),
    ncomp_by_opc = sprintf("by OPC (max: %d)", x$max_ncomp),
    ncomp_fixed = sprintf("fixed: %d", x$ncomp)
  )
  cat("Component selection:", label, "\n")
  invisible(x)
}

context("test-pc_projection")
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# tests/testthat/test-ortho_projection.R

# =============================================================================
# Setup helper
# =============================================================================

.setup_ortho_data <- function(n_xr = 40, n_xu = 20) {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Yr_2 <- NIRsoil$Ciso[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  y_sel <- !is.na(Yr) & !is.na(Yr_2)
  Xr <- Xr[y_sel, ]
  
  Yu <- Yu[!is.na(Yu)]
  Yr_2 <- Yr_2[y_sel]
  Yr <- Yr[y_sel]
  
  Xu <- Xu[seq_len(n_xu), ]
  Yu <- Yu[seq_len(n_xu)]
  Xr <- Xr[seq_len(n_xr), ]
  Yr <- Yr[seq_len(n_xr)]
  Yr_2 <- Yr_2[seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yr_2 = Yr_2, Yu = Yu)
}


# =============================================================================
# Basic ortho_projection tests
# =============================================================================

test_that("ortho_projection with ncomp_by_cumvar works (single matrix)", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  cumvar_value <- 0.999
  
  result <- ortho_projection(
    Xr = d$Xr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
  
  # All but last component should have cumvar < threshold
  test_ncomp <- result$ncomp - 1
  expect_true(all(
    result$variance$x_var["cumulative_explained_var", seq_len(test_ncomp)] < cumvar_value
  ))
})


test_that("ortho_projection with ncomp_by_cumvar works (two matrices)", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  cumvar_value <- 0.999
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_by_cumvar(cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
  expect_equal(nrow(result$scores), nrow(d$Xr) + nrow(d$Xu))
  
  test_ncomp <- result$ncomp - 1
  expect_true(all(
    result$variance$x_var["cumulative_explained_var", seq_len(test_ncomp)] < cumvar_value
  ))
})


test_that("ortho_projection with ncomp_by_var works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  min_var <- 0.001  # equivalent to 1 - 0.999
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_by_var(min_var),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
  
  # All retained components should explain >= min_var
  expect_true(all(
    result$variance$x_var["explained_var", seq_len(result$ncomp)] >= min_var
  ))
})


test_that("ortho_projection with ncomp_fixed works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_fixed(10),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(result$ncomp, 10)
  expect_equal(ncol(result$scores), 10)
})


test_that("ortho_projection with integer ncomp works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = 10,
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(result$ncomp, 10)
})


# =============================================================================
# predict method tests
# =============================================================================

test_that("predict.ortho_projection works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  tol <- 1e-5
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_by_cumvar(0.999),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  # Predicting Xr should match stored scores for Xr rows
  preds <- predict(result, d$Xr)
  stored_xr <- predict(result)[seq_len(nrow(d$Xr)), ]
  
  expect_lt(sum(abs(preds - stored_xr)), tol)
})


# =============================================================================
# ncomp_by_opc tests
# =============================================================================

test_that("ortho_projection with ncomp_by_opc works (PCA)", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    center = TRUE, 
    scale = TRUE,
    method = "pca"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_true("opc_evaluation" %in% names(result))
  
  # ncomp should match the minimum RMSD position
  expect_equal(result$ncomp, as.vector(which.min(result$opc_evaluation[, "rmsd_Yr"])))
  expect_equal(result$ncomp, 7)
})


test_that("ortho_projection pca vs pca_nipals give equivalent results", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  tol <- 1e-5
  
  opc_pca <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    center = TRUE, 
    scale = TRUE,
    method = "pca"
  )
  
  opc_nipals <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 30),
    center = TRUE, 
    scale = TRUE,
    method = "pca_nipals"
  )
  
  # Same number of components selected
  expect_equal(opc_pca$ncomp, opc_nipals$ncomp)
  
  # Scores should be equivalent (up to sign)
  cor_equiv <- sapply(
    seq_len(opc_pca$ncomp),
    function(i) abs(cor(opc_pca$scores[, i], opc_nipals$scores[, i]))
  )
  
  expect_lt(sum(1 - cor_equiv), tol)
})


test_that("ncomp_by_var and ncomp_by_cumvar select consistent components", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  cumvar_value <- 0.999
  
  # Get full variance information via OPC (more components)
  opc_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    center = TRUE, 
    scale = TRUE,
    method = "pca"
  )
  
  cumvar_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_by_cumvar(cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  var_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    ncomp = ncomp_by_var(1 - cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  # Check consistency between selection methods
  # cumvar: counts components where cumvar < threshold, then adds 1
  expect_equal(
    sum(opc_result$variance$x_var["cumulative_explained_var", ] < cumvar_value),
    cumvar_result$ncomp - 1
  )
  
  # var: counts components where var >= threshold
  expect_equal(
    sum(opc_result$variance$x_var["explained_var", ] >= (1 - cumvar_value)),
    var_result$ncomp
  )
})


# =============================================================================
# Multi-response Yr tests
# =============================================================================

test_that("ortho_projection with multi-column Yr works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  Yr_multi <- cbind(name_test_yr = d$Yr, Yr_2 = d$Yr_2)
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = Yr_multi,
    ncomp = ncomp_by_opc(max_ncomp = 30),
    center = TRUE, 
    scale = FALSE,
    method = "pca_nipals"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_true("opc_evaluation" %in% names(result))
  expect_true("rmsd_name_test_yr" %in% colnames(result$opc_evaluation))
  expect_true("rmsd_Yr_2" %in% colnames(result$opc_evaluation))
})


# =============================================================================
# PLS method tests
# =============================================================================

test_that("ortho_projection with method = 'pls' works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    method = "pls",
    scale = TRUE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_true("weights" %in% names(result))
  expect_true("projection_mat" %in% names(result))
  expect_true("Y_loadings" %in% names(result))
})


test_that("ortho_projection with method = 'mpls' works", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = 10,
    method = "mpls",
    scale = TRUE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(result$ncomp, 10)
  expect_equal(result$method, "mpls")
})


test_that("ortho_projection PLS requires Yr", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  expect_error(
    ortho_projection(
      Xr = d$Xr, 
      Xu = d$Xu,
      ncomp = 10,
      method = "pls"
    ),
    "'Yr' is required"
  )
})


test_that("ortho_projection ncomp_by_opc requires Yr", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  expect_error(
    ortho_projection(
      Xr = d$Xr, 
      Xu = d$Xu,
      ncomp = ncomp_by_opc(),
      method = "pca"
    ),
    "'Yr' is required"
  )
})


# =============================================================================
# Validation tests
# =============================================================================

test_that("ortho_projection validates center and scale", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  expect_error(
    ortho_projection(d$Xr, center = "yes"),
    "TRUE or FALSE"
  )
  
  expect_error(
    ortho_projection(d$Xr, scale = "yes"),
    "TRUE or FALSE"
  )
})


test_that("ortho_projection validates method", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_ortho_data()
  
  expect_error(
    ortho_projection(d$Xr, method = "invalid"),
    fixed = '"pca", "pca_nipals", "pls", "mpls", "simpls"'
  )
})


# =============================================================================
# Large dataset tests (skipped on CRAN)
# =============================================================================

test_that("ortho_projection works with larger datasets", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  tol <- 1e-5
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  cumvar_value <- 0.999
  
  # Single matrix
  one_input <- ortho_projection(
    Xr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_equal(ncol(one_input$scores), one_input$ncomp)
  test_ncomp <- one_input$ncomp - 1
  expect_true(all(
    one_input$variance$x_var["cumulative_explained_var", seq_len(test_ncomp)] < cumvar_value
  ))
  
  # Two matrices
  two_input <- ortho_projection(
    Xr, Xu,
    ncomp = ncomp_by_cumvar(cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_equal(ncol(two_input$scores), two_input$ncomp)
  
  # Predictions match
  preds <- sum(abs(
    predict(two_input)[seq_len(nrow(Xr)), ] - predict(two_input, Xr)
  ))
  expect_lt(preds, tol)
  
  # OPC selection
  opc_pca <- ortho_projection(
    Xr, Xu,
    Yr = Yr,
    ncomp = ncomp_by_opc(max_ncomp = 30),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  opc_nipals <- ortho_projection(
    Xr, Xu,
    Yr = Yr,
    ncomp = ncomp_by_opc(max_ncomp = 30),
    center = TRUE, 
    scale = FALSE,
    method = "pca_nipals"
  )
  
  expect_equal(opc_pca$ncomp, as.vector(which.min(opc_pca$opc_evaluation[, "rmsd_Yr"])))
  expect_equal(opc_pca$ncomp, 20)
  
  # PCA and NIPALS should select same number
  expect_equal(opc_pca$ncomp, opc_nipals$ncomp)
  
  # Scores should be equivalent
  cor_equiv <- sapply(
    seq_len(opc_pca$ncomp),
    function(i) abs(cor(opc_pca$scores[, i], opc_nipals$scores[, i]))
  )
  expect_lt(sum(1 - cor_equiv), tol)
  
  # Variance selection consistency
  expect_equal(
    sum(opc_pca$variance$x_var["cumulative_explained_var", ] < cumvar_value),
    two_input$ncomp - 1
  )
  
  var_result <- ortho_projection(
    Xr, Xu,
    ncomp = ncomp_by_var(1 - cumvar_value),
    center = TRUE, 
    scale = FALSE,
    method = "pca"
  )
  
  expect_equal(
    sum(opc_pca$variance$x_var["explained_var", ] >= (1 - cumvar_value)),
    var_result$ncomp
  )
})

context("test-dissimilarity")

# tests/testthat/test-dissimilarity.R
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# =============================================================================
# Setup helper
# =============================================================================

.setup_nirsoil_data <- function(n_xr = 40, n_xu = 20) {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ][seq_len(n_xu), ]
  Xr <- Xr[!is.na(Yr), ][seq_len(n_xr), ]
  Yu <- Yu[!is.na(Yu)][seq_len(n_xu)]
  Yr <- Yr[!is.na(Yr)][seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}



# =============================================================================
# diss_* constructor tests
# =============================================================================

test_that("diss_correlation constructor works", {
  x <- diss_correlation(ws = 11)
  expect_s3_class(x, "diss_method")
})

# =============================================================================
# Tests for diss_euclidean, diss_mahalanobis, diss_cosine constructors
# =============================================================================


# -----------------------------------------------------------------------------
# diss_euclidean constructor tests
# -----------------------------------------------------------------------------

test_that("diss_euclidean returns correct class", {
  m <- diss_euclidean()
  expect_s3_class(m, "diss_euclidean")
  expect_s3_class(m, "diss_method")
})

test_that("diss_euclidean stores parameters correctly", {
  m1 <- diss_euclidean()
  expect_true(m1$center)
  expect_false(m1$scale)
  expect_equal(m1$method, "euclidean")
  
  m2 <- diss_euclidean(center = FALSE, scale = TRUE)
  expect_false(m2$center)
  expect_true(m2$scale)
})

test_that("diss_euclidean computes correct dissimilarities", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  # Test without centering/scaling - compare to stats::dist
  result <- dissimilarity(Xr, diss_method = diss_euclidean(center = FALSE, scale = FALSE))
  
  expect_true(is.matrix(result$dissimilarity))
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xr))
  
  # Diagonal should be zero (self-dissimilarity)
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
  
  # Should be symmetric
  expect_equal(result$dissimilarity, t(result$dissimilarity), tolerance = 1e-10)
})

test_that("diss_euclidean with Xu computes cross-dissimilarity", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  Xu <- NIRsoil$spc[21:30, ]
  
  
  result <- dissimilarity(Xr, Xu, diss_method = diss_euclidean())
  
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xu))
})

test_that("diss_euclidean centering affects results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  result_centered <- dissimilarity(Xr, diss_method = diss_euclidean(center = TRUE))
  
  result_uncentered <- dissimilarity(Xr, diss_method = diss_euclidean(center = FALSE))
  
  # Results should differ when centering is applied
  expect_false(all(result_centered$dissimilarity == result_uncentered$dissimilarity))
})

test_that("diss_euclidean scaling affects results", {
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  result_scaled <- dissimilarity(Xr, diss_method = diss_euclidean(scale = TRUE))
  result_unscaled <- dissimilarity(Xr, diss_method = diss_euclidean(scale = FALSE))
  
  # Results should differ when scaling is applied
  expect_false(all(result_scaled$dissimilarity == result_unscaled$dissimilarity))
})

# -----------------------------------------------------------------------------
# diss_mahalanobis constructor tests
# -----------------------------------------------------------------------------

test_that("diss_mahalanobis returns correct class", {
  m <- diss_mahalanobis()
  expect_s3_class(m, "diss_mahalanobis")
  expect_s3_class(m, "diss_method")
})

test_that("diss_mahalanobis stores parameters correctly", {
  m1 <- diss_mahalanobis()
  expect_true(m1$center)
  expect_false(m1$scale)
  expect_equal(m1$method, "mahalanobis")
  
  m2 <- diss_mahalanobis(center = FALSE, scale = TRUE)
  expect_false(m2$center)
  expect_true(m2$scale)
})

test_that("diss_mahalanobis computes dissimilarities when n > p", {
  # Create data where n > p (more observations than variables)
  skip_on_cran()
  set.seed(123)
  Xr <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  
  result <- dissimilarity(Xr, diss_method = diss_mahalanobis())
  
  expect_true(is.matrix(result$dissimilarity))
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xr))
  
  # Diagonal should be zero
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
  
  # Should be symmetric
  expect_equal(result$dissimilarity, t(result$dissimilarity), tolerance = 1e-10)
})

test_that("diss_mahalanobis errors when covariance is singular", {
  # Create data where n < p (fewer observations than variables) - singular covariance
  set.seed(123)
  Xr <- matrix(rnorm(10 * 100), nrow = 10, ncol = 100)
  
  expect_error(
    dissimilarity(Xr, diss_method = diss_mahalanobis()),
    "For computing the Mahalanobis distance, the total number of observations"
  )
})

test_that("diss_mahalanobis with Xu computes cross-dissimilarity", {
  skip_on_cran()
  set.seed(123)
  Xr <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  Xu <- matrix(rnorm(20 * 10), nrow = 20, ncol = 10)
  
  result <- dissimilarity(Xr, Xu, diss_method = diss_mahalanobis())
  
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xu))
})

test_that("diss_mahalanobis differs from diss_euclidean", {
  set.seed(123)
  Xr <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  
  result_mahal <- dissimilarity(Xr, diss_method = diss_mahalanobis())
  result_euclid <- dissimilarity(Xr, diss_method = diss_euclidean())
  
  # Mahalanobis and Euclidean should give different results
  expect_false(all(result_mahal$dissimilarity == result_euclid$dissimilarity))
})

# -----------------------------------------------------------------------------
# diss_cosine constructor tests
# -----------------------------------------------------------------------------

test_that("diss_cosine returns correct class", {
  m <- diss_cosine()
  expect_s3_class(m, "diss_cosine")
  expect_s3_class(m, "diss_method")
})

test_that("diss_cosine stores parameters correctly", {
  m1 <- diss_cosine()
  expect_true(m1$center)
  expect_false(m1$scale)
  expect_equal(m1$method, "cosine")
  
  m2 <- diss_cosine(center = FALSE, scale = TRUE)
  expect_false(m2$center)
  expect_true(m2$scale)
})

test_that("diss_cosine computes correct dissimilarities", {
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  result <- dissimilarity(Xr, diss_method = diss_cosine())
  
  expect_true(is.matrix(result$dissimilarity))
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xr))
  
  # Diagonal should be zero (self-dissimilarity)
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
  
  # Should be symmetric
  expect_equal(result$dissimilarity, t(result$dissimilarity), tolerance = 1e-10)
})

test_that("diss_cosine values are bounded", {
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  result <- dissimilarity(Xr, diss_method = diss_cosine(center = FALSE))
  
  # Cosine dissimilarity (angle) should be between 0 and pi
  expect_true(all(result$dissimilarity >= 0))
  expect_true(all(result$dissimilarity <= pi + 1e-10))
})

test_that("diss_cosine with Xu computes cross-dissimilarity", {
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  Xu <- NIRsoil$spc[21:30, ]
  
  result <- dissimilarity(Xr, Xu, diss_method = diss_cosine())
  
  expect_equal(nrow(result$dissimilarity), nrow(Xr))
  expect_equal(ncol(result$dissimilarity), nrow(Xu))
})

test_that("diss_cosine differs from diss_euclidean", {
  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[1:20, ]
  
  result_cosine <- dissimilarity(Xr, diss_method = diss_cosine())
  result_euclid <- dissimilarity(Xr, diss_method = diss_euclidean())
  
  # Cosine and Euclidean should give different results
  expect_false(all(result_cosine$dissimilarity == result_euclid$dissimilarity))
})

test_that("diss_cosine handles identical observations", {
  # Two identical rows should have zero dissimilarity
  Xr <- matrix(c(1, 2, 3, 1, 2, 3, 4, 5, 6), nrow = 3, byrow = TRUE)
  
  result <- dissimilarity(Xr, diss_method = diss_cosine(center = FALSE))
  
  # First two rows are identical
  
  expect_equal(result$dissimilarity[1, 2], 0, tolerance = 1e-10)
  expect_equal(result$dissimilarity[2, 1], 0, tolerance = 1e-10)
})

# -----------------------------------------------------------------------------
# Print method tests
# -----------------------------------------------------------------------------

test_that("print.diss_method works for all constructors", {
  expect_output(print(diss_euclidean()), "euclidean")
  expect_output(print(diss_mahalanobis()), "mahalanobis")
  expect_output(print(diss_cosine()), "cosine")
})

# -----------------------------------------------------------------------------
# Edge cases
# -----------------------------------------------------------------------------

test_that("not possible to have a single observation", {
  Xr <- matrix(1:10, nrow = 1)
  
  expect_error(dissimilarity(Xr, diss_method = diss_euclidean()))
})

test_that("constructors handle two observations", {
  Xr <- matrix(1:20, nrow = 2)
  
  result_euclid <- dissimilarity(Xr, diss_method = diss_euclidean())
  expect_equal(dim(result_euclid$dissimilarity), c(2, 2))
  expect_equal(result_euclid$dissimilarity[1, 2], result_euclid$dissimilarity[2, 1])
  
  result_cosine <- dissimilarity(Xr, diss_method = diss_cosine())
  expect_equal(dim(result_cosine$dissimilarity), c(2, 2))
  expect_equal(result_cosine$dissimilarity[1, 2], result_cosine$dissimilarity[2, 1])
})



test_that("diss_pca constructor works", {
  m <- diss_pca()
  expect_s3_class(m, c("diss_pca", "diss_method"))
  expect_s3_class(m$ncomp, "ncomp_by_var")  # default
  expect_equal(m$method, "pca")
  expect_true(m$center)
  expect_false(m$scale)
  expect_false(m$return_projection)
  
  m2 <- diss_pca(
    ncomp = 10, 
    method = "pca_nipals", 
    center = FALSE, 
    scale = TRUE,
    return_projection = TRUE
  )
  expect_s3_class(m2$ncomp, "ncomp_fixed")
  expect_equal(m2$method, "pca_nipals")
  expect_false(m2$center)
  expect_true(m2$scale)
  expect_true(m2$return_projection)
})


test_that("diss_pls constructor works", {
  m <- diss_pls()
  expect_s3_class(m, c("diss_pls", "diss_method"))
  expect_s3_class(m$ncomp, "ncomp_by_opc")  # default
  expect_equal(m$method, "pls")
  expect_true(m$center)  # PLS always centers
  expect_false(m$scale)
  
  m2 <- diss_pls(ncomp = 15, method = "mpls", scale = TRUE)
  expect_s3_class(m2$ncomp, "ncomp_fixed")
  expect_equal(m2$method, "mpls")
  expect_true(m2$scale)
})


test_that("diss_euclidean constructor works", {
  m <- diss_euclidean()
  expect_s3_class(m, c("diss_euclidean", "diss_method"))
  expect_true(m$center)
  expect_false(m$scale)
  
  m2 <- diss_euclidean(center = FALSE, scale = TRUE)
  expect_false(m2$center)
  expect_true(m2$scale)
})


test_that("diss_mahalanobis constructor works", {
  m <- diss_mahalanobis()
  expect_s3_class(m, c("diss_mahalanobis", "diss_method"))
  expect_true(m$center)
  expect_false(m$scale)
})


test_that("diss_cosine constructor works", {
  m <- diss_cosine()
  expect_s3_class(m, c("diss_cosine", "diss_method"))
  expect_true(m$center)
  expect_false(m$scale)
})


test_that("diss_correlation constructor works", {
  m <- diss_correlation()
  expect_s3_class(m, c("diss_correlation", "diss_method"))
  expect_null(m$ws)
  expect_true(m$center)
  expect_false(m$scale)
  
  m2 <- diss_correlation(ws = 41)
  expect_equal(m2$ws, 41L)
})


test_that("diss_correlation validates ws", {
  expect_error(diss_correlation(ws = 2), "greater than 2")
  expect_error(diss_correlation(ws = 4), "odd integer")
  expect_error(diss_correlation(ws = "a"), "single integer")
})


# =============================================================================
# diss_pca dissimilarity tests
# =============================================================================

test_that("diss_pca with ncomp_by_var works", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_pca(
      ncomp = ncomp_by_var(0.01),
      center = TRUE, 
      scale = TRUE,
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_true("projection" %in% names(result))
  expect_true("ncomp" %in% names(result))
  expect_equal(nrow(result$dissimilarity), nrow(d$Xr))
  expect_equal(ncol(result$dissimilarity), nrow(d$Xu))
})


test_that("diss_pca with ncomp_by_opc works", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    diss_method = diss_pca(
      ncomp = ncomp_by_opc(max_ncomp = 15),
      center = TRUE, 
      scale = TRUE,
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_true("projection" %in% names(result))
  expect_true(result$ncomp <= 15)
})


test_that("diss_pca with ncomp_by_opc requires Yr", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  
  d <- .setup_nirsoil_data()
  
  expect_error(
    dissimilarity(
      Xr = d$Xr, 
      Xu = d$Xu,
      diss_method = diss_pca(ncomp = ncomp_by_opc())
    ),
    "'Yr' is required"
  )
})


test_that("diss_pca with ncomp_by_cumvar works", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_pca(
      ncomp = ncomp_by_cumvar(0.99),
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true(result$ncomp >= 1)
})


test_that("diss_pca with fixed ncomp works", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  d <- .setup_nirsoil_data()
  
  # Integer form
  result1 <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10)
  )
  
  # Explicit ncomp_fixed form
  result2 <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_pca(ncomp = ncomp_fixed(10))
  )
  
  expect_equal(result1$ncomp, 10)
  expect_equal(result2$ncomp, 10)
  expect_equal(result1$dissimilarity, result2$dissimilarity)
})


test_that("diss_pca with pca_nipals method works", {
  skip_if_not_installed("prospectr")
  skip_on_cran()
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_pca(
      ncomp = 10,
      method = "pca_nipals",
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_equal(result$ncomp, 10)
})


test_that("diss_pca self-dissimilarity is symmetric with zero diagonal", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 20, n_xu = 0)
  
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_pca(ncomp = 5)
  )
  
  expect_equal(result$dissimilarity, t(result$dissimilarity))
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
})


# =============================================================================
# diss_pls dissimilarity tests
# =============================================================================

test_that("diss_pls with ncomp_by_opc works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  Xr_snv <- prospectr::standardNormalVariate(d$Xr)
  Xu_snv <- prospectr::standardNormalVariate(d$Xu)
  
  result <- dissimilarity(
    Xr = Xr_snv, 
    Xu = Xu_snv,
    Yr = d$Yr,
    diss_method = diss_pls(
      ncomp = ncomp_by_opc(max_ncomp = 15),
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_true("projection" %in% names(result))
  expect_true(result$ncomp <= 15)
})


test_that("diss_pls with ncomp_by_var works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    diss_method = diss_pls(
      ncomp = ncomp_by_var(0.01),
      return_projection = TRUE
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true(result$ncomp >= 1)
})


test_that("diss_pls with fixed ncomp works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    diss_method = diss_pls(ncomp = 15)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_equal(result$ncomp, 15)
})


test_that("diss_pls with mpls method works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    diss_method = diss_pls(
      ncomp = 10,
      method = "mpls"
    )
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_equal(result$ncomp, 10)
})


test_that("diss_pls requires Yr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  expect_error(
    dissimilarity(
      Xr = d$Xr, 
      Xu = d$Xu,
      diss_method = diss_pls()
    ),
    "'Yr' is required"
  )
})


# =============================================================================
# diss_euclidean dissimilarity tests
# =============================================================================

test_that("diss_euclidean works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_euclidean(center = TRUE, scale = TRUE)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_equal(nrow(result$dissimilarity), nrow(d$Xr))
  expect_equal(ncol(result$dissimilarity), nrow(d$Xu))
})


test_that("diss_euclidean matches stats::dist", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 10, n_xu = 0)
  
  # Compute with resemble (no centering/scaling for direct comparison)
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_euclidean(center = FALSE, scale = FALSE)
  )
  
  # resemble scales by 1/sqrt(p), so convert back
  resemble_diss <- result$dissimilarity
  resemble_diss_unscaled <- ((resemble_diss^2) * ncol(d$Xr))^0.5
  
  # Compare with base R
  base_diss <- as.matrix(dist(d$Xr))
  
  expect_lt(sum(abs(resemble_diss_unscaled - base_diss)), 1e-5)
})


test_that("diss_euclidean self-dissimilarity is symmetric", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 20, n_xu = 0)
  
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_euclidean()
  )
  
  expect_equal(result$dissimilarity, t(result$dissimilarity))
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
})


# =============================================================================
# diss_correlation dissimilarity tests
# =============================================================================

test_that("diss_correlation (full) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_correlation(ws = NULL, center = TRUE, scale = FALSE)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_equal(nrow(result$dissimilarity), nrow(d$Xr))
  expect_equal(ncol(result$dissimilarity), nrow(d$Xu))
})


test_that("diss_correlation (moving window) works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_correlation(ws = 11)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
})


test_that("diss_correlation vs one computed with cor()", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_correlation(center = FALSE)
  )
  
  result_base_cor <- 0.5 * (1 - cor(t(d$Xr), t(d$Xu)))
  
  # Results should be similar but not identical 
  expect_lt(
    max(abs(result$dissimilarity - result_base_cor)),
    1e-9
  )
})


# =============================================================================
# diss_cosine dissimilarity tests
# =============================================================================

test_that("diss_cosine works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  result <- dissimilarity(
    Xr = d$Xr, 
    Xu = d$Xu,
    diss_method = diss_cosine(center = TRUE, scale = FALSE)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
  expect_equal(nrow(result$dissimilarity), nrow(d$Xr))
  expect_equal(ncol(result$dissimilarity), nrow(d$Xu))
})


test_that("diss_cosine self-dissimilarity diagonal is zero", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 20, n_xu = 0)
  
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_cosine()
  )
  
  expect_true(all(abs(diag(result$dissimilarity)) < 1e-6))
})


# =============================================================================
# diss_mahalanobis dissimilarity tests
# =============================================================================

test_that("diss_mahalanobis works when n > p", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 40, n_xu = 20)
  
  # Use only first 20 columns to ensure n > p
  result <- dissimilarity(
    Xr = d$Xr[, 1:20], 
    Xu = d$Xu[, 1:20],
    diss_method = diss_mahalanobis(center = TRUE, scale = FALSE)
  )
  
  expect_s3_class(result, "dissimilarity")
  expect_true("dissimilarity" %in% names(result))
})


test_that("diss_mahalanobis errors when covariance is singular", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 10, n_xu = 5)
  
  # Full spectra: p >> n, covariance will be singular
  expect_error(
    dissimilarity(
      Xr = d$Xr,
      Xu = d$Xu,
      diss_method = diss_mahalanobis()
    ),
    "number of observations.*larger than the number of variables"
  )
})


# =============================================================================
# Legacy API tests
# =============================================================================

test_that("character-based diss_method is rejected with helpful message", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data()
  
  expect_error(
    dissimilarity(Xr = d$Xr, Xu = d$Xu, diss_method = "pca"),
    "no longer supported"
  )
  
  expect_error(
    dissimilarity(Xr = d$Xr, Xu = d$Xu, diss_method = "euclid"),
    "diss_euclidean"
  )
})


# =============================================================================
# Large dataset tests (skipped on CRAN)
# =============================================================================

test_that("dissimilarity with large datasets works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  Xr_snv <- prospectr::standardNormalVariate(Xr)
  Xu_snv <- prospectr::standardNormalVariate(Xu)
  
  # PCA with opc
  result_pca <- dissimilarity(
    Xr = Xr, 
    Xu = Xu,
    Yr = Yr,
    diss_method = diss_pca(
      ncomp = ncomp_by_opc(max_ncomp = 30),
      center = TRUE, 
      scale = TRUE,
      return_projection = TRUE
    )
  )
  
  # PLS with opc
  result_pls <- dissimilarity(
    Xr = Xr_snv, 
    Xu = Xu_snv,
    Yr = Yr,
    diss_method = diss_pls(
      ncomp = ncomp_by_opc(max_ncomp = 30),
      return_projection = TRUE
    )
  )
  
  # Euclidean
  result_euclid <- dissimilarity(
    Xr = Xr, 
    Xu = Xu,
    diss_method = diss_euclidean(center = TRUE, scale = TRUE)
  )
  
  # Correlation with moving window
  result_cor <- dissimilarity(
    Xr = Xr, 
    Xu = Xu,
    diss_method = diss_correlation(ws = 11)
  )
  
  expect_s3_class(result_pca, "dissimilarity")
  expect_s3_class(result_pls, "dissimilarity")
  expect_s3_class(result_euclid, "dissimilarity")
  expect_s3_class(result_cor, "dissimilarity")
  
  expect_true(result_pca$ncomp <= 30)
  expect_true(result_pls$ncomp <= 30)
  
  # Verify Euclidean against base R
  result_euclid_self <- dissimilarity(
    Xr = Xu[1:10, ],
    diss_method = diss_euclidean(center = FALSE, scale = FALSE)
  )
  resemble_diss <- result_euclid_self$dissimilarity
  resemble_diss_unscaled <- ((resemble_diss^2) * ncol(Xu))^0.5
  base_diss <- as.matrix(dist(Xu[1:10, ]))
  
  expect_lt(sum(abs(resemble_diss_unscaled - base_diss)), 1e-5)
})


# =============================================================================
# diss_evaluate tests
# =============================================================================

test_that("diss_evaluate works with continuous side_info", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 40, n_xu = 0)
  
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_pca(ncomp = 8)
  )
  
  ev <- diss_evaluate(
    diss = result$dissimilarity, 
    side_info = as.matrix(d$Yr)
  )
  
  expect_true("eval" %in% names(ev))
  expect_true("first_nn" %in% names(ev))
  expect_true("rmsd" %in% colnames(ev$eval))
  expect_true("r" %in% colnames(ev$eval))
})


test_that("diss_evaluate works with multiple side_info columns", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Yr2 <- NIRsoil$Ciso[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  complete <- !is.na(Yr) & !is.na(Yr2)
  Xr <- Xr[complete, ][1:40, ]
  Yr <- Yr[complete][1:40]
  Yr2 <- Yr2[complete][1:40]
  
  result <- dissimilarity(
    Xr = Xr,
    diss_method = diss_pca(ncomp = 8)
  )
  
  ev <- diss_evaluate(
    diss = result$dissimilarity, 
    side_info = cbind(Yr, Yr2)
  )
  
  expect_true("eval" %in% names(ev))
  expect_true("global_eval" %in% names(ev))
  expect_equal(nrow(ev$eval), 2)
})


test_that("diss_evaluate validates inputs", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_nirsoil_data(n_xr = 20, n_xu = 0)
  
  result <- dissimilarity(
    Xr = d$Xr,
    diss_method = diss_pca(ncomp = 5)
  )
  
  # side_info must be a matrix
  expect_error(
    diss_evaluate(result$dissimilarity, side_info = d$Yr),
    "must be a matrix"
  )
  
  # Dimension mismatch
  expect_error(
    diss_evaluate(result$dissimilarity, side_info = as.matrix(d$Yr[1:10])),
    "must match"
  )
})





# =============================================================================
# Tests for cor_diss (internal correlation dissimilarity function)
# =============================================================================

# -----------------------------------------------------------------------------
# Basic functionality tests
# -----------------------------------------------------------------------------

test_that("cor_diss returns correct dimensions for Xr only", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xr))
})

test_that("cor_diss returns correct dimensions with Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr, d$Xu)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xu))
})

test_that("cor_diss diagonal is zero for self-dissimilarity", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr)
  
  expect_true(all(abs(diag(result)) < 1e-6))
})

test_that("cor_diss is symmetric", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr)
  
  expect_equal(result, t(result), tolerance = 1e-10)
})

test_that("cor_diss values are bounded [0, 1]", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr)
  
  expect_true(all(result >= -1e-10))
  expect_true(all(result <= 1 + 1e-10))
})

test_that("cor_diss with moving window returns correct dimensions", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr, ws = 41)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xr))
})

test_that("cor_diss with moving window and Xu works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- cor_diss(d$Xr, d$Xu, ws = 41)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xu))
})

test_that("cor_diss with different ws values gives different results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result_no_ws <- cor_diss(d$Xr)
  result_ws41 <- cor_diss(d$Xr, ws = 41)
  result_ws21 <- cor_diss(d$Xr, ws = 21)
  
  expect_false(all(result_no_ws == result_ws41))
  expect_false(all(result_ws41 == result_ws21))
})

test_that("cor_diss centering and scaling work", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result_centered <- cor_diss(d$Xr, center = TRUE)
  result_scaled <- cor_diss(d$Xr, scale = TRUE)
  result_both <- cor_diss(d$Xr, center = TRUE, scale = TRUE)
  
  expect_true(is.matrix(result_centered))
  expect_true(is.matrix(result_scaled))
  expect_true(is.matrix(result_both))
})

test_that("cor_diss row/column names are correct", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data(n_xr = 5, n_xu = 3)
  
  result_xr <- cor_diss(d$Xr)
  result_xu <- cor_diss(d$Xr, d$Xu)
  
  expect_equal(rownames(result_xr), paste0("Xr_", 1:5))
  expect_equal(colnames(result_xr), paste0("Xr_", 1:5))
  expect_equal(rownames(result_xu), paste0("Xr_", 1:5))
  expect_equal(colnames(result_xu), paste0("Xu_", 1:3))
})

# -----------------------------------------------------------------------------
# Error handling tests
# -----------------------------------------------------------------------------

test_that("cor_diss errors with single variable", {
  skip_on_cran()
  Xr <- matrix(1:10, ncol = 1)
  
  expect_error(cor_diss(Xr), "number of variables must be larger than 1")
})

test_that("cor_diss errors with mismatched columns", {
  skip_on_cran()
  Xr <- matrix(1:20, nrow = 2)
  Xu <- matrix(1:10, nrow = 2)
  
  expect_error(cor_diss(Xr, Xu), "number of columns")
})

test_that("cor_diss errors with NA values", {
  skip_on_cran()
  Xr <- matrix(1:20, nrow = 2)
  Xr[1, 1] <- NA
  
  expect_error(cor_diss(Xr), "missing values")
  
  Xr <- matrix(1:20, nrow = 2)
  Xu <- matrix(1:20, nrow = 2)
  Xu[1, 1] <- NA
  
  expect_error(cor_diss(Xr, Xu), "missing values")
})

test_that("cor_diss errors with invalid logical arguments", {
  skip_on_cran()
  Xr <- matrix(1:20, nrow = 2)
  
  expect_error(cor_diss(Xr, center = "yes"), "'center' argument must be logical")
  expect_error(cor_diss(Xr, scale = 1), "'scale' argument must be logical")
})

test_that("cor_diss errors with invalid ws", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  expect_error(cor_diss(d$Xr, ws = 40), "odd value")
  expect_error(cor_diss(d$Xr, ws = 1), "greater than 2")
  expect_error(cor_diss(d$Xr, ws = ncol(d$Xr) + 1), "smaller than the number of columns")
})

test_that("cor_diss errors with zero standard deviation observation", {
  skip_on_cran()
  Xr <- matrix(rnorm(100), nrow = 10)
  Xr[1, ] <- rep(5, 10)
  
  expect_error(cor_diss(Xr, center = FALSE), "Correlation coefficients cannot be computed")
})

# =============================================================================
# Tests for .f_diss_compute (internal function)
# =============================================================================

test_that(".f_diss_compute works with euclidean method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method <- diss_euclidean(center = TRUE, scale = FALSE)
  result <- .f_diss_compute(d$Xr, Xu = NULL, method = method)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xr))
  expect_true(all(abs(diag(result)) < 1e-6))
  expect_equal(result, t(result), tolerance = 1e-10)
})

test_that(".f_diss_compute works with euclidean and Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method <- diss_euclidean()
  result <- .f_diss_compute(d$Xr, Xu = d$Xu, method = method)
  
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xu))
  expect_equal(rownames(result), paste0("Xr_", seq_len(nrow(d$Xr))))
  expect_equal(colnames(result), paste0("Xu_", seq_len(nrow(d$Xu))))
})

test_that(".f_diss_compute works with euclidean center and scale", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method_none <- diss_euclidean(center = FALSE, scale = FALSE)
  method_center <- diss_euclidean(center = TRUE, scale = FALSE)
  method_scale <- diss_euclidean(center = FALSE, scale = TRUE)
  method_both <- diss_euclidean(center = TRUE, scale = TRUE)
  
  result_none <- .f_diss_compute(d$Xr, method = method_none)
  result_center <- .f_diss_compute(d$Xr, method = method_center)
  result_scale <- .f_diss_compute(d$Xr, method = method_scale)
  result_both <- .f_diss_compute(d$Xr, method = method_both)
  
  # All should produce valid matrices
  expect_true(is.matrix(result_none))
  expect_true(is.matrix(result_center))
  expect_true(is.matrix(result_scale))
  expect_true(is.matrix(result_both))
  
  # Results should differ
  expect_false(all(result_none == result_center))
  expect_false(all(result_none == result_scale))
})

test_that(".f_diss_compute works with mahalanobis method", {
  skip_on_cran()
  
  # Create data where n > p (required for Mahalanobis)
  set.seed(123)
  Xr <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  
  method <- diss_mahalanobis(center = TRUE, scale = FALSE)
  result <- .f_diss_compute(Xr, Xu = NULL, method = method)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(Xr))
  expect_equal(ncol(result), nrow(Xr))
  expect_true(all(abs(diag(result)) < 1e-6))
  expect_equal(result, t(result), tolerance = 1e-10)
})

test_that(".f_diss_compute works with mahalanobis and Xu", {
  skip_on_cran()
  
  set.seed(123)
  Xr <- matrix(rnorm(100 * 10), nrow = 100, ncol = 10)
  Xu <- matrix(rnorm(20 * 10), nrow = 20, ncol = 10)
  
  method <- diss_mahalanobis()
  result <- .f_diss_compute(Xr, Xu = Xu, method = method)
  
  expect_equal(nrow(result), nrow(Xr))
  expect_equal(ncol(result), nrow(Xu))
})

test_that(".f_diss_compute errors with singular covariance for mahalanobis", {
  skip_on_cran()
  
  # Create data where n < p (singular covariance)
  set.seed(123)
  Xr <- matrix(rnorm(10 * 100), nrow = 10, ncol = 100)
  
  method <- diss_mahalanobis()
  
  expect_error(
    .f_diss_compute(Xr, method = method),
    "singular"
  )
})

test_that(".f_diss_compute works with cosine method", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method <- diss_cosine(center = FALSE, scale = FALSE)
  result <- .f_diss_compute(d$Xr, Xu = NULL, method = method)
  
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xr))
  expect_true(all(abs(diag(result)) < 1e-6))
  expect_equal(result, t(result), tolerance = 1e-10)
  
  # Cosine dissimilarity should be bounded [0, pi]
  expect_true(all(result >= -1e-10))
  expect_true(all(result <= pi + 1e-10))
})

test_that(".f_diss_compute works with cosine and Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method <- diss_cosine()
  result <- .f_diss_compute(d$Xr, Xu = d$Xu, method = method)
  
  expect_equal(nrow(result), nrow(d$Xr))
  expect_equal(ncol(result), nrow(d$Xu))
})

test_that(".f_diss_compute cosine handles identical observations", {
  skip_on_cran()
  
  # Two identical rows should have zero dissimilarity
  Xr <- matrix(c(1, 2, 3, 1, 2, 3, 4, 5, 6), nrow = 3, byrow = TRUE)
  
  method <- diss_cosine(center = FALSE, scale = FALSE)
  result <- .f_diss_compute(Xr, method = method)
  
  # First two rows are identical - should be zero
  expect_equal(result[1, 2], 0, tolerance = 1e-10)
  expect_equal(result[2, 1], 0, tolerance = 1e-10)
})

test_that(".f_diss_compute cosine replaces NaN with zero", {
  skip_on_cran()
  
  # Perfectly identical rows can cause NaN in acos due to floating point
  Xr <- matrix(rep(1:10, 3), nrow = 3, byrow = TRUE)
  
  method <- diss_cosine(center = FALSE, scale = FALSE)
  result <- .f_diss_compute(Xr, method = method)
  
  # Should have no NaN values
  expect_false(any(is.nan(result)))
  
  # Diagonal and identical rows should be zero
  expect_true(all(abs(diag(result)) < 1e-10))
})

test_that(".f_diss_compute cosine with center and scale", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  method_none <- diss_cosine(center = FALSE, scale = FALSE)
  method_center <- diss_cosine(center = TRUE, scale = FALSE)
  method_both <- diss_cosine(center = TRUE, scale = TRUE)
  
  result_none <- .f_diss_compute(d$Xr, method = method_none)
  result_center <- .f_diss_compute(d$Xr, method = method_center)
  result_both <- .f_diss_compute(d$Xr, method = method_both)
  
  expect_true(is.matrix(result_none))
  expect_true(is.matrix(result_center))
  expect_true(is.matrix(result_both))
})

test_that(".f_diss_compute row/column names are correct", {
  skip_on_cran()
  
  Xr <- matrix(rnorm(25), nrow = 5)
  Xu <- matrix(rnorm(15), nrow = 3)
  
  method <- diss_euclidean()
  
  result_xr <- .f_diss_compute(Xr, method = method)
  expect_equal(rownames(result_xr), paste0("Xr_", 1:5))
  expect_equal(colnames(result_xr), paste0("Xr_", 1:5))
  
  result_xu <- .f_diss_compute(Xr, Xu = Xu, method = method)
  expect_equal(rownames(result_xu), paste0("Xr_", 1:5))
  expect_equal(colnames(result_xu), paste0("Xu_", 1:3))
})

test_that(".f_diss_compute validates method class", {
  skip_on_cran()
  
  Xr <- matrix(rnorm(20), nrow = 4)
  
  expect_error(.f_diss_compute(Xr, method = "euclidean"))
  expect_error(.f_diss_compute(Xr, method = list(method = "euclidean")))
})

test_that(".f_diss_compute validates method name", {
  skip_on_cran()
  
  Xr <- matrix(rnorm(20), nrow = 4)
  
  # Create a fake diss_method with invalid method name
  fake_method <- structure(
    list(method = "invalid", center = TRUE, scale = FALSE),
    class = c("diss_invalid", "diss_method")
  )
  
  expect_error(.f_diss_compute(Xr, method = fake_method))
})

context("test-pls_projection")
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# =============================================================================
# Setup helper
# =============================================================================

.setup_pls_data <- function(n_xr = 40, n_xu = 20) {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Yr_2 <- NIRsoil$Ciso[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  y_sel <- !is.na(Yr) & !is.na(Yr_2)
  Xr <- Xr[y_sel, ]
  
  Yu <- Yu[!is.na(Yu)]
  Yr_2 <- Yr_2[y_sel]
  Yr <- Yr[y_sel]
  
  Xu <- Xu[seq_len(n_xu), ]
  Yu <- Yu[seq_len(n_xu)]
  Xr <- Xr[seq_len(n_xr), ]
  Yr <- Yr[seq_len(n_xr)]
  Yr_2 <- Yr_2[seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yr_2 = Yr_2, Yu = Yu)
}


# =============================================================================
# PLS with multi-response Y tests
# =============================================================================

test_that("ortho_projection PLS with multi-response Y and ncomp_by_var works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  Yr <- NIRsoil[as.logical(NIRsoil$train), c("Ciso", "Nt")]
  
  # Remove rows with NA in either response
  complete <- complete.cases(Yr)
  Xr <- Xr[complete, ]
  Yr <- Yr[complete, ]
  
  result <- ortho_projection(
    Xr = Xr,
    Yr = as.matrix(Yr),
    ncomp = ncomp_by_var(0.01),
    method = "pls"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
})


test_that("ortho_projection PLS with multi-response Y and ncomp_by_cumvar works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  Yr <- NIRsoil[as.logical(NIRsoil$train), c("Ciso", "Nt")]
  
  complete <- complete.cases(Yr)
  Xr <- Xr[complete, ]
  Yr <- Yr[complete, ]
  
  result <- ortho_projection(
    Xr = Xr,
    Yr = as.matrix(Yr),
    ncomp = ncomp_by_cumvar(0.99),
    method = "pls"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
})


# =============================================================================
# Basic PLS projection tests
# =============================================================================

test_that("ortho_projection with method = 'pls' and ncomp_by_cumvar works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  cumvar_value <- 0.74
  
  result <- ortho_projection(
    Xr = d$Xr,
    Yr = d$Yr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
  
  # All but last component should have cumvar > threshold (for PLS, variance explained grows)
  test_ncomp <- result$ncomp - 1
  if (test_ncomp > 0) {
    expect_true(all(
      result$variance$x_var["cumulative_explained_var_X", seq_len(test_ncomp)] < cumvar_value
    ))
  }
})


test_that("ortho_projection PLS with two matrices works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  cumvar_value <- 0.74
  tol <- 1e-5
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
  expect_equal(nrow(result$scores), nrow(d$Xr) + nrow(d$Xu))
  
  # Predictions for Xr should match stored scores
  preds <- sum(abs(
    predict(result)[seq_len(nrow(d$Xr)), ] - predict(result, d$Xr)
  ))
  expect_lt(preds, tol)
})


test_that("ortho_projection PLS with ncomp_by_var works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  min_var <- 0.26  # 1 - 0.74
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_var(min_var),
    method = "pls",
    scale = FALSE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(ncol(result$scores), result$ncomp)
})


# =============================================================================
# OPC selection tests
# =============================================================================

test_that("ortho_projection PLS with ncomp_by_opc works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    method = "pls",
    scale = TRUE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_true("opc_evaluation" %in% names(result))
  
  # ncomp should match minimum RMSD position
  expect_equal(result$ncomp, as.vector(which.min(result$opc_evaluation[, "rmsd_Yr"])))
  expect_equal(result$ncomp, 8)
})


test_that("ortho_projection PLS variance selection consistency", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  cumvar_value <- 0.74
  
  # Get full variance info via OPC
  opc_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    method = "pls",
    scale = TRUE
  )
  
  cumvar_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  var_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_var(1 - cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  # Check consistency
  expect_equal(
    sum(opc_result$variance$x_var["cumulative_explained_var_X", ] < cumvar_value),
    cumvar_result$ncomp - 1
  )
  
  expect_equal(
    sum(opc_result$variance$x_var["explained_var_X", ] >= (1 - cumvar_value)),
    var_result$ncomp
  )
})


# =============================================================================
# PLS2 (multi-response) with OPC tests
# =============================================================================

test_that("ortho_projection PLS2 with ncomp_by_opc works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  tol <- 1e-6
  Yr_multi <- cbind(d$Yr, d$Yr_2)
  colnames(Yr_multi) <- c("Yr", "Yr_2")
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = Yr_multi,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    method = "pls",
    scale = TRUE
  )
  
  expect_lt(
    sum(abs(result$scores - predict(result, rbind(d$Xr, d$Xu)))), 
    tol,
    label = "proper pls projection of Xr should match stored scores"
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_true("opc_evaluation" %in% names(result))
  
  # Should select based on mean standardized RMSD
  expect_equal(
    result$ncomp, 
    as.vector(which.min(result$opc_evaluation[, "mean_standardized_rmsd_Yr"]))
  )
  
  # Verify RMSD calculation manually
  scores_xr <- result$scores[seq_len(nrow(d$Xr)), ]
  distm <- as.matrix(dist(scale(scores_xr, center = TRUE, scale = TRUE)))
  nn <- apply(distm, MARGIN = 2, FUN = function(x) order(x)[2])
  
  expected_rmsd <- round(
    colMeans((Yr_multi - Yr_multi[nn, ])^2)^0.5, 
    4
  )
  
  actual_rmsd <- round(
    result$opc_evaluation[result$ncomp, c("rmsd_Yr", "rmsd_Yr_2")], 
    4
  )
  
  expect_equal(as.vector(actual_rmsd), as.vector(expected_rmsd))
})


# =============================================================================
# SIMPLS tests
# =============================================================================

test_that("ortho_projection with method = 'simpls' works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  tol <- 1e-6
  d <- .setup_pls_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = ncomp_by_opc(max_ncomp = 15),
    method = "simpls",
    scale = TRUE
  )
  
  expect_lt(
    sum(abs(result$scores - predict(result, rbind(d$Xr, d$Xu)))), 
    tol,
    label = "proper simpls projection of Xr should match stored scores"
  )
  expect_s3_class(result, "ortho_projection")
  expect_equal(result$method, "simpls")
  expect_true("opc_evaluation" %in% names(result))
})


test_that("ortho_projection PLS and SIMPLS produce equivalent results", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()

  pls_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = 10,
    method = "pls",
    scale = TRUE
  )
  
  simpls_result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = 10,
    method = "simpls",
    scale = TRUE
  )
  
  # Predictions should match
  pls_pred <- predict(pls_result, d$Xu)
  simpls_pred <- predict(simpls_result, d$Xu)
  
  for (i in 1:ncol(pls_pred)) {
    expect_gt(
      cor(pls_pred[, i], simpls_pred[, i])^2,
      0.999,
      label = paste("SIMPLS vs NIPALS PLS: projected scores for factor", i, "should corrrelate")
    )
  }
})


# =============================================================================
# Modified PLS tests
# =============================================================================

test_that("ortho_projection with method = 'mpls' works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_pls_data()
  
  result <- ortho_projection(
    Xr = d$Xr, 
    Xu = d$Xu,
    Yr = d$Yr,
    ncomp = 10,
    method = "mpls",
    scale = TRUE
  )
  
  expect_s3_class(result, "ortho_projection")
  expect_equal(result$method, "mpls")
  expect_equal(result$ncomp, 10)
})


# =============================================================================
# Large dataset tests (skipped on CRAN)
# =============================================================================

test_that("ortho_projection PLS works with larger datasets", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  data("NIRsoil", package = "prospectr")
  tol <- 1e-5
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Yr_2 <- NIRsoil$Ciso[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  y_sel <- !is.na(Yr) & !is.na(Yr_2)
  Xr <- Xr[y_sel, ]
  Yu <- Yu[!is.na(Yu)]
  Yr_2 <- Yr_2[y_sel]
  Yr <- Yr[y_sel]
  
  cumvar_value <- 0.991
  
  # Single matrix
  one_input <- ortho_projection(
    Xr = Xr,
    Yr = Yr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    method = "simpls",
    scale = FALSE
  )

  expect_equal(ncol(one_input$scores), one_input$ncomp)
  test_ncomp <- one_input$ncomp - 1
  expect_true(all(
    one_input$variance$x_var["cumulative_explained_var_X", seq_len(test_ncomp)] < cumvar_value
  ))
  
  # Two matrices
  two_input <- ortho_projection(
    Xr = Xr, 
    Xu = Xu,
    Yr = Yr,
    ncomp = ncomp_by_cumvar(cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  expect_equal(ncol(two_input$scores), two_input$ncomp)
  
  # Predictions match
  preds <- sum(abs(
    predict(two_input)[seq_len(nrow(Xr)), ] - predict(two_input, Xr)
  ))
  expect_lt(preds, tol)
  
  # OPC selection
  opc_result <- ortho_projection(
    Xr = Xr, 
    Xu = Xu,
    Yr = Yr,
    ncomp = ncomp_by_opc(max_ncomp = 20),
    method = "pls",
    scale = FALSE
  )
  
  expect_equal(opc_result$ncomp, as.vector(which.min(opc_result$opc_evaluation[, "rmsd_Yr"])))
  expect_equal(opc_result$ncomp, 12)
  
  # Variance selection consistency
  var_result <- ortho_projection(
    Xr = Xr, 
    Xu = Xu,
    Yr = Yr,
    ncomp = ncomp_by_var(1 - cumvar_value),
    method = "pls",
    scale = FALSE
  )
  
  expect_equal(
    sum(opc_result$variance$x_var["cumulative_explained_var_X", ] < cumvar_value),
    two_input$ncomp - 1
  )
  
  expect_equal(
    sum(opc_result$variance$x_var["explained_var_X", ] >= (1 - cumvar_value)),
    var_result$ncomp
  )
  
  # PLS2
  Yr_multi <- cbind(Yr, Yr_2)
  colnames(Yr_multi) <- c("Yr", "Yr_2")
  
  pls2_result <- ortho_projection(
    Xr = Xr, 
    Xu = Xu,
    Yr = Yr_multi,
    ncomp = ncomp_by_opc(max_ncomp = 20),
    method = "pls",
    scale = FALSE
  )
  
  expect_equal(
    pls2_result$ncomp, 
    as.vector(which.min(pls2_result$opc_evaluation[, "mean_standardized_rmsd_Yr"]))
  )
  
  # Verify RMSD calculation
  scores_xr <- pls2_result$scores[seq_len(nrow(Xr)), ]
  distm <- as.matrix(dist(scale(scores_xr, center = TRUE, scale = TRUE)))
  nn <- apply(distm, MARGIN = 2, FUN = function(x) order(x)[2])
  
  expected_rmsd <- round(
    colMeans((Yr_multi - Yr_multi[nn, ])^2)^0.5, 
    4
  )
  
  actual_rmsd <- round(
    pls2_result$opc_evaluation[pls2_result$ncomp, c("rmsd_Yr", "rmsd_Yr_2")], 
    4
  )
  
  expect_equal(as.vector(actual_rmsd), as.vector(expected_rmsd))
})

context("test-ncomp_by-constructors")

# =============================================================================
# ncomp_* constructor tests
# =============================================================================

# test-constructors.R  
test_that("ncomp_by_var returns correct class", {
  x <- ncomp_by_var(0.01)
  expect_s3_class(x, "ncomp_by_var")
})

test_that("ncomp_by_var constructor works", {
  nc <- ncomp_by_var(0.01)
  expect_s3_class(nc, c("ncomp_by_var", "ncomp_selection"))
  expect_equal(nc$min_var, 0.01)
  expect_equal(nc$max_ncomp, 40L)
  
  nc2 <- ncomp_by_var(0.05, max_ncomp = 20)
  expect_equal(nc2$min_var, 0.05)
  expect_equal(nc2$max_ncomp, 20L)
})


test_that("ncomp_by_var validates inputs", {
  expect_error(ncomp_by_var(0), "\\(0, 1\\]")
  expect_error(ncomp_by_var(1.5), "\\(0, 1\\]")
  expect_error(ncomp_by_var("a"), "\\(0, 1\\]")
  expect_error(ncomp_by_var(0.01, max_ncomp = 0), "positive integer")
})


test_that("ncomp_by_cumvar constructor works", {
  nc <- ncomp_by_cumvar(0.99)
  expect_s3_class(nc, c("ncomp_by_cumvar", "ncomp_selection"))
  expect_equal(nc$min_cumvar, 0.99)
  expect_equal(nc$max_ncomp, 40L)
  
  nc2 <- ncomp_by_cumvar(0.95, max_ncomp = 50)
  expect_equal(nc2$min_cumvar, 0.95)
  expect_equal(nc2$max_ncomp, 50L)
})


test_that("ncomp_by_cumvar validates inputs", {
  expect_error(ncomp_by_cumvar(0), "\\(0, 1\\]")
  expect_error(ncomp_by_cumvar(1.5), "\\(0, 1\\]")
  expect_error(ncomp_by_cumvar(0.99, max_ncomp = -1), "positive integer")
})


test_that("ncomp_by_opc constructor works", {
  nc <- ncomp_by_opc()
  expect_s3_class(nc, c("ncomp_by_opc", "ncomp_selection"))
  expect_equal(nc$max_ncomp, 40L)
  
  nc2 <- ncomp_by_opc(max_ncomp = 30)
  expect_equal(nc2$max_ncomp, 30L)
})


test_that("ncomp_by_opc validates inputs", {
  expect_error(ncomp_by_opc(max_ncomp = 0), "positive integer")
  expect_error(ncomp_by_opc(max_ncomp = "a"), "positive integer")
})


test_that("ncomp_fixed constructor works", {
  nc <- ncomp_fixed(10)
  expect_s3_class(nc, c("ncomp_fixed", "ncomp_selection"))
  expect_equal(nc$ncomp, 10L)
})


test_that("ncomp_fixed validates inputs", {
  expect_error(ncomp_fixed(), "required")
  expect_error(ncomp_fixed(0), "positive integer")
  expect_error(ncomp_fixed(-5), "positive integer")
  expect_error(ncomp_fixed("a"), "positive integer")
})

#' @title Build a precomputed library of localised experts using memory-based learning
#' @name liblex
#' @aliases liblex predict.liblex plot.liblex
#' 
#' @description
#' Constructs a library of local predictive models based on memory-based 
#' learning (MBL). For each anchor observation, a local regression model is 
#' fitted using its nearest neighbors from the reference set. This 
#' implementation is based on the methods proposed in 
#' Ramirez-Lopez et al. (2026b).
#'
#' @usage
#' liblex(Xr, Yr, neighbors,
#'        diss_method = diss_pca(ncomp = ncomp_by_opc()),
#'        fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
#'        anchor_indices = NULL, gh = TRUE, group = NULL,
#'        control = liblex_control(), verbose = TRUE, ...)
#'
#' \method{predict}{liblex}(object, newdata, diss_method = NULL,
#'         weighting = c("gaussian", "tricube", "triweight", "triangular",
#'                       "quartic", "parabolic", "cauchy", "none"),
#'         adaptive_bandwidth = TRUE, reliability_weighting = TRUE,
#'         range_prediction_limits = FALSE, residual_cutoff = NULL,
#'         enforce_indices = NULL, probs = c(0.05, 0.25, 0.5, 0.75, 0.95),
#'         verbose = TRUE, allow_parallel = TRUE, blas_threads = 1L, ...)
#'         
#' \method{plot}{liblex}(x, ...)
#'
#' @param Xr A numeric matrix of predictor variables with dimensions `n × p`
#'   (observations in rows, variables in columns). Column names are required.
#' @param Yr A numeric vector or single-column matrix of length `n` containing
#'   the response variable values corresponding to `Xr`. Missing values are
#'   allowed (see Details).
#' @param neighbors A neighbor selection object specifying how to select
#'   neighbors. Use \code{\link{neighbors_k}()} for fixed k-nearest neighbors
#'   or \code{\link{neighbors_diss}()} for dissimilarity threshold-based
#'   selection.
#' @param diss_method For \code{liblex}: either a `diss_*` object specifying 
#'   the dissimilarity method, or a precomputed numeric dissimilarity matrix.
#'   If a `diss_*` object (e.g., [diss_pca()], [diss_pls()], [diss_correlation()],
#'   [diss_euclidean()]), dissimilarities are computed internally. Additional 
#'   parameters (centering, scaling, number of components) are controlled 
#'   within the object.
#'   If a matrix is provided:
#'   \itemize{
#'     \item When `anchor_indices = NULL`: must be square (`n × n`) with zeros 
#'       on the diagonal.
#'     \item When `anchor_indices` is specified (length `m`): must have 
#'       dimensions `n × m`, with `diss_method[anchor_indices, ]` having zeros 
#'       on the diagonal.
#'   }
#'   Default is `diss_pca(ncomp = ncomp_by_opc())`.
#'
#'   For \code{predict.liblex}: a dissimilarity method object created by 
#'   \code{diss_*()} constructors. If not provided, uses the method stored in
#'   \code{object}. Required if \code{object} was built with a precomputed
#'   dissimilarity matrix.
#' @param fit_method A `local_fit` object specifying the local regression 
#'   method. Currently supported: [fit_pls()] and [fit_wapls()]. 
#'   [fit_gpr()] is not yet supported. Default is 
#'   `fit_wapls(min_ncomp = 3, max_ncomp = 15)`.
#' @param anchor_indices An optional integer vector specifying row indices 
#'   of `Xr` around which to build local models. If `NULL` (default), models 
#'   are built for all observations. See Details.
#' @param gh Logical indicating whether to compute the GH distance 
#'   (Mahalanobis distance in PLS score space) for each anchor observation. 
#'   Default is `TRUE`.
#' @param group An optional factor assigning group labels to observations in 
#'   `Xr`. Used for leave-group-out validation to avoid pseudo-replication. 
#'   When an observation is selected for validation, all observations from 
#'   the same group are excluded from model fitting. Default is `NULL` 
#'   (each observation is its own group).
#' @param control A list of control parameters created by [liblex_control()].
#'   Default is `liblex_control()`.
#' @param object A fitted object of class \code{"liblex"}, created by
#'   \code{\link{liblex}} (for \code{predict}).
#' @param newdata A numeric matrix or data frame containing new predictor 
#'   values. Must include all predictors used in \code{object}.
#' @param weighting Character string specifying the kernel weighting function
#'   applied to neighbours when combining predictions. Options are:
#'   \code{"gaussian"} (default), \code{"tricube"}, \code{"triweight"},
#'   \code{"triangular"}, \code{"quartic"}, \code{"parabolic"}, \code{"cauchy"},
#'   or \code{"none"} (equal weights). See Details for kernel definitions.
#' @param adaptive_bandwidth Logical indicating whether to use adaptive 
#'   bandwidth for kernel weighting. When \code{TRUE} (default), the bandwidth 
#'   is set to the maximum dissimilarity within the neighborhood of each observation,
#'   so weights adapt to local density. When \code{FALSE}, a fixed global 
#'   bandwidth is used across all predictions, which may result in uniform 
#'   weights in sparse regions or overly concentrated weights in dense regions.
#' @param reliability_weighting Logical indicating whether to weight expert 
#'   predictions by their estimated reliability. When \code{TRUE} (default), 
#'   the contribution of each experrt is additionally weighted by the inverse of its 
#'   cross-validation residual variance, giving more influence to models that 
#'   performed well during fitting. When \code{FALSE}, only dissimilarity-based 
#'   kernel weights are used.
#' @param probs A numeric vector of probabilities in \eqn{[0, 1]} for computing
#'   weighted quantiles of expert predictions. Default is
#'   \code{c(0.05, 0.25, 0.5, 0.75, 0.95)}.
#' @param range_prediction_limits Logical. If \code{TRUE}, predictions falling
#'   outside the 5th–95th percentile range of neighbour response values are
#'   clipped to those limits. Default is \code{FALSE}.
#' @param residual_cutoff Numeric threshold for excluding models. Models with
#'   absolute residuals exceeding this value are penalized during neighbour
#'   selection. Default is \code{NULL} (no exclusion).
#' @param enforce_indices Optional integer vector specifying model indices that
#'   must always be included in each prediction neighborhood. These models are
#'   assigned the minimum dissimilarity of the neighborhood to ensure selection.
#'   Default is \code{NULL} (no enforced models).
#' @param allow_parallel Logical indicating whether parallel computation is
#'   permitted if a backend is registered. Default is \code{TRUE}.
#' @param verbose Logical indicating whether to display progress messages. 
#'   Default is `TRUE`.
#' @param blas_threads Integer specifying the number of BLAS threads to use.
#'   Default is \code{1L}. Requires the \pkg{RhpcBLASctl} package for thread
#'   control.
#' @param x An object of class `"liblex"` as returned by [liblex()].
#' @param ... Additional arguments (currently unused).
#' 
#' @details
#' By default, local models are constructed for all `n` observations in the
#' reference set. Alternatively, specify a subset of `m` observations 
#' (`m < n`) via `anchor_indices` to reduce computation.
#'
#' Each local model uses neighbors selected from the full reference set, but
#' models are only built for anchor observations. This is useful for large
#' datasets where building models for all observations is computationally
#' prohibitive.
#'
#' When dissimilarity methods depend on `Yr` (e.g., PLS-based distances), the
#' response values of anchor observations are excluded during dissimilarity
#' computation for efficiency. However, anchor response values are always
#' used when fitting local models.
#'
#' The number of anchors must not exceed 90% of `nrow(Xr)`; to build models 
#' for all observations, use `anchor_indices = NULL`.
#'
#' ## Relationship between anchors and neighborhood size
#'
#' The `neighbors` argument controls the neighborhood size (`k`) used both 
#' for fitting local models and for retrieving experts during prediction. 
#' When `anchor_indices` is specified, the number of available experts equals 
#' the number of anchors. If `max(k)` exceeds the number of anchors and 
#' tuning selects a large optimal `k`, prediction will retrieve fewer experts 
#' than specified. For reliable predictions, ensure the number of anchors is 
#' at least as large as the maximum `k` value being evaluated.
#' 
#' ## Missing values in Yr
#' 
#' Missing values in `Yr` are permitted. Observations with missing response
#' values can still serve as neighbors but are excluded from model fitting
#' as target observations.
#'
#' ## GH distance
#' 
#' The GH distance is computed independently from `diss_method` using a PLS
#' projection with optimized component selection. This provides a measure of
#' how far each observation lies from the center of the reference set in the
#' PLS score space.
#'
#' ## Validation and tuning
#' 
#' When `control$mode = "validate"` or `control$tune = TRUE`, nearest-neighbor
#' cross-validation is performed. For each anchor observation, its nearest
#' neighbor is excluded, a model is fitted on remaining neighbors, and the
#' excluded neighbor's response is predicted. This provides validation 
#' statistics for parameter selection.
#'
#' ## Prediction
#' 
#' For each observation in \code{newdata}, the \code{predict} method:
#' \enumerate{
#'   \item Computes dissimilarities to anchor observations (or their 
#'     neighbourhood centres) stored in \code{object}.
#'   \item Selects the \code{k} nearest neighbours based on the optimal 
#'     \code{k} determined during model fitting.
#'   \item Applies kernel weighting based on dissimilarity.
#'   \item Combines expert predictions using weighted averaging.
#' }
#'
#' ## Kernel weighting functions
#' 
#' The weighting functions follow Cleveland and Devlin (1988). Let \eqn{d} be 
#' the normalised dissimilarity (scaled to \eqn{[0, 1]} within the neighbourhood 
#' when \code{adaptive_bandwidth = TRUE}). The available kernels are:
#' \itemize{
#'   \item \code{"gaussian"}: \eqn{w = \exp(-d^2)}
#'   \item \code{"tricube"}: \eqn{w = (1 - d^3)^3}
#'   \item \code{"triweight"}: \eqn{w = (1 - d^2)^3}
#'   \item \code{"triangular"}: \eqn{w = 1 - d}
#'   \item \code{"quartic"}: \eqn{w = (1 - d^2)^2}
#'   \item \code{"parabolic"}: \eqn{w = 1 - d^2}
#'   \item \code{"cauchy"}: \eqn{w = 1 / (1 + d^2)}
#'   \item \code{"none"}: \eqn{w = 1} (equal weights)
#' }
#'
#' @return 
#' \strong{For \code{liblex}:} A list of class `"liblex"` (when 
#' `control$mode = "build"`) or `"liblex_validation"` (when 
#' `control$mode = "validate"`) containing:
#' \itemize{
#'   \item \code{dissimilarity}: List containing the dissimilarity method 
#'     and matrix.
#'   \item \code{fit_method}: Fit constructor from \code{fit_method}.
#'   \item \code{gh}: If `gh = TRUE`, a list with GH distances and the PLS 
#'     projection.
#'   \item \code{results}: Data frame of validation statistics for each 
#'     parameter combination (if validation was performed).
#'   \item \code{best}: The optimal parameter combination based on 
#'     `control$metric`.
#'   \item \code{optimal_params}: List with optimal `k` and `ncomp` values.
#'   \item \code{residuals}: Residuals from predictions using optimal 
#'     parameters.
#'   \item \code{coefficients}: (Build mode only) List of regression 
#'     coefficients: `B0` (intercepts), `B` (slopes).
#'   \item \code{vips}: (Build mode only) Variable importance in projection 
#'     scores.
#'   \item \code{selectivity_ratios}: (Build mode only) Selectivity ratios 
#'     for each predictor.
#'   \item \code{scaling}: (Build mode only) Centering and scaling vectors 
#'     for prediction.
#'   \item \code{neighborhood_stats}: Statistics (response quantiles) for 
#'     each neighborhood size.
#'   \item \code{anchor_indices}: The anchor indices used.
#'   \item \code{neighbors}: The object passed to \code{neighbors}.
#' }
#'
#' \strong{For \code{predict.liblex}:} A list with the following components:
#' \itemize{
#'   \item \code{predictions}: A data frame containing:
#'     \itemize{
#'       \item \code{pred}: Weighted mean predictions.
#'       \item \code{pred_sd}: Weighted standard deviation of expert 
#'         predictions.
#'       \item \code{q*}: Weighted quantiles at probabilities specified by 
#'         \code{probs}.
#'       \item \code{gh}: Global Mahalanobis distance (if computed during 
#'         fitting).
#'       \item \code{min_yr}: Minimum response value (5th percentile) across 
#'         neighbours.
#'       \item \code{max_yr}: Maximum response value (95th percentile) across 
#'         neighbours.
#'       \item \code{below_min}: Logical indicating prediction below 
#'         \code{min_yr}.
#'       \item \code{above_max}: Logical indicating prediction above 
#'         \code{max_yr}.
#'     }
#'   \item \code{neighbors}: A list with:
#'     \itemize{
#'       \item \code{indices}: Matrix of neighbour indices (models) for each 
#'         observation.
#'       \item \code{dissimilarities}: Matrix of corresponding dissimilarity 
#'         scores.
#'     }
#'   \item \code{expert_predictions}: A list with:
#'     \itemize{
#'       \item \code{weights}: Matrix of kernel weights applied to each expert.
#'       \item \code{predictions}: Matrix of raw predictions from each expert.
#'       \item \code{weighted}: Matrix of weighted predictions from each expert.
#'     }
#' }
#'
#' @references 
#' Cleveland, W. S., & Devlin, S. J. (1988). Locally weighted regression:
#' An approach to regression analysis by local fitting.
#' \emph{Journal of the American Statistical Association}, 83(403), 596–610.
#'
#' Naes, T., Isaksson, T., & Kowalski, B. (1990). Locally weighted regression
#' and scatter correction for near-infrared reflectance data.
#' \emph{Analytical Chemistry}, 62(7), 664–673.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. (2013). The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex datasets. Geoderma, 195-196, 
#' 268-279.
#' 
#' Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C.,
#' Perez-Fernandez, E., Plans, M., Breure, T., Behrens, T.,
#' Viscarra Rossel, R., & Peng, Y. (2026b). Rethinking local spectral
#' modelling: From per-query refitting to model libraries. 
#' \emph{Analytica Chimica Acta}, under review.
#'
#' Rajalahti, T., Arneberg, R., Berven, F.S., Myhr, K.M., Ulvik, R.J., 
#' Kvalheim, O.M. (2009). Biomarker discovery in mass spectral profiles by 
#' means of selectivity ratio plot. Chemometrics and Intelligent Laboratory 
#' Systems, 95(1), 35-48.
#'
#' @seealso 
#' [liblex_control()] for control parameters, [neighbors_k()] for neighborhood
#' specification, [diss_pca()], [diss_pls()], [diss_correlation()] for 
#' dissimilarity methods, [fit_pls()], [fit_wapls()] for fitting methods.
#'
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#' 
#' @examples
#' \dontrun{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess spectra
#' NIRsoil$spc_pr <- savitzkyGolay(
#'   detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
#'   m = 1, p = 1, w = 7
#' )
#'
#' # Missing values in the response are allowed
#' train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
#' train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
#' test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
#' test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
#'
#' # Build library
#' model_library <- liblex(
#'   Xr = train_x,
#'   Yr = train_y,
#'   neighbors = neighbors_k(c(30, 40)),
#'   diss_method = diss_correlation(ws = 27, scale = TRUE),
#'   fit_method = fit_wapls(
#'     min_ncomp = 4,
#'     max_ncomp = 17,
#'     scale = FALSE,
#'     method = "mpls"
#'   ),
#'   control = liblex_control(tune = TRUE)
#' )
#'
#' # Visualise neighborhood centroids and samples to predict
#' matplot(
#'   as.numeric(colnames(model_library$scaling$local_x_center)),
#'   t(test_x),
#'   col = rgb(1, 0, 0, 0.3),
#'   lty = 1,
#'   type = "l",
#'   xlab = "Wavelength (nm)",
#'   ylab = "First derivative detrended absorbance"
#' )
#' matlines(
#'   as.numeric(colnames(model_library$scaling$local_x_center)),
#'   t(model_library$scaling$local_x_center),
#'   col = rgb(0, 0, 1, 0.3),
#'   lty = 1,
#'   type = "l"
#' )
#' grid(lty = 1)
#' legend(
#'   "topright",
#'   legend = c("Samples to predict", "Neighborhood centroids"),
#'   col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
#'   lty = 1,
#'   lwd = 2,
#'   bty = "n"
#' )
#'
#' # Predict new observations
#' y_hat_liblex <- predict(model_library, test_x)
#'
#' # Predicted versus observed values
#' lims <- range(y_hat_liblex$predictions$pred, test_y, na.rm = TRUE)
#' plot(
#'   y_hat_liblex$predictions$pred,
#'   test_y,
#'   pch = 16,
#'   col = rgb(0, 0, 0, 0.5),
#'   xlab = "Predicted",
#'   ylab = "Observed",
#'   xlim = lims,
#'   ylim = lims
#' )
#' abline(a = 0, b = 1, col = "red")
#' grid(lty = 1)
#' 
#' ## run liblex in parallel (requires a parallel backend, e.g., doParallel)
#' library(doParallel)
#' n_cores <- min(2, parallel::detectCores() - 1)
#' clust <- makeCluster(n_cores)
#' registerDoParallel(clust)
#' 
#' model_library2 <- liblex(
#'   Xr = train_x,
#'   Yr = train_y,
#'   neighbors = neighbors_k(c(30, 40)),
#'   fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, method = "simpls")
#' )
#' 
#' y_hat_liblex2 <- predict(model_library2, test_x)
#' registerDoSEQ()
#' try(stopCluster(clust))
#' }
#' @export
######################################################################
# resemble
# Copyright (C) 2026 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################


liblex <- function(
    Xr, Yr, neighbors,
    diss_method = diss_pca(ncomp = ncomp_by_opc()),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
    anchor_indices = NULL,
    gh = TRUE, 
    group = NULL,
    control = liblex_control(),
    verbose = TRUE,
    ...
) {
  f_call <- match.call()
  
  # --- Xr validation ---
  if (!is.matrix(Xr)) {
    Xr <- as.matrix(Xr)
  }
  if (is.null(colnames(Xr))) {
    stop("'Xr' must have column names", call. = FALSE)
  }
  if (!is.numeric(Xr)) {
    stop("'Xr' must be numeric", call. = FALSE)
  }
  
  # --- Yr validation ---
  
  if (!is.null(ncol(Yr))) {
    if (ncol(Yr) != 1L) {
      stop("'Yr' must be a numeric vector or single-column matrix (only one response variable allowed at the moment)", call. = FALSE)
    }
  }
  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    stop("'verbose' must be a single TRUE or FALSE value", call. = FALSE)
  }
  if (length(Yr) != nrow(Xr)) {
    stop("'Yr' must have the same number of observations as 'Xr'", call. = FALSE)
  }
  if (!is.matrix(Yr)) {
    Yr <- matrix(Yr, nrow = nrow(Xr))
  }
  if (!is.numeric(Yr) || ncol(Yr) != 1L) {
    stop("'Yr' must be a numeric vector or single-column matrix", call. = FALSE)
  }
  if (nrow(Yr) != nrow(Xr)) {
    stop("'Yr' must have the same number of observations as 'Xr'", call. = FALSE)
  }
  
  # --- neighbors validation ---
  if (missing(neighbors)) {
    stop("'neighbors' is required", call. = FALSE)
  }
  if (!(inherits(neighbors, "neighbors_k") || inherits(neighbors, "neighbors_diss"))) {
    stop("'neighbors' must be a 'neighbors_k' object", call. = FALSE)
  }

  # --- diss_method validation ---
  if (is.matrix(diss_method)) {
    if (!is.numeric(diss_method)) {
      stop("'diss_method' matrix must be numeric", call. = FALSE)
    }
    if (nrow(diss_method) != ncol(diss_method) && is.null(anchor_indices)) {
      stop("'diss_method' matrix must be square", call. = FALSE)
    }
    if (nrow(diss_method) != nrow(Xr)) {
      stop("'diss_method' matrix dimensions must match nrow(Xr)", call. = FALSE)
    }
    precomputed_diss <- TRUE
  } else if (inherits(diss_method, "diss_method")) {
    precomputed_diss <- FALSE
  } else {
    stop("'diss_method' must be a diss_* object or a dissimilarity matrix",
         call. = FALSE)
  }
  
  # --- fit_method validation ---
  if (!inherits(fit_method, "fit_method")) {
    stop("'fit_method' must be a fit_* object (e.g., fit_wapls(), fit_pls())",
         call. = FALSE)
  }
  
  # --- anchor_indices validation ---
  if (is.null(anchor_indices)) {
    if (verbose) {
      message(
        "No 'anchor_indices' provided; building local models for all observations in 'Xr'."
      )
    }
  } else {
    if (!is.numeric(anchor_indices) || any(is.na(anchor_indices))) {
      stop("'anchor_indices' must be a numeric vector without NA", call. = FALSE)
    }
    anchor_indices <- as.integer(anchor_indices)
    if (any(anchor_indices < 1L) || any(anchor_indices > nrow(Xr))) {
      stop("'anchor_indices' must be between 1 and nrow(Xr)", call. = FALSE)
    }
    if (length(anchor_indices) > floor(nrow(Xr) * 0.90)) {
      stop("'anchor_indices' exceeds 90% of nrow(Xr); ",
           "set anchor_indices = NULL to build models for all observations",
           call. = FALSE)
    }
    if (any(duplicated(anchor_indices))) {
      stop("'anchor_indices' contains duplicate values", call. = FALSE)
    }
    maximum_k <- if (inherits(neighbors, "neighbors_k")) {
      max(neighbors$k)
    } else {
      neighbors$k_max
    }
    
    if (maximum_k > length(anchor_indices)) {
      warning(
        "Max. number of neighbors (", maximum_k, ") exceeds number of anchors (", 
        length(anchor_indices), "); prediction may retrieve fewer experts than ",
        "optimal if tuning selects a large k.",
        call. = FALSE
      )
    }
    
  }
  
  # --- gh validation ---
  if (!is.logical(gh) || length(gh) != 1L || is.na(gh)) {
    stop("'gh' must be TRUE or FALSE", call. = FALSE)
  }
  
  # --- control validation ---
  if (!inherits(control, "liblex_control")) {
    stop("'control' must be created by liblex_control()", call. = FALSE)
  }
  
  # --- chunk_size validation (from control) ---
  min_size <- if (!is.null(anchor_indices)) {
    min(nrow(Xr), length(anchor_indices))
  } else {
    nrow(Xr)
  }
  if (control$chunk_size > min_size) {
    stop("'chunk_size' in control cannot exceed ", min_size, call. = FALSE)
  }
  
  # --- verbose validation ---
  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    stop("'verbose' must be TRUE or FALSE", call. = FALSE)
  }
  

  if (!precomputed_diss && inherits(diss_method, c("diss_pca", "diss_pls"))) {
    diss_method$ncomp <- cap_ncomp_to_data(
      diss_method$ncomp,
      n_obs = nrow(Xr),
      n_vars = ncol(Xr),
      verbose = verbose
    )
  }
  
  # --- BLAS thread management ---
  if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
    old_blas_threads <- RhpcBLASctl::blas_get_num_procs()
    if (old_blas_threads != control$blas_threads) {
      RhpcBLASctl::blas_set_num_threads(control$blas_threads)
      on.exit(RhpcBLASctl::blas_set_num_threads(old_blas_threads), add = TRUE)
    }
  }
  
  # ---------------------------------------------------------------------------
  # Parallel setup
  # ---------------------------------------------------------------------------
  "%mydo%" <- get("%do%")
  if (control$allow_parallel && getDoParRegistered()) {
    "%mydo%" <- get("%dopar%")
  }
  
  if (is.matrix(diss_method)) {
    diss_method_type <- "Precomputed dissimilarity matrix"
    
    # --- Validate matrix ---
    if (!is.numeric(diss_method)) {
      stop("'diss_method' matrix must be numeric", call. = FALSE)
    }
    if (nrow(diss_method) != nrow(Xr)) {
      stop("'diss_method' must have nrow equal to nrow(Xr)", call. = FALSE)
    }

    # --- Validate dimensions based on anchor_indices ---
    if (!is.null(anchor_indices)) {
      if (ncol(diss_method) != length(anchor_indices)) {
        stop("'diss_method' must have ncol equal to length(anchor_indices)",
             call. = FALSE)
      }
      if (any(abs(diag(diss_method[anchor_indices, ])) > 1e-8)) {
        stop("'diss_method[anchor_indices, ]' must have zeros on the diagonal",
             call. = FALSE)
      }
      Yr_anchor <- Yr[anchor_indices, , drop = FALSE]
    } else {
      if (ncol(diss_method) != nrow(Xr)) {
        stop("'diss_method' must be square with dimensions equal to nrow(Xr)",
             call. = FALSE)
      }
      if (any(abs(diag(diss_method)) > 1e-8)) {
        stop("'diss_method' diagonal must contain only zeros", call. = FALSE)
      }
      Yr_anchor <- Yr
    }

    # --- Store and free original matrix ---
    dsm <- list(
      diss_method = diss_method_type,
      dissimilarity = diss_method
    )
    rm(diss_method)
    gc()
  } else if (inherits(diss_method, "diss_method")) {
    precomputed_diss <- FALSE
    diss_method_type <- class(diss_method)[1L]

    if (verbose) {
      cat("Computing dissimilarities... \n")
    }

    if (!is.null(anchor_indices)) {
      Yr_anchor <- Yr[anchor_indices, , drop = FALSE]
      if (sum(!is.na(Yr_anchor)) < 3L) {
        stop("At least 3 non-missing values in 'Yr' required for anchor indices",
             call. = FALSE)
      }
      
      # Create copy to avoid modifying user input
      diss_method_internal <- diss_method
      
      # Force projection if needed for GH computation
      if (gh || inherits(diss_method_internal, c("diss_pls")) || inherits(diss_method_internal, c("diss_pca"))) {
        diss_method_internal$return_projection <- TRUE
      }
      
      dsm <- dissimilarity(
        Xr = Xr[-anchor_indices, , drop = FALSE],
        Xu = Xr[anchor_indices, , drop = FALSE],
        diss_method = diss_method_internal,
        Yr = Yr[-anchor_indices, , drop = FALSE]
      )
      
      # Expand dissimilarity matrix to full Xr dimensions
      # (rows for anchor_indices were not computed, fill with NA)
      diss_full <- matrix(NA_real_, nrow(Xr), ncol(dsm$dissimilarity))
      diss_full[-anchor_indices, ] <- dsm$dissimilarity
      dsm$dissimilarity <- diss_full
      rm(diss_full)
      gc()

      # Compute dissimilarities among anchor observations
      if (inherits(diss_method, c("diss_pca", "diss_pls"))) {
        # For projection-based methods, use scores from the existing projection
        n_non_anchor <- nrow(Xr) - length(anchor_indices)
        z_anchor <- dsm$projection$scores[-(1:n_non_anchor), , drop = FALSE]
        z_anchor <- scale(
          z_anchor,
          center = FALSE,
          scale = dsm$projection$scores_sd
        )
        
        # Self-dissimilarity among anchor observations
        dsm$dissimilarity[anchor_indices, ] <- fast_self_euclid(z_anchor)
        rm(z_anchor)
        gc()

        # Reorder projection scores to match original Xr order
        real_order <- order(c(seq_len(nrow(Xr))[-anchor_indices], anchor_indices))
        dsm$projection$scores <- dsm$projection$scores[real_order, , drop = FALSE]
        rownames(dsm$projection$scores) <- paste0("Xr_", seq_len(nrow(Xr)))
        
      } else {
        # For non-projection methods: pre-scale using full Xr, then compute
        # dissimilarity with center/scale disabled
        
        X_anchor <- Xr[anchor_indices, , drop = FALSE]
        
        # Pre-scale using full Xr statistics
        do_center <- isTRUE(diss_method$center)
        do_scale <- isTRUE(diss_method$scale)
        
        if (do_center || do_scale) {
          X_anchor <- scale(
            X_anchor,
            center = if (do_center) get_column_means(Xr) else FALSE,
            scale = if (do_scale) get_column_sds(Xr) else FALSE
          )
        }
        
        # Create modified diss_method with center/scale disabled
        diss_method_anchor <- diss_method
        diss_method_anchor$center <- FALSE
        diss_method_anchor$scale <- FALSE
        
        dsm_anchor <- dissimilarity(
          Xr = X_anchor,
          Xu = NULL,
          diss_method = diss_method_anchor,
          Yr = Yr_anchor
        )
        
        dsm$dissimilarity[anchor_indices, ] <- dsm_anchor$dissimilarity
        rm(dsm_anchor, X_anchor, diss_method_anchor)
        gc()
      }
    } else {
      Yr_anchor <- Yr
      # Create copy to avoid modifying user input
      diss_method_internal <- diss_method
      
      # Force projection for projection-based methods
      if (inherits(diss_method_internal, c("diss_pca", "diss_pls"))) {
        diss_method_internal$return_projection <- TRUE
      }

      dsm <- dissimilarity(
        Xr = Xr,
        Xu = NULL,
        diss_method = diss_method_internal,
        Yr = Yr
      )
    }
    
    sml <- list(diss_method = diss_method)
    dsm <- append(sml, dsm)
  } else {
    stop(
      "'diss_method' must be created by a diss_*() function or be a ",
      "numeric dissimilarity matrix", call. = FALSE
    )
  }
  # --- Compute GH distances (independent of diss_method) ---
  if (gh) {
    # Compute GH distances independently via PLS projection
    # FIXME: this is not ideal, as it requires an additional projection step.
    # in addition the optimisation of components is hard coded
    gh_projection <- ortho_projection(
      Xr = Xr,
      Xu = NULL,
      Yr = Yr,
      ncomp = ncomp_by_opc(min(nrow(Xr), 40L)),
      method = "pls",
      scale = fit_method$scale
    )

    gh_center <- colMeans(gh_projection$scores)
    
    # Use all scores when anchor_indices is NULL
    scores_for_gh <- if (is.null(anchor_indices)) {
      gh_projection$scores
    } else {
      gh_projection$scores[anchor_indices, , drop = FALSE]
    }
    
    gh_all <- as.vector(f_diss(
      scores_for_gh,
      Xu = t(gh_center),
      diss_method = "mahalanobis",
      center = FALSE,
      scale = FALSE
    ))
    
    dsm$gh <- list(
      gh_Xr = gh_all,
      projection = gh_projection
    )
  } else {
    dsm$gh <- NULL
  }

  # Determine maximum neighborhood size
  if (inherits(neighbors, "neighbors_k")) {
    max_k <- max(neighbors$k)
  } else {
    # neighbors_diss
    max_k <- neighbors$k_max
  }

  # Validate against non-missing Yr values
  n_valid_yr <- sum(!is.na(Yr))
  if (n_valid_yr < max_k) {
    stop(
      "Maximum neighborhood size (", max_k, ") exceeds the number of ",
      "non-missing observations in 'Yr' (", n_valid_yr, ")",
      call. = FALSE
    )
  }
  
  if (inherits(neighbors, "neighbors_k")) {
    # kidxmat <- diss_to_neighbors(
    #   diss_matrix = dsm$dissimilarity,
    #   k = max_k,
    #   k_diss = NULL,
    #   k_range = NULL,
    #   spike = -which(is.na(Yr)),
    #   return_dissimilarity = FALSE,
    #   skip_first = FALSE,
    #   keep_self = TRUE
    # )
    k <- neighbors$k
    len_neighbors <- length(k)
    names_nnstats <- paste0(
      "neighbor ", 
      neighbors$method, 
      " index ", 
      seq_along(neighbors$k), 
      " (k = ", neighbors$k, ")"
    )
  }
  
  if (inherits(neighbors, "neighbors_diss")) {
    # kidxmat <- diss_to_neighbors(
    #   diss_matrix = dsm$dissimilarity,
    #   k = NULL,
    #   k_diss = neighbors$threshold,
    #   k_range = c(neighbors$k_min, neighbors$k_max),
    #   spike = -which(is.na(Yr)),
    #   return_dissimilarity = FALSE,
    #   skip_first = FALSE,
    #   keep_self = TRUE
    # )
    len_neighbors <- length(neighbors$threshold)
    
    format_threshold <- function(x, digits = 4) {
      natural <- format(x, scientific = FALSE, trim = TRUE, digits = 8)
      rounded <- formatC(x, format = "f", digits = digits)
      
      decimal_part <- sub("^[^.]*\\.?", "", natural)
      decimal_places <- nchar(decimal_part)
      
      if (decimal_places > digits) {
        paste0(rounded, "...")
      } else {
        natural
      }
    }
    
    names_nnstats <- paste0(
      "neighbor ", 
      neighbors$method, 
      " index ", 
      seq_along(neighbors$threshold), 
      " (diss < ", sapply(neighbors$threshold, format_threshold), ")"
    )
  }

  
  # 
  # 
  # kidxmat2 <- diss_to_neighbors(
  #   diss_matrix = dsm$dissimilarity,
  #   k = NULL, 
  #   k_diss = 0.45, 
  #   k_range = c(5, 50),
  #   spike = -which(is.na(Yr)),
  #   return_dissimilarity = FALSE,
  #   skip_first = FALSE, 
  #   keep_self = TRUE
  # )
  # 
  # sum(abs((kidxmat2$neighbors) - (kidxmat)))
  # 

  # Find indices of the nearest neighbors
  kidxmat <- list(
    neighbors = top_k_order(
      dsm$dissimilarity,
      k = max_k,
      skip = which(is.na(Yr))
    )
  )
  
  # Extract dissimilarities for the selected neighbors
  kidxmat$neighbors_diss <- extract_by_index(dsm$dissimilarity, kidxmat$neighbors)
  
  # kidxmat$neighbors_diss
  # kidxmat$neighbors
  
  # --- group validation ---  
  # For each anchor sample, identify neighbors not in the same group
  # (used for leave-group-out validation)
  if (is.null(group)) {
    group <- as.factor(seq_len(nrow(Xr)))
    kidxgrop <- matrix(
      TRUE, 
      nrow = nrow(kidxmat$neighbors), 
      ncol = ncol(kidxmat$neighbors)
    )
    kidxgrop[1L, ] <- FALSE
  } else {
    if (length(group) != nrow(Xr)) {
      stop("'group' must have length equal to nrow(Xr)", call. = FALSE)
    }
    group <- as.factor(group)
    kidxgrop <- not_in_same_group(kidxmat$neighbors, group = group)
  }
  
  # Compute nearest neighbor statistics (quantiles of Yr in neighborhoods)
  probs <- c(0, 0.05, 0.25, 0.5, 0.75, 0.95, 1.00)
  dnms <- list(
    seq_len(ncol(kidxgrop)),
    paste0(probs * 100, "%")
  )

  nnstats <- vector("list", len_neighbors)
  #######################################################################
  #######################################################################
  #######################################################################
  #######################################################################
  # FIXME: THE compute_nn_quantiles() FUNCTION NEEDS TO BE REFACTORED
  # IN A WAY THAT IT CAN TAKE THE DISSIMILARITY MATRIX AND TO EACH COLUMN IN 
  # kidxmat$neighbors, IT APPLIES THE THRESHOLD IN neighbors$threshold
  # AND THEN COMPUTES THE QUANTILES 
  # OR EVEN BETTER IT TAKES THE FOLLOWIN ks OBJECT:
  
  if (inherits(neighbors, "neighbors_diss")) {
    # The list of number of neighbors per each threshold dissimilarity
    # nrow(ks) is equivalent to the number of thresholds
    # ncol(ks) is equivalent to the observations
    ks <- apply(
      kidxmat$neighbors_diss, 
      MARGIN = 2, 
      FUN = function(x, thresholds) {
        ith_ks <- sapply(thresholds, FUN = function(d, thresholds) sum(x < thresholds, na.rm = TRUE), d = x)
        ith_ks
      }, 
      thresholds = neighbors$threshold
    )
    ks[ks < neighbors$k_min] <- neighbors$k_min
    ks[ks > neighbors$k_max] <- neighbors$k_max
    thresholds_ks <- neighbors$threshold
  } else {
    ks <- matrix(k, nrow = length(k), ncol = nrow(Xr)) 
    thresholds_ks <- neighbors$k
  }

  for (ii in seq_len(len_neighbors)) {
    nnstats[[ii]] <- compute_nn_quantiles(
      kidxmat = kidxmat$neighbors,
      kidxgrop = kidxgrop,
      Yr = Yr,
      k = ks[ii, ],
      probs = probs
    )
    dimnames(nnstats[[ii]]) <- dnms
  }
  names(nnstats) <- names_nnstats

  #######################################################################
  #######################################################################
  #######################################################################
  
  
  # RMSE/(q"75%" - q"25%")
  # Similar to the ratio of performance to inter-quartile distance (RPIQ)
  # Bellon-Maurel, V., Fernandez-Ahumada, E., Palagos, B., Roger, J.M., 
  # McBratney, A., 2010. Critical review of chemometric indicators commonly 
  # used for assessing the quality of the prediction of soil attributes by 
  # NIR spectroscopy. TrAC Trends in Analytical Chemistry, 29(9), pp.1073-1081.
  itq <- vapply(nnstats, function(x) {
    x[, "75%"] - x[, "25%"]
  }, numeric(nrow(nnstats[[1L]])))
  
  # Dissimilarity as predictors setup
  # (Note: diss_usage not yet implemented in liblex, placeholder for future)
  dssm <- NULL
  npredictors <- ncol(Xr)

  # Progress bar setup
  addit <- if (control$mode == "build") 1L else 0L
  if (verbose) {
    pb <- txtProgressBar(min = 0, max = len_neighbors + addit, char = "-")
  }
  
  # PLS component range from fit_method
  if (inherits(fit_method, "fit_wapls")) {
    min_ncomp <- fit_method$min_ncomp
    max_ncomp <- fit_method$max_ncomp
  } else if (inherits(fit_method, "fit_pls")) {
    min_ncomp <- fit_method$ncomp
    max_ncomp <- fit_method$ncomp
  } else {
    stop(
      "'fit_method' must be fit_pls() or fit_wapls(); fit_gpr() is not yet supported",
      call. = FALSE
    )
  }
  
  if (control$tune) {
    # Grid of possible min/max PLS component combinations for optimization
    sgrid <- expand.grid(minpls = min_ncomp:max_ncomp, maxpls = min_ncomp:max_ncomp)
    sgrid <- sgrid[sgrid$minpls <= sgrid$maxpls, ]
    rownames(sgrid) <- seq_len(nrow(sgrid))
    
    # Binary grid to switch components on/off for each combination
    emgrid <- t(
      sapply(
        seq_len(nrow(sgrid)),
        FUN = function(q, x, wv) {
          wv[x[q, 1L]:x[q, 2L]] <- 1L
          wv
        },
        x = sgrid,
        wv = rep(0L, max_ncomp)
      )
    )
    emgrid <- emgrid[, min_ncomp:max_ncomp, drop = FALSE]
  } else {
    emgrid <- matrix(1L, nrow = 1L, ncol = length(min_ncomp:max_ncomp))
    sgrid <- data.frame(
      minpls = min_ncomp,
      maxpls = max_ncomp
    )
  }
  
  if (verbose) cat("Fitting models...\n")
  ## perform the nearest neighbor predictions  
  if (control$mode == "validate" || control$tune) {
    nnpreds <- sapply(
      seq_len(len_neighbors),
      FUN = .get_all_fits,
      Xr = Xr,
      Yr = Yr,
      k = ks,
      ncomp_min = min_ncomp,
      ncomp_max = max_ncomp,
      emgrid = emgrid,
      scale = fit_method$scale,
      max_iter = fit_method$max_iter,
      tol = fit_method$tol,
      algorithm = fit_method$method,
      kidxmat = kidxmat$neighbors,
      kidxgrop = kidxgrop,
      dissimilarity_mat = dssm,
      pb = if (verbose) pb else NULL,
      chunk_size = control$chunk_size
    )
    # Organize the results (in nnpreds)
    # Compute prediction statistics: R², RMSE, bias, and RPIQ-like metric
    pparam <- matrix(NA_real_, nrow(emgrid), 4L)
    
    sstats <- function(y, yhat, itqk, pparam) {
      # Prediction errors (observed - predicted)
      me <- sweep(-yhat, MARGIN = 1L, STATS = y, FUN = "+", check.margin = FALSE)
      
      # R² (squared correlation)
      pparam[, 1L] <- cor(y, yhat, use = "complete.obs")^2
      
      # RMSE (root mean squared error)
      pparam[, 2L] <- colMeans(me^2, na.rm = TRUE)^0.5
      
      # Bias (mean error)
      pparam[, 3L] <- colMeans(me, na.rm = TRUE)
      
      
      # RMSE / IQR: similar to RPIQ (ratio of performance to inter-quartile distance)
      # See Bellon-Maurel et al. (2010)
      pparam[, 4L] <- colMeans(
        sweep(me,
              MARGIN = 1L,
              STATS = itqk,
              FUN = "/",
              check.margin = FALSE)^2,
        na.rm = TRUE
      )^0.5
      
      pparam
    }
    
    # Iterator function to subset columns from multiple matrices simultaneously
    # Used for parallel processing of prediction statistics
    isubset_cols <- function(...) {
      sargs <- names(match.call())[-1L]
      nextEl <- function(..ii..) {
        sapply(sargs, FUN = function(x) get(x)[, ..ii..], simplify = FALSE)
      }
      list(nextElem = nextEl)
    }
    
    # Create iterator for nnpreds and itq matrices
    # Iterator call
    itr <- isubset_cols(nnpreds = nnpreds, itq = itq)
    
    # Then in kpredstats:
    kpredstats <- function(..k.., itr, pparam, y) {
      ne <- itr$nextElem(..k..)
      
      sstats(
        y = y,
        yhat = t(matrix(ne$nnpreds, nrow(pparam))),
        itqk = ne$itq,
        pparam = pparam
      )
    }
    
    # Compute prediction statistics for each neighborhood size
    predperformance <- lapply(
      seq_len(ncol(nnpreds)),
      FUN = kpredstats,
      itr = itr,
      pparam = pparam,
      y = Yr_anchor
    )

    # Combine prediction statistics into a data.frame
    predperformance <- data.frame(do.call("rbind", predperformance))
    colnames(predperformance) <- c("r2", "rmse", "me", "st_rmse")
    
    # Add PLS component range and neighborhood size columns
    
    if (inherits(neighbors, "neighbors_k")) {
      predperformance <- data.frame(
        min_ncomp = rep(sgrid$minpls, times = length(thresholds_ks)),
        max_ncomp = rep(sgrid$maxpls, times = length(thresholds_ks)),
        k = rep(thresholds_ks, each = nrow(pparam)),
        predperformance
      )
      ks_param_name <- "k"
    } else {
      predperformance <- data.frame(
        min_ncomp = rep(sgrid$minpls, times = length(thresholds_ks)),
        max_ncomp = rep(sgrid$maxpls, times = length(thresholds_ks)),
        diss_threshold = rep(thresholds_ks, each = nrow(pparam)),
        k_min = rep(neighbors$k_min, each = nrow(pparam)),
        k_max = rep(neighbors$k_max, each = nrow(pparam)),
        predperformance
      )
      ks_param_name <- "diss_threshold"
    }
      
    
    # Find optimal parameters based on selected metric
    if (control$metric == "rmse") {
      bestp <- predperformance[which.min(predperformance[, "rmse"]), , drop = FALSE][1L, ]
    } else {
      # metric == "r2"
      bestp <- predperformance[which.max(predperformance[, "r2"]), , drop = FALSE][1L, ]
    }
    # Extract optimal parameters
    optimal_param <- bestp[[ks_param_name]]
    optimal_min_ncomp <- bestp$min_ncomp
    optimal_max_ncomp <- bestp$max_ncomp
    
    
    # Extract the vector of predictions corresponding to the best predictions
    # (optimal k, optimal pls range)
    pls_item_idx <- which(
      sgrid$minpls == optimal_min_ncomp & sgrid$maxpls == optimal_max_ncomp
    )
    pls_item_idx <- seq(pls_item_idx, by = nrow(pparam), length.out = length(Yr_anchor))
    
    best_preds <- itr$nextElem(which(thresholds_ks == optimal_param))$nnpreds[pls_item_idx]
    best_preds_residuals <- Yr_anchor - best_preds
    
  } else {
    # No validation: use maximum k and full PLS range
    optimal_param <- max(thresholds_ks)
    optimal_min_ncomp <- min_ncomp
    optimal_max_ncomp <- max_ncomp
    best_preds_residuals <- NULL
    bestp <- NULL
    predperformance <- NULL
    
    if (verbose) {
      setTxtProgressBar(pb, 1L)
    }
  }
  
  if (control$mode == "build") {
    
    # final vector of optimal ks
    optimal_ks <- ks[which(thresholds_ks == optimal_param), ]
    
    # Calculate number of variables for library storage
    n_var <- 1L + (5L * ncol(Xr))
    
    # Template matrix for storing PLS library coefficients
    plslib_template <- matrix(
      NA_real_,
      nrow = n_var,
      ncol = control$chunk_size
    )
    
    # Number of iterations for parallel processing
    n_iter <- ceiling(ncol(dsm$dissimilarity) / control$chunk_size)

    # Build PLS library in parallel (or sequentially if no backend registered)
    # Declare foreach variables to avoid R CMD check NOTE
    ksubsets <- ithbarrio <- iset <- NULL
    plslib <- foreach(
      i = seq_len(n_iter),
      .export = c(
        "ith_pred_subsets",
        "ith_subsets_list",
        "ith_subsets_by_group",
        ".get_all_fits",
        "ith_local_fit",
        "final_fits_cpp"
      ),
      ithbarrio = ith_subsets_list(
        x = Xr,
        y = Yr,
        kindx = kidxmat$neighbors[seq_len(max(optimal_ks)), , drop = FALSE],
        neighborhood_sizes = optimal_ks,
        D = dssm,
        chunk_size = control$chunk_size
      )
    ) %mydo% {
      iplslib <- plslib_template
      for (j in seq_along(ithbarrio)) {
        ij_pls <- final_fits_cpp(
          X = ithbarrio[[j]]$x,
          Y = ithbarrio[[j]]$y,
          new_x = ithbarrio[[j]]$x[1L, , drop = FALSE],
          ncomp_min = optimal_min_ncomp,
          ncomp_max = optimal_max_ncomp,
          scale = fit_method$scale,
          maxiter = fit_method$max_iter,
          tol = fit_method$tol, 
          algorithm = fit_method$method
        )
        iplslib[, j] <- unlist(ij_pls, recursive = FALSE, use.names = FALSE)
      }
      
      # Trim unused columns if last chunk is smaller
      if (j < ncol(iplslib)) {
        iplslib <- iplslib[, seq_len(j), drop = FALSE]
      }
      iplslib
    }

    # Combine parallel results
    plslib <- do.call("cbind", plslib)
    
    if (verbose) {
      setTxtProgressBar(pb, length(thresholds_ks) + 1L)
    }
    
    plslib <- t(plslib)
    
    # Setup column names based on whether dissimilarity is used as predictors
    namesk <- NULL
    npredictors <- ncol(Xr)

    # Extract scaling vectors from library matrix
    xscale <- plslib[, -seq_len((4L * npredictors) + 1L), drop = FALSE]
    plslib <- plslib[, seq_len((4L * npredictors) + 1L), drop = FALSE]
    
    xcenter <- plslib[, -seq_len((3L * npredictors) + 1L), drop = FALSE]
    plslib <- plslib[, seq_len((3L * npredictors) + 1L), drop = FALSE]
    
    # Extract VIPs and selectivity ratios
    plsvips <- plslib[, -c(seq_len(npredictors + 1L), 
                           (ncol(plslib) - npredictors + 1L):ncol(plslib)), 
                      drop = FALSE]
    plssratios <- plslib[, (ncol(plslib) - npredictors + 1L):ncol(plslib), drop = FALSE]
    plslib <- plslib[, seq_len(npredictors + 1L), drop = FALSE]
    
    # Assign column names
    colnames(plslib) <- c("b0", namesk, colnames(Xr))
    colnames(xcenter) <- colnames(xscale) <- c(namesk, colnames(Xr))
    colnames(plssratios) <- colnames(plsvips) <- c(namesk, colnames(Xr))
    
    # Extract regression coefficients
    bs <- list(
      B0 = plslib[, 1L],
      B = plslib[, colnames(plslib) %in% colnames(Xr), drop = FALSE]
    )
    
    
    # Global centering/scaling vectors for prediction
    # (used when applying the library to new observations)
    if (precomputed_diss) {
      global_center <- rep(0, ncol(Xr))
      global_scale <- rep(1, ncol(Xr))
    } else {
      global_center <- if (isTRUE(diss_method$center)) get_column_means(Xr) else rep(0, ncol(Xr))
      global_scale <- if (isTRUE(diss_method$scale)) get_column_sds(Xr) else rep(1, ncol(Xr))
    }
    
    # Scaling information for prediction
    xcenter <- xcenter[, colnames(xcenter) %in% colnames(Xr), drop = FALSE]
    xscale <- xscale[, colnames(xscale) %in% colnames(Xr), drop = FALSE]
    iscale <- list(
      center = global_center,
      scale = global_scale,
      # Adjust local centers by local scales for prediction
      # our pls first scales and then centers, so we need to adjust the 
      # local centers accordingly
      local_x_center = xcenter * xscale, 
      local_x_scale = xscale
    )
    
    param_name <- if (inherits(neighbors, "neighbors_k")) "k" else "diss_threshold"
    optimal_params <- c(
      setNames(list(optimal_param), param_name),
      list(ncomp = c(min = optimal_min_ncomp, max = optimal_max_ncomp))
    )
    
    fresults <- list(
      dissimilarity = dsm[!names(dsm) %in% "gh"],
      fit_method = fit_method,
      gh = dsm$gh,
      results = predperformance,
      best = bestp,
      optimal_params = optimal_params,
      
      
      residuals = best_preds_residuals,
      coefficients = bs,
      vips = plsvips,
      selectivity_ratios = plssratios,
      scaling = iscale,
      neighborhood_stats = nnstats
    )
    
    # Row names for output
    if (!is.null(rownames(Xr))) {
      row_names <- rownames(Xr)
    } else {
      row_names <- seq_len(nrow(Xr))
    }
    
    if (!is.null(anchor_indices)) {
      row_names <- row_names[anchor_indices]
    }
    
    # Assign row names to nearest neighbor statistics
    fresults$neighborhood_stats <- lapply(
      fresults$neighborhood_stats,
      FUN = function(x, nms) {
        rownames(x) <- nms
        x
      },
      nms = row_names
    )
    # Assign row names to library components
    rownames(fresults$vips) <-
      rownames(fresults$selectivity_ratios) <-
      rownames(fresults$coefficients$B) <-
      names(fresults$coefficients$B0) <- row_names
    
    if (!is.null(fresults$coefficients$Bk)) {
      rownames(fresults$coefficients$Bk) <- row_names
    }
  } else {
    
    # Validation only (no library built)
    fresults <- list(
      dissimmilarity = dsm[!names(dsm) %in% "gh"],
      fit_method = fit_method,
      gh = dsm$gh,
      results = predperformance,
      best = bestp,
      residuals = best_preds_residuals,
      neighborhood_stats = nnstats
    )
    
    # Row names for output
    if (!is.null(rownames(Xr))) {
      row_names <- rownames(Xr)
    } else {
      row_names <- seq_len(nrow(Xr))
    }
    
    if (!is.null(anchor_indices)) {
      row_names <- row_names[anchor_indices]
    }
    
    # Assign row names to nearest neighbor statistics
    fresults$neighborhood_stats <- lapply(
      fresults$neighborhood_stats,
      FUN = function(x, nms) {
        rownames(x) <- nms
        x
      },
      nms = row_names
    )
  }
  
  if (verbose) close(pb)
  
  fresults$anchor_indices <- if (!is.null(anchor_indices)) anchor_indices else seq_len(nrow(Xr))
  
  if (control$mode == "validate" || control$tune) {
    names(fresults$residuals) <- row_names
  }
  
  # add the original object passed to neighbors
  fresults$neighbors <- neighbors
  
  attr(fresults, "call") <- f_call
  class(fresults) <- c("liblex", "list")
  fresults
}



#' @aliases liblex
#' @export
predict.liblex <- function(
    object,
    newdata,
    diss_method = NULL,
    weighting = c(
      "gaussian", "tricube", "triweight", "triangular",
      "quartic", "parabolic", "cauchy", "none"
    ),
    adaptive_bandwidth = TRUE,
    reliability_weighting = TRUE,
    range_prediction_limits = FALSE,
    residual_cutoff = NULL,
    enforce_indices = NULL,
    probs = c(0.05, 0.25, 0.5, 0.75, 0.95),
    verbose = TRUE,
    allow_parallel = TRUE,
    blas_threads = 1L,
    ...
) {
  
  # --- Validate allow_parallel ---
  if (!is.logical(allow_parallel) || length(allow_parallel) != 1L) {
    stop("'allow_parallel' must be TRUE or FALSE", call. = FALSE)
  }
  
  # ---------------------------------------------------------------------------
  # Parallel setup
  # ---------------------------------------------------------------------------
  "%mydo%" <- get("%do%")
  if (allow_parallel && getDoParRegistered()) {
    "%mydo%" <- get("%dopar%")
  }
  
  # --- Validate blas_threads ---
  if (!is.numeric(blas_threads) || length(blas_threads) != 1L || 
      blas_threads < 1L || blas_threads != as.integer(blas_threads)) {
    stop("'blas_threads' must be a positive integer", call. = FALSE)
  }
  blas_threads <- as.integer(blas_threads)

  if (requireNamespace("RhpcBLASctl", quietly = TRUE)) {
    old_blas_threads <- blas_get_num_procs()
    if (old_blas_threads != blas_threads) {
      blas_set_num_threads(blas_threads)
      on.exit(blas_set_num_threads(old_blas_threads), add = TRUE)
    }
  } else if (Sys.info()["sysname"] == "Linux" && blas_threads == 1L) {
    message(
      "Tip: Install 'RhpcBLASctl' for optimal performance on Linux:\n",
      "  install.packages('RhpcBLASctl')"
    )
  }
  
  # Validate weighting argument
  weighting <- match.arg(weighting)
  
  # --- Validate object ---
  if (!inherits(object, "liblex")) {
    stop("'object' must be of class 'liblex'", call. = FALSE)
  }
  
  if (is.null(object$coefficients)) {
    stop("Cannot predict: object was built with mode = 'validate'", call. = FALSE)
  }
  
  # --- Validate adaptive_bandwidth ---
  if (!is.logical(adaptive_bandwidth) || length(adaptive_bandwidth) != 1L || is.na(adaptive_bandwidth)) {
    stop("'adaptive_bandwidth' must be TRUE or FALSE", call. = FALSE)
  }
  
  # --- Validate reliability_weighting ---
  if (!is.logical(reliability_weighting) || length(reliability_weighting) != 1L || is.na(reliability_weighting)) {
    stop("'reliability_weighting' must be TRUE or FALSE", call. = FALSE)
  }
  
  # --- Validate probs ---
  if (!is.numeric(probs) || any(is.na(probs)) || any(probs < 0) || any(probs > 1)) {
    stop("'probs' must be a numeric vector with values in [0, 1]", call. = FALSE)
  }
  
  # --- Validate range_prediction_limits ---
  if (!is.logical(range_prediction_limits) || length(range_prediction_limits) != 1L || is.na(range_prediction_limits)) {
    stop("'range_prediction_limits' must be TRUE or FALSE", call. = FALSE)
  }
  
  # --- Validate residual_cutoff ---
  if (!is.null(residual_cutoff)) {
    if (!is.numeric(residual_cutoff) || length(residual_cutoff) != 1L || residual_cutoff <= 0) {
      stop("'residual_cutoff' must be a single positive number or NULL", call. = FALSE)
    }
  }
  
  # --- Validate enforce_indices ---
  if (!is.null(enforce_indices)) {
    if (!is.numeric(enforce_indices) || any(enforce_indices < 1L)) {
      stop("'enforce_indices' must be positive integers or NULL", call. = FALSE)
    }
    enforce_indices <- as.integer(enforce_indices)
    n_models <- nrow(object$coefficients$B)
    if (any(enforce_indices > n_models)) {
      stop("'enforce_indices' contains values exceeding number of models (", n_models, ")", 
           call. = FALSE)
    }
  }
  
  # --- Validate verbose ---
  if (!is.logical(verbose) || length(verbose) != 1L || is.na(verbose)) {
    stop("'verbose' must be TRUE or FALSE", call. = FALSE)
  }
 
  # --- Validate diss_method ---
  stored_diss <- object$dissimilarity$diss_method
  
  if (missing(diss_method)) {
    if (is.character(stored_diss) && stored_diss == "Precomputed dissimilarity matrix") {
      stop(
        "Model built with precomputed dissimilarity matrix; ",
        "'diss_method' argument is required for prediction",
        call. = FALSE
      )
    }
    diss_method <- stored_diss
  } else {
    # Validate user-provided diss_method
    if (!inherits(diss_method, "diss_method")) {
      stop("'diss_method' must be a diss_*() object", call. = FALSE)
    }
    # User provided diss_method — warn if different from stored
    if (inherits(stored_diss, "diss_method") && !identical(diss_method, stored_diss)) {
      warning(
        "Overriding stored 'diss_method' with user-supplied value",
        call. = FALSE
      )
    }
  }

  # --- Validate newdata ---
  required_vars <- colnames(object$coefficients$B)
  
  if (!all(required_vars %in% colnames(newdata))) {
    missing_vars <- setdiff(required_vars, colnames(newdata))
    stop(
      "Missing predictor variables in 'newdata': ",
      paste(missing_vars[seq_len(min(5L, length(missing_vars)))], collapse = ", "),
      if (length(missing_vars) > 5L) paste0(" ... and ", length(missing_vars) - 5L, " more"),
      call. = FALSE
    )
  }
  
  newdata <- as.matrix(newdata[, required_vars, drop = FALSE])
  
  # --- Compute dissimilarities between newdata and anchor observations ---
  if (inherits(diss_method, c("diss_pca", "diss_pls"))) {
    # Projection-based methods: compute distances in score space
    
    # Project newdata onto the stored projection
    scores_newdata <- predict(object$dissimilarity$projection, newdata)
    
    # Get scores for anchor observations
    scores_anchors <- object$dissimilarity$projection$scores[object$anchor_indices, , drop = FALSE]
    
    # Standardize using moments from all training scores (not just anchors)
    score_center <- get_column_means(object$dissimilarity$projection$scores)
    score_scale <- get_column_sds(object$dissimilarity$projection$scores)
    
    scores_anchors_scaled <- scale(scores_anchors, center = score_center, scale = score_scale)
    scores_newdata_scaled <- scale(scores_newdata, center = score_center, scale = score_scale)
    
    # Euclidean distance in standardized score space
    dsmxu <- dissimilarity(
      Xr = scores_anchors_scaled,
      Xu = scores_newdata_scaled,
      diss_method = diss_euclidean(center = FALSE, scale = FALSE)
    )
    
  } else {
    # Non-projection methods: compute distances in original variable space
    # Use stored neighborhood centers as reference points
    
    # Apply stored global centering/scaling
    # Scale the local centers and newdata using the global center and 
    # scaling vectors 
    local_centers_scaled <- scale(
      object$scaling$local_x_center,
      center = object$scaling$center,
      scale = object$scaling$scale
    )
    newdata_scaled <- scale(
      newdata,
      center = object$scaling$center,
      scale = object$scaling$scale
    )
    
    # Copy diss_method to avoid modifying the original object
    # Disable center/scale since already applied above
    diss_method_copy <- diss_method
    diss_method_copy$center <- FALSE
    diss_method_copy$scale <- FALSE
   
    dsmxu <- dissimilarity(
      Xr = local_centers_scaled,
      Xu = newdata_scaled,
      diss_method = diss_method_copy
    )
  }
  
  # --- Compute GH distance for newdata (if available) ---
  gh_newdata <- NULL
  if (!is.null(object$gh)) {
    # Project newdata onto the stored GH projection space
    gh_scores_newdata <- predict(object$gh$projection, newdata)
    
    # GH = distance from each observation to the training centroid
    gh_center <- colMeans(object$gh$projection$scores)
    
    gh_newdata <- as.vector(dissimilarity(
      Xr = gh_scores_newdata,
      Xu = t(gh_center),
      diss_method = diss_euclidean(center = TRUE, scale = TRUE)
    )$dissimilarity)
  }
  
  # --- Progress message for model retrieval ---
  if (verbose) {
    diss_type <- class(diss_method)[1L]
    
    if (inherits(diss_method, "diss_correlation")) {
      if (!is.null(diss_method$ws)) {
        cat("Retrieving models using correlation dissimilarity with window size", 
            diss_method$ws, "...\n")
      } else {
        cat("Retrieving models using correlation dissimilarity with full window...\n")
      }
    } else {
      cat("Retrieving models using", sub("^diss_", "", diss_type), "dissimilarity...\n")
    }
  }
  
  # --- Identify nearest neighbors for each new observation ---
  # Neighbors are sorted by dissimilarity (ascending order)
  
  k_min <- diss_threshold <- NULL
  if (inherits(object$neighbors, "neighbors_diss")) {
    k_min <- object$neighbors$k_min
    k_max <- object$neighbors$k_max
    diss_threshold <- object$optimal_params$diss_threshold
  } else {
    k_min <- k_max <- object$optimal_params$k    
  }
  
  if (k_max > nrow(dsmxu$dissimilarity)) {
    k_max <- nrow(dsmxu$dissimilarity)
    message("Number of available models is", nrow(dsmxu$dissimilarity)) 
  }
  
  if (k_min > nrow(dsmxu$dissimilarity)) {
    k_min <- nrow(dsmxu$dissimilarity)
  }
  
  neighbor_indices <- top_k_neighbors(
    D = dsmxu$dissimilarity,
    k_min = k_min,
    k_max = k_max,
    threshold = diss_threshold
  )
  
  # --- Filter models with high residuals (optional) ---
  # Models exceeding residual_cutoff are penalized with max dissimilarity
  # to prevent their selection as neighbors
  if (!is.null(object$residuals) && !is.null(residual_cutoff)) {
    abs_res <- abs(object$residuals)
    
    # Flag models with residuals exceeding cutoff
    high_residual_flag <- abs_res >= residual_cutoff
    high_residual_flag[is.na(high_residual_flag)] <- TRUE
    
    # Penalize high-residual models by assigning max dissimilarity
    neighbor_diss <- sapply(
      seq_along(neighbor_indices),
      FUN = function(col_idx, diss, nn, flag) {
        knns <- nn[[col_idx]]
        d <- diss[knns, col_idx]
        # Assign max dissimilarity to flagged models
        d[flag[knns]] <- max(d)
        d
      },
      diss = dsmxu$dissimilarity,
      nn = neighbor_indices,
      flag = high_residual_flag
    )
    
  } else {
    # No residual filtering: extract sorted dissimilarities for neighbors
    # neighbor_diss <- apply(
    #   dsmxu$dissimilarity,
    #   MARGIN = 2L,
    #   FUN = sort
    # )[seq_len(k_max), , drop = FALSE]
    
    neighbor_diss <- lapply(
      seq_along(neighbor_indices),
      function(i) dsmxu$dissimilarity[neighbor_indices[[i]], i]
    )
    
    high_residual_flag <- rep(FALSE, nrow(object$coefficients$B))
  }
  
  # --- Enforce specific models into all neighborhoods (optional) ---
  # Enforced models are prepended with minimal dissimilarity to ensure selection
  if (!is.null(enforce_indices)) {
    neighbor_indices <- lapply(neighbor_indices, function(x) c(enforce_indices, x))
    
    # Assign minimal dissimilarity to enforced models
    min_diss_per_obs <- lapply(
      neighbor_diss, 
      FUN = function(x, repetitions) rep(min(x), repetitions), 
      repetitions = length(enforce_indices)
    ) 
    # add the minimal dissimilarity for the enforced models to the existing 
    # neighbor dissimilarities
    neighbor_diss <- Map(c, min_diss_per_obs, neighbor_diss)
  }
  
  # --- Compute neighbor weights ---
  if (weighting == "none") {
    # Equal weights for all neighbors
    dweights <- lapply(neighbor_diss, function(x) rep(1 / length(x), length(x)))
  } else {
    # Normalize dissimilarities to [0, 1] range per observation
    dweights <- lapply(neighbor_diss, function(d) d / max(d))
    
    # Apply kernel weighting function
    sigma <- if (adaptive_bandwidth) NULL else 0.5
    dweights <- compute_pred_weights(dweights, weighting = weighting, sigma = sigma)
  }
  
  # --- Apply reliability weighting (based on model residuals) ---
  if (reliability_weighting && !is.null(object$residuals)) {
    # Extract residuals for neighbor models
    neighbor_residuals <- lapply(neighbor_indices, function(idx) object$residuals[idx])
    
    # Reliability: inverse of squared residual magnitude
    # Scale by median absolute residual to make dimensionless
    eps <- median(abs(object$residuals), na.rm = TRUE) * 0.1
    eps <- max(eps, .Machine$double.eps)  # ensure non-zero
    
    reliability <- lapply(neighbor_residuals, function(r) {
      rel <- 1 / (1 + (r / eps)^2)
      rel[is.na(rel)] <- 0.5  # models with missing residuals get neutral weight
      rel
    })
    
    # Combine distance and reliability weights
    dweights <- Map(`*`, dweights, reliability)
    
    # Re-normalize
    dweights <- lapply(dweights, function(w) {
      s <- sum(w, na.rm = TRUE)
      if (s == 0) s <- 1
      w / s
    })
  }
  
  # --- Prepare coefficient library and scaling parameters ---
  # Standard model without dissimilarity predictors
  coef_library <- cbind(
    object$coefficients$B0,
    object$coefficients$B
  )
  local_scale <- object$scaling$local_x_scale
  
  
  # --- Extract local prediction subsets for each new observation ---
  local_subsets <- ith_pred_subsets(
    plslib = coef_library,
    Xu = newdata,
    xunn = neighbor_indices,
    xscale = local_scale,
    dxrxu = NULL
  )
  
  # --- Compute predictions for each new observation ---
  if (verbose) {
    cat("Computing predictions...\n")
  }
  
  predictions_raw <- foreach(
    i = seq_len(nrow(newdata)),
    iset = local_subsets
  ) %mydo% {
    ith_pred_cpp(
      plslib = iset$iplslib,
      xscale = iset$ixscale,
      Xu = iset$ixu,
      dxrxu = iset$idxrxu
    )
  }
  
  # --- Combine predictions into matrix ---
  max_len <- max(lengths(predictions_raw))
  predictions_raw <- t(do.call("cbind", lapply(predictions_raw, function(x) {
    length(x) <- max_len
    x
  })))
  
  colnames(predictions_raw) <- paste0("expert_", seq_len(ncol(predictions_raw)))
  rownames(predictions_raw) <- rownames(newdata) %||% seq_len(nrow(newdata))
  
  
  
  max_len_weights <- max(lengths(dweights))
  dweights <- t(do.call("cbind", lapply(dweights, function(x) {
    length(x) <- max_len_weights
    x
  })))

  # --- Compute weighted predictions and uncertainty ---
  # Transpose weights to match predictions matrix layout
  rownames(dweights) <- rownames(predictions_raw)
  colnames(dweights) <- colnames(predictions_raw)
  
  # Weighted predictions per expert
  weighted_predictions <- predictions_raw * dweights
  
  # Weighted mean prediction per observation
  pred_mean <- rowSums(weighted_predictions, na.rm = TRUE)
  
  # Weighted standard deviation
  centered_dev <- sweep(predictions_raw, 1L, pred_mean, FUN = "-")^2
  pred_var <- rowSums(centered_dev * dweights, na.rm = TRUE)
  pred_sd <- sqrt(pred_var)

  # Weighted quantiles (exclude last column to avoid edge effects)
  pred_quantiles <- weighted_quantiles(
    predictions_raw,
    weights = dweights,
    probs = probs,
    exclude_last = TRUE
  )
  
  # Assemble predictions data frame
  predictions <- data.frame(
    pred = pred_mean,
    pred_sd = pred_sd,
    pred_quantiles
  )
  
  # --- Add GH distance to predictions ---
  predictions$gh <- gh_newdata
  
  # --- Preserve row names ---
  if (!is.null(rownames(newdata))) {
    rownames(predictions) <- rownames(newdata)
  }
  
  # --- Assemble output list ---
  result <- list(
    predictions = predictions,
    neighbors = list(
      indices = neighbor_indices,
      dissimilarities = neighbor_diss
    ),
    expert_predictions = list(
      weights = dweights,
      predictions = predictions_raw,
      weighted = weighted_predictions
    )
  )

  # --- Compute prediction limits from neighbor statistics ---
  
  if ( inherits(object$neighbors, "neighbors_diss") ) {
    nn_stats_indx <- which(object$neighbors$threshold == object$optimal_params[[1]])
  } else {
    nn_stats_indx <- which(object$neighbors$k == object$optimal_params[[1]])
  }
  
  neighborhood_stats <- object$neighborhood_stats[[nn_stats_indx]]

  # Extract min/max response values across neighbors for each observation
  min_yr <- sapply(
    seq_len(nrow(newdata)),
    FUN = function(i, stats, nn) min(stats[nn[[i]], "5%"]),
    stats = neighborhood_stats,
    nn = neighbor_indices
  )
  max_yr <- sapply(
    seq_len(nrow(newdata)),
    FUN = function(i, stats, nn) max(stats[nn[[i]], "95%"]),
    stats = neighborhood_stats,
    nn = neighbor_indices
  )
  
  result$predictions$min_yr <- min_yr
  result$predictions$max_yr <- max_yr
  
  # Flag predictions outside neighborhood range
  result$predictions$below_min <- result$predictions$pred < min_yr
  result$predictions$above_max <- result$predictions$pred > max_yr
  
  # Clip predictions to neighborhood range (optional)
  if (range_prediction_limits) {
    below <- result$predictions$below_min
    above <- result$predictions$above_max
    result$predictions$pred[below] <- min_yr[below]
    result$predictions$pred[above] <- max_yr[above]
  }
  
  result
}


compute_pred_weights <- function(
    xudss, 
    weighting = "gaussian",
    sigma = NULL,
    gamma = 0.3, 
    normalize = TRUE
) {
  valid_methods <- c(
    "triangular", "quartic", "triweight", "tricube",
    "parabolic", "gaussian", "cauchy"
  )
  weighting <- match.arg(weighting, valid_methods)
  
  # Normalize distances per observation to [0, 1]
  dscaled <- lapply(xudss, function(d) {
    d <- d / max(d)
    pmin(pmax(d, 0), 1)
  })
  
  # Adaptive bandwidth: use median neighbor distance per observation
  if (weighting == "gaussian") {
    if (is.null(sigma)) {
      sigma <- vapply(xudss, function(d) median(d) * 0.5, numeric(1))
      sigma[sigma == 0] <- min(sigma[sigma > 0], na.rm = TRUE)
      sigma[!is.finite(sigma)] <- 0.5
    }
    single_sigma <- length(sigma) == 1L
  }
  
  # Compute kernel weights
  dweights <- switch(
    weighting,
    triangular = lapply(dscaled, function(d) 1 - d),
    parabolic  = lapply(dscaled, function(d) 1 - d^2),
    quartic    = lapply(dscaled, function(d) (1 - d^2)^2),
    triweight  = lapply(dscaled, function(d) (1 - d^2)^3),
    tricube    = lapply(dscaled, function(d) (1 - d^3)^3),
    gaussian   = {
      if (single_sigma) {
        s2 <- 2 * sigma^2
        lapply(dscaled, function(d) exp(-(d^2) / s2))
      } else {
        Map(function(d, s) exp(-(d^2) / (2 * s^2)), dscaled, sigma)
      }
    },
    cauchy = lapply(dscaled, function(d) 1 / (1 + (d / gamma)^2))
  )
  
  dweights <- lapply(dweights, function(w) {
    w[!is.finite(w)] <- 0
    pmax(w, 0)
  })
  
  if (normalize) {
    dweights <- lapply(dweights, function(w) {
      s <- sum(w, na.rm = TRUE)
      if (s == 0) s <- 1
      w / s
    })
  }
  
  attr(dweights, "weighting") <- weighting
  attr(dweights, "sigma") <- sigma
  dweights
}


#' @aliases liblex
#' @export
plot.liblex <- function(x, ...) {
  if (!inherits(x, "liblex")) {
    stop("'x' must be an object of class 'liblex'", call. = FALSE)
  }
  
  
  # Set up 2-panel layout
  
  old_par <- par(no.readonly = TRUE)
  on.exit(par(old_par), add = TRUE)
  nfrows <- (!is.null(x$results)) + (!is.null(x$scaling$local_x_center))
  if (nfrows == 0) {
    stop("Nothing to plot: no results or neighborhood centroids available", call. = FALSE)
  } else {
    par(mfrow = c(nfrows, 1), mar = c(4, 4, 3, 1))
  }
  
    
  # --- Panel 1: Best RMSE per neighborhood parameter ---
  if (!is.null(x$results)) {
    bresult <- NULL

    if (inherits(x$neighbors, "neighbors_k")) {
      for (k in unique(x$results$k)) {
        kth_r <- x$results[x$results$k == k, , drop = FALSE]
        kth_r <- kth_r[which.min(kth_r$rmse), , drop = FALSE]
        bresult <- rbind(bresult, kth_r)
      }
      param <- "k"
      param_name <- "Number of neighbors (k)"
    } else {
      for (thr in unique(x$results$diss_threshold)) {
        kth_r <- x$results[x$results$diss_threshold == thr, , drop = FALSE]
        kth_r <- kth_r[which.min(kth_r$rmse), , drop = FALSE]
        bresult <- rbind(bresult, kth_r)
      }
      param <- "diss_threshold"
      param_name <- "Dissimilarity threshold"
    }
    
    plot(
      bresult[[param]],
      bresult$rmse,
      type = "b",
      pch = 16,
      col = "dodgerblue",
      xlab = param_name,
      ylab = "RMSE",
      main = "Best RMSE at each neighbor selection parameter"
    )
    grid(lty = 1)
    
    # Mark optimal point
    opt_idx <- which.min(bresult$rmse)
    points(
      bresult[[param]][opt_idx],
      bresult$rmse[opt_idx],
      pch = 16,
      col = "red",
      cex = 1.5
    )
  } 
  
  # --- Panel 2: Neighborhood centroids ---
  if (!is.null(x$scaling$local_x_center)) {
    centroids <- x$scaling$local_x_center
    coords <- as.numeric(colnames(centroids))
    
    if (any(is.na(coords))) {
      coords <- seq_len(ncol(centroids))
      xlab_text <- "Variable index"
    } else {
      xlab_text <- "Spectral coordinate"
    }
    
    matplot(
      coords,
      t(centroids),
      col = rgb(0, 0, 1, 0.3),
      lty = 1,
      type = "l",
      xlab = xlab_text,
      ylab = "Intensity",
      main = paste0("Neighborhood centroids (n = ", nrow(centroids), ")")
    )
    grid(lty = 1)
  } 
  
  invisible(NULL)
}

# =============================================================================
# Neighbor selection constructors
# =============================================================================
# Public:
#   neighbors_k()
#   neighbors_diss()
#
# Internal:
#   .new_neighbors()     - shared constructor
#   print.neighbors_k
#   print.neighbors_diss
#   print.neighbors      - fallback


#' @title Neighbor selection methods
#'
#' @description
#' These functions create configuration objects that specify how neighbors
#' are selected for memory-based learning in \code{\link{mbl}} an \code{\link{liblex}}.
#'
#' @name neighbors
#' @aliases neighbors_k neighbors_diss
#'
#' @usage
#' neighbors_k(k)
#'
#' neighbors_diss(threshold, k_min = 4L, k_max = Inf)
#'
#' @param k Integer vector. One or more neighborhood sizes to evaluate.
#'   Values will be sorted in ascending order. Minimum allowed value is 4.
#' @param threshold Numeric vector. One or more dissimilarity thresholds.
#'   Neighbors are selected if their dissimilarity to the target observation
#'   is below the threshold. Values will be sorted in ascending order.
#' @param k_min Integer. Minimum number of neighbors to retain, regardless of
#'   threshold. Default \code{4L}.
#' @param k_max Integer or \code{Inf}. Maximum number of neighbors to retain,
#'   regardless of threshold. Default \code{Inf} (no upper bound other than
#'   the size of the reference set).
#'
#' @details
#' Two strategies are available for neighbor selection:
#'
#' \strong{Fixed-k selection} (\code{neighbors_k}) 
#' 
#' A fixed number of nearest neighbors is selected for each target observation. 
#' Multiple values of \code{k} can be provided to evaluate different 
#' neighborhood sizes.
#'
#' \strong{Dissimilarity-threshold selection} (\code{neighbors_diss}) 
#' 
#' Neighbors are selected based on a dissimilarity threshold. All reference 
#' observations with dissimilarity below the threshold are included. The 
#' \code{k_min} and \code{k_max} arguments provide bounds to ensure a reasonable 
#' neighborhood size regardless of the threshold. Multiple thresholds can be 
#' provided to evaluate different settings.
#'
#' @return
#' An object of class \code{c("neighbors_k", "neighbors")} or
#' \code{c("neighbors_diss", "neighbors")}, containing the validated
#' parameters. Intended to be passed to \code{\link{mbl}}.
#'
#' @seealso \code{\link{mbl}}
#'
#' @examples
#' # Fixed neighborhood sizes
#' neighbors_k(k = 50)
#' neighbors_k(k = c(40, 60, 80, 100, 120))
#'
#' # Dissimilarity threshold with default bounds
#' neighbors_diss(threshold = 0.3)
#'
#' # Dissimilarity threshold with custom bounds
#' neighbors_diss(threshold = c(0.1, 0.2, 0.3), k_min = 10, k_max = 150)
#'
#' @export
neighbors_k <- function(k) {
  
  if (missing(k) || is.null(k)) {
    stop("'k' is required.", call. = FALSE)
  }
  
  if (!is.numeric(k)) {
    stop("'k' must be a numeric vector.", call. = FALSE)
  }
  
  if (any(k != as.integer(k))) {
    stop("'k' must contain integer values.", call. = FALSE)
  }
  
  k <- unique(sort(as.integer(k)))
  
  if (any(k < 4L)) {
    stop("All values in 'k' must be at least 4.", call. = FALSE)
  }
  
  structure(
    list(
      method = "k",
      k = k
    ),
    class = c("neighbors_k", "neighbors")
  )
}


#' @export
neighbors_diss <- function(threshold, k_min = 4L, k_max = Inf) {
  
  if (missing(threshold) || is.null(threshold)) {
    stop("'threshold' is required.", call. = FALSE)
  }
  
  if (!is.numeric(threshold)) {
    stop("'threshold' must be a numeric vector.", call. = FALSE)
  }
  
  if (any(threshold <= 0)) {
    stop("All values in 'threshold' must be positive.", call. = FALSE)
  }
  
  threshold <- unique(sort(threshold))
  
  if (!is.numeric(k_min) || length(k_min) != 1L) {
    stop("'k_min' must be a single integer.", call. = FALSE)
  }
  
  k_min <- as.integer(k_min)
  
  if (k_min < 4L) {
    stop("'k_min' must be at least 4.", call. = FALSE)
  }
  
  if (!is.numeric(k_max) || length(k_max) != 1L) {
    stop("'k_max' must be a single integer or Inf.", call. = FALSE)
  }
  
  if (is.finite(k_max)) {
    k_max <- as.integer(k_max)
  }
  
  if (k_max <= k_min) {
    stop("'k_max' must be greater than 'k_min'.", call. = FALSE)
  }
  
  structure(
    list(
      method = "diss",
      threshold = threshold,
      k_min = k_min,
      k_max = k_max
    ),
    class = c("neighbors_diss", "neighbors")
  )
}


# -----------------------------------------------------------------------------
# Print methods
# -----------------------------------------------------------------------------

#' @export
print.neighbors_k <- function(x, ...) {
  k_str <- if (length(x$k) <= 6L) {
    paste(x$k, collapse = ", ")
  } else {
    paste0(
      paste(x$k[1:3], collapse = ", "),
      ", ..., ",
      paste(x$k[(length(x$k) - 1L):length(x$k)], collapse = ", ")
    )
  }
  cat(
    "Neighbor selection: fixed k\n",
    " k :", k_str, "\n"
  )
  invisible(x)
}


#' @export
print.neighbors_diss <- function(x, ...) {
  thr_str <- if (length(x$threshold) <= 6L) {
    paste(x$threshold, collapse = ", ")
  } else {
    paste0(
      paste(x$threshold[1:3], collapse = ", "),
      ", ..., ",
      paste(x$threshold[(length(x$threshold) - 1L):length(x$threshold)], collapse = ", ")
    )
  }
  k_max_str <- if (is.infinite(x$k_max)) "Inf" else x$k_max
  cat(
    "Neighbor selection: dissimilarity threshold\n",
    "  threshold :", thr_str, "\n",
    "  k_min     :", x$k_min, "\n",
    "  k_max     :", k_max_str, "\n"
  )
  invisible(x)
}


#' @export
print.neighbors <- function(x, ...) {
  cat("Neighbor selection method:", x$method, "\n")
  for (nm in setdiff(names(x), "method")) {
    cat(" ", nm, ":", x[[nm]], "\n")
  }
  invisible(x)
}

context("test-neighbor-constructors")


test_that("neighbors_k validates input", {
  expect_error(neighbors_k(-1))
  expect_s3_class(neighbors_k(10), "neighbors_k")
})


test_that("neighbors_diss validates input", {
  expect_error(neighbors_diss(-1, 50, 100))
  expect_s3_class(neighbors_diss(10), "neighbors_diss")
})


# tests/testthat/test-neighbors-print.R
test_that("print.neighbors_k prints compact output for short k", {
  x <- neighbors_k(c(4, 6, 8))
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection: fixed k", "  k : 4, 6, 8 " 
    )
  )
  expect_identical(ret, x)
})

test_that("print.neighbors_k truncates long k vectors", {
  x <- neighbors_k(c(4, 5, 6, 7, 8, 9, 10))
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection: fixed k", 
      "  k : 4, 5, 6, ..., 9, 10 " 
    )
  )
  expect_identical(ret, x)
})

test_that("print.neighbors_diss prints finite k_max", {
  x <- neighbors_diss(
    threshold = c(0.1, 0.2, 0.3),
    k_min = 4L,
    k_max = 12L
  )
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection: dissimilarity threshold", 
      "   threshold : 0.1, 0.2, 0.3 ",               
      "   k_min     : 4 ", 
      "   k_max     : 12 "        
    )
  )
  expect_identical(ret, x)
})

test_that("print.neighbors_diss truncates long threshold vectors", {
  x <- neighbors_diss(
    threshold = c(0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7),
    k_min = 4L,
    k_max = 10L
  )
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection: dissimilarity threshold", 
      "   threshold : 0.1, 0.2, 0.3, ..., 0.6, 0.7 ",
      "   k_min     : 4 ", 
      "   k_max     : 10 "  
    )
  )
  expect_identical(ret, x)
})

test_that("print.neighbors_diss prints Inf k_max", {
  x <- neighbors_diss(
    threshold = c(0.1, 0.2),
    k_min = 4L,
    k_max = Inf
  )
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection: dissimilarity threshold", 
      "   threshold : 0.1, 0.2 ",
      "   k_min     : 4 ", 
      "   k_max     : Inf "
    )
  )
  expect_identical(ret, x)
})

test_that("print.neighbors prints all fields for generic neighbors object", {
  x <- structure(
    list(
      method = "custom",
      alpha = 0.25,
      beta = 7L
    ),
    class = "neighbors"
  )
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection method: custom ",
      "  alpha : 0.25 ",
      "  beta : 7 "
    )
  )
  expect_identical(ret, x)
})
test_that("print.neighbors handles vector fields in generic neighbors objects", {
  x <- structure(
    list(
      method = "custom",
      ids = c(4L, 5L, 6L)
    ),
    class = "neighbors"
  )
  
  out <- capture.output(ret <- print(x))
  
  expect_identical(
    out,
    c(
      "Neighbor selection method: custom ",
      "  ids : 4 5 6 "
    )
  )
  expect_identical(ret, x)
})

context("test-search_neighbors")
# tests/testthat/test-search_neighbors.R
library(foreach)
library(RhpcBLASctl)
registerDoSEQ()
# =============================================================================
# Setup helper
# =============================================================================

.setup_search_neighbors_data <- function() {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ]
  Xr <- Xr[!is.na(Yr), ]
  Yu <- Yu[!is.na(Yu)]
  Yr <- Yr[!is.na(Yr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}


# =============================================================================
# Constructor validation tests
# =============================================================================

test_that("neighbors_k constructor works", {
  n1 <- neighbors_k(40)
  expect_s3_class(n1, "neighbors_k")
  expect_s3_class(n1, "neighbors")
  expect_equal(n1$k, 40L)
  
  # Multiple k values
  n2 <- neighbors_k(c(10, 20, 30))
  expect_equal(n2$k, c(10L, 20L, 30L))
  
  # Validation: k must be positive
  expect_error(neighbors_k(0))
  expect_error(neighbors_k(-5))
})


test_that("neighbors_diss constructor works", {
  n1 <- neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  expect_s3_class(n1, "neighbors_diss")
  expect_s3_class(n1, "neighbors")
  expect_equal(n1$threshold, 0.1)
  expect_equal(n1$k_min, 10L)
  expect_equal(n1$k_max, 50L)
  
  # Default k_max is Inf
  n2 <- neighbors_diss(threshold = 0.5)
  expect_equal(n2$k_max, Inf)
})


# =============================================================================
# Legacy API rejection tests
# =============================================================================

test_that("search_neighbors rejects legacy arguments", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  # Reject old k argument
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(),
      k = 50,
      neighbors = neighbors_k(50)
    ),
    "k.*k_diss.*k_range.*removed"
  )
  
  # Reject old pc_selection argument
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(),
      neighbors = neighbors_k(50),
      pc_selection = list("manual", 10)
    ),
    "pc_selection.*removed"
  )
  
  # Reject old center/scale arguments
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(),
      neighbors = neighbors_k(50),
      center = TRUE
    ),
    "center.*scale.*removed"
  )
  
  # Reject character-based diss_method
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = "pca",
      neighbors = neighbors_k(50)
    ),
    "Character-based.*no longer supported"
  )
})


test_that("search_neighbors requires neighbors argument", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca()
    ),
    "neighbors.*required"
  )
})


# =============================================================================
# Basic functionality tests
# =============================================================================

test_that("search_neighbors with diss_pca and neighbors_k works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10, method = "pca_nipals"),
    neighbors = neighbors_k(50)
  )
  
  expect_type(nn, "list")
  expect_true("neighbors" %in% names(nn))
  expect_true("neighbors_diss" %in% names(nn))
  expect_true("unique_neighbors" %in% names(nn))
  expect_equal(nrow(nn$neighbors), 50)
  expect_equal(ncol(nn$neighbors), nrow(d$Xu))
})


test_that("search_neighbors with diss_pca and neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10, method = "pca_nipals"),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  
  expect_type(nn, "list")
  expect_true("neighbors" %in% names(nn))
  expect_true("k_diss_info" %in% names(nn))
})


test_that("search_neighbors with spike works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  spikes <- c(5, 10, 15)
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10, method = "pca_nipals"),
    neighbors = neighbors_k(50),
    spike = spikes
  )
  
  # Spiked observations should appear in all neighborhoods
  expect_true(all(rowMeans(nn$neighbors[seq_along(spikes), ]) == spikes))
})


test_that("search_neighbors with spike and neighbors_diss works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  spikes <- c(5, 10, 15)
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10, method = "pca_nipals"),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50),
    spike = spikes
  )
  
  # Spiked observations should appear in all neighborhoods
  expect_true(all(rowMeans(nn$neighbors[seq_along(spikes), ]) == spikes))
})


# =============================================================================
# Dissimilarity method tests
# =============================================================================

test_that("search_neighbors works with diss_pca (SVD)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10),
    Yr = d$Yr,
    neighbors = neighbors_k(10)
  )
  
  expect_type(nn, "list")
  expect_equal(nrow(nn$neighbors), 10)
})


test_that("search_neighbors works with diss_pls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pls(ncomp = 10),
    Yr = d$Yr,
    neighbors = neighbors_k(10)
  )
  
  expect_type(nn, "list")
  expect_equal(nrow(nn$neighbors), 10)
})


test_that("search_neighbors works with diss_correlation", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_correlation(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  
  expect_type(nn, "list")
})


test_that("search_neighbors works with diss_euclidean", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_euclidean(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  
  expect_type(nn, "list")
})


test_that("search_neighbors works with diss_cosine", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_cosine(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  
  expect_type(nn, "list")
})


# =============================================================================
# Return options tests
# =============================================================================

test_that("search_neighbors with return_dissimilarity works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(50),
    return_dissimilarity = TRUE
  )
  
  expect_true("dissimilarity" %in% names(nn))
})


test_that("search_neighbors with return_projection works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(
      ncomp = ncomp_by_opc(40),
      return_projection = TRUE
    ),
    Yr = d$Yr,
    neighbors = neighbors_k(50)
  )
  
  expect_true("projection" %in% names(nn))
})


# =============================================================================
# Expected results tests (skip on CRAN)
# =============================================================================

test_that("search_neighbors produces expected neighbor indices", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  # PCA-based
  nn_pca <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(10)
  )
  expected_sum_pca <- 17853
  expect_equal(sum(nn_pca$neighbors[1, ]), expected_sum_pca)
  
  # PLS-based
  nn_pls <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_pls(ncomp = 10),
    Yr = d$Yr,
    neighbors = neighbors_k(10)
  )
  expected_sum_pls <- 18410
  expect_equal(sum(nn_pls$neighbors[1, ]), expected_sum_pls)
  
  # Correlation-based
  nn_cor <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_correlation(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  expected_sum_cor <- 18036
  expect_equal(sum(nn_cor$neighbors[1, ]), expected_sum_cor)
  
  # Euclidean
  nn_euclid <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_euclidean(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  expected_sum_euclid <- 19586
  expect_equal(sum(nn_euclid$neighbors[1, ]), expected_sum_euclid)
  
  # Cosine
  nn_cosine <- search_neighbors(
    Xr = d$Xr, Xu = d$Xu,
    diss_method = diss_cosine(),
    neighbors = neighbors_diss(threshold = 0.1, k_min = 10, k_max = 50)
  )
  expected_sum_cosine <- 18910
  expect_equal(sum(nn_cosine$neighbors[1, ]), expected_sum_cosine)
})


# =============================================================================
# Edge cases and validation tests
# =============================================================================

test_that("search_neighbors validates k vs nrow(Xr)", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(ncomp = 10),
      neighbors = neighbors_k(nrow(d$Xr) + 10)
    ),
    "cannot exceed"
  )
})


test_that("search_neighbors validates spike indices", {
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  # Out of bounds
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(ncomp = 10),
      neighbors = neighbors_k(50),
      spike = c(1, nrow(d$Xr) + 100)
    ),
    "out-of-bounds"
  )
  
  # Contradictory indices
  expect_error(
    search_neighbors(
      Xr = d$Xr, Xu = d$Xu,
      diss_method = diss_pca(ncomp = 10),
      neighbors = neighbors_k(50),
      spike = c(5, -5)
    ),
    "contradictory"
  )
})


test_that("search_neighbors without Xu searches within Xr", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  d <- .setup_search_neighbors_data()
  
  nn <- search_neighbors(
    Xr = d$Xr,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(20)
  )
  
  expect_type(nn, "list")
  expect_equal(ncol(nn$neighbors), nrow(d$Xr))
  
  # Column names should be Xr* not Xu*
  expect_true(all(grepl("^Xr", colnames(nn$neighbors))))
})

#' @title Global spectral calibration model
#' @name model
#' @aliases model
#' @aliases predict.resemble_model
#' @description
#'
#' \loadmathjax
#'
#' Fits a global calibration model for spectral data using partial least
#' squares (PLS) or Gaussian process regression (GPR). Unlike
#' \code{\link{mbl}}, which builds local models for each prediction,
#' \code{model()} fits a single global model to the reference data.
#'
#' @usage
#' model(Xr, Yr, fit_method, control = model_control(), verbose = TRUE)
#'
#' \method{predict}{resemble_model}(object, newdata, ncomp = NULL, ...)
#'
#' @param Xr A numeric matrix of predictor variables, with observations in rows
#'   and variables in columns. Typically spectral data.
#' @param Yr A numeric matrix with one column containing the response variable
#'   values corresponding to the observations in \code{Xr}.
#' @param fit_method An object created by \code{\link{fit_pls}} or
#'   \code{\link{fit_gpr}} specifying the regression method and its parameters,
#'   including centring and scaling options.
#'   \code{fit_wapls()} is not supported for global models because it requires
#'   target observations to compute weights.
#' @param control A list created by \code{model_control()} specifying the
#'   cross-validation settings. The default is \code{model_control()}.
#' @param verbose Logical indicating whether progress information should be
#'   printed. Default is \code{TRUE}.
#'
#' @param object A \code{resemble_model} object returned by \code{\link{model}}.
#' @param newdata A numeric matrix of new observations with the same number of
#'   columns as the training data.
#' @param ncomp For PLS models, the number of components to use for prediction.
#'   The default is the number used during fitting. Ignored for GPR models. For 
#'   prediction, this can be a vector of integers representing the predictions
#'   coming from models with the requested components. 
#' @param ... Additional arguments, currently unused.
#'
#' @details
#'
#' \code{model()} provides a straightforward interface for fitting global
#' calibration models, in contrast to the local modelling approach implemented
#' in \code{\link{mbl}}. This is useful when:
#' \itemize{
#'   \item{the relationship between spectra and the response is approximately
#'   linear across the full dataset}
#'   \item{a single portable model is needed for deployment}
#'   \item{computational efficiency is prioritised over predictive performance
#'   in heterogeneous datasets}
#' }
#'
#' \subsection{Cross-validation}{
#' When \code{validation_type = "lgo"}, stratified random sampling is used
#' to create training-validation splits that preserve the distribution of the
#' response variable. Cross-validation results include RMSE, R², and
#' standardised RMSE for each component in PLS models, or overall in GPR
#' models.
#' }
#'
#' \subsection{PLS models}{
#' When \code{fit_pls()} is used, the function fits a PLS model with the
#' specified number of components. If cross-validation is enabled, the optimal
#' number of components is selected based on the minimum RMSE.
#' }
#'
#' \subsection{GPR models}{
#' When \code{fit_gpr()} is used, the function fits a Gaussian process
#' regression model with a dot-product covariance function. The noise variance
#' parameter controls regularisation.
#' }
#'
#' @return
#' For \code{model()}, an object of class \code{resemble_model} containing:
#' \itemize{
#'   \item \code{fit_method}: Fitting method object used.
#'   \item \code{control}: Model control object used.
#'   \item \code{model}: Fitted model object.
#'   \item \code{cv_results}: Cross-validation results, if
#'     \code{validation_type = "lgo"}.
#'   \item \code{n_obs}: Number of observations used.
#'   \item \code{n_vars}: Number of predictor variables.
#' }
#'
#' For \code{predict.resemble_model()}, a numeric matrix of predictions. For
#' PLS models, columns correspond to different numbers of components.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @seealso
#' \code{\link{fit_pls}} and \code{\link{fit_gpr}} for specifying the fitting
#' method;
#' \code{\link{model_control}} for controlling aspects of the modelling
#' process;
#' \code{\link{mbl}} for memory-based (local) learning.
#'
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess spectra
#' Xr <- savitzkyGolay(NIRsoil$spc, m = 1, p = 2, w = 11)
#' Yr <- NIRsoil$CEC
#'
#' # Remove missing values
#' ok <- !is.na(Yr)
#' Xr <- Xr[ok, ]
#' Yr <- as.matrix(Yr[ok])
#'
#' # Fit a PLS model with 10 components and cross-validation
#' # Scaling is controlled via fit_pls()
#' pls_mod <- model(
#'   Xr = Xr,
#'   Yr = Yr,
#'   fit_method = fit_pls(ncomp = 10, scale = FALSE),
#'   control = model_control(validation_type = "lgo", number = 10)
#' )
#'
#' # View cross-validation results
#' pls_mod$cv_results
#'
#' # Fit a GPR model (centring/scaling controlled via fit_gpr())
#' gpr_mod <- model(
#'   Xr = Xr,
#'   Yr = Yr,
#'   fit_method = fit_gpr(noise_variance = 0.001, scale = TRUE),
#'   control = model_control(validation_type = "lgo")
#' )
#' }
#'
#' @references
#' de Jong, S. (1993). SIMPLS: An alternative approach to partial least
#' squares regression. \emph{Chemometrics and Intelligent Laboratory Systems},
#' 18(3), 251--263.
#'
#' Rasmussen, C. E., & Williams, C. K. I. (2006). \emph{Gaussian Processes for
#' Machine Learning}. MIT Press.
#'
#' Shenk, J. S., & Westerhaus, M. O. (1991). Population structuring of near
#' infrared spectra and modified partial least squares regression.
#' \emph{Crop Science}, 31(6), 1548--1555.
#' 
# =============================================================================
# model
# =============================================================================

#' @export
model <- function(
    Xr, 
    Yr,
    fit_method,
    control = model_control(),
    verbose = TRUE
) {
  
  call_f <- match.call()
  
  # ---------------------------------------------------------------------------
  # Input validation
  # ---------------------------------------------------------------------------
  if (!is.matrix(Xr)) {
    Xr <- as.matrix(Xr)
  }
  
  if (!is.matrix(Yr)) {
    Yr <- as.matrix(Yr)
  }
  
  if (nrow(Xr) != nrow(Yr)) {
    stop("'Xr' and 'Yr' must have the same number of rows.", call. = FALSE)
  }
  
  if (ncol(Yr) != 1L) {
    stop("'Yr' must have exactly one column.", call. = FALSE)
  }
  
  if (anyNA(Xr)) {
    stop("'Xr' contains missing values.", call. = FALSE)
  }
  
  if (anyNA(Yr)) {
    stop("'Yr' contains missing values.", call. = FALSE)
  }
  
  if (!inherits(fit_method, "fit_method")) {
    stop("'fit_method' must be a fit_*() object (e.g., fit_pls(), fit_gpr()).",
         call. = FALSE)
  }
  
  if (inherits(fit_method, "fit_wapls")) {
    stop(
      "'fit_wapls()' is not supported for global models. ",
      "Use 'fit_pls()' or 'fit_gpr()' instead.",
      call. = FALSE
    )
  }
  
  if (!inherits(control, "model_control")) {
    stop("'control' must be created with model_control().", call. = FALSE)
  }
  
  n_obs  <- nrow(Xr)
  n_vars <- ncol(Xr)
  
  # ---------------------------------------------------------------------------
  # Dispatch to appropriate fitting function
  # ---------------------------------------------------------------------------
  if (inherits(fit_method, "fit_pls")) {
    result <- .fit_model_pls(
      Xr         = Xr,
      Yr         = Yr,
      fit_method = fit_method,
      control    = control,
      verbose    = verbose
    )
  } else if (inherits(fit_method, "fit_gpr")) {
    result <- .fit_model_gpr(
      Xr         = Xr,
      Yr         = Yr,
      fit_method = fit_method,
      control    = control,
      verbose    = verbose
    )
  } else {
    stop("Unsupported 'fit_method' type.", call. = FALSE)
  }
  
  # ---------------------------------------------------------------------------
  # Assemble output
  # ---------------------------------------------------------------------------
  out <- list(
    fit_method = fit_method,
    control    = control,
    model      = result$model,
    cv_results = result$cv_results,
    n_obs      = n_obs,
    n_vars     = n_vars
  )
  attr(out, "call") <- call_f
  class(out) <- "resemble_model"
  out
}


# =============================================================================
# Internal: PLS model fitting
# =============================================================================

.fit_model_pls <- function(
    Xr, Yr, fit_method, control, verbose
) {
  ncomp    <- fit_method$ncomp
  method   <- fit_method$method
  scale    <- fit_method$scale
  max_iter <- fit_method$max_iter
  tol      <- fit_method$tol
  do_cv    <- control$validation_type == "lgo"
  
  cv_results <- NULL
  
  # Cross-validation
  if (do_cv) {
    if (verbose) message("Running cross-validation...")
    
    cv_samples <- sample_stratified(
      y           = Yr,
      p           = control$p,
      number      = control$number,
      group       = NULL,
      replacement = FALSE,
      seed        = NULL
    )
    
    cv_raw <- opls_cv_cpp(
      X             = Xr,
      Y             = Yr,
      scale         = scale,
      method        = "pls",
      mindices      = cv_samples$hold_in,
      pindices      = cv_samples$hold_out,
      min_component = 1L,
      ncomp         = ncomp,
      new_x         = matrix(0, 1, 1),
      maxiter       = max_iter,
      tol           = tol,
      wapls_grid    = matrix(0, 0, 0),
      algorithm     = method,
      statistics    = TRUE
    )
    
    cv_results <- data.frame(
      ncomp      = seq_len(ncomp),
      rmse       = rowMeans(cv_raw$rmse_seg),
      rmse_sd    = apply(cv_raw$rmse_seg, 1, sd),
      st_rmse    = rowMeans(cv_raw$st_rmse_seg),
      r2         = rowMeans(cv_raw$rsq_seg),
      row.names  = NULL
    )
    
    best_ncomp <- which.min(cv_results$rmse)
    cv_results$optimal <- seq_len(ncomp) == best_ncomp
  }
  
  # Fit final model
  if (verbose) message("Fitting model...")
  
  model_fit <- opls_get_all(
    X         = Xr,
    Y         = Yr,
    ncomp     = ncomp,
    scale     = scale,
    maxiter   = max_iter,
    tol       = tol,
    algorithm = method
  )
  
  list(
    model      = model_fit,
    cv_results = cv_results,
    transf     = model_fit$transf
  )
}


# =============================================================================
# Internal: GPR model fitting
# =============================================================================

.fit_model_gpr <- function(
    Xr, Yr, fit_method, control, verbose
) {
  
  noise_variance <- fit_method$noise_variance
  gpr_center     <- fit_method$center
  gpr_scale      <- fit_method$scale
  do_cv          <- control$validation_type == "lgo"
  
  cv_results <- NULL
  
  # Cross-validation
  if (do_cv) {
    if (verbose) message("Running cross-validation...")
    
    cv_samples <- sample_stratified(
      y           = Yr,
      p           = control$p,
      number      = control$number,
      group       = NULL,
      replacement = FALSE,
      seed        = NULL
    )
    
    cv_raw <- gaussian_process_cv(
      X        = Xr,
      Y        = Yr,
      mindices = cv_samples$hold_in,
      pindices = cv_samples$hold_out,
      noisev   = noise_variance,
      scale    = gpr_scale
    )
    
    cv_results <- data.frame(
      rmse      = mean(cv_raw$rmse_seg),
      rmse_sd   = sd(cv_raw$rmse_seg),
      st_rmse   = mean(cv_raw$st_rmse_seg),
      r2        = mean(cv_raw$rsq_seg),
      row.names = NULL
    )
  }
  
  # Fit final model
  if (verbose) message("Fitting model...")
  
  model_fit <- gaussian_process(
    X      = Xr,
    Y      = Yr,
    noisev = noise_variance,
    scale  = gpr_scale
  )
  
  transf <- list(
    Xcenter = model_fit$Xcenter,
    Xscale  = model_fit$Xscale,
    Ycenter = model_fit$Ycenter,
    Yscale  = model_fit$Yscale
  )
  
  list(
    model      = model_fit,
    cv_results = cv_results,
    transf     = transf
  )
}



# =============================================================================
# Predict method
# =============================================================================
#' @export
predict.resemble_model <- function(object, newdata, ncomp = NULL, ...) {
  
  if (!is.matrix(newdata)) {
    newdata <- as.matrix(newdata)
  }
  
  if (ncol(newdata) != object$n_vars) {
    stop(
      "'newdata' must have ", object$n_vars, " columns (same as training data).",
      call. = FALSE
    )
  }
  
  if (inherits(object$fit_method, "fit_pls")) {
    
    if (is.null(ncomp)) {
      ncomp <- object$fit_method$ncomp
      sel_ncomp <- 1:ncomp
    } else {
      sel_ncomp <- ncomp
    }
    
    if (max(ncomp) > object$fit_method$ncomp) {
      stop(
        "'ncomp' (", max(ncomp), ") exceeds the number of components in the model (",
        object$fit_method$ncomp, ").",
        call. = FALSE
      )
    }
    
    yhat <- predict_opls(
      bo = object$model$bo,
      b = object$model$coefficients,
      ncomp = max(ncomp),
      newdata = newdata,
      scale = object$fit_method$scale,
      Xscale = object$model$transf$Xscale
    )
    colnames(yhat) <- paste0("ncomp", seq_len(ncol(yhat)))
    yhat <- yhat[, sel_ncomp, drop = FALSE]
    
  } else if (inherits(object$fit_method, "fit_gpr")) {
    
    yhat <- predict_gaussian_process(
      Xz = object$model$Xz,
      alpha = object$model$alpha,
      newdata = newdata,
      scale = object$model$is_scaled,
      Xcenter = object$model$Xcenter,
      Xscale = object$model$Xscale,
      Ycenter = object$model$Ycenter,
      Yscale = object$model$Yscale
    )
    
    yhat <- as.matrix(yhat)
    colnames(yhat) <- "predicted"
    
  } else {
    stop("Unknown model type.", call. = FALSE)
  }
  
  rownames(yhat) <- rownames(newdata)
  yhat
}

#' @title PCA dissimilarity method constructor
#'
#' @description
#' Creates a configuration object for computing dissimilarity based on
#' Mahalanobis distance in PCA score space.
#'
#' @param ncomp Component selection method. Can be:
#'   \itemize{
#'     \item A positive integer for a fixed number of components
#'     \item \code{\link{ncomp_fixed}(n)}: explicit fixed selection
#'     \item \code{\link{ncomp_by_var}(min_var)}: retain components explaining
#'       at least \code{min_var} variance each (default: \code{ncomp_by_var(0.01)})
#'     \item \code{\link{ncomp_by_cumvar}(min_cumvar)}: retain components until
#'       cumulative variance reaches \code{min_cumvar}
#'     \item \code{\link{ncomp_by_opc}()}: optimize using side information
#'       (\code{Yr} required in \code{dissimilarity()})
#'   }
#' @param method Character. PCA algorithm: \code{"pca"} (default, SVD-based) or
#'   \code{"pca_nipals"} (NIPALS algorithm).
#' @param center Logical. Center data before projection? Default \code{TRUE}.
#' @param scale Logical. Scale data before projection? Default \code{FALSE}.
#' @param return_projection Logical. Return the projection object?
#'   Default \code{FALSE}.
#'
#' @return An object of class \code{c("diss_pca", "diss_method")}.
#'
#' @seealso 
#' Component selection: \code{\link{ncomp_by_var}}, \code{\link{ncomp_by_cumvar}},
#'   \code{\link{ncomp_by_opc}}, \code{\link{ncomp_fixed}}
#'
#' Other dissimilarity methods: \code{\link{diss_pls}}, 
#'   \code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
#'   \code{\link{diss_cosine}}, \code{\link{diss_mahalanobis}}
#'
#' @examples
#' # Fixed number of components
#' diss_pca(ncomp = 10)
#' diss_pca(ncomp = ncomp_fixed(10))
#'
#' # Retain components explaining >= 1% variance each (default)
#' diss_pca(ncomp = ncomp_by_var(0.01))
#'
#' # Retain components until 99% cumulative variance
#' diss_pca(ncomp = ncomp_by_cumvar(0.99))
#'
#' # Optimize using side information (requires Yr)
#' diss_pca(ncomp = ncomp_by_opc(40))
#' diss_pca(ncomp = ncomp_by_opc())
#'
#' # NIPALS algorithm (useful for very large matrices)
#' diss_pca(ncomp = 10, method = "pca_nipals")
#'
#' @export
diss_pca <- function(
    ncomp = ncomp_by_var(0.01),
    method = c("pca", "pca_nipals"),
    center = TRUE,
    scale = FALSE,
    return_projection = FALSE
) {
  ncomp <- .coerce_ncomp(ncomp)
  method <- match.arg(method)
  .validate_logical_args(center, scale, return_projection)
  
  structure(
    list(
      ncomp = ncomp,
      method = method,
      center = center,
      scale = scale,
      return_projection = return_projection
    ),
    class = c("diss_pca", "diss_method")
  )
}


#' @title PLS dissimilarity method constructor
#'
#' @description
#' Creates a configuration object for computing dissimilarity based on
#' Mahalanobis distance in PLS score space. Requires \code{Yr} in
#' \code{dissimilarity()}.
#'
#' @param ncomp Component selection method. Can be:
#'   \itemize{
#'     \item A positive integer for a fixed number of components
#'     \item \code{\link{ncomp_fixed}(n)}: explicit fixed selection
#'     \item \code{\link{ncomp_by_var}(min_var)}: retain components explaining
#'       at least \code{min_var} variance each
#'     \item \code{\link{ncomp_by_cumvar}(min_cumvar)}: retain components until
#'       cumulative variance reaches \code{min_cumvar}
#'     \item \code{\link{ncomp_by_opc}()}: optimize using side information
#'       (default; recommended for PLS since \code{Yr} is already required)
#'   }
#' @param method Character. PLS algorithm: \code{"pls"} (default) or
#'   \code{"mpls"} (modified PLS, Shenk & Westerhaus 1991).
#' @param scale Logical. Scale data? Default \code{FALSE}. Note: PLS always
#'   centers internally.
#' @param return_projection Logical. Return projection object?
#'   Default \code{FALSE}.
#'
#' @return An object of class \code{c("diss_pls", "diss_method")}.
#'
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @seealso 
#' Component selection: \code{\link{ncomp_by_var}}, \code{\link{ncomp_by_cumvar}},
#'   \code{\link{ncomp_by_opc}}, \code{\link{ncomp_fixed}}
#'
#' Other dissimilarity methods: \code{\link{diss_pca}},
#'   \code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
#'   \code{\link{diss_cosine}}, \code{\link{diss_mahalanobis}}
#'
#' @examples
#' # Default: OPC optimization (recommended)
#' diss_pls()
#'
#' # Fixed number of components
#' diss_pls(ncomp = 15)
#'
#' # Custom opc settings
#' diss_pls(ncomp = ncomp_by_opc(max_ncomp = 50))
#'
#' # Modified PLS
#' diss_pls(ncomp = 10, method = "mpls")
#'
#' @export
diss_pls <- function(
    ncomp = ncomp_by_opc(),
    method = c("pls", "mpls"),
    scale = FALSE,
    return_projection = FALSE
) {
  ncomp <- .coerce_ncomp(ncomp)
  method <- match.arg(method)
  .validate_logical_args(TRUE, scale, return_projection)
  
  structure(
    list(
      ncomp = ncomp,
      method = method,
      center = TRUE,
      scale = scale,
      return_projection = return_projection
    ),
    class = c("diss_pls", "diss_method")
  )
}


.format_ncomp <- function(x) {
  if (is.numeric(x)) {
    return(sprintf("fixed: %d", as.integer(x)))
  }
  
  ncomp_class <- class(x)[[1]]
  
  switch(
    ncomp_class,
    ncomp_by_var = sprintf("var >= %s (max: %d)", x$min_var, x$max_ncomp),
    ncomp_by_cumvar = sprintf("cumvar >= %s (max: %d)", x$min_cumvar, x$max_ncomp),
    ncomp_by_opc = sprintf("opc (%.0f%%, max: %d)", x$prop * 100, x$max_ncomp),
    ncomp_fixed = sprintf("fixed: %d", x$ncomp)
  )
}


#' @export
print.diss_pca <- function(x, ...) {
  cat("Dissimilarity: PCA\n")
  cat("  method            :", x$method, "\n")
  cat("  ncomp             :", .format_ncomp(x$ncomp), "\n")
  cat("  center            :", x$center, "\n")
  cat("  scale             :", x$scale, "\n")
  cat("  return_projection :", x$return_projection, "\n")
  invisible(x)
}


#' @export
print.diss_pls <- function(x, ...) {
  cat("Dissimilarity: PLS\n")
  cat("  method            :", x$method, "\n")
  cat("  ncomp             :", .format_ncomp(x$ncomp), "\n")
  cat("  scale             :", x$scale, "\n")
  cat("  return_projection :", x$return_projection, "\n")
  invisible(x)
}


# =============================================================================
# Internal helpers
# =============================================================================

.validate_logical_args <- function(...) {
  args <- list(...)
  nms <- as.character(substitute(list(...)))[-1]
  
  for (i in seq_along(args)) {
    val <- args[[i]]
    if (!is.logical(val) || length(val) != 1L || is.na(val)) {
      stop("'", nms[i], "' must be TRUE or FALSE.", call. = FALSE)
    }
  }
  invisible(NULL)
}


# .validate_pre_k <- function(pre_k) {
#   if (!is.numeric(pre_k) || length(pre_k) != 1L || pre_k < 1L) {
#     stop("'pre_k' must be a positive integer.", call. = FALSE)
#   }
#   invisible(NULL)
# }


# # =============================================================================
# # diss_local_pca — NOT EXPORTED
# # =============================================================================
# 
# diss_local_pca <- function(
#     ncomp = ncomp_by_var(0.01),
#     method = c("pca", "pca_nipals"),
#     pre_k,
#     center = TRUE,
#     scale = FALSE,
#     return_projection = FALSE,
#     allow_parallel = TRUE
# ) {
#   ncomp <- .coerce_ncomp(ncomp)
#   method <- match.arg(method)
#   
#   if (missing(pre_k)) {
#     stop("'pre_k' is required for diss_local_pca().", call. = FALSE)
#   }
#   .validate_pre_k(pre_k)
#   .validate_logical_args(center, scale, return_projection, allow_parallel)
#   
#   structure(
#     list(
#       ncomp = ncomp,
#       method = method,
#       center = center,
#       scale = scale,
#       local = TRUE,
#       pre_k = as.integer(pre_k),
#       return_projection = return_projection,
#       allow_parallel = allow_parallel
#     ),
#     class = c("diss_local_pca", "diss_pca", "diss_method")
#   )
# }


# =============================================================================
# diss_local_pls — NOT EXPORTED
# =============================================================================

# diss_local_pls <- function(
#     ncomp = ncomp_by_opc(),
#     method = c("pls", "mpls"),
#     pre_k,
#     center = TRUE,
#     scale = FALSE,
#     return_projection = FALSE,
#     allow_parallel = TRUE
# ) {
#   ncomp <- .coerce_ncomp(ncomp)
#   method <- match.arg(method)
#   
#   if (missing(pre_k)) {
#     stop("'pre_k' is required for diss_local_pls().", call. = FALSE)
#   }
#   .validate_pre_k(pre_k)
#   .validate_logical_args(center, scale, return_projection, allow_parallel)
#   
#   structure(
#     list(
#       ncomp = ncomp,
#       method = method,
#       center = TRUE,
#       scale = scale,
#       local = TRUE,
#       pre_k = as.integer(pre_k),
#       return_projection = return_projection,
#       allow_parallel = allow_parallel
#     ),
#     class = c("diss_local_pls", "diss_pls", "diss_method")
#   )
# }


# =============================================================================
# Print methods for local (not exported)
# =============================================================================

# #' @export
# print.diss_local_pca <- function(x, ...) {
#   cat("Dissimilarity: local PCA\n")
#   cat("  method            :", x$method, "\n")
#   cat("  ncomp             :", .format_ncomp(x$ncomp), "\n")
#   cat("  pre_k             :", x$pre_k, "\n")
#   cat("  center            :", x$center, "\n")
#   cat("  scale             :", x$scale, "\n")
#   cat("  return_projection :", x$return_projection, "\n")
#   invisible(x)
# }


# #' @export
# print.diss_local_pls <- function(x, ...) {
#   cat("Dissimilarity: local PLS\n")
#   cat("  method            :", x$method, "\n")
#   cat("  ncomp             :", .format_ncomp(x$ncomp), "\n")
#   cat("  pre_k             :", x$pre_k, "\n")
#   cat("  scale             :", x$scale, "\n")
#   cat("  return_projection :", x$return_projection, "\n")
#   invisible(x)
# }

#' @title Correlation and moving correlation dissimilarity measurements  
#' @description
#' \loadmathjax
#' \ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#satble'><img src='figures/lifecycle-stable.svg' alt='Stable lifecycle'></a>}}{\strong{Stable}}
#'
#' Computes correlation and moving correlation dissimilarity matrices.
#' @usage
#' cor_diss(
#'   Xr, Xu = NULL, ws = NULL, 
#'   center = TRUE, scale = FALSE
#' )
#' @param Xr a matrix.
#' @param Xu an optional matrix containing data of a second set of observations.
#' @param ws for moving correlation dissimilarity, an odd integer value which
#' specifies the window size. If \code{ws = NULL}, then the window size will be
#' equal to the number of variables (columns), i.e. instead moving correlation,
#' the normal correlation will be used. See details.
#' @param center a logical indicating if the spectral data \code{Xr} (and
#' \code{Xu} if specified) must be centered. If \code{Xu} is provided, the data
#' is scaled on the basis of \mjeqn{Xr \cup Xu}{Xr U Xu}.
#' @param scale a logical indicating if \code{Xr} (and \code{Xu} if specified)
#' must be scaled. If \code{Xu} is provided the data is scaled on the basis
#' of \mjeqn{Xr \cup Xu}{Xr U Xu}.
#' @details
#' The correlation dissimilarity \mjeqn{d}{d} between two observations
#' \mjeqn{x_i}{x_i} and \mjeqn{x_j}{x_j} is based on the Perason's
#' correlation coefficient (\mjeqn{\rho}{\rho}) and it can be computed as
#' follows:
#'
#' \mjdeqn{d(x_i, x_j) = \frac{1}{2}((1 - \rho(x_i, x_j)))}{d(x_i, x_j) = 1/2 (1 - \rho(x_i, x_j))}
#'
#' The above formula is used when \code{ws = NULL}.
#' On the other hand (when \code{ws != NULL}) the moving correlation
#' dissimilarity between two observations \mjeqn{x_i}{x_i} and \mjeqn{x_j}{x_j}
#' is computed as follows:
#'
#' \mjdeqn{d(x_i, x_j; ws) = \frac{1}{2 ws}\sum_{k=1}^{p-ws}1 - \rho(x_{i,(k:k+ws)}, x_{j,(k:k+ws)})}{d(x_i, x_j) = 1/(2 ws)\sum_(k=1)^{p-ws}(1 - \rho(x_(i,k:k+ws), x_(j,k:k+ws)))}
#'
#' where \mjeqn{ws}{ws} represents a given window size which rolls sequentially
#' from 1 up to \mjeqn{p - ws}{p - ws} and  \mjeqn{p}{p} is the number of
#' variables of the observations.
#'
#' The function does not accept input data containing missing values.
#' @return
#' a matrix of the computed dissimilarities.
#' @author Antoine Stevens and \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
#' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#'
#' cor_diss(Xr = Xr)
#'
#' cor_diss(Xr = Xr, Xu = Xu)
#'
#' cor_diss(Xr = Xr, ws = 41)
#'
#' cor_diss(Xr = Xr, Xu = Xu, ws = 41)
#' }
#' @keywords internal
#' @noRd


######################################################################
# resemble
# Copyrigth (C) 2014 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################

## History:
## 09.03.2014 Leo     The line rslt[is.na(rslt)] <- 0 was added in order
##                    to deal with NaNs produced by the C++ code
## 21.04.2020 Leo     styler applied and Argument scaled renamed to scale
##                    the dimnames of the resulting matrix are now Xr_1... Xr_n
##                    (previusly Xr.1... Xr.n)
## 03.07.2020 Leo     FIXME: diss between the same observation in some values
##                    around 1e-15 are returned

cor_diss <- function(
    Xr, 
    Xu = NULL, 
    ws = NULL, 
    center = TRUE, 
    scale = FALSE
) {
  if (!ncol(Xr) >= 2) {
    stop("For correlation dissimilarity the number of variables must be larger than 1")
  }
  if (!is.null(Xu)) {
    if (ncol(Xu) != ncol(Xr)) {
      stop("The number of columns (variables) in Xr must be equal to the number of columns (variables) in Xu")
    }
    if (sum(is.na(Xu)) > 0) {
      stop("Input data contains missing values")
    }
  }
  
  if (sum(is.na(Xr)) > 0) {
    stop("Matrices with missing values are not accepted")
  }
  
  if (!is.logical(center)) {
    stop("'center' argument must be logical")
  }
  
  if (!is.logical(scale)) {
    stop("'scale' argument must be logical")
  }
  
  if (center | scale) {
    X <- rbind(Xr, Xu)
    
    if (center) {
      X <- sweep(x = X, MARGIN = 2, FUN = "-", STATS = colMeans(X))
    }
    
    if (scale) {
      X <- sweep(x = X, MARGIN = 2, FUN = "/", STATS = get_col_sds(X))
    }
    
    if (!is.null(Xu)) {
      Xu <- X[(nrow(X) - nrow(Xu) + 1):nrow(X), , drop = FALSE]
      Xr <- X[1:(nrow(X) - nrow(Xu)), ]
    } else {
      Xr <- X
    }
    rm(X)
  }
  
  xr_sds <- get_col_sds(t(Xr))
  if (any(xr_sds == 0)) {
    stop(
      paste0(
        "Correlation coefficients cannot be computed. Xr contains ", 
        sum(xr_sds == 0), 
        " observation(s) with a standard deviation of zero."
      )
    )
  }
  if (!is.null(Xu)) {
    xu_sds <- get_col_sds(t(Xu))
    if (any(xr_sds == 0)) {
      stop(
        paste0(
          "Correlation coefficients cannot be computed. Xu contains ", 
          sum(xu_sds == 0), 
          "observation(s) with a standard deviation of zero."
        )
      )
    }
  }
  
  if (is.null(ws)) {
    ws <- ncol(Xr)
  } else {
    
    if (ws < 3 | length(ws) != 1) {
      stop(paste("'ws' must be an odd value greater than 2"))
    }
    if ((ws %% 2) == 0) {
      stop("'ws' must be an odd value")
    }
    if (ws >= ncol(Xr)) {
      stop("'ws' must smaller than the number of columns (variables) in Xr")
    }
  }
  
  if (!is.null(Xu)) {
    rslt <- moving_cor_diss_xy(
      Xu, Xr, ws, 
      compute_block_rows(dim(Xu)),
      compute_block_rows(dim(Xr))
    )
    colnames(rslt) <- paste("Xu", 1:nrow(Xu), sep = "_")
    rownames(rslt) <- paste("Xr", 1:nrow(Xr), sep = "_")
  } else {
    rslt <- moving_cor_diss_self_f64(
      Xr, ws,
      compute_block_rows(dim(Xr))
    )
    rownames(rslt) <- colnames(rslt) <- paste("Xr", 1:nrow(Xr), sep = "_")
  }
  rslt
}

# helpers
.divider <- function(width = 55L) {
  sys_width <- getOption("width")
  if (width > sys_width) width <- sys_width
  paste(rep("_", width), collapse = "")
}

.use_color <- function() {
  getOption("cli.num_colors", if (interactive()) 8L else 1L) > 1L
}

.col_blue <- function(x) {
  if (.use_color()) paste0("\033[34m", x, "\033[39m") else x
}

.col_bold_red <- function(x) {
  if (.use_color()) paste0("\033[1;31m", x, "\033[0m") else x
}

.truncate_call <- function(call_obj, width = 80L) {
  call_str <- deparse(call_obj, width.cutoff = width)
  if (length(call_str) > 3L) {
    call_str <- c(call_str[1:3], "...")
  }
  call_str
}

# -----------------------------------------------------------------------------
# print.liblex
# -----------------------------------------------------------------------------
#' @noRd
#' @export
print.liblex <- function(x, ...) {
  
  div <- .divider()
  
  has_coefficients <- !is.null(x$coefficients)
  
  # Header
  if (has_coefficients) {
    cat(.col_bold_red("--- liblex model library ---"), "\n")
    n_models <- length(x$coefficients$B0)
    cat(.col_blue("Models:    "), n_models, "\n")
    cat(.col_blue("Predictors:"), ncol(x$coefficients$B), "\n")
  } else {
    cat(.col_bold_red("--- liblex validation ---"), "\n")
    n_obs <- nrow(x$dissimilarity$dissimilarity)
    cat(.col_blue("Observations:"), n_obs, "\n")
  }
  cat(div, "\n")
  
  # Dissimilarity
  if (inherits(x$dissimilarity$diss_method, "diss_method")) {
    cat(.col_blue("Dissimilarity"), "\n")
    print(x$dissimilarity$diss_method)
    cat(div, "\n")
  }
  
  cat(.col_blue("Local fit method"), "\n")
  print(x$fit_method)
  cat(div, "\n")
  
  # Optimal parameters
  cat(.col_blue("Optimal parameters"), "\n")
  if (!is.null(x$optimal_params)) {
    if (inherits(x$neighbors, "neighbors_k"))
      cat("  k:    ", x$optimal_params$k, "\n")
    else
      cat("  diss threshold:    ", x$optimal_params$diss_threshold, "\n")
    cat("  ncomp:", x$optimal_params$ncomp["min"], "-", x$optimal_params$ncomp["max"], "\n")
  } else if (!is.null(x$best)) {
    cat("  k:    ", x$best$k, "\n")
    cat("  ncomp:", x$best$min_ncomp, "-", x$best$max_ncomp, "\n")
  }
  cat(div, "\n")
  
  
  
  
  
  
  # Validation statistics
  if (!is.null(x$best)) {
    bresult <- NULL
    if (inherits(x$neighbors, "neighbors_k")) {
      for (k in unique(x$results$k)) {
        kth_r <- x$results[x$results$k == k, , drop = FALSE]
        kth_r <- kth_r[which.min(kth_r$rmse), , drop = FALSE]
        bresult <- rbind(bresult, kth_r)
        bresult <- cbind(k = bresult$k, bresult[, names(bresult) != "k", drop = FALSE])
      }
    } else {
      for (thr in unique(x$results$diss_threshold)) {
        kth_r <- x$results[x$results$diss_threshold == thr, , drop = FALSE]
        kth_r <- kth_r[which.min(kth_r$rmse), , drop = FALSE]
        bresult <- rbind(bresult, kth_r)
        bresult <- cbind(diss_threshold = bresult$diss_threshold, bresult[, names(bresult) != "diss_threshold", drop = FALSE])
      }
    }
    
    cat(.col_blue("Nearest-neighbor validation"), "\n\n")
    cat("Best results per neighbor selection metric", "\n")
    # bresult <- data.frame(
    #   rmse = round(x$best$rmse, 3),
    #   st_rmse = round(x$best$st_rmse, 3),
    #   me = round(x$best$me, 3),
    #   r2 = round(x$best$r2, 3)
    # )
    numeric_cols <- sapply(bresult, is.numeric)
    bresult[numeric_cols] <- lapply(bresult[numeric_cols], function(x) {
      signif(x, digits = 3)
    })
    print(bresult, row.names = FALSE)
    cat(div, "\n")
  }
  
  if (!has_coefficients) {
    cat("\nNote: Rebuild with mode = 'build' to enable predictions.\n")
  }
  
  invisible(x)
}

#' @noRd
#' @export
print.gesearch <- function(x, ...) {
  div <- .divider()
  val <- lapply(x$validation_results, function(xx) xx$results)
  
  # Header
  cat(.col_bold_red("--- gesearch results ---"), "\n")
  cat(.col_blue("Iterations:"), x$complete_iter, "\n")
  cat(.col_blue("Selected:  "), length(x$indices), "\n")
  cat(.col_blue("Removed:   "), max(x$n_removed$cumulative), "\n")
  cat(div, "\n")
  
  # Method info
  cat(.col_blue("Fit method"), "\n")
  print(x$fit_method)
  cat(div, "\n")
  
  # Training validation stats
  if (!is.null(val[[1L]]$train)) {
    cat(.col_blue("Validation (training set)"), "\n")
    for (i in seq_along(val)) {
      if (!is.null(names(val)) && names(val)[i] != "") {
        cat("Response:", names(val)[i], "\n")
      }
      print(as.data.frame(val[[i]]$train), digits = 3, row.names = FALSE)
      if (i < length(val)) cat("\n")
    }
    cat(div, "\n")
  }
  
  # Test validation stats
  if (!is.null(val[[1L]]$test)) {
    cat(.col_blue("Validation (test set)"), "\n")
    for (i in seq_along(val)) {
      if (!is.null(names(val)) && names(val)[i] != "") {
        cat("Response:", names(val)[i], "\n")
      }
      print(as.data.frame(val[[i]]$test), digits = 3, row.names = FALSE)
      if (i < length(val)) cat("\n")
    }
    cat(div, "\n")
  }
  
  invisible(x)
}

# -----------------------------------------------------------------------------
# print.mbl
# -----------------------------------------------------------------------------
#' @noRd
#' @export
print.mbl <- function(x, ...) {
  div <- .divider()
  val <- x$validation_results
  
  # Summary
  cat(.col_bold_red("--- mbl predictions ---"), "\n")
  cat(.col_blue("Predictions:"), x$n_predictions, "\n")
  cat(div, "\n")
  
  # Dissimilarity
  if (inherits(x$dissimilarities$diss_method, "diss_method")) {
    cat(.col_blue("Dissimilarity"), "\n")
    print(x$dissimilarities$diss_method)
    cat(div, "\n")
  }
  
  # Method info
  cat(.col_blue("Local fit method"), "\n")
  print(x$fit_method)
  cat(div, "\n")
  
  # Validation results
  if (!is.null(val$nearest_neighbor_validation)) {
    cat("\n")
    cat(.col_blue("Nearest-neighbor validation"), "\n\n")
    print(as.data.frame(val$nearest_neighbor_validation), digits = 3, row.names = FALSE)
    cat(div, "\n")
  }
  
  if (!is.null(val$local_cross_validation)) {
    cat(.col_blue("Local leave-group-out cross-validation"), "\n\n")
    print(as.data.frame(val$local_cross_validation), digits = 3, row.names = FALSE)
    cat(div, "\n")
  }
  
  if (!is.null(val$Yu_prediction_statistics)) {
    cat(.col_blue("Yu prediction statistics"), "\n\n")
    print(as.data.frame(val$Yu_prediction_statistics), digits = 3, row.names = FALSE)
    cat(div, "\n")
  }
  
  if (!is.null(val$Yr_fitted_statistics)) {
    cat(.col_blue("Yr fitted statistics"), "\n\n")
    print(as.data.frame(val$Yr_fitted_statistics), digits = 3, row.names = FALSE)
    cat(div, "\n")
  }
  
  invisible(x)
}

# =============================================================================
# Print method
# =============================================================================

#' @noRd
#' @export
#' @noRd
#' @export
print.resemble_model <- function(x, ...) {
  div <- .divider()
  
  # Header
  cat(.col_bold_red("--- Global resemble model ---"), "\n")
  cat(.col_blue("Method:      "), x$fit_method$fit_method, "\n")
  cat(.col_blue("Observations:"), x$n_obs, "\n")
  cat(.col_blue("Variables:   "), x$n_vars, "\n")
  cat(div, "\n")
  
  # Fit method details
  cat(.col_blue("Fit method"), "\n")
  print(x$fit_method)
  cat(div, "\n")
  
  # Cross-validation results
  if (!is.null(x$cv_results) && nrow(x$cv_results) > 0) {
    cat(.col_blue("Cross-validation"), "\n")
    if (inherits(x$fit_method, "fit_pls")) {
      best_idx <- which(x$cv_results$optimal)
      cat("  Best ncomp:", best_idx, "\n\n")
    }
    print(
      as.data.frame(x$cv_results[, names(x$cv_results) != "optimal"]),
      digits = 3,
      row.names = FALSE
    )
    cat(div, "\n")
  } else {
    cat(.col_blue("Validation:"), x$control$validation_type, "\n")
    cat(div, "\n")
  }
  
  invisible(x)
}

# =============================================================================
# Tests for print helpers and print methods
# =============================================================================

# =============================================================================
# Setup helper
# =============================================================================

.setup_nirsoil_data <- function(n_xr = 40, n_xu = 20) {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ][seq_len(n_xu), ]
  Xr <- Xr[!is.na(Yr), ][seq_len(n_xr), ]
  Yu <- Yu[!is.na(Yu)][seq_len(n_xu)]
  Yr <- Yr[!is.na(Yr)][seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}


# -----------------------------------------------------------------------------
# Helper function tests
# -----------------------------------------------------------------------------

test_that(".divider returns correct width", {
  skip_on_cran()
  
  div <- .divider(55)
  expect_equal(nchar(div), 55)
  expect_true(all(strsplit(div, "")[[1]] == "_"))
  
  div_short <- .divider(10)
  expect_equal(nchar(div_short), 10)
})

test_that(".divider respects system width", {
  skip_on_cran()
  
  old_width <- getOption("width")
  on.exit(options(width = old_width))
  
  options(width = 40)
  div <- .divider(100)
  expect_equal(nchar(div), 40)
})

test_that(".use_color returns logical", {
  skip_on_cran()
  
  result <- .use_color()
  expect_type(result, "logical")
})

test_that(".col_blue returns string", {
  skip_on_cran()
  
  result <- .col_blue("test")
  expect_type(result, "character")
  expect_true(grepl("test", result))
})

test_that(".col_bold_red returns string", {
  skip_on_cran()
  
  result <- .col_bold_red("test")
  expect_type(result, "character")
  expect_true(grepl("test", result))
})

test_that(".truncate_call truncates long calls", {
  skip_on_cran()
  
  short_call <- quote(foo(x = 1))
  result_short <- .truncate_call(short_call)
  expect_type(result_short, "character")
  
  long_call <- quote(foo(a = 1, b = 2, c = 3, d = 4, e = 5, f = 6, g = 7, h = 8))
  result_long <- .truncate_call(long_call, width = 20)
  expect_true(length(result_long) <= 4)
})

# -----------------------------------------------------------------------------
# print.ortho_projection tests
# -----------------------------------------------------------------------------

test_that("print.ortho_projection works for PCA", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, ncomp = 5)
  out <- capture.output(print(proj))
  
  expect_true(any(grepl("Method:", out)))
  expect_true(any(grepl("pca", out)))
  expect_true(any(grepl("Number of components retained:", out)))
  expect_true(any(grepl("Original variance", out)))
  expect_true(any(grepl("Explained variance", out)))
})

test_that("print.ortho_projection works for PCA with Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Xu = d$Xu, ncomp = 5)
  out <- capture.output(print(proj))
  
  expect_true(any(grepl("Xr; Xu", out)))
})

test_that("print.ortho_projection works for PLS", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Yr = d$Yr, method = "pls", ncomp = 5)
  out <- capture.output(print(proj))
  
  expect_true(any(grepl("pls", out)))
  expect_true(any(grepl("Explained variance in Yr", out)))
})

test_that("print.ortho_projection works for mpls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Yr = d$Yr, method = "mpls", ncomp = 5)
  out <- capture.output(print(proj))
  
  expect_true(any(grepl("mpls", out)))
  expect_true(any(grepl("Explained variance in Yr", out)))
})

test_that("print.ortho_projection returns invisible", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, ncomp = 5)
  
  capture.output(expect_invisible(print(proj)))
})

# -----------------------------------------------------------------------------
# print.liblex tests
# -----------------------------------------------------------------------------

test_that("print.liblex works for model library with wapls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  lib <- liblex(
    Xr = d$Xr, Yr = d$Yr,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(k = c(20, 30)),
    fit_method = fit_wapls(3, 10),
    mode = "build",
    verbose = FALSE
  )
  out <- capture.output(print(lib))
  
  expect_true(any(grepl("liblex model library", out)))
  expect_true(any(grepl("Models:", out)))
  expect_true(any(grepl("Dissimilarity", out)))
  expect_true(any(grepl("Local fit method", out)))
  expect_true(any(grepl("Optimal parameters", out)))
})

test_that("print.liblex works for model library with pls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  lib <- liblex(
    Xr = d$Xr, Yr = d$Yr,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(k = c(20, 30)),
    fit_method = fit_pls(ncomp = 10),
    mode = "build",
    verbose = FALSE
  )
  out <- capture.output(print(lib))
  
  expect_true(any(grepl("liblex model library", out)))
  expect_true(any(grepl("Models:", out)))
  expect_true(any(grepl("Predictors:", out)))
  expect_true(any(grepl("Dissimilarity", out)))
  expect_true(any(grepl("Local fit method", out)))
  expect_true(any(grepl("Optimal parameters", out)))
})

test_that("print.liblex returns invisible", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  lib <- liblex(
    Xr = d$Xr, Yr = d$Yr,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(k = c(20, 30)),
    fit_method = fit_pls(ncomp = 10),
    mode = "build",
    verbose = FALSE
  )
  
  capture.output(expect_invisible(print(lib)))
})

# -----------------------------------------------------------------------------
# print.mbl tests
# -----------------------------------------------------------------------------

test_that("print.mbl works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr,
    Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(k = 30),
    fit_method = fit_pls(ncomp = 10),
    control = mbl_control(validation_type = "NNv"),
    verbose = FALSE
  )
  out <- capture.output(print(result))
  
  expect_true(any(grepl("mbl predictions", out)))
  expect_true(any(grepl("Predictions:", out)))
  expect_true(any(grepl("Dissimilarity", out)))
  expect_true(any(grepl("Local fit method", out)))
})

test_that("print.mbl returns invisible", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- mbl(
    Xr = d$Xr, Yr = d$Yr,
    Xu = d$Xu,
    diss_method = diss_pca(ncomp = 10),
    neighbors = neighbors_k(k = 30),
    fit_method = fit_pls(ncomp = 10),
    control = mbl_control(validation_type = "NNv"),
    verbose = FALSE
  )
  
  capture.output(expect_invisible(print(result)))
})

# -----------------------------------------------------------------------------
# print.resemble_model tests
# -----------------------------------------------------------------------------

test_that("print.resemble_model works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  mod <- model(
    Xr = d$Xr, Yr = d$Yr,
    fit_method = fit_pls(ncomp = 10),
    verbose = FALSE
  )
  out <- capture.output(print(mod))
  
  expect_true(any(grepl("Global resemble model", out)))
  expect_true(any(grepl("Method:", out)))
  expect_true(any(grepl("Observations:", out)))
  expect_true(any(grepl("Variables:", out)))
  expect_true(any(grepl("Fit method", out)))
  expect_true(any(grepl("Cross-validation", out)))
})

test_that("print.resemble_model returns invisible", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  mod <- model(
    Xr = d$Xr, Yr = d$Yr,
    fit_method = fit_pls(ncomp = 10),
    verbose = FALSE
  )
  
  capture.output(expect_invisible(print(mod)))
})

# -----------------------------------------------------------------------------
# print.gesearch tests
# -----------------------------------------------------------------------------

test_that("print.gesearch works", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- gesearch(
    Xr = d$Xr, Yr = d$Yr,
    Xu = d$Xu, Yu = d$Yu,
    k = 20, b = 20,
    fit_method = fit_pls(ncomp = 10),
    verbose = FALSE
  )
  out <- capture.output(print(result))
  
  expect_true(any(grepl("gesearch results", out)))
  expect_true(any(grepl("Iterations:", out)))
  expect_true(any(grepl("Selected:", out)))
  expect_true(any(grepl("Removed:", out)))
  expect_true(any(grepl("Fit method", out)))
})

test_that("print.gesearch returns invisible", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  result <- gesearch(
    Xr = d$Xr, Yr = d$Yr,
    Xu = d$Xu, Yu = d$Yu,
    k = 30, b = 10,
    fit_method = fit_pls(ncomp = 10),
    verbose = FALSE
  )
  
  capture.output(expect_invisible(print(result)))
})

#' @title Control parameters for memory-based learning
#' @description
#' \loadmathjax
#' This function controls various aspects of the memory-based learning process
#' in the \code{\link{mbl}} function.
#'
#' @usage
#' mbl_control(
#'   return_dissimilarity = FALSE,
#'   validation_type = "NNv",
#'   tune_locally = TRUE,
#'   number = 10,
#'   p = 0.75,
#'   range_prediction_limits = TRUE,
#'   allow_parallel = TRUE,
#'   blas_threads = 1L
#' )
#'
#' @param return_dissimilarity Logical indicating whether to return the
#'   dissimilarity matrix between \code{Xr} and \code{Xu}. Default is
#'   \code{FALSE}.
#' @param validation_type Character vector specifying validation method(s):
#'   \itemize{
#'     \item \code{"NNv"}: Leave-nearest-neighbor-out cross-validation (default, faster)
#'     \item \code{"local_cv"}: Local leave-group-out cross-validation
#'     \item \code{"none"}: No validation
#'   }
#'   Multiple methods can be specified (e.g., \code{c("NNv", "local_cv")}).
#'   Default is \code{"NNv"}.
#' @param tune_locally Logical indicating whether to tune PLS components
#'   locally when \code{validation_type = "local_cv"} and using
#'   \code{\link{fit_pls}} or \code{\link{fit_wapls}}. Default is \code{TRUE}.
#' @param number Integer specifying the number of sampling iterations for
#'   \code{"local_cv"} validation. Default is \code{10}.
#' @param p Numeric value between 0 and 1 indicating the proportion of
#'   observations retained at each \code{"local_cv"} iteration. Default is
#'   \code{0.75}.
#' @param range_prediction_limits Logical indicating whether predictions should
#'   be constrained to the range of response values in each neighborhood.
#'   Default is \code{TRUE}.
#' @param allow_parallel Logical indicating whether parallel execution is
#'   allowed via the \pkg{foreach} package. Default is \code{TRUE}.
#' @param blas_threads Integer specifying the number of BLAS threads to use
#'   during \code{mbl()} execution. Default is \code{1L}, which avoids thread
#'   overhead from repeated small matrix operations. Requires the
#'   \pkg{RhpcBLASctl} package to take effect. The original thread count is
#'   restored after \code{mbl()} completes. See Details.
#'
#' @details
#' ## Validation methods
#'
#' \strong{Leave-nearest-neighbor-out cross-validation (\code{"NNv"}):}
#' For each target observation, the nearest neighbor is excluded from the
#' local model, which then predicts that neighbor's value. This is faster
#' than \code{"local_cv"}. If the nearest neighbor belongs to a group
#' (specified via the \code{group} argument in \code{\link{mbl}}), all
#' group members are excluded.
#'
#' \strong{Local leave-group-out cross-validation (\code{"local_cv"}):}
#' The neighborhood is partitioned into subsets via stratified random
#' sampling. Each subset serves as validation data while the remainder
#' fits the model. This repeats \code{number} times, with \code{p}
#' controlling the training proportion. The final error is the average
#' local RMSE.
#'
#' ## BLAS threading
#'
#' On Linux systems with multi-threaded OpenBLAS, the default thread count
#' can cause significant overhead for algorithms like \code{mbl()} that
#' perform many small matrix operations. Setting \code{blas_threads = 1}
#' (the default) eliminates this overhead.
#'
#' This setting requires the \pkg{RhpcBLASctl} package. If not installed,
#' the parameter is ignored and a message is displayed. The original
#' thread count is restored when \code{mbl()} completes.
#'
#' Windows systems typically use single-threaded BLAS by default, so this
#' setting has no effect there.
#'
#' @return A list of class \code{mbl_control} with the specified control parameters.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} and
#' Antoine Stevens
#'
#' @references
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199:43-53.
#'
#' @seealso \code{\link{mbl}}, \code{\link{neighbors_k}},
#' \code{\link{neighbors_diss}}
#'
#' @examples
#' # Default control parameters (NNv validation)
#' mbl_control()
#'
#' # Both validation methods
#' mbl_control(validation_type = c("NNv", "local_cv"))
#'
#' # No validation
#' mbl_control(validation_type = "none")
#'
#' # NNv validation only, no parallel
#' mbl_control(validation_type = "NNv", allow_parallel = FALSE)
#'
#' # Allow more BLAS threads (if needed for other computations)
#' mbl_control(blas_threads = 4)
#'
#' @export

mbl_control <- function(
  return_dissimilarity = FALSE,
  validation_type = "NNv",
  tune_locally = TRUE,
  number = 10,
  p = 0.75,
  range_prediction_limits = TRUE,
  allow_parallel = TRUE,
  blas_threads = 1L
) {
  # Validate return_dissimilarity
  if (!is.logical(return_dissimilarity) || length(return_dissimilarity) != 1L) {
    stop("'return_dissimilarity' must be TRUE or FALSE.", call. = FALSE)
  }

  # Validate and match validation_type
  valid_types <- c("NNv", "local_cv", "none")
  validation_type <- match.arg(validation_type, valid_types, several.ok = TRUE)

  if ("none" %in% validation_type && length(validation_type) > 1L) {
    stop(
      "'validation_type' cannot combine 'none' with other values.",
      call. = FALSE
    )
  }

  # Remove "none" if other methods specified
  if ("none" %in% validation_type && length(validation_type) > 1L) {
    validation_type <- validation_type[validation_type != "none"]
  }

  # Validate tune_locally
  if (!is.logical(tune_locally) || length(tune_locally) != 1L) {
    stop("'tune_locally' must be TRUE or FALSE.", call. = FALSE)
  }

  # Validate local_cv parameters
  if ("local_cv" %in% validation_type) {
    if (!is.numeric(number) || length(number) != 1L || number < 1L) {
      stop("'number' must be a positive integer.", call. = FALSE)
    }

    if (!is.numeric(p) || length(p) != 1L || p <= 0 || p >= 1) {
      stop("'p' must be a numeric value between 0 and 1 (exclusive).",
        call. = FALSE
      )
    }
  }

  # Validate range_prediction_limits
  if (!is.logical(range_prediction_limits) ||
    length(range_prediction_limits) != 1L) {
    stop("'range_prediction_limits' must be TRUE or FALSE.", call. = FALSE)
  }

  # Validate allow_parallel
  if (!is.logical(allow_parallel) || length(allow_parallel) != 1L) {
    stop("'allow_parallel' must be TRUE or FALSE.", call. = FALSE)
  }

  # Validate blas_threads
  if (!is.numeric(blas_threads) || length(blas_threads) != 1L ||
    blas_threads < 1L) {
    stop("'blas_threads' must be a positive integer.", call. = FALSE)
  }

  ctl <- list(
    return_dissimilarity = return_dissimilarity,
    validation_type = validation_type,
    tune_locally = tune_locally,
    number = as.integer(number),
    p = p,
    range_prediction_limits = range_prediction_limits,
    allow_parallel = allow_parallel,
    blas_threads = as.integer(blas_threads)
  )
  class(ctl) <- c("mbl_control", "list")
  ctl
}

.onAttach <- function(lib, pkg) {
  pkg_v <- pkg_info()
  
  mss <- paste0(
    "\033[34m",
    pkg, " version ",
    paste(pkg_v[1:2], collapse = " \U002D\U002D "),
    "\033[39m"
  )
  
  mss2 <- paste0(
    "\033[34mRepository: ",
    pkg_v[, "URL"],
    "\033[39m"
  )
  
  mss3 <- paste0(
    "\033[33m",
    "Note: Version 3.0 has deprecated several functions and arguments.\n",
    "For previous functionality, install version 2.2.5:\n",
    "  remotes::install_github(\"l-ramirez-lopez/resemble@2.2.5\")",
    "\033[39m"
  )
  
  packageStartupMessage(mss)
  packageStartupMessage(mss2)
  packageStartupMessage(mss3)
}

#' @import lifecycle
#' @import Rcpp
#' @import foreach
#' @import iterators
#' @import grDevices
#' @import graphics
#' @import mathjaxr
## usethis namespace: start
#' @importFrom lifecycle deprecate_soft
## usethis namespace: end
#' @importFrom utils setTxtProgressBar txtProgressBar
#' @importFrom stats model.frame model.matrix model.extract na.fail sd reshape
#' @importFrom stats approx median na.omit predict setNames
#' @importFrom utils flush.console tail
#' @importFrom RhpcBLASctl blas_get_num_procs blas_set_num_threads
#' @importFrom foreach foreach %dopar%
#' @useDynLib resemble, .registration=TRUE
#' @description
#' 
#' \ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#maturing'><img src='figures/lifecycle-maturing.svg' alt='Maturing lifecycle'></a>}}{\strong{Maturing}}
#'
#' Functions for dissimilarity assessment, nearest-neighbour search, 
#' memory-based learning, local expert libraries, and evolutionary training subset 
#' search in spectral chemometrics.
#'
#' \if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}
#'
#' @details
#' This is the version
#' `r paste(pkg_info()[1:2], collapse = " -- ")`
#' of the package. It implements a number of functions useful for
#' modeling complex spectral spectra (e.g. NIR, IR).
#' The package includes functions for dimensionality reduction,
#' computing spectral dissimilarity matrices, nearest neighbor search,
#' and modeling spectral data using memory-based learning and evolutionary 
#' search of optimal training subsets in large and complex datasets. This package builds
#' upon the methods presented in 
#' Ramirez-Lopez et al. (2013a) \doi{10.1016/j.geoderma.2012.12.014}, 
#' Ramirez-Lopez et al. (2026a) and Ramirez-Lopez et al. (2026b).
#'
#' Development versions can be found in the github repository of the package
#' at \href{https://github.com/l-ramirez-lopez/resemble}{https://github.com/l-ramirez-lopez/resemble}.
#'
#' The functions available for computing dissimilarity matrices are:
#' \itemize{
#'   \item{\code{\link{dissimilarity}}}: Computes a dissimilarity matrix based on a specified method.
#'   \item{\code{\link{diss_pca}}}: constructor for principal components-based dissimilarity method.
#'   \item{\code{\link{diss_pls}}}: constructor for partial least squares-based dissimilarity method.
#'   \item{\code{\link{diss_correlation}}}: constructor for correlation-based dissimilarity method.
#'   \item{\code{\link{diss_euclidean}}}: constructor for euclidean distance-based dissimilarity method.
#'   \item{\code{\link{diss_mahalanobis}}}: constructor for Mahalanobis distance-based dissimilarity method.
#'   \item{\code{\link{diss_cosine}}}: constructor for cosine-based dissimilarity method.
#'   }
#' The functions available for evaluating dissimilarity matrices are:
#' \itemize{
#'   \item{\code{\link{diss_evaluate}}}: Evaluates the effectiveness of a dissimilarity matrix using side information.
#'   }
#' The functions available for nearest neighbor search:
#' \itemize{
#'   \item{\code{\link{search_neighbors}}}: Search for nearest neighbors of a query spectrum in a reference dataset based on a specified dissimilarity method.
#'   }
#' The functions available for modeling spectral data:
#' \itemize{
#'   \item \code{\link{mbl}}: Memory-based learning for modeling spectral data.
#'   \item \code{\link{gesearch}}: An evolutionary method to search optimal samples in large spectral datasets.
#'   \item \code{\link{liblex}}: Builds a library of reusable localized models.
#'   }
#' The functions available for dimensionality reduction are:
#' \itemize{
#'   \item{\code{\link{ortho_projection}}}: Computes an orthogonal projection of the data based on either principal components or partial least squares.
#'   \item{\code{\link{predict.ortho_projection}}}
#'   }
#' @name resemble-package
#' @aliases resemble-package resemble
#' @title Overview of the functions in the resemble package
#' @references
#'
#' Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
#' Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
#' Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
#' libraries are too complex to search: Evolutionary subset selection for
#' domain-adaptive calibration. \emph{Analytica Chimica Acta}, under review.
#' 
#' Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C.,
#' Perez-Fernandez, E., Plans, M., Breure, T., Behrens, T.,
#' Viscarra Rossel, R., & Peng, Y. (2026b). Rethinking local spectral
#' modelling: From per-query refitting to model libraries. 
#' \emph{Analytica Chimica Acta}, under review.
#' 
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. (2013a). The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. \emph{Geoderma} 195-196,
#' 268-279. 
#'
#' @seealso
#' Useful links:
#' \itemize{
#' \item \url{https://github.com/l-ramirez-lopez/resemble}
#' \item Report bugs at \url{https://github.com/l-ramirez-lopez/resemble/issues}
#' }
#' @author
#'
#' \strong{Maintainer / Creator}: Leonardo Ramirez-Lopez \email{ramirez.lopez.leo@gmail.com}
#'
#' Authors:
#' \itemize{
#' \item Leonardo Ramirez-Lopez (\href{https://orcid.org/0000-0002-5369-5120}{ORCID})
#' \item Antoine Stevens (\href{https://orcid.org/0000-0002-1588-7519}{ORCID})
#' \item Claudio Orellano
#' }
######################################################################
# resemble
# Copyright (C) 2014-2026 Leonardo Ramirez-Lopez
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################
"_PACKAGE"
NULL

#' @title Orthogonal projections
#' @name ortho_projection
#' @aliases ortho_projection predict.ortho_projection plot.ortho_projection
#' @description
#' 
#' \loadmathjax
#' Performs orthogonal projections of high-dimensional data matrices using
#' principal component analysis (PCA) or partial least squares (PLS).
#' 
#' @usage 
#' ortho_projection(
#'   Xr, Xu = NULL, Yr = NULL,
#'   ncomp = ncomp_by_var(0.01),
#'   method = c("pca", "pca_nipals", "pls", "mpls", "simpls"),
#'   center = TRUE,
#'   scale = FALSE,
#'   tol = 1e-6,
#'   max_iter = 1000L,
#'   pc_selection = deprecated(),
#'   ...
#' )
#' 
#' \method{predict}{ortho_projection}(object, newdata, ...)
#'
#' \method{plot}{ortho_projection}(x, col = "#3B82F6", ...)
#' 
#' @param Xr A numeric matrix of reference observations (rows) and variables
#'   (columns).
#' @param Xu An optional matrix of additional observations to project.
#' @param Yr An optional response matrix. Required for PLS methods
#'   (\code{"pls"}, \code{"mpls"}, \code{"simpls"}) and when using 
#'   \code{\link{ncomp_by_opc}()}.
#' @param ncomp Component selection method. Either:
#'   \itemize{
#'     \item A positive integer (equivalent to \code{ncomp_fixed(n)})
#'     \item An \code{ncomp_selection} object: \code{\link{ncomp_by_var}()},
#'           \code{\link{ncomp_by_cumvar}()}, \code{\link{ncomp_by_opc}()}, or
#'           \code{\link{ncomp_fixed}()}
#'   }
#'   Default is \code{ncomp_by_var(0.01)}.
#' @param method A character string specifying the projection method:
#'   \itemize{
#'     \item \code{"pca"}: PCA via singular value decomposition (default)
#'     \item \code{"pca_nipals"}: PCA via NIPALS algorithm
#'     \item \code{"pls"}: PLS via NIPALS algorithm
#'     \item \code{"mpls"}: Modified PLS via NIPALS (Shenk and Westerhaus, 1991)
#'     \item \code{"simpls"}: PLS via SIMPLS algorithm (de Jong, 1993)
#'   }
#' @param center A logical indicating whether to center the data. Default is
#'   \code{TRUE}. PLS methods always center internally regardless of this
#'   setting.
#' @param scale A logical indicating whether to scale the data to unit
#'   variance. Default is \code{FALSE}.
#' @param tol Convergence tolerance for the NIPALS algorithm. Default is
#'   \code{1e-6}. Ignored when \code{method = "simpls"}.
#' @param max_iter Maximum number of iterations for NIPALS. Default is
#'   \code{1000}. Ignored when \code{method = "simpls"}.
#' @param pc_selection `r lifecycle::badge("deprecated")` Use \code{ncomp} instead.
#' @param object Object of class \code{"ortho_projection"}.
#' @param newdata Matrix of new observations to project.
#' @param x An object of class \code{ortho_projection} (as returned by
#'   \code{\link{ortho_projection}}).
#' @param col Color for the plot elements. Default is \code{"#3B82F6"}.
#' @param ... Additional arguments (currently unused).
#'
#' @details
#' ## PCA methods
#'
#' When \code{method = "pca"}, singular value decomposition factorizes the
#' data matrix \mjeqn{X}{X} as:
#'
#' \mjdeqn{X = UDV^{T}}{X = UDV^T}
#'
#' where \mjeqn{U}{U} and \mjeqn{V}{V} are orthogonal matrices (left and right
#' singular vectors), and \mjeqn{D}{D} is a diagonal matrix of singular values.
#' The score matrix is \mjeqn{UD}{UD} and the loadings are \mjeqn{V}{V}.
#'
#' When \code{method = "pca_nipals"}, the non-linear iterative partial least
#' squares (NIPALS) algorithm is used instead.
#'
#' ## PLS methods
#'
#' Three PLS variants are available:
#'
#' \itemize{
#'   \item \code{"pls"}: Standard PLS using the NIPALS algorithm with 
#'     covariance-based weights.
#'   \item \code{"mpls"}: Modified PLS using the NIPALS algorithm with 
#'     correlation-based weights, giving equal influence to all predictors 
#'     regardless of variance (Shenk and Westerhaus, 1991).
#'   \item \code{"simpls"}: SIMPLS algorithm (de Jong, 1993), which deflates 
#'     the cross-product matrix rather than X itself. Computationally faster 
#'     than NIPALS, especially for wide matrices.
#' }
#'
#' ## Component selection
#'
#' When \code{\link{ncomp_by_opc}()} is used, component selection minimizes
#' RMSD (for continuous \code{Yr}) or maximizes kappa (for categorical
#' \code{Yr}) between observations and their nearest neighbors. See
#' \code{\link{diss_evaluate}}.
#'
#' @return
#' An object of class \code{"ortho_projection"} containing:
#' \itemize{
#'   \item \code{scores}: Matrix of projected scores for \code{Xr} (and \code{Xu}).
#'   \item \code{X_loadings}: Matrix of X loadings.
#'   \item \code{Y_loadings}: Matrix of Y loadings (PLS only).
#'   \item \code{weights}: Matrix of PLS weights (PLS only).
#'   \item \code{projection_mat}: Projection matrix for new data (PLS only).
#'   \item \code{variance}: List with original and explained variance.
#'   \item \code{scores_sd}: Standard deviation of scores.
#'   \item \code{ncomp}: Number of components retained.
#'   \item \code{center}: Centering vector used.
#'   \item \code{scale}: Scaling vector used.
#'   \item \code{method}: Projection method used.
#'   \item \code{ncomp_method}: The value passed to the `ncomp` argument.
#'   \item \code{opc_evaluation}: opc optimization results (if applicable).
#' }
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @references
#' de Jong, S. 1993. SIMPLS: An alternative approach to partial least squares 
#' regression. Chemometrics and Intelligent Laboratory Systems 18:251-263.
#'
#' Martens, H. 1991. Multivariate calibration. John Wiley & Sons.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199:43-53.
#'
#' Shenk, J.S., Westerhaus, M.O. 1991. Populations structuring of near
#' infrared spectra and modified partial least squares regression. Crop
#' Science 31:1548-1555.
#'
#' @seealso
#' \code{\link{ncomp_by_var}}, \code{\link{ncomp_by_opc}},
#' \code{\link{diss_evaluate}}, \code{\link{mbl}}
#'
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess
#' sg_det <- savitzkyGolay(
#'   detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
#'   m = 1, p = 1, w = 7
#' )
#'
#' # Split data
#' train_x <- sg_det[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
#' train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]
#' test_x <- sg_det[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]
#'
#' # PCA with fixed components
#' proj <- ortho_projection(train_x, ncomp = 5)
#' 
#' plot(proj)
#'
#' # PCA with variance-based selection
#' proj <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))
#'
#' # PCA with OPC optimization
#' proj <- ortho_projection(train_x, Xu = test_x, Yr = train_y,
#'                          ncomp = ncomp_by_opc(40))
#'
#' #' plot(proj)
#' 
#' # PLS projection (NIPALS)
#' proj <- ortho_projection(train_x, Xu = test_x, Yr = train_y,
#'                          method = "pls", ncomp = ncomp_by_opc(40))
#'
#' # Modified PLS
#' proj <- ortho_projection(train_x, Yr = train_y,
#'                          method = "mpls", ncomp = 10)
#'
#' # SIMPLS (faster for wide matrices)
#' proj <- ortho_projection(train_x, Yr = train_y,
#'                          method = "simpls", ncomp = 10)
#' }
#' @rdname ortho_projection
#' @export
ortho_projection <- function(
    Xr, Xu = NULL, Yr = NULL,
    ncomp = ncomp_by_var(0.01),
    method = c("pca", "pca_nipals", "pls", "mpls", "simpls"),
    center = TRUE,
    scale = FALSE,
    tol = 1e-6,
    max_iter = 1000L,
    pc_selection = deprecated(),
    ...
) {
  
  # ---------------------------------------------------------------------------
  # Handle deprecated pc_selection
  # ---------------------------------------------------------------------------
  if (is_present(pc_selection)) {
    deprecate_warn(
      when = "3.0.0",
      what = "ortho_projection(pc_selection)",
      with = "ortho_projection(ncomp)",
      details = "Use ncomp_fixed(), ncomp_by_var(), ncomp_by_cumvar(), or ncomp_by_opc()."
    )
    ncomp <- .convert_pc_selection_to_ncomp(pc_selection)
  }
  
  # ---------------------------------------------------------------------------
  # Handle legacy method names
  # ---------------------------------------------------------------------------
  if (identical(method, "pca.nipals")) {
    
    deprecate_warn(
      when = "3.0.0",
      what = I('ortho_projection(method = "pca.nipals")'),
      with = I('ortho_projection(method = "pca_nipals")')
    )
    method <- "pca_nipals"
  }
  
  method <- match.arg(method)
  is_pca <- method %in% c("pca", "pca_nipals")
  is_pls <- method %in% c("pls", "mpls", "simpls")
  
  # ---------------------------------------------------------------------------
  # Coerce ncomp
  # ---------------------------------------------------------------------------
  ncomp <- .coerce_ncomp(ncomp)
  
  # ---------------------------------------------------------------------------
  # Validate inputs
  # ---------------------------------------------------------------------------
  if (!is.logical(center) || length(center) != 1L || is.na(center)) {
    stop("'center' must be TRUE or FALSE.")
  }
  if (!is.logical(scale) || length(scale) != 1L || is.na(scale)) {
    stop("'scale' must be TRUE or FALSE.")
  }
  
  if (is_pls && is.null(Yr)) {
    stop("'Yr' is required for PLS methods ('pls', 'mpls').")
  }
  
  if (inherits(ncomp, "ncomp_by_opc") && is.null(Yr)) {
    stop("'Yr' is required when using ncomp_by_opc().")
  }
  
  # ---------------------------------------------------------------------------
  # Dispatch to implementation
  # ---------------------------------------------------------------------------
  if (is_pls) {
    proj <- .pls_projection_impl(
      Xr = Xr, Xu = Xu, Yr = Yr,
      ncomp = ncomp,
      method = method,
      scale = scale,
      tol = tol,
      max_iter = max_iter
    )
  } else {
    proj <- .pca_projection_impl(
      Xr = Xr, Xu = Xu, Yr = Yr,
      ncomp = ncomp,
      method = method,
      center = center,
      scale = scale,
      tol = tol,
      max_iter = max_iter
    )
  }
  
  proj$method <- method
  proj$ncomp_method <- ncomp
  class(proj) <- c("ortho_projection", "list")
  proj
}


# =============================================================================
# Deprecated wrappers
# =============================================================================

#' @keywords internal
#' @noRd
pc_projection <- function(
    Xr, Xu = NULL, Yr = NULL,
    pc_selection = list(method = "var", value = 0.01),
    center = TRUE, scale = FALSE,
    method = "pca",
    tol = 1e-6, max_iter = 1000,
    ...
) {
  deprecate_warn(
    when = "3.0.0",
    what = "pc_projection()",
    with = "ortho_projection()"
  )
  
  ncomp <- .convert_pc_selection_to_ncomp(pc_selection)
  new_method <- if (method == "pca.nipals") "pca_nipals" else "pca"
  
  ortho_projection(
    Xr = Xr, Xu = Xu, Yr = Yr,
    ncomp = ncomp,
    method = new_method,
    center = center,
    scale = scale,
    tol = tol,
    max_iter = max_iter
  )
}


#' @keywords internal
#' @noRd
pls_projection <- function(
    Xr, Xu = NULL, Yr,
    pc_selection = list(method = "opc", value = min(dim(Xr), 40)),
    scale = FALSE,
    method = "pls",
    tol = 1e-6, max_iter = 1000,
    ...
) {
  deprecate_warn(
    when = "3.0.0",
    what = "pls_projection()",
    with = "ortho_projection()"
  )
  
  ncomp <- .convert_pc_selection_to_ncomp(pc_selection)
  
  ortho_projection(
    Xr = Xr, Xu = Xu, Yr = Yr,
    ncomp = ncomp,
    method = method,
    scale = scale,
    tol = tol,
    max_iter = max_iter
  )
}


# =============================================================================
# predict method
# =============================================================================

#' @rdname ortho_projection
#' @export
predict.ortho_projection <- function(object, newdata, ...) {
  if (missing(newdata)) {
    return(object$scores)
  }
  
  nms_orig <- colnames(object$X_loadings)
  nms_new <- colnames(newdata)
  
  if (any(!nms_orig %in% nms_new)) {
    stop("Variables missing in 'newdata' that are required for projection.")
  }
  
  is_pca <- grepl("pca", object$method)
  
  if (is_pca) {
    newdata <- sweep(newdata, 2L, object$center, FUN = "-")
    newdata <- sweep(newdata, 2L, object$scale, FUN = "/")
    scores <- newdata %*% t(object$X_loadings)
    colnames(scores) <- rownames(object$X_loadings)
    rownames(scores) <- rownames(newdata)
    return(scores)
  } else {
    scores <- project_opls(
      projection_mat = object$projection_mat,
      ncomp = ncol(object$projection_mat),
      newdata = newdata,
      scale = TRUE,
      Xcenter = object$center,
      Xscale = object$scale
    )
    colnames(scores) <- paste0("pls_", seq_len(ncol(scores)))
    rownames(scores) <- rownames(newdata)
    return(scores)
  }
}


# =============================================================================
# Helper: Convert legacy pc_selection to ncomp_selection
# =============================================================================

.convert_pc_selection_to_ncomp <- function(pc_selection) {
  # Handle character shorthand
  if (is.character(pc_selection) && length(pc_selection) == 1L) {
    pc_selection <- switch(pc_selection,
                           opc = list(method = "opc", value = 40L),
                           cumvar = list(method = "cumvar", value = 0.99),
                           var = list(method = "var", value = 0.01),
                           manual = list(method = "manual", value = 40L),
                           stop("Unknown pc_selection method: ", pc_selection)
    )
  }
  
  sel_method <- if (!is.null(pc_selection$method)) {
    pc_selection$method
  } else {
    pc_selection[[1]]
  }
  
  value <- if (!is.null(pc_selection$value)) {
    pc_selection$value
  } else {
    pc_selection[[2]]
  }
  
  switch(sel_method,
         manual = ncomp_fixed(as.integer(value)),
         var = ncomp_by_var(min_var = value, max_ncomp = 40L),
         cumvar = ncomp_by_cumvar(min_cumvar = value, max_ncomp = 40L),
         opc = ncomp_by_opc(max_ncomp = as.integer(value)),
         stop("Unknown pc_selection method: ", sel_method)
  )
}




# =============================================================================
# Helper: Get max_ncomp from ncomp_selection object
# =============================================================================

.get_max_ncomp <- function(ncomp) {
  if (inherits(ncomp, "ncomp_fixed")) {
    return(ncomp$ncomp)
  }
  ncomp$max_ncomp
}

# =============================================================================
# Helper: Bridge ncomp_selection to legacy pc_selection list
# =============================================================================

.ncomp_to_pc_selection <- function(ncomp_obj) {
  stopifnot(inherits(ncomp_obj, "ncomp_selection"))
  
  switch(
    class(ncomp_obj)[[1]],
    ncomp_fixed = list(method = "manual", value = ncomp_obj$ncomp),
    ncomp_by_var = list(method = "var", value = ncomp_obj$min_var),
    ncomp_by_cumvar = list(method = "cumvar", value = ncomp_obj$min_cumvar),
    ncomp_by_opc = list(method = "opc", value = ncomp_obj$max_ncomp),
    stop("Unknown ncomp_selection class: ", class(ncomp_obj)[[1]])
  )
}



# =============================================================================
# Internal PCA implementation
# =============================================================================

.pca_projection_impl <- function(
    Xr, Xu, Yr, ncomp, method, center, scale, tol, max_iter
) {
  # Bridge to legacy format
  pc_selection <- .ncomp_to_pc_selection(ncomp)
  pc_selection_method <- pc_selection$method
  max_comp <- .get_max_ncomp(ncomp)
  
  # Validation
  if (!is.null(Yr)) {
    Yr <- as.matrix(Yr)
    if (nrow(Yr) != nrow(Xr)) {
      stop("Number of rows in 'Yr' must match 'Xr'.")
    }
  }
  
  ny <- if (!is.null(Yr)) ncol(Yr) else 0L
  
  if (!is.null(Xu) && ncol(Xr) != ncol(Xu)) {
    stop("Number of columns in 'Xr' and 'Xu' must match.")
  }
  
  effective_rows_xr <- nrow(Xr)
  effective_rows_xu <- if (is.null(Xu)) 0L else nrow(Xu)
  n_cols <- ncol(Xr)
  
  # Combine Xr and Xu
  X <- rbind(Xr, Xu)
  
  # Cap max_comp at data dimensions
  max_comp <- min(max_comp, nrow(X), ncol(X))
  
  # Center
  if (center) {
    mean_vector <- colMeans(X)
    X0 <- sweep(X, 2L, mean_vector, FUN = "-")
  } else {
    mean_vector <- rep(0, n_cols)
    X0 <- X
  }
  
  # Scale
  if (scale) {
    sd_vector <- get_column_sds(X0)
    X0 <- sweep(X0, 2L, sd_vector, FUN = "/")
  } else {
    sd_vector <- rep(1, n_cols)
  }
  
  # Compute PCA
  if (method == "pca") {
    sv <- svd(X0, nu = max_comp, nv = max_comp)
    sv$d <- sv$d[1:max_comp]
    
    if (length(sv$d) == 1L) {
      pc_scores <- sv$u %*% as.matrix(sv$d)
    } else {
      pc_scores <- sv$u %*% diag(sv$d)
    }
    pc_loadings <- t(sv$v)
    
    xvariance <- overall_var(X0)
    explained_v <- (sv$d)^2 / xvariance
    cumulative_v <- cumsum(explained_v)
    variance <- rbind(
      var = (sv$d)^2,
      explained_var = explained_v,
      cumulative_explained_var = cumulative_v
    )
  } else {
    # pca_nipals
    nipals_result <- pca_nipals(
      X = X0,
      ncomp = max_comp,
      center = FALSE,
      scale = FALSE,
      maxiter = max_iter,
      tol = tol,
      pcSelmethod = pc_selection_method,
      pcSelvalue = pc_selection$value
    )
    
    pc_scores <- nipals_result$pc_scores
    pc_loadings <- nipals_result$pc_loadings
    xvariance <- nipals_result$original_x_variance
    variance <- rbind(
      var = nipals_result$pc_variance[1, ],
      explained_var = nipals_result$pc_variance[2, ],
      cumulative_explained_var = nipals_result$pc_variance[3, ]
    )
  }
  
  # Assign names
  colnames(pc_scores) <- paste0("pc_", seq_len(ncol(pc_scores)))
  rownames(pc_scores) <- c(
    paste0("Xr_", seq_len(effective_rows_xr)),
    if (effective_rows_xu > 0) paste0("Xu_", seq_len(effective_rows_xu))
  )
  colnames(pc_loadings) <- colnames(X0)
  rownames(pc_loadings) <- paste0("pc_", seq_len(nrow(pc_loadings)))
  colnames(variance) <- paste0("pc_", seq_len(ncol(variance)))
  
  # Select components
  opc_evaluation <- NULL
  
  if (pc_selection_method == "opc") {
    if (is.null(Yr) || !is.matrix(Yr)) {
      stop("'Yr' must be a matrix when using ncomp_by_opc().")
    }
    if (nrow(Yr) != effective_rows_xr) {
      stop("Number of rows in 'Yr' must match 'Xr'.")
    }
    if (!is.null(colnames(Yr)) && any(duplicated(colnames(Yr)))) {
      stop("Column names in 'Yr' must be unique.")
    }
    if (is.null(colnames(Yr))) {
      colnames(Yr) <- paste0("Yr_", seq_len(ny))
    }
    
    results <- eval_multi_pc_diss(
      pc_scores[seq_len(effective_rows_xr), 1:max_comp, drop = FALSE],
      side_info = Yr,
      method = "pc",
      check_dims = FALSE
    )
    selected_pcs <- results$best_pc
    opc_evaluation <- results$result
    
  } else if (pc_selection_method == "cumvar") {
    selected_pcs <- sum(variance["cumulative_explained_var", ] < pc_selection$value) + 1L
    selected_pcs <- min(selected_pcs, ncol(variance))
    
  } else if (pc_selection_method == "var") {
    selected_pcs <- sum(variance["explained_var", ] >= pc_selection$value)
    selected_pcs <- max(1L, selected_pcs)
    
  } else {
    # manual
    selected_pcs <- min(pc_selection$value, ncol(pc_scores))
  }
  
  # Subset to selected components
  scores_sd <- get_column_sds(pc_scores[, 1:selected_pcs, drop = FALSE])
  colnames(scores_sd) <- colnames(pc_scores)[1:selected_pcs]
  rownames(scores_sd) <- "sd"
  
  result <- list(
    scores = pc_scores[, 1:selected_pcs, drop = FALSE],
    X_loadings = pc_loadings[1:selected_pcs, , drop = FALSE],
    variance = list(
      original_x_var = xvariance,
      x_var = variance[, 1:selected_pcs, drop = FALSE]
    ),
    scores_sd = scores_sd,
    ncomp = selected_pcs,
    center = mean_vector,
    scale = sd_vector
  )
  
  if (!is.null(opc_evaluation)) {
    result$opc_evaluation <- opc_evaluation
  }
  
  result
}


# =============================================================================
# Internal PLS implementation
# =============================================================================

.pls_projection_impl <- function(
    Xr, Xu, Yr, ncomp, method, scale, tol, max_iter
) {
  # Bridge to legacy format
  pc_selection <- .ncomp_to_pc_selection(ncomp)
  pc_selection_method <- pc_selection$method
  max_comp <- .get_max_ncomp(ncomp)
  
  # Validation
  Yr <- as.matrix(Yr)
  if (!is.numeric(Yr)) {
    stop("'Yr' must be numeric.")
  }
  if (nrow(Yr) != nrow(Xr)) {
    stop("Number of rows in 'Yr' must match 'Xr'.")
  }
  
  if (!is.null(Xu) && ncol(Xr) != ncol(Xu)) {
    stop("Number of columns in 'Xr' and 'Xu' must match.")
  }
  
  ny <- ncol(Yr)
  # Handle missing values in Yr
  nas <- rowSums(is.na(Yr)) > 0
  X0 <- Xr
  Y0 <- Yr
  non_nas_yr <- seq_len(nrow(Xr))
  nas_yr <- integer(0)
  Xout <- NULL
  if (any(nas)) {
    nas_yr <- which(nas)
    non_nas_yr <- which(!nas)
    Xout <- Xr[nas_yr, , drop = FALSE]
    X0 <- Xr[non_nas_yr, , drop = FALSE]
    Y0 <- Yr[non_nas_yr, , drop = FALSE]
    
    if (pc_selection_method %in% c("opc", "manual")) {
      if (min(dim(X0)) < max_comp) {
        stop(
          "Missing values in 'Yr' reduce available observations. ",
          "The number of components exceeds available observations."
        )
      }
    }
  }
  effective_rows_xr <- nrow(X0)
  max_comp <- min(max_comp, effective_rows_xr, ncol(X0))
  
  # Set up for C++ backend
  cpp_method <- if (pc_selection_method %in% c("opc", "manual")) {
    "manual"
  } else {
    pc_selection_method
  }
  
  cpp_value <- if (pc_selection_method %in% c("opc", "manual")) {
    max_comp - 1L
  } else {
    pc_selection$value
  }
  # Run PLS
  plsp <- opls_for_projection(
    X = X0,
    Y = Y0,
    ncomp = max_comp,
    scale = scale,
    maxiter = max_iter,
    tol = tol,
    pcSelmethod = cpp_method,
    pcSelvalue = cpp_value,
    algorithm = method
  )
  
  max_comp <- plsp$ncomp
  
  # OPC selection if requested
  opc_evaluation <- NULL
  if (pc_selection_method == "opc") {
    if (is.null(colnames(Yr))) {
      colnames(Y0) <- colnames(Yr) <- paste0("Yr_", seq_len(ny))
    }
    if (any(duplicated(colnames(Yr)))) {
      stop("Column names in 'Yr' must be unique.")
    }
    
    results <- eval_multi_pc_diss(
      plsp$scores[, 1:max_comp, drop = FALSE],
      side_info = Y0,
      method = "pls",
      check_dims = FALSE
    )
    max_comp <- results$best_pc
    opc_evaluation <- results$result
  }
  
  # Extract selected components
  pls_variance <- plsp$variance$x_var[, 1:max_comp, drop = FALSE]
  weights <- plsp$weights[1:max_comp, , drop = FALSE]
  scores <- plsp$scores[, 1:max_comp, drop = FALSE]
  X_loadings <- plsp$X_loadings[1:max_comp, , drop = FALSE]
  Y_loadings <- plsp$Y_loadings[1:max_comp, , drop = FALSE]
  projection_mat <- plsp$projection_mat[, 1:max_comp, drop = FALSE]
  
  # Assign names
  comp_names <- paste0("pls_", seq_len(max_comp))
  rownames(projection_mat) <- colnames(X_loadings) <- colnames(X0)
  colnames(projection_mat) <- rownames(X_loadings) <- comp_names
  rownames(Y_loadings) <- comp_names
  colnames(Y_loadings) <- paste0("Y_pls_", seq_len(ny))
  rownames(weights) <- comp_names
  colnames(weights) <- colnames(X0)
  colnames(pls_variance) <- comp_names
  rownames(pls_variance) <- c("var", "explained_var_X", "cumulative_explained_var_X")
  
  yex <- plsp$variance$y_var[, 1:max_comp, drop = FALSE]
  colnames(yex) <- comp_names
  if (ny > 1) {
    rownames(yex) <- paste0("cumulative_explained_var_", colnames(Yr))
  } else {
    rownames(yex) <- "cumulative_explained_var_Yr"
  }
  
  # Handle observations with missing Yr
  if (length(nas_yr) > 0) {
    scores_full <- matrix(NA_real_, nrow(Xr), ncol(scores))
    scores_full[nas_yr, ] <- project_opls(
      projection_mat = projection_mat,
      ncomp = max_comp,
      newdata = Xout,
      scale = scale,
      Xcenter = plsp$transf$Xcenter,
      Xscale = plsp$transf$Xscale
    )
    scores_full[non_nas_yr, ] <- scores
    scores <- scores_full
  }
  
  rownames(scores) <- paste0("Xr_", seq_len(nrow(scores)))
  colnames(scores) <- comp_names
  
  # Project Xu if provided
  if (!is.null(Xu)) {
    if (is.vector(Xu)) Xu <- t(Xu)
    scores_Xu <- project_opls(
      projection_mat = projection_mat,
      ncomp = max_comp,
      newdata = Xu,
      scale = scale,
      Xcenter = plsp$transf$Xcenter,
      Xscale = plsp$transf$Xscale
    )
    colnames(scores_Xu) <- comp_names
    rownames(scores_Xu) <- paste0("Xu_", seq_len(nrow(Xu)))
    scores <- rbind(scores, scores_Xu)
  }
  
  # Ensure scale is a matrix
  if (!nrow(plsp$transf$Xscale)) {
    plsp$transf$Xscale <- matrix(1, 1L, length(plsp$transf$Xcenter))
  }
  
  scores_sd <- get_column_sds(scores)
  colnames(scores_sd) <- colnames(scores)
  rownames(scores_sd) <- "sd"
  
  result <- list(
    scores = scores,
    X_loadings = X_loadings,
    Y_loadings = Y_loadings,
    weights = weights,
    projection_mat = projection_mat,
    variance = list(
      original_x_var = plsp$variance$original_x_var,
      x_var = pls_variance,
      y_var = yex
    ),
    scores_sd = scores_sd,
    ncomp = max_comp,
    center = plsp$transf$Xcenter,
    scale = plsp$transf$Xscale
  )
  
  if (!is.null(opc_evaluation)) {
    result$opc_evaluation <- opc_evaluation
  }
  
  result
}


#' @aliases ortho_projection
#' @importFrom graphics barplot grid par plot segments
#' @export
plot.ortho_projection <- function(x, col = "#3B82F6", ...) {
  
  if (inherits(x$ncomp_method, "ncomp_by_opc")) {
    .plot_opc_evaluation(x, col = col, ...)
  } else {
    .plot_variance(x, col = col, ...)
  }
  
  invisible(NULL)
}

# --- internal: OPC evaluation plot --------------------------------------------
.plot_opc_evaluation <- function(x, col, ...) {
  opc_eval <- x$opc_evaluation
  tpl <- opc_eval[, c(1, ncol(opc_eval))]
  metric <- colnames(tpl)[2]
  
  ylab <- switch(
    metric,
    mean_standardized_rmsd_Yr = "Mean standardized RMSD",
    rmsd_Yr = "RMSD",
    rmsd = "RMSD",
    kappa = "Kappa",
    metric
  )
  
  plot(
    tpl,
    type = "p",
    pch = 16,
    col = col,
    ylab = ylab,
    xlab = "Number of components",
    las = 1,
    ...
  )
  grid(col = "#33415580", lty = 1)
  segments(tpl[, 1], 0, tpl[, 1], tpl[, 2], col = col)
  
  # Mark optimal
  opt_idx <- which.min(tpl[, 2])
  if (metric == "kappa") {
    opt_idx <- which.max(tpl[, 2])
  }
  points(tpl[opt_idx, 1], tpl[opt_idx, 2], pch = 16, col = "#F59E0B", cex = 1.5)
}

# --- internal: variance plot --------------------------------------------------
.plot_variance <- function(x, col, ...) {
  x_var <- x$variance$x_var
  
  opar <- par(mfrow = c(1, 2), mar = c(4, 4, 2, 1))
  on.exit(par(opar))
  
  # Individual variance
  ind_var <- x_var[grep("^explained_var", rownames(x_var)), ]
  barplot(
    ind_var,
    horiz = FALSE,
    names.arg = colnames(x_var),
    ylim = c(0, 1),
    ylab = "Explained variance",
    xlab = "Component",
    col = col,
    border = NA,
    las = 1,
    ...
  )
  
  # Cumulative variance
  cum_var <- x_var[grep("cumulative", rownames(x_var)), ]
  barplot(
    cum_var,
    horiz = FALSE,
    names.arg = colnames(x_var),
    ylim = c(0, 1),
    ylab = "Cumulative explained variance",
    xlab = "Component",
    col = col,
    border = NA,
    las = 1,
    ...
  )
}

context("test-print.ortho_projection")

.setup_nirsoil_data <- function(n_xr = 40, n_xu = 20) {
  data("NIRsoil", package = "prospectr")
  
  Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
  Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
  Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  Xu <- Xu[!is.na(Yu), ][seq_len(n_xu), ]
  Xr <- Xr[!is.na(Yr), ][seq_len(n_xr), ]
  Yu <- Yu[!is.na(Yu)][seq_len(n_xu)]
  Yr <- Yr[!is.na(Yr)][seq_len(n_xr)]
  
  list(Xr = Xr, Xu = Xu, Yr = Yr, Yu = Yu)
}


# =============================================================================
# Tests for print.ortho_projection
# =============================================================================

test_that("print.ortho_projection works for PCA", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, ncomp = 5)
  
  expect_output(print(proj), "Method:")
  expect_output(print(proj), "pca")
  expect_output(print(proj), "Number of components retained:")
  expect_output(print(proj), "5")
  expect_output(print(proj), "Original variance")
  expect_output(print(proj), "Explained variance")
})

test_that("print.ortho_projection works for PCA with Xu", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Xu = d$Xu, ncomp = 5)
  
  expect_output(print(proj), "Xr; Xu")
})

test_that("print.ortho_projection works for PLS", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Yr = d$Yr, method = "pls", ncomp = 5)
  
  expect_output(print(proj), "Method:")
  expect_output(print(proj), "pls")
  expect_output(print(proj), "Explained variance in Yr")
})

test_that("print.ortho_projection works for mpls", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  d <- .setup_nirsoil_data()
  
  proj <- ortho_projection(d$Xr, Yr = d$Yr, method = "mpls", ncomp = 5)
  
  expect_output(print(proj), "mpls")
  expect_output(print(proj), "Explained variance in Yr")
})

# #' @title Print method for an object of class \code{ortho_diss}
# #' @description Prints the content of an object of class \code{ortho_diss}
# #' @usage \method{print}{local_ortho_diss}(x, ...)
# #' @param x an object of class \code{local_ortho_diss} (returned by
# #' \code{ortho_diss} when it uses \code{.local = TRUE}).
# #' @param ... arguments to be passed to methods (not yet functional).
# #' @author Leonardo Ramirez-Lopez and Antoine Stevens
# #' @keywords internal
# #' @export
# #' @noRd

######################################################################
# resemble
# Copyright (C) 2014 Leonardo Ramirez-Lopez and Antoine Stevens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
######################################################################

## History:
## 09.03.2014 Leo     The tryMod function was removed

# print.local_ortho_diss <- function(x, ...) {
#   obj <- x
#   if (is.list(obj)) {
#     obj$dissimilarity <- as.matrix(round(obj$dissimilarity, getOption("digits")))
#     obj$dissimilarity[is.na(obj$dissimilarity)] <- "*"
#     dm <- format(obj$dissimilarity, digits = getOption("digits"), justify = "right")
#     print(list(
#       n.components = object$ncomp,
#       loc_ncomp = object$neighborhood_info$local_ncomp,
#       dissimilarity = noquote(dm)
#     ))
#     cat("*: Not a (local) neighbor")
#   }
#   if (is.matrix(obj)) {
#     object <- as.matrix(round(obj, getOption("digits")))
#     object[is.na(obj)] <- "*"
#     dm <- format(object, digits = getOption("digits"), justify = "right")
#     print(dm, quote = FALSE)
#     cat("*: not a neighbor")
#   }
# }

# #' @title Print method for an object of class \code{local_ortho_diss}
# #' @param x \code{local_ortho_diss} matrix
# # #' @param rows the indices of the rows
# #' @param columns the indices of the columns
# #' @param drop drop argument
# #' @param ... not used
# #' @description prints the subsets of local_ortho_diss objects
# #' @keywords internal
# #' @noRd
# #' @export
# "[.local_ortho_diss" <- function(x, rows, columns, drop = FALSE, ...) {
#   object <- x
#   if (!is.logical(drop)) {
#     drop <- FALSE
#   }
#   class(object) <- NULL
#   obj <- object[rows, columns, drop = drop]
#   if (!drop) {
#     class(obj) <- "local_ortho_diss"
#   }
#   return(obj)
# }

#' @title Evaluate dissimilarity matrices
#'
#' @description
#' \loadmathjax
#' `r lifecycle::badge("stable")`
#'
#' Evaluates a dissimilarity matrix by comparing each observation to its
#' nearest neighbor based on side information. For continuous variables,
#' RMSD and correlation are computed; for categorical variables, the kappa
#' index is used.
#' 
#' @usage
#' diss_evaluate(diss, side_info)
#' 
#' @param diss A symmetric dissimilarity matrix. Alternatively, a vector
#'   containing the lower triangle values (as returned by \code{\link[stats]{dist}}).
#' @param side_info A matrix of side information corresponding to the
#'   observations. Can be numeric (one or more columns) or character (single
#'   column for categorical data).
#'
#' @details
#' This function assesses whether a dissimilarity matrix captures meaningful
#' structure by examining the side information of nearest neighbor pairs
#' (Ramirez-Lopez et al., 2013). If observations that are similar in the
#' dissimilarity space also have similar side information values, the
#' dissimilarity is considered effective.
#'
#' For numeric \code{side_info}, the root mean square of differences (RMSD)
#' between each observation and its nearest neighbor is computed:
#'
#' \mjdeqn{j(i) = NN(x_i, X^{\{-i\}})}{j(i) = NN(x_i, X^{-i})}
#' \mjdeqn{RMSD = \sqrt{\frac{1}{m} \sum_{i=1}^{m} (y_i - y_{j(i)})^2}}{RMSD = sqrt(1/m sum (y_i - y_j(i))^2)}
#'
#' where \mjeqn{NN(x_i, X^{-i})}{NN(x_i, X^{-i})} returns the index of the
#' nearest neighbor of observation \mjeqn{i}{i} (excluding itself),
#' \mjeqn{y_i}{y_i} is the side information value for observation \mjeqn{i}{i},
#' and \mjeqn{m}{m} is the number of observations.
#'
#' For categorical \code{side_info}, the kappa index is computed:
#'
#' \mjdeqn{\kappa = \frac{p_o - p_e}{1 - p_e}}{kappa = (p_o - p_e) / (1 - p_e)}
#'
#' where \mjeqn{p_o}{p_o} is the observed agreement and \mjeqn{p_e}{p_e} is the
#' agreement expected by chance.
#'
#' @return A list with the following components:
#' \describe{
#'   \item{eval}{For numeric side information: a matrix with columns
#'     \code{rmsd} and \code{r} (correlation). For categorical: a matrix
#'     with column \code{kappa}.}
#'   \item{global_eval}{If multiple numeric side information variables are
#'     provided, summary statistics across variables.}
#'   \item{first_nn}{A matrix with the original side information and the
#'     side information of each observation's nearest neighbor.}
#' }
#'
#' @seealso \code{\link{dissimilarity}}, \code{\link{ncomp_by_opc}}
#'
#' @references
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex datasets. Geoderma 195-196,
#' 268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199, 43-53.
#'
#' @author
#' \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' sg <- savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
#' NIRsoil$spc <- sg
#'
#' Yr <- NIRsoil$Nt[as.logical(NIRsoil$train)]
#' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#'
#' # Compute PCA-based dissimilarity
#' d <- dissimilarity(Xr, diss_method = diss_pca(ncomp = 8))
#'
#' # Evaluate using side information
#' ev <- diss_evaluate(d$dissimilarity, side_info = as.matrix(Yr))
#' ev$eval
#'
#' # Evaluate with multiple side information variables
#' Yr_2 <- NIRsoil$CEC[as.logical(NIRsoil$train)]
#' ev_2 <- diss_evaluate(d$dissimilarity, side_info = cbind(Yr, Yr_2))
#' ev_2$eval
#' ev_2$global_eval
#' }
#'
#' @export
diss_evaluate <- function(diss, side_info) {
  # ---------------------------------------------------------------------------
  # Validate side_info
  # ---------------------------------------------------------------------------
  if (!is.matrix(side_info)) {
    stop("'side_info' must be a matrix.", call. = FALSE)
  }
  
  if (is.character(side_info)) {
    if (ncol(side_info) > 1L) {
      stop(
        "For categorical 'side_info', only one column is allowed.",
        call. = FALSE
      )
    }
    side_info <- as.factor(side_info)
    if (nlevels(side_info) < 2L) {
      stop(
        "'side_info' must have at least two categories.",
        call. = FALSE
      )
    }
    get_eval <- .eval_categorical
  } else {
    get_eval <- .eval_continuous
  }
  
  if (any(colSums(!is.na(side_info)) < 4L)) {
    stop(
      "Each 'side_info' variable must have at least 4 non-missing values.",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # Validate diss
  # ---------------------------------------------------------------------------
  if (!is.vector(diss) && !is.matrix(diss)) {
    stop("'diss' must be a matrix or a vector.", call. = FALSE)
  }
  
  if (is.vector(diss)) {
    expected_len <- (nrow(side_info)^2 - nrow(side_info)) / 2
    if (length(diss) != expected_len) {
      stop(
        "Length of 'diss' does not match a lower triangular matrix ",
        "for ", nrow(side_info), " observations.",
        call. = FALSE
      )
    }
    find_nearest <- which_min_vector
  } else {
    if (nrow(diss) != ncol(diss)) {
      stop("'diss' matrix must be square.", call. = FALSE)
    }
    if (nrow(diss) != nrow(side_info)) {
      stop(
        "Number of rows in 'diss' must match 'side_info'.",
        call. = FALSE
      )
    }
    find_nearest <- which_min
  }
  
  # ---------------------------------------------------------------------------
  # Find nearest neighbors and evaluate
  # ---------------------------------------------------------------------------
  nearest_idx <- find_nearest(diss)
  
  cnms <- colnames(side_info)
  if (is.null(cnms)) {
    cnms <- paste0("side_info_", seq_len(ncol(side_info)))
  }
  eval_result <- get_eval(y = side_info, indices_closest = nearest_idx)
  rownames(eval_result) <- cnms
  
  # ---------------------------------------------------------------------------
  # Build output
  # ---------------------------------------------------------------------------
  first_nn <- side_info[nearest_idx, , drop = FALSE]
  colnames(first_nn) <- paste0("nn_", cnms)
  
  if (ncol(side_info) > 1L) {
    sds <- get_column_sds(side_info[complete.cases(side_info), , drop = FALSE])
    global_eval <- data.frame(
      mean_standardized_rmsd = mean(eval_result[, "rmsd"] / sds),
      mean_r = mean(eval_result[, "r"])
    )
    
    result <- list(
      eval = eval_result,
      global_eval = global_eval,
      first_nn = cbind(side_info, first_nn)
    )
  } else {
    result <- list(
      eval = eval_result,
      first_nn = cbind(side_info, first_nn)
    )
  }
  
  result
}


#' @rdname diss_evaluate
#' @param d Deprecated. Use \code{diss} in \code{diss_evaluate()} instead.
#' @export
sim_eval <- function(d, side_info) {
  .Deprecated("diss_evaluate")
  diss_evaluate(diss = d, side_info = side_info)
}


# =============================================================================
# Internal helpers
# =============================================================================

#' @keywords internal
.eval_categorical <- function(y, indices_closest) {
  tab <- as.matrix(table(y[!is.na(y)], y[indices_closest][!is.na(y)]))
  total <- sum(tab)
  tab <- tab / total
  p <- rowSums(tab) %*% t(colSums(tab))
  pra <- sum(diag(tab))
  pre <- sum(diag(p))
  kappa <- t((pra - pre) / (1 - pre))
  colnames(kappa) <- "kappa"
  kappa
}

#' @keywords internal
#' @importFrom stats cor
.eval_continuous <- function(y, indices_closest) {
  eval_one <- function(col_idx) {
    obs <- y[, col_idx]
    nn <- y[indices_closest, col_idx]
    rmsd <- sqrt(mean((nn - obs)^2, na.rm = TRUE))
    if (sd(nn, na.rm = TRUE) == 0 || sd(obs, na.rm = TRUE) == 0) {
      r <- NA_real_
    } else {
      r <- cor(nn, obs, use = "complete.obs")
    }
    c(rmsd = rmsd, r = r)
  }
  
  result <- vapply(
    seq_len(ncol(y)),
    eval_one,
    FUN.VALUE = numeric(2)
  )
  t(result)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @title A fast distance algorithm for two matrices written in C++ 
#' @description Computes distances between two data matrices using 
#' "euclid", "cor", "cosine" 
#' @usage 
#' fast_diss(X, Y, method)
#' @param X a matrix
#' @param Y a matrix
#' @param method a \code{string} with possible values "euclid", "cor", "cosine"
#' @return a distance matrix
#' @keywords internal
#' @noRd
#' @author Antoine Stevens and Leonardo Ramirez-Lopez
fast_diss <- function(X, Y, method) {
    .Call(`_resemble_fast_diss`, X, Y, method)
}

#' @title A fast algorithm of (squared) Euclidean cross-distance for vectors written in C++ 
#' @description A fast (parallel for linux) algorithm of (squared) Euclidean cross-distance for vectors written in C++ 
#' @usage 
#' fast_diss_vector(X)
#' @param X a vector.
#' @return a vector of distance (lower triangle of the distance matrix, stored by column)
#' @details used internally in ortho_projection
#' @author Antoine Stevens
#' @keywords internal 
#' @noRd
fast_diss_vector <- function(X) {
    .Call(`_resemble_fast_diss_vector`, X)
}

#' @title Symmetric Euclidean Distance Matrix from a Single Matrix
#' @description Computes the symmetric pairwise Euclidean distance matrix
#' from the rows of a single matrix. Only the upper triangle is computed,
#' then mirrored to the lower triangle to reduce redundant operations.
#'
#' Squared distances are calculated first, then square-rooted. Tiny values
#' below 1e-14 are clamped to zero to ensure numerical stability.
#'
#' @param X An `n x p` numeric matrix where each row is a p-dimensional
#' observation.
#'
#' @return An `n x n` symmetric matrix of Euclidean distances.
#'
#' @details
#' This function is designed for efficient in-memory computation of full
#' distance matrices using only the upper triangle to avoid redundant work.
#' It applies the square root only once per distance and avoids unnecessary
#' writes to the diagonal (which is always 0). The implementation leverages
#' memory locality and avoids data duplication.
#'
#' Suitable for moderate to large `n`. If OpenMP is enabled, the upper-triangle
#' loop can optionally be parallelized for additional speedup on multicore systems.
#' Note that the function assumes that the input matrix `X` is not empty and
#' @keywords internal
#' @noRd
fast_self_euclid <- function(X) {
    .Call(`_resemble_fast_self_euclid`, X)
}

#' @title Rolling correlation distance between X and Y
#' @param X Numeric matrix m×T
#' @param Y Numeric matrix n×T
#' @param w Window size (odd)
#' @param block_x Tile size for rows of X (default 1024)
#' @param block_y Tile size for rows of Y (default 1024)
#' @return n×m distance matrix (R double matrix)
#' @noRd
#' @keywords internal
moving_cor_diss_xy <- function(X, Y, w, block_x = 1024L, block_y = 1024L) {
    .Call(`_resemble_moving_cor_diss_xy`, X, Y, w, block_x, block_y)
}

#' @title Rolling correlation distance within X (upper-triangle, tiled, serial-optimized)
#' @description Mean rolling-window correlation distance among rows of X.
#'              Serial only; computes upper-triangle tiles densely and mirrors.
#' @param X Numeric matrix (m x T)
#' @param w Odd window size
#' @param block_rows Tile size in rows (default 1024)
#' @return m x m symmetric distance matrix
#' @keywords internal
#' @noRd
moving_cor_diss_self_f64 <- function(X, w, block_rows = 1024L) {
    .Call(`_resemble_moving_cor_diss_self_f64`, X, w, block_rows)
}

#' @title Rolling correlation distance within X (templated, OpenMP; precision)
#' @param X Numeric matrix (m x T)
#' @param w Window size (odd in)
#' @param block_rows Tile size (default 1024)
#' @return m×m distance matrix (double for R)
#' @keywords internal
#' @noRd
moving_cor_diss_self <- function(X, w, block_rows = 1024L) {
    .Call(`_resemble_moving_cor_diss_self`, X, w, block_rows)
}

#' @title A function to compute row-wise index of minimum values of a square distance matrix
#' @description For internal use only
#' @usage 
#' which_min(X)
#' @param X a square matrix of distances
#' @return a vector of the indices of the minimum value in each row of the input matrix
#' @details Used internally to find the nearest neighbors
#' @keywords internal
#' @noRd
#' @author Antoine Stevens 
which_min <- function(X) {
    .Call(`_resemble_which_min`, X)
}

#' @title A function to compute indices of minimum values of a distance vector
#' @description For internal use only
#' @usage 
#' which_min_vector(X)
#' @param X a vector of distances 
#' @return a vector of the indices of the nearest neighbors
#' @details 
#' Used internally to find the nearest neighbors. 
#' It searches in lower (or upper) triangular matrix. Therefore this must be the format of the 
#' input data. The piece of code int \code{len = (sqrt(X.size()*8+1)+1)/2} generated an error in CRAN
#' since \code{sqrt} cannot be applied to integers.
#' @keywords internal
#' @noRd
#' @author Antoine Stevens 
which_min_vector <- function(X) {
    .Call(`_resemble_which_min_vector`, X)
}

#' @title Top-k Order Indices for Matrix Columns
#' @description 
#' Internal helper function. For each column in a numeric matrix, 
#' returns the indices of the top-\code{k} smallest values (i.e., closest neighbors).
#'
#' @details 
#' This function is implemented in C++ using \code{Rcpp}. It performs a partial sort 
#' on each column to retrieve the \code{k} smallest elements efficiently. This is 
#' particularly useful when working with large dissimilarity matrices, where 
#' performance is critical.
#'
#' @param mat A numeric matrix (e.g., dissimilarity matrix).
#' @param k Integer. Number of top smallest elements to return for each column.
#'
#' @return An integer matrix of dimensions \code{k × ncol(mat)}. Each column contains 
#' the 1-based row indices of the top-\code{k} smallest values in the corresponding column of \code{mat}.
#'
#' @keywords internal
#' @noRd
top_k_order <- function(mat, k, skip = as.integer( c())) {
    .Call(`_resemble_top_k_order`, mat, k, skip)
}

#' @title Extract Column Elements by Row Indices
#' @description
#' Efficiently extracts specific elements from each column of a numeric matrix 
#' based on provided row indices, using Rcpp for performance.
#'
#' @param mat A numeric matrix from which values will be extracted.
#' @param idx An integer matrix of the same number of columns as \code{mat}. 
#' Each column of \code{idx} contains the row indices to extract from the 
#' corresponding column of \code{mat}.
#'
#' @details
#' For each column \code{j}, this function returns a vector of values 
#' \code{mat[idx[, j], j]}. It is implemented in C++ using Rcpp for speed 
#' and is suitable for large-scale operations (e.g., tens of thousands of rows).
#'
#' Indices in \code{idx} are assumed to be 1-based (R-style). Internally, 
#' they are adjusted for 0-based indexing in C++.
#'
#' @return A numeric matrix of the same dimension as \code{idx}, where each 
#' element is extracted from \code{mat} according to the corresponding 
#' row and column in \code{idx}.
#'
#' @keywords internal
#' @noRd
extract_by_index <- function(mat, idx) {
    .Call(`_resemble_extract_by_index`, mat, idx)
}

#' @title Group Disparity Mask for Nearest Neighbors (C++)
#' @description
#' Internal C++ function that creates a logical matrix indicating whether
#' each neighbor in `kidxmat` belongs to a different group than the sample.
#'
#' @param kidxmat An integer matrix of neighbor indices. Each column
#'        corresponds to a sample, and each row to a ranked neighbor.
#'        Indices refer to rows in the `group` vector.
#' @param group An integer vector representing the group assignment
#'        for each sample. Its length must match the number of rows
#'        in the reference data.
#'
#' @return
#' A logical matrix of the same dimensions as `kidxmat`, where each
#' element is `TRUE` if the corresponding neighbor belongs to a
#' different group than the target sample, and `FALSE` otherwise.
#'
#' @details
#' For each column in `kidxmat`, corresponding to a sample,
#' the function:
#' \itemize{
#'   \item Extracts the group assignment of the sample.
#'   \item Checks whether each neighbor (by index) belongs to the same group.
#'   \item Stores `TRUE` if the neighbor is from a different group,
#'         `FALSE` otherwise.
#' }
#' The result can be used as a mask to filter out same-group neighbors.
#'
#' @note
#' Parallelized using OpenMP for performance.
#' Assumes 1-based indexing from R (internally adjusted).
#' @noRd
#' @keywords internal
#' @author Leonardo Ramirez-Lopez
not_in_same_group <- function(kidxmat, group) {
    .Call(`_resemble_not_in_same_group`, kidxmat, group)
}

#' @title Quantile Statistics from Nearest Neighbors (C++)
#' @description
#' Internal C++ function to compute quantiles of response values
#' among filtered nearest neighbors for each sample.
#'
#' @param kidxmat An integer matrix of nearest neighbor indices.
#'        Each column corresponds to one observation (e.g., in Xu),
#'        and rows represent the ranks of the neighbors.
#' @param kidxgrop A logical matrix of the same shape as `kidxmat`
#'        indicating which neighbors should be retained (`TRUE`)
#'        based on group membership.
#' @param Yr A numeric vector of reference response values. The
#'        values are indexed by `kidxmat` to compute quantiles.
#' @param k An integer specifying the number of neighbors to use.
#'        Either a single value (applied to all observations) or a
#'        vector of length `ncol(kidxmat)` (per-observation neighborhood sizes).
#' @param probs A numeric vector of probabilities for which to
#'        compute quantiles (e.g., `c(0, 0.05, 0.5, 0.95, 1)`).
#'
#' @return
#' A numeric matrix of size `(length(k) * ncol(kidxmat)) × length(probs)`
#' where each row contains the quantiles for a particular observation
#' and each column corresponds to a probability in `probs`.
#'
#' @details
#' For each observation, the function:
#' \itemize{
#'   \item Selects its top-k nearest neighbors.
#'   \item Filters neighbors based on the logical group mask.
#'   \item Extracts corresponding values from `Yr`.
#'   \item Computes requested quantiles.
#' }
#' Missing values in `Yr` are ignored. If no valid neighbors are
#' found for a given sample, all quantiles will be set to `NA`.
#'
#' @note
#' This is an internal function intended for performance-critical
#' applications. It should not be exported to the user.
#' @noRd
#' @keywords internal
#' @author Leonardo Ramirez-Lopez
compute_nn_quantiles <- function(kidxmat, kidxgrop, Yr, k, probs) {
    .Call(`_resemble_compute_nn_quantiles`, kidxmat, kidxgrop, Yr, k, probs)
}

#' @title Find Nearest Neighbors by Dissimilarity
#' 
#' @description
#' Internal C++ function to identify nearest neighbors from a dissimilarity
#' matrix, with optional thresholding. Uses partial sorting for efficiency
#' when only the top-k neighbors are needed.
#'
#' @param D A numeric matrix of dissimilarities (n × m), where rows are
#'   reference observations and columns are query observations.
#' @param k_min Integer. Minimum number of neighbors to return per 
#'   observation. When \code{threshold} is \code{NULL}, exactly \code{k_min}
#'   neighbors are returned. When \code{threshold} is specified, at least
#'   \code{k_min} neighbors are returned even if fewer fall below the
#'   threshold.
#' @param k_max Integer. Maximum number of neighbors to consider. Only the
#'   \code{k_max} nearest neighbors are evaluated against the threshold.
#'   Must satisfy \code{k_max >= k_min} and \code{k_max <= nrow(D)}.
#' @param threshold Either \code{NULL} or a numeric scalar. When \code{NULL},
#'   exactly \code{k_min} neighbors are returned. When specified, all
#'   neighbors among the top \code{k_max} with dissimilarity ≤ \code{threshold}
#'   are returned, subject to the \code{k_min} floor.
#'
#' @return A list of length \code{ncol(D)}. Each element is an integer vector
#'   of 1-indexed row indices identifying the nearest neighbors for that
#'   column, sorted by increasing dissimilarity.
#'
#' @details
#' For each column (query observation), the function:
#' \enumerate{
#'   \item Performs a partial sort to identify the \code{k_max} smallest
#'     dissimilarities in O(n + k log k) time.
#'   \item If \code{threshold} is \code{NULL}, returns the first \code{k_min}
#'     indices.
#'   \item If \code{threshold} is specified, counts how many of the top
#'     \code{k_max} neighbors have dissimilarity ≤ \code{threshold}, then
#'     returns \code{max(k_min, n_below)} indices.
#' }
#'
#' @section Parallel execution:
#' The function uses OpenMP for parallel computation across columns.
#' Thread count is controlled by the \code{OMP_NUM_THREADS} environment
#' variable.
#'
#' @note
#' Tie-breaking may differ slightly from R's \code{order()} function, which
#' uses stable sorting. In practice, exact ties in dissimilarity matrices
#' are rare.
#'
#' @keywords internal
#' @noRd
#' @author Leonardo Ramirez-Lopez
top_k_neighbors <- function(D, k_min, k_max, threshold = NULL) {
    .Call(`_resemble_top_k_neighbors`, D, k_min, k_max, threshold)
}

#' @title Function for identifiying the column in a matrix with the largest standard deviation
#' @description Identifies the column with the largest standard deviation. For internal use only!
#' @usage get_col_largest_sd(X)
#' @param X a matrix.
#' @return a value indicating the index of the column with the largest standard deviation. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
get_col_largest_sd <- function(X) {
    .Call(`_resemble_get_col_largest_sd`, X)
}

#' @title Function for computing the standard deviation of each column in a matrix
#' @description Computes the standard deviation of each column in a matrix. For internal use only!
#' @usage get_column_sds(X)
#' @param X a a matrix.
#' @return a vector of standard deviation values. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
get_column_sds <- function(X) {
    .Call(`_resemble_get_column_sds`, X)
}

#' @title Function for computing the overall variance of a matrix
#' @description Computes the variance of a matrix. For internal use only!
#' @usage overall_var(X)
#' @param X a matrix.
#' @return a vector of standard deviation values. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
overall_var <- function(X) {
    .Call(`_resemble_overall_var`, X)
}

#' @title Function for computing the mean of each column in a matrix
#' @description Computes the mean of each column in a matrix. For internal use only!
#' @usage get_column_means(X)
#' @param X a a matrix.
#' @return a vector of mean values. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
get_column_means <- function(X) {
    .Call(`_resemble_get_column_means`, X)
}

#' @title Function for computing the maximum of each column
#' @description Computes the max of each column in a matrix. For internal use only!
#' @usage get_column_maxs(X)
#' @param X a a matrix.
#' @return a vector of mean values. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
get_column_maxs <- function(X) {
    .Call(`_resemble_get_column_maxs`, X)
}

#' @title Function for computing sum of each column in a matrix
#' @description Computes the sum of each column in a matrix. For internal use only!
#' @usage get_column_sums(X)
#' @param X a matrix.
#' @return a vector of standard deviation values. 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
get_column_sums <- function(X) {
    .Call(`_resemble_get_column_sums`, X)
}

#' @title Computes the weights for pls regressions
#' @description
#' This is an internal function that computes the weights required for obtaining
#' each vector of pls scores. Implementation is done in C++ for improved performance.
#' @param X a numeric matrix of spectral data.
#' @param Y a matrix of one column with the response variable.
#' @param algorithm a character string indicating what method to use. Options are:
#' \code{'pls'} for pls (using covariance between X and Y), 
#' \code{'mpls'} for modified pls (using correlation between X and Y as in 
#' Shenk and Westerhaus, 1991; Westerhaus 2014) or
#' \code{'xls'} for extended pls (as implemented in BUCHI NIRWise PLUS software).
#' @param xls_min_w an integer indicating the minimum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 3 (as in BUCHI NIRWise PLUS software).
#' @param xls_max_w an integer indicating the maximum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 15 (as in BUCHI NIRWise PLUS software).
#' @author Leonardo Ramirez-Lopez and Claudio Orellano
#' @references
#' Shenk, J. S., & Westerhaus, M. O. (1991). Populations structuring of 
#' near infrared spectra and modified partial least squares regression. 
#' Crop Science, 31(6), 1548-1555.
#' 
#' Westerhaus, M. (2014). Eastern Analytical Symposium Award for outstanding 
#' Wachievements in near infrared spectroscopy: my contributions to 
#' Wnear infrared spectroscopy. NIR news, 25(8), 16-20.
#' @return a `matrix` of one column containing the weights.
#' @keywords internal 
#' @noRd
get_weights <- function(X, Y, algorithm = "pls", xls_min_w = 3L, xls_max_w = 15L) {
    .Call(`_resemble_get_weights`, X, Y, algorithm, xls_min_w, xls_max_w)
}

#' @title Internal Cpp function for computing the weights of the PLS components 
#' necessary for weighted average PLS
#' @description For internal use only!. 
#' @usage
#' get_local_pls_weights(projection_mat, 
#'           xloadings, 
#'           coefficients, 
#'           new_x, 
#'           min_component, 
#'           max_component, 
#'           scale, 
#'           Xcenter, 
#'           Xscale)
#' @param projection_mat the projection matrix generated either by the \code{opls} function.
#' @param xloadings .
#' @param coefficients the matrix of regression coefficients.
#' @param new_x a matrix of one new spectra to be predicted.
#' @param min_component an integer indicating the minimum number of pls components.
#' @param max_component an integer indicating the maximum number of pls components.
#' @param scale a logical indicating whether the matrix of predictors used to create the regression model was scaled.
#' @param Xcenter a vector with the values that must be used for centering \code{newdata}.
#' @param Xscale if \code{scale = TRUE} a vector with the values that must be used for scaling \code{newdata}.
#' @return a matrix of one row with the weights for each component between the max. and min. specified. 
#' @author Leonardo Ramirez-Lopez
#' @noRd
#' @keywords internal
get_local_pls_weights <- function(projection_mat, xloadings, coefficients, new_x, min_component, max_component, scale, Xcenter, Xscale) {
    .Call(`_resemble_get_local_pls_weights`, projection_mat, xloadings, coefficients, new_x, min_component, max_component, scale, Xcenter, Xscale)
}

#' @title orthogonal scores algorithn of partial leat squares (opls) projection
#' @description Computes orthogonal socres partial least squares (opls) 
#' projection with the NIPALS algorithm. It allows multiple response variables.
#' Although the main use of the function is for projection, it also retrieves 
#' regression coefficients. NOTE: For internal use only!
#' @usage 
#' opls_for_projection(X, Y, ncomp, scale,
#'                     maxiter, tol,
#'                     pcSelmethod = "var",
#'                     pcSelvalue = 0.01, 
#'                     algorithm = "pls", 
#'                     xls_min_w = 3, 
#'                     xls_max_w = 15)
#' @param X a matrix of predictor variables.
#' @param Y a matrix of either a single or multiple response variables.
#' @param ncomp the number of pls components.
#' @param scale logical indicating whether \code{X} must be scaled.
#' @param maxiter maximum number of iterations.
#' @param tol limit for convergence of the algorithm in the nipals algorithm.
#' @param pcSelmethod if \code{regression = TRUE}, the method for selecting the 
#' number of components. 
#' Options are: \code{'manual'}, \code{'cumvar'} (for selecting the number of 
#' principal components based on a given  cumulative amount of explained 
#' variance) and \code{'var'} (for selecting the number of principal components 
#' based on a given amount of explained variance). Default is \code{'cumvar'}.
#' @param pcSelvalue a numerical value that complements the selected method 
#' (\code{pcSelmethod}). 
#' If \code{'cumvar'} is chosen (default), \code{pcSelvalue} must be a value 
#' (larger than 0 and below 1) indicating the maximum amount of cumulative 
#' variance that the retained components should explain. Default is 0.99. 
#' If \code{'var'} is chosen, \code{pcSelvalue} must be a value (larger than 0 
#' and below 1) indicating that components that explain (individually) 
#' a variance lower than this threshold must be excluded. If \code{'manual'} 
#' is chosen, \code{pcSelvalue} has no effect and the number of components 
#' retrieved are the one specified in \code{ncomp}.
#' @param algorithm (for weights computation) a character string indicating 
#' what method to use. Options are:
#' \code{'pls'} for pls (using covariance between X and Y), 
#' \code{'mpls'} for modified pls (using correlation between X and Y) or
#' \code{'xls'} for extended pls (as implemented in BUCHI NIRWise PLUS software).
#' @param xls_min_w (for weights computation) an integer indicating the minimum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 3 (as in BUCHI NIRWise PLUS software).
#' @param xls_max_w (for weights computation) an integer indicating the maximum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 15 (as in BUCHI NIRWise PLUS software).
#' @return a list containing the following elements:
#' \itemize{
#' \item{\code{coefficients}: the matrix of regression coefficients.}
#' \item{\code{bo}: a matrix of one row containing the intercepts for 
#' each component.}
#' \item{\code{scores}: the matrix of scores.}
#' \item{\code{X_loadings}: the matrix of X loadings.}
#' \item{\code{Y_loadings}: the matrix of Y loadings.}
#' \item{\code{projection_mat}: the projection matrix.}
#' \item{\code{Y}: the \code{Y} input.}
#' \item{\code{variance}: a \code{list} conating two objects: \code{x_var} 
#' and \code{y_var}. 
#' These objects contain information on the explained variance for the \code{X} 
#' and \code{Y} matrices respectively.}
#' \item{\code{transf}: a \code{list} conating two objects: \code{Xcenter} 
#' and \code{Xscale}}. 
#' \item{\code{weights}: the matrix of wheights.}
#' }
#' @author Leonardo Ramirez-Lopez
#' @noRd
#' @keywords internal
opls_for_projection <- function(X, Y, ncomp, scale, maxiter, tol, pcSelmethod = "var", pcSelvalue = 0.01, algorithm = "pls", xls_min_w = 3L, xls_max_w = 15L) {
    .Call(`_resemble_opls_for_projection`, X, Y, ncomp, scale, maxiter, tol, pcSelmethod, pcSelvalue, algorithm, xls_min_w, xls_max_w)
}

#' @title orthogonal scores algorithn of partial leat squares (opls_get_all)
#' @description Computes orthogonal socres partial least squares (opls_get_all) 
#' regressions with the NIPALS algorithm. It retrives a comprehensive set of
#' pls outputs (e.g. vip and sensivity radius). It allows multiple response 
#' variables. NOTE: For internal use only!
#' @usage 
#' opls_get_all(X, 
#'              Y, 
#'              ncomp, 
#'              scale, 
#'              maxiter, 
#'              tol, 
#'              algorithm = "pls", 
#'              xls_min_w = 3, 
#'              xls_max_w = 15)
#' @param X a matrix of predictor variables.
#' @param Y a matrix of either a single or multiple response variables.
#' @param ncomp the number of pls components.
#' @param scale logical indicating whether \code{X} must be scaled.
#' @param maxiter maximum number of iterations.
#' @param tol limit for convergence of the algorithm in the nipals algorithm.
#' @param algorithm (for weights computation) a character string indicating 
#' what method to use. Options are:
#' \code{'pls'} for pls (using covariance between X and Y), 
#' \code{'mpls'} for modified pls (using correlation between X and Y) or
#' \code{'xls'} for extended pls (as implemented in BUCHI NIRWise PLUS software).
#' @param xls_min_w (for weights computation) an integer indicating the minimum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 3 (as in BUCHI NIRWise PLUS software).
#' @param xls_max_w (for weights computation) an integer indicating the maximum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 15 (as in BUCHI NIRWise PLUS software).
#' @return a list containing the following elements:
#' \itemize{
#' \item{\code{ncomp}: the number of components used.}
#' \item{\code{coefficients}: the matrix of regression coefficients.}
#' \item{\code{bo}: a matrix of one row containing the intercepts for each component.}
#' \item{\code{scores}: the matrix of scores.}
#' \item{\code{X_loadings}: the matrix of X loadings.}
#' \item{\code{Y_loadings}: the matrix of Y loadings.}
#' \item{\code{vip}: the projection matrix.}
#' \item{\code{selectivity_ratio}: the matrix of selectivity ratio (see Rajalahti, Tarja, et al. 2009).}
#' \item{\code{Y}: the \code{Y} input.}
#' \item{\code{variance}: a \code{list} conating two objects: \code{x_var} and \code{y_var}. 
#' These objects contain information on the explained variance for the \code{X} and \code{Y} matrices respectively.}
#' \item{\code{transf}: a \code{list} conating two objects: \code{Xcenter} and \code{Xscale}}. 
#' \item{\code{weights}: the matrix of wheights.}
#' } 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
opls_get_all <- function(X, Y, ncomp, scale, maxiter, tol, algorithm = "pls", xls_min_w = 3L, xls_max_w = 15L) {
    .Call(`_resemble_opls_get_all`, X, Y, ncomp, scale, maxiter, tol, algorithm, xls_min_w, xls_max_w)
}

#' @title orthogonal scores algorithn of partial leat squares (opls)
#' @description Computes orthogonal socres partial least squares (opls) 
#' regressions with the NIPALS algorithm. It allows multiple response variables. 
#' It does not return the variance information of the components. NOTE: For 
#' internal use only!
#' @usage 
#' opls(X, 
#'      Y, 
#'      ncomp, 
#'      scale, 
#'      maxiter, 
#'      tol, 
#'      algorithm = "pls", 
#'      xls_min_w = 3, 
#'      xls_max_w = 15)
#'      
#' @param X a matrix of predictor variables.
#' @param Y a matrix of either a single or multiple response variables.
#' @param ncomp the number of pls components.
#' @param scale logical indicating whether \code{X} must be scaled.
#' @param maxiter maximum number of iterations.
#' @param tol limit for convergence of the algorithm in the nipals algorithm.
#' @param algorithm (for weights computation) a character string indicating 
#' what method to use. Options are:
#' \code{'pls'} for pls (using covariance between X and Y), 
#' \code{'mpls'} for modified pls (using correlation between X and Y) or
#' \code{'xls'} for extended pls (as implemented in BUCHI NIRWise PLUS software).
#' @param xls_min_w (for weights computation) an integer indicating the minimum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 3 (as in BUCHI NIRWise PLUS software).
#' @param xls_max_w (for weights computation) an integer indicating the maximum window size for the "xls"
#' method. Only used if \code{algorithm = 'xls'}. Default is 15 (as in BUCHI NIRWise PLUS software).
#' @return a list containing the following elements:
#' \itemize{
#' \item{\code{coefficients}: the matrix of regression coefficients.}
#' \item{\code{bo}: a matrix of one row containing the intercepts for each component.}
#' \item{\code{scores}: the matrix of scores.}
#' \item{\code{X_loadings}: the matrix of X loadings.}
#' \item{\code{Y_loadings}: the matrix of Y loadings.}
#' \item{\code{projection_mat}: the projection matrix.}
#' \item{\code{Y}: the \code{Y} input.}
#' \item{\code{transf}: a \code{list} conating two objects: \code{Xcenter} and \code{Xscale}}. 
#' \item{\code{weights}: the matrix of wheights.}} 
#' @author Leonardo Ramirez-Lopez
#' @noRd
#' @keywords internal
opls <- function(X, Y, ncomp, scale, maxiter, tol, algorithm = "pls", xls_min_w = 3L, xls_max_w = 15L) {
    .Call(`_resemble_opls`, X, Y, ncomp, scale, maxiter, tol, algorithm, xls_min_w, xls_max_w)
}

#' @title Fast orthogonal scores algorithm of partial least squares (PLS)
#' @description Computes orthogonal scores partial least squares (PLS) 
#' regression using either NIPALS or SIMPLS algorithm. Supports multiple 
#' response variables. In contrast to \code{opls}, this function omits 
#' auxiliary outputs (e.g. scores, explained variance) not required for 
#' local regression. For internal use only.
#' @usage 
#' opls_get_basics(X, Y, ncomp, scale, 
#'                 maxiter, tol, 
#'                 algorithm = "pls", 
#'                 xls_min_w = 3, 
#'                 xls_max_w = 15)
#' @param X a matrix of predictor variables.
#' @param Y a matrix of either a single or multiple response variables.
#' @param ncomp the number of PLS components.
#' @param scale logical indicating whether \code{X} must be scaled.
#' @param maxiter maximum number of iterations (only used for NIPALS-based 
#' algorithms: \code{'pls'}, \code{'mpls'}, \code{'xls'}).
#' @param tol convergence tolerance for the NIPALS algorithm (only used for 
#' NIPALS-based algorithms).
#' @param algorithm a character string indicating the PLS algorithm to use:
#' \itemize{
#'   \item{\code{'pls'}: standard PLS using covariance between X and Y for 
#'     weight computation (NIPALS algorithm).}
#'   \item{\code{'mpls'}: modified PLS using correlation between X and Y for 
#'     weight computation (NIPALS algorithm). See Shenk and Westerhaus (1991).}
#'   \item{\code{'xls'}: extended PLS as implemented in BUCHI NIRWise PLUS 
#'     software (NIPALS algorithm).}
#'   \item{\code{'simpls'}: SIMPLS algorithm (de Jong, 1993). Computationally 
#'     faster as it avoids iterative X deflation. Parameters \code{maxiter}, 
#'     \code{tol}, \code{xls_min_w}, and \code{xls_max_w} are ignored.}
#' }
#' @param xls_min_w an integer indicating the minimum window size for the 
#' \code{'xls'} method. Only used if \code{algorithm = 'xls'}. Default is 3.
#' @param xls_max_w an integer indicating the maximum window size for the 
#' \code{'xls'} method. Only used if \code{algorithm = 'xls'}. Default is 15.
#' @return a list containing:
#' \itemize{
#'   \item{\code{ncomp}: the number of PLS components.}
#'   \item{\code{coefficients}: the matrix of regression coefficients.}
#'   \item{\code{bo}: a matrix containing the intercepts for each component.}
#'   \item{\code{X_loadings}: the matrix of X loadings.}
#'   \item{\code{Y_loadings}: the matrix of Y loadings.}
#'   \item{\code{projection_mat}: the projection matrix for computing scores 
#'     from new data.}
#'   \item{\code{transf}: a list containing:
#'     \itemize{
#'       \item{\code{Xcenter}: row vector of column means used for centering.}
#'       \item{\code{Xscale}: row vector of column standard deviations used 
#'         for scaling (ones if \code{scale = FALSE}).}
#'     }
#'   }
#'   \item{\code{weights}: the matrix of PLS weights.}
#' }
#' @references
#' de Jong, S. (1993). SIMPLS: An alternative approach to partial least 
#' squares regression. Chemometrics and Intelligent Laboratory Systems, 
#' 18(3), 251-263.
#' 
#' Shenk, J.S., & Westerhaus, M.O. (1991). Populations structuring of near 
#' infrared spectra and modified partial least squares regression. Crop 
#' Science, 31(6), 1548-1555.
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
opls_get_basics <- function(X, Y, ncomp, scale, maxiter, tol, algorithm = "pls", xls_min_w = 3L, xls_max_w = 15L) {
    .Call(`_resemble_opls_get_basics`, X, Y, ncomp, scale, maxiter, tol, algorithm, xls_min_w, xls_max_w)
}

#' @title Prediction function for the \code{opls} and \code{fopls} functions
#' @description Predicts response values based on a model generated by either by \code{opls} or the \code{fopls} functions. 
#' For internal use only!. 
#' @usage predict_opls(bo, b, ncomp, newdata, scale, Xscale)
#' @param bo a numeric value indicating the intercept.
#' @param b the matrix of regression coefficients.
#' @param ncomp an integer value indicating how may components must be used in the prediction.
#' @param newdata a matrix containing the predictor variables.
#' @param scale a logical indicating whether the matrix of predictors used to create the regression model was scaled.
#' @param Xscale if \code{scale = TRUE} a matrix of one row with the values that must be used for scaling \code{newdata}.
#' @return a matrix of predicted values.
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
predict_opls <- function(bo, b, ncomp, newdata, scale, Xscale) {
    .Call(`_resemble_predict_opls`, bo, b, ncomp, newdata, scale, Xscale)
}

#' @title Projection function for the \code{opls} function
#' @description Projects new spectra onto a PLS space based on a model generated by either by \code{opls} or the \code{opls2} functions. 
#' For internal use only!. 
#' @usage project_opls(projection_mat, ncomp, newdata, scale, Xcenter, Xscale)
#' @param projection_mat the projection matrix generated by the \code{opls} function.
#' @param ncomp an integer value indicating how may components must be used in the prediction.
#' @param newdata a matrix containing the predictor variables.
#' @param scale a logical indicating whether the matrix of predictors used to create the regression model was scaled.
#' @param Xscale if \code{scale = TRUE} a matrix of one row with the values that must be used for scaling \code{newdata}.
#' @param Xcenter a matrix of one row with the values that must be used for centering \code{newdata}.
#' @return a matrix corresponding to the new spectra projected onto the PLS space 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
project_opls <- function(projection_mat, ncomp, newdata, scale, Xcenter, Xscale) {
    .Call(`_resemble_project_opls`, projection_mat, ncomp, newdata, scale, Xcenter, Xscale)
}

#' @title Projection to pls and then re-construction
#' @description Projects spectra onto a PLS space and then reconstructs it back.
#' @usage reconstruction_error(x, 
#'                             projection_mat, 
#'                             xloadings, 
#'                             scale, 
#'                             Xcenter, 
#'                             Xscale, 
#'                             scale_back = FALSE)
#' @param x a matrix to project.
#' @param projection_mat the projection matrix generated by the \code{opls_get_basics} function.
#' @param xloadings the loadings matrix generated by the \code{opls_get_basics} function.
#' @param scale logical indicating if scaling is required
#' @param Xcenter a matrix of one row with the centering values
#' @param Xscale a matrix of one row with the scaling values
#' @param scale_back compute the reconstruction error after de-centering the 
#' data and de-scaling it.
#' @return a matrix of 1 row and 1 column.
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
reconstruction_error <- function(x, projection_mat, xloadings, scale, Xcenter, Xscale, scale_back = FALSE) {
    .Call(`_resemble_reconstruction_error`, x, projection_mat, xloadings, scale, Xcenter, Xscale, scale_back)
}

#' @title Internal Cpp function for performing leave-group-out cross-validations for pls regression 
#' @description For internal use only!. 
#' @usage opls_cv_cpp(X, Y, scale, method, 
#'                   mindices, pindices, 
#'                   min_component, ncomp, 
#'                   new_x, 
#'                   maxiter, tol, 
#'                   wapls_grid, 
#'                   algorithm, 
#'                   statistics = TRUE)
#' @param X a matrix of predictor variables.
#' @param Y a matrix of a single response variable.
#' @param scale a logical indicating whether the matrix of predictors 
#' (\code{X}) must be scaled.
#' @param method the method used for regression. One of the following options: 
#' \code{'pls'} or \code{'wapls'} or \code{'completewapls1p'}.
#' @param mindices a matrix with \code{n} rows and \code{m} columns where 
#' \code{m} is equivalent to the number of resampling iterations. The elements 
#' of each column indicate the indices of the observations to be used for 
#' modeling at each iteration.
#' @param pindices a matrix with \code{k} rows and \code{m} columns where 
#' \code{m} is equivalent to the number of 
#' resampling iterations. The elements of each column indicate the indices of 
#' the observations to be used for predicting at each iteration.
#' @param min_component an integer indicating the number of minimum pls 
#' components (if the \code{method = 'pls'}).
#' @param ncomp an integer indicating the number of pls components.
#' @param new_x a matrix of one row corresponding to the observation to be 
#' predicted (if the \code{method = 'wapls'}).
#' @param maxiter maximum number of iterations.
#' @param tol limit for convergence of the algorithm in the nipals algorithm.
#' @param wapls_grid the grid on which the search for the best combination of 
#' minimum and maximum pls factors of \code{'wapls'} is based on in case 
#' \code{method = 'completewapls1p'}.
#' @param algorithm either pls (\code{'pls'}) or modified pls (\code{'mpls'}). 
#' See \code{get_weigths} function.
#' @param statistics a logical value indicating whether the precision and 
#' accuracy statistics are to be returned, otherwise the predictions for each 
#' validation segment are retrieved.
#' @return 
#' if \code{statistics = true} a list containing the following one-row matrices:
#' \itemize{
#' \item{\code{rmse_seg}: the RMSEs.}
#' \item{\code{st_rmse_seg}: the standardized RMSEs.}
#' \item{\code{rsq_seg}: the coefficients of determination.}
#' } 
#' 
#' if \code{statistics = false} a list containing the following one-row matrices:
#' \itemize{
#' \item{\code{predictions}: the predictions of each of the validation 
#' segments in \code{pindices}. Each column in \code{pindices} contains the 
#' validation indices of a segment.}
#' \item{\code{st_rmse_seg}: the standardized RMSEs.}
#' \item{\code{rsq_seg}: the coefficients of determination.}
#' } 
#' 
#' If \code{method = "wapls"}, data of the pls weights are output in this 
#' list(\code{compweights}).
#'
#' If \code{method = "completewapls1"}, data of all the combination of 
#' components passed in \code{wapls_grid} are 
#' output in this list(\code{complete_compweights}).
#' 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
opls_cv_cpp <- function(X, Y, scale, method, mindices, pindices, min_component, ncomp, new_x, maxiter, tol, wapls_grid, algorithm, statistics = TRUE) {
    .Call(`_resemble_opls_cv_cpp`, X, Y, scale, method, mindices, pindices, min_component, ncomp, new_x, maxiter, tol, wapls_grid, algorithm, statistics)
}

#' @title Orthogonal scores algorithm of partial least squares for gesearch
#' @description Computes orthogonal scores partial least squares (PLS) 
#' regression using either NIPALS or SIMPLS algorithm. This function is 
#' optimised for the \code{gesearch} evolutionary search and computes only 
#' the outputs required for weakness score evaluation: predictions, 
#' reconstruction error, and score-space dissimilarity.
#' 
#' NOTE: This function supports only a single response variable (PLS1). 
#' For internal use only.
#' @usage 
#' opls_gesearch(Xr, Yr, Xu, ncomp, scale,
#'               response = FALSE, reconstruction = TRUE,
#'               similarity = TRUE, fresponse = TRUE,
#'               algorithm = "pls")
#'         
#' @param Xr a matrix of predictor variables for the reference/training set.
#' @param Yr a single-column matrix of the response variable for the 
#' reference/training set. Only single-response (PLS1) is supported.
#' @param Xu a matrix of predictor variables for the target/test set.
#' @param ncomp the number of PLS components.
#' @param scale logical indicating whether \code{Xr} and \code{Xu} must be 
#' scaled. Centering is always applied using parameters derived from 
#' \code{Xr}.
#' @param response logical indicating whether to compute predictions for 
#' \code{Xu}. Used for the response weakness score (\code{w_r}) in 
#' \code{gesearch}. Default is \code{FALSE}.
#' @param reconstruction logical indicating whether to compute the 
#' reconstruction error of \code{Xu}. Used for the reconstruction weakness 
#' score (\code{w_q}) in \code{gesearch}. Default is \code{TRUE}.
#' @param similarity logical indicating whether to compute the distance 
#' between \code{Xr} and \code{Xu} in the PLS score space. Used for the 
#' similarity weakness score (\code{w_d}) in \code{gesearch}. Default is 
#' \code{TRUE}.
#' @param fresponse logical indicating whether to compute the proportion of 
#' response variance not explained by the model. Default is \code{TRUE}.
#' @param algorithm a character string indicating the PLS algorithm to use:
#' \itemize{
#'   \item{\code{'pls'}: standard PLS using covariance between X and Y for 
#'     weight computation (NIPALS algorithm).}
#'   \item{\code{'mpls'}: modified PLS using correlation between X and Y for 
#'     weight computation (NIPALS algorithm). See Shenk and Westerhaus (1991).}
#'   \item{\code{'simpls'}: SIMPLS algorithm (de Jong, 1993). Computationally 
#'     faster as it avoids iterative X deflation.}
#' }
#' @return a list containing:
#' \itemize{
#'   \item{\code{ncomp}: the number of components used.}
#'   \item{\code{pred_response}: predictions for \code{Xu} (only if 
#'     \code{response = TRUE}).}
#'   \item{\code{rmse_reconstruction}: RMSE of the spectral reconstruction 
#'     for \code{Xu} (only if \code{reconstruction = TRUE}).}
#'   \item{\code{score_dissimilarity}: mean Euclidean distance between 
#'     \code{Xr} and \code{Xu} scores in Mahalanobis-scaled PLS space 
#'     (only if \code{similarity = TRUE}).}
#'   \item{\code{residual_variance}: proportion of response variance not 
#'     explained by the model (only if \code{fresponse = TRUE}).}
#' }
#' @details
#' This function is designed for repeated evaluation within the 
#' \code{gesearch} evolutionary search algorithm, where it may be called 
#' ~10^5 times per run. It computes only the outputs necessary for 
#' calculating weakness scores:
#' \itemize{
#'   \item{Response weakness (\code{w_r}): prediction RMSE on the target, 
#'     requires \code{response = TRUE}.}
#'   \item{Reconstruction weakness (\code{w_q}): spectral reconstruction 
#'     error, requires \code{reconstruction = TRUE}.}
#'   \item{Similarity weakness (\code{w_d}): Mahalanobis distance in score 
#'     space, requires \code{similarity = TRUE}.}
#' }
#' 
#' Preprocessing applies scaling (if requested) followed by centering, 
#' using parameters derived from \code{Xr}. The same transformation is 
#' applied to \code{Xu}.
#' 
#' The \code{'simpls'} algorithm is faster than NIPALS-based methods 
#' (\code{'pls'}, \code{'mpls'}) as it avoids iterative X deflation. 
#' However, reconstruction errors and score-space distances differ 
#' numerically between algorithms (rankings are typically similar). 
#' Do not mix algorithms within a single \code{gesearch} run.
#' 
#' @references
#' de Jong, S. (1993). SIMPLS: An alternative approach to partial least 
#' squares regression. Chemometrics and Intelligent Laboratory Systems, 
#' 18(3), 251-263.
#' 
#' Shenk, J.S., & Westerhaus, M.O. (1991). Populations structuring of 
#' near infrared spectra and modified partial least squares regression. 
#' Crop Science, 31(6), 1548-1555.
#' 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
opls_gesearch <- function(Xr, Yr, Xu, ncomp, scale, response = FALSE, reconstruction = TRUE, similarity = TRUE, fresponse = TRUE, algorithm = "pls") {
    .Call(`_resemble_opls_gesearch`, Xr, Yr, Xu, ncomp, scale, response, reconstruction, similarity, fresponse, algorithm)
}

#' @title Gaussian process regression with linear kernel (gaussian_process)
#' @description Carries out a gaussian process regression with a linear kernel (dot product). For internal use only!
#' @usage gaussian_process(X, Y, noisev, scale) 
#' @param X a matrix of predictor variables
#' @param Y a matrix with a single response variable
#' @param noisev a value indicating the variance of the noise for Gaussian process regression. Default is 0.001. a matrix with a single response variable
#' @param scale a logical indicating whether both the predictors 
#' and the response variable must be scaled to zero mean and unit variance.
#' @return a list containing the following elements:
#' \itemize{
#' \item{\code{b}: the regression coefficients.}
#' \item{\code{Xz}: the (final transformed) matrix of predictor variables.}
#' \item{\code{alpha}: the alpha matrix.}
#' \item{\code{is.scaled}: logical indicating whether both the predictors and response variable were scaled to zero mean and unit variance.}
#' \item{\code{Xcenter}: if matrix of predictors was scaled, the centering vector used for \code{X}.}
#' \item{\code{Xscale}: if matrix of predictors was scaled, the scaling vector used for \code{X}.}
#' \item{\code{Ycenter}: if matrix of predictors was scaled, the centering vector used for \code{Y}.}
#' \item{\code{Yscale}: if matrix of predictors was scaled, the scaling vector used for \code{Y}.}
#' }
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
gaussian_process <- function(X, Y, noisev = 0.001, scale = TRUE) {
    .Call(`_resemble_gaussian_process`, X, Y, noisev, scale)
}

#' @title Prediction function for the \code{gaussian_process} function (Gaussian process regression with dot product covariance)
#' @description Predicts response values based on a model generated by the \code{gaussian_process} function (Gaussian process regression with dot product covariance). For internal use only!. 
#' @usage predict_gaussian_process(Xz, alpha, newdata, scale, Xcenter, Xscale, Ycenter, Yscale)
#' @param newdata a matrix containing the predictor variables
#' @param scale a logical indicating whether the matrix of predictors used to create the regression model 
#' (in the \code{gaussian_process} function) was scaled
#' @param Xcenter if \code{center = TRUE} a matrix of one row with the values that must be used for centering \code{newdata}.
#' @param Xscale if \code{scale = TRUE} a matrix of one row with the values that must be used for scaling \code{newdata}.
#' @param Ycenter if \code{center = TRUE} a matrix of one row with the values that must be used for accounting for the centering of the response variable.
#' @param Yscale if \code{scale = TRUE} a matrix of one row with the values that must be used  for accounting for the scaling of the response variable.
#' @return a matrix of predicted values
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
predict_gaussian_process <- function(Xz, alpha, newdata, scale, Xcenter, Xscale, Ycenter, Yscale) {
    .Call(`_resemble_predict_gaussian_process`, Xz, alpha, newdata, scale, Xcenter, Xscale, Ycenter, Yscale)
}

#' @title Internal Cpp function for performing leave-group-out cross 
#' validations for gaussian process
#' @description For internal use only!. 
#' @usage gaussian_process_cv(X, Y, mindices, pindices, noisev = 0.001,  
#' scale = TRUE, statistics = TRUE)
#' @param X a matrix of predictor variables.
#' @param Y a matrix of a single response variable.
#' @param mindices a matrix with \code{n} rows and \code{m} columns where \code{m} is equivalent to the number of 
#' resampling iterations. The elements of each column indicate the indices of the observations to be used for modeling at each 
#' iteration.
#' @param pindices a matrix with \code{k} rows and \code{m} columns where \code{m} is equivalent to the number of 
#' resampling iterations. The elements of each column indicate the indices of the observations to be used for predicting at each 
#' iteration.
#' @param scale a logical indicating whether both the predictors 
#' and the response variable must be scaled to zero mean and unit variance.
#' @param statistics a logical value indicating whether the precision and 
#' accuracy statistics are to be returned, otherwise the predictions for each 
#' validation segment are retrieved.
#' @return a list containing the following one-row matrices:
#' \itemize{
#' \item{\code{rmse.seg}: the RMSEs.}
#' \item{\code{st.rmse.seg}: the standardized RMSEs.}
#' \item{\code{rsq.seg}: the coefficients of determination.}
#' } 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
gaussian_process_cv <- function(X, Y, mindices, pindices, noisev = 0.001, scale = TRUE, statistics = TRUE) {
    .Call(`_resemble_gaussian_process_cv`, X, Y, mindices, pindices, noisev, scale, statistics)
}

#' @title Principal components based on  the non-linear iterative partial least squares (nipals) algorithm
#' @description Computes orthogonal socres partial least squares (opls) regressions with the NIPALS algorithm. It allows multiple response variables. 
#' For internal use only!
#' @usage 
#' pca_nipals(X, ncomp, center, scale,
#'            maxiter, tol,
#'            pcSelmethod = "var",
#'            pcSelvalue = 0.01)
#' @param X a matrix of predictor variables.
#' @param ncomp the number of pls components.
#' @param scale logical indicating whether \code{X} must be scaled.
#' @param maxiter maximum number of iterations.
#' @param tol limit for convergence of the algorithm in the nipals algorithm.
#' @param pcSelmethod the method for selecting the number of components. 
#' Options are: \code{'cumvar'} (for selecting the number of principal components based on a given 
#' cumulative amount of explained variance) and \code{"var"} (for selecting the number of principal 
#' components based on a given amount of explained variance). Default is \code{'var'}
#' @param pcSelvalue a numerical value that complements the selected method (\code{pcSelmethod}). 
#' If \code{"cumvar"} is chosen, it must be a value (larger than 0 and below 1) indicating the maximum 
#' amount of cumulative variance that the retained components should explain. If \code{"var"} is chosen, 
#' it must be a value (larger than 0 and below 1) indicating that components that explain (individually) 
#' a variance lower than this threshold must be excluded. If \code{"manual"} is chosen, it must be a value 
#' specifying the desired number of principal components to retain. Default is 0.01.
#' @return a list containing the following elements:
#' \itemize{
#' \item{\code{pc_scores}: a matrix of principal component scores.}
#' \item{\code{pc_loadings}: a matrix of of principal component loadings.}
#' \item{\code{variance}: a matrix of the variance of the principal components.} 
#' \item{\code{scale}: a \code{list} conating two objects: \code{center} and \code{scale}, which correspond to the vectors used to center and scale the input matrix.} 
#' } 
#' @author Leonardo Ramirez-Lopez
#' @keywords internal 
#' @noRd
pca_nipals <- function(X, ncomp, center, scale, maxiter, tol, pcSelmethod = "var", pcSelvalue = 0.01) {
    .Call(`_resemble_pca_nipals`, X, ncomp, center, scale, maxiter, tol, pcSelmethod, pcSelvalue)
}

#' @title Internal: Fit a local weighted PLS model and predict for a query point
#'
#' @description
#' Fits a local Partial Least Squares (PLS) model using a neighborhood subset
#' and computes a weighted prediction for a target sample. The weighting is
#' done over multiple components using a provided evaluation grid.
#'
#' @param X Numeric matrix of predictors from the local neighborhood
#'   (observations in rows, variables in columns).
#' @param Y Numeric matrix (single column) of corresponding response values.
#' @param xval Numeric matrix (single row) representing the query sample to
#'   predict.
#' @param emgrid Numeric matrix used to weight component-wise predictions.
#' @param ncomp_max Integer. Maximum number of PLS components to fit.
#' @param ncomp_min Integer. Minimum number of PLS components to use in
#'   prediction.
#' @param scale Logical. Whether to scale predictors before PLS fitting.
#' @param max_iter Numeric. Maximum number of iterations for the PLS algorithm.
#' @param tol Numeric. Convergence tolerance for the PLS algorithm.
#' @param algorithm Character. PLS algorithm to use: \code{"mpls"} (default),
#'   \code{"pls"} (nipals), or \code{"simpls"}.
#'
#' @return
#' A numeric vector of weighted predictions (length equal to number of rows
#' in \code{emgrid}).
#'
#' @details
#' The function performs the following steps:
#' \enumerate{
#'   \item Fits a PLS model on \code{X} and \code{Y} with up to
#'         \code{ncomp_max} components.
#'   \item Extracts centering and scaling parameters from the fitted model.
#'   \item Computes component weights for \code{xval} using
#'         \code{get_local_pls_weights()}.
#'   \item Predicts component-wise responses using \code{predict_opls()}.
#'   \item Applies \code{emgrid} weights scaled by component weights.
#'   \item Returns row-normalized weighted average of predictions.
#' }
#'
#' @author Leonardo Ramirez-Lopez
#' @keywords internal
#' @noRd
ith_local_fit <- function(X, Y, xval, emgrid, ncomp_max, ncomp_min, scale, max_iter, tol, algorithm = "mpls") {
    .Call(`_resemble_ith_local_fit`, X, Y, xval, emgrid, ncomp_max, ncomp_min, scale, max_iter, tol, algorithm)
}

#' @title Compute Final Local PLS Model Outputs
#' @description
#' This function calculates the final local model coefficients, intercept,
#' scaled variable importance (VIP), and selectivity ratio for a single observation
#' using a weighted combination of PLS components.
#'
#' @param X A matrix of predictor variables used for calibration.
#' @param Y A matrix of response variables used for calibration.
#' @param new_x A single observation (1 x p) of predictor variables to compute weights.
#' @param ncomp_min The minimum number of PLS components to include in the final model.
#' @param ncomp_max The maximum number of PLS components to include in the final model.
#' @param scale Logical indicating whether to scale the data.
#' @param maxiter Maximum number of iterations allowed during the NIPALS algorithm.
#' @param tol Tolerance threshold for convergence in the iterative algorithm.
#' @param algorithm \code{'mpls'} (defalt), \code{'pls'} (nipals), \code{'simpls'}.
#'
#' @return A list with the following elements:
#' \itemize{
#'   \item \code{ib0}: Intercept term, computed as a weighted sum of component-specific intercepts.
#'   \item \code{ibs}: Weighted regression coefficients.
#'   \item \code{ivips}: Weighted variable importance (VIP) scores, scaled by component SDs.
#'   \item \code{isratio}: Weighted selectivity ratios, scaled by component SDs.
#'   \item \code{Xscale}: Scaling vector used to scale \code{X} (if \code{scale = TRUE}).
#' }
#'
#' @keywords internal
#' @noRd
final_fits_cpp <- function(X, Y, new_x, ncomp_min, ncomp_max, scale, maxiter, tol, algorithm = "mpls") {
    .Call(`_resemble_final_fits_cpp`, X, Y, new_x, ncomp_min, ncomp_max, scale, maxiter, tol, algorithm)
}

#' @title Internal: Predict using local PLS coefficients and optional
#' dissimilarities
#'
#' @description
#' Computes predictions for a new observation using local PLS models
#' represented by coefficients (\code{plslib}). The prediction is based on
#' inverse-scaled feature values. If a dissimilarity vector is provided, it is
#' prepended to the input features before inverse scaling.
#'
#' @param plslib A numeric matrix of PLS model coefficients (n_models × p+1).
#'   First column is the intercept; remaining columns are coefficients for 
#'   scaled features.
#' @param xscale A numeric matrix of scaling values (n_models × p), same 
#'   number of columns as \code{plslib[,-1]}.
#' @param Xu A numeric vector of length p representing the query sample to 
#'   be predicted.
#' @param dxrxu Optional numeric vector of dissimilarities between \code{Xu} 
#'   and reference samples. If provided, prepended to \code{Xu} as additional 
#'   predictive features. Default is \code{R_NilValue} (NULL).
#'
#' @return A numeric vector of length \code{nrow(plslib)} containing the 
#'   predicted response for \code{Xu} from each local model.
#'
#' @details
#' For each local model (row of \code{plslib}), the function:
#' \enumerate{
#'   \item Optionally prepends \code{dxrxu} to \code{Xu}
#'   \item Computes inverse-scaled features: \code{Xu / xscale[i,]}
#'   \item Computes prediction: \code{intercept + sum(coefficients * scaled_features)}
#' }
#'
#' @note
#' This is an internal function optimised for performance in the prediction
#' loop of \code{predict.liblex}.
#'
#' @keywords internal
#' @noRd
#' @author Leonardo Ramirez-Lopez
ith_pred_cpp <- function(plslib, xscale, Xu, dxrxu = NULL) {
    .Call(`_resemble_ith_pred_cpp`, plslib, xscale, Xu, dxrxu)
}

context("test-pls-cpp-consistency")

test_that("PLS (NIPALS) implementations produce consistent results with scale = TRUE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  tol2 <- 1e-5
  
  # Fit models
  pls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = TRUE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = TRUE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "pls"
  )
  
  # Predictions
  pls_basics_predict <- resemble:::predict_opls(
    bo = pls_basics$bo, 
    b = pls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = pls_basics$transf$Xscale
  )
  
  pls_opls_predict <- resemble:::predict_opls(
    bo = pls_opls$bo, 
    b = pls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = pls_opls$transf$Xscale
  )
  
  # Reconstruction error
  pls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = pls_opls$projection_mat, 
    xloadings = pls_opls$X_loadings,
    scale = TRUE,
    Xcenter = pls_opls$transf$Xcenter,
    Xscale = pls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(pls_projection$weights - pls_basics$weights)), 
    tol,
    label = "PLS scale=TRUE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(pls_basics$weights - pls_opls$weights)), 
    tol,
    label = "PLS scale=TRUE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(pls_basics_predict - pls_opls_predict)), 
    tol,
    label = "PLS scale=TRUE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(pls_projection$scores - pls_opls$scores)), 
    tol,
    label = "PLS scale=TRUE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(pls_gesearch$pred_response - pls_opls_predict[, ncomp_test])), 
    tol,
    label = "PLS scale=TRUE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(pls_gesearch$rmse_reconstruction) - drop(pls_opls_rec)), 
    tol,
    label = "PLS scale=TRUE: opls_gesearch vs reconstruction_error"
  )
})


test_that("PLS (NIPALS) implementations produce consistent results with scale = FALSE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  
  # Fit models
  pls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = FALSE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = FALSE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "pls"
  )
  
  # Predictions
  pls_basics_predict <- resemble:::predict_opls(
    bo = pls_basics$bo, 
    b = pls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = pls_basics$transf$Xscale
  )
  
  pls_opls_predict <- resemble:::predict_opls(
    bo = pls_opls$bo, 
    b = pls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = pls_opls$transf$Xscale
  )
  
  # Reconstruction error
  pls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = pls_opls$projection_mat, 
    xloadings = pls_opls$X_loadings,
    scale = FALSE,
    Xcenter = pls_opls$transf$Xcenter,
    Xscale = pls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(pls_projection$weights - pls_basics$weights)), 
    tol,
    label = "PLS scale=FALSE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(pls_basics$weights - pls_opls$weights)), 
    tol,
    label = "PLS scale=FALSE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(pls_basics_predict - pls_opls_predict)), 
    tol,
    label = "PLS scale=FALSE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(pls_projection$scores - pls_opls$scores)), 
    tol,
    label = "PLS scale=FALSE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(pls_gesearch$pred_response - pls_opls_predict[, ncomp_test])), 
    tol,
    label = "PLS scale=FALSE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(pls_gesearch$rmse_reconstruction) - drop(pls_opls_rec)), 
    tol,
    label = "PLS scale=FALSE: opls_gesearch vs reconstruction_error"
  )
  
  # Xscale should be all ones when scale = FALSE
  expect_true(
    all(pls_basics$transf$Xscale == 1),
    label = "PLS scale=FALSE: Xscale should be all ones (opls_get_basics)"
  )
  
  expect_true(
    all(pls_opls$transf$Xscale == 1),
    label = "PLS scale=FALSE: Xscale should be all ones (opls)"
  )
})


test_that("SIMPLS implementations produce consistent results with scale = TRUE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  tol2 <- 1e-5
  
  # Fit models
  simpls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = TRUE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = TRUE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "simpls"
  )
  
  # Predictions
  simpls_basics_predict <- resemble:::predict_opls(
    bo = simpls_basics$bo, 
    b = simpls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = simpls_basics$transf$Xscale
  )
  
  simpls_opls_predict <- resemble:::predict_opls(
    bo = simpls_opls$bo, 
    b = simpls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = simpls_opls$transf$Xscale
  )
  
  # Reconstruction error
  simpls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = simpls_opls$projection_mat, 
    xloadings = simpls_opls$X_loadings,
    scale = TRUE,
    Xcenter = simpls_opls$transf$Xcenter,
    Xscale = simpls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(simpls_projection$weights - simpls_basics$weights)), 
    tol,
    label = "SIMPLS scale=TRUE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(simpls_basics$weights - simpls_opls$weights)), 
    tol,
    label = "SIMPLS scale=TRUE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(simpls_basics_predict - simpls_opls_predict)), 
    tol2,
    label = "SIMPLS scale=TRUE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(simpls_projection$scores - simpls_opls$scores)), 
    tol,
    label = "SIMPLS scale=TRUE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(simpls_gesearch$pred_response - simpls_opls_predict[, ncomp_test])), 
    tol2,
    label = "SIMPLS scale=TRUE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(simpls_gesearch$rmse_reconstruction) - drop(simpls_opls_rec)), 
    tol,
    label = "SIMPLS scale=TRUE: opls_gesearch vs reconstruction_error"
  )
})


test_that("SIMPLS implementations produce consistent results with scale = FALSE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  tol2 <- 1e-5
  
  # Fit models
  simpls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = FALSE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = FALSE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "simpls"
  )
  
  # Predictions
  simpls_basics_predict <- resemble:::predict_opls(
    bo = simpls_basics$bo, 
    b = simpls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = simpls_basics$transf$Xscale
  )
  
  simpls_opls_predict <- resemble:::predict_opls(
    bo = simpls_opls$bo, 
    b = simpls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = simpls_opls$transf$Xscale
  )
  
  # Reconstruction error
  simpls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = simpls_opls$projection_mat, 
    xloadings = simpls_opls$X_loadings,
    scale = FALSE,
    Xcenter = simpls_opls$transf$Xcenter,
    Xscale = simpls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(simpls_projection$weights - simpls_basics$weights)), 
    tol,
    label = "SIMPLS scale=FALSE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(simpls_basics$weights - simpls_opls$weights)), 
    tol,
    label = "SIMPLS scale=FALSE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(simpls_basics_predict - simpls_opls_predict)), 
    tol2,
    label = "SIMPLS scale=FALSE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(simpls_projection$scores - simpls_opls$scores)), 
    tol,
    label = "SIMPLS scale=FALSE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(simpls_gesearch$pred_response - simpls_opls_predict[, ncomp_test])), 
    tol2,
    label = "SIMPLS scale=FALSE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(simpls_gesearch$rmse_reconstruction) - drop(simpls_opls_rec)), 
    tol,
    label = "SIMPLS scale=FALSE: opls_gesearch vs reconstruction_error"
  )
  
  # Xscale should be all ones when scale = FALSE
  expect_true(
    all(simpls_basics$transf$Xscale == 1),
    label = "SIMPLS scale=FALSE: Xscale should be all ones (opls_get_basics)"
  )
  
  expect_true(
    all(simpls_opls$transf$Xscale == 1),
    label = "SIMPLS scale=FALSE: Xscale should be all ones (opls)"
  )
})


test_that("mpls implementations produce consistent results with scale = TRUE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  
  # Fit models
  mpls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = TRUE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = TRUE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "mpls"
  )
  
  # Predictions
  mpls_basics_predict <- resemble:::predict_opls(
    bo = mpls_basics$bo, 
    b = mpls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = mpls_basics$transf$Xscale
  )
  
  mpls_opls_predict <- resemble:::predict_opls(
    bo = mpls_opls$bo, 
    b = mpls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = mpls_opls$transf$Xscale
  )
  
  # Reconstruction error
  mpls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = mpls_opls$projection_mat, 
    xloadings = mpls_opls$X_loadings,
    scale = TRUE,
    Xcenter = mpls_opls$transf$Xcenter,
    Xscale = mpls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(mpls_projection$weights - mpls_basics$weights)), 
    tol,
    label = "mpls scale=TRUE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(mpls_basics$weights - mpls_opls$weights)), 
    tol,
    label = "mpls scale=TRUE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(mpls_basics_predict - mpls_opls_predict)), 
    tol,
    label = "mpls scale=TRUE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(mpls_projection$scores - mpls_opls$scores)), 
    tol,
    label = "mpls scale=TRUE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(mpls_gesearch$pred_response - mpls_opls_predict[, ncomp_test])), 
    tol,
    label = "mpls scale=TRUE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(mpls_gesearch$rmse_reconstruction) - drop(mpls_opls_rec)), 
    tol,
    label = "mpls scale=TRUE: opls_gesearch vs reconstruction_error"
  )
})


test_that("mpls implementations produce consistent results with scale = FALSE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol <- 1e-8
  
  # Fit models
  mpls_basics <- resemble:::opls_get_basics(
    X = train_x, 
    Y = as.matrix(train_y), 
    ncomp = ncomp_test,
    scale = FALSE,
    maxiter = 100,
    tol = 1e-06, 
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_opls <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_projection <- resemble:::opls_for_projection(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,            
    maxiter = 1000,
    tol = 1e-06,
    pcSelmethod = "manual",
    pcSelvalue = ncomp_test, 
    algorithm = "mpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  mpls_gesearch <- resemble:::opls_gesearch(
    Xr = train_x, 
    Yr = as.matrix(train_y), 
    Xu = test_x, 
    ncomp = ncomp_test,
    scale = FALSE,     
    response = TRUE, 
    reconstruction = TRUE,
    similarity = TRUE,
    fresponse = TRUE,
    algorithm = "mpls"
  )
  
  # Predictions
  mpls_basics_predict <- resemble:::predict_opls(
    bo = mpls_basics$bo, 
    b = mpls_basics$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = mpls_basics$transf$Xscale
  )
  
  mpls_opls_predict <- resemble:::predict_opls(
    bo = mpls_opls$bo, 
    b = mpls_opls$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = mpls_opls$transf$Xscale
  )
  
  # Reconstruction error
  mpls_opls_rec <- resemble:::reconstruction_error(
    x = test_x, 
    projection_mat = mpls_opls$projection_mat, 
    xloadings = mpls_opls$X_loadings,
    scale = FALSE,
    Xcenter = mpls_opls$transf$Xcenter,
    Xscale = mpls_opls$transf$Xscale,
    scale_back = TRUE
  )
  
  # Tests
  expect_lt(
    sum(abs(mpls_projection$weights - mpls_basics$weights)), 
    tol,
    label = "mpls scale=FALSE: opls_for_projection vs opls_get_basics weights"
  )
  
  expect_lt(
    sum(abs(mpls_basics$weights - mpls_opls$weights)), 
    tol,
    label = "mpls scale=FALSE: opls_get_basics vs opls weights"
  )
  
  expect_lt(
    sum(abs(mpls_basics_predict - mpls_opls_predict)), 
    tol,
    label = "mpls scale=FALSE: predict_opls consistency"
  )
  
  expect_lt(
    sum(abs(mpls_projection$scores - mpls_opls$scores)), 
    tol,
    label = "mpls scale=FALSE: opls_for_projection vs opls scores"
  )
  
  expect_lt(
    sum(abs(mpls_gesearch$pred_response - mpls_opls_predict[, ncomp_test])), 
    tol,
    label = "mpls scale=FALSE: opls_gesearch vs opls predictions"
  )
  
  expect_lt(
    abs(drop(mpls_gesearch$rmse_reconstruction) - drop(mpls_opls_rec)), 
    tol,
    label = "mpls scale=FALSE: opls_gesearch vs reconstruction_error"
  )
  
  # Xscale should be all ones when scale = FALSE
  expect_true(
    all(mpls_basics$transf$Xscale == 1),
    label = "mpls scale=FALSE: Xscale should be all ones (opls_get_basics)"
  )
  
  expect_true(
    all(mpls_opls$transf$Xscale == 1),
    label = "mpls scale=FALSE: Xscale should be all ones (opls)"
  )
})


test_that("PLS and SIMPLS produce equivalent predictions with scale = TRUE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40
  tol2 <- tol <- 1e-5  # Looser tolerance for cross-algorithm comparison
  tol3 <- 1e-3  # Looser tolerance for coefficients comparison
  # Fit PLS (NIPALS)
  pls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  # Fit SIMPLS
  simpls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  # Predictions
  pls_predict <- resemble:::predict_opls(
    bo = pls_fit$bo, 
    b = pls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = pls_fit$transf$Xscale
  )
  
  simpls_predict <- resemble:::predict_opls(
    bo = simpls_fit$bo, 
    b = simpls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = TRUE,
    Xscale = simpls_fit$transf$Xscale
  )
  
  # PLS and SIMPLS should produce equivalent predictions
  expect_lt(
    sum(abs(pls_predict - simpls_predict)), 
    tol2,
    label = "PLS vs SIMPLS scale=TRUE: predictions equivalence"
  )
  
  # Coefficients should also match
  expect_lt(
    sum(abs(pls_fit$coefficients - simpls_fit$coefficients)), 
    tol3,
    label = "PLS vs SIMPLS scale=TRUE: coefficients equivalence"
  )
})


test_that("PLS and SIMPLS produce equivalent predictions with scale = FALSE", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  tol2 <- 1e-5
  tol3 <- 1e-3 
  # Preprocessing
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det

  # Split data
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  test_idx <- NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)
  
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  test_x <- NIRsoil$spc_pr[test_idx, ]
  
  ncomp_test <- 40

  # Fit PLS (NIPALS)
  pls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  # Fit SIMPLS
  simpls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  # Predictions
  pls_predict <- resemble:::predict_opls(
    bo = pls_fit$bo, 
    b = pls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = pls_fit$transf$Xscale
  )
  
  simpls_predict <- resemble:::predict_opls(
    bo = simpls_fit$bo, 
    b = simpls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = test_x,
    scale = FALSE,
    Xscale = simpls_fit$transf$Xscale
  )
  
  # PLS and SIMPLS should produce equivalent predictions
  expect_lt(
    sum(abs(pls_predict - simpls_predict)), 
    tol2,
    label = "PLS vs SIMPLS scale=FALSE: predictions equivalence"
  )
  
  # Coefficients should also match
  expect_lt(
    sum(abs(pls_fit$coefficients - simpls_fit$coefficients)), 
    tol3,
    label = "PLS vs SIMPLS scale=FALSE: coefficients equivalence"
  )
})


test_that("PLS overfits training data with many components (scale = TRUE)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 0, p = 1, w = 17
  )
  NIRsoil$spc_pr <- sg_det
  
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  
  ncomp_test <- 200 #min(dim(train_x)) - 1
  
  # Fit PLS (NIPALS)
  pls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_pred <- resemble:::predict_opls(
    bo = pls_fit$bo, 
    b = pls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = train_x,
    scale = TRUE,
    Xscale = pls_fit$transf$Xscale
  )
  
  pred <- pls_pred[, ncomp_test]
  obs <- train_y
  residuals <- pred - obs
  # With 100 components, training predictions should nearly perfectly match observations
  # Any systematic bias or corruption would show up here
  
  # Mean bias should be essentially zero (not shifted up or down)
  expect_lt(abs(mean(residuals)), 1e-5, 
            label = "PLS scale=TRUE: no systematic bias")
  
  # use relative error
  expect_lt(max(abs(residuals)) / diff(range(obs)), 0.05, 
            label = "PLS scale=TRUE: max relative error < 5%")
  
  # R2 should be essentially 1
  expect_gt(cor(pred, obs)^2, 0.98, 
            label = "PLS scale=TRUE: near-perfect fit")
  
  # Slope of pred vs obs should be ~1 (no scaling issues)
  slope <- cov(pred, obs) / var(obs)
  expect_lt(abs(slope - 1), 0.98, 
            label = "PLS scale=TRUE: slope near 1")
  
  # Fit SIMPLS
  simpls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = TRUE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_pred <- resemble:::predict_opls(
    bo = simpls_fit$bo, 
    b = simpls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = train_x,
    scale = TRUE,
    Xscale = simpls_fit$transf$Xscale
  )
  
  pred_simpls <- simpls_pred[, ncomp_test]
  residuals_simpls <- pred_simpls - obs
  
  expect_lt(abs(mean(residuals_simpls)), 1e-5, 
            label = "SIMPLS scale=TRUE: no systematic bias")
  expect_lt(max(abs(residuals_simpls)) / diff(range(obs)), 0.05, 
            label = "SIMPLS scale=TRUE: max relative error < 5%")
  expect_gt(cor(pred_simpls, obs)^2, 0.98, 
            label = "SIMPLS scale=TRUE: near-perfect fit")
  
  slope_simpls <- cov(pred_simpls, obs) / var(obs)
  expect_lt(abs(slope_simpls - 1), 0.98, 
            label = "SIMPLS scale=TRUE: slope near 1")
  
  # PLS and SIMPLS should be equivalent
  expect_lt(max(abs(pred - pred_simpls)), 1e-5,
            label = "PLS vs SIMPLS scale=TRUE: equivalent predictions")
})


test_that("PLS overfits training data with many components (scale = FALSE)", {
  skip_on_cran()
  skip_if_not_installed("prospectr")
  
  library(prospectr)
  data(NIRsoil, package = "prospectr")
  
  sg_det <- savitzkyGolay(
    detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
    m = 1, p = 1, w = 7
  )
  NIRsoil$spc_pr <- sg_det
  
  train_idx <- NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)
  train_x <- NIRsoil$spc_pr[train_idx, ]
  train_y <- NIRsoil$Ciso[train_idx]
  
  ncomp_test <- floor(0.75 * min(dim(train_x)) - 1)
  
  # Fit PLS (NIPALS)
  pls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "pls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  pls_pred <- resemble:::predict_opls(
    bo = pls_fit$bo, 
    b = pls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = train_x,
    scale = FALSE,
    Xscale = pls_fit$transf$Xscale
  )
  
  pred <- pls_pred[, ncomp_test]
  obs <- train_y
  residuals <- pred - obs
  # plot(pred, obs)
  # abline(0, 1)
  
  expect_lt(abs(mean(residuals)), 1e-5, 
            label = "PLS scale=FALSE: no systematic bias")
  expect_lt(max(abs(residuals)) / diff(range(obs)), 0.05, 
            label = "PLS scale=FALSE: max relative error < 5%")
  expect_gt(cor(pred, obs)^2, 0.9999, 
            label = "PLS scale=FALSE: near-perfect fit")
  
  slope <- cov(pred, obs) / var(obs)
  expect_lt(abs(slope - 1), 1e-4, 
            label = "PLS scale=FALSE: slope near 1")
  
  # Fit SIMPLS
  simpls_fit <- resemble:::opls(
    X = train_x, 
    Y = as.matrix(train_y),
    ncomp = ncomp_test,
    scale = FALSE,           
    maxiter = 1000,
    tol = 1e-06,
    algorithm = "simpls", 
    xls_min_w = 3, 
    xls_max_w = 15
  )
  
  simpls_pred <- resemble:::predict_opls(
    bo = simpls_fit$bo, 
    b = simpls_fit$coefficients, 
    ncomp = ncomp_test, 
    newdata = train_x,
    scale = FALSE,
    Xscale = simpls_fit$transf$Xscale
  )
  
  pred_simpls <- simpls_pred[, ncomp_test]
  residuals_simpls <- pred_simpls - obs
  
  expect_lt(abs(mean(residuals_simpls)), 1e-5, 
            label = "SIMPLS scale=FALSE: no systematic bias")
  expect_lt(max(abs(residuals)) / diff(range(obs)), 0.05, 
            label = "SIMPLS scale=FALSE: max relative error < 5%")
  expect_gt(cor(pred_simpls, obs)^2, 0.9999, 
            label = "SIMPLS scale=FALSE: near-perfect fit")
  
  slope_simpls <- cov(pred_simpls, obs) / var(obs)
  expect_lt(abs(slope_simpls - 1), 1e-3, 
            label = "SIMPLS scale=FALSE: slope near 1")
  
  # PLS and SIMPLS should be equivalent
  expect_lt(mean(abs(pred - pred_simpls)), 0.04,
            label = "PLS vs SIMPLS scale=FALSE: equivalent predictions")
})

#' @title Control parameters for liblex
#'
#' @description
#' Specifies control parameters for the \code{\link{liblex}} function, including
#' output options, validation settings, tuning behavior, and parallel execution.
#'
#' @param return_dissimilarity A logical indicating whether the dissimilarity
#'   matrix should be returned in the output. Default is \code{FALSE}. Setting
#'   to \code{TRUE} can be useful for diagnostics but increases memory usage
#'   for large libraries.
#' @param mode A character string specifying the operation mode:
#'   \describe{
#'     \item{\code{"build"}}{(default) Builds the library of local models. If
#'       \code{tune = TRUE}, validation is performed first to find optimal
#'       parameters, then the library is built using those parameters. If
#'       \code{tune = FALSE}, the library is built directly using the
#'       parameters provided to \code{\link{liblex}}.}
#'     \item{\code{"validate"}}{Performs validation only without building the
#'       library. Useful for parameter exploration or when testing different
#'       configurations before committing to the full library build.}
#'   }
#' @param tune A logical indicating whether to optimize parameters via
#'   nearest-neighbor validation. Default is \code{FALSE}. When \code{TRUE},
#'   the function evaluates all combinations of \code{k} values and PLS
#'   component ranges, selecting the combination that minimizes RMSE (or
#'   maximizes R², depending on \code{metric}). See Details.
#' @param metric A character string specifying the performance metric used
#'   for parameter selection when \code{tune = TRUE}. Options are:
#'   \describe{
#'     \item{\code{"rmse"}}{(default) Root mean squared error (minimized).}
#'     \item{\code{"r2"}}{Coefficient of determination (maximized).}
#'   }
#'   Ignored when \code{tune = FALSE}.
#' @param chunk_size An integer specifying the number of local models to
#'   process per parallel task. Default is \code{1L}. Increasing this value
#'   reduces parallel overhead but may cause load imbalance. For large
#'   libraries, values between 10 and 50 often provide a good trade-off
#'   between overhead and efficiency.
#' @param allow_parallel A logical indicating whether parallel execution is
#'   permitted. Default is \code{TRUE}. Parallelization is applied to the
#'   model fitting loop using the \pkg{foreach} package. Requires a registered
#'   parallel backend (e.g., via \pkg{doParallel}).
#' @param blas_threads An integer specifying the number of threads for BLAS
#'   operations. Default is \code{1L}, which avoids thread contention when
#'   using parallel processing. Requires the \pkg{RhpcBLASctl} package to
#'   take effect. On Linux systems with multi-threaded BLAS (e.g., OpenBLAS),
#'   setting this to 1 can substantially improve performance when
#'   \code{allow_parallel = TRUE}.
#'
#' @details
#' \subsection{Nearest-neighbor validation}{
#' When \code{tune = TRUE} or \code{mode = "validate"}, the function performs
#' nearest-neighbor validation (NNv) to assess model performance. For each
#' observation in the reference set (or anchor set, if specified), the
#' procedure:
#' \enumerate{
#'   \item Identifies the k nearest neighbors of the target observation.
#'   \item Excludes the target observation (and any observations in the same
#'     group, if \code{group} is specified) from the neighbor set.
#'   \item Fits a local model using the remaining neighbors.
#'   \item Predicts the response value for the excluded target observation.
#'   \item Computes prediction errors across all observations.
#' }
#'
#' This leave-one-out style validation provides an estimate of prediction
#' performance without requiring a separate test set. When \code{tune = TRUE},
#' the parameter combination (number of neighbors and PLS component range)
#' yielding the best performance according to \code{metric} is selected for
#' building the final library.
#' }
#'
#' \subsection{Mode and tune combinations}{
#' \tabular{lll}{
#'   \code{mode} \tab \code{tune} \tab Behavior \cr
#'   \code{"build"} \tab \code{FALSE} \tab Build library using parameters as
#'     provided \cr
#'   \code{"build"} \tab \code{TRUE} \tab Validate, find optimal parameters,
#'     build library \cr
#'   \code{"validate"} \tab \code{FALSE} \tab Validate only, report
#'     performance statistics \cr
#'   \code{"validate"} \tab \code{TRUE} \tab Validate, report statistics with
#'     optimal parameters identified \cr
#' }
#' }
#'
#' \subsection{Parallel chunk size}{
#' The \code{chunk_size} parameter controls granularity of parallel work
#' distribution. When \code{allow_parallel = TRUE} and a parallel backend is
#' registered:
#' \itemize{
#'   \item \code{chunk_size = 1}: Each local model is a separate parallel task.
#'     Maximum parallelism but higher scheduling overhead.
#'   \item \code{chunk_size > 1}: Multiple models are processed sequentially
#'     within each parallel task. Reduces overhead and improves memory
#'     locality, but may cause load imbalance if the number of models is not
#'     evenly divisible.
#' }
#' When \code{allow_parallel = FALSE}, \code{chunk_size} has no effect.
#' }
#'
#' @return A list of class \code{liblex_control} containing the validated
#'   control parameters.
#'
#' @seealso \code{\link{liblex}}, \code{\link{mbl_control}}
#'
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#'
#' @examples
#' # Default settings: build library without tuning
#' liblex_control()
#'
#' # Tune parameters before building
#' liblex_control(tune = TRUE)
#'
#' # Validate only for parameter exploration
#' liblex_control(mode = "validate", tune = TRUE)
#'
#' # Include dissimilarity matrix in output
#' liblex_control(return_dissimilarity = TRUE)
#'
#' # Larger chunks for reduced parallel overhead
#' liblex_control(chunk_size = 20L)
#'
#' # Parallel settings for Linux with OpenBLAS
#' liblex_control(allow_parallel = TRUE, blas_threads = 1L)
#'
#' @export
liblex_control <- function(
    return_dissimilarity = FALSE,
    mode = c("build", "validate"),
    tune = FALSE,
    metric = c("rmse", "r2"),
    chunk_size = 1L,
    allow_parallel = TRUE,
    blas_threads = 1L
) {
  mode <- match.arg(mode)
  metric <- match.arg(metric)
  
  if (!is.logical(return_dissimilarity) || length(return_dissimilarity) != 1L ||
      is.na(return_dissimilarity)) {
    stop("'return_dissimilarity' must be TRUE or FALSE", call. = FALSE)
  }
  
  if (!is.logical(tune) || length(tune) != 1L || is.na(tune)) {
    stop("'tune' must be TRUE or FALSE", call. = FALSE)
  }
  
  if (!is.numeric(chunk_size) || length(chunk_size) != 1L ||
      is.na(chunk_size) || chunk_size < 1L) {
    stop("'chunk_size' must be a positive integer", call. = FALSE)
  }
  chunk_size <- as.integer(chunk_size)
  
  if (!is.logical(allow_parallel) || length(allow_parallel) != 1L ||
      is.na(allow_parallel)) {
    stop("'allow_parallel' must be TRUE or FALSE", call. = FALSE)
  }
  
  if (!is.numeric(blas_threads) || length(blas_threads) != 1L ||
      is.na(blas_threads) || blas_threads < 1L) {
    stop("'blas_threads' must be a positive integer", call. = FALSE)
  }
  blas_threads <- as.integer(blas_threads)
  
  structure(
    list(
      return_dissimilarity = return_dissimilarity,
      mode = mode,
      tune = tune,
      metric = metric,
      chunk_size = chunk_size,
      allow_parallel = allow_parallel,
      blas_threads = blas_threads
    ),
    class = "liblex_control"
  )
}

# =============================================================================
# Euclidean, Mahalanobis and cosine dissimilarity — redesigned API
# =============================================================================
# Public:   diss_euclidean()    — method constructor
#           diss_mahalanobis()  — method constructor
#           diss_cosine()       — method constructor
# Internal: .f_diss_compute()   — shared raw computation, not exported


# -----------------------------------------------------------------------------
#' @title Euclidean dissimilarity method constructor
#'
#' @description
#' Creates a configuration object for computing Euclidean dissimilarity.
#' Pass the result to \code{dissimilarity()} to compute the dissimilarity
#' matrix.
#'
#' The scaled Euclidean dissimilarity between two observations \eqn{x_i} and
#' \eqn{x_j} is:
#'
#' \deqn{d(x_i, x_j) = \sqrt{\frac{1}{p} \sum_{k=1}^{p}(x_{i,k} - x_{j,k})^2}}
#'
#' where \eqn{p} is the number of variables. Results are equivalent to
#' \code{stats::dist()} but scaled by \eqn{1/p}.
#'
#' @param center Logical. Center the data before computing distances?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{TRUE}.
#' @param scale Logical. Scale the data before computing distances?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{FALSE}.
#'
#' @return An object of class \code{c("diss_euclidean", "diss_method")}.
#' 
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} 
#' 
#' @seealso \code{\link{dissimilarity}}, \code{\link{diss_mahalanobis}},
#'   \code{\link{diss_cosine}}
#' @examples
#' m <- diss_euclidean()
#' m <- diss_euclidean(center = FALSE, scale = TRUE)
#' @export
diss_euclidean <- function(center = TRUE, scale = FALSE) {
  .new_diss_method("euclidean", center = center, scale = scale)
}


# -----------------------------------------------------------------------------
#' @title Mahalanobis dissimilarity method constructor
#'
#' @description
#' Creates a configuration object for computing Mahalanobis dissimilarity.
#' Pass the result to \code{dissimilarity()} to compute the dissimilarity
#' matrix.
#'
#' The Mahalanobis distance is computed by first transforming the data into
#' Mahalanobis space via a factorization of the inverse covariance matrix
#' \eqn{M^{-1} = W^{T}W} (using SVD), then applying Euclidean distance in
#' that transformed space:
#'
#' \deqn{d(x_i, x_j) = \sqrt{\frac{1}{p}(x_i - x_j)M^{-1}(x_i - x_j)^T}}
#'
#' @section Important limitations:
#' The covariance matrix will be singular — and the distance therefore
#' uncomputable — when the number of observations is smaller than the number
#' of variables, or when variables are perfectly collinear. This is common
#' with raw spectral data; consider using \code{diss_euclidean()} on
#' PCA scores instead.
#'
#' @param center Logical. Center the data before computing distances?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{TRUE}.
#' @param scale Logical. Scale the data before computing distances?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{FALSE}.
#'
#' @return An object of class \code{c("diss_mahalanobis", "diss_method")}.
#' 
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} 
#' 
#' @seealso \code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
#'   \code{\link{diss_cosine}}
#' @examples
#' m <- diss_mahalanobis()
#' m <- diss_mahalanobis(center = TRUE, scale = TRUE)
#' @export
diss_mahalanobis <- function(center = TRUE, scale = FALSE) {
  .new_diss_method("mahalanobis", center = center, scale = scale)
}


# -----------------------------------------------------------------------------
#' @title Cosine dissimilarity method constructor
#'
#' @description
#' Creates a configuration object for computing cosine dissimilarity
#' (also known as spectral angle mapper). Pass the result to
#' \code{dissimilarity()} to compute the dissimilarity matrix.
#'
#' The cosine dissimilarity between two observations \eqn{x_i} and
#' \eqn{x_j} is:
#'
#' \deqn{c(x_i, x_j) = \cos^{-1}
#'   \frac{\sum_{k=1}^{p} x_{i,k}\, x_{j,k}}
#'        {\sqrt{\sum_{k=1}^{p} x_{i,k}^{2}}\,
#'         \sqrt{\sum_{k=1}^{p} x_{j,k}^{2}}}}
#'
#' where \eqn{p} is the number of variables.
#'
#' @param center Logical. Center the data before computing dissimilarities?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{TRUE}.
#' @param scale Logical. Scale the data before computing dissimilarities?
#'   Applied jointly to \code{Xr} and \code{Xu} if both are provided.
#'   Default \code{FALSE}.
#'
#' @return An object of class \code{c("diss_cosine", "diss_method")}.
#' 
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} 
#' 
#' @seealso \code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
#'   \code{\link{diss_mahalanobis}}
#' @examples
#' m <- diss_cosine()
#' m <- diss_cosine(center = FALSE)
#' @export
diss_cosine <- function(center = TRUE, scale = FALSE) {
  .new_diss_method("cosine", center = center, scale = scale)
}




# =============================================================================
# Internal shared compute function — NOT exported
# =============================================================================
#
# Called by dissimilarity() after data-level validation is done there.
# Assumes:
#   - Xr and Xu are numeric matrices with no NAs
#   - ncol(Xu) == ncol(Xr) if Xu is provided
#   - for Mahalanobis: nrow(rbind(Xr, Xu)) > ncol(Xr) — checked in dissimilarity()
#
.f_diss_compute <- function(Xr, Xu = NULL, method) {
  
  stopifnot(inherits(method, "diss_method"))
  stopifnot(method$method %in% c("euclidean", "mahalanobis", "cosine"))
  
  center      <- method$center
  scale       <- method$scale
  method_name <- method$method
  
  # --- internal method label used for C++ dispatch --------------------------
  # Mahalanobis is transformed to Euclidean space before calling fast_diss(),
  # so the C++ backend always sees either "euclid" or "cosine"
  cpp_method <- if (method_name == "cosine") "cosine" else "euclid"
  
  # --- preprocessing --------------------------------------------------------
  # Note: Euclidean and Mahalanobis always go through rbind() because
  # Mahalanobis needs the full combined X to estimate the covariance matrix.
  # Cosine only needs it if center or scale are requested.
  needs_combined <- center || scale || method_name %in% c("euclidean", "mahalanobis")
  
  if (needs_combined) {
    X <- rbind(Xr, Xu)  # rbind(x, NULL) == x, no special casing needed
    
    if (center) {
      X <- sweep(X, MARGIN = 2, STATS = colMeans(X), FUN = "-")
    }
    if (scale) {
      X <- sweep(X, MARGIN = 2, STATS = get_col_sds(X), FUN = "/")
    }
    
    # --- Mahalanobis transform -----------------------------------------------
    # Project X into Mahalanobis space; afterwards treat as Euclidean
    if (method_name == "mahalanobis") {
      X <- try(euclid_to_mahal(X, sm_method = "svd"), silent = TRUE)
      if (!is.matrix(X)) {
        stop(
          "The covariance matrix is exactly singular and cannot be inverted. ",
          "The Mahalanobis distance cannot be computed. ",
          "Consider using diss_euclidean() on PCA scores instead."
        )
      }
    }
    
    if (!is.null(Xu)) {
      n_xu <- nrow(Xu)
      Xu   <- X[(nrow(X) - n_xu + 1L):nrow(X), , drop = FALSE]
      Xr   <- X[1L:(nrow(X) - n_xu), , drop = FALSE]
    } else {
      Xr <- X
    }
    rm(X)
  }
  
  # --- compute dissimilarity ------------------------------------------------
  if (!is.null(Xu)) {
    # abs() guard: C++ backend can return small negative values (~-1e-14)
    # due to floating point reuse in memory
    result <- abs(fast_diss(Xu, Xr, cpp_method))
    if (cpp_method == "euclid") {
      result <- sqrt(result / ncol(Xr))
    }
    rownames(result) <- paste("Xr", seq_len(nrow(Xr)), sep = "_")
    colnames(result) <- paste("Xu", seq_len(nrow(Xu)), sep = "_")
  } else {
    result <- abs(fast_diss(Xr, Xr, cpp_method))
    if (cpp_method == "euclid") {
      result <- sqrt(result / ncol(Xr))
    }
    rownames(result) <- colnames(result) <- paste("Xr", seq_len(nrow(Xr)), sep = "_")
  }
  
  # --- cosine NaN guard ------------------------------------------------------
  # The C++ backend occasionally returns NaN for identical observations
  # due to floating point precision in the arccos computation.
  # These are self-distances and are correctly zero.
  if (method_name == "cosine") {
    result[is.nan(result)] <- 0
  }
  
  result
}

context("test-euclidean")

test_that("dist() equivalence to resemble euclidean", {

  skip_if_not_installed("prospectr")
  data("NIRsoil", package = "prospectr")
  Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
  
  tol <- 1e-3
  dist_m <- as.matrix(dist(Xr, method = "euclidean"))
  diss_m <- dissimilarity(
    Xr = Xr, diss_method = diss_euclidean(center = FALSE, scale = FALSE)
  )$dissimilarity
  diss_m <- sqrt(diss_m^2 * ncol(Xr))
  expect_lt(sum(abs(diss_m - dist_m)), tol)
})

#' @title Compute dissimilarity matrices
#'
#' @description
#' \loadmathjax
#' Computes dissimilarity matrices between observations using various methods.
#' This is the main interface for dissimilarity computation in the resemble
#' package.
#'
#' @param Xr A numeric matrix of reference observations (rows) and variables
#'   (columns).
#' @param Xu Optional matrix of additional observations with the same variables.
#' @param diss_method A dissimilarity method object created by one of:
#'   \itemize{
#'     \item \code{\link{diss_pca}()}: Mahalanobis distance in PCA space
#'     \item \code{\link{diss_pls}()}: Mahalanobis distance in PLS space
#'     \item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
#'     \item \code{\link{diss_euclidean}()}: Euclidean distance
#'     \item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
#'     \item \code{\link{diss_cosine}()}: Cosine dissimilarity
#'   }
#'   Default is \code{diss_pca()}.
#' @param Yr Optional response matrix. Required for PLS methods and when using
#'   \code{ncomp_by_opc()}.
#'
#' @details
#' The function dispatches to the appropriate internal computation based on the
#' class of \code{diss_method}. Each method constructor (e.g., \code{diss_pca()})
#' encapsulates all method-specific parameters including component selection,
#' centering, scaling, and whether to return projections.
#'
#' \subsection{Output dimensions}{
#' When only \code{Xr} is provided, the function computes pairwise dissimilarities
#' among all observations in \code{Xr}, returning a symmetric
#' \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xr)} matrix.
#'
#' When both \code{Xr} and \code{Xu} are provided, the function computes
#' dissimilarities between each observation in \code{Xr} and each observation
#' in \code{Xu}, returning a \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xu)}
#' matrix where element \mjeqn{(i, j)}{(i, j)} is the dissimilarity between the
#' \mjeqn{i}{i}-th observation in \code{Xr} and the \mjeqn{j}{j}-th observation
#' in \code{Xu}.
#' }
#'
#' \subsection{Mahalanobis distance}{
#' Note that \code{diss_mahalanobis()} computes Mahalanobis distance directly on
#' the input variables. This requires the covariance matrix to be invertible,
#' which fails when the number of variables exceeds the number of observations
#' or when variables are highly correlated (common in spectral data). For such
#' cases, use \code{diss_pca()} or \code{diss_pls()} instead.
#' }
#'
#' @return A list of class \code{"dissimilarity"} containing:
#' \describe{
#'   \item{dissimilarity}{The computed dissimilarity matrix. Dimensions are
#'     \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xr)} when \code{Xu = NULL},
#'     or \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xu)} otherwise.}
#'   \item{diss_method}{The \code{diss_*} constructor object used for computation.}
#'   \item{center}{Vector used to center the data.}
#'   \item{scale}{Vector used to scale the data.}
#'   \item{ncomp}{Number of components used (for projection methods).}
#'   \item{projection}{If \code{return_projection = TRUE} in the method
#'     constructor, the \code{ortho_projection} object.}
#' }
#' 
#' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} 
#' 
#' @seealso
#' \code{\link{diss_pca}}, \code{\link{diss_pls}}, 
#' \code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
#' \code{\link{diss_mahalanobis}}, \code{\link{diss_cosine}}
#' 
#'
#' @references
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
#' 268-279.
#'
#' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
#' with soil vis-NIR spectra. Geoderma 199, 43-53.
#'
#'
#' @examples
#' \donttest{
#' library(prospectr)
#' data(NIRsoil)
#'
#' # Preprocess
#' sg <- savitzkyGolay(NIRsoil$spc, m = 1, p = 4, w = 15)
#'
#' Xr <- sg[as.logical(NIRsoil$train), ]
#' Xu <- sg[!as.logical(NIRsoil$train), ]
#' Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
#' Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
#'
#' Xu <- Xu[!is.na(Yu), ]
#' Xr <- Xr[!is.na(Yr), ]
#' Yr <- Yr[!is.na(Yr)]
#'
#' # PCA-based dissimilarity with variance-based selection
#' d1 <- dissimilarity(Xr, Xu, diss_method = diss_pca())
#'
#' # PCA with OPC selection (requires Yr)
#' d2 <- dissimilarity(Xr, Xu,
#'   Yr = Yr,
#'   diss_method = diss_pca(
#'     ncomp = ncomp_by_opc(30),
#'     return_projection = TRUE
#'   )
#' )
#'
#' # PLS-based dissimilarity 
#' d3 <- dissimilarity(
#'   Xr, Xu,
#'   Yr = Yr,
#'   diss_method = diss_pls(
#'     ncomp = ncomp_by_opc(30)
#'   )
#' )
#'
#' # Euclidean distance
#' d4 <- dissimilarity(Xr, Xu, diss_method = diss_euclidean())
#'
#' # Correlation dissimilarity with moving window
#' d5 <- dissimilarity(Xr, Xu, diss_method = diss_correlation(ws = 41))
#'
#' # Mahalanobis distance (use only when n > p and low collinearity)
#' # d6 <- dissimilarity(Xr[, 1:20], Xu[, 1:20],
#' #                     diss_method = diss_mahalanobis())
#' }
#'
#' @export
dissimilarity <- function(
  Xr,
  Xu = NULL,
  diss_method = diss_pca(),
  Yr = NULL
) {
  # ---------------------------------------------------------------------------
  # Handle legacy character-based diss_method
  # ---------------------------------------------------------------------------
  if (is.character(diss_method)) {
    stop(
      "Character-based 'diss_method' is no longer supported.\n\n",
      "Use method constructors instead:\n\n",
      "  Old API                          -> New API\n",
      "  -------                             -------\n",
      "  diss_method = \"pca\"               -> diss_pca()\n",
      "  diss_method = \"pca.nipals\"        -> diss_pca(method = \"pca_nipals\")\n",
      "  diss_method = \"pls\"               -> diss_pls()\n",
      "  diss_method = \"mpls\"              -> diss_pls(method = \"mpls\")\n",
      "  diss_method = \"euclid\"            -> diss_euclidean()\n",
      "  diss_method = \"cosine\"            -> diss_cosine()\n",
      "  diss_method = \"cor\"               -> diss_correlation()\n\n",
      "  diss_method = \"sid\"               -> Deprecated\n",
      "  pc_selection = list(\"opc\", 30)    -> ncomp = ncomp_by_opc(30)\n",
      "  gh = TRUE                           -> Deprecated\n",
      call. = FALSE
    )
  }
  
  # ---------------------------------------------------------------------------
  # check Xr and Xu
  # ---------------------------------------------------------------------------
  if (is.numeric(Xr) && is.null(dim(Xr))) {
    Xr <- matrix(Xr, nrow = 1)
  }
  
  if (!is.null(Xu)) {
    if (is.numeric(Xu) && is.null(dim(Xu))) {
      Xu <- matrix(Xu, nrow = 1)
    }
    if (ncol(Xr) != ncol(Xu)) {
      stop("'Xr' and 'Xu' must have the same number of columns.", call. = FALSE)
    }
  } else {
    if (nrow(Xr) < 2L) {
      stop("'Xr' must have at least 2 rows when 'Xu' is not provided.", call. = FALSE)
    }
  }
  
  # ---------------------------------------------------------------------------
  # Validate diss_method
  # ---------------------------------------------------------------------------
  if (!inherits(diss_method, "diss_method")) {
    stop(
      "'diss_method' must be a dissimilarity method object.\n",
      "Use one of: diss_pca(), diss_pls(), diss_correlation(),",  
      "diss_euclidean(), diss_mahalanobis(), diss_cosine()."
    )
  }

  # ---------------------------------------------------------------------------
  # Validate Yr requirements
  # ---------------------------------------------------------------------------
  method_class <- class(diss_method)[[1]]

  requires_yr <- method_class == "diss_pls" ||
    (method_class == "diss_pca" && inherits(diss_method$ncomp, "ncomp_by_opc"))

  if (requires_yr && is.null(Yr)) {
    stop("'Yr' is required for this dissimilarity method.")
  }

  # ---------------------------------------------------------------------------
  # Dispatch to compute function
  # ---------------------------------------------------------------------------
  result <- switch(
    method_class,
    diss_pca = .diss_pca_compute(Xr, Xu, Yr, diss_method),
    diss_pls = .diss_pls_compute(Xr, Xu, Yr, diss_method),
    diss_euclidean = .diss_euclidean_compute(Xr, Xu, diss_method),
    diss_mahalanobis = .diss_mahalanobis_compute(Xr, Xu, diss_method),
    diss_cosine = .diss_cosine_compute(Xr, Xu, diss_method),
    diss_correlation = .diss_correlation_compute(Xr, Xu, diss_method),
    stop("Unknown dissimilarity method: ", method_class)
  )
  
  # Fix diagonal for self-dissimilarity (avoid floating-point artifacts)
  if (is.null(Xu)) {
    diag(result$dissimilarity) <- 0
  }
  
  result$diss_method <- diss_method
  
  if (diss_method$center) {
    result$center <- drop(get_column_means(rbind(Xr, Xu)))
  } else {
    result$center <- NULL
  }
  
  if (diss_method$scale) {
    result$scale <- get_col_sds(rbind(Xr, Xu))
  } else {
    result$scale <- NULL
  }  

  class(result) <- c("resemble_dissimilarity", "dissimilarity", "list")
  result
}


# =============================================================================
# Internal compute functions
# =============================================================================

.diss_pca_compute <- function(Xr, Xu, Yr, method) {
  proj <- ortho_projection(
    Xr = Xr,
    Xu = Xu,
    Yr = Yr,
    ncomp = method$ncomp,
    method = method$method,
    center = method$center,
    scale = method$scale
  )

  # Standardize scores by SD before computing Euclidean distance
  scores <- proj$scores
  scores_sd <- apply(scores, 2L, sd)
  scores_sd[scores_sd < .Machine$double.eps] <- 1
  scores_std <- sweep(scores, 2L, scores_sd, "/")

  # Split scores
  n_xr <- nrow(Xr)
  scores_xr <- scores_std[seq_len(n_xr), , drop = FALSE]
  scores_xu <- if (!is.null(Xu)) {
    scores_std[(n_xr + 1L):nrow(scores_std), , drop = FALSE]
  } else {
    NULL
  }

  # Euclidean distance on standardized scores
  dmat <- f_diss(
    Xr = scores_xr,
    Xu = scores_xu,
    diss_method = "euclid",
    center = FALSE,
    scale = FALSE
  )

  result <- list(
    dissimilarity = dmat,
    ncomp = proj$ncomp
  )

  if (method$return_projection) {
    result$projection <- proj
  }

  result
}


.diss_pls_compute <- function(Xr, Xu, Yr, method) {
  if (is.null(Yr)) {
    stop("'Yr' is required for PLS dissimilarity.")
  }

  proj <- ortho_projection(
    Xr = Xr,
    Xu = Xu,
    Yr = Yr,
    ncomp = method$ncomp,
    method = method$method,
    center = TRUE,
    scale = method$scale
  )

  scores <- proj$scores
  n_xr <- nrow(Xr)

  scores_xr <- scores[seq_len(n_xr), , drop = FALSE]
  scores_xu <- if (!is.null(Xu)) {
    scores[(n_xr + 1L):nrow(scores), , drop = FALSE]
  } else {
    NULL
  }

  # Mahalanobis distance in PLS space
  dmat <- f_diss(
    Xr = scores_xr,
    Xu = scores_xu,
    diss_method = "mahalanobis",
    center = FALSE,
    scale = FALSE
  )

  result <- list(
    dissimilarity = dmat,
    ncomp = proj$ncomp
  )

  if (method$return_projection) {
    result$projection <- proj
  }
  
  result
}


.diss_euclidean_compute <- function(Xr, Xu, method) {
  dmat <- f_diss(
    Xr = Xr,
    Xu = Xu,
    diss_method = "euclid",
    center = method$center,
    scale = method$scale
  )

  list(dissimilarity = dmat)
}


.diss_mahalanobis_compute <- function(Xr, Xu, method) {
  dmat <- f_diss(
    Xr = Xr,
    Xu = Xu,
    diss_method = "mahalanobis",
    center = method$center,
    scale = method$scale
  )

  list(dissimilarity = dmat)
}


.diss_cosine_compute <- function(Xr, Xu, method) {
  dmat <- f_diss(
    Xr = Xr,
    Xu = Xu,
    diss_method = "cosine",
    center = method$center,
    scale = method$scale
  )

  list(dissimilarity = dmat)
}


.diss_correlation_compute <- function(Xr, Xu, method) {
  dmat <- .cor_diss_compute(
    Xr = Xr,
    Xu = Xu,
    method = method
  )

  result <- list(dissimilarity = dmat)

  if (!is.null(method$ws)) {
    result$ws <- method$ws
  }

  result
}

# =============================================================================
# Helper: Coerce ncomp argument
# =============================================================================

.coerce_ncomp <- function(ncomp) {
  if (is.numeric(ncomp) && length(ncomp) == 1L) {
    if (is.na(ncomp) || ncomp < 1L) {
      stop("'ncomp' must be a positive integer.")
    }
    return(ncomp_fixed(as.integer(ncomp)))
  }
  if (!inherits(ncomp, "ncomp_selection")) {
    stop(
      "'ncomp' must be a positive integer or an ncomp_*() object.\n",
      "See ?ncomp_by_var, ?ncomp_by_cumvar, ?ncomp_by_opc, ?ncomp_fixed."
    )
  }
  ncomp
}


# =============================================================================
# Print method
# =============================================================================

#' @export
print.resemble_dissimilarity <- function(x, ...) {
  dmat <- x$dissimilarity
  cat("Dissimilarity matrix\n")
  cat("  Dimensions:", nrow(dmat), "x", ncol(dmat), "\n\n")
  cat("Constructor:\n")
  print(x$diss_method)
  if (!is.null(x$ncomp)) {
    cat("\nUsed ncomp: ", x$ncomp, "\n")
  }
  invisible(x)
}

# -----------------------------------------------------------------------------
# Shared constructor helper — not exported
# Avoids repeating the same center/scale/precision validation three times
# -----------------------------------------------------------------------------
.new_diss_method <- function(method_name, center, scale, extra = list()) {
  
  if (!is.logical(center) || length(center) != 1L) {
    stop("'center' must be a single logical value (TRUE or FALSE).")
  }
  if (!is.logical(scale) || length(scale) != 1L) {
    stop("'scale' must be a single logical value (TRUE or FALSE).")
  }
  
  structure(
    c(list(method = method_name, center = center, scale = scale), extra),
    class = c(paste0("diss_", method_name), "diss_method")
  )
}

# -----------------------------------------------------------------------------
# Shared print method for all three — dispatches on "diss_method" base class
# -----------------------------------------------------------------------------
#' @export
print.diss_method <- function(x, ...) {
  cat(
    "Dissimilarity method:", x$method, "\n",
    " center :", x$center, "\n",
    " scale  :", x$scale, "\n"
  )
  invisible(x)
}

#' #' @title A function for computing dissimilarity matrices from orthogonal
#' #' projections (ortho_diss)
#' #' @description
#' #' \loadmathjax
#' #' This function computes dissimilarities (in an orthogonal space) between
#' #' either observations in a given set or between observations in two different
#' #' sets.The dissimilarities are computed based on either principal component
#' #' projection or partial least squares projection of the data. After projecting
#' #' the data, the Mahalanobis distance is applied.
#' #' @usage
#' #' ortho_diss(Xr, Xu = NULL,
#' #'            Yr = NULL,
#' #'            pc_selection = list(method = "var", value = 0.01),
#' #'            diss_method = "pca",
#' #'            .local = FALSE,
#' #'            pre_k,
#' #'            center = TRUE,
#' #'            scale = FALSE,
#' #'            compute_all = FALSE,
#' #'            return_projection = FALSE,
#' #'            allow_parallel = TRUE, ...)
#' #' @param Xr a matrix containing \code{n} reference observations rows and
#' #' \code{p} variablescolumns.
#' #' @param Xu an optional matrix containing data of a second set of observations
#' #' with \code{p} variables/columns.
#' #' @param Yr a matrix of \code{n} rows and one or more columns (variables) with
#' #' side information corresponding to the observations in \code{Xr} (e.g. response
#' #' variables). It can be numeric with multiple variables/columns, or character
#' #' with one single column. This argument is
#' #' required if:
#' #' \itemize{
#' #' \item{\code{diss_method == 'pls'}: \code{Yr} is required to project the variables
#' #' to orthogonal directions such that the covariance between the extracted pls
#' #' components and \code{Yr} is maximized.}
#' #' \item{\code{pc_selection$method == 'opc'}: \code{Yr}  is required to optimize
#' #' the number of components. The optimal number of projected components is the one
#' #' for which its distance matrix minimizes the differences between the \code{Yr}
#' #' value of each observation and the \code{Yr} value of its closest observation.
#' #' See \code{\link{diss_evaluate}}.}
#' #' }
#' #' @param pc_selection a list of length 2 which specifies the method to be used
#' #' for optimizing the number of components (principal components or pls factors)
#' #' to be retained. This list must contain two elements (in the following order):
#' #' \code{method} (a character indicating the method for selecting the number of
#' #' components) and \code{value} (a numerical value that complements the selected
#' #' method). The methods available are:
#' #' \itemize{
#' #'        \item{\code{"opc"}: optimized principal component selection based on
#' #'        Ramirez-Lopez et al. (2013a, 2013b). The optimal number of components
#' #'        (of a given set of observations) is the one for which its distance
#' #'        matrix minimizes the differences between the \code{Yr} value of each
#' #'        observation and the \code{Yr} value of its closest observation. In this
#' #'        case, \code{value} must be a value (larger than 0 and
#' #'        below \code{min(nrow(Xr)} \code{+ nrow(Xu),} \code{ncol(Xr))} indicating the maximum
#' #'        number of principal components to be tested. See the
#' #'        \code{\link{ortho_projection}} function for more details.}
#' 
#' #'        \item{\code{"cumvar"}: selection of the principal components based
#' #'        on a given cumulative amount of explained variance. In this case,
#' #'        \code{value} must be a value (larger than 0 and below or equal to 1)
#' #'        indicating the minimum amount of cumulative variance that the
#' #'        combination of retained components should explain.}
#' 
#' #'        \item{\code{"var"}: selection of the principal components based
#' #'        on a given amount of explained variance. In this case,
#' #'        \code{value} must be a value (larger than 0 and below or equal to 1)
#' #'        indicating the minimum amount of variance that a single component
#' #'        should explain in order to be retained.}
#' 
#' #'        \item{\code{"manual"}: for manually specifying a fix number of
#' #'        principal components. In this case, \code{value} must be a value
#' #'        (larger than 0 and
#' #'        below the minimum dimension of \code{Xr} or \code{Xr} and \code{Xu}
#' #'        combined).
#' #'        indicating the minimum amount of variance that a component should
#' #'        explain in order to be retained.}
#' #'        }
#' #' Default is \code{list(method = "var", value = 0.01)}.
#' #'
#' #' Optionally, the \code{pc_selection} argument admits \code{"opc"} or
#' #' \code{"cumvar"} or \code{"var"} or \code{"manual"} as a single character
#' #' string. In such case, the default \code{"value"} when either \code{"opc"} or
#' #' \code{"manual"} are used is 40. When \code{"cumvar"} is used the default
#' #' \code{"value"} is set to 0.99 and when \code{"var"} is used, the default
#' #' \code{"value"} is set to 0.01.
#' #' @param diss_method a character value indicating the type of projection on which
#' #' the dissimilarities must be computed. This argument is equivalent to
#' #' \code{method} argument in the \code{\link{ortho_projection}} function.
#' #' Options are:
#' #' \itemize{
#' #' \item{\code{"pca"}: principal component analysis using the singular value
#' #' decomposition algorithm)}
#' #' \item{\code{"pca.nipals"}: principal component analysis using
#' #' the non-linear iterative partial least squares algorithm.}
#' #' \item{\code{"pls"}: partial least squares.}
#' #' \item{\code{"mpls"}: modified partial least squares (Shenk and Westerhaus,
#' #' 1991 and Westerhaus, 2014).}
#' #' }
#' #' See the \code{\link{ortho_projection}} function for further details on the
#' #' projection methods.
#' #' @param .local a logical indicating whether or not to compute the dissimilarities
#' #' locally (i.e. projecting locally the data) by using the \code{pre_k} nearest
#' #' neighbor observations of each target observation. Default is \code{FALSE}. See details.
#' #' @param pre_k if \code{.local = TRUE} a numeric integer value which indicates the
#' #' number of nearest neighbors to (pre-)retain for each observation to
#' #' compute the (local) orthogonal dissimilarities to each observation in its
#' #' neighborhhod.
#' #' @param center a logical indicating if the \code{Xr} and \code{Xu} must be
#' #' centered. If \code{Xu} is provided the data is centered around the mean of
#' #' the pooled \code{Xr} and \code{Xu} matrices (\mjeqn{Xr \cup Xu}{Xr U Xu}). For
#' #' dissimilarity computations based on pls, the data is always centered for
#' #' the projections.
#' #' @param scale a logical indicating if the \code{Xr} and \code{Xu} must be
#' #' scaled. If \code{Xu} is provided the data is scaled based on the standard
#' #' deviation of the the pooled \code{Xr} and \code{Xu} matrices (\mjeqn{Xr \cup Xu}{Xr U Xu}).
#' #' if \code{center = TRUE}, scaling is applied after centering.
#' #' @param compute_all a logical. In case \code{Xu} is specified it indicates
#' #' whether or not the distances between all the elements resulting from the
#' #' pooled \code{Xr} and \code{Xu} matrices (\mjeqn{Xr \cup Xu}{Xr U Xu} must be computed).
#' #' @param return_projection a logical. If \code{TRUE} the `ortho_projection` object
#' #' on which the dissimilarities are computed will be returned. Default is \code{FALSE}. Note that
#' #' for \code{.local = TRUE} only the initial projection is returned (i.e. local
#' #' projections are not).
#' #' @param allow_parallel a logical (default TRUE). It allows parallel computing
#' #' of the local distance matrices (i.e. when \code{.local = TRUE}). This is done
#' #' via \code{\link[foreach]{foreach}} function of the 'foreach' package.
#' #' @param ... additional arguments to be passed to the
#' #' \code{\link{ortho_projection}} function.
#' #' @details
#' #' When \code{.local = TRUE}, first a global dissimilarity matrix is computed based on
#' #' the parameters specified. Then, by using this matrix for each target
#' #' observation, a given set of nearest neighbors (\code{pre_k}) are identified.
#' #' These neighbors (together with the target observation) are projected
#' #' (from the original data space) onto a (local) orthogonal space (using the
#' #' same parameters specified in the function). In this projected space the
#' #' Mahalanobis distance between the target observation and its neighbors is
#' #' recomputed. A missing value is assigned to the observations that do not belong to
#' #' this set of neighbors (non-neighbor observations).
#' #' In this case the dissimilarity matrix cannot be considered as a distance
#' #' metric since it does not necessarily satisfies the symmetry condition for
#' #' distance matrices (i.e. given two observations \mjeqn{x_i}{x_i} and \mjeqn{x_j}{x_j}, the local
#' #' dissimilarity (\mjeqn{d}{d}) between them is relative since generally
#' #' \mjeqn{d(x_i, x_j) \neq d(x_j, x_i)}{d(x_i, x_j) ne d(x_j, x_i)}). On the other hand, when
#' #' \code{.local = FALSE}, the dissimilarity matrix obtained can be considered as
#' #' a distance matrix.
#' #'
#' #' In the cases where \code{"Yr"} is required to compute the dissimilarities and
#' #' if \code{.local = TRUE}, care must be taken as some neighborhoods might
#' #' not have enough observations with non-missing \code{"Yr"} values, which might retrieve
#' #' unreliable dissimilarity computations.
#' #'
#' #' If \code{"opc"} or \code{"manual"} are used in \code{pc_selection$method}
#' #' and \code{.local = TRUE}, the minimum number of observations with non-missing
#' #' \code{"Yr"} values at each neighborhood is determined by
#' #' \code{pc_selection$value} (i.e. the maximum number of components to compute).
#' #'
#' #'
#' #' @return a \code{list} of class \code{ortho_diss} with the following elements:
#' #' \itemize{
#' #'  \item{\code{ncomp}: the number of components (either principal
#' #'  components or partial least squares components) used for computing the
#' #'  global dissimilarities.}
#' #'  \item{\code{global_variance_info}: the information about the expalined
#' #'  variance(s) of the projection. When \code{.local = TRUE}, the information
#' #'  corresponds to the global projection done prior computing the local
#' #'  projections.}
#' #'  \item{\code{local_ncomp}: if \code{.local = TRUE}, a data.frame
#' #'  which specifies the number of local components (either principal components
#' #'  or partial least squares components) used for computing the dissimilarity
#' #'  between each target observation and its neighbor observations.}
#' #'  \item{\code{dissimilarity}: the computed dissimilarity matrix. If
#' #'  \code{.local = FALSE} a distance matrix. If \code{.local = TRUE} a matrix of
#' #'  class \code{local_ortho_diss}. In this case, each column represent the dissimilarity
#' #'  between a target observation and its neighbor observations.}
#' #'  \item{\code{projection}: if \code{return_projection = TRUE},
#' #'  an \code{ortho_projection} object.}
#' #'  }
#' #' @author \href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
#' #' @references
#' #' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
#' #' Scholten, T. 2013a. The spectrum-based learner: A new local approach for
#' #' modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196, 268-279.
#' #'
#' #' Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
#' #' J. A. M.,  Scholten, T. 2013b. Distance and similarity-search metrics for use
#' #' with soil vis-NIR spectra. Geoderma 199, 43-53.
#' #' @seealso \code{\link{ortho_projection}}, \code{\link{diss_evaluate}}
#' #' @examples
#' #' library(prospectr)
#' #' data(NIRsoil)
#' #'
#' #' Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
#' #' Yu <- NIRsoil[!as.logical(NIRsoil$train), "CEC", drop = FALSE]
#' #' Yr <- NIRsoil[as.logical(NIRsoil$train), "CEC", drop = FALSE]
#' #' Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]
#' #'
#' #' Xu <- Xu[!is.na(Yu), ]
#' #' Yu <- Yu[!is.na(Yu), , drop = FALSE]
#' #'
#' #' Xr <- Xr[!is.na(Yr), ]
#' #' Yr <- Yr[!is.na(Yr), , drop = FALSE]
#' #'
#' #' # Computation of the orthogonal dissimilarity matrix using the
#' #' # default parameters
#' #' pca_diss <- ortho_diss(Xr, Xu)
#' #'
#' #' # Computation of a principal component dissimilarity matrix using
#' #' # the "opc" method for the selection of the principal components
#' #' pca_diss_optim <- ortho_diss(
#' #'   Xr, Xu, Yr,
#' #'   pc_selection = list("opc", 40),
#' #'   compute_all = TRUE
#' #' )
#' #'
#' #' # Computation of a partial least squares (PLS) dissimilarity
#' #' # matrix using the "opc" method for the selection of the PLS
#' #' # components
#' #' pls_diss_optim <- ortho_diss(
#' #'   Xr = Xr, Xu = Xu,
#' #'   Yr = Yr,
#' #'   pc_selection = list("opc", 40),
#' #'   diss_method = "pls"
#' #' )
#' #' @keywords internal
#' #' @noRd
#' 
#' #######################################################################
#' # resemble
#' # Copyright (C) 2014 Leonardo Ramirez-Lopez and Antoine Stevens
#' #
#' # This program is free software; you can redistribute it and/or modify
#' # it under the terms of the GNU General Public License as published by
#' # the Free Software Foundation; either version 2 of the License, or
#' # any later version.
#' #
#' # This program is distributed in the hope that it will be useful,
#' # but WITHOUT ANY WARRANTY; without even the implied warranty of
#' # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#' # GNU General Public License for more details.
#' #######################################################################
#' 
#' ortho_diss <- function(Xr, Xu = NULL,
#'                        Yr = NULL,
#'                        pc_selection = list(method = "var", value = 0.01),
#'                        diss_method = "pca",
#'                        .local = FALSE,
#'                        pre_k,
#'                        center = TRUE, scale = FALSE,
#'                        compute_all = FALSE,
#'                        return_projection = FALSE,
#'                        allow_parallel = TRUE,
#'                        ...) {
#'   if (!is.logical(.local)) {
#'     stop("'.local'must be logical")
#'   }
#' 
#'   if (!is.logical(compute_all)) {
#'     stop("'compute_all' must be logical")
#'   }
#' 
#'   if (!is.logical(center)) {
#'     stop("'center' must be logical")
#'   }
#' 
#'   if (!is.logical(scale)) {
#'     stop("'scale' must be logical")
#'   }
#' 
#'   if (.local) {
#'     if (pre_k < 4) {
#'       stop("pre_k cannot be smaller than 4")
#'     }
#'     if (pre_k > nrow(Xr)) {
#'       stop("pre_k cannot be larger than the numer of elements in Xr")
#'     }
#'   }
#' 
#'   if (!is.null(Yr)) {
#'     Yr <- as.matrix(Yr)
#'   } else {
#'     if (pc_selection[[1]] == "opc" | diss_method %in% c("pls", "mpls")) {
#'       stop("Yr must be provided when the 'opc' is used in pc_selection is used or diss_method is one of 'pls' or 'mpls'")
#'     }
#'   }
#' 
#'   projection <- ortho_projection(
#'     Xr = Xr,
#'     Yr = Yr,
#'     Xu = Xu,
#'     method = diss_method,
#'     pc_selection = pc_selection,
#'     center = center,
#'     scale = scale, ...
#'   )
#' 
#'   scores <- projection$scores
#'   if (diss_method %in% c("pca", "pca.nipals")) {
#'     scores <- sweep(projection$scores,
#'       MARGIN = 2,
#'       STATS = projection$scores_sd,
#'       FUN = "/"
#'     )
#'     use_distance_method <- "euclid"
#'   } else {
#'     use_distance_method <- "mahalanobis"
#'   }
#' 
#'   n_components <- projection$ncomp
#' 
#'   if (is.null(Xu) | compute_all) {
#'     distnc <- f_diss(
#'       Xr = scores,
#'       Xu = NULL,
#'       diss_method = use_distance_method,
#'       center = FALSE, scale = FALSE
#'     )
#'     dimnames(distnc) <- list(
#'       rownames(scores),
#'       rownames(scores)
#'     )
#'   }
#' 
#'   if (!compute_all & !is.null(Xu)) {
#'     distnc <- f_diss(
#'       Xr = scores[1:nrow(Xr), , drop = FALSE],
#'       Xu = scores[(1 + nrow(Xr)):nrow(scores), , drop = FALSE],
#'       diss_method = use_distance_method,
#'       center = FALSE, scale = FALSE
#'     )
#'     dimnames(distnc) <- list(
#'       rownames(scores[1:nrow(Xr), , drop = FALSE]),
#'       rownames(scores[(1 + nrow(Xr)):nrow(scores), , drop = FALSE])
#'     )
#'   }
#' 
#'   if (.local) {
#'     n_xr <- nrow(Xr)
#'     if (!is.null(Xu) & compute_all) {
#'       pre_k <- pre_k + 1
#'       Xr <- rbind(Xr, Xu)
#'       Yr <- rbind(Yr, matrix(rep(NA, nrow(Xu))))
#'     }
#' 
#'     if (is.null(Xu)) {
#'       pre_k <- pre_k + 1
#'     }
#' 
#'     neighborhood_info <- k0_indices <- apply(distnc,
#'       MARGIN = 2,
#'       FUN = order
#'     )[1:pre_k, ]
#'     d_dimnames <- dimnames(distnc)
#'     rm(distnc)
#' 
#'     neighborhood_info[k0_indices <= n_xr] <- paste0(
#'       "Xr_",
#'       neighborhood_info[k0_indices <= n_xr]
#'     )
#'     neighborhood_info[k0_indices > n_xr] <- paste0(
#'       "Xu_",
#'       neighborhood_info[k0_indices > n_xr]
#'     )
#'     neighborhood_info <- data.frame(
#'       do.call(
#'         "rbind",
#'         strsplit(colnames(k0_indices), "_")
#'       ),
#'       t(neighborhood_info)
#'     )
#'     colnames(neighborhood_info) <- c(
#'       "Set", "Index",
#'       paste0("k_", 1:nrow(k0_indices))
#'     )
#' 
#'     if ("opc" %in% pc_selection | diss_method == "pls") {
#'       # k0_index_matrix <- (apply(distnc[, (1 + nrow(Xr)):ncol(distnc)], 2, order))[1:pre_k, ]
#'       have_response <- colSums(!apply(k0_indices,
#'         MARGIN = 2,
#'         FUN = function(i, y) is.na(y[i, ]),
#'         y = Yr
#'       ))
#' 
#'       if (any(c("opc", "manual") %in% pc_selection)) {
#'         insufficient <- which(have_response < pc_selection[[2]])
#'         if (length(insufficient) > 0) {
#'           inms <- names(insufficient)
#' 
#'           mss <- format_xr_xu_indices(xr_xu_names = inms)
#' 
#'           omessage <- paste0(
#'             "Local disssimilarity matrix cannot be computed as ",
#'             "the following observation \n",
#'             "  indices have insufficient neighbors with Yr values:",
#'             mss$xr_mss,
#'             mss$Xu_mss
#'           )
#'           stop(omessage)
#'         }
#'       }
#' 
#'       if (any(have_response < 4)) {
#'         inms <- names(which(have_response < 3))
#' 
#'         mss <- format_xr_xu_indices(xr_xu_names = inms)
#' 
#'         omessage <- paste0(
#'           "Local disssimilarity matrix cannot be computed as ",
#'           "the following observation \n",
#'           "  indices have less than 3 neighbors with Yr values:",
#'           mss$xr_mss,
#'           mss$Xu_mss
#'         )
#' 
#'         stop(omessage)
#'       }
#'       cnt <- 100 * have_response / nrow(k0_indices)
#' 
#'       below_f1 <- sum(cnt != 100)
#'       if (below_f1 > 0) {
#'         omessage <- paste0(
#'           "The neighborhoods of ", below_f1,
#'           " observations contain missing 'Yr' values."
#'         )
#' 
#'         below_f2 <- sum(cnt < 50)
#'         if (below_f2 > 0) {
#'           omessage2 <- paste0(
#'             "\nFor ", below_f2, " of these observations ",
#'             "Yr is missing for more than 50% of their neighbors."
#'           )
#'           omessage <- paste0(omessage, omessage2)
#'         }
#'         omessage <- paste0(omessage, "\nCheck ...$neighborhood_info")
#'         message(omessage)
#'       }
#'       neighborhood_info <- cbind(
#'         missing_Yr = nrow(k0_indices) - have_response,
#'         neighborhood_info
#'       )
#'     }
#' 
#'     local_d <- local_ortho_diss(
#'       k_index_matrix = k0_indices,
#'       Xr = Xr,
#'       Yr = Yr,
#'       Xu = Xu,
#'       diss_method = diss_method,
#'       pc_selection = pc_selection,
#'       center = center,
#'       scale = scale,
#'       allow_parallel = allow_parallel, ...
#'     )
#' 
#'     dimnames(local_d$dissimilarity_m) <- d_dimnames
#' 
#'     neighborhood_info <- cbind(neighborhood_info[, 2:1],
#'       local_ncomp = local_d$local_ncomp,
#'       neighborhood_info[, -c(1:2)]
#'     )
#' 
#'     resultsList <- list(
#'       ncomp = n_components,
#'       global_variance_info = projection$variance,
#'       neighborhood_info = neighborhood_info,
#'       dissimilarity = local_d$dissimilarity_m
#'     )
#' 
#'     if (return_projection) {
#'       resultsList$projection <- projection
#'     }
#' 
#'     class(resultsList) <- c("ortho_diss", "list")
#'     class(resultsList$dissimilarity) <- c("local_ortho_diss", "matrix")
#'     return(resultsList)
#'   } else {
#'     resultsList <- list(
#'       ncomp = n_components,
#'       global_variance_info = projection$variance,
#'       dissimilarity = distnc
#'     )
#'     if (return_projection) {
#'       resultsList$projection <- projection
#'     }
#' 
#'     class(resultsList) <- c("ortho_diss", "list")
#'     class(resultsList$dissimilarity) <- c("ortho_diss", "matrix")
#'     return(resultsList)
#'   }
#' }

---
title: "Nearest-neighbor search"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    toc-location: left
    number-sections: true
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{4 Nearest-neighbor search}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---


```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
```



:::: {.columns}
::: {.column width="70%"}
> *We are caught in an inescapable network of mutuality, tied in a single garment of destiny* -- Martin Luther King Jr
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::


# Searching for neighbors

In the package, $k$-nearest neighbor search is used to identify, within a reference set of observations, a group of spectrally similar observations for another set of observations. For a given observation, the most similar observations are referred to as its nearest neighbors and are typically identified using dissimilarity measures.

In `resemble`, k-nearest neighbor search is implemented in the `search_neighbors()` function. This function relies on `dissimilarity()` to compute the dissimilarity matrix used to identify the neighbors. If `Xu` is not provided, the search is performed within `Xr` itself, while automatically excluding each observation from its own neighborhood.

Neighbors can be selected in two ways, each specified by a constructor function:

* `neighbors_k()`: by defining a fixed number of neighbors ($k$)
* `neighbors_diss()`: by defining a dissimilarity threshold ($d_{\text{th}}$)

Readers are encouraged to review the section on dissimilarity measures first, as it provides the basis for the examples presented in this section.


# Neighbor selection methods

## Fixed-$k$ selection

`neighbors_k()` selects a fixed number of nearest neighbors for each target observation:

```{r neighbors-k-example, eval = FALSE}
neighbors_k(k = 50)
neighbors_k(k = c(40, 60, 80, 100))
```

Multiple values of $k$ can be provided to evaluate different neighborhood sizes. The minimum allowed value is 4, ensuring sufficient observations for local model fitting.

## Dissimilarity-threshold selection

`neighbors_diss()` selects neighbors whose dissimilarity falls below a threshold $d_\text{th}$:

```{r neighbors-diss-example, eval = FALSE}
neighbors_diss(threshold = 0.3)
neighbors_diss(threshold = c(0.1, 0.2, 0.3), k_min = 10, k_max = 150)
```

The `k_min` and `k_max` arguments constrain the neighborhood size. If fewer than `k_min` neighbors fall below the threshold, the closest `k_min` observations are retained regardless. If more than `k_max` qualify, only the closest `k_max` are kept.



# Using a specific number of neighbors

This means that the neighboring observations are retained regardless their dissimilarity/distance to the target observation. Each target observation for which its neighbors are to be found ends up with the same neighborhood size ($k$). A drawback of this approach is that observations that are in fact largely dissimilar to the target observation might end up in its neighborhood. This is because the requirement for building the neighborhood is based on its size and not on the similarity of the retained observations to the target one.

Here is an example that demonstrates how `search_neighbors()` can be used to search in the `train_x` set the spectral neighbors of the `test_x` set:

```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

First, take three observations from the test set at random and search for their 
nearest neighbors in the training set using PCA-based dissimilarity with only 2 
components and a fixed neighborhood size of 30:

```{r}
set.seed(8011)
rnd_idc <- sample(nrow(test_x), 3)

k_fixed <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2, return_projection = TRUE),
  neighbors = neighbors_k(30)
)
```

The, for the same 3 observations in the test set, search for their nearest neighbors in the training set using the same previous dissimilarity but use a dissimilarity threshold of 0.25 to reatin the neighbors:

```{r}
k_diss <- search_neighbors(
  Xr = train_x,
  Xu = test_x[rnd_idc, ],
  diss_method = diss_pca(ncomp = 2),
  neighbors = neighbors_diss(threshold = 0.25)
)
```

Identify the indices of the test observations in the projection space of the 
PCA-based dissimilarity (their row names always start with "Xu_"):

```{r}
test_scores_indices <- grep("^Xu_", rownames(k_fixed$projection$scores))
```

Plot the projection scores of the training set and highlight the neighbors found with the two methods (fixed $k$ and dissimilarity threshold) as well as the test observations:

```{r}
#| label: fig-neighbors
#| fig-cap: "Nearest neighbors identified using the same dissimilarity metric but retained using two different selection methods."
#| fig-width: 8.5
#| fig-height: 4.5

old_par <- par(no.readonly = TRUE)
on.exit(par(old_par), add = TRUE)

par(mfrow = c(1, 2), mar = c(4, 4, 3, 1))

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Fixed k-nearest neighbors"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_fixed$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)

plot(
  k_fixed$projection$scores,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.3),
  main = "Neighbors selected by a \nthreshold dissimilarity"
)
grid(lty = 1)
points(
  k_fixed$projection$scores[k_diss$unique_neighbors, ],
  pch = 16,
  col = "dodgerblue"
)
points(
  k_fixed$projection$scores[test_scores_indices, ],
  pch = 16,
  cex = 1.5,
  col = "red"
)
```


@fig-neighbors illustrates the difference between the two neighbor-selection strategies. In the fixed-(k) approach, each test sample is assigned the same number of neighbors, but the spatial extent of the resulting neighborhoods varies. For example, one neighborhood is very compact because many similar observations are located close to the test sample, whereas the other two are more dispersed because a larger region must be searched to retrieve the required number of neighbors. In contrast, the dissimilarity-threshold approach produces neighborhoods with a more comparable spatial extent, since neighbors are retained according to a common dissimilarity cutoff rather than a fixed count. As a result, the number of selected neighbors varies across test samples. In @fig-neighbors, the neighborhood of the left test sample is denser and contains more neighbors, whereas the neighborhood of the right test sample is sparser and contains fewer neighbors.

Exploring what the `search_neighbors()` function returns is useful to understand how the neighbors are stored and how to access them. For example, the following code shows how to access the neighbors found with the fixed-$k$ approach:

```{r knn-pca, eval = FALSE}
# matrix of neighbors
k_fixed$neighbors

# matrix of neighbor distances (dissimilarity scores)
k_fixed$neighbors_diss

# the index (in the training set) of the first two closest neighbors found in 
# training for the first observation in testing:
k_fixed$neighbors[1:2, 1, drop = FALSE]

# the distances of the two closest neighbors found in 
# training for the first observation in testing:
k_fixed$neighbors_diss[1:2, 1, drop = FALSE]

# the indices in training that fall in any of the 
# neighborhoods of testing
k_fixed$unique_neighbors
```

In the above code, `k_fixed$neighbors` is a matrix showing the results of the neighbors found. This is a matrix of neighbor indices where every column represents an observation in the testing set while every row represents the neighbor index (in descending order). Every entry represents the index of the neighbor observation in the training set. The `k_fixed$neighbors_diss` matrix shows the dissimilarity scores corresponding to the neighbors found. For example, for the first observation in `test_x` its closest observation found in `train_x` corresponds to the one with index `k_fixed$neighbors[1, 1]` which has a dissimilarity score of `k_fixed$neighbors_diss[1, 1]`.

Note that in the method where neighbors are selected by a threshold dissimilarity, the matrix `k_diss$neighbors` will contain many `NA` values because the number of neighbors retained for each observation in testing is not constant. The number of neighbors retained for each observation in testing can be easily visualized with a histogram on `k_diss$k_diss_info$final_n_k`.

Neighbor search can also be conducted with all the dissimilarity measures described in previous sections. The neighbors retrieved will then depend on the dissimilarity method used. Thus, it is recommended to evaluate carefully what dissimilarity metric to use before neighbor search.

Here are other examples of neighbor search based on other dissimilarity measures:

```{r}
#| eval: TRUE
#| label: knn-other-methods
# using PC dissimilarity with optimal selection of components
knn_opc <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(),
    scale = TRUE,
    return_projection = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using PLS dissimilarity with optimal selection of components
knn_pls <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pls(
    ncomp = ncomp_by_opc(),
    scale = TRUE
  ),
  Yr = train_y,
  neighbors = neighbors_k(50)
)

# using correlation dissimilarity
knn_cor <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(),
  neighbors = neighbors_k(50)
)

# using moving window correlation dissimilarity
knn_mw <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_correlation(ws = 51),
  neighbors = neighbors_k(50)
)
```

# Using a dissimilarity threshold

Here, the neighboring observations to be retained must have a dissimilarity score less or equal to a given dissimilarity threshold ($d_\text{th}$). Therefore, the neighborhood size of the target observations is not constant. A drawback with this approach is that choosing a meaningful $d_\text{th}$ can be difficult, especially because its value is largely influenced by the dissimilarity method used. Thresholds are not comparable across methods: a threshold of 0.3 for correlation dissimilarity has no correspondence to 0.3 for Euclidean or PCA-based dissimilarity. Thresholds must be calibrated empirically for each method.

Furthermore, without bounds, some neighborhoods retrieved by certain thresholds might be of a very small size or even empty, which constrains any type of analysis within such neighborhoods. On the other hand, some neighborhoods might end up with large sizes which might include either redundant observations or in some other cases where $d_\text{th}$ is too large the complexity in the neighborhood might be large.

In `neighbors_diss()`, $d_\text{th}$ is controlled by the argument `threshold`. This argument is accompanied by the arguments `k_min` and `k_max` which are used to control the minimum and maximum neighborhood sizes. For example, if a neighborhood size is below the minimum size `k_min`, the function automatically ignores $d_\text{th}$ and retrieves the `k_min` closest observations. Similarly, if the neighborhood size is above the maximum size `k_max`, the function automatically ignores $d_\text{th}$ and retrieves only a maximum of `k_max` neighbors.

In the package, we can use `search_neighbors()` to find in the `train_x` set the neighbors of the `test_x` set which dissimilarity scores are less or equal to a user-defined threshold:

```{r}
#| eval: TRUE
#| label: knn-threshold-example
#| results: hide
# a dissimilarity threshold
d_th <- 1

# the minimum number of observations required in each neighborhood
k_min <- 20

# the maximum number of observations allowed in each neighborhood
k_max <- 300

dnn_pca <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(scale = TRUE),
  neighbors = neighbors_diss(threshold = d_th, k_min = k_min, k_max = k_max)
)

# matrix of neighbors. The minimum number of indices is 20 (given by k_min)
# and the maximum number of indices is 300 (given by k_max).
# NAs indicate "not a neighbor"
dnn_pca$neighbors

# this reports how many neighbors were found for each observation in 
# testing using the input distance threshold (column n_k) and how 
# many were finally selected (column final_n_k)
dnn_pca$k_diss_info

# matrix of neighbor distances
dnn_pca$neighbors_diss

# the indices in training that fall in any of the 
# neighborhoods of testing
dnn_pca$unique_neighbors
```

In the code above, the size of the neighborhoods is not constant, the size variability can be easily visualized with a histogram on `dnn_pca$k_diss_info$final_n_k` (@fig-knn-hist). The histogram shows that many neighborhoods were reset to a size of `k_min` or to a size of `k_max`.

```{r}
#| eval: TRUE
#| label: fig-knn-hist
#| fig-cap: "Histogram of the final neighborhood sizes after applying the dissimilarity threshold and the minimum and maximum neighborhood size constraints."
#| fig-width: 5
#| fig-height: 4
hist(
  dnn_pca$k_diss_info$final_n_k,
  breaks = 20,
  xlab = "Final neighborhood size",
  main = "",
  col = "dodgerblue"
)
```

# Spiking the neighborhoods

In the package, _spiking_ refers to forcing specific observations to be included in the neighborhoods. For example, if we are searching in the `train_x` set the neighbors of the `test_x` set, and if we want to force certain observations in `train_x` to be included in the neighborhood of each observation in `test_x`, we can use the `spike` argument in `search_neighbors()`. For that, in this argument we will need to pass the indices of `train_x` that will be forced into the neighborhoods. Positive indices are forced into all neighborhoods; negative indices are excluded from all neighborhoods.

The following example demonstrates how to do that:

```{r}
#| eval: TRUE
#| label: knn-spike-example
# the indices of the observations that we want to "invite" to every neighborhood
forced_guests <- c(1, 5, 8, 9)

# using PC dissimilarity with optimal selection of components
knn_spiked <- search_neighbors(
  Xr = train_x,
  Xu = test_x,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(20)
  ),
  Yr = train_y,
  neighbors = neighbors_k(50),
  spike = forced_guests
)

# check the first 8 neighbors found in training for the 
# first 2 observations in testing
knn_spiked$neighbors[1:8, 1:2]
```

The previous code shows that the indices specified in `forced_guests` are always selected as part of every neighborhood.

Spiking may be useful when prior knowledge indicates that certain observations should always be considered (for example, calibration standards or known reference samples) regardless of their spectral dissimilarity.

---
title: "Simple global models"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
knitr:
  opts_chunk:
    class-output: "ansi"
vignette: >
  %\VignetteIndexEntry{5 Simple global models}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *It is vain to do with more what can be done with less* -- William of Ockham
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::


```{r}
#| message: false
#| echo: false
library(resemble)
```

# Introduction

The `model()` function provides a simple interface for building global calibration models using partial least squares (PLS) or Gaussian process regression (GPR). Unlike `mbl()`, which builds local models for each prediction, `model()` fits a single model to the entire reference set.

# Fitting methods

Regression methods are specified via constructor functions. The available methods are:

| Constructor | Method | Description | Reference |
|-------------|--------|-------------|-----------|
| `fit_pls(method = "pls")` | PLS | Non-linear iterative partial least squares (NIPALS) | @wold1975soft |
| `fit_pls(method = "simpls")` | SIMPLS | Straightforward implementation of PLS | @de1993simpls |
| `fit_pls(method = "mpls")` | Modified PLS | Uses correlation instead of covariance for weights | @shenk1991populations |
| `fit_wapls(method = "mpls")` | Weighted average PLS | Predictions are a weighted average of multiple models from multiple PLS factors. __NOT YET AVAIABLE FOR THE__ `model()` __FUNCTION__ | @shenk1997investigation |
| `fit_gpr()` | GPR | Gaussian process regression with dot product kernel | @rasmussen2006gaussian |

```{r}
#| label: fit-methods
# PLS with 15 components
fit_pls(ncomp = 15)

# SIMPLS with scaling
fit_pls(ncomp = 15, method = "simpls", scale = TRUE)

# mPLS with scaling
fit_pls(ncomp = 15, method = "mpls")

# GPR with default noise variance
fit_gpr()
```

# Example

## Data preparation

```{r}
#| label: data-prep
#| message: false
library(prospectr)

wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split data
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

# Remove missing values
ok_train <- !is.na(train_y)
ok_test  <- !is.na(test_y)
train_x <- train_x[ok_train, ]
train_y <- train_y[ok_train]
test_x  <- test_x[ok_test, ]
test_y  <- test_y[ok_test]
```

## Fitting a PLS model

```{r}
#| label: fit-pls

set.seed(1124) # guarantee same CV splits for all methods
pls_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_pls(ncomp = 12, method = "mpls", scale = TRUE),
 control = model_control(validation_type = "lgo", number = 10)
)

pls_mod
```

## Fitting a GPR model

```{r}
#| label: fit-gpr
set.seed(1124) # guarantee same CV splits for all methods
gpr_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_gpr(noise_variance = 0.5),
 control = model_control(validation_type = "lgo", number = 10)
)

gpr_mod
```

## Prediction

```{r}
#| label: predict
# Predict using optimal number of components (from CV)
pls_pred <- predict(pls_mod, newdata = test_x)

# Predict with GPR
gpr_pred <- predict(gpr_mod, newdata = test_x)

# Compare predictions
data.frame(
  observed = test_y,
  pls = pls_pred[, which(pls_mod$cv_results$optimal)],
  gpr = as.vector(gpr_pred)
) |> head(10)
```

## Validation statistics

```{r}
#| label: validation
# Function to compute stats
eval_pred <- function(obs, pred) {
  data.frame(
    rmse = sqrt(mean((obs - pred)^2)),
    r2 = cor(obs, pred)^2
  )
}
```

```{r}
## PLS evaluation
eval_pred(test_y, pls_pred[, which(pls_mod$cv_results$optimal)])
```

```{r}
## GPR evaluation
eval_pred(test_y, gpr_pred)
```

# References {-}

---
title: "Essential concepts and setup"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{1 Essential concepts and setup}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *Think Globally, Fit Locally* -- [@saul2003think]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Spectroscopic data analysis plays a central role in many environmental, agricultural, and food-related applications. Techniques such as near-infrared (NIR), mid-infrared (IR), and other forms of diffuse reflectance spectroscopy provide rapid, non-destructive, and cost-efficient measurements that can be used to infer chemical, physical, or biological properties of complex matrices, including soils, plant materials, and food products. In quantitative applications, these measurements are typically linked to reference laboratory values through empirical calibration models.

As spectral databases grow in size and diversity, their effective use becomes increasingly challenging. Large spectral libraries often contain substantial heterogeneity, domain shifts, redundant observations, and samples that are only locally informative for a given prediction problem. Under these conditions, global modelling strategies are often insufficient on their own, and methods based on dimensionality reduction, dissimilarity analysis, neighbour retrieval, local modelling, and targeted sample selection become essential.

The `resemble` package provides a framework for sample retrieval and local learning in spectral chemometrics. It is designed to support the analysis of large and complex spectral datasets through tools for projection-based representation, dissimilarity computation, neighbourhood search, memory-based learning, evolutionary subset search, and retrieval-based modelling with pre-computed local models. The package therefore supports both classical local modelling workflows and newer strategies for exploiting spectral libraries as structured resources for predictive modelling. 

The functions presented here are implemented based on the methods described in @ramirezlopez2026a, @ramirezlopez2026b, and @ramirez2013spectrum.

The main functionalities of `resemble` include:

- orthogonal projection of spectral data using principal component analyssis (PCA) and partial least squares (PLS) methods

- computation and evaluation of spectral dissimilarity measures

- nearest-neighbour search in spectral reference sets

- memory-based learning and local regression

- evolutionary sample search for context-specific calibration

- retrieval-based modelling using libraries of localised experts


# Citing the package

Simply type and you will get the info you need:
```{r}
#| eval: true
citation(package = "resemble")
```

# Dataset used across the vignettes

The vignettes in `resemble` use the soil near-infrared (NIR) spectral dataset 
provided in the [`prospectr`](https://CRAN.R-project.org/package=prospectr) 
package [@stevens2020introduction]. This dataset is used because soils are among 
the most complex matrices analyzed by NIR spectroscopy. It was originally used 
in the *Chimiométrie 2006* challenge [@pierna2008soil].

The dataset contains NIR absorbance spectra for 825 dried and sieved soil 
samples collected from agricultural fields across the Walloon region of Belgium. 
In `R`, the data are stored in a `data.frame` with the following structure:

* **Response variables**:
  
  * **`Nt`**: total nitrogen (g/kg dry soil); available for 645 samples and 
    missing for 180.
  
  * **`Ciso`**: carbon (g/100 g dry soil); available for 732 samples and 
    missing for 93. 
  
  * **`CEC`**: cation exchange capacity (meq/100 g dry soil); available for 447 
    samples and missing for 378.

* **Predictor variables (`spc`)**: the spectral predictors are stored in the matrix 
  `NIRsoil$spc`, embedded within the data frame. These variables contain NIR 
  absorbance spectra measured from 1100 to 2498 nm at 2 nm intervals. Each 
  column name corresponds to a wavelength value (in nm).

* **Set indicator (`Set`)**: a binary variable indicating whether a sample belongs to 
  the training set (`1`, 618 samples) or the test set (`0`, 207 samples).
  
  
Load the necessary packages and data: 
```{r}
#| label: libraries
#| message: false
library(resemble)
library(prospectr)
```

The dataset can be loaded into R as follows: 
```{r}
#| message: false
#| results: hide
data(NIRsoil)
dim(NIRsoil)
str(NIRsoil)
```

# Spectral preprocessing

Throughout the vignettes, the same preprocessing workflow is used to improve the suitability of the spectra for quantitative analysis. In particular, the goal is to reduce unwanted baseline variation and enhance local spectral features that may be informative for modeling. The preprocessing steps are implemented using the [`prospectr`](https://CRAN.R-project.org/package=prospectr) package [@stevens2020introduction].

The following steps are applied:

1. **Detrending** is applied first to reduce broad baseline shifts and curvature effects across the spectra.

2. A **first-order Savitzky–Golay derivative** [@Savitzky1964] is then computed to emphasize local spectral features and reduce remaining additive effects.

```{r}
#| label: NIRsoil
#| message: false
# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
```

```{r}
#| label: fig-plotspectra
#| fig-cap: "Raw spectral absorbance data (top) and first derivative of the absorbance spectra (bottom)."
#| fig-align: center
#| fig-width: 7
#| fig-height: 7
#| echo: false
# Hex sticker palette#
bg_dark <- NA #"#0F172A"      # background
blue <- "#3B82F6"         # border / spiral high
amber <- "#F59E0B"        # spiral low
amber_light <- "#FBBF24"  # text accent
slate <- "#64748B"        # muted elements

old_par <- par("mfrow", "mar", "bg")
par(mfrow = c(2, 1), mar = c(4, 4, 1, 4), bg = bg_dark)

new_wavs <- as.matrix(as.numeric(colnames(NIRsoil$spc_p)))

text_col <- "black"#"white"

# Plot 1: Raw spectra
plot(range(wavs), range(NIRsoil$spc), col = NA,
     xlab = "",
     ylab = "Absorbance",
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = wavs, y = t(NIRsoil$spc), 
         lty = 1, col = paste0(blue, "33"))

# Plot 2: First derivative
plot(range(new_wavs), range(NIRsoil$spc_p), col = NA,
     xlab = "Wavelengths, nm",
     ylab = expression(d(detrended~A)/d*lambda),
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = new_wavs, y = t(NIRsoil$spc_p), 
         lty = 1, col = paste0(amber, "33"))

par(old_par)
```

Both the raw absorbance spectra and the preprocessed spectra are shown in 
@fig-plotspectra. The preprocessed spectra, obtained as the first derivative of 
detrended absorbance, are used as the predictor variables in all examples 
throughout this document.

For illustration purposes, the `NIRsoil` data are divided into training and test 
subsets. In the examples that require a response variable, `Ciso` is used to 
demonstrate the functionality of the package.

```{r}
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

The notation used throughout the `resemble` package for arguments referring to 
training and test observations is as follows:

* **Training observations**:

  * `Xr` denotes the matrix of predictor variables in the reference/training set.
  
  * `Yr` denotes the response variable(s) in the reference/training set. In the
    context of this package, `Yr` may also be referred to as **side
    information**, that is, variables associated with the training observations
    that can support or guide optimization during modeling, even when they are
    not directly used as model inputs. For example, as shown in later sections,
    `Yr` can be used in principal component analysis to help determine the
    optimal number of components.

* **Test observations**:

  * `Xu` denotes the matrix of predictor variables in the unknown/test set.
  
  * `Yu` denotes the response variable(s) in the unknown/test set.


# References {-} 

---
title: "Estimating dissimilarity between spectra"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    toc-location: left
    number-sections: true
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{3 Estimating dissimilarity between spectra}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
```


:::: {.columns}
::: {.column width="70%"}
> *Similarity has two faces: causal and derivative* -- [@tversky1977features]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Dissimilarity measures quantify how different observations are from one another. 
In spectroscopy, dissimilarity matrices are commonly used for exploratory 
analysis, outlier detection, nearest-neighbor search, sample selection, and 
local calibration or local learning methods. In the `resemble` package, 
`dissimilarity()` is the main interface for computing dissimilarity matrices.

In `resemble`, the term *dissimilarity* is preferred over *distance* to 
emphasize that the implemented measures are not necessarily proper metric 
distances. For reference, a metric distance is typically expected to satisfy 
three properties: minimality (or identity), meaning that the distance between an 
observation and itself is zero and no greater than the distance between distinct 
observations; symmetry, meaning that the distance from $a$ to $b$ is the 
same as the distance from $b$ to $a$; and the triangle inequality, meaning 
that the distance between two observations cannot exceed the sum of their 
distances through a third observation. Some of the measures implemented in 
`resemble` do not satisfy all of these properties. For example, correlation 
dissimilarity does not satisfy the triangle inequality, and cosine dissimilarity 
can be zero for non-identical observations.

The `dissimilarity()` function includes an argument called `diss_method`, which 
specifies how dissimilarity is computed. This argument takes a dissimilarity 
constructor object that defines the corresponding method and its parameters. The 
package provides a variety of such constructors.

The advantages of using `dissimilarity()` function over standard methods in `R` 
are twofold: _i_) it allows the computation of dissimilarities between 
observations in two different sets (e.g., training and test sets) without needing 
to combine them into a single dataset; and _ii_) it consolidates various 
dissimilarity measures under a common interface, making it easier to switch
between methods and compare results. 

## Main notation

Here, the following notation is used:

- $Xr$ is the training set of spectra (or predictors), with dimensions $n \times p$ (observations $\times$ variables).

- $Yr$ is the training set of response variables, with dimensions $n \times m$ (observations $\times$ response variables). Typically, $Yr$ includes only one response variable only.

- $Xu$ is the test set of spectra, with dimensions $n_u \times p$ (observations $\times$ variables).


## Available dissimilarity construtors

The `dissimilarity()` function accepts method constructor objects that fully 
specify the dissimilarity computation:
 
| Constructor | Description |
|-------------|-------------|
| `diss_pca()` | Mahalanobis distance in PCA score space |
| `diss_pls()` | Mahalanobis distance in PLS score space (requires `Yr`) |
| `diss_correlation()` | Correlation-based dissimilarity |
| `diss_euclidean()` | Euclidean distance |
| `diss_mahalanobis()` | Mahalanobis distance in original space (requires $n > p$) |
| `diss_cosine()` | Cosine dissimilarity (spectral angle mapper) |


In all of these constructors, preprocessing instructions can be specified through the `center` and `scale` arguments. Setting `center = TRUE` and `scale = TRUE` indicates that the data should be centered and scaled before computing dissimilarity. In the case of `diss_pls()`, however, centering is always performed internally, and therefore a `center` argument is not provided for that constructor. Where available, all the constructors use `center = TRUE` and `scale = FALSE` as defaults. 

```{r}
#| message: false
#| echo: false
library(resemble)
```

```{r}
diss_correlation()
```

```{r}
diss_correlation(center = TRUE, scale = TRUE)
```


# Dissimilarity in orthogonal projection spaces

Spectral data typically lie on a lower-dimensional manifold embedded in the 
high-dimensional measurement space. Measuring dissimilarity in this reduced 
space (via PCA or PLS projections) is often more meaningful than operating 
directly on raw spectra [@ramirez2013distance].
 
Given two observations $t_i$ and $t_j$ already projected onto their score space, 
their Mahalanobis distance/dissimilarity is:

$$d(x_i, x_j) = d(t_i, t_j) = \sqrt{\frac{1}{o} (t_i - t_j) \, C^{-1} (t_i - t_j)^\top}$$

where $o$ is the number of components and $C^{-1}$ is the inverse covariance of 
the scores. Because PCA scores are orthogonal, their covariance matrix ((C)) is diagonal, with the variance of each score on the diagonal. As a result, the Mahalanobis distance in PCA score space reduces to a variance-scaled Euclidean distance, obtained by standardizing each score by its standard deviation before computing the Euclidean distance. After this standardization, the covariance matrix becomes the identity matrix, with ones on the diagonal, so the Mahalanobis distance is exactly the Euclidean distance in the standardized PCA score space [@de2000mahalanobis].

One important aspect to note in this computation is that the Mahalanobis distance is scaled by the number of components ($o$) to ensure that the resulting dissimilarity values are on a comparable scale regardless of the number of components retained. This scaling prevents the distance from artificially increasing as more components are included, which could otherwise lead to misleading interpretations of similarity or dissimilarity between observations. This scaling is often overlooked in other `R` implementations. 


```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

## _PCA-based dissimilarity_
```{r}
#| label: pca-diss
#| eval: TRUE

# Default: variance-based component selection (ncomp_by_var(0.01))
d_pca <- dissimilarity(train_x, diss_method = diss_pca())

# With OPC-based component selection (requires Yr)
d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30))
)

# Between training and test sets
d_pca_tr_ts <- dissimilarity(
 train_x,
 Xu = test_x,
 Yr = train_y,
 diss_method = diss_pca(
   ncomp = ncomp_by_opc(30),
   return_projection = TRUE
 )
)

d_pca_tr_ts
```

To access the dissimilarity matrix: 
```{r}
# Example of the first 5 rows and columns of the dissimilarity 
# matrix between training and test sets
# Rows: training observations; 
# Columns: test observations
d_pca_tr_ts$dissimilarity[1:5, 1:5]
```

The projected latent variables used to asses the dissimilarity between training and test sets can be obtained by setting `return_projection = TRUE` in the PCA constructor. This allows us to visualize the relative positions of training and test observations in the PCA score space, which can provide insights into their spectral similarity and potential model performance.

```{r}
d_pca_tr_ts$projection
```


## _PLS-based dissimilarity_

PLS projections maximise covariance with the response variable, making them 
particularly effective when the goal is to find spectrally similar observations 
that also have similar response values.
```{r pls-diss, eval = FALSE}
# Default: OPC-based component selection
d_pls <- dissimilarity(
  train_x,
  Xu = test_x,
  Yr = train_y,
  diss_method = diss_pls()
)

# Fixed number of components
d_pls_fixed <- dissimilarity(
  train_x,
  Xu = test_x,
  Yr = train_y,
  diss_method = diss_pls(ncomp = 15)
)
```

# Correlation dissimilarity

## _Conventional correlation dissimilarity_

Correlation dissimilarity is based on the Pearson correlation coefficient 
$\rho$ between observations:

$$d(x_i, x_j) = \frac{1}{2}(1 - \rho(x_i, x_j))$$

Values range from 0 (identical shape) to 1 (perfectly anticorrelated). This 
measure captures spectral shape regardless of amplitude differences.
```{r cor-diss, eval = FALSE}
# Standard correlation dissimilarity
d_cor <- dissimilarity(train_x, diss_method = diss_correlation())
d_cor
```

The conventional correlation dissimilarity can be also obtained with the standard `cor()` function in `R` as follows:

```{r}
correlation_baser <- 0.5 * (1 - cor(t(train_x)))
corrrelation_resemble <- dissimilarity(
  train_x, diss_method = diss_correlation(center = FALSE)
)
# the maximum discreepancy between the two implementations should 
# be very close to zero
max(abs(corrrelation_resemble$dissimilarity - correlation_baser))
```

However, the correlation dissimilarity implemented in `resemble` is computationally more efficient than the corresponding approach in base `R`. This gain in efficiency comes from optimized internal routines and more efficient handling of high-dimensional data, making it better suited for spectral datasets with many variables. The following code can be run locally to compare the computational performance of the two implementations:

```{r}
#| eval: false
# Compare computational speed: resemble vs base R correlation dissimilarity
n_iter <- 50

# Transpose once before timing (fair comparison)
train_x_t <- t(train_x)

# resemble implementation
time_resemble <- system.time({
  for (i in seq_len(n_iter)) {
    cor_resemble <- dissimilarity(train_x, diss_method = diss_correlation(center = FALSE))
  }
})

# base R implementation
time_base <- system.time({
  for (i in seq_len(n_iter)) {
    cor_base <- 0.5 * (1 - cor(train_x_t))
  }
})

# Results
data.frame(
  method = c("resemble", "base R"),
  elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
  per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
)
```

## Moving window correlation dissimilarity

Full-spectrum correlation can obscure localised spectral differences: a 
chemically meaningful distinction confined to a narrow wavelength region may 
contribute little to the overall correlation when averaged over hundreds of 
variables. The moving window variant addresses this by computing correlation 
over sliding windows of width `ws`, then averaging across all windows:

$$d(x_i, x_j; w) = \frac{1}{2w} \sum_{k=1}^{p-w} \left(1 - \rho(x_{i,k:k+w}, x_{j,k:k+w})\right)$$

This accumulates evidence from multiple local comparisons, making the 
dissimilarity more sensitive to region-specific absorption features. It also 
reduces sensitivity to broadband baseline offsets that can inflate full-spectrum 
correlation without reflecting true compositional similarity.

The trade-off is computational cost and the need to select an appropriate window 
size. Moving-window correlation dissimilarity can be computed using the `ws` 
argument in the `diss_correlation()` constructor as in the following example: 

```{r cor-diss-mw, eval = FALSE}
# Moving window correlation (window size must be odd)
d_cor_mw <- dissimilarity(
  train_x,
  Xu = test_x,
  diss_method = diss_correlation(ws = 51)
)
d_cor_mw
```

## Parallel execution for correlation-based dissimilarity (advanced users)

The internal functions for computing correlation dissimilarity in `resemble` have been optimised for computational efficiency and can take advantage of parallel processing when available. The underlying C++ implementation uses OpenMP for parallel computation.

By default, OpenMP will use all available cores. To explicitly control the number of threads:

```{r omp-threads-use, eval = FALSE}
# Use 4 threads
Sys.setenv(OMP_NUM_THREADS = 4)

# Or use all available cores
Sys.setenv(OMP_NUM_THREADS = parallel::detectCores())
```

When running `dissimilarity()` from within an existing parallel backend (e.g., `foreach`, `future`, `mclapply`), nested parallelism can cause thread oversubscription (where more threads are spawned than available cores) leading to degraded performance. Additionally, if R is linked against a multi-threaded BLAS library (OpenBLAS, MKL), its threads may compete with OpenMP threads.

To avoid these issues in nested parallel contexts:

```{r omp-threads-restrict, eval = FALSE}
# Limit OpenMP threads
Sys.setenv(OMP_NUM_THREADS = 1)

# If using OpenBLAS
Sys.setenv(OPENBLAS_NUM_THREADS = 1)

# If using MKL
Sys.setenv(MKL_NUM_THREADS = 1)
```

Alternatively, the `RhpcBLASctl` package provides programmatic control:

```{r blas-threads, eval = FALSE}
RhpcBLASctl::blas_set_num_threads(1)
RhpcBLASctl::omp_set_num_threads(1)
```

# Euclidean dissimilarity

The Euclidean distance between two observations $x_i$ and $x_j$ is computed as:

$$d(x_i, x_j) = \sqrt{\frac{1}{p} \sum_{k=1}^{p} (x_{i,k} - x_{j,k})^2}$$

where $p$ is the number of variables. The implementation in `resemble` scales by 
$1/p$, making results comparable across datasets with different numbers of 
variables.

```{r euclid-diss, eval = FALSE}
d_euclid <- dissimilarity(train_x, Xu = test_x, diss_method = diss_euclidean())
d_euclid
```

Euclidean distance treats each variable equally, which can be problematic with 
spectral data where variables are highly correlated and unevenly informative. 
Correlated variables effectively receive higher weight, and noisy regions 
contribute as much as chemically informative ones. For these reasons, 
`diss_pca()` or `diss_pls()` are generally more appropriate for spectral data.

Note that here, the Euclidean distance is also scaled depending on the number of variables, which is not the case in the base `R` implementation (`dist(method = "euclidean"`). This scaling ensures that the resulting dissimilarity values are on a comparable scale regardless of the number of variables, preventing the distance from artificially increasing as more variables are included, which could otherwise lead to misleading interpretations of similarity or dissimilarity between observations.


The base `R` function `dist()` and `dissimilarity()` in `resemble` with the `diss_euclidean()` constructor compute equivalent Euclidean dissimilarities, but their results differ by a scaling factor. In `resemble`, the Euclidean dissimilarity is scaled by the number of variables, whereas `dist()` returns the unscaled Euclidean distance. 

```{r}
ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
ed_baser <- as.matrix(dist(train_x, method = "euclidean"))

# scaling the base R Euclidean distance by the number of variables:
ed_baser_scaled <- sqrt((ed_baser^2) / ncol(train_x))

# differences between the two implementations
max(abs(ed_baser_scaled - ed_resemble$dissimilarity))
```

Another important aspect to note is that the Euclidean dissimilarity computations 
in `resemble` tend to be more computationally efficient than the base `R` implementation, especially for larger datasets. This efficiency is achieved through optimized algorithms and data handling strategies that reduce computational overhead, making it more suitable for high-dimensional spectral data. Execute the following code in your local environment to compare the computational speed of the two implementations:

```{r}
#| label: euclid-speed
#| eval: false
# Compare computational speed: resemble vs base R
n_iter <- 50

# resemble implementation
time_resemble <- system.time({
  for (i in seq_len(n_iter)) {
    ed_resemble <- dissimilarity(train_x, diss_method = diss_euclidean(center = FALSE))
  }
})

# base R implementation
time_base <- system.time({
  for (i in seq_len(n_iter)) {
    ed_base <- as.matrix(dist(train_x, method = "euclidean"))
  }
})

# Results
data.frame(
  method = c("resemble", "base R"),
  elapsed_sec = c(time_resemble["elapsed"], time_base["elapsed"]),
  per_call_ms = c(time_resemble["elapsed"], time_base["elapsed"]) / n_iter * 1000
)
```

 
# Cosine dissimilarity

Cosine dissimilarity, also known as the spectral angle mapper (SAM), measures 
the angle between two observation vectors:

$$d(x_i, x_j) = \cos^{-1} \frac{\sum_{k=1}^{p} x_{i,k} \, x_{j,k}}{\sqrt{\sum_{k=1}^{p} x_{i,k}^2} \sqrt{\sum_{k=1}^{p} x_{j,k}^2}}$$

This measure captures spectral shape independently of overall amplitude. Two 
spectra that differ only by a multiplicative scaling factor will have zero 
cosine dissimilarity. This property makes it useful when intensity variations 
arise from path length differences or sample concentration rather than 
compositional changes.

```{r cosine-diss, eval = FALSE}
d_cosine <- dissimilarity(train_x, Xu = test_x, diss_method = diss_cosine())
d_cosine
```


# Evaluating dissimilarity methods

A dissimilarity measure is useful if observations that are close in the
spectral space also tend to be similar in relevant sample properties. In
`resemble`, this can be assessed with `diss_evaluate()`, which evaluates a
dissimilarity matrix by comparing each observation with its first nearest
neighbor using external information provided through `side_info`.

Let $j(i)$ denote the index of the nearest neighbor of observation $i$,
excluding the observation itself:

$$
j(i) = NN(x_i, X^{\{-i\}})
$$

where $NN(x_i, X^{\{-i\}})$ returns the nearest neighbor of $x_i$ among all
other observations.

For numeric side information, such as `Ciso` in the `NIRSoil` dataset, 
`diss_evaluate()` summarizes the quality of the dissimilarity matrix using two criteria:

1. The root mean squared difference (RMSD) between each observation and its
   nearest neighbor:
$$
RMSD = \sqrt{\frac{1}{m} \sum_{i=1}^{m} \left(y_i - y_{j(i)}\right)^2},
$$
where $y_i$ is the side-information value of observation $i$,
$y_{j(i)}$ is the value of its nearest neighbor, and $m$ is the number of
observations.

2. The correlation between the observed side-information values and the values
   of their nearest neighbors.

Lower RMSD and higher correlation indicate that the dissimilarity measure is
more effective at placing observations with similar properties close to one
another.

For categorical side information, `diss_evaluate()` uses Cohen's kappa index:

$$
\kappa = \frac{p_o - p_e}{1 - p_e}
$$

where \(p_o\) is the observed agreement and \(p_e\) is the agreement expected
by chance.

In the present example, `Ciso` is numeric, so the evaluation focuses on RMSD
and correlation.

The code below computes several dissimilarity matrices using different methods:

```{r}
# Compute dissimilarity matrices using different methods
d_pca_var <- dissimilarity(train_x, diss_method = diss_pca(scale = TRUE))

d_pca_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pca(ncomp = ncomp_by_opc(30), scale = TRUE)
)

d_pls_opc <- dissimilarity(
  train_x,
  Yr = train_y,
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE)
)

d_cor <- dissimilarity(
  train_x,
  diss_method = diss_correlation(ws = 51, scale = TRUE)
)

d_euclid <- dissimilarity(
  train_x,
  diss_method = diss_euclidean(scale = TRUE)
)

d_cosine <- dissimilarity(
  train_x,
  diss_method = diss_cosine(scale = TRUE)
)
```

The following code evaluates each dissimilarity matrix using `Ciso` as side
information:

```{r}
# Evaluate each method using Ciso as side information
side_info <- as.matrix(train_y)

eval_results <- list(
  "PCA (var)" = diss_evaluate(
    d_pca_var$dissimilarity, side_info
  ),
  "PCA (opc)" = diss_evaluate(
    d_pca_opc$dissimilarity, side_info
  ),
  "PLS (opc)" = diss_evaluate(
    d_pls_opc$dissimilarity, side_info
  ),
  correlation = diss_evaluate(
    d_cor$dissimilarity, side_info
  ),
  Euclidean = diss_evaluate(
    d_euclid$dissimilarity, side_info
  ),
  cosine = diss_evaluate(
    d_cosine$dissimilarity, side_info
  )
)

# Extract RMSD and correlation for each method
comparison <- do.call(rbind, lapply(names(eval_results), function(nm) {
  data.frame(
    method = nm,
    rmsd   = eval_results[[nm]]$eval[, "rmsd"],
    r      = eval_results[[nm]]$eval[, "r"]
  )
}))
comparison
```

The resulting summary allows comparison of how well each method retrieves
nearest neighbors that are also similar in terms of `Ciso`. In general, methods
with lower RMSD and higher correlation provide a more useful representation of
spectral proximity for this variable. Methods that incorporate `Yr` during the
projection step, such as PCA or PLS with OPC-based component selection, often
perform better because the latent space is optimized to preserve similarities
that are relevant to the side information.

The next figure provides a graphical view of this evaluation. For each method,
the x-axis shows the `Ciso` value of the first nearest neighbor of each
observation, while the y-axis shows the observed `Ciso` value of the
corresponding sample. Points closer to the 1:1 line indicate better agreement
between observations and their nearest neighbors in terms of the side
information.

```{r}
#| label: diss-eval-plot
#| fig-cap: "Comparison of dissimilarity methods based on first nearest-neighbor evaluation using `Ciso` as side information. In each panel, the x-axis shows the `Ciso` value of the nearest neighbor identified from the spectral dissimilarity matrix, and the y-axis shows the observed `Ciso` value of the corresponding sample. The dashed line represents perfect agreement (y = x)."
#| fig-align: "center"
#| fig-width: 8
#| fig-height: 6
#| echo: true
#| fig-show: hold
blue  <- "#3B82F6"
amber <- "#F59E0B"
slate_grid <- "#33415540"

old_par <- par(mfrow = c(2, 3), mar = c(4, 4, 3, 1))

for (nm in names(eval_results)) {
  xy <- eval_results[[nm]]$first_nn
  plot(
    xy[, 2], xy[, 1],
    xlab = "Ciso (1-NN)",
    ylab = "Ciso",
    col = adjustcolor(blue, alpha.f = 0.5),
    pch = 16,
    main = nm
  )
  grid(lty = 1, col = slate_grid)
  abline(0, 1, col = amber, lwd = 2)
}

par(old_par)
```


# References {-} 

---
title: "Building a library of models with liblex"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{8 Building a library of models with liblex}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| message: false
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
library(resemble)
library(prospectr)
```

:::: {.columns}
::: {.column width="70%"}
> *All models are wrong, but some are useful* -- [@box1976science]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Traditional memory-based learning (MBL) methods fit a new local model for each 
query observation. While effective, this per-query refitting can be 
computationally expensive and requires access to the full reference library at 
prediction time.

The `liblex()` function (*library of localised experts*) takes a different 
approach: it pre-computes a library of local models during a build phase, then 
retrieves and combines relevant experts at prediction time without refitting. 
This design offers several advantages:
 
- **Computational efficiency**: Models are fitted once; prediction requires only 
  retrieval and weighted averaging.
- **Privacy preservation**: Only model coefficients and centroids need to be 
  stored (not the original training spectra).
- **Intrinsic uncertainty**: Disagreement among retrieved experts provides a 
  natural measure of prediction uncertainty.
- **Interpretability**: Each expert retains its regression coefficients and 
  variable importance, enabling inspection of local relationships.

The method is introduced and described in detail in @ramirezlopez2026b.

# How liblex works

The `liblex` algorithm operates in two phases:

## Build phase

1. **Anchor selection**: A subset of reference observations (anchors) is 
   selected. Each anchor will generate one local expert model. By default, 
   all observations serve as anchors; for large datasets, a representative 
   subset can be specified via `anchor_indices`.

2. **Neighborhood construction**: For each anchor, the $k$ nearest neighbors 
   are identified from the full reference set using a dissimilarity measure. 
   Note that neighbors are drawn from all reference observations, not just 
   anchors.

3. **Local model fitting**: A partial least squares (PLS) regression model 
   (or variant such as weigthed average PLS or modified PLS) is fitted within each neighborhood, producing regression coefficients. The neighborhood centroid is stored for use during prediction.

4. **Validation** (optional): Nearest-neighbor cross-validation assesses 
   performance and optimizes hyperparameters ($k$ and PLS component range).

## Prediction phase

1. **Expert retrieval**: For a new observation, dissimilarities to all stored 
   centroids are computed, and the $k$ nearest experts are retrieved. The 
   same optimal $k$ determined during fitting is used here, so the number of 
   anchors should be at least as large as the maximum $k$ being evaluated.

2. **Weighted combination**: Each retrieved expert generates a prediction. 
   These predictions are combined via distance-based kernel weighting, with 
   closer experts receiving higher weights.

3. **Uncertainty estimation**: The weighted variance across expert predictions 
   provides a per-sample uncertainty estimate, reflecting disagreement among 
   retrieved experts.

# Data preparation

```{r}
#| label: data-prep
#| message: false
library(resemble)
library(prospectr)

data(NIRsoil)

# Wavelengths
wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split into training and test sets
# Note: missing values in the response are allowed in liblex
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

cat("Training set:", nrow(train_x), "observations\n")
cat("Test set:", nrow(test_x), "observations\n")
```

# Core components

## Neighbor selection

Neighborhood size is specified using `neighbors_k()` for fixed-$k$ selection 
or `neighbors_diss()` for threshold-based selection. Multiple values can be 
provided for hyperparameter tuning.

```{r}
#| label: neighbors-example1
# Fixed neighborhood sizes to evaluate
neighbors_k(k = seq(40, 120, by = 20))
```


```{r}
#| label: neighbors-example2
# Threshold-based selection with bounds
neighbors_diss(
  threshold = seq(0.05, 0.5, length.out = 10), 
  k_min = 40, 
  k_max = 150
)
```

## Dissimilarity methods

The dissimilarity measure determines how neighbors are identified. Available 
methods include `diss_pca()`, `diss_pls()`, `diss_correlation()`, 
`diss_cosine()`, etc (see the `dissimilarity()` function for all methods). The moving-window correlation dissimilarity is particularly effective for spectral data:

```{r}
#| label: diss-example
# Moving-window correlation dissimilarity (recommended for spectra)
diss_correlation(ws = 37, scale = TRUE)

# PCA-based Mahalanobis distance with optimized components
diss_pca()
```

## Fitting method

The default method in the `liblex()` function to fit the local regressions is the weighted average PLS (waPLS), which combines predictions from multiple PLS component counts. The `fit_wapls()` constructor specifies the 
component range and algorithm:

```{r}
#| label: fit-method-example
# waPLS with modified PLS algorithm and scaling
fit_wapls(
  min_ncomp = 3, 
  max_ncomp = 15, 
  scale = TRUE, 
  method = "mpls"
)
```

## Control parameters

The `liblex_control()` function specifies operational settings:

| Parameter | Description |
|-----------|-------------|
| `mode` | `"build"` (fit models) or `"validate"` (evaluate only) |
| `tune` | If `TRUE`, optimize hyperparameters via nearest-neighbor validation |
| `metric` | Optimization metric: `"rmse"` or `"r2"` |
| `allow_parallel` | Enable parallel computation |

```{r}
#| label: control-example
#| eval: false
# Build library with hyperparameter tuning
liblex_control(mode = "build", tune = TRUE)
```

# Building a library of models

## Using fixed neighborhood size(s)

The following example builds a library using fixed-$k$ neighbor selection with 
moving-window correlation dissimilarity:



```{r}
#| eval: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE)
)

ciso_lib_k
```

```{r}
#| label: liblex-fixed-k
#| echo: false
#| message: false
#| warning: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_k
```

The plot method for liblex objects visualizes the performance of the fitted experts across different neighborhood sizes and shows the centroids of the neighborhoods. @fig-liblex shows the resulting plots for the library built with fixed neighborhood sizes. The top panel compares the performance of the best model obtained for each neighborhood size based on the RMSE, while the bottom panel displays the centroids of the neighborhoods, which are used later in prediction to select the models to be used.

```{r}
#| label: fig-liblex
#| fig-cap: "Top: Best model obtained for each neighborhood size (based on the RMSE). Bottom: Centroids of the neighborhoods, usled later in prediction to selet the models to be used."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 6.5
plot(ciso_lib_k)
```

The centroids in @fig-liblex represent the average spectra of the neighbors for each expert. During prediction, the similarity between the new observation and these centroids is used to determine which experts to retrieve and how to weight their predictions.

## Using dissimilarity thresholds

Alternatively, neighbors can be selected based on a dissimilarity threshold. 
The `k_min` and `k_max` arguments ensure reasonable neighborhood sizes:


```{r}
#| label: liblex-threshold
#| echo: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_thr
```

```{r}
#| eval: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE)
)

ciso_lib_thr
```

## Visualizing neighborhood centroids

The stored neighborhood centroids can be compared against the spectra to be 
predicted. Good coverage of the prediction space by the centroids indicates 
that relevant experts are available:

```{r}
#| label: fig-centroids
#| fig-cap: "Neighborhood centroids (blue) compared to test spectra (red)"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center
wavs_pr <- as.numeric(colnames(ciso_lib_k$scaling$local_x_center))

matplot(
  wavs_pr,
  t(test_x),
  col = rgb(1, 0, 0, 0.3),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance"
)

matlines(
  wavs_pr,
  t(ciso_lib_k$scaling$local_x_center),
  col = rgb(0, 0, 1, 0.3),
  lty = 1
)

grid(lty = 1)

legend(
  "topright",
  legend = c("Samples to predict", "Neighborhood centroids"),
  col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
  lty = 1,
  lwd = 2,
  bty = "n"
)
```




## Visualizing the regression models

The stored regression coefficients provide insight into how different spectral 
regions contribute to predictions across the library. Each expert model has 
its own set of coefficients, reflecting the local relationships learned within 
its neighborhood. @fig-regression-coefficients shows an example of how the regression models can be visualized.

```{r}
#| label: fig-regression-coefficients
#| fig-cap: "Regression coefficients (top) and intercept distribution (bottom) of the local experts"
#| fig-width: 8
#| fig-height: 7
#| fig-align: center

par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Regression coefficients across wavelengths
models <- ciso_lib_k$coefficients

matplot(
  x = wavs_pr,
  y = t(models$B),
  type = "l",
  lty = 1,
  col = rgb(0.23, 0.51, 0.96, 0.15),
  xlab = "Wavelength (nm)",
  ylab = "Regression coefficient",
  main = paste0("Regression coefficients (", nrow(models$B), " experts)")
)

# Add mean coefficient profile
lines(wavs_pr, colMeans(models$B), col = "red", lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Individual experts", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96, 0.6), "red"),
  lty = 1,
  lwd = c(1, 2),
  bty = "n"
)

# Distribution of intercepts
plot(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96),
  lwd = 2,
  xlab = "Intercept value",
  ylab = "Density",
  main = "Distribution of intercepts"
)
polygon(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96, 0.2),
  border = NA
)
abline(v = mean(models$B0, na.rm = TRUE), col = "red", lty = 2, lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Density", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96), "red"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)

par(mfrow = c(1, 1))
```

The top panel shows regression coefficients for each expert model across the 
spectral range, with the mean profile highlighted in red. Regions with high 
coefficient variability indicate wavelengths where local relationships differ 
substantially across the reference set. The bottom panel shows the distribution 
of intercepts, reflecting the range of baseline predictions across experts.


# Choosing anchor samples

For large libraries, building models for every observation may be 
computationally prohibitive. The `anchor_indices` argument allows specifying a 
subset of observations as anchors while still using the full library for 
neighbor retrieval.

## Using k-means sampling

The `naes()` function from the `prospectr` package implements the _k_-means sampling algorithm, which is a reasonable approach for selecting 
representative anchor samples:

```{r}
#| label: kmeans-anchors
# Select 350 representative anchors using k-means sampling
# on the first 20 principal components
set.seed(1124)
kms <- naes(
  train_x, 
  k = 350, 
  pc = 20, 
  iter.max = 100,
  .center = TRUE, 
  .scale = TRUE
)

anchor_km <- kms$model

cat("Selected", length(anchor_km), "anchors via k-means\n")
```

## Building the library with selected anchors

```{r}
#| label: liblex-anchored
#| echo: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_anchored
```



```{r}
#| eval: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE)
)

ciso_lib_anchored
```
the final number of models here can be verifyied by looking at the number of rows in the `ciso_lib_anchored$coefficients$B` matrix, which contains regression coefficients for each expert. 

```{r}
# Number of experts in the anchored library
n_experts <- nrow(ciso_lib_anchored$coefficients$B)
cat("Number of experts in the anchored library:", n_experts, "\n")
```


Is good practice to compare the spectra of the samples to be predicted against the centroids of the models stored in the library. @fig-centroids-anchored shows the centroids of the anchored library compared to the test spectra. This visualization helps assess whether the selected anchors provide good coverage of the spectral space of the test samples, which is crucial for reliable predictions.



```{r}
#| label: fig-centroids-anchored
#| fig-cap: "Neighborhood centroids from anchored library compared to test spectra"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center

wavs_pr <- as.numeric(colnames(ciso_lib_anchored$scaling$local_x_center))
n_experts <- nrow(ciso_lib_anchored$scaling$local_x_center)
n_test <- nrow(test_x)

# Plot test spectra first (background)
matplot(
  wavs_pr,
  t(test_x),
  col = rgb(0.8, 0.2, 0.2, 0.2),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance",
  main = paste0("Coverage: ", n_experts, " experts vs ", n_test, " test samples")
)

# Overlay centroids
matlines(
  wavs_pr,
  t(ciso_lib_anchored$scaling$local_x_center),
  col = rgb(0.23, 0.51, 0.96, 0.3),
  lty = 1
)
grid(lty = 1)

legend(
  "topright",
  legend = c(
    paste0("Test samples (n = ", n_test, ")"),
    paste0("Centroids (n = ", n_experts, ")")
  ),
  col = c(
    rgb(0.8, 0.2, 0.2, 0.5),
    rgb(0.23, 0.51, 0.96, 0.5)
  ),
  lty = 1,
  lwd = c(1, 1, 2, 2),
  bty = "n"
)
```


# Prediction

The `predict()` method retrieves relevant experts for each new observation and 
combines their predictions:

```{r}
#| label: predict-basic
ciso_pred <- predict(ciso_lib_k, newdata = test_x, verbose = FALSE)

# Prediction output structure
names(ciso_pred)

# Main predictions with uncertainty
head(ciso_pred$predictions)
```

## Prediction output

The prediction result contains:

| Component | Description |
|-----------|-------------|
| `pred` | Weighted mean prediction |
| `pred_sd` | Weighted standard deviation (uncertainty proxy) |
| `pred_q*` | Weighted quantiles of expert predictions |
| `gh` | GH distance (Mahalanobis distance in PLS space) |
| `min_yr`, `max_yr` | Response range in the neighborhood |
| `below_min`, `above_max` | Flags for extrapolation |

## Weighting options

The `weighting` argument controls how expert predictions are combined:

```{r}
#| label: weighting-options
#| eval: false
# Gaussian kernel (default)
predict(ciso_lib_k, newdata = test_x, weighting = "gaussian")

# Tricubic kernel (similar to LOCAL algorithm)
predict(ciso_lib_k, newdata = test_x, weighting = "tricube")

# Equal weights
predict(ciso_lib_k, newdata = test_x, weighting = "none")
```

Available kernels include `"gaussian"`, `"tricube"`, `"triweight"`, 
`"triangular"`, `"quartic"`, `"parabolic"`, and `"cauchy"`.

## Enforcing specific experts

The `enforce_indices` argument ensures certain experts are always included in 
the prediction neighborhood. This is useful when some anchors are known to be 
from the same domain as the target observations.

```{r}
#| label: enforce-indices
#| eval: false
# Always include specific experts in predictions
predict(ciso_lib_k, newdata = test_x, enforce_indices = c(1, 5, 10))
```

# Validation and performance

## Evaluation metrics

```{r}
#| label: evaluation
pred_values <- ciso_pred$predictions$pred
pred_sd <- ciso_pred$predictions$pred_sd

# Performance metrics (excluding missing values)
complete <- !is.na(test_y)
r2 <- cor(pred_values[complete], test_y[complete])^2
rmse <- sqrt(mean((pred_values[complete] - test_y[complete])^2))

cat("R²:  ", round(r2, 3), "\n")
cat("RMSE:", round(rmse, 3), "\n")
```

## Uncertainty as a quality filter

The prediction standard deviation (`pred_sd`) reflects disagreement among 
experts and can be used to identify unreliable predictions:

```{r}
#| label: uncertainty-filter
# Filter predictions by uncertainty threshold
unc_threshold <- quantile(pred_sd[complete], 0.75)
reliable <- complete & (pred_sd < unc_threshold)

r2_filtered <- cor(pred_values[reliable], test_y[reliable])^2
rmse_filtered <- sqrt(mean((pred_values[reliable] - test_y[reliable])^2))

cat("After filtering high-uncertainty predictions:\n")
cat("  Retained:", sum(reliable), "/", sum(complete), "observations\n")
cat("  R²:      ", round(r2_filtered, 3), "\n")
cat("  RMSE:    ", round(rmse_filtered, 3), "\n")
```

## Visualization

```{r}
#| label: fig-validation
#| fig-cap: "Predicted versus observed values"
#| fig-width: 6
#| fig-height: 5
#| fig-align: center
lims <- range(pred_values[complete], test_y[complete], na.rm = TRUE)

plot(
  pred_values[complete],
  test_y[complete],
  pch = 16,
  col = rgb(0, 0, 0, 0.5),
  xlab = "Predicted",
  ylab = "Observed",
  xlim = lims,
  ylim = lims
)

abline(0, 1, col = "red", lwd = 2)
grid(lty = 1)
```

# Comparison with classical MBL

The key difference between `liblex()` and `mbl()` is when models are fitted:

| Aspect | `mbl()` | `liblex()` |
|--------|---------|------------|
| Model fitting | Per query (on demand) | Once (build phase) |
| Prediction | Fit + predict | Retrieve + combine |
| Storage | Reference library | Coefficients + centroids |
| Uncertainty | Typically requires resampling | Intrinsic (expert dispersion) |

For applications requiring repeated predictions on new data, `liblex()` offers 
computational advantages once the library is built.

# Parallel processing

::: {.callout-note}
These functions support parallel execution via the `foreach` and `doParallel`
packages. However, parallel execution is only beneficial when the workload per
iteration is large enough to outweigh the overhead of spawning worker processes
and serialising data between them. In practice this means large prediction or
reference sets (typically hundreds of observations or more), large
neighbourhoods, and many PLS components. For small datasets, sequential
execution is invariably faster. When in doubt, benchmark both before committing
to a parallel workflow.
:::

The following example may not be faster than sequential execution due to the relatively small size of the dataset, but it illustrates how to set up parallel processing for `liblex()`:

 
```{r}
#| label: parallel-example
#| eval: false
library(doParallel)

# Register parallel backend
n_cores <- min(parallel::detectCores() - 1, 4)
cl <- makeCluster(n_cores)
registerDoParallel(cl)

# Build library with parallel processing
ciso_lib_parallel <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 120, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(
    tune = TRUE,
    allow_parallel = TRUE,
    chunk_size = 10
  )
)

# Clean up
stopCluster(cl)
registerDoSEQ()
```

The `chunk_size` parameter in `liblex_control()` controls how many models are 
processed per parallel task. Larger values reduce scheduling overhead but may 
cause load imbalance.

# Summary

The `liblex()` function provides a retrieval-based approach to memory-based 
learning that separates model building from prediction. Key features include:

- Pre-computed library of local waPLS experts
- Retrieval-gated prediction via centroid similarity
- Intrinsic uncertainty quantification through expert dispersion
- Support for anchor subsampling to handle large libraries (e.g., via `naes()`)
- Flexible weighting schemes for combining expert predictions

For large spectral libraries where predictions need to be made repeatedly, 
this approach offers efficiency and interpretability advantages over 
traditional per-query refitting methods.

# References {-}

---
title: "Building a library of models with liblex"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{8 Building a library of models with liblex}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| message: false
#| echo: false
library(resemble)
library(prospectr)
Sys.setenv(OMP_NUM_THREADS = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *All models are wrong, but some are useful* -- [@box1976science]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Traditional memory-based learning (MBL) methods fit a new local model for each 
query observation. While effective, this per-query refitting can be 
computationally expensive and requires access to the full reference library at 
prediction time.

The `liblex()` function (*library of localised experts*) takes a different 
approach: it pre-computes a library of local models during a build phase, then 
retrieves and combines relevant experts at prediction time without refitting. 
This design offers several advantages:
 
- **Computational efficiency**: Models are fitted once; prediction requires only 
  retrieval and weighted averaging.
- **Privacy preservation**: Only model coefficients and centroids need to be 
  stored (not the original training spectra).
- **Intrinsic uncertainty**: Disagreement among retrieved experts provides a 
  natural measure of prediction uncertainty.
- **Interpretability**: Each expert retains its regression coefficients and 
  variable importance, enabling inspection of local relationships.

The method is introduced and described in detail in @ramirezlopez2026b.

# How liblex works

The `liblex` algorithm operates in two phases:

## Build phase

1. **Anchor selection**: A subset of reference observations (anchors) is 
   selected. Each anchor will generate one local expert model. By default, 
   all observations serve as anchors; for large datasets, a representative 
   subset can be specified via `anchor_indices`.

2. **Neighborhood construction**: For each anchor, the $k$ nearest neighbors 
   are identified from the full reference set using a dissimilarity measure. 
   Note that neighbors are drawn from all reference observations, not just 
   anchors.

3. **Local model fitting**: A partial least squares (PLS) regression model 
   (or variant such as weigthed average PLS or modified PLS) is fitted within each neighborhood, producing regression coefficients. The neighborhood centroid is stored for use during prediction.

4. **Validation** (optional): Nearest-neighbor cross-validation assesses 
   performance and optimizes hyperparameters ($k$ and PLS component range).

## Prediction phase

1. **Expert retrieval**: For a new observation, dissimilarities to all stored 
   centroids are computed, and the $k$ nearest experts are retrieved. The 
   same optimal $k$ determined during fitting is used here, so the number of 
   anchors should be at least as large as the maximum $k$ being evaluated.

2. **Weighted combination**: Each retrieved expert generates a prediction. 
   These predictions are combined via distance-based kernel weighting, with 
   closer experts receiving higher weights.

3. **Uncertainty estimation**: The weighted variance across expert predictions 
   provides a per-sample uncertainty estimate, reflecting disagreement among 
   retrieved experts.

# Data preparation

```{r}
#| label: data-prep
#| message: false
library(resemble)
library(prospectr)

data(NIRsoil)

# Wavelengths
wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split into training and test sets
# Note: missing values in the response are allowed in liblex
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

cat("Training set:", nrow(train_x), "observations\n")
cat("Test set:", nrow(test_x), "observations\n")
```

# Core components

## Neighbor selection

Neighborhood size is specified using `neighbors_k()` for fixed-$k$ selection 
or `neighbors_diss()` for threshold-based selection. Multiple values can be 
provided for hyperparameter tuning.

```{r}
#| label: neighbors-example1
# Fixed neighborhood sizes to evaluate
neighbors_k(k = seq(40, 120, by = 20))
```


```{r}
#| label: neighbors-example2
# Threshold-based selection with bounds
neighbors_diss(
  threshold = seq(0.05, 0.5, length.out = 10), 
  k_min = 40, 
  k_max = 150
)
```

## Dissimilarity methods

The dissimilarity measure determines how neighbors are identified. Available 
methods include `diss_pca()`, `diss_pls()`, `diss_correlation()`, 
`diss_cosine()`, etc (see the `dissimilarity()` function for all methods). The moving-window correlation dissimilarity is particularly effective for spectral data:

```{r}
#| label: diss-example
# Moving-window correlation dissimilarity (recommended for spectra)
diss_correlation(ws = 37, scale = TRUE)

# PCA-based Mahalanobis distance with optimized components
diss_pca()
```

## Fitting method

The default method in the `liblex()` function to fit the local regressions is the weighted average PLS (waPLS), which combines predictions from multiple PLS component counts. The `fit_wapls()` constructor specifies the 
component range and algorithm:

```{r}
#| label: fit-method-example
# waPLS with modified PLS algorithm and scaling
fit_wapls(
  min_ncomp = 3, 
  max_ncomp = 15, 
  scale = TRUE, 
  method = "mpls"
)
```

## Control parameters

The `liblex_control()` function specifies operational settings:

| Parameter | Description |
|-----------|-------------|
| `mode` | `"build"` (fit models) or `"validate"` (evaluate only) |
| `tune` | If `TRUE`, optimize hyperparameters via nearest-neighbor validation |
| `metric` | Optimization metric: `"rmse"` or `"r2"` |
| `allow_parallel` | Enable parallel computation |

```{r}
#| label: control-example
#| eval: false
# Build library with hyperparameter tuning
liblex_control(mode = "build", tune = TRUE)
```

# Building a library of models

## Using fixed neighborhood size(s)

The following example builds a library using fixed-$k$ neighbor selection with 
moving-window correlation dissimilarity:



```{r}
#| eval: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE)
)

ciso_lib_k
```

```{r}
#| label: liblex-fixed-k
#| cache: true
#| echo: false
#| message: false
#| warning: false
ciso_lib_k <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_k
```

The plot method for liblex objects visualizes the performance of the fitted experts across different neighborhood sizes and shows the centroids of the neighborhoods. @fig-liblex shows the resulting plots for the library built with fixed neighborhood sizes. The top panel compares the performance of the best model obtained for each neighborhood size based on the RMSE, while the bottom panel displays the centroids of the neighborhoods, which are used later in prediction to select the models to be used.

```{r}
#| label: fig-liblex
#| fig-cap: "Top: Best model obtained for each neighborhood size (based on the RMSE). Bottom: Centroids of the neighborhoods, usled later in prediction to selet the models to be used."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 6.5
plot(ciso_lib_k)
```

The centroids in @fig-liblex represent the average spectra of the neighbors for each expert. During prediction, the similarity between the new observation and these centroids is used to determine which experts to retrieve and how to weight their predictions.

## Using dissimilarity thresholds

Alternatively, neighbors can be selected based on a dissimilarity threshold. 
The `k_min` and `k_max` arguments ensure reasonable neighborhood sizes:


```{r}
#| label: liblex-threshold
#| cache: true
#| echo: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_thr
```

```{r}
#| cache: true
#| eval: false
ciso_lib_thr <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_diss(
    threshold = seq(0.05, 0.5, length.out = 5), 
    k_min = 40, 
    k_max = 150
  ),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE)
)

ciso_lib_thr
```

## Visualizing neighborhood centroids

The stored neighborhood centroids can be compared against the spectra to be 
predicted. Good coverage of the prediction space by the centroids indicates 
that relevant experts are available:

```{r}
#| label: fig-centroids
#| fig-cap: "Neighborhood centroids (blue) compared to test spectra (red)"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center
wavs_pr <- as.numeric(colnames(ciso_lib_k$scaling$local_x_center))

matplot(
  wavs_pr,
  t(test_x),
  col = rgb(1, 0, 0, 0.3),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance"
)

matlines(
  wavs_pr,
  t(ciso_lib_k$scaling$local_x_center),
  col = rgb(0, 0, 1, 0.3),
  lty = 1
)

grid(lty = 1)

legend(
  "topright",
  legend = c("Samples to predict", "Neighborhood centroids"),
  col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
  lty = 1,
  lwd = 2,
  bty = "n"
)
```




## Visualizing the regression models

The stored regression coefficients provide insight into how different spectral 
regions contribute to predictions across the library. Each expert model has 
its own set of coefficients, reflecting the local relationships learned within 
its neighborhood. @fig-regression-coefficients shows an example of how the regression models can be visualized.

```{r}
#| label: fig-regression-coefficients
#| fig-cap: "Regression coefficients (top) and intercept distribution (bottom) of the local experts"
#| fig-width: 8
#| fig-height: 7
#| fig-align: center

par(mfrow = c(2, 1), mar = c(4, 4, 2, 1))

# Regression coefficients across wavelengths
models <- ciso_lib_k$coefficients

matplot(
  x = wavs_pr,
  y = t(models$B),
  type = "l",
  lty = 1,
  col = rgb(0.23, 0.51, 0.96, 0.15),
  xlab = "Wavelength (nm)",
  ylab = "Regression coefficient",
  main = paste0("Regression coefficients (", nrow(models$B), " experts)")
)

# Add mean coefficient profile
lines(wavs_pr, colMeans(models$B), col = "red", lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Individual experts", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96, 0.6), "red"),
  lty = 1,
  lwd = c(1, 2),
  bty = "n"
)

# Distribution of intercepts
plot(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96),
  lwd = 2,
  xlab = "Intercept value",
  ylab = "Density",
  main = "Distribution of intercepts"
)
polygon(
  density(models$B0, na.rm = TRUE),
  col = rgb(0.23, 0.51, 0.96, 0.2),
  border = NA
)
abline(v = mean(models$B0, na.rm = TRUE), col = "red", lty = 2, lwd = 2)
grid(lty = 1)
legend(
  "topright",
  legend = c("Density", "Mean"),
  col = c(rgb(0.23, 0.51, 0.96), "red"),
  lty = c(1, 2),
  lwd = 2,
  bty = "n"
)

par(mfrow = c(1, 1))
```

The top panel shows regression coefficients for each expert model across the 
spectral range, with the mean profile highlighted in red. Regions with high 
coefficient variability indicate wavelengths where local relationships differ 
substantially across the reference set. The bottom panel shows the distribution 
of intercepts, reflecting the range of baseline predictions across experts.


# Choosing anchor samples

For large libraries, building models for every observation may be 
computationally prohibitive. The `anchor_indices` argument allows specifying a 
subset of observations as anchors while still using the full library for 
neighbor retrieval.

## Using k-means sampling

The `naes()` function from the `prospectr` package implements the _k_-means sampling algorithm, which is a reasonable approach for selecting 
representative anchor samples:

```{r}
#| label: kmeans-anchors
# Select 350 representative anchors using k-means sampling
# on the first 20 principal components
set.seed(1124)
kms <- naes(
  train_x, 
  k = 350, 
  pc = 20, 
  iter.max = 100,
  .center = TRUE, 
  .scale = TRUE
)

anchor_km <- kms$model

cat("Selected", length(anchor_km), "anchors via k-means\n")
```

## Building the library with selected anchors

```{r}
#| label: liblex-anchored
#| cache: true
#| echo: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE),
  verbose = FALSE
)

ciso_lib_anchored
```



```{r}
#| cache: true
#| eval: false
ciso_lib_anchored <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 80, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  anchor_indices = anchor_km,
  control = liblex_control(tune = TRUE)
)

ciso_lib_anchored
```
the final number of models here can be verifyied by looking at the number of rows in the `ciso_lib_anchored$coefficients$B` matrix, which contains regression coefficients for each expert. 

```{r}
# Number of experts in the anchored library
n_experts <- nrow(ciso_lib_anchored$coefficients$B)
cat("Number of experts in the anchored library:", n_experts, "\n")
```


Is good practice to compare the spectra of the samples to be predicted against the centroids of the models stored in the library. @fig-centroids-anchored shows the centroids of the anchored library compared to the test spectra. This visualization helps assess whether the selected anchors provide good coverage of the spectral space of the test samples, which is crucial for reliable predictions.



```{r}
#| label: fig-centroids-anchored
#| fig-cap: "Neighborhood centroids from anchored library compared to test spectra"
#| fig-width: 8
#| fig-height: 5
#| fig-align: center

wavs_pr <- as.numeric(colnames(ciso_lib_anchored$scaling$local_x_center))
n_experts <- nrow(ciso_lib_anchored$scaling$local_x_center)
n_test <- nrow(test_x)

# Plot test spectra first (background)
matplot(
  wavs_pr,
  t(test_x),
  col = rgb(0.8, 0.2, 0.2, 0.2),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance",
  main = paste0("Coverage: ", n_experts, " experts vs ", n_test, " test samples")
)

# Overlay centroids
matlines(
  wavs_pr,
  t(ciso_lib_anchored$scaling$local_x_center),
  col = rgb(0.23, 0.51, 0.96, 0.3),
  lty = 1
)
grid(lty = 1)

legend(
  "topright",
  legend = c(
    paste0("Test samples (n = ", n_test, ")"),
    paste0("Centroids (n = ", n_experts, ")")
  ),
  col = c(
    rgb(0.8, 0.2, 0.2, 0.5),
    rgb(0.23, 0.51, 0.96, 0.5)
  ),
  lty = 1,
  lwd = c(1, 1, 2, 2),
  bty = "n"
)
```


# Prediction

The `predict()` method retrieves relevant experts for each new observation and 
combines their predictions:

```{r}
#| label: predict-basic
ciso_pred <- predict(ciso_lib_k, newdata = test_x, verbose = FALSE)

# Prediction output structure
names(ciso_pred)

# Main predictions with uncertainty
head(ciso_pred$predictions)
```

## Prediction output

The prediction result contains:

| Component | Description |
|-----------|-------------|
| `pred` | Weighted mean prediction |
| `pred_sd` | Weighted standard deviation (uncertainty proxy) |
| `pred_q*` | Weighted quantiles of expert predictions |
| `gh` | GH distance (Mahalanobis distance in PLS space) |
| `min_yr`, `max_yr` | Response range in the neighborhood |
| `below_min`, `above_max` | Flags for extrapolation |

## Weighting options

The `weighting` argument controls how expert predictions are combined:

```{r}
#| label: weighting-options
#| eval: false
# Gaussian kernel (default)
predict(ciso_lib_k, newdata = test_x, weighting = "gaussian")

# Tricubic kernel (similar to LOCAL algorithm)
predict(ciso_lib_k, newdata = test_x, weighting = "tricube")

# Equal weights
predict(ciso_lib_k, newdata = test_x, weighting = "none")
```

Available kernels include `"gaussian"`, `"tricube"`, `"triweight"`, 
`"triangular"`, `"quartic"`, `"parabolic"`, and `"cauchy"`.

## Enforcing specific experts

The `enforce_indices` argument ensures certain experts are always included in 
the prediction neighborhood. This is useful when some anchors are known to be 
from the same domain as the target observations.

```{r}
#| label: enforce-indices
#| eval: false
# Always include specific experts in predictions
predict(ciso_lib_k, newdata = test_x, enforce_indices = c(1, 5, 10))
```

# Validation and performance

## Evaluation metrics

```{r}
#| label: evaluation
pred_values <- ciso_pred$predictions$pred
pred_sd <- ciso_pred$predictions$pred_sd

# Performance metrics (excluding missing values)
complete <- !is.na(test_y)
r2 <- cor(pred_values[complete], test_y[complete])^2
rmse <- sqrt(mean((pred_values[complete] - test_y[complete])^2))

cat("R²:  ", round(r2, 3), "\n")
cat("RMSE:", round(rmse, 3), "\n")
```

## Uncertainty as a quality filter

The prediction standard deviation (`pred_sd`) reflects disagreement among 
experts and can be used to identify unreliable predictions:

```{r}
#| label: uncertainty-filter
# Filter predictions by uncertainty threshold
unc_threshold <- quantile(pred_sd[complete], 0.75)
reliable <- complete & (pred_sd < unc_threshold)

r2_filtered <- cor(pred_values[reliable], test_y[reliable])^2
rmse_filtered <- sqrt(mean((pred_values[reliable] - test_y[reliable])^2))

cat("After filtering high-uncertainty predictions:\n")
cat("  Retained:", sum(reliable), "/", sum(complete), "observations\n")
cat("  R²:      ", round(r2_filtered, 3), "\n")
cat("  RMSE:    ", round(rmse_filtered, 3), "\n")
```

## Visualization

```{r}
#| label: fig-validation
#| fig-cap: "Predicted versus observed values"
#| fig-width: 6
#| fig-height: 5
#| fig-align: center
lims <- range(pred_values[complete], test_y[complete], na.rm = TRUE)

plot(
  pred_values[complete],
  test_y[complete],
  pch = 16,
  col = rgb(0, 0, 0, 0.5),
  xlab = "Predicted",
  ylab = "Observed",
  xlim = lims,
  ylim = lims
)

abline(0, 1, col = "red", lwd = 2)
grid(lty = 1)
```

# Comparison with classical MBL

The key difference between `liblex()` and `mbl()` is when models are fitted:

| Aspect | `mbl()` | `liblex()` |
|--------|---------|------------|
| Model fitting | Per query (on demand) | Once (build phase) |
| Prediction | Fit + predict | Retrieve + combine |
| Storage | Reference library | Coefficients + centroids |
| Uncertainty | Typically requires resampling | Intrinsic (expert dispersion) |

For applications requiring repeated predictions on new data, `liblex()` offers 
computational advantages once the library is built.

# Parallel processing

::: {.callout-note}
These functions support parallel execution via the `foreach` and `doParallel`
packages. However, parallel execution is only beneficial when the workload per
iteration is large enough to outweigh the overhead of spawning worker processes
and serialising data between them. In practice this means large prediction or
reference sets (typically hundreds of observations or more), large
neighbourhoods, and many PLS components. For small datasets, sequential
execution is invariably faster. When in doubt, benchmark both before committing
to a parallel workflow.
:::

The following example may not be faster than sequential execution due to the relatively small size of the dataset, but it illustrates how to set up parallel processing for `liblex()`:

 
```{r}
#| label: parallel-example
#| eval: false
library(doParallel)

# Register parallel backend
n_cores <- min(parallel::detectCores() - 1, 4)
cl <- makeCluster(n_cores)
registerDoParallel(cl)

# Build library with parallel processing
ciso_lib_parallel <- liblex(
  Xr = train_x, 
  Yr = train_y, 
  neighbors = neighbors_k(seq(40, 120, by = 20)),
  diss_method = diss_correlation(ws = 37, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 3,
    max_ncomp = 15,
    scale = TRUE,
    method = "mpls"
  ),
  control = liblex_control(
    tune = TRUE,
    allow_parallel = TRUE,
    chunk_size = 10
  )
)

# Clean up
stopCluster(cl)
registerDoSEQ()
```

The `chunk_size` parameter in `liblex_control()` controls how many models are 
processed per parallel task. Larger values reduce scheduling overhead but may 
cause load imbalance.

# Summary

The `liblex()` function provides a retrieval-based approach to memory-based 
learning that separates model building from prediction. Key features include:

- Pre-computed library of local waPLS experts
- Retrieval-gated prediction via centroid similarity
- Intrinsic uncertainty quantification through expert dispersion
- Support for anchor subsampling to handle large libraries (e.g., via `naes()`)
- Flexible weighting schemes for combining expert predictions

For large spectral libraries where predictions need to be made repeatedly, 
this approach offers efficiency and interpretability advantages over 
traditional per-query refitting methods.

# References {-}

---
title: "Classical memory-based learning (local modelling)"
author: 
  - name: Leonardo Ramirez-Lopez;    
    email: ramirez.lopez.leo@gmail.com  
  - name: Alexandre M.J.-C. Wadoux
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
knitr:
  opts_chunk:
    class-output: "ansi"
vignette: >
  %\VignetteIndexEntry{6 Classical memory-based learning (local modelling)}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *Think Globally, Fit Locally* -- [@saul2003think]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::




```{r}
#| message: false
#| echo: false
library(resemble)
```



# Memory-based learning

Memory-based learning (MBL) describes a family of local learning methods that
are well suited to complex and heterogeneous spectral datasets
[@ramirez2013spectrum]. In MBL, instead of fitting a single global regression
function, a local regression model is fitted for each target observation using
its nearest neighbors identified in a calibration or reference set. Although
the global relationship between $X$ and $Y$ may be complex, MBL approximates it
through a collection of simpler local models, each assumed to be valid within a
restricted region of the predictor space [@mitchell1997machine].

For a set of $m$ observations requiring prediction, this can be expressed as

$$
\hat{y}_i = \hat{f}_i(x_i; \theta_i), \qquad i = 1, \ldots, m
$$

where $\hat{f}_i$ denotes the local prediction function fitted for observation
$i$, and $\theta_i$ represents the corresponding model parameters, such as the
number of latent variables in a local PLS model.

In this sense, MBL can be viewed as a collection of local prediction functions,

$$
\hat{f} = \{\hat{f}_1, \ldots, \hat{f}_m\},
$$

constructed on demand for the observations to be predicted. The @fig-mbl illustrates the basic steps in MBL for a set of five observations ($m = 5$).

::: {#fig-mbl fig-align="center"}
![](MBL.gif){width="65%"}

Example of the main steps in memory-based learning for predicting a response variable for five observations in a $p$-dimensional space.
:::


There are four basic elements that must be specified in any MBL algorithm:

1. **A dissimilarity measure**: This is required for neighbor search. Ideally, the dissimilarity measure should also reflect differences that are relevant to the response variable for which local models are to be built. For example, in soil NIR spectroscopy, spectral dissimilarities between soil samples should reflect, at least to some extent, their compositional differences. If the dissimilarity measure fails to capture this relevant notion of similarity, the resulting MBL models are more likely to show poor predictive performance.

2. **How many neighbors to use**: The neighborhood size used to fit the local models must be chosen carefully. Very small neighborhoods may be overly sensitive to noise and outliers, which can reduce model robustness [@ramirez2013spectrum]. They may also contain insufficient variability to adequately capture the relationship between predictors and response. In contrast, overly large neighborhoods may include observations from more heterogeneous regions of the predictor space, which can introduce more complex relationships and reduce local model accuracy.

3. **How to use the dissimilarity information**: Dissimilarity information can be used in different ways:
   - *Ignored*, in which case it is used only to retrieve neighbors [e.g., the LOCAL algorithm, @shenk1997investigation].
   - *Used to weight the training observations* according to their dissimilarity to the target observation [e.g., locally weighted PLS regression, @naes1990locally].
   - *Used as a source of additional predictors* [@ramirez2013spectrum]. In this case, the pairwise dissimilarity matrix among the $k$ neighbors is also included in the local predictor matrix. This $k \times k$ matrix is combined with the original $p$ predictor variables, resulting in a final predictor matrix of dimensions $k \times (k + p)$. For prediction of the target observation, the $p$ spectral variables are combined with the vector of dissimilarities between the target observation and its neighbors. In some cases, this strategy can improve predictive performance. The combined predictor matrix for the neighborhood can be represented as:

$$
\begin{bmatrix}
0_{1,1} & d_{1,2} & \cdots & d_{1,k} & x_{1,1} & x_{1,2} & \cdots & x_{1,p} \\
d_{2,1} & 0_{2,2} & \cdots & d_{2,k} & x_{2,1} & x_{2,2} & \cdots & x_{2,p} \\
\vdots  & \vdots  & \ddots & \vdots  & \vdots  & \vdots  & \ddots & \vdots \\
d_{k,1} & d_{k,2} & \cdots & 0_{k,k} & x_{k,1} & x_{k,2} & \cdots & x_{k,p}
\end{bmatrix}
$$

   where $d_{i,j}$ represents the dissimilarity between the $i$th and $j$th neighbors.

4. **How to fit the local models**: This is determined by the regression method used within each neighborhood. In many spectroscopic applications, these local models are linear, because the relationship between predictors and response is often assumed to be approximately linear within a restricted local region.

In the literature, MBL is often referred to as *local modeling*. However, local modeling also includes other strategies, such as cluster-based modeling or geographically segmented modeling. In this sense, MBL can be regarded as one particular type of local modeling [@ramirez2013spectrum].

The `mbl()` function in the `resemble` package provides a flexible framework for building customized memory-based learners. It allows the user to choose among different dissimilarity measures, neighborhood-selection methods, ways of using dissimilarity information, and regression methods for fitting the local models.

Readers are encouraged to review the sections on dissimilarity measures and k-nearest neighbor search first, as they provide the basis for the examples presented in this section.

The `mbl()` function can be described in terms of the four basic elements introduced above:

1. *The dissimilarity measure*: This is controlled by the `diss_method` argument of `mbl()`. The available methods are the same as those described in the dissimilarity section and are specified through constructor functions such as `diss_pca()`, `diss_pls()`, and `diss_correlation()`.

2. *How many neighbors to use*: This is controlled by the `neighbors` argument, which accepts either `neighbors_k()` for fixed neighborhood sizes or `neighbors_diss()` for threshold-based selection. Multiple neighborhood sizes can be passed to `neighbors_k()`, and multiple dissimilarity thresholds can be passed to `neighbors_diss()`, allowing several settings to be evaluated in a single run. For `neighbors_diss()`, the `k_min` and `k_max` arguments define the minimum and maximum allowed neighborhood sizes.

3. *How to use the dissimilarity information*: This is controlled by the `diss_usage` argument. If `"none"` is used, the dissimilarity information is used only for neighbor retrieval. If `"weights"` is used, the neighbors are weighted according to their dissimilarity to the target observation using a tricubic weighting function. If `"predictors"` is used, the dissimilarity information is incorporated as an additional set of predictors.

4. *How to fit the local models*: This is controlled by the `fit_method` argument, which takes a fitting-method object specifying the regression method and its parameters. Three methods are currently available: partial least squares (PLS) regression, weighted average partial least squares regression (WAPLS) [@shenk1997investigation], and Gaussian process regression (GPR) with a dot-product covariance function. 

# Fitting methods and constructors

The regressions available for fitting the local models in `mbl()` are specified through fitting-method objects. These objects are created using constructor functions such as `fit_pls()`, `fit_wapls()`, and `fit_gpr()`. Each of these functions accepts specific arguments that control the parameters of the corresponding regression method. For example, `fit_pls()` accepts the `ncomp` argument to specify the number of latent variables in the PLS model, while `fit_wapls()` accepts `min_ncomp` and `max_ncomp` to define a range of latent variable numbers for optimization. The following examples illustrate how to construct these fitting-method objects:

```{r}
#| label: fit-methods
#| eval: true

# Creates an object with instructions to build PLS models
my_pls <- fit_pls(ncomp = 15)
my_pls

# Creates an object with instructions to build WAPLS models
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 20)
my_wapls

# Creates an object with instructions to build GPR models
my_gpr <- fit_gpr()
my_gpr
```

# The `mbl_control()` function and validation types

The `mbl_control()` function is used to create control objects that define
validation settings and selected optimization options for `mbl()`. In
particular, the `validation_type` argument specifies how local model
performance is evaluated.

Two validation methods are available:

- **Leave-nearest-neighbor-out validation (`"NNv"`)**: For each target
  observation, its nearest neighbor within the selected neighborhood is
  excluded, a local model is fitted using the remaining neighbors, and the
  excluded neighbor is then predicted. The predicted values obtained for these
  excluded nearest neighbors are compared with their observed reference values.

- **Local leave-group-out cross-validation (`"local_cv"`)**: For each local
  neighborhood, observations are repeatedly split into calibration and
  validation subsets. At each iteration, a local model is fitted to the
  calibration subset and used to predict the validation subset, from which a
  local root mean squared error is computed. This procedure is repeated
  multiple times, and the final local error is calculated as the average of the
  root mean squared errors across iterations. In `mbl_control()`, the number of
  iterations is controlled by the `number` argument, while `p` defines the
  proportion of observations retained in each calibration subset.

The following examples illustrate how to create control objects for validation
in `mbl()`:

```{r}
#| label: validation-control
#| eval: true
# Create an object with instructions to conduct both validation types
# "NNv" and "local_cv"
two_val_control <- mbl_control(
  validation_type = c("NNv", "local_cv"),
  number = 10,
  p = 0.75
)
```

The object `two_val_control` stores the settings for both validation methods,
`"NNv"` and `"local_cv"`. For `"local_cv"`, the number of iterations is set to
`r two_val_control$number`, and the proportion of neighbors retained in each
local calibration subset is `r 100 * two_val_control$p`%.

# Model fitting and prediction

With the main components of `mbl()` now defined, the function can be used to
predict response values for the observations in `test_x` using `train_x` as the
reference set. The following configuration reproduces the LOCAL algorithm
[@shenk1997investigation]:

```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

```{r}
#| label: mbl-local
#| eval: true
#| results: hide
# Define the neighborhood sizes to test
my_ks <- seq(80, 160, by = 40)

# Define how to use the dissimilarity information (ignore it)
ignore_diss <- "none"

# Define the regression method to be used at each neighborhood
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 25, scale = TRUE)

# For the moment use only "NNv" validation (faster)
nnv_val_control <- mbl_control()

# Predict Total Carbon
# (remove missing values)
local_ciso <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_k(my_ks),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  gh = TRUE,
  control = nnv_val_control
)
```

In the above code, the arguments `Xr`, `Yr` are used to specify the reference or training set of predictors variables and the reponse while `Xu` is used to specify the target or test set. The argument `Yu` (not yet introduced here) is also available to specify the response values for the target set, which can be used for validation purposes. 

The subscript $u$ denotes the target domain (partially "unknown"), whereas $r$ denotes the reference library. 


The `local_ciso` object can now be examined:

```{r}
#| label: fig-localresultsciso
#| fig-cap: "MBL results for Total Carbon predictions using the LOCAL algorithm. NNv: nearest-neighbor cross-validation."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4
plot(local_ciso, main = "")
local_ciso
```

```{r}
#| label: bestk
#| eval: true
#| echo: false
#| results: hide
bestk <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bestk <- local_ciso$validation_results$nearest_neighbor_validation$k[bestk]
```

According to the results of the previous example, the neighborhood size that minimizes the root mean squared error (RMSE) under nearest-neighbor validation is `r bestk`. The corresponding LOCAL model can now be used to generate predictions for the `test_x` dataset:

```{r}
#| label: get-predictions
#| eval: true
#| results: hide
#| fig-show: hide
bki <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bk <- local_ciso$validation_results$nearest_neighbor_validation$k[bki]

# All the prediction results are stored in:
local_ciso$results

# The get_predictions function makes easier to retrieve the
# predictions from the previous object
ciso_hat <- as.matrix(get_predictions(local_ciso))[, bki]
```

@fig-plot-predictions shows the predicted values for Total Carbon obtained with the LOCAL algorithm using the best neighborhood size according to nearest-neighbor validation, plotted against the reference values in `test_y`. 

```{r}
#| label: fig-plot-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation."
#| eval: true
#| fig-width: 5
#| fig-height: 5
# Plot predicted vs reference
rng <- range(ciso_hat, test_y, na.rm = TRUE)
plot(ciso_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a fixed k", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid(lty = 1)
abline(0, 1, col = "red")
```

The prediction root mean squared error is then:

```{r}
#| eval: true
# Prediction RMSE:
sqrt(mean((ciso_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_hat, test_y, use = "complete.obs")^2
```

Similar results are obtained when the optimization of the neighborhoods is based on distance thresholds:

```{r}
#| label: mbl-diss-threshold
#| eval: true
#| results: hide
#| fig-show: hide
# Create a vector of dissimilarity thresholds to evaluate
# since the correlation dissimilarity will be used
# these thresholds need to be > 0 and <= 1
dths <- seq(0.025, 0.15, by = 0.025)

# Indicate the minimum and maximum sizes allowed for the neighborhood
k_min <- 30
k_max <- 150

local_ciso_diss <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dths, k_min = k_min, k_max = k_max),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  control = nnv_val_control
)
```

```{r}
#| label: plot-diss-results
#| eval: false
plot(local_ciso_diss)
```

```{r}
#| label: bestd
#| eval: true
#| echo: false
#| results: hide
bestd <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bestd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bestd]
```

```{r}
#| label: show-diss-results
#| eval: true
local_ciso_diss
```

The best correlation dissimilarity threshold is `r bestd`. The column "p_bounded" in the table of validation results, indicates the percentage of neighborhoods for which the size was reset either to the `k_min` or `k_max` values in the `neighbors` object passed to the `neighbors` argument. The final validation lot is shown in @fig-diss-predictions.

```{r}
#| label: diss-predictions
#| eval: true
#| results: hide
#| fig-show: hide
# Best distance threshold
bdi <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bdi]

# Predictions for the best distance
ciso_diss_hat <- as.matrix(get_predictions(local_ciso_diss))[, bdi]
```

```{r}
#| label: fig-diss-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation and distance thresholds."
#| eval: true
# Plot predicted vs reference
plot(ciso_diss_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a distance threshold \nfor neighbor retrieval", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid()
abline(0, 1, col = "red")
```
```{r}
# RMSE
sqrt(mean((ciso_diss_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_diss_hat, test_y, use = "complete.obs")^2
```

## Additional examples

Here we provide few additional examples of some MBL configurations where we make use of another response variable available in the dataset: soil cation exchange capacity (CEC). This variable is perhaps more challenging to predict in comparison to Total Carbon. @tbl-addexamples provides a summary of the configurations tested in the following code examples.


| Abbreviation | Dissimilarity method | Dissimilarity usage | Local regression |
|:-------------|:---------------------|:--------------------|:-----------------|
| `mbl_cor` (LOCAL) | Correlation | None | Weighted average PLS |
| `mbl_pc` | optimized PC | Source of predictors | Weighted average PLS |
| `mbl_pls` | optimized PLS | None | Weighted average PLS |
| `mbl_gpr` | optimized PC | Source of predictors | Gaussian process |
: Basic description of the different MBL configurations in the examples to predict Cation Exchange Capacity (CEC). {#tbl-addexamples}

```{r}
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_cec <- NIRsoil$CEC[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_cec  <- NIRsoil$CEC[NIRsoil$train == 0]
```


```{r}
#| label: cec-examples
#| results: hide
#| eval: true
#| fig-show: hide
# Define the WAPLS fitting method
my_wapls <- fit_wapls(min_ncomp = 2, max_ncomp = 25, scale = FALSE)

# mbl_cor: LOCAL algorithm with correlation dissimilarity
dth_cor <- seq(0.01, 0.3, by = 0.08)
mbl_cor <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_cor, k_min = 80, k_max = 150),
  diss_method = diss_correlation(),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pc: PCA dissimilarity with dissimilarity matrix as predictors
dth_pc <- seq(0.05, 1, by = 0.4)
mbl_pc <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pls: PLS dissimilarity without dissimilarity predictors
mbl_pls <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_gpr: Gaussian process regression with PCA dissimilarity as predictors
mbl_gpr <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = fit_gpr(),
  control = nnv_val_control
)
```

Collect the predictions for each configuration:

```{r}
#| label: collect-predictions
#| eval: true
#| fig-show: hide
# Get the indices of the best results according to
# nearest neighbor validation statistics
c_val_name <- "validation_results"
c_nn_val_name <- "nearest_neighbor_validation"

bi_cor <- which.min(mbl_cor[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pc  <- which.min(mbl_pc[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pls <- which.min(mbl_pls[[c_val_name]][[c_nn_val_name]]$rmse)
bi_gpr <- which.min(mbl_gpr[[c_val_name]][[c_nn_val_name]]$rmse)

preds <- cbind(
  get_predictions(mbl_cor)[, bi_cor],
  get_predictions(mbl_pc)[, bi_pc],
  get_predictions(mbl_pls)[, bi_pls],
  get_predictions(mbl_gpr)[, bi_gpr]
)

colnames(preds) <- c("mbl_cor", "mbl_pc", "mbl_pls", "mbl_gpr")
preds <- as.matrix(preds)

# R2s
cor(test_cec, preds, use = "complete.obs")^2

# RMSEs
colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5
```

@fig-mblcomparisons illustrate the prediction results obtained for CEC with each of the MBL configurations tested.

```{r}
#| label: cec-plots-code
#| eval: false
#| fig-show: hide
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2))

plot(test_cec, preds[, "mbl_cor"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_cor (LOCAL)")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_pc"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_pc")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_pls"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_pls")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_gpr"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_gpr")
abline(0, 1, col = "red")

par(old_par)
```

```{r}
#| label: fig-mblcomparisons
#| fig-cap: "CEC prediction results for the different MBL configurations tested"
#| fig-align: center
#| fig-width: 8
#| fig-height: 8
#| echo: false
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2), pch = 16, mar = c(4, 4, 4, 4))

my_cols <- c(
  "mbl_cor" = "#D42B08CC",
  "mbl_pc"  = "#750E3380",
  "mbl_pls" = "#EFBF4780",
  "mbl_gpr" = "#5177A180"
)

# R2s
r2s <- drop(round(cor(test_cec, preds, use = "complete.obs")^2, 2))

# RMSEs
rmses <- round(colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5, 2)

plot_titles <- c(
  "mbl_cor" = "mbl_cor (LOCAL)",
  "mbl_pc"  = "mbl_pc",
  "mbl_pls" = "mbl_pls",
  "mbl_gpr" = "mbl_gpr"
)

p <- sapply(
  colnames(preds),
  FUN = function(label, y, yhats, cols, rsq, rmse, titles) {
    plot(
      x = y,
      y = yhats[, label],
      xlim = range(y, na.rm = TRUE),
      ylim = range(y, na.rm = TRUE),
      xlab = "CEC, meq/100g",
      ylab = "Predicted CEC, meq/100g",
      col = cols[label]
    )
    title(titles[label])
    title(paste0("\n\n\n RMSE: ", rmse[label], "; R²: ", rsq[label]), cex.main = 1)
    grid(col = "#80808080", lty = 1)
    abline(0, 1, col = "#FF1A0080")
  },
  y = test_cec,
  yhats = preds,
  cols = my_cols,
  rsq = r2s,
  rmse = rmses,
  titles = plot_titles
)

par(old_par)
```

## Using `Yu` argument

If response values for the prediction set are available, the `Yu` argument can be used to directly validate the predictions produced by `mbl()`. These values are used only for validation and are not involved in any optimization or modeling step. The argument can be used as follows:

```{r}
#| label: yu-argument
#| eval: true
# Use Yu argument to validate the predictions
pc_pred_cec_yu <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_k(80),
  diss_method = diss_pca(scale = TRUE),
  diss_usage = "none",
  verbose = FALSE,
  control = mbl_control()
)

pc_pred_cec_yu
```

# Supported parallel processing

The `mbl()` function uses `foreach()` from the [foreach package](https://CRAN.R-project.org/package=foreach) to iterate over the rows in `Xu`. In the example below, the [doParallel package](https://CRAN.R-project.org/package=doParallel) is used to register the cores for parallel execution. The [doSNOW package](https://CRAN.R-project.org/package=doSNOW) can also be used. This example uses parallel processing to predict CEC.

::: {.callout-note}
These functions support parallel execution via the `foreach` and `doParallel`
packages. However, parallel execution is only beneficial when the workload per
iteration is large enough to outweigh the overhead of spawning worker processes
and serialising data between them. In practice this means large prediction or
reference sets (typically hundreds of observations or more), large
neighbourhoods, and many PLS components. For small datasets, sequential
execution is invariably faster. When in doubt, benchmark both before committing
to a parallel workflow.
:::

The following example may not be faster than sequential execution due to the relatively small size of the dataset, but it illustrates how to set up parallel processing for `mbl()`:

```{r}
#| label: parallel-example
#| eval: false
# Running the mbl function using multiple cores

# Execute with two cores, if available, ...
n_cores <- 2

# ... if not then go with 1 core
if (parallel::detectCores() < 2) {
  n_cores <- 1
}

# Set the number of cores
library(doParallel)
clust <- makeCluster(n_cores)
registerDoParallel(clust)

# Alternatively:
# library(doSNOW)
# clust <- makeCluster(n_cores, type = "SOCK")
# registerDoSNOW(clust)
# getDoParWorkers()

pc_pred_cec <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_k(seq(40, 100, by = 10)),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  control = mbl_control()
)

# Go back to sequential processing
registerDoSEQ()
try(stopCluster(clust))

pc_pred_cec
```


# References {-} 

---
title: "Classical memory-based learning (local modelling)"
author: 
  - name: Leonardo Ramirez-Lopez;    
    email: ramirez.lopez.leo@gmail.com  
  - name: Alexandre M.J.-C. Wadoux
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
knitr:
  opts_chunk:
    class-output: "ansi"
vignette: >
  %\VignetteIndexEntry{6 Classical memory-based learning (local modelling)}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *Think Globally, Fit Locally* -- [@saul2003think]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::




```{r}
#| message: false
#| echo: false
library(resemble)
```



# Memory-based learning

Memory-based learning (MBL) describes a family of local learning methods that
are well suited to complex and heterogeneous spectral datasets
[@ramirez2013spectrum]. In MBL, instead of fitting a single global regression
function, a local regression model is fitted for each target observation using
its nearest neighbors identified in a calibration or reference set. Although
the global relationship between $X$ and $Y$ may be complex, MBL approximates it
through a collection of simpler local models, each assumed to be valid within a
restricted region of the predictor space [@mitchell1997machine].

For a set of $m$ observations requiring prediction, this can be expressed as

$$
\hat{y}_i = \hat{f}_i(x_i; \theta_i), \qquad i = 1, \ldots, m
$$

where $\hat{f}_i$ denotes the local prediction function fitted for observation
$i$, and $\theta_i$ represents the corresponding model parameters, such as the
number of latent variables in a local PLS model.

In this sense, MBL can be viewed as a collection of local prediction functions,

$$
\hat{f} = \{\hat{f}_1, \ldots, \hat{f}_m\},
$$

constructed on demand for the observations to be predicted. The @fig-mbl illustrates the basic steps in MBL for a set of five observations ($m = 5$).

::: {#fig-mbl fig-align="center"}
![](MBL.gif){width="65%"}

Example of the main steps in memory-based learning for predicting a response variable for five observations in a $p$-dimensional space.
:::


There are four basic elements that must be specified in any MBL algorithm:

1. **A dissimilarity measure**: This is required for neighbor search. Ideally, the dissimilarity measure should also reflect differences that are relevant to the response variable for which local models are to be built. For example, in soil NIR spectroscopy, spectral dissimilarities between soil samples should reflect, at least to some extent, their compositional differences. If the dissimilarity measure fails to capture this relevant notion of similarity, the resulting MBL models are more likely to show poor predictive performance.

2. **How many neighbors to use**: The neighborhood size used to fit the local models must be chosen carefully. Very small neighborhoods may be overly sensitive to noise and outliers, which can reduce model robustness [@ramirez2013spectrum]. They may also contain insufficient variability to adequately capture the relationship between predictors and response. In contrast, overly large neighborhoods may include observations from more heterogeneous regions of the predictor space, which can introduce more complex relationships and reduce local model accuracy.

3. **How to use the dissimilarity information**: Dissimilarity information can be used in different ways:
   - *Ignored*, in which case it is used only to retrieve neighbors [e.g., the LOCAL algorithm, @shenk1997investigation].
   - *Used to weight the training observations* according to their dissimilarity to the target observation [e.g., locally weighted PLS regression, @naes1990locally].
   - *Used as a source of additional predictors* [@ramirez2013spectrum]. In this case, the pairwise dissimilarity matrix among the $k$ neighbors is also included in the local predictor matrix. This $k \times k$ matrix is combined with the original $p$ predictor variables, resulting in a final predictor matrix of dimensions $k \times (k + p)$. For prediction of the target observation, the $p$ spectral variables are combined with the vector of dissimilarities between the target observation and its neighbors. In some cases, this strategy can improve predictive performance. The combined predictor matrix for the neighborhood can be represented as:

$$
\begin{bmatrix}
0_{1,1} & d_{1,2} & \cdots & d_{1,k} & x_{1,1} & x_{1,2} & \cdots & x_{1,p} \\
d_{2,1} & 0_{2,2} & \cdots & d_{2,k} & x_{2,1} & x_{2,2} & \cdots & x_{2,p} \\
\vdots  & \vdots  & \ddots & \vdots  & \vdots  & \vdots  & \ddots & \vdots \\
d_{k,1} & d_{k,2} & \cdots & 0_{k,k} & x_{k,1} & x_{k,2} & \cdots & x_{k,p}
\end{bmatrix}
$$

   where $d_{i,j}$ represents the dissimilarity between the $i$th and $j$th neighbors.

4. **How to fit the local models**: This is determined by the regression method used within each neighborhood. In many spectroscopic applications, these local models are linear, because the relationship between predictors and response is often assumed to be approximately linear within a restricted local region.

In the literature, MBL is often referred to as *local modeling*. However, local modeling also includes other strategies, such as cluster-based modeling or geographically segmented modeling. In this sense, MBL can be regarded as one particular type of local modeling [@ramirez2013spectrum].

The `mbl()` function in the `resemble` package provides a flexible framework for building customized memory-based learners. It allows the user to choose among different dissimilarity measures, neighborhood-selection methods, ways of using dissimilarity information, and regression methods for fitting the local models.

Readers are encouraged to review the sections on dissimilarity measures and k-nearest neighbor search first, as they provide the basis for the examples presented in this section.

The `mbl()` function can be described in terms of the four basic elements introduced above:

1. *The dissimilarity measure*: This is controlled by the `diss_method` argument of `mbl()`. The available methods are the same as those described in the dissimilarity section and are specified through constructor functions such as `diss_pca()`, `diss_pls()`, and `diss_correlation()`.

2. *How many neighbors to use*: This is controlled by the `neighbors` argument, which accepts either `neighbors_k()` for fixed neighborhood sizes or `neighbors_diss()` for threshold-based selection. Multiple neighborhood sizes can be passed to `neighbors_k()`, and multiple dissimilarity thresholds can be passed to `neighbors_diss()`, allowing several settings to be evaluated in a single run. For `neighbors_diss()`, the `k_min` and `k_max` arguments define the minimum and maximum allowed neighborhood sizes.

3. *How to use the dissimilarity information*: This is controlled by the `diss_usage` argument. If `"none"` is used, the dissimilarity information is used only for neighbor retrieval. If `"weights"` is used, the neighbors are weighted according to their dissimilarity to the target observation using a tricubic weighting function. If `"predictors"` is used, the dissimilarity information is incorporated as an additional set of predictors.

4. *How to fit the local models*: This is controlled by the `fit_method` argument, which takes a fitting-method object specifying the regression method and its parameters. Three methods are currently available: partial least squares (PLS) regression, weighted average partial least squares regression (WAPLS) [@shenk1997investigation], and Gaussian process regression (GPR) with a dot-product covariance function. 

# Fitting methods and constructors

The regressions available for fitting the local models in `mbl()` are specified through fitting-method objects. These objects are created using constructor functions such as `fit_pls()`, `fit_wapls()`, and `fit_gpr()`. Each of these functions accepts specific arguments that control the parameters of the corresponding regression method. For example, `fit_pls()` accepts the `ncomp` argument to specify the number of latent variables in the PLS model, while `fit_wapls()` accepts `min_ncomp` and `max_ncomp` to define a range of latent variable numbers for optimization. The following examples illustrate how to construct these fitting-method objects:

```{r}
#| label: fit-methods
#| eval: true

# Creates an object with instructions to build PLS models
my_pls <- fit_pls(ncomp = 15)
my_pls

# Creates an object with instructions to build WAPLS models
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 20)
my_wapls

# Creates an object with instructions to build GPR models
my_gpr <- fit_gpr()
my_gpr
```

# The `mbl_control()` function and validation types

The `mbl_control()` function is used to create control objects that define
validation settings and selected optimization options for `mbl()`. In
particular, the `validation_type` argument specifies how local model
performance is evaluated.

Two validation methods are available:

- **Leave-nearest-neighbor-out validation (`"NNv"`)**: For each target
  observation, its nearest neighbor within the selected neighborhood is
  excluded, a local model is fitted using the remaining neighbors, and the
  excluded neighbor is then predicted. The predicted values obtained for these
  excluded nearest neighbors are compared with their observed reference values.

- **Local leave-group-out cross-validation (`"local_cv"`)**: For each local
  neighborhood, observations are repeatedly split into calibration and
  validation subsets. At each iteration, a local model is fitted to the
  calibration subset and used to predict the validation subset, from which a
  local root mean squared error is computed. This procedure is repeated
  multiple times, and the final local error is calculated as the average of the
  root mean squared errors across iterations. In `mbl_control()`, the number of
  iterations is controlled by the `number` argument, while `p` defines the
  proportion of observations retained in each calibration subset.

The following examples illustrate how to create control objects for validation
in `mbl()`:

```{r}
#| label: validation-control
#| eval: true
# Create an object with instructions to conduct both validation types
# "NNv" and "local_cv"
two_val_control <- mbl_control(
  validation_type = c("NNv", "local_cv"),
  number = 10,
  p = 0.75
)
```

The object `two_val_control` stores the settings for both validation methods,
`"NNv"` and `"local_cv"`. For `"local_cv"`, the number of iterations is set to
`r two_val_control$number`, and the proportion of neighbors retained in each
local calibration subset is `r 100 * two_val_control$p`%.

# Model fitting and prediction

With the main components of `mbl()` now defined, the function can be used to
predict response values for the observations in `test_x` using `train_x` as the
reference set. The following configuration reproduces the LOCAL algorithm
[@shenk1997investigation]:

```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

```{r}
#| label: mbl-local
#| eval: true
#| results: hide
# Define the neighborhood sizes to test
my_ks <- seq(80, 160, by = 40)

# Define how to use the dissimilarity information (ignore it)
ignore_diss <- "none"

# Define the regression method to be used at each neighborhood
my_wapls <- fit_wapls(min_ncomp = 3, max_ncomp = 25, scale = TRUE)

# For the moment use only "NNv" validation (faster)
nnv_val_control <- mbl_control()

# Predict Total Carbon
# (remove missing values)
local_ciso <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_k(my_ks),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  gh = TRUE,
  control = nnv_val_control
)
```

In the above code, the arguments `Xr`, `Yr` are used to specify the reference or training set of predictors variables and the reponse while `Xu` is used to specify the target or test set. The argument `Yu` (not yet introduced here) is also available to specify the response values for the target set, which can be used for validation purposes. 

The subscript $u$ denotes the target domain (partially "unknown"), whereas $r$ denotes the reference library. 


The `local_ciso` object can now be examined:

```{r}
#| label: fig-localresultsciso
#| fig-cap: "MBL results for Total Carbon predictions using the LOCAL algorithm. NNv: nearest-neighbor cross-validation."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4
plot(local_ciso, main = "")
local_ciso
```

```{r}
#| label: bestk
#| eval: true
#| echo: false
#| results: hide
bestk <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bestk <- local_ciso$validation_results$nearest_neighbor_validation$k[bestk]
```

According to the results of the previous example, the neighborhood size that minimizes the root mean squared error (RMSE) under nearest-neighbor validation is `r bestk`. The corresponding LOCAL model can now be used to generate predictions for the `test_x` dataset:

```{r}
#| label: get-predictions
#| eval: true
#| results: hide
#| fig-show: hide
bki <- which.min(local_ciso$validation_results$nearest_neighbor_validation$rmse)
bk <- local_ciso$validation_results$nearest_neighbor_validation$k[bki]

# All the prediction results are stored in:
local_ciso$results

# The get_predictions function makes easier to retrieve the
# predictions from the previous object
ciso_hat <- as.matrix(get_predictions(local_ciso))[, bki]
```

@fig-plot-predictions shows the predicted values for Total Carbon obtained with the LOCAL algorithm using the best neighborhood size according to nearest-neighbor validation, plotted against the reference values in `test_y`. 

```{r}
#| label: fig-plot-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation."
#| eval: true
#| fig-width: 5
#| fig-height: 5
# Plot predicted vs reference
rng <- range(ciso_hat, test_y, na.rm = TRUE)
plot(ciso_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a fixed k", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid(lty = 1)
abline(0, 1, col = "red")
```

The prediction root mean squared error is then:

```{r}
#| eval: true
# Prediction RMSE:
sqrt(mean((ciso_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_hat, test_y, use = "complete.obs")^2
```

Similar results are obtained when the optimization of the neighborhoods is based on distance thresholds:

```{r}
#| label: mbl-diss-threshold
#| eval: true
#| results: hide
#| fig-show: hide
# Create a vector of dissimilarity thresholds to evaluate
# since the correlation dissimilarity will be used
# these thresholds need to be > 0 and <= 1
dths <- seq(0.025, 0.15, by = 0.025)

# Indicate the minimum and maximum sizes allowed for the neighborhood
k_min <- 30
k_max <- 150

local_ciso_diss <- mbl(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dths, k_min = k_min, k_max = k_max),
  diss_method = diss_correlation(center = FALSE),
  diss_usage = ignore_diss,
  fit_method = my_wapls,
  control = nnv_val_control
)
```

```{r}
#| label: plot-diss-results
#| eval: false
plot(local_ciso_diss)
```

```{r}
#| label: bestd
#| eval: true
#| echo: false
#| results: hide
bestd <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bestd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bestd]
```

```{r}
#| label: show-diss-results
#| eval: true
local_ciso_diss
```

The best correlation dissimilarity threshold is `r bestd`. The column "p_bounded" in the table of validation results, indicates the percentage of neighborhoods for which the size was reset either to the `k_min` or `k_max` values in the `neighbors` object passed to the `neighbors` argument. The final validation lot is shown in @fig-diss-predictions.

```{r}
#| label: diss-predictions
#| eval: true
#| results: hide
#| fig-show: hide
# Best distance threshold
bdi <- which.min(local_ciso_diss$validation_results$nearest_neighbor_validation$rmse)
bd <- local_ciso_diss$validation_results$nearest_neighbor_validation$k_diss[bdi]

# Predictions for the best distance
ciso_diss_hat <- as.matrix(get_predictions(local_ciso_diss))[, bdi]
```

```{r}
#| label: fig-diss-predictions
#| fig-cap: "Predicted vs reference values for Total Carbon using the LOCAL algorithm with the best neighborhood size according to nearest-neighbor validation and distance thresholds."
#| eval: true
# Plot predicted vs reference
plot(ciso_diss_hat, test_y,
     xlim = rng,
     ylim = rng,
     xlab = "Predicted Total Carbon, %",
     ylab = "Total Carbon, %",
     main = "LOCAL using a distance threshold \nfor neighbor retrieval", 
     cex = 1.5,
     pch = 16, col = rgb(0.5, 0.5, 0.5, 0.6))
grid()
abline(0, 1, col = "red")
```
```{r}
# RMSE
sqrt(mean((ciso_diss_hat - test_y)^2, na.rm = TRUE))

# Squared R
cor(ciso_diss_hat, test_y, use = "complete.obs")^2
```

## Additional examples

Here we provide few additional examples of some MBL configurations where we make use of another response variable available in the dataset: soil cation exchange capacity (CEC). This variable is perhaps more challenging to predict in comparison to Total Carbon. @tbl-addexamples provides a summary of the configurations tested in the following code examples.


| Abbreviation | Dissimilarity method | Dissimilarity usage | Local regression |
|:-------------|:---------------------|:--------------------|:-----------------|
| `mbl_cor` (LOCAL) | Correlation | None | Weighted average PLS |
| `mbl_pc` | optimized PC | Source of predictors | Weighted average PLS |
| `mbl_pls` | optimized PLS | None | Weighted average PLS |
| `mbl_gpr` | optimized PC | Source of predictors | Gaussian process |
: Basic description of the different MBL configurations in the examples to predict Cation Exchange Capacity (CEC). {#tbl-addexamples}

```{r}
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_cec <- NIRsoil$CEC[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_cec  <- NIRsoil$CEC[NIRsoil$train == 0]
```


```{r}
#| label: cec-examples
#| results: hide
#| eval: true
#| fig-show: hide
# Define the WAPLS fitting method
my_wapls <- fit_wapls(min_ncomp = 2, max_ncomp = 25, scale = FALSE)

# mbl_cor: LOCAL algorithm with correlation dissimilarity
dth_cor <- seq(0.01, 0.3, by = 0.08)
mbl_cor <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_cor, k_min = 80, k_max = 150),
  diss_method = diss_correlation(),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pc: PCA dissimilarity with dissimilarity matrix as predictors
dth_pc <- seq(0.05, 1, by = 0.4)
mbl_pc <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_pls: PLS dissimilarity without dissimilarity predictors
mbl_pls <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pls(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  fit_method = my_wapls,
  control = nnv_val_control
)

# mbl_gpr: Gaussian process regression with PCA dissimilarity as predictors
mbl_gpr <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_diss(threshold = dth_pc, k_min = 80, k_max = 150),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "predictors",
  fit_method = fit_gpr(),
  control = nnv_val_control
)
```

Collect the predictions for each configuration:

```{r}
#| label: collect-predictions
#| eval: true
#| fig-show: hide
# Get the indices of the best results according to
# nearest neighbor validation statistics
c_val_name <- "validation_results"
c_nn_val_name <- "nearest_neighbor_validation"

bi_cor <- which.min(mbl_cor[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pc  <- which.min(mbl_pc[[c_val_name]][[c_nn_val_name]]$rmse)
bi_pls <- which.min(mbl_pls[[c_val_name]][[c_nn_val_name]]$rmse)
bi_gpr <- which.min(mbl_gpr[[c_val_name]][[c_nn_val_name]]$rmse)

preds <- cbind(
  get_predictions(mbl_cor)[, bi_cor],
  get_predictions(mbl_pc)[, bi_pc],
  get_predictions(mbl_pls)[, bi_pls],
  get_predictions(mbl_gpr)[, bi_gpr]
)

colnames(preds) <- c("mbl_cor", "mbl_pc", "mbl_pls", "mbl_gpr")
preds <- as.matrix(preds)

# R2s
cor(test_cec, preds, use = "complete.obs")^2

# RMSEs
colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5
```

@fig-mblcomparisons illustrate the prediction results obtained for CEC with each of the MBL configurations tested.

```{r}
#| label: cec-plots-code
#| eval: false
#| fig-show: hide
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2))

plot(test_cec, preds[, "mbl_cor"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_cor (LOCAL)")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_pc"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_pc")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_pls"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_pls")
abline(0, 1, col = "red")

plot(test_cec, preds[, "mbl_gpr"],
     xlab = "CEC, meq/100g",
     ylab = "Predicted CEC, meq/100g", 
     main = "mbl_gpr")
abline(0, 1, col = "red")

par(old_par)
```

```{r}
#| label: fig-mblcomparisons
#| fig-cap: "CEC prediction results for the different MBL configurations tested"
#| fig-align: center
#| fig-width: 8
#| fig-height: 8
#| echo: false
old_par <- par("mfrow", "mar")
par(mfrow = c(2, 2), pch = 16, mar = c(4, 4, 4, 4))

my_cols <- c(
  "mbl_cor" = "#D42B08CC",
  "mbl_pc"  = "#750E3380",
  "mbl_pls" = "#EFBF4780",
  "mbl_gpr" = "#5177A180"
)

# R2s
r2s <- drop(round(cor(test_cec, preds, use = "complete.obs")^2, 2))

# RMSEs
rmses <- round(colMeans((preds - test_cec)^2, na.rm = TRUE)^0.5, 2)

plot_titles <- c(
  "mbl_cor" = "mbl_cor (LOCAL)",
  "mbl_pc"  = "mbl_pc",
  "mbl_pls" = "mbl_pls",
  "mbl_gpr" = "mbl_gpr"
)

p <- sapply(
  colnames(preds),
  FUN = function(label, y, yhats, cols, rsq, rmse, titles) {
    plot(
      x = y,
      y = yhats[, label],
      xlim = range(y, na.rm = TRUE),
      ylim = range(y, na.rm = TRUE),
      xlab = "CEC, meq/100g",
      ylab = "Predicted CEC, meq/100g",
      col = cols[label]
    )
    title(titles[label])
    title(paste0("\n\n\n RMSE: ", rmse[label], "; R²: ", rsq[label]), cex.main = 1)
    grid(col = "#80808080", lty = 1)
    abline(0, 1, col = "#FF1A0080")
  },
  y = test_cec,
  yhats = preds,
  cols = my_cols,
  rsq = r2s,
  rmse = rmses,
  titles = plot_titles
)

par(old_par)
```

## Using `Yu` argument

If response values for the prediction set are available, the `Yu` argument can be used to directly validate the predictions produced by `mbl()`. These values are used only for validation and are not involved in any optimization or modeling step. The argument can be used as follows:

```{r}
#| label: yu-argument
#| eval: true
# Use Yu argument to validate the predictions
pc_pred_cec_yu <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  Yu = test_cec,
  neighbors = neighbors_k(80),
  diss_method = diss_pca(scale = TRUE),
  diss_usage = "none",
  verbose = FALSE,
  control = mbl_control()
)

pc_pred_cec_yu
```

# Supported parallel processing

The `mbl()` function uses `foreach()` from the [foreach package](https://CRAN.R-project.org/package=foreach) to iterate over the rows in `Xu`. In the example below, the [doParallel package](https://CRAN.R-project.org/package=doParallel) is used to register the cores for parallel execution. The [doSNOW package](https://CRAN.R-project.org/package=doSNOW) can also be used. This example uses parallel processing to predict CEC.

::: {.callout-note}
These functions support parallel execution via the `foreach` and `doParallel`
packages. However, parallel execution is only beneficial when the workload per
iteration is large enough to outweigh the overhead of spawning worker processes
and serialising data between them. In practice this means large prediction or
reference sets (typically hundreds of observations or more), large
neighbourhoods, and many PLS components. For small datasets, sequential
execution is invariably faster. When in doubt, benchmark both before committing
to a parallel workflow.
:::

The following example may not be faster than sequential execution due to the relatively small size of the dataset, but it illustrates how to set up parallel processing for `mbl()`:

```{r}
#| label: parallel-example
#| eval: false
# Running the mbl function using multiple cores

# Execute with two cores, if available, ...
n_cores <- 2

# ... if not then go with 1 core
if (parallel::detectCores() < 2) {
  n_cores <- 1
}

# Set the number of cores
library(doParallel)
clust <- makeCluster(n_cores)
registerDoParallel(clust)

# Alternatively:
# library(doSNOW)
# clust <- makeCluster(n_cores, type = "SOCK")
# registerDoSNOW(clust)
# getDoParWorkers()

pc_pred_cec <- mbl(
  Xr = train_x[!is.na(train_cec), ],
  Yr = train_cec[!is.na(train_cec)],
  Xu = test_x,
  neighbors = neighbors_k(seq(40, 100, by = 10)),
  diss_method = diss_pca(ncomp = ncomp_by_opc(), scale = TRUE),
  diss_usage = "none",
  control = mbl_control()
)

# Go back to sequential processing
registerDoSEQ()
try(stopCluster(clust))

pc_pred_cec
```


# References {-} 

---
title: "Essential concepts and setup"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{1 Essential concepts and setup}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *Think Globally, Fit Locally* -- [@saul2003think]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Spectroscopic data analysis plays a central role in many environmental, agricultural, and food-related applications. Techniques such as near-infrared (NIR), mid-infrared (IR), and other forms of diffuse reflectance spectroscopy provide rapid, non-destructive, and cost-efficient measurements that can be used to infer chemical, physical, or biological properties of complex matrices, including soils, plant materials, and food products. In quantitative applications, these measurements are typically linked to reference laboratory values through empirical calibration models.

As spectral databases grow in size and diversity, their effective use becomes increasingly challenging. Large spectral libraries often contain substantial heterogeneity, domain shifts, redundant observations, and samples that are only locally informative for a given prediction problem. Under these conditions, global modelling strategies are often insufficient on their own, and methods based on dimensionality reduction, dissimilarity analysis, neighbour retrieval, local modelling, and targeted sample selection become essential.

The `resemble` package provides a framework for sample retrieval and local learning in spectral chemometrics. It is designed to support the analysis of large and complex spectral datasets through tools for projection-based representation, dissimilarity computation, neighbourhood search, memory-based learning, evolutionary subset search, and retrieval-based modelling with pre-computed local models. The package therefore supports both classical local modelling workflows and newer strategies for exploiting spectral libraries as structured resources for predictive modelling. 

The functions presented here are implemented based on the methods described in @ramirezlopez2026a, @ramirezlopez2026b, and @ramirez2013spectrum.

The main functionalities of `resemble` include:

- orthogonal projection of spectral data using principal component analysis (PCA) and partial least squares (PLS) methods

- computation and evaluation of spectral dissimilarity measures

- nearest-neighbour search in spectral reference sets

- memory-based learning and local regression

- evolutionary sample search for context-specific calibration

- retrieval-based modelling using libraries of localised experts


# Citing the package

Simply type and you will get the info you need:
```{r}
#| eval: true
citation(package = "resemble")
```

# Dataset used across the vignettes

The vignettes in `resemble` use the soil near-infrared (NIR) spectral dataset 
provided in the [`prospectr`](https://CRAN.R-project.org/package=prospectr) 
package [@stevens2020introduction]. This dataset is used because soils are among 
the most complex matrices analyzed by NIR spectroscopy. It was originally used 
in the *Chimiométrie 2006* challenge [@pierna2008soil].

The dataset contains NIR absorbance spectra for 825 dried and sieved soil 
samples collected from agricultural fields across the Walloon region of Belgium. 
In `R`, the data are stored in a `data.frame` with the following structure:

* **Response variables**:
  
  * **`Nt`**: total nitrogen (g/kg dry soil); available for 645 samples and 
    missing for 180.
  
  * **`Ciso`**: carbon (g/100 g dry soil); available for 732 samples and 
    missing for 93. 
  
  * **`CEC`**: cation exchange capacity (meq/100 g dry soil); available for 447 
    samples and missing for 378.

* **Predictor variables (`spc`)**: the spectral predictors are stored in the matrix 
  `NIRsoil$spc`, embedded within the data frame. These variables contain NIR 
  absorbance spectra measured from 1100 to 2498 nm at 2 nm intervals. Each 
  column name corresponds to a wavelength value (in nm).

* **Set indicator (`Set`)**: a binary variable indicating whether a sample belongs to 
  the training set (`1`, 618 samples) or the test set (`0`, 207 samples).
  
  
Load the necessary packages and data: 
```{r}
#| label: libraries
#| message: false
library(resemble)
library(prospectr)
```

The dataset can be loaded into R as follows: 
```{r}
#| message: false
#| results: hide
data(NIRsoil)
dim(NIRsoil)
str(NIRsoil)
```

# Spectral preprocessing

Throughout the vignettes, the same preprocessing workflow is used to improve the suitability of the spectra for quantitative analysis. In particular, the goal is to reduce unwanted baseline variation and enhance local spectral features that may be informative for modeling. The preprocessing steps are implemented using the [`prospectr`](https://CRAN.R-project.org/package=prospectr) package [@stevens2020introduction].

The following steps are applied:

1. **Detrending** is applied first to reduce broad baseline shifts and curvature effects across the spectra.

2. A **first-order Savitzky–Golay derivative** [@Savitzky1964] is then computed to emphasize local spectral features and reduce remaining additive effects.

```{r}
#| label: NIRsoil
#| message: false
# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
```

```{r}
#| label: fig-plotspectra
#| fig-cap: "Raw spectral absorbance data (top) and first derivative of the absorbance spectra (bottom)."
#| fig-align: center
#| fig-width: 7
#| fig-height: 7
#| echo: false
# Hex sticker palette#
bg_dark <- NA #"#0F172A"      # background
blue <- "#3B82F6"         # border / spiral high
amber <- "#F59E0B"        # spiral low
amber_light <- "#FBBF24"  # text accent
slate <- "#64748B"        # muted elements

old_par <- par("mfrow", "mar", "bg")
par(mfrow = c(2, 1), mar = c(4, 4, 1, 4), bg = bg_dark)

new_wavs <- as.matrix(as.numeric(colnames(NIRsoil$spc_p)))

text_col <- "black"#"white"

# Plot 1: Raw spectra
plot(range(wavs), range(NIRsoil$spc), col = NA,
     xlab = "",
     ylab = "Absorbance",
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = wavs, y = t(NIRsoil$spc), 
         lty = 1, col = paste0(blue, "33"))

# Plot 2: First derivative
plot(range(new_wavs), range(NIRsoil$spc_p), col = NA,
     xlab = "Wavelengths, nm",
     ylab = expression(d(detrended~A)/d*lambda),
     col.lab = text_col, col.axis = text_col)
rect(par("usr")[1], par("usr")[3], par("usr")[2], par("usr")[4], col = "#1E293B")
grid(lty = 1, col = "#334155")
matlines(x = new_wavs, y = t(NIRsoil$spc_p), 
         lty = 1, col = paste0(amber, "33"))

par(old_par)
```

Both the raw absorbance spectra and the preprocessed spectra are shown in 
@fig-plotspectra. The preprocessed spectra, obtained as the first derivative of 
detrended absorbance, are used as the predictor variables in all examples 
throughout this document.

For illustration purposes, the `NIRsoil` data are divided into training and test 
subsets. In the examples that require a response variable, `Ciso` is used to 
demonstrate the functionality of the package.

```{r}
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

The notation used throughout the `resemble` package for arguments referring to 
training and test observations is as follows:

* **Training observations**:

  * `Xr` denotes the matrix of predictor variables in the reference/training set.
  
  * `Yr` denotes the response variable(s) in the reference/training set. In the
    context of this package, `Yr` may also be referred to as **side
    information**, that is, variables associated with the training observations
    that can support or guide optimization during modeling, even when they are
    not directly used as model inputs. For example, as shown in later sections,
    `Yr` can be used in principal component analysis to help determine the
    optimal number of components.

* **Test observations**:

  * `Xu` denotes the matrix of predictor variables in the unknown/test set.
  
  * `Yu` denotes the response variable(s) in the unknown/test set.


# References {-} 

---
title: "Dimensionality reduction of spectral data"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    toc-location: left
    number-sections: true
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{2 Dimensionality reduction of spectral data}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
```


:::: {.columns}
::: {.column width="70%"}
> *Probability mass concentrates near regions that have a much smaller dimensionality than the original space where the data lives* -- [@bengio2013representation]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Exploratory analysis of spectral data is often affected by the curse of dimensionality. In near-infrared (NIR) spectroscopy, for example, a single spectrum may contain measurements at hundreds to thousands of highly correlated wavelengths. Tasks such as pattern detection, similarity assessment, and outlier detection therefore often require reducing the dimensionality of the spectra while retaining the most relevant information.

Principal component analysis (PCA) and partial least squares (PLS) are among the most widely used methods for dimensionality reduction in spectroscopic analysis. Both methods are based on the idea that the meaningful structure of spectral data lies in a lower-dimensional subspace. They seek to construct a projection that transforms the original variables into a smaller set of latent variables that provides a more compact representation of the data while preserving its dominant structure.

The main difference between PCA and PLS lies in the criterion used to define the latent variables. In PCA, the objective is to obtain orthogonal components that explain as much variation in the predictor matrix as possible. In PLS, the objective is to obtain latent variables that maximize the covariance between the predictor matrix and one or more external variables, such as response variables or side information.

In PCA and PLS, the input spectra ($X$, of dimensions $n \times d$) are decomposed into a score matrix ($T$) and a loading matrix ($P$), such that

$$
X = TP^\top + E
$$

where the dimensions of $T$ and $P$ are $n \times o$ and $d \times o$, respectively, $o$ is the number of retained components, and $E$ is the residual matrix (reconstruction error). In PCA, the columns of $P$ are orthonormal ($P^\top P = I$), which allows new observations to be projected directly onto the latent space. For centered data, the maximum number of components that can be retained is $\min(n - 1, d)$.

Once a projection model has been estimated, new observations can be projected onto the same latent space. If $X_{\mathrm{new}}$ denotes a matrix of new spectra preprocessed in the same way as the original data, their scores can be obtained as

$$
T_{\mathrm{new}} = X_{\mathrm{new}} P
$$

where $P$ contains the loading vectors defining the projection space.

The `resemble` package provides the function `ortho_projection()` to compute orthogonal projections of spectral data using PCA- and PLS-based methods. These projections can be used for exploratory analysis, visualization, dissimilarity analysis, neighbor search, and predictive modeling.





# Methods available in the `resemble` package

In the `resemble` package, orthogonal projections based on PCA and PLS are computed with the `ortho_projection()` function.

The function provides the following algorithms for PCA:

* `"pca"`: standard PCA computed by singular value decomposition (SVD).  

* `"pca_nipals"`: PCA computed with the nonlinear iterative partial least squares (NIPALS) algorithm [@wold1975soft].

For PLS, the available algorithms are:

* `"pls"`: standard partial least squares decomposition computed with the NIPALS algorithm. The projection is guided by external variables provided in `Yr`, so that the latent variables maximize the covariance between the spectral predictors and the side information.

* `"mpls"`: modified partial least squares, also computed with the NIPALS algorithm. Unlike standard PLS, which derives weights from covariances, this method uses correlations between `Xr` and `Yr`, reducing the influence of predictor variance on the projection [@shenk1991populations].

* `"simpls"`: partial least squares computed with the SIMPLS algorithm [@de1993simpls], which is often computationally more efficient than NIPALS for high-dimensional data.



# Selecting the number of components

One of the critical decisions in dimensionality reduction is determining how many latent variables to extract. Retaining too few components risks discarding meaningful information, while retaining too many may introduce noise or lead to overfitting in subsequent modeling steps.

The `resemble` package provides a set of constructor functions for specifying component selection criteria. These functions return specification objects that are passed to the `ncomp` argument of `ortho_projection()` and related functions:

| Function | Criterion |
|----------|-----------|
| `ncomp_by_var()` | Individual variance threshold |
| `ncomp_by_cumvar()` | Cumulative variance threshold |
| `ncomp_by_opc()` | Optimal selection via nearest-neighbor evaluation |
| `ncomp_fixed()` | Fixed number of components |

Alternatively, passing a positive integer directly to `ncomp` is equivalent to using `ncomp_fixed()`.

## Variance-based selection

The simplest approach is to retain components based on the amount of variance they explain.

### Individual variance threshold (`ncomp_by_var()`)

Retains all components that individually explain at least a specified proportion of the total variance:

```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

In the following examples, the default method (PCA with SVD) is used to compute the projection, but the same component selection criteria can be applied to any of the available methods.


```{r}
#| eval: true
# Retain components that individually explain at least 1% of variance
proj_var <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))
proj_var
```

The above configuration of `ncomp_by_var(0.01)` retained a total of `r proj_var$ncomp` components which are the ones that individually explain at least 1% of the total variance in the data. 


### Cumulative variance threshold (`ncomp_by_cumvar()`)

Retains the minimum number of components needed to reach a specified cumulative proportion of explained variance:

```{r}
#| eval: true
# Retain enough components to explain at least 99% of variance
proj_cumvar <- ortho_projection(train_x, ncomp = ncomp_by_cumvar(0.99))
proj_cumvar
```

The above configuration of `ncomp_by_cumvar(0.99)` retained a total of `r proj_cumvar$ncomp` components which are the minimum number of components needed to explain at least 99% of the total variance in the data.

## Optimal component selection (`ncomp_by_opc()`)

Variance-based criteria consider only the structure of the predictor matrix. However, in supervised contexts, the goal is often to retain components that are relevant for capturing information realted to a response variable. The _optimal component selection_ method ("opc") addresses this by incorporating side information.

The method is based on the assumption that if two observations are similar in the spectral space, they should also be similar in terms of their _side information_ [@ramirez2013spectrum]. For a sequence of retained components $o = 1, 2, \ldots, k$, the function computes a dissimilarity matrix at each step and identifies the nearest spectral neighbor for each observation. The side information values of each observation are then compared with those of its nearest neighbor.

For continuous side information, the root mean squared difference (RMSD) is computed:

$$
j(i) = \mathrm{NN}(x_{r_i}, X_r^{\{-i\}})
$$

$$
\mathrm{RMSD} = \sqrt{\frac{1}{m} \sum_{i=1}^m (y_i - y_{j(i)})^2}
$$

where $j(i)$ is the index of the nearest neighbor of observation $i$ (excluding itself), $y_i$ is the side information value for observation $i$, $y_{j(i)}$ is the side information value for the nearest neighbor, and $m$ is the total number of observations.

For categorical side information, the kappa index is used instead of RMSD.

The optimal number of components is the one that minimizes RMSD (or maximizes kappa). Note that in this case, the side information is represented by the response variable (`Ciso` in the `NIRSoil` dataset):

```{r}
#| eval: true
# Optimal component selection using side information
# using default max_ncomp which is 40
proj_opc <- ortho_projection(train_x, Yr = train_y, ncomp = ncomp_by_opc())
proj_opc
```
Using the OPC method, the optimal number of components is `r proj_opc$ncomp`. Although this may appear large, these components are required to adequately represent similarities among observations in terms of the side information (here, the `Ciso` variable). In this sense, OPC provides a semi-supervised approach for extracting meaningful latent variables, where the side information is used only to determine the optimal number of components.

The "opc" method is particularly useful because it selects latent variables that are genuinely informative with respect to the response variable. By contrast, standard component-selection methods often underestimate the number of components needed to preserve similarities among observations in terms of the response.

Note that this method can take multiple variables as side information. 

By default, up to 40 latent variables are evaluated. If fewer than 40 spectral variables are supplied to `ortho_projection()`, the number of spectral variables is used as the upper limit for the number of latent variables to test.

If fewer than 40 latent variables are to be evaluated, the `ncomp_by_opc()` 
method can be used to set a smaller upper limit. For example:

```{r}
# this needs to be passed to the ncomp argument in ortho_projection()
ncomp_by_opc(max_ncomp = 15)
```


## Fixed number of components

When the number of components is known in advance, it can be specified directly:

```{r}
#| eval: false
# Using ncomp_fixed()
proj_fixed <- ortho_projection(train_x, ncomp = ncomp_fixed(10))

# Equivalent: passing an integer directly
proj_fixed <- ortho_projection(train_x, ncomp = 10)
```

# Dimentionality reduction methods in  `ortho_projection()`

The `ortho_projection()` function computes orthogonal projections based on PCA and PLS.

## Available methods

The function provides the following algorithms for PCA:

* `"pca"`: standard PCA computed by singular value decomposition (SVD).
* `"pca_nipals"`: PCA computed with the nonlinear iterative partial least squares (NIPALS) algorithm [@wold1975soft].

For PLS, the available algorithms are:

* `"pls"`: standard partial least squares decomposition computed with the NIPALS algorithm. The projection is guided by external variables provided in `Yr`, so that the latent variables maximize the covariance between the spectral predictors and the side information.
* `"mpls"`: modified partial least squares, computed with the NIPALS algorithm. Weights are derived from correlations rather than covariances between `Xr` and `Yr`, giving equal influence to all predictors regardless of variance [@shenk1991populations].
* `"simpls"`: partial least squares computed with the SIMPLS algorithm [@de1993simpls], which is often computationally more efficient than NIPALS for high-dimensional data.

## Example: PCA projection

```{r}
#| results: hide
# PCA with default component selection (ncomp_by_var(0.01))
pca_train <- ortho_projection(train_x, method = "pca")
pca_train
```

```{r}
#| label: fig-pca-variance
#| fig-cap: "Individual contribution to the explained variance for each component (left) and cumulative variance explained by the principal components (right)."
#| fig-width: 7
#| fig-height: 3
plot(pca_train)
```

In this dataset, `r pca_train$ncomp` components are required to explain approximately `r round(100 * sum(pca_train$variance$x_var[2,]), 0)`% of the original variance in the spectra (@fig-pca-variance).

The NIPALS algorithm provides an alternative to SVD and can be faster when only a few components are required:

```{r}
#| results: hide
# PCA using NIPALS
pca_nipals_train <- ortho_projection(train_x, method = "pca_nipals")
```

## Example: PLS projection

For PLS methods, side information (`Yr`) must be provided:

```{r}
#| results: hide
# PLS using SIMPLS with 10 components
pls_train <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = 10
)
pls_train
```

Observations with missing values in `Yr` are held out during model fitting, then projected using the resulting projection matrix and pooled with the results.

## Example: Optimal component selection

```{r}
#| results: hide
# PCA with optimal component selection
pca_opc <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40)
)
pca_opc
```

```{r}
#| label: fig-pca-opc
#| fig-cap: "RMSD between observations and their nearest neighbors as a function of the number of principal components."
#| fig-width: 5
#| fig-height: 4
plot(pca_opc, col = "#F59E0B")
```

In this example, `r pca_opc$ncomp` components minimize the RMSD for Total Carbon (@fig-pca-opc).

# Projecting new data

Both PCA and PLS generate projection matrices that can be used to project new observations onto the latent space. Use the `predict()` method:

```{r}
#| results: hide
# Fit PLS projection model
pls_model <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

# Project test data
pls_test_scores <- predict(pls_model, newdata = test_x)
```

```{r}
#| label: fig-pc-proj
#| fig-width: 4.5
#| fig-height: 5
#| fig-cap: "Projection of the training observations in the first two PLS components (blue) and projection of the test observations onto the same latent space (red)."
plot(
  pls_model$scores[, 1:2], 
  col = rgb(0.231, 0.51, 0.965, 0.3), 
  pch = 16, cex = 1.5,
  xlab = "PLS1", ylab = "PLS2"
)
grid(lty = 1)
points(
  pls_test_scores[, 1:2], 
  col = rgb(0.961, 0.620, 0.043, 0.7), 
  pch = 16, cex = 1.5
)
legend(
  "topright",
  legend = c("Training projection", "Predicted projection"),
  col = c(
    rgb(0.231, 0.51, 0.965, 0.3),
    rgb(0.961, 0.620, 0.043, 0.7)
  ),
  pch = 16,
  pt.cex = 1.5,
  box.col = "black"
)
```



Alternatively, training and test data can be projected in a single call by passing the test data to `Xu`:

```{r}
#| results: hide
# Project training and test data simultaneously
pca_both <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)
```

For PCA, the training and test data are pooled, projected jointly, and then split for the final results. For `ncomp_by_opc()`, dissimilarity matrices are computed only on the training data with available side information.

# Multivariate side information

The `ortho_projection()` function accepts multiple variables as side information:

```{r}
#| results: hide
#| eval: false
train_y2 <- NIRsoil$Nt[NIRsoil$train == 1]
train_y3 <- NIRsoil$CEC[NIRsoil$train == 1]

# PLS with multivariate side information
pls_multi <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = cbind(train_y, train_y2, train_y3),
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

plot(pls_multi)
```

When multiple variables are provided, the PLS2 algorithm is used [@wold1983multivariate]. For optimal component selection, the RMSD is computed for each variable, standardized, and averaged. The evaluation results for each variable are accessible via `$opc_evaluation`.

For categorical side information, the kappa index is used instead of RMSD.  

# References {-}

---
title: "Evolutionary subset search and modeling using gesearch"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{7 Evolutionary subset search using gesearch}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

:::: {.columns}
::: {.column width="70%"}
> *Particular individuals do not recur, but their building blocks do* -- [@holland1995hidden]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::


```{r}
#| message: false
#| echo: false
library(resemble)
Sys.setenv(OMP_NUM_THREADS = 2)
```



# Introduction

Large reference datasets, such as spectral libraries, can support model development when only a limited number of target-domain samples are available. The `gesearch()` function implements the _gesearch_ algorithm described in @ramirezlopez2026a, which builds on the sample-selection idea introduced by @lobsey2017rs. The algorithm uses an evolutionary search to identify a subset of library samples that is most relevant to a target domain and then uses the selected subset to fit a context-specific model.

In its current implementation, `gesearch()` uses partial least squares (PLS) models to evaluate candidate subsets. Sample selection is guided by how well a subset predicts the target response, reconstructs the target spectra, or aligns with the target domain in latent space. The method can therefore be used for calibration transfer or domain adaptation, with or without response values from the target domain.

It is important to note that the _gesearch_ algorithm is designed to identify a single subset of samples that is most relevant to the target domain. Accordingly, some data from the target domain must be available. In addition, __an important assumption underlying__ ___gesearch___ __is that the relationships between spectra and response values in the target domain are expected to be predominantly linear__ [@ramirezlopez2026a], even though the search process itself is non-linear.

This vignette introduces the main workflow for `gesearch`, including how to define the search, control the selection process, inspect the selected subsets, and generate predictions from the final model. It also highlights key tuning arguments such as `k`, `b`, `retain`, and the control settings available through `gesearch_control()`.

# Glossary and conventions

As _gesearch_ is an evolutionary search algorithm, it draws on terminology from evolutionary optimization and sample selection. This glossary summarizes the main terms used throughout the vignette.

| Term | Meaning |
|:-----|:--------|
| Target set ($\mathcal{D}_u$) | The population of interest (`Xu`, `Yu`), typically small ($m$ samples). Response values may be partially or entirely unavailable. The subscript $u$ denotes the target domain (partially "unknown").|
| Spectral library ($\mathcal{D}_r$) | A large, heterogeneous collection of candidate samples for model training (`Xr`, `Yr`), with $n \gg m$. The subscript $r$ denotes the reference library.|
| Gene | A sample from the spectral library. |
| Gene pool | The set of genes currently eligible for selection. |
| Individual | A subset of `k` genes used to fit a candidate PLS model. |
| Population | The collection of all individuals in a given generation. |
| Silenced gene | A gene permanently excluded from all subsequent generations. |
| Weakness score | A measure of how poorly a gene contributes to model performance for the target set. |
| Generation | One iteration of the evolutionary cycle. |
| Incidence count | The number of individuals in which a gene appears. |

# The _gesearch_ algorithm: How it works

## Evolutionary search overview

The algorithm treats sample selection as an evolutionary process:

1. **Initialization**: Create a population of individuals, each containing `k` genes randomly selected from the library. A gene, and therefore its corresponding sample, can appear in multiple individuals. The representation factor `b` controls how often each gene is expected to appear across the population, so that genes are represented approximately equally.

2. **Evaluation**: Fit a PLS model for each individual. One or more figures of merit are then computed for that model using the available target samples. The weakness score of a gene is defined as the average performance of all individuals in which that gene is active. This procedure is repeated for all genes in the active gene pool.

3. **Silencing**: Permanently mark genes with high weakness scores as inactive, preventing them from being selected in subsequent generations. In practical terms, the corresponding samples are permanently excluded from further selection.

4. **Reproduction**: Generate new individuals through crossover (combining genes from two parents) and mutation (adding genes from the remaining pool).

5. **Termination**: Stop when the gene pool shrinks to `target_size`.

The surviving genes form the final calibration subset, which is expected to deliver near-optimal performance for the target population.

## Weakness score computation

The key idea behind `gesearch` is that gene quality is assessed indirectly: each individual (subset) receives a performance score, and a gene's weakness is defined as the average performance across all individuals containing that gene.

### Step 1: Evaluate each individual

For each individual $S_i^{(\gamma)}$ in generation $\gamma$, a PLS model is fitted using its `k` active genes. This model is then evaluated on the target spectra $\mathbf{X}_u$, producing an individual-level score, for example, prediction error, reconstruction error, or distance to the target set.

### Step 2: Aggregate to gene-level weakness

Each gene appears in multiple individuals, with its frequency controlled by the representation factor `b`. The weakness of gene $g_j$ is the mean of the individual-level scores across all individuals containing that gene:

$$w^{(\gamma)}(g_j) = \frac{1}{h_j^{(\gamma)}} \sum_{i:\, g_j \in S_i^{(\gamma)}} \text{score}(S_i^{(\gamma)})$$

where $h_j^{(\gamma)}$ is the incidence count, that is, the number of individuals containing gene $g_j$.

### Interpretation

A gene that consistently appears in poorly performing individuals accumulates a high weakness score. Conversely, a gene that contributes to well-performing individuals has a low weakness score. This averaging mechanism helps isolate the marginal contribution of each gene despite the combinatorial nature of subset selection.

## Weakness functions

The `optimization` argument controls which weakness criteria are applied. Multiple criteria can be combined, for example `c("reconstruction", "similarity")`. At least one criterion must be specified:

| Criterion | Individual-level score | Requires `Yu`? |
|:----------|:-----------------------|:---------------|
| `"reconstruction"` (default) | Error of reconstructing $\mathbf{X}_u$ from the PLS latent space | No |
| `"similarity"` (default) | Mahalanobis distance between the target and training score centroids | No |
| `"response"` | RMSE of predicting $\mathbf{y}_u$ | Yes |
| `"range"` | Penalty for predictions outside `Yu_lims` | No (bounds only) |

## Gene silencing

After weakness scores are computed, genes are retained or silenced based on the `retain` parameter and the retention strategy specified in `gesearch_control()`:

- **Probability-based** (`retain_by = "probability"`, default): Genes with weakness scores below the `retain` quantile are kept. This approach is more robust to outliers.
- **Proportion-based** (`retain_by = "proportion"`): A fixed proportion of genes with the lowest weakness scores is kept.

Silenced genes are permanently excluded:

$$G^{(\gamma+1)} = G^{(\gamma)} \setminus U^{(\gamma)}$$

## Final model

Evolution terminates when the gene pool falls to `target_size` or when no additional genes can be silenced (stagnation). A final PLS model is then fitted using the surviving genes, with the optimal number of components selected by cross-validation independently of the fixed `ncomp` used during evolution.

# Key parameters

| Parameter | Description |
|:----------|:------------|
| `k` | Number of active genes per individual |
| `b` | Gene-representation factor, that is, the target average number of individuals in which each gene appears |
| `retain` | Retention threshold, that is, the proportion of genes kept per generation (values > 0.9 are recommended) |
| `target_size` | Minimum gene pool size at which evolution terminates |
| `ncomp` | Number of PLS components fixed during evolution for comparability |

The population size at generation $\gamma$ is approximately:

$$p^{(\gamma)} \approx \frac{b \cdot n^{(\gamma)}}{k}$$

where $n^{(\gamma)}$ is the number of active genes remaining.

# Practical considerations

- **Keep `k` moderate**: Smaller subsets make the contribution of each sample more distinguishable, improving the sensitivity of weakness-score estimation. However, `k` must remain large enough to fit stable PLS models. When `k`becomes too large, the effect of each sample may get diluted.

- **Keep `b` high**: A higher representation factor ensures each gene appears in more individuals, yielding more robust weakness estimates. Values above 40 are generally recommended.

- **Keep `retain` high but not too high**: Values above 0.9 (e.g., 0.95–0.98) ensure that only the weakest genes are silenced at each generation, reducing the risk of prematurely discarding useful samples. Lower values accelerate convergence but may degrade selection quality.

- **Multiple weakness criteria accelerate convergence**: When multiple optimisation criteria are used (e.g., `"reconstruction"`, `"similarity"`, `"response"`), more genes are silenced per generation, speeding up evolution. However, this also increases the risk of over-silencing if criteria are redundant or noisy.

- **Set `target_size` according to practical constraints**: The target should reflect the expected number of relevant samples in the library and the requirements of downstream model fitting and validation. A larger `target_size` preserves more candidates but may retain less relevant samples; a smaller value yields a more focused subset but risks excluding useful samples.



# Example workflow

The `NIRSoil` dataset is also used here to illustrate how to perform an evolutionary subset search with `gesearch()`. The workflow includes defining the search, controlling the selection process, inspecting the selected subsets, and generating predictions from the final model.

It is important to note that the `NIRSoil` dataset may not be an ideal test case for `gesearch()`, because of two reasons: _i_) the target set is a subset of the library and the sample size is relatively small. _ii_) the test set, which could be used as the target set, appears to be dominated by non-linear relationships between spectra and response values, which may limit the performance of the search. This is mainly because the dataset originates from samples from a large geographical area, which may introduce complex interactions between soil properties and spectral features. 


## The dataset, the target set, and its practical constraints

To better illustrate the main workflow and key parameters of the method, the target set was extracted from the test set using a simple filtering strategy intended to reduce potential non-linearities between spectra and response values. Spectra that were substantially dissimilar to the median spectrum of the test set were removed. Although this does not guarantee the elimination of non-linear relationships, it provides a straightforward way to define a more spectrally homogeneous target set for demostration purposes.

```{r}
#| message: false
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Missing values in the response are allowed
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

md <- dissimilarity(
  test_x,
  apply(test_x, 2, median),
  diss_method = diss_correlation(ws = 101, center = T, scale = T)
)

threshold <- 0.4
```

```{r}
#| label: fig-diss-threshold
#| fig-cap: "Left: Dissimilarity of the test spectra to the median test spectrum. Samples above the threshold (dashed line and red) were excluded from the target set as a simple attempt to reduce spectral heterogeneity and potential non-linearities in the spectra-response relationship. Right: Spectra of the samples above (red, to be excluded) and below (blue, to be kept) the dissimilarity threshold."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 2, 1))

plot(
  seq_along(md$dissimilarity),
  md$dissimilarity, 
  ylim = c(0, 1), 
  col = "steelblue",
  pch = 16,
  cex = 1.5,
  xlab = "Sample index", 
  ylab = "Dissimilarity to median spectrum"
)
points(
  which(md$dissimilarity >= threshold),
  md$dissimilarity[md$dissimilarity >= threshold],
  col = "tomato",
  pch = 16,
  cex = 1.5
)
abline(h = threshold, col = "tomato", lty = 2, lwd = 2)
grid(lty = 1)

matplot(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity >= threshold, ]),
  col = "tomato",
  ylab = "Preprocessed spectra",
  xlab = "Wavelengths, nm",
  lty = 1,
  type = "l"
)
grid(lty = 1)
matlines(
  as.numeric(colnames(train_x)),
  t(test_x[md$dissimilarity < threshold, ]),
  col = "steelblue",
  lty = 1,
  type = "l"
)

par(op)
```

To construct the final target set, samples whose dissimilarity to the median spectrum falls below the correlation-dissimilarity threshold (`r threshold`) are retained. This set is then split into two subsets: a very small subset, used to guide the search within `gesearch()`, and a second subset, used to validate the models built from the reference samples selected by `gesearch()`. The use of only a very small subset to drive the search reflects a typical practical constraint in target-domain applications, where only a limited number of samples with laboratory reference measurements are available within a short time frame. The "target" subsets are then obtained as follows: 

```{r}
keep <- md$dissimilarity < threshold

cat("Number of samples retained:", sum(keep))

test_x_test <- test_x[keep, ]
test_y_test <- test_y[keep]

# sample a very small subset to guide the search, 
# and use the rest for testing the final model
set.seed(1124)
ref_ind <- sample(sum(keep), 8)
test_x_ref <- test_x_test[ref_ind, ]
test_y_ref <- test_y_test[ref_ind]

test_x_test <- test_x_test[-ref_ind, ]
test_y_test <- test_y_test[-ref_ind]
```

## A simple linear model for predicting in the target set (baseline)

Here, the full training set is combined with the small subset of target samples to fit a simple PLS model. This serves as a baseline for comparison with the models generated by `gesearch()`. The optimal number of components is selected by cross-validation, and the resulting model is evaluated on the test subset of the target set.

```{r}
set.seed(1124)

pls_model <- model(
  Xr = rbind(
    train_x[!is.na(train_y), ],
    test_x_ref[!is.na(test_y_ref), ]
  ), 
  Yr = c(
    train_y[!is.na(train_y)],
    test_y_ref[!is.na(test_y_ref)]
  ),
  fit_method = fit_pls(ncomp = 15, method = "simpls", scale = TRUE),
  control = model_control(validation_type = "lgo", number = 10)
)

best_ncomp <- which(pls_model$cv_results$optimal)

pred <- predict(pls_model, test_x_test, ncomp = best_ncomp)
```

A quick function for computing the $R^2$ and root mean square error ($RMSE$) of the predictions is defined here for convenience: 

```{r}
reg_metrics <- function(obs, pred, na.rm = TRUE) {
  if (na.rm) {
    ok <- complete.cases(obs, pred)
    obs <- obs[ok]
    pred <- pred[ok]
  }
  c(RMSE = sqrt(mean((obs - pred)^2)), R2 = cor(obs, pred)^2)
}
```

The $RMSE$ and $R^2$ of the globla PLS model:
```{r}
reg_metrics(test_y_test, pred)
```

## The model built with the samples found by _gesearch_ 

The `gesearch()` function is then used to identify a subset of samples from the library that is most relevant to the target set. In the following example, the search is guided by three optimisation criteria: reconstruction error, similarity in latent space, and response prediction error. The parameters used are `k = 50`, meaning that each individual contains 50 active genes or samples; `b = 100`, meaning that each gene is expected to appear in approximately 100 individuals; and `retain = 0.98`, meaning that, for each weakness metric, samples with weakness scores above the 98th quantile are silenced. The search terminates when the gene pool shrinks to `target_size = 80` samples. During evolution, a PLS model with a fixed number of components (`ncomp = 15`) is used for comparability across individuals, but the optimal number of components for the final model is selected by cross-validation independently of this fixed value.

Using the `gesearch_control()` additional aspects of the search can be controlled, such as the strategy for retaining genes based on their weakness scores. In this example, a probability-based retention strategy is used, where genes with weakness scores below the `retain` quantile are kept. This approach is more robust to outliers compared to a fixed proportion-based strategy. Here, tuning the number of factors for each individual is also disbaled for comparability across individuals, but it can be enabled by setting `tune = TRUE` in the control settings. The way the final models (and also the intermediate models during evolution in case `tune = TRUE`) is controlled by the parameters `number` and `p`, which specify the number of groups and the proportion of observations per group for leave-group-out cross-validation, respectively.

**Note** that `tune = FALSE` is recommended for most applications (and it is the default), as it enables a more direct comparison of the performance of different subsets during evolution. This is especially advisable when the search is guided by the `"reconstruction"` and/or `"similarity"` criteria. The reason is that tuning targets only the minimization of the RMSE of the response variable. As a result, models with different numbers of components may be selected during evolution. Models with too few components may reconstruct the target set poorly, which can in turn bias the weakness scores and, consequently, the selection of samples. This may also artificially increase the apparent similarity of some subsets to the target set in latent space. In additon, when `tune = FALSE`, the search process is considerable faster than if that is set to `TRUE`.

```{r}
my_control <- gesearch_control(
  retain_by = "probability",
  tune = FALSE,
  number = 10L,
  p = 0.75
)
```

```{r}
#| eval: true
#| echo: false 
gs <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = test_y_ref, # the available target domain response (accepts NAs)
  k = 50, b = 100, retain = 0.95,
  target_size = 120,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "response"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)
```


```{r}
#| eval: false
gs <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = test_y_ref, # the available target domain response (accepts NAs)
  k = 50, b = 100, retain = 0.95,
  target_size = 120,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "response"),
  control = my_control,
  seed = 1124
)
```

@fig-evolution-gs presents the evolution of the weakness scores through the generations of the search. 

```{r}
#| eval: false
plot(gs)
```

```{r}
#| label: fig-evolution-gs
#| echo: false 
#| fig-cap: "Evolution of the three weakness metrics (response, reconstruction and dissimilarity) used in the search."
#| fig-align: "center"
#| fig-width: 8.5
#| fig-height: 5.5
plot(gs, main = NA)
```

The final model is validated in two ways. First, it is internally validated on the selected samples using leave-group-out cross-validation. Second, a model fitted only on the selected samples is applied to the target samples that drove the search and for which response values are available (`Yu`).

The validation results can be accessed from the final object. For the internal leave-group-out cross-validation (CV) carried out on the training set, the following summary metrics are provided: the mean RMSE across CV groups (`rmse_cv`), the standardised RMSE (`st_rmse_cv`), computed as the RMSE divided by the range of the response values (maximum minus minimum), the standard deviation of the RMSE across CV groups (`rmse_sd_cv`), and the mean (R^2) across CV groups (`r2_cv`).

```{r}
# the internal leave-group out CV results for the data found
gs$validation_results[[1]]$results$train
```


```{r}
# the CV on the target samples that drove the search (note that this is not a proper test set, as these samples were used to guide the search, but it can provide some insight into the performance of the final model on the target domain)
gs$validation_results[[1]]$results$test
```

Assume that the final model is chosen as the one with the lowest CV RMSE obtained from the samples found:

```{r}
best_ncomp <- which.min(gs$validation_results[[1]]$results$train$rmse_cv)
```

The predictions on the samples that represent the target set and that did not participate in the search process:

```{r}
pred_gs <- predict(gs, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs[, best_ncomp])
```

@fig-pred-comparison compares the observed versus predicted values for the model fitted using all available training samples and the model fitted using the subset of samples selected by `gesearch()`.


```{r}
#| label: fig-pred-comparison
#| fig-cap: "Comparison of observed versus predicted total carbon values. Left: model fitted using all available training samples. Right: model fitted using the subset of samples selected by gesearch()."
#| fig-align: "center"
#| fig-width: 7.5
#| fig-height: 4.5
op <- par(mfrow = c(1, 2), mar = c(4.5, 4.5, 3, 1))

rng <- range(pred, pred_gs, test_y_test, na.rm = TRUE)

plot(
  pred, test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with all training \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

plot(
  pred_gs[, best_ncomp], test_y_test,
  xlim = rng,
  ylim = rng,
  xlab = "Predicted Total Carbon, %",
  ylab = "Observed Total Carbon, %",
  main = "Model with gesearch-selected \nsamples",
  cex = 1.5,
  pch = 16,
  col = rgb(0.5, 0.5, 0.5, 0.6)
)
grid(lty = 1)
abline(0, 1, col = "red")

par(op)
```




## Further examples of _gesearch_ 

### No response values available for the target set

The search process can be conducted when no information on the response variable is available for the target set:

```{r}
#| eval: true
#| echo: false 
gs_nr <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)
```


```{r}
#| eval: false
gs_nr <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity"),
  control = my_control,
  seed = 1124
)
```


```{r}
best_ncomp_nr <- which.min(gs_nr$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr
pred_gs_nr <- predict(gs_nr, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr[, best_ncomp_nr])
```


### No response values in the target set but with information about the response range

In some cases, the response values for the target set may not be available, but the range of possible response values may be known. This information can be used to guide the search by penalizing predictions that fall outside the specified range. The `Yu_lims` argument can be used for this purpose, and the `"range"` criterion must be included in the `optimization` argument. The `"range"` option penalizes samples that tend to generate models that predict values for the target spectra that fall outside the specified limits.

```{r}
#| eval: true
#| echo: false 
gs_nr2 <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  Yu_lims = c(0, 5), # the response range in the target domain
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "range"),
  control = my_control,
  seed = 1124, 
  verbose = FALSE
)
```


```{r}
#| eval: false
gs_nr2 <- gesearch(
  Xr = train_x[!is.na(train_y), ], # the spectra of the reference "set/library"
  Yr = train_y[!is.na(train_y)],   # the response of the reference "set/library"
  Xu = test_x_ref, # the available target domain spectra
  Yu = NULL, # NO available target domain response (accepts NAs)
  Yu_lims = c(0, 5), # the response range in the target domain
  k = 50, b = 80, retain = 0.98,
  target_size = 160,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "range"),
  control = my_control,
  seed = 1124
)
```


```{r}
best_ncomp_nr2 <- which.min(gs_nr2$validation_results[[1]]$results$train$rmse_cv)
best_ncomp_nr2
pred_gs_nr2 <- predict(gs_nr2, test_x_test)[[1]]

reg_metrics(test_y_test, pred_gs_nr2[, best_ncomp_nr2])
```

# Supported parallel processing 

The `gesearch()` function supports parallel execution for evaluating individuals during the evolutionary search. To improve memory efficiency, the data are not sent in full to each worker. Instead, only the subsets of samples required for the individuals being evaluated are extracted and passed to the parallel workers. The `pchunks` argument controls how many individuals are processed together at each iterator step during parallel execution. Its effect on speed and memory usage depends on the size of the subsets and the available computing resources. By default, `pchunks = 1L`, meaning that individuals are processed one at a time. Increasing `pchunks` can reduce parallelisation overhead, but it may also increase memory usage because more data may need to be sent to each worker at once. The optimal value depends on the dataset and computational environment, and may require some experimentation.

::: {.callout-note}
`gesearch()` supports parallel execution via the `foreach` and `doParallel`
packages. Parallel execution is most beneficial when the number of individuals
in the population is large, as each individual is evaluated independently
across iterations of the evolutionary search. For small populations or short
runs the overhead of spawning worker processes and serialising data between
them can outweigh the computational savings, making sequential execution
faster. When in doubt, benchmark both before committing to a parallel workflow.
:::

In the example below, the [doParallel package](https://CRAN.R-project.org/package=doParallel) is used to register the cores for parallel execution. The [doSNOW package](https://CRAN.R-project.org/package=doSNOW) can also be used. The example illustrates how to run `gesearch()` using multiple cores. The example may not be faster than sequential execution:

```{r}
#| label: parallel-gesearch
#| eval: false
# Running gesearch() using multiple cores

# Execute with two cores, if available, ...
n_cores <- 2

# ... if not then go with 1 core
if (parallel::detectCores() < 2) {
  n_cores <- 1
}

# Set the number of cores
library(doParallel)
clust <- makeCluster(n_cores)
registerDoParallel(clust)

# Alternatively:
# library(doSNOW)
# clust <- makeCluster(n_cores, type = "SOCK")
# registerDoSNOW(clust)
# getDoParWorkers()

my_control <- gesearch_control(
  retain_by = "probability",
  tune = FALSE,
  number = 10L,
  p = 0.75
)

gs_p <- gesearch(
  Xr = train_x[!is.na(train_y), ],
  Yr = train_y[!is.na(train_y)],
  Xu = test_x_ref,
  Yu = test_y_ref,
  k = 50,
  b = 100,
  retain = 0.98,
  target_size = 80,
  fit_method = fit_pls(ncomp = 20, method = "simpls", scale = TRUE),
  optimization = c("reconstruction", "similarity", "response"),
  control = my_control,
  seed = 1124,
  pchunks = 2L
)

# Go back to sequential processing
registerDoSEQ()
try(stopCluster(clust))

gs_p
```

Parallel processing is particularly useful when the search requires fitting a large number of intermediate models per generation. However, the computational gain depends on the number of available cores, the size of the population, and the cost of fitting each individual model.


# References {-} 

---
title: "Dimensionality reduction of spectral data"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    toc-location: left
    number-sections: true
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
vignette: >
  %\VignetteIndexEntry{2 Dimensionality reduction of spectral data}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
Sys.setenv(OMP_NUM_THREADS = 2)
Sys.setenv(OMP_THREAD_LIMIT = 2)
options(mc.cores = 2)
```


:::: {.columns}
::: {.column width="70%"}
> *Probability mass concentrates near regions that have a much smaller dimensionality than the original space where the data lives* -- [@bengio2013representation]
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::

# Introduction

Exploratory analysis of spectral data is often affected by the curse of dimensionality. In near-infrared (NIR) spectroscopy, for example, a single spectrum may contain measurements at hundreds to thousands of highly correlated wavelengths. Tasks such as pattern detection, similarity assessment, and outlier detection therefore often require reducing the dimensionality of the spectra while retaining the most relevant information.

Principal component analysis (PCA) and partial least squares (PLS) are among the most widely used methods for dimensionality reduction in spectroscopic analysis. Both methods are based on the idea that the meaningful structure of spectral data lies in a lower-dimensional subspace. They seek to construct a projection that transforms the original variables into a smaller set of latent variables that provides a more compact representation of the data while preserving its dominant structure.

The main difference between PCA and PLS lies in the criterion used to define the latent variables. In PCA, the objective is to obtain orthogonal components that explain as much variation in the predictor matrix as possible. In PLS, the objective is to obtain latent variables that maximize the covariance between the predictor matrix and one or more external variables, such as response variables or side information.

In PCA and PLS, the input spectra ($X$, of dimensions $n \times d$) are decomposed into a score matrix ($T$) and a loading matrix ($P$), such that

$$
X = TP^\top + E
$$

where the dimensions of $T$ and $P$ are $n \times o$ and $d \times o$, respectively, $o$ is the number of retained components, and $E$ is the residual matrix (reconstruction error). In PCA, the columns of $P$ are orthonormal ($P^\top P = I$), which allows new observations to be projected directly onto the latent space. For centered data, the maximum number of components that can be retained is $\min(n - 1, d)$.

Once a projection model has been estimated, new observations can be projected onto the same latent space. If $X_{\mathrm{new}}$ denotes a matrix of new spectra preprocessed in the same way as the original data, their scores can be obtained as

$$
T_{\mathrm{new}} = X_{\mathrm{new}} P
$$

where $P$ contains the loading vectors defining the projection space.

The `resemble` package provides the function `ortho_projection()` to compute orthogonal projections of spectral data using PCA- and PLS-based methods. These projections can be used for exploratory analysis, visualization, dissimilarity analysis, neighbor search, and predictive modeling.





# Methods available in the `resemble` package

In the `resemble` package, orthogonal projections based on PCA and PLS are computed with the `ortho_projection()` function.

The function provides the following algorithms for PCA:

* `"pca"`: standard PCA computed by singular value decomposition (SVD).  

* `"pca_nipals"`: PCA computed with the nonlinear iterative partial least squares (NIPALS) algorithm [@wold1975soft].

For PLS, the available algorithms are:

* `"pls"`: standard partial least squares decomposition computed with the NIPALS algorithm. The projection is guided by external variables provided in `Yr`, so that the latent variables maximize the covariance between the spectral predictors and the side information.

* `"mpls"`: modified partial least squares, also computed with the NIPALS algorithm. Unlike standard PLS, which derives weights from covariances, this method uses correlations between `Xr` and `Yr`, reducing the influence of predictor variance on the projection [@shenk1991populations].

* `"simpls"`: partial least squares computed with the SIMPLS algorithm [@de1993simpls], which is often computationally more efficient than NIPALS for high-dimensional data.



# Selecting the number of components

One of the critical decisions in dimensionality reduction is determining how many latent variables to extract. Retaining too few components risks discarding meaningful information, while retaining too many may introduce noise or lead to overfitting in subsequent modeling steps.

The `resemble` package provides a set of constructor functions for specifying component selection criteria. These functions return specification objects that are passed to the `ncomp` argument of `ortho_projection()` and related functions:

| Function | Criterion |
|----------|-----------|
| `ncomp_by_var()` | Individual variance threshold |
| `ncomp_by_cumvar()` | Cumulative variance threshold |
| `ncomp_by_opc()` | Optimal selection via nearest-neighbor evaluation |
| `ncomp_fixed()` | Fixed number of components |

Alternatively, passing a positive integer directly to `ncomp` is equivalent to using `ncomp_fixed()`.

## Variance-based selection

The simplest approach is to retain components based on the amount of variance they explain.

### Individual variance threshold (`ncomp_by_var()`)

Retains all components that individually explain at least a specified proportion of the total variance:

```{r}
#| message: false
library(resemble)
library(prospectr)

# obtain a numeric vector of the wavelengths at which spectra is recorded 
wavs <- as.numeric(colnames(NIRsoil$spc))

# pre-process the spectra:
# - use detrend
# - use first order derivative
diff_order <- 1
poly_order <- 1
window <- 7

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = diff_order, p = poly_order, w = window
)
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]

test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]
```

In the following examples, the default method (PCA with SVD) is used to compute the projection, but the same component selection criteria can be applied to any of the available methods.


```{r}
#| eval: true
# Retain components that individually explain at least 1% of variance
proj_var <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))
proj_var
```

The above configuration of `ncomp_by_var(0.01)` retained a total of `r proj_var$ncomp` components which are the ones that individually explain at least 1% of the total variance in the data. 


### Cumulative variance threshold (`ncomp_by_cumvar()`)

Retains the minimum number of components needed to reach a specified cumulative proportion of explained variance:

```{r}
#| eval: true
# Retain enough components to explain at least 99% of variance
proj_cumvar <- ortho_projection(train_x, ncomp = ncomp_by_cumvar(0.99))
proj_cumvar
```

The above configuration of `ncomp_by_cumvar(0.99)` retained a total of `r proj_cumvar$ncomp` components which are the minimum number of components needed to explain at least 99% of the total variance in the data.

## Optimal component selection (`ncomp_by_opc()`)

Variance-based criteria consider only the structure of the predictor matrix. However, in supervised contexts, the goal is often to retain components that are relevant for capturing information realted to a response variable. The _optimal component selection_ method ("opc") addresses this by incorporating side information.

The method is based on the assumption that if two observations are similar in the spectral space, they should also be similar in terms of their _side information_ [@ramirez2013spectrum]. For a sequence of retained components $o = 1, 2, \ldots, k$, the function computes a dissimilarity matrix at each step and identifies the nearest spectral neighbor for each observation. The side information values of each observation are then compared with those of its nearest neighbor.

For continuous side information, the root mean squared difference (RMSD) is computed:

$$
j(i) = \mathrm{NN}(x_{r_i}, X_r^{\{-i\}})
$$

$$
\mathrm{RMSD} = \sqrt{\frac{1}{m} \sum_{i=1}^m (y_i - y_{j(i)})^2}
$$

where $j(i)$ is the index of the nearest neighbor of observation $i$ (excluding itself), $y_i$ is the side information value for observation $i$, $y_{j(i)}$ is the side information value for the nearest neighbor, and $m$ is the total number of observations.

For categorical side information, the kappa index is used instead of RMSD.

The optimal number of components is the one that minimizes RMSD (or maximizes kappa). Note that in this case, the side information is represented by the response variable (`Ciso` in the `NIRSoil` dataset):

```{r}
#| eval: true
# Optimal component selection using side information
# using default max_ncomp which is 40
proj_opc <- ortho_projection(train_x, Yr = train_y, ncomp = ncomp_by_opc())
proj_opc
```
Using the OPC method, the optimal number of components is `r proj_opc$ncomp`. Although this may appear large, these components are required to adequately represent similarities among observations in terms of the side information (here, the `Ciso` variable). In this sense, OPC provides a semi-supervised approach for extracting meaningful latent variables, where the side information is used only to determine the optimal number of components.

The "opc" method is particularly useful because it selects latent variables that are genuinely informative with respect to the response variable. By contrast, standard component-selection methods often underestimate the number of components needed to preserve similarities among observations in terms of the response.

Note that this method can take multiple variables as side information. 

By default, up to 40 latent variables are evaluated. If fewer than 40 spectral variables are supplied to `ortho_projection()`, the number of spectral variables is used as the upper limit for the number of latent variables to test.

If fewer than 40 latent variables are to be evaluated, the `ncomp_by_opc()` 
method can be used to set a smaller upper limit. For example:

```{r}
# this needs to be passed to the ncomp argument in ortho_projection()
ncomp_by_opc(max_ncomp = 15)
```


## Fixed number of components

When the number of components is known in advance, it can be specified directly:

```{r}
#| eval: false
# Using ncomp_fixed()
proj_fixed <- ortho_projection(train_x, ncomp = ncomp_fixed(10))

# Equivalent: passing an integer directly
proj_fixed <- ortho_projection(train_x, ncomp = 10)
```

# Dimentionality reduction methods in  `ortho_projection()`

The `ortho_projection()` function computes orthogonal projections based on PCA and PLS.

## Available methods

The function provides the following algorithms for PCA:

* `"pca"`: standard PCA computed by singular value decomposition (SVD).
* `"pca_nipals"`: PCA computed with the nonlinear iterative partial least squares (NIPALS) algorithm [@wold1975soft].

For PLS, the available algorithms are:

* `"pls"`: standard partial least squares decomposition computed with the NIPALS algorithm. The projection is guided by external variables provided in `Yr`, so that the latent variables maximize the covariance between the spectral predictors and the side information.
* `"mpls"`: modified partial least squares, computed with the NIPALS algorithm. Weights are derived from correlations rather than covariances between `Xr` and `Yr`, giving equal influence to all predictors regardless of variance [@shenk1991populations].
* `"simpls"`: partial least squares computed with the SIMPLS algorithm [@de1993simpls], which is often computationally more efficient than NIPALS for high-dimensional data.

## Example: PCA projection

```{r}
#| results: hide
# PCA with default component selection (ncomp_by_var(0.01))
pca_train <- ortho_projection(train_x, method = "pca")
pca_train
```

```{r}
#| label: fig-pca-variance
#| fig-cap: "Individual contribution to the explained variance for each component (left) and cumulative variance explained by the principal components (right)."
#| fig-width: 7
#| fig-height: 3
plot(pca_train)
```

In this dataset, `r pca_train$ncomp` components are required to explain approximately `r round(100 * sum(pca_train$variance$x_var[2,]), 0)`% of the original variance in the spectra (@fig-pca-variance).

The NIPALS algorithm provides an alternative to SVD and can be faster when only a few components are required:

```{r}
#| results: hide
# PCA using NIPALS
pca_nipals_train <- ortho_projection(train_x, method = "pca_nipals")
```

## Example: PLS projection

For PLS methods, side information (`Yr`) must be provided:

```{r}
#| results: hide
# PLS using SIMPLS with 10 components
pls_train <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = 10
)
pls_train
```

Observations with missing values in `Yr` are held out during model fitting, then projected using the resulting projection matrix and pooled with the results.

## Example: Optimal component selection

```{r}
#| results: hide
# PCA with optimal component selection
pca_opc <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40)
)
pca_opc
```

```{r}
#| label: fig-pca-opc
#| fig-cap: "RMSD between observations and their nearest neighbors as a function of the number of principal components."
#| fig-width: 5
#| fig-height: 4
plot(pca_opc, col = "#F59E0B")
```

In this example, `r pca_opc$ncomp` components minimize the RMSD for Total Carbon (@fig-pca-opc).

# Projecting new data

Both PCA and PLS generate projection matrices that can be used to project new observations onto the latent space. Use the `predict()` method:

```{r}
#| results: hide
# Fit PLS projection model
pls_model <- ortho_projection(
  train_x,
  Yr = train_y,
  method = "simpls",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

# Project test data
pls_test_scores <- predict(pls_model, newdata = test_x)
```

```{r}
#| label: fig-pc-proj
#| fig-width: 4.5
#| fig-height: 5
#| fig-cap: "Projection of the training observations in the first two PLS components (blue) and projection of the test observations onto the same latent space (red)."
plot(
  pls_model$scores[, 1:2], 
  col = rgb(0.231, 0.51, 0.965, 0.3), 
  pch = 16, cex = 1.5,
  xlab = "PLS1", ylab = "PLS2"
)
grid(lty = 1)
points(
  pls_test_scores[, 1:2], 
  col = rgb(0.961, 0.620, 0.043, 0.7), 
  pch = 16, cex = 1.5
)
legend(
  "topright",
  legend = c("Training projection", "Predicted projection"),
  col = c(
    rgb(0.231, 0.51, 0.965, 0.3),
    rgb(0.961, 0.620, 0.043, 0.7)
  ),
  pch = 16,
  pt.cex = 1.5,
  box.col = "black"
)
```



Alternatively, training and test data can be projected in a single call by passing the test data to `Xu`:

```{r}
#| results: hide
# Project training and test data simultaneously
pca_both <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = train_y,
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)
```

For PCA, the training and test data are pooled, projected jointly, and then split for the final results. For `ncomp_by_opc()`, dissimilarity matrices are computed only on the training data with available side information.

# Multivariate side information

The `ortho_projection()` function accepts multiple variables as side information:

```{r}
#| results: hide
#| eval: false
train_y2 <- NIRsoil$Nt[NIRsoil$train == 1]
train_y3 <- NIRsoil$CEC[NIRsoil$train == 1]

# PLS with multivariate side information
pls_multi <- ortho_projection(
  train_x,
  Xu = test_x,
  Yr = cbind(train_y, train_y2, train_y3),
  method = "pca",
  ncomp = ncomp_by_opc(40),
  scale = TRUE
)

plot(pls_multi)
```

When multiple variables are provided, the PLS2 algorithm is used [@wold1983multivariate]. For optimal component selection, the RMSD is computed for each variable, standardized, and averaged. The evaluation results for each variable are accessible via `$opc_evaluation`.

For categorical side information, the kappa index is used instead of RMSD.  

# References {-}

---
title: "Simple global models"
author: 
 - name: Leonardo Ramirez-Lopez
   email: ramirez.lopez.leo@gmail.com
date: today
bibliography: resemble.bib
csl: elsevier-harvard.csl  
format:
  html:
    toc: true
    toc-depth: 3
    number-sections: true
    toc-location: left
    code-overflow: wrap
    smooth-scroll: true
    html-math-method: mathjax
knitr:
  opts_chunk:
    class-output: "ansi"
vignette: >
  %\VignetteIndexEntry{5 Simple global models}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{quarto::html}
---

```{r}
#| echo: false
options(cli.num_colors = 1)
Sys.setenv(OMP_NUM_THREADS = 2)
```

:::: {.columns}
::: {.column width="70%"}
> *It is vain to do with more what can be done with less* -- William of Ockham
:::
::: {.column width="30%"}
<img src="logo.png" style="width: 150%; max-width: 150px;">
:::
::::


```{r}
#| message: false
#| echo: false
library(resemble)
```

# Introduction

The `model()` function provides a simple interface for building global calibration models using partial least squares (PLS) or Gaussian process regression (GPR). Unlike `mbl()`, which builds local models for each prediction, `model()` fits a single model to the entire reference set.

# Fitting methods

Regression methods are specified via constructor functions. The available methods are:

| Constructor | Method | Description | Reference |
|-------------|--------|-------------|-----------|
| `fit_pls(method = "pls")` | PLS | Non-linear iterative partial least squares (NIPALS) | @wold1975soft |
| `fit_pls(method = "simpls")` | SIMPLS | Straightforward implementation of PLS | @de1993simpls |
| `fit_pls(method = "mpls")` | Modified PLS | Uses correlation instead of covariance for weights | @shenk1991populations |
| `fit_wapls(method = "mpls")` | Weighted average PLS | Predictions are a weighted average of multiple models from multiple PLS factors. __NOT YET AVAIABLE FOR THE__ `model()` __FUNCTION__ | @shenk1997investigation |
| `fit_gpr()` | GPR | Gaussian process regression with dot product kernel | @rasmussen2006gaussian |

```{r}
#| label: fit-methods
# PLS with 15 components
fit_pls(ncomp = 15)

# SIMPLS with scaling
fit_pls(ncomp = 15, method = "simpls", scale = TRUE)

# mPLS with scaling
fit_pls(ncomp = 15, method = "mpls")

# GPR with default noise variance
fit_gpr()
```

# Example

## Data preparation

```{r}
#| label: data-prep
#| message: false
library(prospectr)

wavs <- as.numeric(colnames(NIRsoil$spc))

# Preprocess: detrend + first derivative
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = wavs),
  m = 1, p = 1, w = 7
)

# Split data
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

# Remove missing values
ok_train <- !is.na(train_y)
ok_test  <- !is.na(test_y)
train_x <- train_x[ok_train, ]
train_y <- train_y[ok_train]
test_x  <- test_x[ok_test, ]
test_y  <- test_y[ok_test]
```

## Fitting a PLS model

```{r}
#| label: fit-pls

set.seed(1124) # guarantee same CV splits for all methods
pls_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_pls(ncomp = 12, method = "mpls", scale = TRUE),
 control = model_control(validation_type = "lgo", number = 10)
)

pls_mod
```

## Fitting a GPR model

```{r}
#| label: fit-gpr
set.seed(1124) # guarantee same CV splits for all methods
gpr_mod <- model(
 Xr = train_x,
 Yr = train_y,
 fit_method = fit_gpr(noise_variance = 0.5),
 control = model_control(validation_type = "lgo", number = 10)
)

gpr_mod
```

## Prediction

```{r}
#| label: predict
# Predict using optimal number of components (from CV)
pls_pred <- predict(pls_mod, newdata = test_x)

# Predict with GPR
gpr_pred <- predict(gpr_mod, newdata = test_x)

# Compare predictions
data.frame(
  observed = test_y,
  pls = pls_pred[, which(pls_mod$cv_results$optimal)],
  gpr = as.vector(gpr_pred)
) |> head(10)
```

## Validation statistics

```{r}
#| label: validation
# Function to compute stats
eval_pred <- function(obs, pred) {
  data.frame(
    rmse = sqrt(mean((obs - pred)^2)),
    r2 = cor(obs, pred)^2
  )
}
```

```{r}
## PLS evaluation
eval_pred(test_y, pls_pred[, which(pls_mod$cv_results$optimal)])
```

```{r}
## GPR evaluation
eval_pred(test_y, gpr_pred)
```

# References {-}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ncomp_by_methods.R
\name{ncomp_selection}
\alias{ncomp_selection}
\alias{ncomp_by_var}
\alias{ncomp_by_cumvar}
\alias{ncomp_by_opc}
\alias{ncomp_fixed}
\title{Component selection methods}
\usage{
ncomp_by_var(min_var = 0.01, max_ncomp = 40L)

ncomp_by_cumvar(min_cumvar = 0.99, max_ncomp = 40L)

ncomp_by_opc(max_ncomp = 40L)

ncomp_fixed(ncomp)
}
\arguments{
\item{min_var}{Numeric in (0, 1]. Minimum variance a single component must
explain to be retained.}

\item{max_ncomp}{Positive integer. Maximum number of components to
compute or evaluate.}

\item{min_cumvar}{Numeric in (0, 1]. Minimum cumulative variance that the
retained components must explain.}

\item{ncomp}{Positive integer. Exact number of components to use.}
}
\value{
An object of class \code{"ncomp_selection"} with a subclass indicating the
method:
\itemize{
\item \code{ncomp_by_var}: class \code{c("ncomp_by_var", "ncomp_selection")}
\item \code{ncomp_by_cumvar}: class \code{c("ncomp_by_cumvar", "ncomp_selection")}
\item \code{ncomp_by_opc}: class \code{c("ncomp_by_opc", "ncomp_selection")}
\item \code{ncomp_fixed}: class \code{c("ncomp_fixed", "ncomp_selection")}
}
}
\description{
Constructor functions for specifying how to select the number of components
in projection-based dissimilarity methods (\code{\link[=diss_pca]{diss_pca()}}, \code{\link[=diss_pls]{diss_pls()}}).
}
\details{
Four selection methods are available:

\describe{
\item{\code{ncomp_by_var()}}{Retains components that individually explain at
least \code{min_var} proportion of variance.}
\item{\code{ncomp_by_cumvar()}}{Retains the minimum number of components whose
combined explained variance reaches \code{min_cumvar}.
}
\item{\code{ncomp_by_opc()}}{Optimized principal component selection based on
side information (Ramirez-Lopez et al., 2013). The optimal number of
components minimizes the RMSD between each observation's response and
its nearest neighbor's response in the projected space. Requires \code{Yr}.}
\item{\code{ncomp_fixed()}}{Uses exactly \code{ncomp} components with no automatic
selection. Equivalent to passing an integer directly.}
}

At runtime, \code{max_ncomp} is capped at \code{min(max_ncomp, nrow(X), ncol(X))}.
}
\examples{
# Retain components explaining >= 1\% variance each
ncomp_by_var(0.01)

# Retain enough components for 99\% cumulative variance
ncomp_by_cumvar(0.99)

# Optimize using side information (requires Yr)
ncomp_by_opc(max_ncomp = 40)

# Fix at exactly 10 components
ncomp_fixed(10)

# Usage in dissimilarity constructors
diss_pca(ncomp = ncomp_by_var(0.01))
diss_pca(ncomp = ncomp_by_opc())
diss_pca(ncomp = 10)  

}
\references{
Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
268-279.
}
\seealso{
\code{\link[=diss_pca]{diss_pca()}}, \code{\link[=diss_pls]{diss_pls()}}, \code{\link[=dissimilarity]{dissimilarity()}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/neighbors_methods.R
\name{neighbors}
\alias{neighbors}
\alias{neighbors_k}
\alias{neighbors_diss}
\title{Neighbor selection methods}
\usage{
neighbors_k(k)

neighbors_diss(threshold, k_min = 4L, k_max = Inf)
}
\arguments{
\item{k}{Integer vector. One or more neighborhood sizes to evaluate.
Values will be sorted in ascending order. Minimum allowed value is 4.}

\item{threshold}{Numeric vector. One or more dissimilarity thresholds.
Neighbors are selected if their dissimilarity to the target observation
is below the threshold. Values will be sorted in ascending order.}

\item{k_min}{Integer. Minimum number of neighbors to retain, regardless of
threshold. Default \code{4L}.}

\item{k_max}{Integer or \code{Inf}. Maximum number of neighbors to retain,
regardless of threshold. Default \code{Inf} (no upper bound other than
the size of the reference set).}
}
\value{
An object of class \code{c("neighbors_k", "neighbors")} or
\code{c("neighbors_diss", "neighbors")}, containing the validated
parameters. Intended to be passed to \code{\link{mbl}}.
}
\description{
These functions create configuration objects that specify how neighbors
are selected for memory-based learning in \code{\link{mbl}} an \code{\link{liblex}}.
}
\details{
Two strategies are available for neighbor selection:

\strong{Fixed-k selection} (\code{neighbors_k})

A fixed number of nearest neighbors is selected for each target observation.
Multiple values of \code{k} can be provided to evaluate different
neighborhood sizes.

\strong{Dissimilarity-threshold selection} (\code{neighbors_diss})

Neighbors are selected based on a dissimilarity threshold. All reference
observations with dissimilarity below the threshold are included. The
\code{k_min} and \code{k_max} arguments provide bounds to ensure a reasonable
neighborhood size regardless of the threshold. Multiple thresholds can be
provided to evaluate different settings.
}
\examples{
# Fixed neighborhood sizes
neighbors_k(k = 50)
neighbors_k(k = c(40, 60, 80, 100, 120))

# Dissimilarity threshold with default bounds
neighbors_diss(threshold = 0.3)

# Dissimilarity threshold with custom bounds
neighbors_diss(threshold = c(0.1, 0.2, 0.3), k_min = 10, k_max = 150)

}
\seealso{
\code{\link{mbl}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/resemble.R
\docType{package}
\name{resemble-package}
\alias{resemble-package}
\alias{resemble}
\title{Overview of the functions in the resemble package}
\description{
\ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#maturing'><img src='figures/lifecycle-maturing.svg' alt='Maturing lifecycle'></a>}}{\strong{Maturing}}

Functions for dissimilarity assessment, nearest-neighbour search,
memory-based learning, local expert libraries, and evolutionary training subset
search in spectral chemometrics.

\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}
}
\details{
This is the version
3.0.0 -- vortex
of the package. It implements a number of functions useful for
modeling complex spectral spectra (e.g. NIR, IR).
The package includes functions for dimensionality reduction,
computing spectral dissimilarity matrices, nearest neighbor search,
and modeling spectral data using memory-based learning and evolutionary
search of optimal training subsets in large and complex datasets. This package builds
upon the methods presented in
Ramirez-Lopez et al. (2013a) \doi{10.1016/j.geoderma.2012.12.014},
Ramirez-Lopez et al. (2026a) and Ramirez-Lopez et al. (2026b).

Development versions can be found in the github repository of the package
at \href{https://github.com/l-ramirez-lopez/resemble}{https://github.com/l-ramirez-lopez/resemble}.

The functions available for computing dissimilarity matrices are:
\itemize{
\item{\code{\link{dissimilarity}}}: Computes a dissimilarity matrix based on a specified method.
\item{\code{\link{diss_pca}}}: constructor for principal components-based dissimilarity method.
\item{\code{\link{diss_pls}}}: constructor for partial least squares-based dissimilarity method.
\item{\code{\link{diss_correlation}}}: constructor for correlation-based dissimilarity method.
\item{\code{\link{diss_euclidean}}}: constructor for euclidean distance-based dissimilarity method.
\item{\code{\link{diss_mahalanobis}}}: constructor for Mahalanobis distance-based dissimilarity method.
\item{\code{\link{diss_cosine}}}: constructor for cosine-based dissimilarity method.
}
The functions available for evaluating dissimilarity matrices are:
\itemize{
\item{\code{\link{diss_evaluate}}}: Evaluates the effectiveness of a dissimilarity matrix using side information.
}
The functions available for nearest neighbor search:
\itemize{
\item{\code{\link{search_neighbors}}}: Search for nearest neighbors of a query spectrum in a reference dataset based on a specified dissimilarity method.
}
The functions available for modeling spectral data:
\itemize{
\item \code{\link{mbl}}: Memory-based learning for modeling spectral data.
\item \code{\link{gesearch}}: An evolutionary method to search optimal samples in large spectral datasets.
\item \code{\link{liblex}}: Builds a library of reusable localized models.
}
The functions available for dimensionality reduction are:
\itemize{
\item{\code{\link{ortho_projection}}}: Computes an orthogonal projection of the data based on either principal components or partial least squares.
\item{\code{\link{predict.ortho_projection}}}
}
}
\references{
Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
libraries are too complex to search: Evolutionary subset selection for
domain-adaptive calibration. \emph{Analytica Chimica Acta}, under review.

Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C.,
Perez-Fernandez, E., Plans, M., Breure, T., Behrens, T.,
Viscarra Rossel, R., & Peng, Y. (2026b). Rethinking local spectral
modelling: From per-query refitting to model libraries.
\emph{Analytica Chimica Acta}, under review.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. (2013a). The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. \emph{Geoderma} 195-196,
268-279.
}
\seealso{
Useful links:
\itemize{
\item \url{https://github.com/l-ramirez-lopez/resemble}
\item Report bugs at \url{https://github.com/l-ramirez-lopez/resemble/issues}
}
}
\author{
\strong{Maintainer / Creator}: Leonardo Ramirez-Lopez \email{ramirez.lopez.leo@gmail.com}

Authors:
\itemize{
\item Leonardo Ramirez-Lopez (\href{https://orcid.org/0000-0002-5369-5120}{ORCID})
\item Antoine Stevens (\href{https://orcid.org/0000-0002-1588-7519}{ORCID})
\item Claudio Orellano
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ortho_projection.R
\name{ortho_projection}
\alias{ortho_projection}
\alias{predict.ortho_projection}
\alias{plot.ortho_projection}
\title{Orthogonal projections}
\usage{
ortho_projection(
  Xr, Xu = NULL, Yr = NULL,
  ncomp = ncomp_by_var(0.01),
  method = c("pca", "pca_nipals", "pls", "mpls", "simpls"),
  center = TRUE,
  scale = FALSE,
  tol = 1e-6,
  max_iter = 1000L,
  pc_selection = deprecated(),
  ...
)

\method{predict}{ortho_projection}(object, newdata, ...)

\method{plot}{ortho_projection}(x, col = "#3B82F6", ...)

\method{predict}{ortho_projection}(object, newdata, ...)
}
\arguments{
\item{Xr}{A numeric matrix of reference observations (rows) and variables
(columns).}

\item{Xu}{An optional matrix of additional observations to project.}

\item{Yr}{An optional response matrix. Required for PLS methods
(\code{"pls"}, \code{"mpls"}, \code{"simpls"}) and when using
\code{\link{ncomp_by_opc}()}.}

\item{ncomp}{Component selection method. Either:
\itemize{
\item A positive integer (equivalent to \code{ncomp_fixed(n)})
\item An \code{ncomp_selection} object: \code{\link{ncomp_by_var}()},
\code{\link{ncomp_by_cumvar}()}, \code{\link{ncomp_by_opc}()}, or
\code{\link{ncomp_fixed}()}
}
Default is \code{ncomp_by_var(0.01)}.}

\item{method}{A character string specifying the projection method:
\itemize{
\item \code{"pca"}: PCA via singular value decomposition (default)
\item \code{"pca_nipals"}: PCA via NIPALS algorithm
\item \code{"pls"}: PLS via NIPALS algorithm
\item \code{"mpls"}: Modified PLS via NIPALS (Shenk and Westerhaus, 1991)
\item \code{"simpls"}: PLS via SIMPLS algorithm (de Jong, 1993)
}}

\item{center}{A logical indicating whether to center the data. Default is
\code{TRUE}. PLS methods always center internally regardless of this
setting.}

\item{scale}{A logical indicating whether to scale the data to unit
variance. Default is \code{FALSE}.}

\item{tol}{Convergence tolerance for the NIPALS algorithm. Default is
\code{1e-6}. Ignored when \code{method = "simpls"}.}

\item{max_iter}{Maximum number of iterations for NIPALS. Default is
\code{1000}. Ignored when \code{method = "simpls"}.}

\item{pc_selection}{\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}} Use \code{ncomp} instead.}

\item{...}{Additional arguments (currently unused).}

\item{object}{Object of class \code{"ortho_projection"}.}

\item{newdata}{Matrix of new observations to project.}

\item{x}{An object of class \code{ortho_projection} (as returned by
\code{\link{ortho_projection}}).}

\item{col}{Color for the plot elements. Default is \code{"#3B82F6"}.}
}
\value{
An object of class \code{"ortho_projection"} containing:
\itemize{
\item \code{scores}: Matrix of projected scores for \code{Xr} (and \code{Xu}).
\item \code{X_loadings}: Matrix of X loadings.
\item \code{Y_loadings}: Matrix of Y loadings (PLS only).
\item \code{weights}: Matrix of PLS weights (PLS only).
\item \code{projection_mat}: Projection matrix for new data (PLS only).
\item \code{variance}: List with original and explained variance.
\item \code{scores_sd}: Standard deviation of scores.
\item \code{ncomp}: Number of components retained.
\item \code{center}: Centering vector used.
\item \code{scale}: Scaling vector used.
\item \code{method}: Projection method used.
\item \code{ncomp_method}: The value passed to the \code{ncomp} argument.
\item \code{opc_evaluation}: opc optimization results (if applicable).
}
}
\description{
\loadmathjax
Performs orthogonal projections of high-dimensional data matrices using
principal component analysis (PCA) or partial least squares (PLS).
}
\details{
\subsection{PCA methods}{

When \code{method = "pca"}, singular value decomposition factorizes the
data matrix \mjeqn{X}{X} as:

\mjdeqn{X = UDV^{T}}{X = UDV^T}

where \mjeqn{U}{U} and \mjeqn{V}{V} are orthogonal matrices (left and right
singular vectors), and \mjeqn{D}{D} is a diagonal matrix of singular values.
The score matrix is \mjeqn{UD}{UD} and the loadings are \mjeqn{V}{V}.

When \code{method = "pca_nipals"}, the non-linear iterative partial least
squares (NIPALS) algorithm is used instead.
}

\subsection{PLS methods}{

Three PLS variants are available:

\itemize{
\item \code{"pls"}: Standard PLS using the NIPALS algorithm with
covariance-based weights.
\item \code{"mpls"}: Modified PLS using the NIPALS algorithm with
correlation-based weights, giving equal influence to all predictors
regardless of variance (Shenk and Westerhaus, 1991).
\item \code{"simpls"}: SIMPLS algorithm (de Jong, 1993), which deflates
the cross-product matrix rather than X itself. Computationally faster
than NIPALS, especially for wide matrices.
}
}

\subsection{Component selection}{

When \code{\link{ncomp_by_opc}()} is used, component selection minimizes
RMSD (for continuous \code{Yr}) or maximizes kappa (for categorical
\code{Yr}) between observations and their nearest neighbors. See
\code{\link{diss_evaluate}}.
}
}
\examples{
\donttest{
library(prospectr)
data(NIRsoil)

# Preprocess
sg_det <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Split data
train_x <- sg_det[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]
test_x <- sg_det[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]

# PCA with fixed components
proj <- ortho_projection(train_x, ncomp = 5)

plot(proj)

# PCA with variance-based selection
proj <- ortho_projection(train_x, ncomp = ncomp_by_var(0.01))

# PCA with OPC optimization
proj <- ortho_projection(train_x, Xu = test_x, Yr = train_y,
                         ncomp = ncomp_by_opc(40))

#' plot(proj)

# PLS projection (NIPALS)
proj <- ortho_projection(train_x, Xu = test_x, Yr = train_y,
                         method = "pls", ncomp = ncomp_by_opc(40))

# Modified PLS
proj <- ortho_projection(train_x, Yr = train_y,
                         method = "mpls", ncomp = 10)

# SIMPLS (faster for wide matrices)
proj <- ortho_projection(train_x, Yr = train_y,
                         method = "simpls", ncomp = 10)
}
}
\references{
de Jong, S. 1993. SIMPLS: An alternative approach to partial least squares
regression. Chemometrics and Intelligent Laboratory Systems 18:251-263.

Martens, H. 1991. Multivariate calibration. John Wiley & Sons.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199:43-53.

Shenk, J.S., Westerhaus, M.O. 1991. Populations structuring of near
infrared spectra and modified partial least squares regression. Crop
Science 31:1548-1555.
}
\seealso{
\code{\link{ncomp_by_var}}, \code{\link{ncomp_by_opc}},
\code{\link{diss_evaluate}}, \code{\link{mbl}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.ortho_projection.R
\name{print.ortho_projection}
\alias{print.ortho_projection}
\title{Print method for an object of class \code{ortho_projection}}
\usage{
\method{print}{ortho_projection}(x, ...)
}
\arguments{
\item{x}{an object of class \code{ortho_projection} (as returned by the
\code{ortho_projection} function).}

\item{...}{arguments to be passed to methods (not yet functional).}
}
\description{
Prints the contents of an object of class \code{ortho_projection}
}
\author{
Leonardo Ramirez-Lopez
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_control.R
\name{model_control}
\alias{model_control}
\title{Control parameters for global model fitting}
\usage{
model_control(validation_type = c("lgo", "none"), number = 10L, p = 0.75)
}
\arguments{
\item{validation_type}{a character string specifying the validation method:
\itemize{
\item{\code{"lgo"}: Leave-group-out cross-validation. At each iteration,
a proportion \code{p} of observations is retained for training and the
remainder is used for validation. This is repeated \code{number} times.}
\item{\code{"none"}: No cross-validation is performed.}
}}

\item{number}{an integer indicating the number of cross-validation
iterations. Only used when \code{validation_type = "lgo"}. Default is 10.}

\item{p}{a numeric value between 0 and 1 indicating the proportion of
observations to retain for training at each cross-validation iteration.
Only used when \code{validation_type = "lgo"}. Default is 0.75.}
}
\value{
A list of class \code{"model_control"} containing the specified
parameters.
}
\description{
Specifies cross-validation settings for the \code{\link{model}} function.
}
\examples{
# Default settings (leave-group-out CV with 10 iterations)
model_control()

# No cross-validation
model_control(validation_type = "none")

# Custom CV settings
model_control(validation_type = "lgo", number = 20, p = 0.80)

}
\seealso{
\code{\link{model}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gesearch_control.R
\name{gesearch_control}
\alias{gesearch_control}
\title{Control parameters for gesearch}
\usage{
gesearch_control(
  retain_by = c("probability", "proportion"),
  percentile_type = 7L,
  tune = FALSE,
  number = 10L,
  p = 0.75,
  stagnation_limit = 5L,
  allow_parallel = TRUE,
  blas_threads = 1L
)
}
\arguments{
\item{retain_by}{A character string specifying how training observations are
selected at each iteration:
\describe{
\item{\code{"probability"}}{(default) Retains observations with errors
below a percentile estimated from a given probability. More robust to
outliers.}
\item{\code{"proportion"}}{Retains a fixed proportion of observations
with lowest errors.}
}}

\item{percentile_type}{An integer between 1 and 9 specifying the quantile
algorithm when \code{retain_by = "probability"}. Passed to the \code{type}
argument of \code{\link[stats]{quantile}}. Default is 7. Ignored when
\code{retain_by = "proportion"}.}

\item{tune}{A logical indicating whether to tune regression parameters
(e.g., number of PLS components) via cross-validation at each iteration.
Increases computation time substantially. Default is \code{FALSE}.}

\item{number}{An integer specifying the number of groups for leave-group-out
cross-validation. Default is 10. This is used for validating the final
models built the samples found. When \code{tune = TRUE}, this value is also
used in the internal CV for tuning regression parameters at each iteration.}

\item{p}{A numeric value in (0, 1) specifying the proportion of observations
per group in leave-group-out cross-validation. Default is 0.75. When
\code{tune = TRUE}, this value is also used in the internal CV for tuning
regression parameters at each iteration.}

\item{stagnation_limit}{An integer specifying the maximum number of
consecutive iterations with no change in gene pool size before early
termination. Prevents infinite loops when target size cannot be reached.
Default is 5.}

\item{allow_parallel}{A logical indicating whether to enable parallel
processing for internal resampling and calibration. The parallel backend
must be registered by the user. Default is \code{TRUE}.}

\item{blas_threads}{An integer specifying the number of BLAS threads to use
during computation. Default is 1, which avoids multi-threaded OpenBLAS
overhead on Linux. Requires \pkg{RhpcBLASctl}. See Details.}
}
\value{
A list of class \code{"gesearch_control"} containing the specified
parameters.
}
\description{
Creates a control object specifying algorithm parameters for
\code{\link{gesearch}}.
}
\details{
\subsection{Retention strategies}{

When \code{retain_by = "probability"} (default), observations with errors
below a percentile threshold are retained. The percentile is computed using
\code{\link[stats]{quantile}} with \code{probs} set to the \code{retain}
value from \code{\link{gesearch}}. This approach is more robust when outlier
observations have extreme error values.

When \code{retain_by = "proportion"}, a fixed fraction of observations
(specified by the \code{retain} argument in \code{\link{gesearch}}) with the
lowest associated errors are kept at each iteration.
}

\subsection{Cross-validation for tuning}{

When \code{tune = TRUE}, leave-group-out cross-validation is used to select
optimal regression parameters at each iteration. The \code{number} argument
controls how many CV groups are formed, and \code{p} controls the proportion
of observations in each group.
}

\subsection{BLAS threading}{

On Linux systems with multi-threaded OpenBLAS, the default thread count
can cause significant overhead for algorithms that perform many small
matrix operations (like the iterative PLS fits in \code{gesearch}).
Setting \code{blas_threads = 1} (the default) eliminates this overhead.

This setting requires the \pkg{RhpcBLASctl} package. If not installed,
the parameter is ignored and a message is displayed. The original thread
count is restored when \code{\link{gesearch}} completes.
}
}
\examples{
# Default parameters (probability-based retention)
gesearch_control()

# Proportion-based retention
gesearch_control(retain_by = "proportion")

# Enable parameter tuning with custom CV settings
gesearch_control(tune = TRUE, number = 5, p = 0.8)

}
\references{
Lobsey, C.R., Viscarra Rossel, R.A., Roudier, P., Hedley, C.B. 2017.
rs-local data-mines information from spectral libraries to improve local
calibrations. European Journal of Soil Science 68:840-852.
}
\seealso{
\code{\link{gesearch}}, \code{\link[stats]{quantile}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mbl.R
\name{mbl}
\alias{mbl}
\alias{plot.mbl}
\alias{get_predictions}
\title{Memory-based learning (mbl)}
\usage{
mbl(Xr, Yr, Xu, Yu = NULL,
    neighbors,
    diss_method = diss_pca(ncomp = ncomp_by_opc()),
    diss_usage = c("none", "predictors", "weights"),
    fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
    spike = NULL, group = NULL,
    gh = FALSE,
    control = mbl_control(),
    verbose = TRUE, seed = NULL, ...)

\method{plot}{mbl}(x, what = c("validation", "gh"), metric = "rmse", ncomp = c(1, 2), ...)

get_predictions(x)

\method{plot}{mbl}(x, what = c("validation", "gh"), metric = "rmse", ncomp = c(1, 2), ...)
}
\arguments{
\item{Xr}{A matrix of predictor variables for the reference data
(observations in rows, variables in columns). Column names are required.}

\item{Yr}{A numeric vector or single-column matrix of response values
corresponding to \code{Xr}. NA values are not permitted.}

\item{Xu}{A matrix of predictor variables for the data to be predicted
(observations in rows, variables in columns). Must have the same column
names as \code{Xr}.}

\item{Yu}{An optional numeric vector or single-column matrix of response
values corresponding to \code{Xu}. Used for computing prediction
statistics. Default is \code{NULL}.}

\item{neighbors}{A neighbor selection object specifying how to select
neighbors. Use \code{\link{neighbors_k}()} for fixed k-nearest neighbors
or \code{\link{neighbors_diss}()} for dissimilarity threshold-based
selection.}

\item{diss_method}{A dissimilarity method object or a precomputed
dissimilarity matrix. Available constructors:
\itemize{
\item \code{\link{diss_pca}()}: Mahalanobis distance in PCA score space.
This is the default where the number of components is optimized using
side information (see \code{ncomp_by_opc}()).
\item \code{\link{diss_pls}()}: Mahalanobis distance in PLS score space
\item \code{\link{diss_euclidean}()}: Euclidean distance
\item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
\item \code{\link{diss_cosine}()}: Cosine dissimilarity
\item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
}
A precomputed matrix can also be passed. When \code{diss_usage =
  "predictors"}, it must be square with dimensions
\code{(nrow(Xr) + nrow(Xu))} and zeros on the diagonal. Otherwise, it must
have \code{nrow(Xr)} rows and \code{nrow(Xu)} columns.}

\item{diss_usage}{How dissimilarity information is used in local models:
\itemize{
\item \code{"none"} (default): dissimilarities used only for neighbor
selection
\item \code{"predictors"}: local dissimilarity matrix columns added as
predictors
\item \code{"weights"}: neighbors weighted by dissimilarity using a
tricubic function
}}

\item{fit_method}{A local fitting method object. Available constructors:
\itemize{
\item \code{\link{fit_pls}()}: Partial least squares regression
\item \code{\link{fit_wapls}()}: Weighted average PLS (default)
\item \code{\link{fit_gpr}()}: Gaussian process regression
}}

\item{spike}{An integer vector indicating indices of observations in
\code{Xr} to force into (positive values) or exclude from (negative values)
all neighborhoods. Default is \code{NULL}. Spiking does not change
neighborhood size; forced observations displace the most distant neighbors.}

\item{group}{An optional factor assigning group labels to \code{Xr}
observations (e.g., measurement batches). Used to avoid pseudo-replication
in cross-validation: when one observation is held out, all observations
from its group are also removed.}

\item{gh}{Logical indicating whether to compute global Mahalanobis (GH)
distances. Default is \code{FALSE}. GH distances measure how far each
observation lies from the center of the reference set in PLS score space.
The computation uses a fixed methodology: PLS projection with the number
of components selected via \code{\link{ncomp_by_opc}()} (capped at 40).
This is independent of the \code{diss_method} argument.}

\item{control}{A list from \code{\link{mbl_control}()} specifying validation
type, tuning options, and other settings.}

\item{verbose}{Logical indicating whether to display a progress bar.
Default is \code{TRUE}. Not shown during parallel execution.}

\item{seed}{An integer for random number generation, enabling reproducible
cross-validation results. Default is \code{NULL}.}

\item{...}{Additional arguments (currently unused).}

\item{x}{An object of class \code{mbl} (as returned by \code{\link{mbl}}).}

\item{what}{Character vector specifying what to plot. Options are
\code{"validation"} (validation statistics) and/or \code{"gh"} (PLS scores
used for GH distance computation). Default is both.}

\item{metric}{Character string specifying which validation statistic to plot.
Options are \code{"rmse"}, \code{"st_rmse"}, or \code{"r2"}. Only used when
\code{"validation"} is in \code{what}.}

\item{ncomp}{Integer vector of length 1 or 2 specifying which PLS components
to plot. Default is \code{c(1, 2)}. Only used when \code{"gh"} is in
\code{what}.}
}
\value{
\subsection{mbl}{

For \code{mbl()}, a list of class \code{mbl} containing:
\itemize{
\item \code{control}: control parameters from \code{control}
\item \code{fit_method}: fit constructor from \code{fit_method}
\item \code{Xu_neighbors}: list with neighbor indices and dissimilarities
\item \code{dissimilarities}: dissimilarity method and matrix (if
\code{return_dissimilarity = TRUE} in \code{control})
\item \code{n_predictions}: number of predictions made
\item \code{gh}: GH distances for \code{Xr} and \code{Xu} (if
\code{gh = TRUE})
\item \code{validation_results}: validation statistics by method
\item \code{results}: list of data.frame objects with predictions, one per
neighborhood size
\item \code{seed}: the seed value used
}

Each results table contains:
\itemize{
\item \code{o_index}: observation index
\item \code{k}: number of neighbors used
\item \code{k_diss}, \code{k_original}: (\code{neighbors_diss} only)
threshold and original count
\item \code{ncomp}: (\code{fit_pls} only) number of PLS components
\item \code{min_ncomp}, \code{max_ncomp}: (\code{fit_wapls} only)
component range
\item \code{yu_obs}, \code{pred}: observed and predicted values
\item \code{yr_min_obs}, \code{yr_max_obs}: response range in neighborhood
\item \code{index_nearest_in_Xr}, \code{index_farthest_in_Xr}: neighbor
indices
\item \code{y_nearest}, \code{y_farthest}: neighbor response values
\item \code{diss_nearest}, \code{diss_farthest}: neighbor dissimilarities
\item \code{y_nearest_pred}: (NNv validation) leave-one-out prediction
\item \code{loc_rmse_cv}, \code{loc_st_rmse_cv}: (local_cv validation) CV
statistics
\item \code{loc_ncomp}: (local dissimilarity only) components used locally
}
}

\subsection{Get predictions}{

The \code{get_predictions()} function extracts predicted values from an
object of class \code{mbl}. It returns a \code{data.frame} containing the
predictions.
}
}
\description{
\loadmathjax
Memory-based learning (a.k.a. instance-based learning or local regression)
is a non-linear lazy learning approach for predicting a response variable
from predictor variables. For each observation in a prediction set, a local
regression is fitted using a subset of similar observations (nearest
neighbors) from a reference set. This function does not produce a global
model.
}
\details{
\subsection{Spiking}{

The \code{spike} argument forces specific reference observations into or out
of neighborhoods. Positive indices are always included; negative indices are
always excluded. When observations are forced in, the most distant neighbors
are displaced to maintain neighborhood size. See Guerrero et al. (2010).
}

\subsection{Dissimilarity usage}{

When \code{diss_usage = "predictors"}, the local dissimilarity matrix columns
are appended as additional predictor variables, which can improve predictions
(Ramirez-Lopez et al., 2013a).

When \code{diss_usage = "weights"}, neighbors are weighted using a tricubic
function (Cleveland and Devlin, 1988; Naes et al., 1990):

\mjdeqn{W_{j} = (1 - v^{3})^{3}}{W_j = (1 - v^3)^3}

where \mjeqn{v = d(xr_i, xu_j) / \max(d)}{v = d(xr_i, xu_j) / max(d)}.
}

\subsection{GH distance}{

The global Mahalanobis distance (GH) measures how far each observation lies
from the center of the reference set. It is always computed using a PLS
projection with the number of components optimized via
\code{\link{ncomp_by_opc}()} (maximum 40 components or \code{nrow(Xr)},
whichever is smaller). This methodology is fixed and independent of the
\code{diss_method} specified for neighbor selection.

GH distances are useful for identifying extrapolation: observations with
high GH values lie far from the calibration space and may yield unreliable
predictions.
}

\subsection{Grouping}{

The \code{group} argument enables leave-group-out cross-validation. When
\code{validation_type = "local_cv"} in \code{\link{mbl_control}()}, the
\code{p} parameter refers to the proportion of groups (not observations)
retained per iteration.
}

\subsection{Deprecated arguments}{

The following arguments from previous versions of \code{resemble} are no
longer supported and will throw an error if used: \code{k}, \code{k_diss},
\code{k_range}, \code{method}, \code{pc_selection}, \code{center},
\code{scale}, and \code{documentation}. See the current argument list for
their replacements.
}
}
\examples{
\dontrun{
library(prospectr)
data(NIRsoil)

# Preprocess: detrend + first derivative with Savitzky-Golay
sg_det <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)
NIRsoil$spc_pr <- sg_det

# Split data
test_x <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$CEC), ]
test_y <- NIRsoil$CEC[NIRsoil$train == 0 & !is.na(NIRsoil$CEC)]
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$CEC), ]
train_y <- NIRsoil$CEC[NIRsoil$train == 1 & !is.na(NIRsoil$CEC)]

# Example 1: Spectrum-based learner (Ramirez-Lopez et al., 2013)
ctrl <- mbl_control(validation_type = "NNv")

sbl <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  neighbors = neighbors_k(seq(40, 140, by = 20)),
  diss_method = diss_pca(ncomp = ncomp_by_opc(40)),
  fit_method = fit_gpr(),
  control = ctrl
)
sbl
plot(sbl)
get_predictions(sbl)

# Example 2: With known Yu
sbl_2 <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  Yu = test_y,
  neighbors = neighbors_k(seq(40, 140, by = 20)),
  fit_method = fit_gpr(),
  control = ctrl
)
plot(sbl_2)

# Example 3: LOCAL algorithm (Shenk et al., 1997)
local_algo <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  Yu = test_y,
  neighbors = neighbors_k(seq(40, 140, by = 20)),
  diss_method = diss_correlation(),
  diss_usage = "none",
  fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
  control = ctrl
)
plot(local_algo)

# Example 4: Using dissimilarity as predictors
local_algo_2 <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  Yu = test_y,
  neighbors = neighbors_k(seq(40, 140, by = 20)),
  diss_method = diss_pca(ncomp = ncomp_by_opc(40)),
  diss_usage = "predictors",
  fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
  control = ctrl
)
plot(local_algo_2)

# Example 5: Parallel execution
library(doParallel)
n_cores <- min(2, parallel::detectCores() - 1)
clust <- makeCluster(n_cores)
registerDoParallel(clust)

local_algo_par <- mbl(
  Xr = train_x,
  Yr = train_y,
  Xu = test_x,
  Yu = test_y,
  neighbors = neighbors_k(seq(40, 140, by = 20)),
  diss_method = diss_correlation(),
  fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
  control = ctrl
)

registerDoSEQ()
try(stopCluster(clust))
}

}
\references{
Cleveland, W. S., and Devlin, S. J. 1988. Locally weighted regression: an
approach to regression analysis by local fitting. Journal of the American
Statistical Association 83:596-610.

Guerrero, C., Zornoza, R., Gomez, I., Mataix-Beneyto, J. 2010. Spiking of
NIR regional models using observations from target sites: Effect of model
size on prediction accuracy. Geoderma 158:66-77.

Naes, T., Isaksson, T., Kowalski, B. 1990. Locally weighted regression and
scatter correction for near-infrared reflectance data. Analytical Chemistry
62:664-673.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199:43-53.

Rasmussen, C.E., Williams, C.K. 2006. Gaussian Processes for Machine
Learning. MIT Press.

Shenk, J., Westerhaus, M., Berzaghi, P. 1997. Investigation of a LOCAL
calibration procedure for near infrared instruments. Journal of Near
Infrared Spectroscopy 5:223-232.
}
\seealso{
\code{\link{mbl_control}}, \code{\link{neighbors_k}},
\code{\link{neighbors_diss}}, \code{\link{diss_pca}}, \code{\link{diss_pls}},
\code{\link{fit_pls}}, \code{\link{fit_wapls}}, \code{\link{fit_gpr}},
\code{\link{search_neighbors}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} and
Antoine Stevens
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_evaluate.R
\name{diss_evaluate}
\alias{diss_evaluate}
\alias{sim_eval}
\title{Evaluate dissimilarity matrices}
\usage{
diss_evaluate(diss, side_info)

sim_eval(d, side_info)
}
\arguments{
\item{diss}{A symmetric dissimilarity matrix. Alternatively, a vector
containing the lower triangle values (as returned by \code{\link[stats]{dist}}).}

\item{side_info}{A matrix of side information corresponding to the
observations. Can be numeric (one or more columns) or character (single
column for categorical data).}

\item{d}{Deprecated. Use \code{diss} in \code{diss_evaluate()} instead.}
}
\value{
A list with the following components:
\describe{
\item{eval}{For numeric side information: a matrix with columns
\code{rmsd} and \code{r} (correlation). For categorical: a matrix
with column \code{kappa}.}
\item{global_eval}{If multiple numeric side information variables are
provided, summary statistics across variables.}
\item{first_nn}{A matrix with the original side information and the
side information of each observation's nearest neighbor.}
}
}
\description{
\loadmathjax
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#stable}{\figure{lifecycle-stable.svg}{options: alt='[Stable]'}}}{\strong{[Stable]}}

Evaluates a dissimilarity matrix by comparing each observation to its
nearest neighbor based on side information. For continuous variables,
RMSD and correlation are computed; for categorical variables, the kappa
index is used.
}
\details{
This function assesses whether a dissimilarity matrix captures meaningful
structure by examining the side information of nearest neighbor pairs
(Ramirez-Lopez et al., 2013). If observations that are similar in the
dissimilarity space also have similar side information values, the
dissimilarity is considered effective.

For numeric \code{side_info}, the root mean square of differences (RMSD)
between each observation and its nearest neighbor is computed:

\mjdeqn{j(i) = NN(x_i, X^{\{-i\}})}{j(i) = NN(x_i, X^{-i})}
\mjdeqn{RMSD = \sqrt{\frac{1}{m} \sum_{i=1}^{m} (y_i - y_{j(i)})^2}}{RMSD = sqrt(1/m sum (y_i - y_j(i))^2)}

where \mjeqn{NN(x_i, X^{-i})}{NN(x_i, X^{-i})} returns the index of the
nearest neighbor of observation \mjeqn{i}{i} (excluding itself),
\mjeqn{y_i}{y_i} is the side information value for observation \mjeqn{i}{i},
and \mjeqn{m}{m} is the number of observations.

For categorical \code{side_info}, the kappa index is computed:

\mjdeqn{\kappa = \frac{p_o - p_e}{1 - p_e}}{kappa = (p_o - p_e) / (1 - p_e)}

where \mjeqn{p_o}{p_o} is the observed agreement and \mjeqn{p_e}{p_e} is the
agreement expected by chance.
}
\examples{
\donttest{
library(prospectr)
data(NIRsoil)

sg <- savitzkyGolay(NIRsoil$spc, p = 3, w = 11, m = 0)
NIRsoil$spc <- sg

Yr <- NIRsoil$Nt[as.logical(NIRsoil$train)]
Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]

# Compute PCA-based dissimilarity
d <- dissimilarity(Xr, diss_method = diss_pca(ncomp = 8))

# Evaluate using side information
ev <- diss_evaluate(d$dissimilarity, side_info = as.matrix(Yr))
ev$eval

# Evaluate with multiple side information variables
Yr_2 <- NIRsoil$CEC[as.logical(NIRsoil$train)]
ev_2 <- diss_evaluate(d$dissimilarity, side_info = cbind(Yr, Yr_2))
ev_2$eval
ev_2$global_eval
}

}
\references{
Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex datasets. Geoderma 195-196,
268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199, 43-53.
}
\seealso{
\code{\link{dissimilarity}}, \code{\link{ncomp_by_opc}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pkg_info.R
\name{pkg_info}
\alias{pkg_info}
\title{Get the package version info}
\usage{
pkg_info(pkg = "resemble")
}
\arguments{
\item{pkg}{the package name i.e "resemble"}
}
\description{
returns package info.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dissimilarity.R
\name{dissimilarity}
\alias{dissimilarity}
\title{Compute dissimilarity matrices}
\usage{
dissimilarity(Xr, Xu = NULL, diss_method = diss_pca(), Yr = NULL)
}
\arguments{
\item{Xr}{A numeric matrix of reference observations (rows) and variables
(columns).}

\item{Xu}{Optional matrix of additional observations with the same variables.}

\item{diss_method}{A dissimilarity method object created by one of:
\itemize{
\item \code{\link{diss_pca}()}: Mahalanobis distance in PCA space
\item \code{\link{diss_pls}()}: Mahalanobis distance in PLS space
\item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
\item \code{\link{diss_euclidean}()}: Euclidean distance
\item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
\item \code{\link{diss_cosine}()}: Cosine dissimilarity
}
Default is \code{diss_pca()}.}

\item{Yr}{Optional response matrix. Required for PLS methods and when using
\code{ncomp_by_opc()}.}
}
\value{
A list of class \code{"dissimilarity"} containing:
\describe{
\item{dissimilarity}{The computed dissimilarity matrix. Dimensions are
\code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xr)} when \code{Xu = NULL},
or \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xu)} otherwise.}
\item{diss_method}{The \code{diss_*} constructor object used for computation.}
\item{center}{Vector used to center the data.}
\item{scale}{Vector used to scale the data.}
\item{ncomp}{Number of components used (for projection methods).}
\item{projection}{If \code{return_projection = TRUE} in the method
constructor, the \code{ortho_projection} object.}
}
}
\description{
\loadmathjax
Computes dissimilarity matrices between observations using various methods.
This is the main interface for dissimilarity computation in the resemble
package.
}
\details{
The function dispatches to the appropriate internal computation based on the
class of \code{diss_method}. Each method constructor (e.g., \code{diss_pca()})
encapsulates all method-specific parameters including component selection,
centering, scaling, and whether to return projections.

\subsection{Output dimensions}{
When only \code{Xr} is provided, the function computes pairwise dissimilarities
among all observations in \code{Xr}, returning a symmetric
\code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xr)} matrix.

When both \code{Xr} and \code{Xu} are provided, the function computes
dissimilarities between each observation in \code{Xr} and each observation
in \code{Xu}, returning a \code{nrow(Xr)} \mjeqn{\times}{x} \code{nrow(Xu)}
matrix where element \mjeqn{(i, j)}{(i, j)} is the dissimilarity between the
\mjeqn{i}{i}-th observation in \code{Xr} and the \mjeqn{j}{j}-th observation
in \code{Xu}.
}

\subsection{Mahalanobis distance}{
Note that \code{diss_mahalanobis()} computes Mahalanobis distance directly on
the input variables. This requires the covariance matrix to be invertible,
which fails when the number of variables exceeds the number of observations
or when variables are highly correlated (common in spectral data). For such
cases, use \code{diss_pca()} or \code{diss_pls()} instead.
}
}
\examples{
\donttest{
library(prospectr)
data(NIRsoil)

# Preprocess
sg <- savitzkyGolay(NIRsoil$spc, m = 1, p = 4, w = 15)

Xr <- sg[as.logical(NIRsoil$train), ]
Xu <- sg[!as.logical(NIRsoil$train), ]
Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]

Xu <- Xu[!is.na(Yu), ]
Xr <- Xr[!is.na(Yr), ]
Yr <- Yr[!is.na(Yr)]

# PCA-based dissimilarity with variance-based selection
d1 <- dissimilarity(Xr, Xu, diss_method = diss_pca())

# PCA with OPC selection (requires Yr)
d2 <- dissimilarity(Xr, Xu,
  Yr = Yr,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(30),
    return_projection = TRUE
  )
)

# PLS-based dissimilarity 
d3 <- dissimilarity(
  Xr, Xu,
  Yr = Yr,
  diss_method = diss_pls(
    ncomp = ncomp_by_opc(30)
  )
)

# Euclidean distance
d4 <- dissimilarity(Xr, Xu, diss_method = diss_euclidean())

# Correlation dissimilarity with moving window
d5 <- dissimilarity(Xr, Xu, diss_method = diss_correlation(ws = 41))

# Mahalanobis distance (use only when n > p and low collinearity)
# d6 <- dissimilarity(Xr[, 1:20], Xu[, 1:20],
#                     diss_method = diss_mahalanobis())
}

}
\references{
Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199, 43-53.
}
\seealso{
\code{\link{diss_pca}}, \code{\link{diss_pls}},
\code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
\code{\link{diss_mahalanobis}}, \code{\link{diss_cosine}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_neighbors.R
\name{search_neighbors}
\alias{search_neighbors}
\title{Search neighbors in a reference set}
\usage{
search_neighbors(Xr, Xu = NULL,
                 diss_method = diss_pca(), Yr = NULL,
                 neighbors, spike = NULL,
                 return_dissimilarity = FALSE, 
                 k, k_diss, k_range, pc_selection, 
                 center, scale, documentation, ...
                 )
}
\arguments{
\item{Xr}{A numeric matrix of reference observations (rows) and variables
(columns) where the neighbor search is conducted.}

\item{Xu}{Optional matrix of observations for which neighbors are to be
searched in \code{Xr}.}

\item{diss_method}{A dissimilarity method object created by one of:
\itemize{
\item \code{\link{diss_pca}()}: Mahalanobis distance in PCA space
\item \code{\link{diss_pls}()}: Mahalanobis distance in PLS space
\item \code{\link{diss_correlation}()}: Correlation-based dissimilarity
\item \code{\link{diss_euclidean}()}: Euclidean distance
\item \code{\link{diss_mahalanobis}()}: Mahalanobis distance
\item \code{\link{diss_cosine}()}: Cosine dissimilarity
}
Default is \code{diss_pca()}.}

\item{Yr}{Optional response matrix. Required for PLS methods and when using
\code{ncomp_by_opc()}.}

\item{neighbors}{A neighbor selection object created by:
\itemize{
\item \code{\link{neighbors_k}()}: Select k nearest neighbors
\item \code{\link{neighbors_diss}()}: Select neighbors by dissimilarity threshold
}}

\item{spike}{Optional integer vector indicating observations in \code{Xr}
to force into (positive indices) or exclude from (negative indices)
neighborhoods.}

\item{return_dissimilarity}{Logical indicating whether to return the
dissimilarity matrix. Default is \code{FALSE}.}

\item{k}{Deprecated.}

\item{k_diss}{Deprecated.}

\item{k_range}{Deprecated.}

\item{pc_selection}{Deprecated.}

\item{center}{Deprecated.}

\item{scale}{Deprecated.}

\item{documentation}{Deprecated.}

\item{...}{Additional arguments (currently unused).}
}
\value{
A list containing:
\describe{
\item{neighbors}{Matrix of \code{Xr} indices for each query observation's
neighbors, sorted by dissimilarity (columns = query observations).}
\item{neighbors_diss}{Matrix of dissimilarity scores corresponding to
\code{neighbors}.}
\item{unique_neighbors}{Vector of unique \code{Xr} indices that appear
in any neighborhood.}
\item{k_diss_info}{If \code{neighbors_diss()} was used, a \code{data.frame}
with columns for observation index, number of neighbors found, and final
number after applying bounds.}
\item{dissimilarity}{If \code{return_dissimilarity = TRUE}, the full
dissimilarity object.}
\item{projection}{If the dissimilarity method includes
\code{return_projection = TRUE}, the projection object.}
\item{gh}{If the dissimilarity method includes \code{gh = TRUE}, the
GH distances.}
}
}
\description{
\loadmathjax
Searches for the nearest neighbors of observations in a reference set or
between two sets of observations.
}
\details{
This function is useful for reducing large reference sets by identifying
only relevant neighbors before running \code{\link{mbl}}.

If \code{Xu} is not provided, the function searches for neighbors within
\code{Xr} itself (excluding self-matches). If \code{Xu} is provided,
neighbors of each observation in \code{Xu} are searched in \code{Xr}.

The \code{spike} argument allows forcing specific observations into or out
of all neighborhoods. Positive indices are always included; negative indices
are always excluded.
}
\examples{
\donttest{
library(prospectr)
data(NIRsoil)

Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]

Xu <- Xu[!is.na(Yu), ]
Yu <- Yu[!is.na(Yu)]
Xr <- Xr[!is.na(Yr), ]
Yr <- Yr[!is.na(Yr)]

# Correlation-based neighbor search with k neighbors
ex1 <- search_neighbors(
  Xr = Xr, Xu = Xu,
  diss_method = diss_correlation(),
  neighbors = neighbors_k(40)
)

# PCA-based with OPC selection
ex2 <- search_neighbors(
  Xr = Xr, Xu = Xu,
  diss_method = diss_pca(
    ncomp = ncomp_by_opc(40),
    scale = TRUE,
    return_projection = TRUE
  ),
  Yr = Yr,
  neighbors = neighbors_k(50)
)

# Observations not in any neighborhood
setdiff(seq_len(nrow(Xr)), ex2$unique_neighbors)

# Dissimilarity threshold-based selection
ex3 <- search_neighbors(
  Xr = Xr, Xu = Xu,
  diss_method = diss_pls(
    ncomp = ncomp_by_opc(40),
    scale = TRUE
  ),
  Yr = Yr,
  neighbors = neighbors_diss(threshold = 0.5, k_min = 10, k_max = 100)
)
}

}
\references{
Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196,
268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199, 43-53.
}
\seealso{
\code{\link{dissimilarity}}, \code{\link{mbl}},
\code{\link{neighbors_k}}, \code{\link{neighbors_diss}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fit_methods.R
\name{fit_methods}
\alias{fit_methods}
\alias{fit_pls}
\alias{fit_wapls}
\alias{fit_gpr}
\title{Local fitting method constructors}
\usage{
fit_pls(ncomp, method = c("pls", "mpls", "simpls"), 
        scale = FALSE, max_iter = 100L, tol = 1e-6)

fit_wapls(min_ncomp, max_ncomp, method = c("mpls", "pls", "simpls"),
          scale = FALSE, max_iter = 100L, tol = 1e-6)

fit_gpr(noise_variance = 0.001, center = TRUE, scale = TRUE)
}
\arguments{
\item{ncomp}{an integer indicating the number of PLS components to use
in local regressions when \code{fit_pls} is used.}

\item{min_ncomp}{an integer indicating the minimum number of PLS components
to use in local regressions when \code{fit_wapls} is used. See details.}

\item{max_ncomp}{an integer indicating the maximum number of PLS components
to use in local regressions when \code{fit_wapls} is used. See details.}

\item{method}{a character string indicating the PLS algorithm to use.
Options are:
\itemize{
\item{\code{'pls'}: standard PLS using covariance between X and Y for
weight computation (NIPALS algorithm).}
\item{\code{'mpls'}: modified PLS using correlation between X and Y for
weight computation (NIPALS algorithm). See Shenk and Westerhaus (1991).}
\item{\code{'simpls'}: SIMPLS algorithm (de Jong, 1993). Computationally
faster as it avoids iterative X deflation. Parameters \code{max_iter}
and \code{tol} are ignored when this method is used.}
}
Default is \code{'pls'} for \code{fit_pls} and \code{'mpls'} for
\code{fit_wapls}.}

\item{scale}{logical indicating whether predictors must be scaled.
Default is \code{FALSE} for PLS methods and \code{TRUE} for GPR.}

\item{max_iter}{an integer indicating the maximum number of iterations
for convergence in the NIPALS algorithm. Only used when
\code{method = 'pls'} or \code{method = 'mpls'}. Default is 100.}

\item{tol}{a numeric value indicating the convergence tolerance for
calculating scores in the NIPALS algorithm. Only used when
\code{method = 'pls'} or \code{method = 'mpls'}. Default is 1e-6.}

\item{noise_variance}{a numeric value indicating the variance of the noise
for Gaussian process local regressions (\code{fit_gpr}). Default is 0.001.}

\item{center}{logical indicating whether predictors should be centered
before fitting. Only used for \code{fit_gpr}. Default is \code{TRUE}.}
}
\value{
An object of class \code{c("fit_<method>", "fit_method")}
containing the specified parameters. This object is passed to
\code{\link{mbl}} to configure local model fitting.
}
\description{
\loadmathjax

These functions create configuration objects that specify how local
regression models are fitted within the \code{\link{mbl}} function.
}
\details{
These functions create configuration objects that are passed to
\code{\link{mbl}} to specify how local regression models are fitted.

There are three fitting methods available:

\subsection{Partial least squares (\code{fit_pls})}{
Uses orthogonal scores partial least squares regression. Three algorithm
variants are available:
\itemize{
\item{\strong{Standard PLS} (\code{method = 'pls'}): Uses the NIPALS
algorithm with covariance-based weights.}
\item{\strong{Modified PLS} (\code{method = 'mpls'}): Uses the NIPALS
algorithm with correlation-based weights. Proposed by Shenk and
Westerhaus (1991), this approach gives equal influence to all
predictors regardless of their variance scale.}
\item{\strong{SIMPLS} (\code{method = 'simpls'}): Uses the SIMPLS
algorithm (de Jong, 1993), which deflates the cross-product matrix
rather than X itself. This is computationally faster, especially for
wide matrices, and produces identical predictions to standard PLS.}
}
The only parameter to optimise is the number of PLS components
(\code{ncomp}).
}

\subsection{Weighted average PLS (\code{fit_wapls})}{
This method was developed by Shenk et al. (1997) and is used as the
regression method in the LOCAL algorithm. It fits multiple PLS models
using different numbers of components (from \code{min_ncomp} to
\code{max_ncomp}). The final prediction is a weighted average of
predictions from all models, where the weight for component \mjeqn{j}{j}
is:

\mjdeqn{w_{j} = \frac{1}{s_{1:j} \times g_{j}}}{w_j = 1/(s_{1:j} * g_j)}

where \mjeqn{s_{1:j}}{s_{1:j}} is the root mean square of the spectral
reconstruction error of the target observation(s) when \mjeqn{j}{j} PLS
components are used, and \mjeqn{g_{j}}{g_j} is the root mean square of
the squared regression coefficients for the \mjeqn{j}{j}th component.

The same algorithm variants (\code{'pls'}, \code{'mpls'}, \code{'simpls'})
are available. The default is \code{'mpls'} following the original LOCAL
implementation.
}

\subsection{Gaussian process regression (\code{fit_gpr})}{
Gaussian process regression is a non-parametric Bayesian method
characterised by a mean and covariance function. This implementation uses
a dot product covariance.

The prediction vector \mjeqn{A}{A} is computed from training data
(\mjeqn{X}{X}, \mjeqn{Y}{Y}) as:

\mjdeqn{A = (X X^{T} + \sigma^2 I)^{-1} Y}{A = (X X^T + sigma^2 I)^{-1} Y}

where \mjeqn{\sigma^2}{sigma^2} is the noise variance and \mjeqn{I}{I} is
the identity matrix. Prediction for a new observation
\mjeqn{x_{u}}{x_u} is:

\mjdeqn{\hat{y}_{u} = x_{u} X^{T} A}{hat y_u = x_u X^T A}

The only parameter is the noise variance (\code{noise_variance}).
}
}
\examples{
# PLS with 10 components using standard algorithm
fit_pls(ncomp = 10)

# PLS with modified algorithm (correlation-based weights)
fit_pls(ncomp = 10, method = "mpls")

# PLS with SIMPLS (faster, no iteration)
fit_pls(ncomp = 10, method = "simpls")

# Weighted average PLS (LOCAL-style)
fit_wapls(min_ncomp = 3, max_ncomp = 12)

# Weighted average PLS with SIMPLS
fit_wapls(min_ncomp = 3, max_ncomp = 15, method = "simpls")

# Gaussian process regression
fit_gpr()
fit_gpr(noise_variance = 0.01)

}
\references{
de Jong, S. (1993). SIMPLS: An alternative approach to partial least
squares regression. Chemometrics and Intelligent Laboratory Systems,
18(3), 251-263.

Rasmussen, C.E., Williams, C.K. (2006). Gaussian Processes for Machine
Learning. MIT Press.

Shenk, J.S., & Westerhaus, M.O. (1991). Populations structuring of near
infrared spectra and modified partial least squares regression. Crop
Science, 31(6), 1548-1555.

Shenk, J., Westerhaus, M., & Berzaghi, P. (1997). Investigation of a LOCAL
calibration procedure for near infrared instruments. Journal of Near
Infrared Spectroscopy, 5, 223-232.

Westerhaus, M. (2014). Eastern Analytical Symposium Award for outstanding
achievements in near infrared spectroscopy: my contributions to near
infrared spectroscopy. NIR news, 25(8), 16-20.
}
\seealso{
\code{\link{mbl}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liblex.R
\name{liblex}
\alias{liblex}
\alias{predict.liblex}
\alias{plot.liblex}
\title{Build a precomputed library of localised experts using memory-based learning}
\usage{
liblex(Xr, Yr, neighbors,
       diss_method = diss_pca(ncomp = ncomp_by_opc()),
       fit_method = fit_wapls(min_ncomp = 3, max_ncomp = 15),
       anchor_indices = NULL, gh = TRUE, group = NULL,
       control = liblex_control(), verbose = TRUE, ...)

\method{predict}{liblex}(object, newdata, diss_method = NULL,
        weighting = c("gaussian", "tricube", "triweight", "triangular",
                      "quartic", "parabolic", "cauchy", "none"),
        adaptive_bandwidth = TRUE, reliability_weighting = TRUE,
        range_prediction_limits = FALSE, residual_cutoff = NULL,
        enforce_indices = NULL, probs = c(0.05, 0.25, 0.5, 0.75, 0.95),
        verbose = TRUE, allow_parallel = TRUE, blas_threads = 1L, ...)
        
\method{plot}{liblex}(x, ...)
}
\arguments{
\item{Xr}{A numeric matrix of predictor variables with dimensions \verb{n × p}
(observations in rows, variables in columns). Column names are required.}

\item{Yr}{A numeric vector or single-column matrix of length \code{n} containing
the response variable values corresponding to \code{Xr}. Missing values are
allowed (see Details).}

\item{neighbors}{A neighbor selection object specifying how to select
neighbors. Use \code{\link{neighbors_k}()} for fixed k-nearest neighbors
or \code{\link{neighbors_diss}()} for dissimilarity threshold-based
selection.}

\item{diss_method}{For \code{liblex}: either a \verb{diss_*} object specifying
the dissimilarity method, or a precomputed numeric dissimilarity matrix.
If a \verb{diss_*} object (e.g., \code{\link[=diss_pca]{diss_pca()}}, \code{\link[=diss_pls]{diss_pls()}}, \code{\link[=diss_correlation]{diss_correlation()}},
\code{\link[=diss_euclidean]{diss_euclidean()}}), dissimilarities are computed internally. Additional
parameters (centering, scaling, number of components) are controlled
within the object.
If a matrix is provided:
\itemize{
\item When \code{anchor_indices = NULL}: must be square (\verb{n × n}) with zeros
on the diagonal.
\item When \code{anchor_indices} is specified (length \code{m}): must have
dimensions \verb{n × m}, with \code{diss_method[anchor_indices, ]} having zeros
on the diagonal.
}
Default is \code{diss_pca(ncomp = ncomp_by_opc())}.

For \code{predict.liblex}: a dissimilarity method object created by
\code{diss_*()} constructors. If not provided, uses the method stored in
\code{object}. Required if \code{object} was built with a precomputed
dissimilarity matrix.}

\item{fit_method}{A \code{local_fit} object specifying the local regression
method. Currently supported: \code{\link[=fit_pls]{fit_pls()}} and \code{\link[=fit_wapls]{fit_wapls()}}.
\code{\link[=fit_gpr]{fit_gpr()}} is not yet supported. Default is
\code{fit_wapls(min_ncomp = 3, max_ncomp = 15)}.}

\item{anchor_indices}{An optional integer vector specifying row indices
of \code{Xr} around which to build local models. If \code{NULL} (default), models
are built for all observations. See Details.}

\item{gh}{Logical indicating whether to compute the GH distance
(Mahalanobis distance in PLS score space) for each anchor observation.
Default is \code{TRUE}.}

\item{group}{An optional factor assigning group labels to observations in
\code{Xr}. Used for leave-group-out validation to avoid pseudo-replication.
When an observation is selected for validation, all observations from
the same group are excluded from model fitting. Default is \code{NULL}
(each observation is its own group).}

\item{control}{A list of control parameters created by \code{\link[=liblex_control]{liblex_control()}}.
Default is \code{liblex_control()}.}

\item{verbose}{Logical indicating whether to display progress messages.
Default is \code{TRUE}.}

\item{...}{Additional arguments (currently unused).}

\item{object}{A fitted object of class \code{"liblex"}, created by
\code{\link{liblex}} (for \code{predict}).}

\item{newdata}{A numeric matrix or data frame containing new predictor
values. Must include all predictors used in \code{object}.}

\item{weighting}{Character string specifying the kernel weighting function
applied to neighbours when combining predictions. Options are:
\code{"gaussian"} (default), \code{"tricube"}, \code{"triweight"},
\code{"triangular"}, \code{"quartic"}, \code{"parabolic"}, \code{"cauchy"},
or \code{"none"} (equal weights). See Details for kernel definitions.}

\item{adaptive_bandwidth}{Logical indicating whether to use adaptive
bandwidth for kernel weighting. When \code{TRUE} (default), the bandwidth
is set to the maximum dissimilarity within the neighborhood of each observation,
so weights adapt to local density. When \code{FALSE}, a fixed global
bandwidth is used across all predictions, which may result in uniform
weights in sparse regions or overly concentrated weights in dense regions.}

\item{reliability_weighting}{Logical indicating whether to weight expert
predictions by their estimated reliability. When \code{TRUE} (default),
the contribution of each experrt is additionally weighted by the inverse of its
cross-validation residual variance, giving more influence to models that
performed well during fitting. When \code{FALSE}, only dissimilarity-based
kernel weights are used.}

\item{probs}{A numeric vector of probabilities in \eqn{[0, 1]} for computing
weighted quantiles of expert predictions. Default is
\code{c(0.05, 0.25, 0.5, 0.75, 0.95)}.}

\item{range_prediction_limits}{Logical. If \code{TRUE}, predictions falling
outside the 5th–95th percentile range of neighbour response values are
clipped to those limits. Default is \code{FALSE}.}

\item{residual_cutoff}{Numeric threshold for excluding models. Models with
absolute residuals exceeding this value are penalized during neighbour
selection. Default is \code{NULL} (no exclusion).}

\item{enforce_indices}{Optional integer vector specifying model indices that
must always be included in each prediction neighborhood. These models are
assigned the minimum dissimilarity of the neighborhood to ensure selection.
Default is \code{NULL} (no enforced models).}

\item{allow_parallel}{Logical indicating whether parallel computation is
permitted if a backend is registered. Default is \code{TRUE}.}

\item{blas_threads}{Integer specifying the number of BLAS threads to use.
Default is \code{1L}. Requires the \pkg{RhpcBLASctl} package for thread
control.}

\item{x}{An object of class \code{"liblex"} as returned by \code{\link[=liblex]{liblex()}}.}
}
\value{
\strong{For \code{liblex}:} A list of class \code{"liblex"} (when
\code{control$mode = "build"}) or \code{"liblex_validation"} (when
\code{control$mode = "validate"}) containing:
\itemize{
\item \code{dissimilarity}: List containing the dissimilarity method
and matrix.
\item \code{fit_method}: Fit constructor from \code{fit_method}.
\item \code{gh}: If \code{gh = TRUE}, a list with GH distances and the PLS
projection.
\item \code{results}: Data frame of validation statistics for each
parameter combination (if validation was performed).
\item \code{best}: The optimal parameter combination based on
\code{control$metric}.
\item \code{optimal_params}: List with optimal \code{k} and \code{ncomp} values.
\item \code{residuals}: Residuals from predictions using optimal
parameters.
\item \code{coefficients}: (Build mode only) List of regression
coefficients: \code{B0} (intercepts), \code{B} (slopes).
\item \code{vips}: (Build mode only) Variable importance in projection
scores.
\item \code{selectivity_ratios}: (Build mode only) Selectivity ratios
for each predictor.
\item \code{scaling}: (Build mode only) Centering and scaling vectors
for prediction.
\item \code{neighborhood_stats}: Statistics (response quantiles) for
each neighborhood size.
\item \code{anchor_indices}: The anchor indices used.
\item \code{neighbors}: The object passed to \code{neighbors}.
}

\strong{For \code{predict.liblex}:} A list with the following components:
\itemize{
\item \code{predictions}: A data frame containing:
\itemize{
\item \code{pred}: Weighted mean predictions.
\item \code{pred_sd}: Weighted standard deviation of expert
predictions.
\item \code{q*}: Weighted quantiles at probabilities specified by
\code{probs}.
\item \code{gh}: Global Mahalanobis distance (if computed during
fitting).
\item \code{min_yr}: Minimum response value (5th percentile) across
neighbours.
\item \code{max_yr}: Maximum response value (95th percentile) across
neighbours.
\item \code{below_min}: Logical indicating prediction below
\code{min_yr}.
\item \code{above_max}: Logical indicating prediction above
\code{max_yr}.
}
\item \code{neighbors}: A list with:
\itemize{
\item \code{indices}: Matrix of neighbour indices (models) for each
observation.
\item \code{dissimilarities}: Matrix of corresponding dissimilarity
scores.
}
\item \code{expert_predictions}: A list with:
\itemize{
\item \code{weights}: Matrix of kernel weights applied to each expert.
\item \code{predictions}: Matrix of raw predictions from each expert.
\item \code{weighted}: Matrix of weighted predictions from each expert.
}
}
}
\description{
Constructs a library of local predictive models based on memory-based
learning (MBL). For each anchor observation, a local regression model is
fitted using its nearest neighbors from the reference set. This
implementation is based on the methods proposed in
Ramirez-Lopez et al. (2026b).
}
\details{
By default, local models are constructed for all \code{n} observations in the
reference set. Alternatively, specify a subset of \code{m} observations
(\code{m < n}) via \code{anchor_indices} to reduce computation.

Each local model uses neighbors selected from the full reference set, but
models are only built for anchor observations. This is useful for large
datasets where building models for all observations is computationally
prohibitive.

When dissimilarity methods depend on \code{Yr} (e.g., PLS-based distances), the
response values of anchor observations are excluded during dissimilarity
computation for efficiency. However, anchor response values are always
used when fitting local models.

The number of anchors must not exceed 90\% of \code{nrow(Xr)}; to build models
for all observations, use \code{anchor_indices = NULL}.
\subsection{Relationship between anchors and neighborhood size}{

The \code{neighbors} argument controls the neighborhood size (\code{k}) used both
for fitting local models and for retrieving experts during prediction.
When \code{anchor_indices} is specified, the number of available experts equals
the number of anchors. If \code{max(k)} exceeds the number of anchors and
tuning selects a large optimal \code{k}, prediction will retrieve fewer experts
than specified. For reliable predictions, ensure the number of anchors is
at least as large as the maximum \code{k} value being evaluated.
}

\subsection{Missing values in Yr}{

Missing values in \code{Yr} are permitted. Observations with missing response
values can still serve as neighbors but are excluded from model fitting
as target observations.
}

\subsection{GH distance}{

The GH distance is computed independently from \code{diss_method} using a PLS
projection with optimized component selection. This provides a measure of
how far each observation lies from the center of the reference set in the
PLS score space.
}

\subsection{Validation and tuning}{

When \code{control$mode = "validate"} or \code{control$tune = TRUE}, nearest-neighbor
cross-validation is performed. For each anchor observation, its nearest
neighbor is excluded, a model is fitted on remaining neighbors, and the
excluded neighbor's response is predicted. This provides validation
statistics for parameter selection.
}

\subsection{Prediction}{

For each observation in \code{newdata}, the \code{predict} method:
\enumerate{
\item Computes dissimilarities to anchor observations (or their
neighbourhood centres) stored in \code{object}.
\item Selects the \code{k} nearest neighbours based on the optimal
\code{k} determined during model fitting.
\item Applies kernel weighting based on dissimilarity.
\item Combines expert predictions using weighted averaging.
}
}

\subsection{Kernel weighting functions}{

The weighting functions follow Cleveland and Devlin (1988). Let \eqn{d} be
the normalised dissimilarity (scaled to \eqn{[0, 1]} within the neighbourhood
when \code{adaptive_bandwidth = TRUE}). The available kernels are:
\itemize{
\item \code{"gaussian"}: \eqn{w = \exp(-d^2)}
\item \code{"tricube"}: \eqn{w = (1 - d^3)^3}
\item \code{"triweight"}: \eqn{w = (1 - d^2)^3}
\item \code{"triangular"}: \eqn{w = 1 - d}
\item \code{"quartic"}: \eqn{w = (1 - d^2)^2}
\item \code{"parabolic"}: \eqn{w = 1 - d^2}
\item \code{"cauchy"}: \eqn{w = 1 / (1 + d^2)}
\item \code{"none"}: \eqn{w = 1} (equal weights)
}
}
}
\examples{
\dontrun{
library(prospectr)
data(NIRsoil)

# Preprocess spectra
NIRsoil$spc_pr <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)

# Missing values in the response are allowed
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1, ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1]
test_x  <- NIRsoil$spc_pr[NIRsoil$train == 0, ]
test_y  <- NIRsoil$Ciso[NIRsoil$train == 0]

# Build library
model_library <- liblex(
  Xr = train_x,
  Yr = train_y,
  neighbors = neighbors_k(c(30, 40)),
  diss_method = diss_correlation(ws = 27, scale = TRUE),
  fit_method = fit_wapls(
    min_ncomp = 4,
    max_ncomp = 17,
    scale = FALSE,
    method = "mpls"
  ),
  control = liblex_control(tune = TRUE)
)

# Visualise neighborhood centroids and samples to predict
matplot(
  as.numeric(colnames(model_library$scaling$local_x_center)),
  t(test_x),
  col = rgb(1, 0, 0, 0.3),
  lty = 1,
  type = "l",
  xlab = "Wavelength (nm)",
  ylab = "First derivative detrended absorbance"
)
matlines(
  as.numeric(colnames(model_library$scaling$local_x_center)),
  t(model_library$scaling$local_x_center),
  col = rgb(0, 0, 1, 0.3),
  lty = 1,
  type = "l"
)
grid(lty = 1)
legend(
  "topright",
  legend = c("Samples to predict", "Neighborhood centroids"),
  col = c(rgb(1, 0, 0, 0.8), rgb(0, 0, 1, 0.8)),
  lty = 1,
  lwd = 2,
  bty = "n"
)

# Predict new observations
y_hat_liblex <- predict(model_library, test_x)

# Predicted versus observed values
lims <- range(y_hat_liblex$predictions$pred, test_y, na.rm = TRUE)
plot(
  y_hat_liblex$predictions$pred,
  test_y,
  pch = 16,
  col = rgb(0, 0, 0, 0.5),
  xlab = "Predicted",
  ylab = "Observed",
  xlim = lims,
  ylim = lims
)
abline(a = 0, b = 1, col = "red")
grid(lty = 1)

## run liblex in parallel (requires a parallel backend, e.g., doParallel)
library(doParallel)
n_cores <- min(2, parallel::detectCores() - 1)
clust <- makeCluster(n_cores)
registerDoParallel(clust)

model_library2 <- liblex(
  Xr = train_x,
  Yr = train_y,
  neighbors = neighbors_k(c(30, 40)),
  fit_method = fit_wapls(min_ncomp = 4, max_ncomp = 17, method = "simpls")
)

y_hat_liblex2 <- predict(model_library2, test_x)
registerDoSEQ()
try(stopCluster(clust))
}
}
\references{
Cleveland, W. S., & Devlin, S. J. (1988). Locally weighted regression:
An approach to regression analysis by local fitting.
\emph{Journal of the American Statistical Association}, 83(403), 596–610.

Naes, T., Isaksson, T., & Kowalski, B. (1990). Locally weighted regression
and scatter correction for near-infrared reflectance data.
\emph{Analytical Chemistry}, 62(7), 664–673.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. (2013). The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex datasets. Geoderma, 195-196,
268-279.

Ramirez-Lopez, L., Metz, M., Lesnoff, M., Orellano, C.,
Perez-Fernandez, E., Plans, M., Breure, T., Behrens, T.,
Viscarra Rossel, R., & Peng, Y. (2026b). Rethinking local spectral
modelling: From per-query refitting to model libraries.
\emph{Analytica Chimica Acta}, under review.

Rajalahti, T., Arneberg, R., Berven, F.S., Myhr, K.M., Ulvik, R.J.,
Kvalheim, O.M. (2009). Biomarker discovery in mass spectral profiles by
means of selectivity ratio plot. Chemometrics and Intelligent Laboratory
Systems, 95(1), 35-48.
}
\seealso{
\code{\link[=liblex_control]{liblex_control()}} for control parameters, \code{\link[=neighbors_k]{neighbors_k()}} for neighborhood
specification, \code{\link[=diss_pca]{diss_pca()}}, \code{\link[=diss_pls]{diss_pls()}}, \code{\link[=diss_correlation]{diss_correlation()}} for
dissimilarity methods, \code{\link[=fit_pls]{fit_pls()}}, \code{\link[=fit_wapls]{fit_wapls()}} for fitting methods.
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/liblex_control.R
\name{liblex_control}
\alias{liblex_control}
\title{Control parameters for liblex}
\usage{
liblex_control(
  return_dissimilarity = FALSE,
  mode = c("build", "validate"),
  tune = FALSE,
  metric = c("rmse", "r2"),
  chunk_size = 1L,
  allow_parallel = TRUE,
  blas_threads = 1L
)
}
\arguments{
\item{return_dissimilarity}{A logical indicating whether the dissimilarity
matrix should be returned in the output. Default is \code{FALSE}. Setting
to \code{TRUE} can be useful for diagnostics but increases memory usage
for large libraries.}

\item{mode}{A character string specifying the operation mode:
\describe{
\item{\code{"build"}}{(default) Builds the library of local models. If
\code{tune = TRUE}, validation is performed first to find optimal
parameters, then the library is built using those parameters. If
\code{tune = FALSE}, the library is built directly using the
parameters provided to \code{\link{liblex}}.}
\item{\code{"validate"}}{Performs validation only without building the
library. Useful for parameter exploration or when testing different
configurations before committing to the full library build.}
}}

\item{tune}{A logical indicating whether to optimize parameters via
nearest-neighbor validation. Default is \code{FALSE}. When \code{TRUE},
the function evaluates all combinations of \code{k} values and PLS
component ranges, selecting the combination that minimizes RMSE (or
maximizes R², depending on \code{metric}). See Details.}

\item{metric}{A character string specifying the performance metric used
for parameter selection when \code{tune = TRUE}. Options are:
\describe{
\item{\code{"rmse"}}{(default) Root mean squared error (minimized).}
\item{\code{"r2"}}{Coefficient of determination (maximized).}
}
Ignored when \code{tune = FALSE}.}

\item{chunk_size}{An integer specifying the number of local models to
process per parallel task. Default is \code{1L}. Increasing this value
reduces parallel overhead but may cause load imbalance. For large
libraries, values between 10 and 50 often provide a good trade-off
between overhead and efficiency.}

\item{allow_parallel}{A logical indicating whether parallel execution is
permitted. Default is \code{TRUE}. Parallelization is applied to the
model fitting loop using the \pkg{foreach} package. Requires a registered
parallel backend (e.g., via \pkg{doParallel}).}

\item{blas_threads}{An integer specifying the number of threads for BLAS
operations. Default is \code{1L}, which avoids thread contention when
using parallel processing. Requires the \pkg{RhpcBLASctl} package to
take effect. On Linux systems with multi-threaded BLAS (e.g., OpenBLAS),
setting this to 1 can substantially improve performance when
\code{allow_parallel = TRUE}.}
}
\value{
A list of class \code{liblex_control} containing the validated
control parameters.
}
\description{
Specifies control parameters for the \code{\link{liblex}} function, including
output options, validation settings, tuning behavior, and parallel execution.
}
\details{
\subsection{Nearest-neighbor validation}{
When \code{tune = TRUE} or \code{mode = "validate"}, the function performs
nearest-neighbor validation (NNv) to assess model performance. For each
observation in the reference set (or anchor set, if specified), the
procedure:
\enumerate{
\item Identifies the k nearest neighbors of the target observation.
\item Excludes the target observation (and any observations in the same
group, if \code{group} is specified) from the neighbor set.
\item Fits a local model using the remaining neighbors.
\item Predicts the response value for the excluded target observation.
\item Computes prediction errors across all observations.
}

This leave-one-out style validation provides an estimate of prediction
performance without requiring a separate test set. When \code{tune = TRUE},
the parameter combination (number of neighbors and PLS component range)
yielding the best performance according to \code{metric} is selected for
building the final library.
}

\subsection{Mode and tune combinations}{
\tabular{lll}{
\code{mode} \tab \code{tune} \tab Behavior \cr
\code{"build"} \tab \code{FALSE} \tab Build library using parameters as
provided \cr
\code{"build"} \tab \code{TRUE} \tab Validate, find optimal parameters,
build library \cr
\code{"validate"} \tab \code{FALSE} \tab Validate only, report
performance statistics \cr
\code{"validate"} \tab \code{TRUE} \tab Validate, report statistics with
optimal parameters identified \cr
}
}

\subsection{Parallel chunk size}{
The \code{chunk_size} parameter controls granularity of parallel work
distribution. When \code{allow_parallel = TRUE} and a parallel backend is
registered:
\itemize{
\item \code{chunk_size = 1}: Each local model is a separate parallel task.
Maximum parallelism but higher scheduling overhead.
\item \code{chunk_size > 1}: Multiple models are processed sequentially
within each parallel task. Reduces overhead and improves memory
locality, but may cause load imbalance if the number of models is not
evenly divisible.
}
When \code{allow_parallel = FALSE}, \code{chunk_size} has no effect.
}
}
\examples{
# Default settings: build library without tuning
liblex_control()

# Tune parameters before building
liblex_control(tune = TRUE)

# Validate only for parameter exploration
liblex_control(mode = "validate", tune = TRUE)

# Include dissimilarity matrix in output
liblex_control(return_dissimilarity = TRUE)

# Larger chunks for reduced parallel overhead
liblex_control(chunk_size = 20L)

# Parallel settings for Linux with OpenBLAS
liblex_control(allow_parallel = TRUE, blas_threads = 1L)

}
\seealso{
\code{\link{liblex}}, \code{\link{mbl_control}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mbl_control.R
\name{mbl_control}
\alias{mbl_control}
\title{Control parameters for memory-based learning}
\usage{
mbl_control(
  return_dissimilarity = FALSE,
  validation_type = "NNv",
  tune_locally = TRUE,
  number = 10,
  p = 0.75,
  range_prediction_limits = TRUE,
  allow_parallel = TRUE,
  blas_threads = 1L
)
}
\arguments{
\item{return_dissimilarity}{Logical indicating whether to return the
dissimilarity matrix between \code{Xr} and \code{Xu}. Default is
\code{FALSE}.}

\item{validation_type}{Character vector specifying validation method(s):
\itemize{
\item \code{"NNv"}: Leave-nearest-neighbor-out cross-validation (default, faster)
\item \code{"local_cv"}: Local leave-group-out cross-validation
\item \code{"none"}: No validation
}
Multiple methods can be specified (e.g., \code{c("NNv", "local_cv")}).
Default is \code{"NNv"}.}

\item{tune_locally}{Logical indicating whether to tune PLS components
locally when \code{validation_type = "local_cv"} and using
\code{\link{fit_pls}} or \code{\link{fit_wapls}}. Default is \code{TRUE}.}

\item{number}{Integer specifying the number of sampling iterations for
\code{"local_cv"} validation. Default is \code{10}.}

\item{p}{Numeric value between 0 and 1 indicating the proportion of
observations retained at each \code{"local_cv"} iteration. Default is
\code{0.75}.}

\item{range_prediction_limits}{Logical indicating whether predictions should
be constrained to the range of response values in each neighborhood.
Default is \code{TRUE}.}

\item{allow_parallel}{Logical indicating whether parallel execution is
allowed via the \pkg{foreach} package. Default is \code{TRUE}.}

\item{blas_threads}{Integer specifying the number of BLAS threads to use
during \code{mbl()} execution. Default is \code{1L}, which avoids thread
overhead from repeated small matrix operations. Requires the
\pkg{RhpcBLASctl} package to take effect. The original thread count is
restored after \code{mbl()} completes. See Details.}
}
\value{
A list of class \code{mbl_control} with the specified control parameters.
}
\description{
\loadmathjax
This function controls various aspects of the memory-based learning process
in the \code{\link{mbl}} function.
}
\details{
\subsection{Validation methods}{

\strong{Leave-nearest-neighbor-out cross-validation (\code{"NNv"}):}
For each target observation, the nearest neighbor is excluded from the
local model, which then predicts that neighbor's value. This is faster
than \code{"local_cv"}. If the nearest neighbor belongs to a group
(specified via the \code{group} argument in \code{\link{mbl}}), all
group members are excluded.

\strong{Local leave-group-out cross-validation (\code{"local_cv"}):}
The neighborhood is partitioned into subsets via stratified random
sampling. Each subset serves as validation data while the remainder
fits the model. This repeats \code{number} times, with \code{p}
controlling the training proportion. The final error is the average
local RMSE.
}

\subsection{BLAS threading}{

On Linux systems with multi-threaded OpenBLAS, the default thread count
can cause significant overhead for algorithms like \code{mbl()} that
perform many small matrix operations. Setting \code{blas_threads = 1}
(the default) eliminates this overhead.

This setting requires the \pkg{RhpcBLASctl} package. If not installed,
the parameter is ignored and a message is displayed. The original
thread count is restored when \code{mbl()} completes.

Windows systems typically use single-threaded BLAS by default, so this
setting has no effect there.
}
}
\examples{
# Default control parameters (NNv validation)
mbl_control()

# Both validation methods
mbl_control(validation_type = c("NNv", "local_cv"))

# No validation
mbl_control(validation_type = "none")

# NNv validation only, no parallel
mbl_control(validation_type = "NNv", allow_parallel = FALSE)

# Allow more BLAS threads (if needed for other computations)
mbl_control(blas_threads = 4)

}
\references{
Ramirez-Lopez, L., Behrens, T., Schmidt, K., Stevens, A., Dematte, J.A.M.,
Scholten, T. 2013a. The spectrum-based learner: A new local approach for
modeling soil vis-NIR spectra of complex data sets. Geoderma 195-196:268-279.

Ramirez-Lopez, L., Behrens, T., Schmidt, K., Viscarra Rossel, R., Dematte,
J.A.M., Scholten, T. 2013b. Distance and similarity-search metrics for use
with soil vis-NIR spectra. Geoderma 199:43-53.
}
\seealso{
\code{\link{mbl}}, \code{\link{neighbors_k}},
\code{\link{neighbors_diss}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez} and
Antoine Stevens
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_projection.R
\name{diss_pca}
\alias{diss_pca}
\title{PCA dissimilarity method constructor}
\usage{
diss_pca(
  ncomp = ncomp_by_var(0.01),
  method = c("pca", "pca_nipals"),
  center = TRUE,
  scale = FALSE,
  return_projection = FALSE
)
}
\arguments{
\item{ncomp}{Component selection method. Can be:
\itemize{
\item A positive integer for a fixed number of components
\item \code{\link{ncomp_fixed}(n)}: explicit fixed selection
\item \code{\link{ncomp_by_var}(min_var)}: retain components explaining
at least \code{min_var} variance each (default: \code{ncomp_by_var(0.01)})
\item \code{\link{ncomp_by_cumvar}(min_cumvar)}: retain components until
cumulative variance reaches \code{min_cumvar}
\item \code{\link{ncomp_by_opc}()}: optimize using side information
(\code{Yr} required in \code{dissimilarity()})
}}

\item{method}{Character. PCA algorithm: \code{"pca"} (default, SVD-based) or
\code{"pca_nipals"} (NIPALS algorithm).}

\item{center}{Logical. Center data before projection? Default \code{TRUE}.}

\item{scale}{Logical. Scale data before projection? Default \code{FALSE}.}

\item{return_projection}{Logical. Return the projection object?
Default \code{FALSE}.}
}
\value{
An object of class \code{c("diss_pca", "diss_method")}.
}
\description{
Creates a configuration object for computing dissimilarity based on
Mahalanobis distance in PCA score space.
}
\examples{
# Fixed number of components
diss_pca(ncomp = 10)
diss_pca(ncomp = ncomp_fixed(10))

# Retain components explaining >= 1\% variance each (default)
diss_pca(ncomp = ncomp_by_var(0.01))

# Retain components until 99\% cumulative variance
diss_pca(ncomp = ncomp_by_cumvar(0.99))

# Optimize using side information (requires Yr)
diss_pca(ncomp = ncomp_by_opc(40))
diss_pca(ncomp = ncomp_by_opc())

# NIPALS algorithm (useful for very large matrices)
diss_pca(ncomp = 10, method = "pca_nipals")

}
\seealso{
Component selection: \code{\link{ncomp_by_var}}, \code{\link{ncomp_by_cumvar}},
\code{\link{ncomp_by_opc}}, \code{\link{ncomp_fixed}}

Other dissimilarity methods: \code{\link{diss_pls}},
\code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
\code{\link{diss_cosine}}, \code{\link{diss_mahalanobis}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_correlation.R
\name{diss_correlation}
\alias{diss_correlation}
\title{Correlation dissimilarity method constructor}
\usage{
diss_correlation(ws = NULL, center = TRUE, scale = FALSE)
}
\arguments{
\item{ws}{Either \code{NULL} (default) or an odd integer greater than 2.
When \code{NULL}, standard Pearson correlation dissimilarity is used.
When an odd integer is provided, a moving (rolling) correlation
dissimilarity is computed using a window of that size.}

\item{center}{Logical. Should the data be mean-centered before computing
dissimilarities? Centering is applied jointly to \code{Xr} and \code{Xu}
(if provided) based on their combined column means. Default is \code{TRUE}.}

\item{scale}{Logical. Should the data be scaled (divided by column standard
deviations) before computing dissimilarities? Scaling is applied jointly
to \code{Xr} and \code{Xu} (if provided). Default is \code{FALSE}.}
}
\value{
An object of class \code{c("diss_correlation", "diss_method")} — a
list holding the validated method parameters. Intended to be passed to
\code{dissimilarity()}, not used directly.
}
\description{
Creates a configuration object that fully specifies a correlation (or moving
correlation) dissimilarity method. Pass the result to \code{dissimilarity()}
to compute the dissimilarity matrix.
}
\details{
The correlation dissimilarity between two observations \eqn{x_i} and
\eqn{x_j} is:

\deqn{d(x_i, x_j) = \frac{1}{2}(1 - \rho(x_i, x_j))}

where \eqn{\rho} is the Pearson correlation coefficient. This is used when
\code{ws = NULL}.

When \code{ws} is specified, the moving correlation dissimilarity is:

\deqn{d(x_i, x_j; ws) = \frac{1}{2\,ws} \sum_{k=1}^{p - ws}
  \bigl(1 - \rho(x_{i,(k:k+ws)},\, x_{j,(k:k+ws)})\bigr)}

where \eqn{ws} is the window size and \eqn{p} is the number of variables.
}
\section{Parallel execution}{

The underlying C++ implementation uses OpenMP for parallel computation.
Thread count is controlled by the \code{OMP_NUM_THREADS} environment
variable. To limit threads (e.g., when calling from within a parallel
backend):
\preformatted{
Sys.setenv(OMP_NUM_THREADS = 1)
}
}

\examples{
# Standard correlation dissimilarity
m <- diss_correlation()

# Moving correlation with window size 41
m <- diss_correlation(ws = 41)

# Without centering
m <- diss_correlation(center = FALSE)

}
\seealso{
\code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
\code{\link{diss_mahalanobis}}, \code{\link{diss_cosine}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gesearch.R
\name{gesearch}
\alias{gesearch}
\alias{gesearch.default}
\alias{gesearch.formula}
\alias{predict.gesearch}
\alias{plot.gesearch}
\title{Evolutionary sample search for context-specific calibrations}
\usage{
\method{gesearch}{default}(Xr, Yr, Xu, Yu = NULL, Yu_lims = NULL,
         k, b, retain = 0.95, target_size = k,
         fit_method = fit_pls(ncomp = 10),
         optimization = "reconstruction",
         group = NULL, control = gesearch_control(),
         intermediate_models = FALSE,
         verbose = TRUE, seed = NULL, pchunks = 1L, ...)

\method{gesearch}{formula}(formula, train, test, k, b, target_size, fit_method,
         ..., na_action = na.pass)

\method{predict}{gesearch}(object, newdata, type = "response",
         what = c("final", "all_generations"), ...)

\method{plot}{gesearch}(x, which = c("weakness", "removed"), ...)
}
\arguments{
\item{Xr}{A numeric matrix of predictor variables for the reference data
(observations in rows, variables in columns).}

\item{Yr}{A numeric vector or single-column matrix of response values
corresponding to \code{Xr}. Only one response variable is supported.}

\item{Xu}{A numeric matrix of predictor variables for target observations
(same structure as \code{Xr}).}

\item{Yu}{An optional numeric vector or single-column matrix of response
values for \code{Xu}. Required when \code{optimization} includes
\code{"response"}. Default is \code{NULL}.}

\item{Yu_lims}{A numeric vector of length 2 specifying expected response
limits for the target population. Used with \code{optimization = "range"}.}

\item{k}{An integer specifying the number of samples in each resampling
subset (gene size).}

\item{b}{An integer specifying the target average number of times each
training sample is evaluated per iteration. Higher values (e.g., >40)
produce more stable results but increase computation time.}

\item{retain}{A numeric value in (0, 1] specifying the proportion of samples
retained per iteration. Default is 0.95. Values >0.9 are recommended for
stability. See \code{\link{gesearch_control}} for retention strategy.}

\item{target_size}{An integer specifying the target number of selected
samples (gene pool size). Must be >= \code{k}. Default is \code{k}.}

\item{fit_method}{A fit method object created with \code{\link{fit_pls}}.
Specifies the regression model and scaling used during the search.
Currently only \code{fit_pls()} is supported.}

\item{optimization}{A character vector specifying optimization criteria:
\itemize{
\item \code{"reconstruction"}: (default) Retains samples based on
spectral reconstruction error of \code{Xu} in PLS space.
\item \code{"response"}: Retains samples based on RMSE of predicting
\code{Yu}. Requires \code{Yu}.
\item \code{"similarity"}: Retains samples based on Mahalanobis distance
between \code{Xu} and training samples in PLS score space.
\item \code{"range"}: Removes samples producing predictions outside
\code{Yu_lims}.
}
Multiple criteria can be combined, e.g.,
\code{c("reconstruction", "similarity")}.}

\item{group}{An optional factor assigning group labels to training
observations. Used for leave-group-out cross-validation to avoid
pseudo-replication.}

\item{control}{A list created with \code{\link{gesearch_control}} containing
additional algorithm parameters.}

\item{intermediate_models}{A logical indicating whether to store models for
each intermediate generation. Default is \code{FALSE}.}

\item{verbose}{A logical indicating whether to print progress information.
Default is \code{TRUE}.}

\item{seed}{An integer for random number generation to ensure
reproducibility. Default is \code{NULL}.}

\item{pchunks}{An integer specifying the chunk size used for memory-efficient
parallel processing. Larger values divide the workload into smaller pieces,
which can help reduce memory pressure. Default is 1L.}

\item{formula}{A \code{\link[stats]{formula}} defining the model.}

\item{train}{A data.frame containing training data with model variables.}

\item{test}{A data.frame containing test data with model variables.}

\item{na_action}{A function for handling missing values in training data.
Default is \code{\link[stats]{na.pass}}.}

\item{object}{A fitted \code{gesearch} object (for \code{predict}).}

\item{newdata}{A matrix or data.frame of new observations. For formula-fitted
models, a data.frame containing all predictor variables is accepted. For
non-formula models, a matrix is required.}

\item{type}{A character string specifying the prediction type. Currently only
\code{"response"} is supported.}

\item{what}{A character string specifying which models to use for prediction:
\code{"final"} (default) for predictions from final models only, or
\code{"all_generations"} for predictions from all intermediate generations
plus the final models.}

\item{x}{A \code{gesearch} object (for \code{plot}).}

\item{which}{Character string specifying what to plot:
\code{"weakness"} (maximum weakness scores per generation) or
\code{"removed"} (cumulative samples removed).}

\item{...}{Additional arguments passed to methods.}
}
\value{
\strong{For \code{gesearch}:} A list of class \code{"gesearch"} containing:
\itemize{
\item \code{x_local}: Matrix of predictors for selected samples.
\item \code{y_local}: Vector of responses for selected samples.
\item \code{indices}: Indices of selected samples from original training set.
\item \code{complete_iter}: Number of completed iterations.
\item \code{iter_weakness}: List with iteration-level weakness statistics.
\item \code{samples}: List of sample indices retained at each iteration.
\item \code{n_removed}: data.frame of samples removed per iteration.
\item \code{control}: Copy of control parameters.
\item \code{fit_method}: Fit constructor from \code{fit_method}.
\item \code{validation_results}: Cross-validation in the training only set
validation on the test set using models built only with the samples found.
\item \code{final_models}: Final PLS model containing coefficients, loadings,
scores, VIP, and selectivity ratios.
\item \code{intermediate_models}: List of models per generation (if
\code{intermediate_models = TRUE}).
\item \code{seed}: RNG seed used.
}

\strong{For \code{predict.gesearch}:}
\itemize{
\item If \code{what = "final"}: a prediction matrix with
\code{nrow(newdata)} rows and one column per PLS component.
\item If \code{what = "all_generations"}: a named list of generations,
where each generation contains a prediction matrix as above.
}
}
\description{
Implements an evolutionary search algorithm that selects a subset from large
reference datasets (e.g., spectral libraries) to build context-specific
calibrations. The algorithm iteratively removes weak or non-informative
samples based on prediction error, spectral reconstruction error, or
dissimilarity criteria. This implementation is based on the methods proposed
in Ramirez-Lopez et al. (2026a).
}
\details{
The \code{gesearch} algorithm requires a large reference dataset (\code{Xr})
where the sample search is conducted, target observations (\code{Xu}), and
three tuning parameters: \code{k}, \code{b}, and \code{retain}.

The target observations (\code{Xu}) should represent the population of
interest. These may be selected via algorithms like Kennard-Stone when
response values are unavailable.

The algorithm iteratively removes weak samples from \code{Xr} based on:
\itemize{
\item Increased RMSE when predicting \code{Yu}
\item Increased PLS reconstruction error on \code{Xu}
\item Increased dissimilarity to \code{Xu} in PLS space
}

A resampling scheme identifies samples that consistently appear in
high-error subsets. These are labeled weak and removed. The process
continues until approximately \code{target_size} samples remain.

The \code{gesearch()} function also returns a final model fitted on the selected
samples, which can be used for prediction. This model is internally validated
by cross-validation using only the selected samples from the training/reference
set. If \code{Yu} is available, a model fitted only on the selected reference samples
is first used to predict the target samples. The final model is then refitted
using both the selected reference samples and the target samples used to guide
the search, provided that response values are available for those target samples.
\subsection{Parameter guidance}{

\itemize{
\item \code{k}: Number of samples per resampling subset. See Lobsey et
al. (2017) for guidance.
\item \code{b}: Resampling intensity. Higher values increase stability
but computational cost.
\item \code{retain}: Proportion retained per iteration. Values >0.9
recommended.
}
}

\subsection{Prediction}{

The \code{predict} method generates predictions from a fitted
\code{gesearch} object. If the model was fitted with a formula,
\code{newdata} is validated and transformed to the appropriate model matrix.

When \code{what = "all_generations"}, the return value is a named list with
one element per generation, where each element contains a prediction
matrix. This option requires \code{intermediate_models = TRUE} during
fitting.
}
}
\examples{
\dontrun{
library(prospectr)
data(NIRsoil)

# Preprocess
sg_det <- savitzkyGolay(
  detrend(NIRsoil$spc, wav = as.numeric(colnames(NIRsoil$spc))),
  m = 1, p = 1, w = 7
)
NIRsoil$spc_pr <- sg_det

# Split data
train_x <- NIRsoil$spc_pr[NIRsoil$train == 1 & !is.na(NIRsoil$Ciso), ]
train_y <- NIRsoil$Ciso[NIRsoil$train == 1 & !is.na(NIRsoil$Ciso)]
test_x <- NIRsoil$spc_pr[NIRsoil$train == 0 & !is.na(NIRsoil$Ciso), ]
test_y <- NIRsoil$Ciso[NIRsoil$train == 0 & !is.na(NIRsoil$Ciso)]

# Basic search with reconstruction and similarity optimizations
gs <- gesearch(
  Xr = train_x, Yr = train_y,
  Xu = test_x, Yu = test_y,
  k = 50, b = 100, retain = 0.97,
  target_size = 200,
  fit_method = fit_pls(ncomp = 15, method = "mpls"),
  optimization = c("reconstruction", "similarity"),
  control = gesearch_control(retain_by = "probability"),
  seed = 42
)

# Predict
preds <- predict(gs, test_x)

# Plot progress
plot(gs)
plot(gs, which = "removed")

# With reconstruction and response optimization (requires Yu)
gs_response <- gesearch(
  Xr = train_x, Yr = train_y,
  Xu = test_x, Yu = test_y,
  k = 50, b = 100, retain = 0.97,
  target_size = 200,
  fit_method = fit_pls(ncomp = 15),
  optimization = c("reconstruction", "response"),
  seed = 42
)

# Parallel processing
library(doParallel)
n_cores <- min(2, parallel::detectCores() - 1)
cl <- makeCluster(n_cores)
registerDoParallel(cl)

gs_parallel <- gesearch(
  Xr = train_x, Yr = train_y,
  Xu = test_x,
  k = 50, b = 100, retain = 0.97,
  target_size = 200,
  fit_method = fit_pls(ncomp = 15),
  pchunks = 3,
  seed = 42
)

stopCluster(cl)
registerDoSEQ()
}

}
\references{
Lobsey, C.R., Viscarra Rossel, R.A., Roudier, P., Hedley, C.B. 2017.
rs-local data-mines information from spectral libraries to improve local
calibrations. European Journal of Soil Science 68:840-852.

Kennard, R.W., Stone, L.A. 1969. Computer aided design of experiments.
Technometrics 11:137-148.

Rajalahti, T., Arneberg, R., Berven, F.S., Myhr, K.M., Ulvik, R.J.,
Kvalheim, O.M. 2009. Biomarker discovery in mass spectral profiles by means
of selectivity ratio plot. Chemometrics and Intelligent Laboratory Systems
95:35-48.

Ramirez-Lopez, L., Viscarra Rossel, R., Behrens, T., Orellano, C.,
Perez-Fernandez, E., Kooijman, L., Wadoux, A. M. J.-C., Breure, T.,
Summerauer, L., Safanelli, J. L., & Plans, M. (2026a). When spectral
libraries are too complex to search: Evolutionary subset selection for
domain-adaptive calibration. \emph{Analytica Chimica Acta}, under review.
}
\seealso{
\code{\link{fit_pls}}, \code{\link{gesearch_control}}, \code{\link{mbl}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez},
Claudio Orellano,
\href{https://orcid.org/0000-0001-5416-4520}{Craig Lobsey},
\href{https://orcid.org/0000-0003-1540-4748}{Raphael Viscarra Rossel}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{model}
\alias{model}
\alias{predict.resemble_model}
\title{Global spectral calibration model}
\usage{
model(Xr, Yr, fit_method, control = model_control(), verbose = TRUE)

\method{predict}{resemble_model}(object, newdata, ncomp = NULL, ...)
}
\arguments{
\item{Xr}{A numeric matrix of predictor variables, with observations in rows
and variables in columns. Typically spectral data.}

\item{Yr}{A numeric matrix with one column containing the response variable
values corresponding to the observations in \code{Xr}.}

\item{fit_method}{An object created by \code{\link{fit_pls}} or
\code{\link{fit_gpr}} specifying the regression method and its parameters,
including centring and scaling options.
\code{fit_wapls()} is not supported for global models because it requires
target observations to compute weights.}

\item{control}{A list created by \code{model_control()} specifying the
cross-validation settings. The default is \code{model_control()}.}

\item{verbose}{Logical indicating whether progress information should be
printed. Default is \code{TRUE}.}

\item{object}{A \code{resemble_model} object returned by \code{\link{model}}.}

\item{newdata}{A numeric matrix of new observations with the same number of
columns as the training data.}

\item{ncomp}{For PLS models, the number of components to use for prediction.
The default is the number used during fitting. Ignored for GPR models. For
prediction, this can be a vector of integers representing the predictions
coming from models with the requested components.}

\item{...}{Additional arguments, currently unused.}
}
\value{
For \code{model()}, an object of class \code{resemble_model} containing:
\itemize{
\item \code{fit_method}: Fitting method object used.
\item \code{control}: Model control object used.
\item \code{model}: Fitted model object.
\item \code{cv_results}: Cross-validation results, if
\code{validation_type = "lgo"}.
\item \code{n_obs}: Number of observations used.
\item \code{n_vars}: Number of predictor variables.
}

For \code{predict.resemble_model()}, a numeric matrix of predictions. For
PLS models, columns correspond to different numbers of components.
}
\description{
\loadmathjax

Fits a global calibration model for spectral data using partial least
squares (PLS) or Gaussian process regression (GPR). Unlike
\code{\link{mbl}}, which builds local models for each prediction,
\code{model()} fits a single global model to the reference data.
}
\details{
\code{model()} provides a straightforward interface for fitting global
calibration models, in contrast to the local modelling approach implemented
in \code{\link{mbl}}. This is useful when:
\itemize{
\item{the relationship between spectra and the response is approximately
linear across the full dataset}
\item{a single portable model is needed for deployment}
\item{computational efficiency is prioritised over predictive performance
in heterogeneous datasets}
}

\subsection{Cross-validation}{
When \code{validation_type = "lgo"}, stratified random sampling is used
to create training-validation splits that preserve the distribution of the
response variable. Cross-validation results include RMSE, R², and
standardised RMSE for each component in PLS models, or overall in GPR
models.
}

\subsection{PLS models}{
When \code{fit_pls()} is used, the function fits a PLS model with the
specified number of components. If cross-validation is enabled, the optimal
number of components is selected based on the minimum RMSE.
}

\subsection{GPR models}{
When \code{fit_gpr()} is used, the function fits a Gaussian process
regression model with a dot-product covariance function. The noise variance
parameter controls regularisation.
}
}
\examples{
\donttest{
library(prospectr)
data(NIRsoil)

# Preprocess spectra
Xr <- savitzkyGolay(NIRsoil$spc, m = 1, p = 2, w = 11)
Yr <- NIRsoil$CEC

# Remove missing values
ok <- !is.na(Yr)
Xr <- Xr[ok, ]
Yr <- as.matrix(Yr[ok])

# Fit a PLS model with 10 components and cross-validation
# Scaling is controlled via fit_pls()
pls_mod <- model(
  Xr = Xr,
  Yr = Yr,
  fit_method = fit_pls(ncomp = 10, scale = FALSE),
  control = model_control(validation_type = "lgo", number = 10)
)

# View cross-validation results
pls_mod$cv_results

# Fit a GPR model (centring/scaling controlled via fit_gpr())
gpr_mod <- model(
  Xr = Xr,
  Yr = Yr,
  fit_method = fit_gpr(noise_variance = 0.001, scale = TRUE),
  control = model_control(validation_type = "lgo")
)
}

}
\references{
de Jong, S. (1993). SIMPLS: An alternative approach to partial least
squares regression. \emph{Chemometrics and Intelligent Laboratory Systems},
18(3), 251--263.

Rasmussen, C. E., & Williams, C. K. I. (2006). \emph{Gaussian Processes for
Machine Learning}. MIT Press.

Shenk, J. S., & Westerhaus, M. O. (1991). Population structuring of near
infrared spectra and modified partial least squares regression.
\emph{Crop Science}, 31(6), 1548--1555.
}
\seealso{
\code{\link{fit_pls}} and \code{\link{fit_gpr}} for specifying the fitting
method;
\code{\link{model_control}} for controlling aspects of the modelling
process;
\code{\link{mbl}} for memory-based (local) learning.
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_methods.R
\name{diss_cosine}
\alias{diss_cosine}
\title{Cosine dissimilarity method constructor}
\usage{
diss_cosine(center = TRUE, scale = FALSE)
}
\arguments{
\item{center}{Logical. Center the data before computing dissimilarities?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{TRUE}.}

\item{scale}{Logical. Scale the data before computing dissimilarities?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{FALSE}.}
}
\value{
An object of class \code{c("diss_cosine", "diss_method")}.
}
\description{
Creates a configuration object for computing cosine dissimilarity
(also known as spectral angle mapper). Pass the result to
\code{dissimilarity()} to compute the dissimilarity matrix.

The cosine dissimilarity between two observations \eqn{x_i} and
\eqn{x_j} is:

\deqn{c(x_i, x_j) = \cos^{-1}
  \frac{\sum_{k=1}^{p} x_{i,k}\, x_{j,k}}
       {\sqrt{\sum_{k=1}^{p} x_{i,k}^{2}}\,
        \sqrt{\sum_{k=1}^{p} x_{j,k}^{2}}}}

where \eqn{p} is the number of variables.
}
\examples{
m <- diss_cosine()
m <- diss_cosine(center = FALSE)
}
\seealso{
\code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
\code{\link{diss_mahalanobis}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sid.R
\name{sid}
\alias{sid}
\title{A function for computing the spectral information divergence between
spectra (sid)}
\usage{
sid(Xr, Xu = NULL,
    mode = "density",
    center = FALSE, scale = FALSE,
    kernel = "gaussian",
    n = if(mode == "density") round(0.5 * ncol(Xr)),
    bw = "nrd0",
    reg = 1e-04,
    ...)
}
\arguments{
\item{Xr}{a matrix containing the spectral (reference) data.}

\item{Xu}{an optional matrix containing the spectral data of a second set of
observations.}

\item{mode}{the method to be used for computing the spectral information
divergence. Options are \code{"density"} (default) for computing the divergence
values on the density distributions of the spectral observations, and
\code{"feature"} for computing the divergence vales on the spectral variables.
See details.}

\item{center}{a logical indicating if the computations must be carried out on
the centred \code{X} and \code{Xu} (if specified) matrices. If
\code{mode = "feature"} centring is not carried out since this option does
not accept negative values which are generated after centring the matrices.
Default is FALSE. See details.}

\item{scale}{a logical indicating if the computations must be carried out on
the variance scaled \code{X} and \code{Xu} (if specified) matrices. Default
is TRUE.}

\item{kernel}{if \code{mode = "density"} a character string indicating the
smoothing kernel to be used. It must be one of \code{"gaussian"} (default),
\code{"rectangular"}, \code{"triangular"}, \code{"epanechnikov"},
\code{"biweight"}, \code{"cosine"} or \code{"optcosine"}. See the
\code{\link[stats]{density}} function of the \code{stats} package.}

\item{n}{if \code{mode = "density"} a numerical value indicating the number
of equally spaced points at which the density is to be estimated. See the
\code{\link[stats]{density}} function of the \code{stats} package for further
details. Default is \code{round(0.5 * ncol(X))}.}

\item{bw}{if \code{mode = "density"} a numerical value indicating the
smoothing kernel bandwidth to be used. Optionally the character string
\code{"nrd0"} can be used, it computes the bandwidth using the \code{bw.nrd0}
function of the \code{stats} package (see \code{bw.nrd0}). See the
\code{\link[stats]{density}} and the \code{bw.nrd0} functions for more
details. By default \code{"nrd0"} is used, in this case the bandwidth is
computed as \code{bw.nrd0(as.vector(X))}, if \code{Xu} is specified the
bandwidth is computed as \code{bw.nrd0(as.vector(rbind(X, Xu)))}.}

\item{reg}{a numerical value larger than 0 which indicates a regularization
parameter. Values (probabilities) below this threshold are replaced by this
value for numerical stability. Default is 1e-4.}

\item{...}{additional arguments to be passed to the
\code{\link[stats]{density}} function of the base package.}
}
\value{
a \code{list} with the following components:
\itemize{
\item{\code{sid}: if only \code{"X"} is specified (i.e. \code{Xu = NULL}),
a square symmetric matrix of SID distances between all the components in
\code{"X"}. If both \code{"X"} and \code{"Xu"} are specified, a matrix
of SID distances between the components in \code{"X"} and the components
in \code{"Xu"}) where the rows represent the objects in \code{"X"} and the
columns represent the objects in \code{"Xu"}}
\item{\code{Xr}: the (centered and/or scaled if specified) spectral
\code{X} matrix}
\item{\code{Xu}: the (centered and/or scaled if specified) spectral
\code{Xu} matrix}
\item{\code{densityDisXr}: if \code{mode = "density"}, the computed
density distributions of \code{Xr}}
\item{\code{densityDisXu}: if \code{mode = "density"}, the computed
density distributions of \code{Xu}}
}
}
\description{
\loadmathjax
\ifelse{html}{\out{<a href='https://www.tidyverse.org/lifecycle/#experimental'><img src='figures/lifecycle-experimental.svg' alt='Experimental lifecycle'></a>}}{\strong{Experimental}}

This function computes the spectral information divergence/dissimilarity between
spectra based on the kullback-leibler divergence algorithm (see details).
}
\details{
This function computes the spectral information divergence (distance)
between spectra.
When \code{mode = "density"}, the function first computes the probability
distribution of each spectrum which result in a matrix of density
distribution estimates. The density distributions of all the observations in
the datasets are compared based on the kullback-leibler divergence algorithm.
When \code{mode = "feature"}, the kullback-leibler divergence between all
the observations is computed directly on the spectral variables.
The spectral information divergence (SID) algorithm (Chang, 2000) uses the
Kullback-Leibler divergence (\mjeqn{KL}{KL}) or relative entropy
(Kullback and Leibler, 1951) to account for the vis-NIR information provided
by each spectrum. The SID between two spectra (\mjeqn{x_{i}}{x_i} and
\mjeqn{x_{j}}{x_j}) is computed as follows:

\mjdeqn{sid(x_{i},x_{j}) = KL(x_{i} \left |\right | x_{j}) + KL(x_{j} \left |\right | x_{i})}{sid(x_i,x_j) = KL(x_i  |\ | x_j) + KL(x_j  |\ | x_i)}

\mjdeqn{sid(x_{i},x_{j}) = \sum_{l=1}^{k} p_l \ log(\frac{p_l}{q_l}) + \sum_{l=1}^{k} q_l \ log(\frac{q_l}{p_l})}{sid(x_i,x_j) = \sum_{l=1}^{k} p_l \ log(p_l/q_l) + \sum_{l=1}^{k} q_l \ log(q_l/p_l)}

where \mjeqn{k}{k} represents the number of variables or spectral features,
\mjeqn{p}{p} and \mjeqn{q}{q} are the probability vectors of \mjeqn{x_{i}}{x_i} and
\mjeqn{x_{i}}{x_j} respectively which are calculated as:

\mjdeqn{p = \frac{x_i}{\sum_{l=1}^{k} x_{i,l}}}{p = x_i/{\sum_{l=1}^{k} x_{i,l}}}

\mjdeqn{q = \frac{x_j}{\sum_{l=1}^{k} x_{j,l}}}{q = x_j1/\sum_{l=1}^{k} x_{j,l}}

From the above equations it can be seen that the original SID algorithm
assumes that all the components in the data matrices are nonnegative.
Therefore centering cannot be applied when \code{mode = "feature"}. If a
data matrix with negative values is provided and \code{mode = "feature"},
the \code{sid} function automatically scales the matrix as follows:

\mjdeqn{X_s = \frac{X-min(X)}{max(X)-min(X)}}{X_s = {X-min(X)}/{max(X)-min(X)}}

or

\mjdeqn{X_{s} = \frac{X-min(X, Xu)}{max(X, Xu)-min(X, Xu)}}{X_{s} = {X-min(X, Xu)}/{max(X, Xu)-min(X, Xu)}}

\mjdeqn{Xu_{s} = \frac{Xu-min(X, Xu)}{max(X, Xu)-min(X, Xu)}}{Xu_{s} = {Xu-min(X, Xu)}/{max(X, Xu)-min(X, Xu)}}

if \code{Xu} is specified. The 0 values are replaced by a regularization
parameter (\code{reg} argument) for numerical stability.
The default of the \code{sid} function is to compute the SID based on the
density distributions of the spectra (\code{mode = "density"}). For each
spectrum in \code{X} the density distribution is computed using the
\code{\link[stats]{density}} function of the \code{stats} package.
The 0 values of the estimated density distributions of the spectra are
replaced by a regularization parameter (\code{"reg"} argument) for numerical
stability. Finally the divergence between the computed spectral histogramas
is computed using the SID algorithm. Note that if \code{mode = "density"},
the \code{sid} function will accept negative values and matrix centering
will be possible.
}
\examples{
\donttest{
library(prospectr)

data(NIRsoil)

Xu <- NIRsoil$spc[!as.logical(NIRsoil$train), ]
Yu <- NIRsoil$CEC[!as.logical(NIRsoil$train)]
Yr <- NIRsoil$CEC[as.logical(NIRsoil$train)]
Xr <- NIRsoil$spc[as.logical(NIRsoil$train), ]

Xu <- Xu[!is.na(Yu), ]
Xr <- Xr[!is.na(Yr), ]

# Example 1
# Compute the SID distance between all the observations in Xr
xr_sid <- sid(Xr)
xr_sid

# Example 2
# Compute the SID distance between the observations in Xr and the observations
# in Xu
xr_xu_sid <- sid(Xr, Xu)
xr_xu_sid
}
}
\references{
Chang, C.I. 2000. An information theoretic-based approach to
spectral variability, similarity and discriminability for hyperspectral
image analysis. IEEE Transactions on Information Theory 46, 1927-1932.
}
\seealso{
\code{\link[stats]{density}}
}
\author{
Leonardo Ramirez-Lopez
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_methods.R
\name{diss_euclidean}
\alias{diss_euclidean}
\title{Euclidean dissimilarity method constructor}
\usage{
diss_euclidean(center = TRUE, scale = FALSE)
}
\arguments{
\item{center}{Logical. Center the data before computing distances?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{TRUE}.}

\item{scale}{Logical. Scale the data before computing distances?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{FALSE}.}
}
\value{
An object of class \code{c("diss_euclidean", "diss_method")}.
}
\description{
Creates a configuration object for computing Euclidean dissimilarity.
Pass the result to \code{dissimilarity()} to compute the dissimilarity
matrix.

The scaled Euclidean dissimilarity between two observations \eqn{x_i} and
\eqn{x_j} is:

\deqn{d(x_i, x_j) = \sqrt{\frac{1}{p} \sum_{k=1}^{p}(x_{i,k} - x_{j,k})^2}}

where \eqn{p} is the number of variables. Results are equivalent to
\code{stats::dist()} but scaled by \eqn{1/p}.
}
\examples{
m <- diss_euclidean()
m <- diss_euclidean(center = FALSE, scale = TRUE)
}
\seealso{
\code{\link{dissimilarity}}, \code{\link{diss_mahalanobis}},
\code{\link{diss_cosine}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_projection.R
\name{diss_pls}
\alias{diss_pls}
\title{PLS dissimilarity method constructor}
\usage{
diss_pls(
  ncomp = ncomp_by_opc(),
  method = c("pls", "mpls"),
  scale = FALSE,
  return_projection = FALSE
)
}
\arguments{
\item{ncomp}{Component selection method. Can be:
\itemize{
\item A positive integer for a fixed number of components
\item \code{\link{ncomp_fixed}(n)}: explicit fixed selection
\item \code{\link{ncomp_by_var}(min_var)}: retain components explaining
at least \code{min_var} variance each
\item \code{\link{ncomp_by_cumvar}(min_cumvar)}: retain components until
cumulative variance reaches \code{min_cumvar}
\item \code{\link{ncomp_by_opc}()}: optimize using side information
(default; recommended for PLS since \code{Yr} is already required)
}}

\item{method}{Character. PLS algorithm: \code{"pls"} (default) or
\code{"mpls"} (modified PLS, Shenk & Westerhaus 1991).}

\item{scale}{Logical. Scale data? Default \code{FALSE}. Note: PLS always
centers internally.}

\item{return_projection}{Logical. Return projection object?
Default \code{FALSE}.}
}
\value{
An object of class \code{c("diss_pls", "diss_method")}.
}
\description{
Creates a configuration object for computing dissimilarity based on
Mahalanobis distance in PLS score space. Requires \code{Yr} in
\code{dissimilarity()}.
}
\examples{
# Default: OPC optimization (recommended)
diss_pls()

# Fixed number of components
diss_pls(ncomp = 15)

# Custom opc settings
diss_pls(ncomp = ncomp_by_opc(max_ncomp = 50))

# Modified PLS
diss_pls(ncomp = 10, method = "mpls")

}
\seealso{
Component selection: \code{\link{ncomp_by_var}}, \code{\link{ncomp_by_cumvar}},
\code{\link{ncomp_by_opc}}, \code{\link{ncomp_fixed}}

Other dissimilarity methods: \code{\link{diss_pca}},
\code{\link{diss_correlation}}, \code{\link{diss_euclidean}},
\code{\link{diss_cosine}}, \code{\link{diss_mahalanobis}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diss_methods.R
\name{diss_mahalanobis}
\alias{diss_mahalanobis}
\title{Mahalanobis dissimilarity method constructor}
\usage{
diss_mahalanobis(center = TRUE, scale = FALSE)
}
\arguments{
\item{center}{Logical. Center the data before computing distances?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{TRUE}.}

\item{scale}{Logical. Scale the data before computing distances?
Applied jointly to \code{Xr} and \code{Xu} if both are provided.
Default \code{FALSE}.}
}
\value{
An object of class \code{c("diss_mahalanobis", "diss_method")}.
}
\description{
Creates a configuration object for computing Mahalanobis dissimilarity.
Pass the result to \code{dissimilarity()} to compute the dissimilarity
matrix.

The Mahalanobis distance is computed by first transforming the data into
Mahalanobis space via a factorization of the inverse covariance matrix
\eqn{M^{-1} = W^{T}W} (using SVD), then applying Euclidean distance in
that transformed space:

\deqn{d(x_i, x_j) = \sqrt{\frac{1}{p}(x_i - x_j)M^{-1}(x_i - x_j)^T}}
}
\section{Important limitations}{

The covariance matrix will be singular — and the distance therefore
uncomputable — when the number of observations is smaller than the number
of variables, or when variables are perfectly collinear. This is common
with raw spectral data; consider using \code{diss_euclidean()} on
PCA scores instead.
}

\examples{
m <- diss_mahalanobis()
m <- diss_mahalanobis(center = TRUE, scale = TRUE)
}
\seealso{
\code{\link{dissimilarity}}, \code{\link{diss_euclidean}},
\code{\link{diss_cosine}}
}
\author{
\href{https://orcid.org/0000-0002-5369-5120}{Leonardo Ramirez-Lopez}
}
